/* Generated by Cython 3.0.10 */

/* BEGIN: Cython Metadata
{
    "distutils": {
        "depends": [
            "src/viur/datastore/simdjson.h"
        ],
        "extra_compile_args": [
            "-std=c++11"
        ],
        "include_dirs": [
            "src/viur/datastore"
        ],
        "language": "c++",
        "name": "viur.datastore.transport",
        "sources": [
            "src/viur/datastore/transport.pyx",
            "src/viur/datastore/simdjson.cpp"
        ]
    },
    "module_name": "viur.datastore.transport"
}
END: Cython Metadata */

#ifndef PY_SSIZE_T_CLEAN
#define PY_SSIZE_T_CLEAN
#endif /* PY_SSIZE_T_CLEAN */
#if defined(CYTHON_LIMITED_API) && 0
  #ifndef Py_LIMITED_API
    #if CYTHON_LIMITED_API+0 > 0x03030000
      #define Py_LIMITED_API CYTHON_LIMITED_API
    #else
      #define Py_LIMITED_API 0x03030000
    #endif
  #endif
#endif

#include "Python.h"
#ifndef Py_PYTHON_H
    #error Python headers needed to compile C extensions, please install development version of Python.
#elif PY_VERSION_HEX < 0x02070000 || (0x03000000 <= PY_VERSION_HEX && PY_VERSION_HEX < 0x03030000)
    #error Cython requires Python 2.7+ or Python 3.3+.
#else
#if defined(CYTHON_LIMITED_API) && CYTHON_LIMITED_API
#define __PYX_EXTRA_ABI_MODULE_NAME "limited"
#else
#define __PYX_EXTRA_ABI_MODULE_NAME ""
#endif
#define CYTHON_ABI "3_0_10" __PYX_EXTRA_ABI_MODULE_NAME
#define __PYX_ABI_MODULE_NAME "_cython_" CYTHON_ABI
#define __PYX_TYPE_MODULE_PREFIX __PYX_ABI_MODULE_NAME "."
#define CYTHON_HEX_VERSION 0x03000AF0
#define CYTHON_FUTURE_DIVISION 1
#include <stddef.h>
#ifndef offsetof
  #define offsetof(type, member) ( (size_t) & ((type*)0) -> member )
#endif
#if !defined(_WIN32) && !defined(WIN32) && !defined(MS_WINDOWS)
  #ifndef __stdcall
    #define __stdcall
  #endif
  #ifndef __cdecl
    #define __cdecl
  #endif
  #ifndef __fastcall
    #define __fastcall
  #endif
#endif
#ifndef DL_IMPORT
  #define DL_IMPORT(t) t
#endif
#ifndef DL_EXPORT
  #define DL_EXPORT(t) t
#endif
#define __PYX_COMMA ,
#ifndef HAVE_LONG_LONG
  #define HAVE_LONG_LONG
#endif
#ifndef PY_LONG_LONG
  #define PY_LONG_LONG LONG_LONG
#endif
#ifndef Py_HUGE_VAL
  #define Py_HUGE_VAL HUGE_VAL
#endif
#define __PYX_LIMITED_VERSION_HEX PY_VERSION_HEX
#if defined(GRAALVM_PYTHON)
  /* For very preliminary testing purposes. Most variables are set the same as PyPy.
     The existence of this section does not imply that anything works or is even tested */
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #define CYTHON_COMPILING_IN_LIMITED_API 0
  #define CYTHON_COMPILING_IN_GRAAL 1
  #define CYTHON_COMPILING_IN_NOGIL 0
  #undef CYTHON_USE_TYPE_SLOTS
  #define CYTHON_USE_TYPE_SLOTS 0
  #undef CYTHON_USE_TYPE_SPECS
  #define CYTHON_USE_TYPE_SPECS 0
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #if PY_VERSION_HEX < 0x03050000
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #undef CYTHON_USE_UNICODE_INTERNALS
  #define CYTHON_USE_UNICODE_INTERNALS 0
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #undef CYTHON_AVOID_BORROWED_REFS
  #define CYTHON_AVOID_BORROWED_REFS 1
  #undef CYTHON_ASSUME_SAFE_MACROS
  #define CYTHON_ASSUME_SAFE_MACROS 0
  #undef CYTHON_UNPACK_METHODS
  #define CYTHON_UNPACK_METHODS 0
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_GIL
  #define CYTHON_FAST_GIL 0
  #undef CYTHON_METH_FASTCALL
  #define CYTHON_METH_FASTCALL 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #ifndef CYTHON_PEP487_INIT_SUBCLASS
    #define CYTHON_PEP487_INIT_SUBCLASS (PY_MAJOR_VERSION >= 3)
  #endif
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 1
  #undef CYTHON_USE_MODULE_STATE
  #define CYTHON_USE_MODULE_STATE 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE 0
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC 0
  #endif
  #undef CYTHON_USE_FREELISTS
  #define CYTHON_USE_FREELISTS 0
#elif defined(PYPY_VERSION)
  #define CYTHON_COMPILING_IN_PYPY 1
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #define CYTHON_COMPILING_IN_LIMITED_API 0
  #define CYTHON_COMPILING_IN_GRAAL 0
  #define CYTHON_COMPILING_IN_NOGIL 0
  #undef CYTHON_USE_TYPE_SLOTS
  #define CYTHON_USE_TYPE_SLOTS 0
  #ifndef CYTHON_USE_TYPE_SPECS
    #define CYTHON_USE_TYPE_SPECS 0
  #endif
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #if PY_VERSION_HEX < 0x03050000
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #undef CYTHON_USE_UNICODE_INTERNALS
  #define CYTHON_USE_UNICODE_INTERNALS 0
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #undef CYTHON_AVOID_BORROWED_REFS
  #define CYTHON_AVOID_BORROWED_REFS 1
  #undef CYTHON_ASSUME_SAFE_MACROS
  #define CYTHON_ASSUME_SAFE_MACROS 0
  #undef CYTHON_UNPACK_METHODS
  #define CYTHON_UNPACK_METHODS 0
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_GIL
  #define CYTHON_FAST_GIL 0
  #undef CYTHON_METH_FASTCALL
  #define CYTHON_METH_FASTCALL 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #ifndef CYTHON_PEP487_INIT_SUBCLASS
    #define CYTHON_PEP487_INIT_SUBCLASS (PY_MAJOR_VERSION >= 3)
  #endif
  #if PY_VERSION_HEX < 0x03090000
    #undef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #elif !defined(CYTHON_PEP489_MULTI_PHASE_INIT)
    #define CYTHON_PEP489_MULTI_PHASE_INIT 1
  #endif
  #undef CYTHON_USE_MODULE_STATE
  #define CYTHON_USE_MODULE_STATE 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE (PY_VERSION_HEX >= 0x030400a1 && PYPY_VERSION_NUM >= 0x07030C00)
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC 0
  #endif
  #undef CYTHON_USE_FREELISTS
  #define CYTHON_USE_FREELISTS 0
#elif defined(CYTHON_LIMITED_API)
  #ifdef Py_LIMITED_API
    #undef __PYX_LIMITED_VERSION_HEX
    #define __PYX_LIMITED_VERSION_HEX Py_LIMITED_API
  #endif
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #define CYTHON_COMPILING_IN_LIMITED_API 1
  #define CYTHON_COMPILING_IN_GRAAL 0
  #define CYTHON_COMPILING_IN_NOGIL 0
  #undef CYTHON_CLINE_IN_TRACEBACK
  #define CYTHON_CLINE_IN_TRACEBACK 0
  #undef CYTHON_USE_TYPE_SLOTS
  #define CYTHON_USE_TYPE_SLOTS 0
  #undef CYTHON_USE_TYPE_SPECS
  #define CYTHON_USE_TYPE_SPECS 1
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #undef CYTHON_USE_ASYNC_SLOTS
  #define CYTHON_USE_ASYNC_SLOTS 0
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #undef CYTHON_USE_UNICODE_INTERNALS
  #define CYTHON_USE_UNICODE_INTERNALS 0
  #ifndef CYTHON_USE_UNICODE_WRITER
    #define CYTHON_USE_UNICODE_WRITER 0
  #endif
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #undef CYTHON_ASSUME_SAFE_MACROS
  #define CYTHON_ASSUME_SAFE_MACROS 0
  #undef CYTHON_UNPACK_METHODS
  #define CYTHON_UNPACK_METHODS 0
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_GIL
  #define CYTHON_FAST_GIL 0
  #undef CYTHON_METH_FASTCALL
  #define CYTHON_METH_FASTCALL 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #ifndef CYTHON_PEP487_INIT_SUBCLASS
    #define CYTHON_PEP487_INIT_SUBCLASS 1
  #endif
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #undef CYTHON_USE_MODULE_STATE
  #define CYTHON_USE_MODULE_STATE 1
  #ifndef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE 0
  #endif
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC 0
  #endif
  #undef CYTHON_USE_FREELISTS
  #define CYTHON_USE_FREELISTS 0
#elif defined(Py_GIL_DISABLED) || defined(Py_NOGIL)
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #define CYTHON_COMPILING_IN_LIMITED_API 0
  #define CYTHON_COMPILING_IN_GRAAL 0
  #define CYTHON_COMPILING_IN_NOGIL 1
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #ifndef CYTHON_USE_TYPE_SPECS
    #define CYTHON_USE_TYPE_SPECS 0
  #endif
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #ifndef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #ifndef CYTHON_USE_PYLONG_INTERNALS
    #define CYTHON_USE_PYLONG_INTERNALS 0
  #endif
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_GIL
  #define CYTHON_FAST_GIL 0
  #ifndef CYTHON_METH_FASTCALL
    #define CYTHON_METH_FASTCALL 1
  #endif
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #ifndef CYTHON_PEP487_INIT_SUBCLASS
    #define CYTHON_PEP487_INIT_SUBCLASS 1
  #endif
  #ifndef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT 1
  #endif
  #ifndef CYTHON_USE_MODULE_STATE
    #define CYTHON_USE_MODULE_STATE 0
  #endif
  #ifndef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE 1
  #endif
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC 1
  #endif
  #ifndef CYTHON_USE_FREELISTS
    #define CYTHON_USE_FREELISTS 0
  #endif
#else
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_CPYTHON 1
  #define CYTHON_COMPILING_IN_LIMITED_API 0
  #define CYTHON_COMPILING_IN_GRAAL 0
  #define CYTHON_COMPILING_IN_NOGIL 0
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #ifndef CYTHON_USE_TYPE_SPECS
    #define CYTHON_USE_TYPE_SPECS 0
  #endif
  #ifndef CYTHON_USE_PYTYPE_LOOKUP
    #define CYTHON_USE_PYTYPE_LOOKUP 1
  #endif
  #if PY_MAJOR_VERSION < 3
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #ifndef CYTHON_USE_PYLONG_INTERNALS
    #define CYTHON_USE_PYLONG_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_PYLIST_INTERNALS
    #define CYTHON_USE_PYLIST_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #if PY_VERSION_HEX < 0x030300F0 || PY_VERSION_HEX >= 0x030B00A2
    #undef CYTHON_USE_UNICODE_WRITER
    #define CYTHON_USE_UNICODE_WRITER 0
  #elif !defined(CYTHON_USE_UNICODE_WRITER)
    #define CYTHON_USE_UNICODE_WRITER 1
  #endif
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #ifndef CYTHON_FAST_THREAD_STATE
    #define CYTHON_FAST_THREAD_STATE 1
  #endif
  #ifndef CYTHON_FAST_GIL
    #define CYTHON_FAST_GIL (PY_MAJOR_VERSION < 3 || PY_VERSION_HEX >= 0x03060000 && PY_VERSION_HEX < 0x030C00A6)
  #endif
  #ifndef CYTHON_METH_FASTCALL
    #define CYTHON_METH_FASTCALL (PY_VERSION_HEX >= 0x030700A1)
  #endif
  #ifndef CYTHON_FAST_PYCALL
    #define CYTHON_FAST_PYCALL 1
  #endif
  #ifndef CYTHON_PEP487_INIT_SUBCLASS
    #define CYTHON_PEP487_INIT_SUBCLASS 1
  #endif
  #if PY_VERSION_HEX < 0x03050000
    #undef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #elif !defined(CYTHON_PEP489_MULTI_PHASE_INIT)
    #define CYTHON_PEP489_MULTI_PHASE_INIT 1
  #endif
  #ifndef CYTHON_USE_MODULE_STATE
    #define CYTHON_USE_MODULE_STATE 0
  #endif
  #if PY_VERSION_HEX < 0x030400a1
    #undef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE 0
  #elif !defined(CYTHON_USE_TP_FINALIZE)
    #define CYTHON_USE_TP_FINALIZE 1
  #endif
  #if PY_VERSION_HEX < 0x030600B1
    #undef CYTHON_USE_DICT_VERSIONS
    #define CYTHON_USE_DICT_VERSIONS 0
  #elif !defined(CYTHON_USE_DICT_VERSIONS)
    #define CYTHON_USE_DICT_VERSIONS  (PY_VERSION_HEX < 0x030C00A5)
  #endif
  #if PY_VERSION_HEX < 0x030700A3
    #undef CYTHON_USE_EXC_INFO_STACK
    #define CYTHON_USE_EXC_INFO_STACK 0
  #elif !defined(CYTHON_USE_EXC_INFO_STACK)
    #define CYTHON_USE_EXC_INFO_STACK 1
  #endif
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC 1
  #endif
  #ifndef CYTHON_USE_FREELISTS
    #define CYTHON_USE_FREELISTS 1
  #endif
#endif
#if !defined(CYTHON_FAST_PYCCALL)
#define CYTHON_FAST_PYCCALL  (CYTHON_FAST_PYCALL && PY_VERSION_HEX >= 0x030600B1)
#endif
#if !defined(CYTHON_VECTORCALL)
#define CYTHON_VECTORCALL  (CYTHON_FAST_PYCCALL && PY_VERSION_HEX >= 0x030800B1)
#endif
#define CYTHON_BACKPORT_VECTORCALL (CYTHON_METH_FASTCALL && PY_VERSION_HEX < 0x030800B1)
#if CYTHON_USE_PYLONG_INTERNALS
  #if PY_MAJOR_VERSION < 3
    #include "longintrepr.h"
  #endif
  #undef SHIFT
  #undef BASE
  #undef MASK
  #ifdef SIZEOF_VOID_P
    enum { __pyx_check_sizeof_voidp = 1 / (int)(SIZEOF_VOID_P == sizeof(void*)) };
  #endif
#endif
#ifndef __has_attribute
  #define __has_attribute(x) 0
#endif
#ifndef __has_cpp_attribute
  #define __has_cpp_attribute(x) 0
#endif
#ifndef CYTHON_RESTRICT
  #if defined(__GNUC__)
    #define CYTHON_RESTRICT __restrict__
  #elif defined(_MSC_VER) && _MSC_VER >= 1400
    #define CYTHON_RESTRICT __restrict
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_RESTRICT restrict
  #else
    #define CYTHON_RESTRICT
  #endif
#endif
#ifndef CYTHON_UNUSED
  #if defined(__cplusplus)
    /* for clang __has_cpp_attribute(maybe_unused) is true even before C++17
     * but leads to warnings with -pedantic, since it is a C++17 feature */
    #if ((defined(_MSVC_LANG) && _MSVC_LANG >= 201703L) || __cplusplus >= 201703L)
      #if __has_cpp_attribute(maybe_unused)
        #define CYTHON_UNUSED [[maybe_unused]]
      #endif
    #endif
  #endif
#endif
#ifndef CYTHON_UNUSED
# if defined(__GNUC__)
#   if !(defined(__cplusplus)) || (__GNUC__ > 3 || (__GNUC__ == 3 && __GNUC_MINOR__ >= 4))
#     define CYTHON_UNUSED __attribute__ ((__unused__))
#   else
#     define CYTHON_UNUSED
#   endif
# elif defined(__ICC) || (defined(__INTEL_COMPILER) && !defined(_MSC_VER))
#   define CYTHON_UNUSED __attribute__ ((__unused__))
# else
#   define CYTHON_UNUSED
# endif
#endif
#ifndef CYTHON_UNUSED_VAR
#  if defined(__cplusplus)
     template<class T> void CYTHON_UNUSED_VAR( const T& ) { }
#  else
#    define CYTHON_UNUSED_VAR(x) (void)(x)
#  endif
#endif
#ifndef CYTHON_MAYBE_UNUSED_VAR
  #define CYTHON_MAYBE_UNUSED_VAR(x) CYTHON_UNUSED_VAR(x)
#endif
#ifndef CYTHON_NCP_UNUSED
# if CYTHON_COMPILING_IN_CPYTHON
#  define CYTHON_NCP_UNUSED
# else
#  define CYTHON_NCP_UNUSED CYTHON_UNUSED
# endif
#endif
#ifndef CYTHON_USE_CPP_STD_MOVE
  #if defined(__cplusplus) && (\
    __cplusplus >= 201103L || (defined(_MSC_VER) && _MSC_VER >= 1600))
    #define CYTHON_USE_CPP_STD_MOVE 1
  #else
    #define CYTHON_USE_CPP_STD_MOVE 0
  #endif
#endif
#define __Pyx_void_to_None(void_result) ((void)(void_result), Py_INCREF(Py_None), Py_None)
#ifdef _MSC_VER
    #ifndef _MSC_STDINT_H_
        #if _MSC_VER < 1300
            typedef unsigned char     uint8_t;
            typedef unsigned short    uint16_t;
            typedef unsigned int      uint32_t;
        #else
            typedef unsigned __int8   uint8_t;
            typedef unsigned __int16  uint16_t;
            typedef unsigned __int32  uint32_t;
        #endif
    #endif
    #if _MSC_VER < 1300
        #ifdef _WIN64
            typedef unsigned long long  __pyx_uintptr_t;
        #else
            typedef unsigned int        __pyx_uintptr_t;
        #endif
    #else
        #ifdef _WIN64
            typedef unsigned __int64    __pyx_uintptr_t;
        #else
            typedef unsigned __int32    __pyx_uintptr_t;
        #endif
    #endif
#else
    #include <stdint.h>
    typedef uintptr_t  __pyx_uintptr_t;
#endif
#ifndef CYTHON_FALLTHROUGH
  #if defined(__cplusplus)
    /* for clang __has_cpp_attribute(fallthrough) is true even before C++17
     * but leads to warnings with -pedantic, since it is a C++17 feature */
    #if ((defined(_MSVC_LANG) && _MSVC_LANG >= 201703L) || __cplusplus >= 201703L)
      #if __has_cpp_attribute(fallthrough)
        #define CYTHON_FALLTHROUGH [[fallthrough]]
      #endif
    #endif
    #ifndef CYTHON_FALLTHROUGH
      #if __has_cpp_attribute(clang::fallthrough)
        #define CYTHON_FALLTHROUGH [[clang::fallthrough]]
      #elif __has_cpp_attribute(gnu::fallthrough)
        #define CYTHON_FALLTHROUGH [[gnu::fallthrough]]
      #endif
    #endif
  #endif
  #ifndef CYTHON_FALLTHROUGH
    #if __has_attribute(fallthrough)
      #define CYTHON_FALLTHROUGH __attribute__((fallthrough))
    #else
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
  #if defined(__clang__) && defined(__apple_build_version__)
    #if __apple_build_version__ < 7000000
      #undef  CYTHON_FALLTHROUGH
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
#endif
#ifdef __cplusplus
  template <typename T>
  struct __PYX_IS_UNSIGNED_IMPL {static const bool value = T(0) < T(-1);};
  #define __PYX_IS_UNSIGNED(type) (__PYX_IS_UNSIGNED_IMPL<type>::value)
#else
  #define __PYX_IS_UNSIGNED(type) (((type)-1) > 0)
#endif
#if CYTHON_COMPILING_IN_PYPY == 1
  #define __PYX_NEED_TP_PRINT_SLOT  (PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x030A0000)
#else
  #define __PYX_NEED_TP_PRINT_SLOT  (PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000)
#endif
#define __PYX_REINTERPRET_FUNCION(func_pointer, other_pointer) ((func_pointer)(void(*)(void))(other_pointer))

#ifndef __cplusplus
  #error "Cython files generated with the C++ option must be compiled with a C++ compiler."
#endif
#ifndef CYTHON_INLINE
  #if defined(__clang__)
    #define CYTHON_INLINE __inline__ __attribute__ ((__unused__))
  #else
    #define CYTHON_INLINE inline
  #endif
#endif
template<typename T>
void __Pyx_call_destructor(T& x) {
    x.~T();
}
template<typename T>
class __Pyx_FakeReference {
  public:
    __Pyx_FakeReference() : ptr(NULL) { }
    __Pyx_FakeReference(const T& ref) : ptr(const_cast<T*>(&ref)) { }
    T *operator->() { return ptr; }
    T *operator&() { return ptr; }
    operator T&() { return *ptr; }
    template<typename U> bool operator ==(const U& other) const { return *ptr == other; }
    template<typename U> bool operator !=(const U& other) const { return *ptr != other; }
    template<typename U> bool operator==(const __Pyx_FakeReference<U>& other) const { return *ptr == *other.ptr; }
    template<typename U> bool operator!=(const __Pyx_FakeReference<U>& other) const { return *ptr != *other.ptr; }
  private:
    T *ptr;
};

#define __PYX_BUILD_PY_SSIZE_T "n"
#define CYTHON_FORMAT_SSIZE_T "z"
#if PY_MAJOR_VERSION < 3
  #define __Pyx_BUILTIN_MODULE_NAME "__builtin__"
  #define __Pyx_DefaultClassType PyClass_Type
  #define __Pyx_PyCode_New(a, p, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a+k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
#else
  #define __Pyx_BUILTIN_MODULE_NAME "builtins"
  #define __Pyx_DefaultClassType PyType_Type
#if CYTHON_COMPILING_IN_LIMITED_API
    static CYTHON_INLINE PyObject* __Pyx_PyCode_New(int a, int p, int k, int l, int s, int f,
                                                    PyObject *code, PyObject *c, PyObject* n, PyObject *v,
                                                    PyObject *fv, PyObject *cell, PyObject* fn,
                                                    PyObject *name, int fline, PyObject *lnos) {
        PyObject *exception_table = NULL;
        PyObject *types_module=NULL, *code_type=NULL, *result=NULL;
        #if __PYX_LIMITED_VERSION_HEX < 0x030B0000
        PyObject *version_info;
        PyObject *py_minor_version = NULL;
        #endif
        long minor_version = 0;
        PyObject *type, *value, *traceback;
        PyErr_Fetch(&type, &value, &traceback);
        #if __PYX_LIMITED_VERSION_HEX >= 0x030B0000
        minor_version = 11;
        #else
        if (!(version_info = PySys_GetObject("version_info"))) goto end;
        if (!(py_minor_version = PySequence_GetItem(version_info, 1))) goto end;
        minor_version = PyLong_AsLong(py_minor_version);
        Py_DECREF(py_minor_version);
        if (minor_version == -1 && PyErr_Occurred()) goto end;
        #endif
        if (!(types_module = PyImport_ImportModule("types"))) goto end;
        if (!(code_type = PyObject_GetAttrString(types_module, "CodeType"))) goto end;
        if (minor_version <= 7) {
            (void)p;
            result = PyObject_CallFunction(code_type, "iiiiiOOOOOOiOO", a, k, l, s, f, code,
                          c, n, v, fn, name, fline, lnos, fv, cell);
        } else if (minor_version <= 10) {
            result = PyObject_CallFunction(code_type, "iiiiiiOOOOOOiOO", a,p, k, l, s, f, code,
                          c, n, v, fn, name, fline, lnos, fv, cell);
        } else {
            if (!(exception_table = PyBytes_FromStringAndSize(NULL, 0))) goto end;
            result = PyObject_CallFunction(code_type, "iiiiiiOOOOOOOiOO", a,p, k, l, s, f, code,
                          c, n, v, fn, name, name, fline, lnos, exception_table, fv, cell);
        }
    end:
        Py_XDECREF(code_type);
        Py_XDECREF(exception_table);
        Py_XDECREF(types_module);
        if (type) {
            PyErr_Restore(type, value, traceback);
        }
        return result;
    }
    #ifndef CO_OPTIMIZED
    #define CO_OPTIMIZED 0x0001
    #endif
    #ifndef CO_NEWLOCALS
    #define CO_NEWLOCALS 0x0002
    #endif
    #ifndef CO_VARARGS
    #define CO_VARARGS 0x0004
    #endif
    #ifndef CO_VARKEYWORDS
    #define CO_VARKEYWORDS 0x0008
    #endif
    #ifndef CO_ASYNC_GENERATOR
    #define CO_ASYNC_GENERATOR 0x0200
    #endif
    #ifndef CO_GENERATOR
    #define CO_GENERATOR 0x0020
    #endif
    #ifndef CO_COROUTINE
    #define CO_COROUTINE 0x0080
    #endif
#elif PY_VERSION_HEX >= 0x030B0000
  static CYTHON_INLINE PyCodeObject* __Pyx_PyCode_New(int a, int p, int k, int l, int s, int f,
                                                    PyObject *code, PyObject *c, PyObject* n, PyObject *v,
                                                    PyObject *fv, PyObject *cell, PyObject* fn,
                                                    PyObject *name, int fline, PyObject *lnos) {
    PyCodeObject *result;
    PyObject *empty_bytes = PyBytes_FromStringAndSize("", 0);
    if (!empty_bytes) return NULL;
    result =
      #if PY_VERSION_HEX >= 0x030C0000
        PyUnstable_Code_NewWithPosOnlyArgs
      #else
        PyCode_NewWithPosOnlyArgs
      #endif
        (a, p, k, l, s, f, code, c, n, v, fv, cell, fn, name, name, fline, lnos, empty_bytes);
    Py_DECREF(empty_bytes);
    return result;
  }
#elif PY_VERSION_HEX >= 0x030800B2 && !CYTHON_COMPILING_IN_PYPY
  #define __Pyx_PyCode_New(a, p, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_NewWithPosOnlyArgs(a, p, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
#else
  #define __Pyx_PyCode_New(a, p, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
#endif
#endif
#if PY_VERSION_HEX >= 0x030900A4 || defined(Py_IS_TYPE)
  #define __Pyx_IS_TYPE(ob, type) Py_IS_TYPE(ob, type)
#else
  #define __Pyx_IS_TYPE(ob, type) (((const PyObject*)ob)->ob_type == (type))
#endif
#if PY_VERSION_HEX >= 0x030A00B1 || defined(Py_Is)
  #define __Pyx_Py_Is(x, y)  Py_Is(x, y)
#else
  #define __Pyx_Py_Is(x, y) ((x) == (y))
#endif
#if PY_VERSION_HEX >= 0x030A00B1 || defined(Py_IsNone)
  #define __Pyx_Py_IsNone(ob) Py_IsNone(ob)
#else
  #define __Pyx_Py_IsNone(ob) __Pyx_Py_Is((ob), Py_None)
#endif
#if PY_VERSION_HEX >= 0x030A00B1 || defined(Py_IsTrue)
  #define __Pyx_Py_IsTrue(ob) Py_IsTrue(ob)
#else
  #define __Pyx_Py_IsTrue(ob) __Pyx_Py_Is((ob), Py_True)
#endif
#if PY_VERSION_HEX >= 0x030A00B1 || defined(Py_IsFalse)
  #define __Pyx_Py_IsFalse(ob) Py_IsFalse(ob)
#else
  #define __Pyx_Py_IsFalse(ob) __Pyx_Py_Is((ob), Py_False)
#endif
#define __Pyx_NoneAsNull(obj)  (__Pyx_Py_IsNone(obj) ? NULL : (obj))
#if PY_VERSION_HEX >= 0x030900F0 && !CYTHON_COMPILING_IN_PYPY
  #define __Pyx_PyObject_GC_IsFinalized(o) PyObject_GC_IsFinalized(o)
#else
  #define __Pyx_PyObject_GC_IsFinalized(o) _PyGC_FINALIZED(o)
#endif
#ifndef CO_COROUTINE
  #define CO_COROUTINE 0x80
#endif
#ifndef CO_ASYNC_GENERATOR
  #define CO_ASYNC_GENERATOR 0x200
#endif
#ifndef Py_TPFLAGS_CHECKTYPES
  #define Py_TPFLAGS_CHECKTYPES 0
#endif
#ifndef Py_TPFLAGS_HAVE_INDEX
  #define Py_TPFLAGS_HAVE_INDEX 0
#endif
#ifndef Py_TPFLAGS_HAVE_NEWBUFFER
  #define Py_TPFLAGS_HAVE_NEWBUFFER 0
#endif
#ifndef Py_TPFLAGS_HAVE_FINALIZE
  #define Py_TPFLAGS_HAVE_FINALIZE 0
#endif
#ifndef Py_TPFLAGS_SEQUENCE
  #define Py_TPFLAGS_SEQUENCE 0
#endif
#ifndef Py_TPFLAGS_MAPPING
  #define Py_TPFLAGS_MAPPING 0
#endif
#ifndef METH_STACKLESS
  #define METH_STACKLESS 0
#endif
#if PY_VERSION_HEX <= 0x030700A3 || !defined(METH_FASTCALL)
  #ifndef METH_FASTCALL
     #define METH_FASTCALL 0x80
  #endif
  typedef PyObject *(*__Pyx_PyCFunctionFast) (PyObject *self, PyObject *const *args, Py_ssize_t nargs);
  typedef PyObject *(*__Pyx_PyCFunctionFastWithKeywords) (PyObject *self, PyObject *const *args,
                                                          Py_ssize_t nargs, PyObject *kwnames);
#else
  #if PY_VERSION_HEX >= 0x030d00A4
  #  define __Pyx_PyCFunctionFast PyCFunctionFast
  #  define __Pyx_PyCFunctionFastWithKeywords PyCFunctionFastWithKeywords
  #else
  #  define __Pyx_PyCFunctionFast _PyCFunctionFast
  #  define __Pyx_PyCFunctionFastWithKeywords _PyCFunctionFastWithKeywords
  #endif
#endif
#if CYTHON_METH_FASTCALL
  #define __Pyx_METH_FASTCALL METH_FASTCALL
  #define __Pyx_PyCFunction_FastCall __Pyx_PyCFunctionFast
  #define __Pyx_PyCFunction_FastCallWithKeywords __Pyx_PyCFunctionFastWithKeywords
#else
  #define __Pyx_METH_FASTCALL METH_VARARGS
  #define __Pyx_PyCFunction_FastCall PyCFunction
  #define __Pyx_PyCFunction_FastCallWithKeywords PyCFunctionWithKeywords
#endif
#if CYTHON_VECTORCALL
  #define __pyx_vectorcallfunc vectorcallfunc
  #define __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET  PY_VECTORCALL_ARGUMENTS_OFFSET
  #define __Pyx_PyVectorcall_NARGS(n)  PyVectorcall_NARGS((size_t)(n))
#elif CYTHON_BACKPORT_VECTORCALL
  typedef PyObject *(*__pyx_vectorcallfunc)(PyObject *callable, PyObject *const *args,
                                            size_t nargsf, PyObject *kwnames);
  #define __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET  ((size_t)1 << (8 * sizeof(size_t) - 1))
  #define __Pyx_PyVectorcall_NARGS(n)  ((Py_ssize_t)(((size_t)(n)) & ~__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET))
#else
  #define __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET  0
  #define __Pyx_PyVectorcall_NARGS(n)  ((Py_ssize_t)(n))
#endif
#if PY_MAJOR_VERSION >= 0x030900B1
#define __Pyx_PyCFunction_CheckExact(func)  PyCFunction_CheckExact(func)
#else
#define __Pyx_PyCFunction_CheckExact(func)  PyCFunction_Check(func)
#endif
#define __Pyx_CyOrPyCFunction_Check(func)  PyCFunction_Check(func)
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_CyOrPyCFunction_GET_FUNCTION(func)  (((PyCFunctionObject*)(func))->m_ml->ml_meth)
#elif !CYTHON_COMPILING_IN_LIMITED_API
#define __Pyx_CyOrPyCFunction_GET_FUNCTION(func)  PyCFunction_GET_FUNCTION(func)
#endif
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_CyOrPyCFunction_GET_FLAGS(func)  (((PyCFunctionObject*)(func))->m_ml->ml_flags)
static CYTHON_INLINE PyObject* __Pyx_CyOrPyCFunction_GET_SELF(PyObject *func) {
    return (__Pyx_CyOrPyCFunction_GET_FLAGS(func) & METH_STATIC) ? NULL : ((PyCFunctionObject*)func)->m_self;
}
#endif
static CYTHON_INLINE int __Pyx__IsSameCFunction(PyObject *func, void *cfunc) {
#if CYTHON_COMPILING_IN_LIMITED_API
    return PyCFunction_Check(func) && PyCFunction_GetFunction(func) == (PyCFunction) cfunc;
#else
    return PyCFunction_Check(func) && PyCFunction_GET_FUNCTION(func) == (PyCFunction) cfunc;
#endif
}
#define __Pyx_IsSameCFunction(func, cfunc)   __Pyx__IsSameCFunction(func, cfunc)
#if __PYX_LIMITED_VERSION_HEX < 0x030900B1
  #define __Pyx_PyType_FromModuleAndSpec(m, s, b)  ((void)m, PyType_FromSpecWithBases(s, b))
  typedef PyObject *(*__Pyx_PyCMethod)(PyObject *, PyTypeObject *, PyObject *const *, size_t, PyObject *);
#else
  #define __Pyx_PyType_FromModuleAndSpec(m, s, b)  PyType_FromModuleAndSpec(m, s, b)
  #define __Pyx_PyCMethod  PyCMethod
#endif
#ifndef METH_METHOD
  #define METH_METHOD 0x200
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Malloc)
  #define PyObject_Malloc(s)   PyMem_Malloc(s)
  #define PyObject_Free(p)     PyMem_Free(p)
  #define PyObject_Realloc(p)  PyMem_Realloc(p)
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
  #define __Pyx_PyCode_HasFreeVars(co)  (PyCode_GetNumFree(co) > 0)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno)
#else
  #define __Pyx_PyCode_HasFreeVars(co)  (PyCode_GetNumFree(co) > 0)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno)  (frame)->f_lineno = (lineno)
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
  #define __Pyx_PyThreadState_Current PyThreadState_Get()
#elif !CYTHON_FAST_THREAD_STATE
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#elif PY_VERSION_HEX >= 0x030d00A1
  #define __Pyx_PyThreadState_Current PyThreadState_GetUnchecked()
#elif PY_VERSION_HEX >= 0x03060000
  #define __Pyx_PyThreadState_Current _PyThreadState_UncheckedGet()
#elif PY_VERSION_HEX >= 0x03000000
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#else
  #define __Pyx_PyThreadState_Current _PyThreadState_Current
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
static CYTHON_INLINE void *__Pyx_PyModule_GetState(PyObject *op)
{
    void *result;
    result = PyModule_GetState(op);
    if (!result)
        Py_FatalError("Couldn't find the module state");
    return result;
}
#endif
#define __Pyx_PyObject_GetSlot(obj, name, func_ctype)  __Pyx_PyType_GetSlot(Py_TYPE(obj), name, func_ctype)
#if CYTHON_COMPILING_IN_LIMITED_API
  #define __Pyx_PyType_GetSlot(type, name, func_ctype)  ((func_ctype) PyType_GetSlot((type), Py_##name))
#else
  #define __Pyx_PyType_GetSlot(type, name, func_ctype)  ((type)->name)
#endif
#if PY_VERSION_HEX < 0x030700A2 && !defined(PyThread_tss_create) && !defined(Py_tss_NEEDS_INIT)
#include "pythread.h"
#define Py_tss_NEEDS_INIT 0
typedef int Py_tss_t;
static CYTHON_INLINE int PyThread_tss_create(Py_tss_t *key) {
  *key = PyThread_create_key();
  return 0;
}
static CYTHON_INLINE Py_tss_t * PyThread_tss_alloc(void) {
  Py_tss_t *key = (Py_tss_t *)PyObject_Malloc(sizeof(Py_tss_t));
  *key = Py_tss_NEEDS_INIT;
  return key;
}
static CYTHON_INLINE void PyThread_tss_free(Py_tss_t *key) {
  PyObject_Free(key);
}
static CYTHON_INLINE int PyThread_tss_is_created(Py_tss_t *key) {
  return *key != Py_tss_NEEDS_INIT;
}
static CYTHON_INLINE void PyThread_tss_delete(Py_tss_t *key) {
  PyThread_delete_key(*key);
  *key = Py_tss_NEEDS_INIT;
}
static CYTHON_INLINE int PyThread_tss_set(Py_tss_t *key, void *value) {
  return PyThread_set_key_value(*key, value);
}
static CYTHON_INLINE void * PyThread_tss_get(Py_tss_t *key) {
  return PyThread_get_key_value(*key);
}
#endif
#if PY_MAJOR_VERSION < 3
    #if CYTHON_COMPILING_IN_PYPY
        #if PYPY_VERSION_NUM < 0x07030600
            #if defined(__cplusplus) && __cplusplus >= 201402L
                [[deprecated("`with nogil:` inside a nogil function will not release the GIL in PyPy2 < 7.3.6")]]
            #elif defined(__GNUC__) || defined(__clang__)
                __attribute__ ((__deprecated__("`with nogil:` inside a nogil function will not release the GIL in PyPy2 < 7.3.6")))
            #elif defined(_MSC_VER)
                __declspec(deprecated("`with nogil:` inside a nogil function will not release the GIL in PyPy2 < 7.3.6"))
            #endif
            static CYTHON_INLINE int PyGILState_Check(void) {
                return 0;
            }
        #else  // PYPY_VERSION_NUM < 0x07030600
        #endif  // PYPY_VERSION_NUM < 0x07030600
    #else
        static CYTHON_INLINE int PyGILState_Check(void) {
            PyThreadState * tstate = _PyThreadState_Current;
            return tstate && (tstate == PyGILState_GetThisThreadState());
        }
    #endif
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030d0000 || defined(_PyDict_NewPresized)
#define __Pyx_PyDict_NewPresized(n)  ((n <= 8) ? PyDict_New() : _PyDict_NewPresized(n))
#else
#define __Pyx_PyDict_NewPresized(n)  PyDict_New()
#endif
#if PY_MAJOR_VERSION >= 3 || CYTHON_FUTURE_DIVISION
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_TrueDivide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceTrueDivide(x,y)
#else
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_Divide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceDivide(x,y)
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX > 0x030600B4 && PY_VERSION_HEX < 0x030d0000 && CYTHON_USE_UNICODE_INTERNALS
#define __Pyx_PyDict_GetItemStrWithError(dict, name)  _PyDict_GetItem_KnownHash(dict, name, ((PyASCIIObject *) name)->hash)
static CYTHON_INLINE PyObject * __Pyx_PyDict_GetItemStr(PyObject *dict, PyObject *name) {
    PyObject *res = __Pyx_PyDict_GetItemStrWithError(dict, name);
    if (res == NULL) PyErr_Clear();
    return res;
}
#elif PY_MAJOR_VERSION >= 3 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07020000)
#define __Pyx_PyDict_GetItemStrWithError  PyDict_GetItemWithError
#define __Pyx_PyDict_GetItemStr           PyDict_GetItem
#else
static CYTHON_INLINE PyObject * __Pyx_PyDict_GetItemStrWithError(PyObject *dict, PyObject *name) {
#if CYTHON_COMPILING_IN_PYPY
    return PyDict_GetItem(dict, name);
#else
    PyDictEntry *ep;
    PyDictObject *mp = (PyDictObject*) dict;
    long hash = ((PyStringObject *) name)->ob_shash;
    assert(hash != -1);
    ep = (mp->ma_lookup)(mp, name, hash);
    if (ep == NULL) {
        return NULL;
    }
    return ep->me_value;
#endif
}
#define __Pyx_PyDict_GetItemStr           PyDict_GetItem
#endif
#if CYTHON_USE_TYPE_SLOTS
  #define __Pyx_PyType_GetFlags(tp)   (((PyTypeObject *)tp)->tp_flags)
  #define __Pyx_PyType_HasFeature(type, feature)  ((__Pyx_PyType_GetFlags(type) & (feature)) != 0)
  #define __Pyx_PyObject_GetIterNextFunc(obj)  (Py_TYPE(obj)->tp_iternext)
#else
  #define __Pyx_PyType_GetFlags(tp)   (PyType_GetFlags((PyTypeObject *)tp))
  #define __Pyx_PyType_HasFeature(type, feature)  PyType_HasFeature(type, feature)
  #define __Pyx_PyObject_GetIterNextFunc(obj)  PyIter_Next
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
  #define __Pyx_SetItemOnTypeDict(tp, k, v) PyObject_GenericSetAttr((PyObject*)tp, k, v)
#else
  #define __Pyx_SetItemOnTypeDict(tp, k, v) PyDict_SetItem(tp->tp_dict, k, v)
#endif
#if CYTHON_USE_TYPE_SPECS && PY_VERSION_HEX >= 0x03080000
#define __Pyx_PyHeapTypeObject_GC_Del(obj)  {\
    PyTypeObject *type = Py_TYPE((PyObject*)obj);\
    assert(__Pyx_PyType_HasFeature(type, Py_TPFLAGS_HEAPTYPE));\
    PyObject_GC_Del(obj);\
    Py_DECREF(type);\
}
#else
#define __Pyx_PyHeapTypeObject_GC_Del(obj)  PyObject_GC_Del(obj)
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
  #define CYTHON_PEP393_ENABLED 1
  #define __Pyx_PyUnicode_READY(op)       (0)
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GetLength(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) PyUnicode_ReadChar(u, i)
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   ((void)u, 1114111U)
  #define __Pyx_PyUnicode_KIND(u)         ((void)u, (0))
  #define __Pyx_PyUnicode_DATA(u)         ((void*)u)
  #define __Pyx_PyUnicode_READ(k, d, i)   ((void)k, PyUnicode_ReadChar((PyObject*)(d), i))
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GetLength(u))
#elif PY_VERSION_HEX > 0x03030000 && defined(PyUnicode_KIND)
  #define CYTHON_PEP393_ENABLED 1
  #if PY_VERSION_HEX >= 0x030C0000
    #define __Pyx_PyUnicode_READY(op)       (0)
  #else
    #define __Pyx_PyUnicode_READY(op)       (likely(PyUnicode_IS_READY(op)) ?\
                                                0 : _PyUnicode_Ready((PyObject *)(op)))
  #endif
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_LENGTH(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) PyUnicode_READ_CHAR(u, i)
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   PyUnicode_MAX_CHAR_VALUE(u)
  #define __Pyx_PyUnicode_KIND(u)         ((int)PyUnicode_KIND(u))
  #define __Pyx_PyUnicode_DATA(u)         PyUnicode_DATA(u)
  #define __Pyx_PyUnicode_READ(k, d, i)   PyUnicode_READ(k, d, i)
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  PyUnicode_WRITE(k, d, i, (Py_UCS4) ch)
  #if PY_VERSION_HEX >= 0x030C0000
    #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GET_LENGTH(u))
  #else
    #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03090000
    #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != (likely(PyUnicode_IS_READY(u)) ? PyUnicode_GET_LENGTH(u) : ((PyCompactUnicodeObject *)(u))->wstr_length))
    #else
    #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != (likely(PyUnicode_IS_READY(u)) ? PyUnicode_GET_LENGTH(u) : PyUnicode_GET_SIZE(u)))
    #endif
  #endif
#else
  #define CYTHON_PEP393_ENABLED 0
  #define PyUnicode_1BYTE_KIND  1
  #define PyUnicode_2BYTE_KIND  2
  #define PyUnicode_4BYTE_KIND  4
  #define __Pyx_PyUnicode_READY(op)       (0)
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_SIZE(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) ((Py_UCS4)(PyUnicode_AS_UNICODE(u)[i]))
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   ((sizeof(Py_UNICODE) == 2) ? 65535U : 1114111U)
  #define __Pyx_PyUnicode_KIND(u)         ((int)sizeof(Py_UNICODE))
  #define __Pyx_PyUnicode_DATA(u)         ((void*)PyUnicode_AS_UNICODE(u))
  #define __Pyx_PyUnicode_READ(k, d, i)   ((void)(k), (Py_UCS4)(((Py_UNICODE*)d)[i]))
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  (((void)(k)), ((Py_UNICODE*)d)[i] = (Py_UNICODE) ch)
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GET_SIZE(u))
#endif
#if CYTHON_COMPILING_IN_PYPY
  #define __Pyx_PyUnicode_Concat(a, b)      PyNumber_Add(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  PyNumber_Add(a, b)
#else
  #define __Pyx_PyUnicode_Concat(a, b)      PyUnicode_Concat(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  ((unlikely((a) == Py_None) || unlikely((b) == Py_None)) ?\
      PyNumber_Add(a, b) : __Pyx_PyUnicode_Concat(a, b))
#endif
#if CYTHON_COMPILING_IN_PYPY
  #if !defined(PyUnicode_DecodeUnicodeEscape)
    #define PyUnicode_DecodeUnicodeEscape(s, size, errors)  PyUnicode_Decode(s, size, "unicode_escape", errors)
  #endif
  #if !defined(PyUnicode_Contains) || (PY_MAJOR_VERSION == 2 && PYPY_VERSION_NUM < 0x07030500)
    #undef PyUnicode_Contains
    #define PyUnicode_Contains(u, s)  PySequence_Contains(u, s)
  #endif
  #if !defined(PyByteArray_Check)
    #define PyByteArray_Check(obj)  PyObject_TypeCheck(obj, &PyByteArray_Type)
  #endif
  #if !defined(PyObject_Format)
    #define PyObject_Format(obj, fmt)  PyObject_CallMethod(obj, "__format__", "O", fmt)
  #endif
#endif
#define __Pyx_PyString_FormatSafe(a, b)   ((unlikely((a) == Py_None || (PyString_Check(b) && !PyString_CheckExact(b)))) ? PyNumber_Remainder(a, b) : __Pyx_PyString_Format(a, b))
#define __Pyx_PyUnicode_FormatSafe(a, b)  ((unlikely((a) == Py_None || (PyUnicode_Check(b) && !PyUnicode_CheckExact(b)))) ? PyNumber_Remainder(a, b) : PyUnicode_Format(a, b))
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyString_Format(a, b)  PyUnicode_Format(a, b)
#else
  #define __Pyx_PyString_Format(a, b)  PyString_Format(a, b)
#endif
#if PY_MAJOR_VERSION < 3 && !defined(PyObject_ASCII)
  #define PyObject_ASCII(o)            PyObject_Repr(o)
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBaseString_Type            PyUnicode_Type
  #define PyStringObject               PyUnicodeObject
  #define PyString_Type                PyUnicode_Type
  #define PyString_Check               PyUnicode_Check
  #define PyString_CheckExact          PyUnicode_CheckExact
#ifndef PyObject_Unicode
  #define PyObject_Unicode             PyObject_Str
#endif
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyBaseString_Check(obj) PyUnicode_Check(obj)
  #define __Pyx_PyBaseString_CheckExact(obj) PyUnicode_CheckExact(obj)
#else
  #define __Pyx_PyBaseString_Check(obj) (PyString_Check(obj) || PyUnicode_Check(obj))
  #define __Pyx_PyBaseString_CheckExact(obj) (PyString_CheckExact(obj) || PyUnicode_CheckExact(obj))
#endif
#if CYTHON_COMPILING_IN_CPYTHON
  #define __Pyx_PySequence_ListKeepNew(obj)\
    (likely(PyList_CheckExact(obj) && Py_REFCNT(obj) == 1) ? __Pyx_NewRef(obj) : PySequence_List(obj))
#else
  #define __Pyx_PySequence_ListKeepNew(obj)  PySequence_List(obj)
#endif
#ifndef PySet_CheckExact
  #define PySet_CheckExact(obj)        __Pyx_IS_TYPE(obj, &PySet_Type)
#endif
#if PY_VERSION_HEX >= 0x030900A4
  #define __Pyx_SET_REFCNT(obj, refcnt) Py_SET_REFCNT(obj, refcnt)
  #define __Pyx_SET_SIZE(obj, size) Py_SET_SIZE(obj, size)
#else
  #define __Pyx_SET_REFCNT(obj, refcnt) Py_REFCNT(obj) = (refcnt)
  #define __Pyx_SET_SIZE(obj, size) Py_SIZE(obj) = (size)
#endif
#if CYTHON_ASSUME_SAFE_MACROS
  #define __Pyx_PySequence_ITEM(o, i) PySequence_ITEM(o, i)
  #define __Pyx_PySequence_SIZE(seq)  Py_SIZE(seq)
  #define __Pyx_PyTuple_SET_ITEM(o, i, v) (PyTuple_SET_ITEM(o, i, v), (0))
  #define __Pyx_PyList_SET_ITEM(o, i, v) (PyList_SET_ITEM(o, i, v), (0))
  #define __Pyx_PyTuple_GET_SIZE(o) PyTuple_GET_SIZE(o)
  #define __Pyx_PyList_GET_SIZE(o) PyList_GET_SIZE(o)
  #define __Pyx_PySet_GET_SIZE(o) PySet_GET_SIZE(o)
  #define __Pyx_PyBytes_GET_SIZE(o) PyBytes_GET_SIZE(o)
  #define __Pyx_PyByteArray_GET_SIZE(o) PyByteArray_GET_SIZE(o)
#else
  #define __Pyx_PySequence_ITEM(o, i) PySequence_GetItem(o, i)
  #define __Pyx_PySequence_SIZE(seq)  PySequence_Size(seq)
  #define __Pyx_PyTuple_SET_ITEM(o, i, v) PyTuple_SetItem(o, i, v)
  #define __Pyx_PyList_SET_ITEM(o, i, v) PyList_SetItem(o, i, v)
  #define __Pyx_PyTuple_GET_SIZE(o) PyTuple_Size(o)
  #define __Pyx_PyList_GET_SIZE(o) PyList_Size(o)
  #define __Pyx_PySet_GET_SIZE(o) PySet_Size(o)
  #define __Pyx_PyBytes_GET_SIZE(o) PyBytes_Size(o)
  #define __Pyx_PyByteArray_GET_SIZE(o) PyByteArray_Size(o)
#endif
#if __PYX_LIMITED_VERSION_HEX >= 0x030d00A1
  #define __Pyx_PyImport_AddModuleRef(name) PyImport_AddModuleRef(name)
#else
  static CYTHON_INLINE PyObject *__Pyx_PyImport_AddModuleRef(const char *name) {
      PyObject *module = PyImport_AddModule(name);
      Py_XINCREF(module);
      return module;
  }
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyIntObject                  PyLongObject
  #define PyInt_Type                   PyLong_Type
  #define PyInt_Check(op)              PyLong_Check(op)
  #define PyInt_CheckExact(op)         PyLong_CheckExact(op)
  #define __Pyx_Py3Int_Check(op)       PyLong_Check(op)
  #define __Pyx_Py3Int_CheckExact(op)  PyLong_CheckExact(op)
  #define PyInt_FromString             PyLong_FromString
  #define PyInt_FromUnicode            PyLong_FromUnicode
  #define PyInt_FromLong               PyLong_FromLong
  #define PyInt_FromSize_t             PyLong_FromSize_t
  #define PyInt_FromSsize_t            PyLong_FromSsize_t
  #define PyInt_AsLong                 PyLong_AsLong
  #define PyInt_AS_LONG                PyLong_AS_LONG
  #define PyInt_AsSsize_t              PyLong_AsSsize_t
  #define PyInt_AsUnsignedLongMask     PyLong_AsUnsignedLongMask
  #define PyInt_AsUnsignedLongLongMask PyLong_AsUnsignedLongLongMask
  #define PyNumber_Int                 PyNumber_Long
#else
  #define __Pyx_Py3Int_Check(op)       (PyLong_Check(op) || PyInt_Check(op))
  #define __Pyx_Py3Int_CheckExact(op)  (PyLong_CheckExact(op) || PyInt_CheckExact(op))
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBoolObject                 PyLongObject
#endif
#if PY_MAJOR_VERSION >= 3 && CYTHON_COMPILING_IN_PYPY
  #ifndef PyUnicode_InternFromString
    #define PyUnicode_InternFromString(s) PyUnicode_FromString(s)
  #endif
#endif
#if PY_VERSION_HEX < 0x030200A4
  typedef long Py_hash_t;
  #define __Pyx_PyInt_FromHash_t PyInt_FromLong
  #define __Pyx_PyInt_AsHash_t   __Pyx_PyIndex_AsHash_t
#else
  #define __Pyx_PyInt_FromHash_t PyInt_FromSsize_t
  #define __Pyx_PyInt_AsHash_t   __Pyx_PyIndex_AsSsize_t
#endif
#if CYTHON_USE_ASYNC_SLOTS
  #if PY_VERSION_HEX >= 0x030500B1
    #define __Pyx_PyAsyncMethodsStruct PyAsyncMethods
    #define __Pyx_PyType_AsAsync(obj) (Py_TYPE(obj)->tp_as_async)
  #else
    #define __Pyx_PyType_AsAsync(obj) ((__Pyx_PyAsyncMethodsStruct*) (Py_TYPE(obj)->tp_reserved))
  #endif
#else
  #define __Pyx_PyType_AsAsync(obj) NULL
#endif
#ifndef __Pyx_PyAsyncMethodsStruct
    typedef struct {
        unaryfunc am_await;
        unaryfunc am_aiter;
        unaryfunc am_anext;
    } __Pyx_PyAsyncMethodsStruct;
#endif

#if defined(_WIN32) || defined(WIN32) || defined(MS_WINDOWS)
  #if !defined(_USE_MATH_DEFINES)
    #define _USE_MATH_DEFINES
  #endif
#endif
#include <math.h>
#ifdef NAN
#define __PYX_NAN() ((float) NAN)
#else
static CYTHON_INLINE float __PYX_NAN() {
  float value;
  memset(&value, 0xFF, sizeof(value));
  return value;
}
#endif
#if defined(__CYGWIN__) && defined(_LDBL_EQ_DBL)
#define __Pyx_truncl trunc
#else
#define __Pyx_truncl truncl
#endif

#define __PYX_MARK_ERR_POS(f_index, lineno) \
    { __pyx_filename = __pyx_f[f_index]; (void)__pyx_filename; __pyx_lineno = lineno; (void)__pyx_lineno; __pyx_clineno = __LINE__;  (void)__pyx_clineno; }
#define __PYX_ERR(f_index, lineno, Ln_error) \
    { __PYX_MARK_ERR_POS(f_index, lineno) goto Ln_error; }

#ifdef CYTHON_EXTERN_C
    #undef __PYX_EXTERN_C
    #define __PYX_EXTERN_C CYTHON_EXTERN_C
#elif defined(__PYX_EXTERN_C)
    #ifdef _MSC_VER
    #pragma message ("Please do not define the '__PYX_EXTERN_C' macro externally. Use 'CYTHON_EXTERN_C' instead.")
    #else
    #warning Please do not define the '__PYX_EXTERN_C' macro externally. Use 'CYTHON_EXTERN_C' instead.
    #endif
#else
    #define __PYX_EXTERN_C extern "C++"
#endif

#define __PYX_HAVE__viur__datastore__transport
#define __PYX_HAVE_API__viur__datastore__transport
/* Early includes */
#include <stdint.h>
#include <string.h>
#include <stdio.h>
#include "string_view"
#include "simdjson.h"
#include "ios"
#include "new"
#include "stdexcept"
#include "typeinfo"
#ifdef _OPENMP
#include <omp.h>
#endif /* _OPENMP */

#if defined(PYREX_WITHOUT_ASSERTIONS) && !defined(CYTHON_WITHOUT_ASSERTIONS)
#define CYTHON_WITHOUT_ASSERTIONS
#endif

typedef struct {PyObject **p; const char *s; const Py_ssize_t n; const char* encoding;
                const char is_unicode; const char is_str; const char intern; } __Pyx_StringTabEntry;

#define __PYX_DEFAULT_STRING_ENCODING_IS_ASCII 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_UTF8 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT (PY_MAJOR_VERSION >= 3 && __PYX_DEFAULT_STRING_ENCODING_IS_UTF8)
#define __PYX_DEFAULT_STRING_ENCODING ""
#define __Pyx_PyObject_FromString __Pyx_PyBytes_FromString
#define __Pyx_PyObject_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#define __Pyx_uchar_cast(c) ((unsigned char)c)
#define __Pyx_long_cast(x) ((long)x)
#define __Pyx_fits_Py_ssize_t(v, type, is_signed)  (\
    (sizeof(type) < sizeof(Py_ssize_t))  ||\
    (sizeof(type) > sizeof(Py_ssize_t) &&\
          likely(v < (type)PY_SSIZE_T_MAX ||\
                 v == (type)PY_SSIZE_T_MAX)  &&\
          (!is_signed || likely(v > (type)PY_SSIZE_T_MIN ||\
                                v == (type)PY_SSIZE_T_MIN)))  ||\
    (sizeof(type) == sizeof(Py_ssize_t) &&\
          (is_signed || likely(v < (type)PY_SSIZE_T_MAX ||\
                               v == (type)PY_SSIZE_T_MAX)))  )
static CYTHON_INLINE int __Pyx_is_valid_index(Py_ssize_t i, Py_ssize_t limit) {
    return (size_t) i < (size_t) limit;
}
#if defined (__cplusplus) && __cplusplus >= 201103L
    #include <cstdlib>
    #define __Pyx_sst_abs(value) std::abs(value)
#elif SIZEOF_INT >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) abs(value)
#elif SIZEOF_LONG >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) labs(value)
#elif defined (_MSC_VER)
    #define __Pyx_sst_abs(value) ((Py_ssize_t)_abs64(value))
#elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define __Pyx_sst_abs(value) llabs(value)
#elif defined (__GNUC__)
    #define __Pyx_sst_abs(value) __builtin_llabs(value)
#else
    #define __Pyx_sst_abs(value) ((value<0) ? -value : value)
#endif
static CYTHON_INLINE Py_ssize_t __Pyx_ssize_strlen(const char *s);
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject*);
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject*, Py_ssize_t* length);
static CYTHON_INLINE PyObject* __Pyx_PyByteArray_FromString(const char*);
#define __Pyx_PyByteArray_FromStringAndSize(s, l) PyByteArray_FromStringAndSize((const char*)s, l)
#define __Pyx_PyBytes_FromString        PyBytes_FromString
#define __Pyx_PyBytes_FromStringAndSize PyBytes_FromStringAndSize
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char*);
#if PY_MAJOR_VERSION < 3
    #define __Pyx_PyStr_FromString        __Pyx_PyBytes_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#else
    #define __Pyx_PyStr_FromString        __Pyx_PyUnicode_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyUnicode_FromStringAndSize
#endif
#define __Pyx_PyBytes_AsWritableString(s)     ((char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsWritableSString(s)    ((signed char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsWritableUString(s)    ((unsigned char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsString(s)     ((const char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsSString(s)    ((const signed char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsUString(s)    ((const unsigned char*) PyBytes_AS_STRING(s))
#define __Pyx_PyObject_AsWritableString(s)    ((char*)(__pyx_uintptr_t) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableSString(s)    ((signed char*)(__pyx_uintptr_t) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableUString(s)    ((unsigned char*)(__pyx_uintptr_t) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsSString(s)    ((const signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsUString(s)    ((const unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_FromCString(s)  __Pyx_PyObject_FromString((const char*)s)
#define __Pyx_PyBytes_FromCString(s)   __Pyx_PyBytes_FromString((const char*)s)
#define __Pyx_PyByteArray_FromCString(s)   __Pyx_PyByteArray_FromString((const char*)s)
#define __Pyx_PyStr_FromCString(s)     __Pyx_PyStr_FromString((const char*)s)
#define __Pyx_PyUnicode_FromCString(s) __Pyx_PyUnicode_FromString((const char*)s)
#define __Pyx_PyUnicode_FromOrdinal(o)       PyUnicode_FromOrdinal((int)o)
#define __Pyx_PyUnicode_AsUnicode            PyUnicode_AsUnicode
#define __Pyx_NewRef(obj) (Py_INCREF(obj), obj)
#define __Pyx_Owned_Py_None(b) __Pyx_NewRef(Py_None)
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b);
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject*);
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject*);
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x);
#define __Pyx_PySequence_Tuple(obj)\
    (likely(PyTuple_CheckExact(obj)) ? __Pyx_NewRef(obj) : PySequence_Tuple(obj))
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject*);
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t);
static CYTHON_INLINE Py_hash_t __Pyx_PyIndex_AsHash_t(PyObject*);
#if CYTHON_ASSUME_SAFE_MACROS
#define __pyx_PyFloat_AsDouble(x) (PyFloat_CheckExact(x) ? PyFloat_AS_DOUBLE(x) : PyFloat_AsDouble(x))
#else
#define __pyx_PyFloat_AsDouble(x) PyFloat_AsDouble(x)
#endif
#define __pyx_PyFloat_AsFloat(x) ((float) __pyx_PyFloat_AsDouble(x))
#if PY_MAJOR_VERSION >= 3
#define __Pyx_PyNumber_Int(x) (PyLong_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Long(x))
#else
#define __Pyx_PyNumber_Int(x) (PyInt_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Int(x))
#endif
#if CYTHON_USE_PYLONG_INTERNALS
  #if PY_VERSION_HEX >= 0x030C00A7
  #ifndef _PyLong_SIGN_MASK
    #define _PyLong_SIGN_MASK 3
  #endif
  #ifndef _PyLong_NON_SIZE_BITS
    #define _PyLong_NON_SIZE_BITS 3
  #endif
  #define __Pyx_PyLong_Sign(x)  (((PyLongObject*)x)->long_value.lv_tag & _PyLong_SIGN_MASK)
  #define __Pyx_PyLong_IsNeg(x)  ((__Pyx_PyLong_Sign(x) & 2) != 0)
  #define __Pyx_PyLong_IsNonNeg(x)  (!__Pyx_PyLong_IsNeg(x))
  #define __Pyx_PyLong_IsZero(x)  (__Pyx_PyLong_Sign(x) & 1)
  #define __Pyx_PyLong_IsPos(x)  (__Pyx_PyLong_Sign(x) == 0)
  #define __Pyx_PyLong_CompactValueUnsigned(x)  (__Pyx_PyLong_Digits(x)[0])
  #define __Pyx_PyLong_DigitCount(x)  ((Py_ssize_t) (((PyLongObject*)x)->long_value.lv_tag >> _PyLong_NON_SIZE_BITS))
  #define __Pyx_PyLong_SignedDigitCount(x)\
        ((1 - (Py_ssize_t) __Pyx_PyLong_Sign(x)) * __Pyx_PyLong_DigitCount(x))
  #if defined(PyUnstable_Long_IsCompact) && defined(PyUnstable_Long_CompactValue)
    #define __Pyx_PyLong_IsCompact(x)     PyUnstable_Long_IsCompact((PyLongObject*) x)
    #define __Pyx_PyLong_CompactValue(x)  PyUnstable_Long_CompactValue((PyLongObject*) x)
  #else
    #define __Pyx_PyLong_IsCompact(x)     (((PyLongObject*)x)->long_value.lv_tag < (2 << _PyLong_NON_SIZE_BITS))
    #define __Pyx_PyLong_CompactValue(x)  ((1 - (Py_ssize_t) __Pyx_PyLong_Sign(x)) * (Py_ssize_t) __Pyx_PyLong_Digits(x)[0])
  #endif
  typedef Py_ssize_t  __Pyx_compact_pylong;
  typedef size_t  __Pyx_compact_upylong;
  #else
  #define __Pyx_PyLong_IsNeg(x)  (Py_SIZE(x) < 0)
  #define __Pyx_PyLong_IsNonNeg(x)  (Py_SIZE(x) >= 0)
  #define __Pyx_PyLong_IsZero(x)  (Py_SIZE(x) == 0)
  #define __Pyx_PyLong_IsPos(x)  (Py_SIZE(x) > 0)
  #define __Pyx_PyLong_CompactValueUnsigned(x)  ((Py_SIZE(x) == 0) ? 0 : __Pyx_PyLong_Digits(x)[0])
  #define __Pyx_PyLong_DigitCount(x)  __Pyx_sst_abs(Py_SIZE(x))
  #define __Pyx_PyLong_SignedDigitCount(x)  Py_SIZE(x)
  #define __Pyx_PyLong_IsCompact(x)  (Py_SIZE(x) == 0 || Py_SIZE(x) == 1 || Py_SIZE(x) == -1)
  #define __Pyx_PyLong_CompactValue(x)\
        ((Py_SIZE(x) == 0) ? (sdigit) 0 : ((Py_SIZE(x) < 0) ? -(sdigit)__Pyx_PyLong_Digits(x)[0] : (sdigit)__Pyx_PyLong_Digits(x)[0]))
  typedef sdigit  __Pyx_compact_pylong;
  typedef digit  __Pyx_compact_upylong;
  #endif
  #if PY_VERSION_HEX >= 0x030C00A5
  #define __Pyx_PyLong_Digits(x)  (((PyLongObject*)x)->long_value.ob_digit)
  #else
  #define __Pyx_PyLong_Digits(x)  (((PyLongObject*)x)->ob_digit)
  #endif
#endif
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
#include <string.h>
static int __Pyx_sys_getdefaultencoding_not_ascii;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    PyObject* ascii_chars_u = NULL;
    PyObject* ascii_chars_b = NULL;
    const char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    if (strcmp(default_encoding_c, "ascii") == 0) {
        __Pyx_sys_getdefaultencoding_not_ascii = 0;
    } else {
        char ascii_chars[128];
        int c;
        for (c = 0; c < 128; c++) {
            ascii_chars[c] = (char) c;
        }
        __Pyx_sys_getdefaultencoding_not_ascii = 1;
        ascii_chars_u = PyUnicode_DecodeASCII(ascii_chars, 128, NULL);
        if (!ascii_chars_u) goto bad;
        ascii_chars_b = PyUnicode_AsEncodedString(ascii_chars_u, default_encoding_c, NULL);
        if (!ascii_chars_b || !PyBytes_Check(ascii_chars_b) || memcmp(ascii_chars, PyBytes_AS_STRING(ascii_chars_b), 128) != 0) {
            PyErr_Format(
                PyExc_ValueError,
                "This module compiled with c_string_encoding=ascii, but default encoding '%.200s' is not a superset of ascii.",
                default_encoding_c);
            goto bad;
        }
        Py_DECREF(ascii_chars_u);
        Py_DECREF(ascii_chars_b);
    }
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    Py_XDECREF(ascii_chars_u);
    Py_XDECREF(ascii_chars_b);
    return -1;
}
#endif
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT && PY_MAJOR_VERSION >= 3
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_DecodeUTF8(c_str, size, NULL)
#else
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_Decode(c_str, size, __PYX_DEFAULT_STRING_ENCODING, NULL)
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
#include <string.h>
static char* __PYX_DEFAULT_STRING_ENCODING;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) (const char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    __PYX_DEFAULT_STRING_ENCODING = (char*) malloc(strlen(default_encoding_c) + 1);
    if (!__PYX_DEFAULT_STRING_ENCODING) goto bad;
    strcpy(__PYX_DEFAULT_STRING_ENCODING, default_encoding_c);
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    return -1;
}
#endif
#endif


/* Test for GCC > 2.95 */
#if defined(__GNUC__)     && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))
  #define likely(x)   __builtin_expect(!!(x), 1)
  #define unlikely(x) __builtin_expect(!!(x), 0)
#else /* !__GNUC__ or GCC < 2.95 */
  #define likely(x)   (x)
  #define unlikely(x) (x)
#endif /* __GNUC__ */
static CYTHON_INLINE void __Pyx_pretend_to_initialize(void* ptr) { (void)ptr; }

#if !CYTHON_USE_MODULE_STATE
static PyObject *__pyx_m = NULL;
#endif
static int __pyx_lineno;
static int __pyx_clineno = 0;
static const char * __pyx_cfilenm = __FILE__;
static const char *__pyx_filename;

/* #### Code section: filename_table ### */

static const char *__pyx_f[] = {
  "src/viur/datastore/transport.pyx",
  "type.pxd",
};
/* #### Code section: utility_code_proto_before_types ### */
/* ForceInitThreads.proto */
#ifndef __PYX_FORCE_INIT_THREADS
  #define __PYX_FORCE_INIT_THREADS 0
#endif

/* #### Code section: numeric_typedefs ### */
/* #### Code section: complex_type_declarations ### */
/* #### Code section: type_declarations ### */

/*--- Type declarations ---*/
struct __pyx_obj_4viur_9datastore_9transport___pyx_scope_struct__genexpr;
struct __pyx_opt_args_4viur_9datastore_9transport_toEntityStructure;

/* "viur/datastore/transport.pyx":284
 *     return key
 * 
 * cdef inline object toEntityStructure(simdjsonElement v, boolean_type isInitial = False):             # <<<<<<<<<<<<<<
 *     """
 *         Parses a simdJsonElement into the corresponding python datatypes.
 */
struct __pyx_opt_args_4viur_9datastore_9transport_toEntityStructure {
  int __pyx_n;
  bool isInitial;
};

/* "viur/datastore/transport.pyx":584
 *         keys = [keys]
 *         isMulti = False
 *     if any(key.is_partial or key.kind is None for key in keys):             # <<<<<<<<<<<<<<
 *         raise InvalidArgumentError("Keys must not be partial or kind-less")
 *     accessLog = currentDbAccessLog.get()
 */
struct __pyx_obj_4viur_9datastore_9transport___pyx_scope_struct__genexpr {
  PyObject_HEAD
  PyObject *__pyx_genexpr_arg_0;
  PyObject *__pyx_v_key;
};

/* #### Code section: utility_code_proto ### */

/* --- Runtime support code (head) --- */
/* Refnanny.proto */
#ifndef CYTHON_REFNANNY
  #define CYTHON_REFNANNY 0
#endif
#if CYTHON_REFNANNY
  typedef struct {
    void (*INCREF)(void*, PyObject*, Py_ssize_t);
    void (*DECREF)(void*, PyObject*, Py_ssize_t);
    void (*GOTREF)(void*, PyObject*, Py_ssize_t);
    void (*GIVEREF)(void*, PyObject*, Py_ssize_t);
    void* (*SetupContext)(const char*, Py_ssize_t, const char*);
    void (*FinishContext)(void**);
  } __Pyx_RefNannyAPIStruct;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNanny = NULL;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname);
  #define __Pyx_RefNannyDeclarations void *__pyx_refnanny = NULL;
#ifdef WITH_THREAD
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          if (acquire_gil) {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), (__LINE__), (__FILE__));\
              PyGILState_Release(__pyx_gilstate_save);\
          } else {\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), (__LINE__), (__FILE__));\
          }
  #define __Pyx_RefNannyFinishContextNogil() {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __Pyx_RefNannyFinishContext();\
              PyGILState_Release(__pyx_gilstate_save);\
          }
#else
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), (__LINE__), (__FILE__))
  #define __Pyx_RefNannyFinishContextNogil() __Pyx_RefNannyFinishContext()
#endif
  #define __Pyx_RefNannyFinishContextNogil() {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __Pyx_RefNannyFinishContext();\
              PyGILState_Release(__pyx_gilstate_save);\
          }
  #define __Pyx_RefNannyFinishContext()\
          __Pyx_RefNanny->FinishContext(&__pyx_refnanny)
  #define __Pyx_INCREF(r)  __Pyx_RefNanny->INCREF(__pyx_refnanny, (PyObject *)(r), (__LINE__))
  #define __Pyx_DECREF(r)  __Pyx_RefNanny->DECREF(__pyx_refnanny, (PyObject *)(r), (__LINE__))
  #define __Pyx_GOTREF(r)  __Pyx_RefNanny->GOTREF(__pyx_refnanny, (PyObject *)(r), (__LINE__))
  #define __Pyx_GIVEREF(r) __Pyx_RefNanny->GIVEREF(__pyx_refnanny, (PyObject *)(r), (__LINE__))
  #define __Pyx_XINCREF(r)  do { if((r) == NULL); else {__Pyx_INCREF(r); }} while(0)
  #define __Pyx_XDECREF(r)  do { if((r) == NULL); else {__Pyx_DECREF(r); }} while(0)
  #define __Pyx_XGOTREF(r)  do { if((r) == NULL); else {__Pyx_GOTREF(r); }} while(0)
  #define __Pyx_XGIVEREF(r) do { if((r) == NULL); else {__Pyx_GIVEREF(r);}} while(0)
#else
  #define __Pyx_RefNannyDeclarations
  #define __Pyx_RefNannySetupContext(name, acquire_gil)
  #define __Pyx_RefNannyFinishContextNogil()
  #define __Pyx_RefNannyFinishContext()
  #define __Pyx_INCREF(r) Py_INCREF(r)
  #define __Pyx_DECREF(r) Py_DECREF(r)
  #define __Pyx_GOTREF(r)
  #define __Pyx_GIVEREF(r)
  #define __Pyx_XINCREF(r) Py_XINCREF(r)
  #define __Pyx_XDECREF(r) Py_XDECREF(r)
  #define __Pyx_XGOTREF(r)
  #define __Pyx_XGIVEREF(r)
#endif
#define __Pyx_Py_XDECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; Py_XDECREF(tmp);\
    } while (0)
#define __Pyx_XDECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_XDECREF(tmp);\
    } while (0)
#define __Pyx_DECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_DECREF(tmp);\
    } while (0)
#define __Pyx_CLEAR(r)    do { PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);} while(0)
#define __Pyx_XCLEAR(r)   do { if((r) != NULL) {PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);}} while(0)

/* PyErrExceptionMatches.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_ExceptionMatches(err) __Pyx_PyErr_ExceptionMatchesInState(__pyx_tstate, err)
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err);
#else
#define __Pyx_PyErr_ExceptionMatches(err)  PyErr_ExceptionMatches(err)
#endif

/* PyThreadStateGet.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyThreadState_declare  PyThreadState *__pyx_tstate;
#define __Pyx_PyThreadState_assign  __pyx_tstate = __Pyx_PyThreadState_Current;
#if PY_VERSION_HEX >= 0x030C00A6
#define __Pyx_PyErr_Occurred()  (__pyx_tstate->current_exception != NULL)
#define __Pyx_PyErr_CurrentExceptionType()  (__pyx_tstate->current_exception ? (PyObject*) Py_TYPE(__pyx_tstate->current_exception) : (PyObject*) NULL)
#else
#define __Pyx_PyErr_Occurred()  (__pyx_tstate->curexc_type != NULL)
#define __Pyx_PyErr_CurrentExceptionType()  (__pyx_tstate->curexc_type)
#endif
#else
#define __Pyx_PyThreadState_declare
#define __Pyx_PyThreadState_assign
#define __Pyx_PyErr_Occurred()  (PyErr_Occurred() != NULL)
#define __Pyx_PyErr_CurrentExceptionType()  PyErr_Occurred()
#endif

/* PyErrFetchRestore.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_Clear() __Pyx_ErrRestore(NULL, NULL, NULL)
#define __Pyx_ErrRestoreWithState(type, value, tb)  __Pyx_ErrRestoreInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)    __Pyx_ErrFetchInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  __Pyx_ErrRestoreInState(__pyx_tstate, type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)    __Pyx_ErrFetchInState(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A6
#define __Pyx_PyErr_SetNone(exc) (Py_INCREF(exc), __Pyx_ErrRestore((exc), NULL, NULL))
#else
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#endif
#else
#define __Pyx_PyErr_Clear() PyErr_Clear()
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#define __Pyx_ErrRestoreWithState(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestoreInState(tstate, type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchInState(tstate, type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)  PyErr_Fetch(type, value, tb)
#endif

/* PyObjectGetAttrStr.proto */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GetAttrStr(o,n) PyObject_GetAttr(o,n)
#endif

/* PyObjectGetAttrStrNoError.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStrNoError(PyObject* obj, PyObject* attr_name);

/* GetBuiltinName.proto */
static PyObject *__Pyx_GetBuiltinName(PyObject *name);

/* TupleAndListFromArray.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyList_FromArray(PyObject *const *src, Py_ssize_t n);
static CYTHON_INLINE PyObject* __Pyx_PyTuple_FromArray(PyObject *const *src, Py_ssize_t n);
#endif

/* IncludeStringH.proto */
#include <string.h>

/* BytesEquals.proto */
static CYTHON_INLINE int __Pyx_PyBytes_Equals(PyObject* s1, PyObject* s2, int equals);

/* UnicodeEquals.proto */
static CYTHON_INLINE int __Pyx_PyUnicode_Equals(PyObject* s1, PyObject* s2, int equals);

/* fastcall.proto */
#if CYTHON_AVOID_BORROWED_REFS
    #define __Pyx_Arg_VARARGS(args, i) PySequence_GetItem(args, i)
#elif CYTHON_ASSUME_SAFE_MACROS
    #define __Pyx_Arg_VARARGS(args, i) PyTuple_GET_ITEM(args, i)
#else
    #define __Pyx_Arg_VARARGS(args, i) PyTuple_GetItem(args, i)
#endif
#if CYTHON_AVOID_BORROWED_REFS
    #define __Pyx_Arg_NewRef_VARARGS(arg) __Pyx_NewRef(arg)
    #define __Pyx_Arg_XDECREF_VARARGS(arg) Py_XDECREF(arg)
#else
    #define __Pyx_Arg_NewRef_VARARGS(arg) arg
    #define __Pyx_Arg_XDECREF_VARARGS(arg)
#endif
#define __Pyx_NumKwargs_VARARGS(kwds) PyDict_Size(kwds)
#define __Pyx_KwValues_VARARGS(args, nargs) NULL
#define __Pyx_GetKwValue_VARARGS(kw, kwvalues, s) __Pyx_PyDict_GetItemStrWithError(kw, s)
#define __Pyx_KwargsAsDict_VARARGS(kw, kwvalues) PyDict_Copy(kw)
#if CYTHON_METH_FASTCALL
    #define __Pyx_Arg_FASTCALL(args, i) args[i]
    #define __Pyx_NumKwargs_FASTCALL(kwds) PyTuple_GET_SIZE(kwds)
    #define __Pyx_KwValues_FASTCALL(args, nargs) ((args) + (nargs))
    static CYTHON_INLINE PyObject * __Pyx_GetKwValue_FASTCALL(PyObject *kwnames, PyObject *const *kwvalues, PyObject *s);
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030d0000
    CYTHON_UNUSED static PyObject *__Pyx_KwargsAsDict_FASTCALL(PyObject *kwnames, PyObject *const *kwvalues);
  #else
    #define __Pyx_KwargsAsDict_FASTCALL(kw, kwvalues) _PyStack_AsDict(kwvalues, kw)
  #endif
    #define __Pyx_Arg_NewRef_FASTCALL(arg) arg  /* no-op, __Pyx_Arg_FASTCALL is direct and this needs
                                                   to have the same reference counting */
    #define __Pyx_Arg_XDECREF_FASTCALL(arg)
#else
    #define __Pyx_Arg_FASTCALL __Pyx_Arg_VARARGS
    #define __Pyx_NumKwargs_FASTCALL __Pyx_NumKwargs_VARARGS
    #define __Pyx_KwValues_FASTCALL __Pyx_KwValues_VARARGS
    #define __Pyx_GetKwValue_FASTCALL __Pyx_GetKwValue_VARARGS
    #define __Pyx_KwargsAsDict_FASTCALL __Pyx_KwargsAsDict_VARARGS
    #define __Pyx_Arg_NewRef_FASTCALL(arg) __Pyx_Arg_NewRef_VARARGS(arg)
    #define __Pyx_Arg_XDECREF_FASTCALL(arg) __Pyx_Arg_XDECREF_VARARGS(arg)
#endif
#if CYTHON_COMPILING_IN_CPYTHON && CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
#define __Pyx_ArgsSlice_VARARGS(args, start, stop) __Pyx_PyTuple_FromArray(&__Pyx_Arg_VARARGS(args, start), stop - start)
#define __Pyx_ArgsSlice_FASTCALL(args, start, stop) __Pyx_PyTuple_FromArray(&__Pyx_Arg_FASTCALL(args, start), stop - start)
#else
#define __Pyx_ArgsSlice_VARARGS(args, start, stop) PyTuple_GetSlice(args, start, stop)
#define __Pyx_ArgsSlice_FASTCALL(args, start, stop) PyTuple_GetSlice(args, start, stop)
#endif

/* RaiseArgTupleInvalid.proto */
static void __Pyx_RaiseArgtupleInvalid(const char* func_name, int exact,
    Py_ssize_t num_min, Py_ssize_t num_max, Py_ssize_t num_found);

/* RaiseDoubleKeywords.proto */
static void __Pyx_RaiseDoubleKeywordsError(const char* func_name, PyObject* kw_name);

/* ParseKeywords.proto */
static int __Pyx_ParseOptionalKeywords(PyObject *kwds, PyObject *const *kwvalues,
    PyObject **argnames[],
    PyObject *kwds2, PyObject *values[], Py_ssize_t num_pos_args,
    const char* function_name);

/* ArgTypeTest.proto */
#define __Pyx_ArgTypeTest(obj, type, none_allowed, name, exact)\
    ((likely(__Pyx_IS_TYPE(obj, type) | (none_allowed && (obj == Py_None)))) ? 1 :\
        __Pyx__ArgTypeTest(obj, type, name, exact))
static int __Pyx__ArgTypeTest(PyObject *obj, PyTypeObject *type, const char *name, int exact);

/* PyDictVersioning.proto */
#if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
#define __PYX_DICT_VERSION_INIT  ((PY_UINT64_T) -1)
#define __PYX_GET_DICT_VERSION(dict)  (((PyDictObject*)(dict))->ma_version_tag)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)\
    (version_var) = __PYX_GET_DICT_VERSION(dict);\
    (cache_var) = (value);
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP) {\
    static PY_UINT64_T __pyx_dict_version = 0;\
    static PyObject *__pyx_dict_cached_value = NULL;\
    if (likely(__PYX_GET_DICT_VERSION(DICT) == __pyx_dict_version)) {\
        (VAR) = __pyx_dict_cached_value;\
    } else {\
        (VAR) = __pyx_dict_cached_value = (LOOKUP);\
        __pyx_dict_version = __PYX_GET_DICT_VERSION(DICT);\
    }\
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj);
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj);
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version);
#else
#define __PYX_GET_DICT_VERSION(dict)  (0)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP)  (VAR) = (LOOKUP);
#endif

/* GetModuleGlobalName.proto */
#if CYTHON_USE_DICT_VERSIONS
#define __Pyx_GetModuleGlobalName(var, name)  do {\
    static PY_UINT64_T __pyx_dict_version = 0;\
    static PyObject *__pyx_dict_cached_value = NULL;\
    (var) = (likely(__pyx_dict_version == __PYX_GET_DICT_VERSION(__pyx_d))) ?\
        (likely(__pyx_dict_cached_value) ? __Pyx_NewRef(__pyx_dict_cached_value) : __Pyx_GetBuiltinName(name)) :\
        __Pyx__GetModuleGlobalName(name, &__pyx_dict_version, &__pyx_dict_cached_value);\
} while(0)
#define __Pyx_GetModuleGlobalNameUncached(var, name)  do {\
    PY_UINT64_T __pyx_dict_version;\
    PyObject *__pyx_dict_cached_value;\
    (var) = __Pyx__GetModuleGlobalName(name, &__pyx_dict_version, &__pyx_dict_cached_value);\
} while(0)
static PyObject *__Pyx__GetModuleGlobalName(PyObject *name, PY_UINT64_T *dict_version, PyObject **dict_cached_value);
#else
#define __Pyx_GetModuleGlobalName(var, name)  (var) = __Pyx__GetModuleGlobalName(name)
#define __Pyx_GetModuleGlobalNameUncached(var, name)  (var) = __Pyx__GetModuleGlobalName(name)
static CYTHON_INLINE PyObject *__Pyx__GetModuleGlobalName(PyObject *name);
#endif

/* PyObjectCall.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw);
#else
#define __Pyx_PyObject_Call(func, arg, kw) PyObject_Call(func, arg, kw)
#endif

/* GetTopmostException.proto */
#if CYTHON_USE_EXC_INFO_STACK && CYTHON_FAST_THREAD_STATE
static _PyErr_StackItem * __Pyx_PyErr_GetTopmostException(PyThreadState *tstate);
#endif

/* SaveResetException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_ExceptionSave(type, value, tb)  __Pyx__ExceptionSave(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#define __Pyx_ExceptionReset(type, value, tb)  __Pyx__ExceptionReset(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
#else
#define __Pyx_ExceptionSave(type, value, tb)   PyErr_GetExcInfo(type, value, tb)
#define __Pyx_ExceptionReset(type, value, tb)  PyErr_SetExcInfo(type, value, tb)
#endif

/* FastTypeChecks.proto */
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_TypeCheck(obj, type) __Pyx_IsSubtype(Py_TYPE(obj), (PyTypeObject *)type)
#define __Pyx_TypeCheck2(obj, type1, type2) __Pyx_IsAnySubtype2(Py_TYPE(obj), (PyTypeObject *)type1, (PyTypeObject *)type2)
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b);
static CYTHON_INLINE int __Pyx_IsAnySubtype2(PyTypeObject *cls, PyTypeObject *a, PyTypeObject *b);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject *type);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *type1, PyObject *type2);
#else
#define __Pyx_TypeCheck(obj, type) PyObject_TypeCheck(obj, (PyTypeObject *)type)
#define __Pyx_TypeCheck2(obj, type1, type2) (PyObject_TypeCheck(obj, (PyTypeObject *)type1) || PyObject_TypeCheck(obj, (PyTypeObject *)type2))
#define __Pyx_PyErr_GivenExceptionMatches(err, type) PyErr_GivenExceptionMatches(err, type)
#define __Pyx_PyErr_GivenExceptionMatches2(err, type1, type2) (PyErr_GivenExceptionMatches(err, type1) || PyErr_GivenExceptionMatches(err, type2))
#endif
#define __Pyx_PyErr_ExceptionMatches2(err1, err2)  __Pyx_PyErr_GivenExceptionMatches2(__Pyx_PyErr_CurrentExceptionType(), err1, err2)
#define __Pyx_PyException_Check(obj) __Pyx_TypeCheck(obj, PyExc_Exception)

/* GetException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_GetException(type, value, tb)  __Pyx__GetException(__pyx_tstate, type, value, tb)
static int __Pyx__GetException(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#else
static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb);
#endif

/* PyFunctionFastCall.proto */
#if CYTHON_FAST_PYCALL
#if !CYTHON_VECTORCALL
#define __Pyx_PyFunction_FastCall(func, args, nargs)\
    __Pyx_PyFunction_FastCallDict((func), (args), (nargs), NULL)
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject **args, Py_ssize_t nargs, PyObject *kwargs);
#endif
#define __Pyx_BUILD_ASSERT_EXPR(cond)\
    (sizeof(char [1 - 2*!(cond)]) - 1)
#ifndef Py_MEMBER_SIZE
#define Py_MEMBER_SIZE(type, member) sizeof(((type *)0)->member)
#endif
#if !CYTHON_VECTORCALL
#if PY_VERSION_HEX >= 0x03080000
  #include "frameobject.h"
#if PY_VERSION_HEX >= 0x030b00a6 && !CYTHON_COMPILING_IN_LIMITED_API
  #ifndef Py_BUILD_CORE
    #define Py_BUILD_CORE 1
  #endif
  #include "internal/pycore_frame.h"
#endif
  #define __Pxy_PyFrame_Initialize_Offsets()
  #define __Pyx_PyFrame_GetLocalsplus(frame)  ((frame)->f_localsplus)
#else
  static size_t __pyx_pyframe_localsplus_offset = 0;
  #include "frameobject.h"
  #define __Pxy_PyFrame_Initialize_Offsets()\
    ((void)__Pyx_BUILD_ASSERT_EXPR(sizeof(PyFrameObject) == offsetof(PyFrameObject, f_localsplus) + Py_MEMBER_SIZE(PyFrameObject, f_localsplus)),\
     (void)(__pyx_pyframe_localsplus_offset = ((size_t)PyFrame_Type.tp_basicsize) - Py_MEMBER_SIZE(PyFrameObject, f_localsplus)))
  #define __Pyx_PyFrame_GetLocalsplus(frame)\
    (assert(__pyx_pyframe_localsplus_offset), (PyObject **)(((char *)(frame)) + __pyx_pyframe_localsplus_offset))
#endif
#endif
#endif

/* PyObjectCallMethO.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg);
#endif

/* PyObjectFastCall.proto */
#define __Pyx_PyObject_FastCall(func, args, nargs)  __Pyx_PyObject_FastCallDict(func, args, (size_t)(nargs), NULL)
static CYTHON_INLINE PyObject* __Pyx_PyObject_FastCallDict(PyObject *func, PyObject **args, size_t nargs, PyObject *kwargs);

/* PyObject_Str.proto */
#define __Pyx_PyObject_Str(obj)\
    (likely(PyString_CheckExact(obj)) ? __Pyx_NewRef(obj) : PyObject_Str(obj))

/* IterFinish.proto */
static CYTHON_INLINE int __Pyx_IterFinish(void);

/* PyObjectCallNoArg.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func);

/* PyObjectCallOneArg.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg);

/* PyObjectGetMethod.proto */
static int __Pyx_PyObject_GetMethod(PyObject *obj, PyObject *name, PyObject **method);

/* PyObjectCallMethod0.proto */
static PyObject* __Pyx_PyObject_CallMethod0(PyObject* obj, PyObject* method_name);

/* RaiseNeedMoreValuesToUnpack.proto */
static CYTHON_INLINE void __Pyx_RaiseNeedMoreValuesError(Py_ssize_t index);

/* RaiseTooManyValuesToUnpack.proto */
static CYTHON_INLINE void __Pyx_RaiseTooManyValuesError(Py_ssize_t expected);

/* UnpackItemEndCheck.proto */
static int __Pyx_IternextUnpackEndCheck(PyObject *retval, Py_ssize_t expected);

/* RaiseNoneIterError.proto */
static CYTHON_INLINE void __Pyx_RaiseNoneNotIterableError(void);

/* UnpackTupleError.proto */
static void __Pyx_UnpackTupleError(PyObject *, Py_ssize_t index);

/* UnpackTuple2.proto */
#define __Pyx_unpack_tuple2(tuple, value1, value2, is_tuple, has_known_size, decref_tuple)\
    (likely(is_tuple || PyTuple_Check(tuple)) ?\
        (likely(has_known_size || PyTuple_GET_SIZE(tuple) == 2) ?\
            __Pyx_unpack_tuple2_exact(tuple, value1, value2, decref_tuple) :\
            (__Pyx_UnpackTupleError(tuple, 2), -1)) :\
        __Pyx_unpack_tuple2_generic(tuple, value1, value2, has_known_size, decref_tuple))
static CYTHON_INLINE int __Pyx_unpack_tuple2_exact(
    PyObject* tuple, PyObject** value1, PyObject** value2, int decref_tuple);
static int __Pyx_unpack_tuple2_generic(
    PyObject* tuple, PyObject** value1, PyObject** value2, int has_known_size, int decref_tuple);

/* dict_iter.proto */
static CYTHON_INLINE PyObject* __Pyx_dict_iterator(PyObject* dict, int is_dict, PyObject* method_name,
                                                   Py_ssize_t* p_orig_length, int* p_is_dict);
static CYTHON_INLINE int __Pyx_dict_iter_next(PyObject* dict_or_iter, Py_ssize_t orig_length, Py_ssize_t* ppos,
                                              PyObject** pkey, PyObject** pvalue, PyObject** pitem, int is_dict);

/* pyfrozenset_new.proto */
static CYTHON_INLINE PyObject* __Pyx_PyFrozenSet_New(PyObject* it);

/* PySetContains.proto */
static CYTHON_INLINE int __Pyx_PySet_ContainsTF(PyObject* key, PyObject* set, int eq);

/* ListAppend.proto */
#if CYTHON_USE_PYLIST_INTERNALS && CYTHON_ASSUME_SAFE_MACROS
static CYTHON_INLINE int __Pyx_PyList_Append(PyObject* list, PyObject* x) {
    PyListObject* L = (PyListObject*) list;
    Py_ssize_t len = Py_SIZE(list);
    if (likely(L->allocated > len) & likely(len > (L->allocated >> 1))) {
        Py_INCREF(x);
        #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030d0000
        L->ob_item[len] = x;
        #else
        PyList_SET_ITEM(list, len, x);
        #endif
        __Pyx_SET_SIZE(list, len + 1);
        return 0;
    }
    return PyList_Append(list, x);
}
#else
#define __Pyx_PyList_Append(L,x) PyList_Append(L,x)
#endif

/* DictGetItem.proto */
#if PY_MAJOR_VERSION >= 3 && !CYTHON_COMPILING_IN_PYPY
static PyObject *__Pyx_PyDict_GetItem(PyObject *d, PyObject* key);
#define __Pyx_PyObject_Dict_GetItem(obj, name)\
    (likely(PyDict_CheckExact(obj)) ?\
     __Pyx_PyDict_GetItem(obj, name) : PyObject_GetItem(obj, name))
#else
#define __Pyx_PyDict_GetItem(d, key) PyObject_GetItem(d, key)
#define __Pyx_PyObject_Dict_GetItem(obj, name)  PyObject_GetItem(obj, name)
#endif

/* ListCompAppend.proto */
#if CYTHON_USE_PYLIST_INTERNALS && CYTHON_ASSUME_SAFE_MACROS
static CYTHON_INLINE int __Pyx_ListComp_Append(PyObject* list, PyObject* x) {
    PyListObject* L = (PyListObject*) list;
    Py_ssize_t len = Py_SIZE(list);
    if (likely(L->allocated > len)) {
        Py_INCREF(x);
        #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030d0000
        L->ob_item[len] = x;
        #else
        PyList_SET_ITEM(list, len, x);
        #endif
        __Pyx_SET_SIZE(list, len + 1);
        return 0;
    }
    return PyList_Append(list, x);
}
#else
#define __Pyx_ListComp_Append(L,x) PyList_Append(L,x)
#endif

/* AssertionsEnabled.proto */
#if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX < 0x02070600 && !defined(Py_OptimizeFlag)
  #define __Pyx_init_assertions_enabled()  (0)
  #define __pyx_assertions_enabled()  (1)
#elif CYTHON_COMPILING_IN_LIMITED_API  ||  (CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030C0000)
  static int __pyx_assertions_enabled_flag;
  #define __pyx_assertions_enabled() (__pyx_assertions_enabled_flag)
  static int __Pyx_init_assertions_enabled(void) {
    PyObject *builtins, *debug, *debug_str;
    int flag;
    builtins = PyEval_GetBuiltins();
    if (!builtins) goto bad;
    debug_str = PyUnicode_FromStringAndSize("__debug__", 9);
    if (!debug_str) goto bad;
    debug = PyObject_GetItem(builtins, debug_str);
    Py_DECREF(debug_str);
    if (!debug) goto bad;
    flag = PyObject_IsTrue(debug);
    Py_DECREF(debug);
    if (flag == -1) goto bad;
    __pyx_assertions_enabled_flag = flag;
    return 0;
  bad:
    __pyx_assertions_enabled_flag = 1;
    return -1;
  }
#else
  #define __Pyx_init_assertions_enabled()  (0)
  #define __pyx_assertions_enabled()  (!Py_OptimizeFlag)
#endif

/* PyObjectFormatAndDecref.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_FormatSimpleAndDecref(PyObject* s, PyObject* f);
static CYTHON_INLINE PyObject* __Pyx_PyObject_FormatAndDecref(PyObject* s, PyObject* f);

/* JoinPyUnicode.proto */
static PyObject* __Pyx_PyUnicode_Join(PyObject* value_tuple, Py_ssize_t value_count, Py_ssize_t result_ulength,
                                      Py_UCS4 max_char);

/* RaiseException.proto */
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause);

/* MoveIfSupported.proto */
#if CYTHON_USE_CPP_STD_MOVE
  #include <utility>
  #define __PYX_STD_MOVE_IF_SUPPORTED(x) std::move(x)
#else
  #define __PYX_STD_MOVE_IF_SUPPORTED(x) x
#endif

/* PyObjectSetAttrStr.proto */
#if CYTHON_USE_TYPE_SLOTS
#define __Pyx_PyObject_DelAttrStr(o,n) __Pyx_PyObject_SetAttrStr(o, n, NULL)
static CYTHON_INLINE int __Pyx_PyObject_SetAttrStr(PyObject* obj, PyObject* attr_name, PyObject* value);
#else
#define __Pyx_PyObject_DelAttrStr(o,n)   PyObject_DelAttr(o,n)
#define __Pyx_PyObject_SetAttrStr(o,n,v) PyObject_SetAttr(o,n,v)
#endif

/* SliceObject.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetSlice(
        PyObject* obj, Py_ssize_t cstart, Py_ssize_t cstop,
        PyObject** py_start, PyObject** py_stop, PyObject** py_slice,
        int has_cstart, int has_cstop, int wraparound);

/* PySequenceContains.proto */
static CYTHON_INLINE int __Pyx_PySequence_ContainsTF(PyObject* item, PyObject* seq, int eq) {
    int result = PySequence_Contains(seq, item);
    return unlikely(result < 0) ? result : (result == (eq == Py_EQ));
}

/* PyObjectCall2Args.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call2Args(PyObject* function, PyObject* arg1, PyObject* arg2);

/* PyObjectCallMethod1.proto */
static PyObject* __Pyx_PyObject_CallMethod1(PyObject* obj, PyObject* method_name, PyObject* arg);

/* append.proto */
static CYTHON_INLINE int __Pyx_PyObject_Append(PyObject* L, PyObject* x);

/* GetItemInt.proto */
#define __Pyx_GetItemInt(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Fast(o, (Py_ssize_t)i, is_list, wraparound, boundscheck) :\
    (is_list ? (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL) :\
               __Pyx_GetItemInt_Generic(o, to_py_func(i))))
#define __Pyx_GetItemInt_List(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_List_Fast(o, (Py_ssize_t)i, wraparound, boundscheck) :\
    (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck);
#define __Pyx_GetItemInt_Tuple(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Tuple_Fast(o, (Py_ssize_t)i, wraparound, boundscheck) :\
    (PyErr_SetString(PyExc_IndexError, "tuple index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck);
static PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j);
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i,
                                                     int is_list, int wraparound, int boundscheck);

/* PyIntCompare.proto */
static CYTHON_INLINE int __Pyx_PyInt_BoolEqObjC(PyObject *op1, PyObject *op2, long intval, long inplace);

/* ListExtend.proto */
static CYTHON_INLINE int __Pyx_PyList_Extend(PyObject* L, PyObject* v) {
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030d0000
    PyObject* none = _PyList_Extend((PyListObject*)L, v);
    if (unlikely(!none))
        return -1;
    Py_DECREF(none);
    return 0;
#else
    return PyList_SetSlice(L, PY_SSIZE_T_MAX, PY_SSIZE_T_MAX, v);
#endif
}

/* PyUnicode_Unicode.proto */
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_Unicode(PyObject *obj);

/* GCCDiagnostics.proto */
#if !defined(__INTEL_COMPILER) && defined(__GNUC__) && (__GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 6))
#define __Pyx_HAS_GCC_DIAGNOSTIC
#endif

/* BuildPyUnicode.proto */
static PyObject* __Pyx_PyUnicode_BuildFromAscii(Py_ssize_t ulength, char* chars, int clength,
                                                int prepend_sign, char padding_char);

/* CIntToPyUnicode.proto */
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_From_Py_ssize_t(Py_ssize_t value, Py_ssize_t width, char padding_char, char format_char);

/* ObjectGetItem.proto */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject *__Pyx_PyObject_GetItem(PyObject *obj, PyObject *key);
#else
#define __Pyx_PyObject_GetItem(obj, key)  PyObject_GetItem(obj, key)
#endif

/* RaiseUnboundLocalError.proto */
static CYTHON_INLINE void __Pyx_RaiseUnboundLocalError(const char *varname);

/* pep479.proto */
static void __Pyx_Generator_Replace_StopIteration(int in_async_gen);

/* UnpackUnboundCMethod.proto */
typedef struct {
    PyObject *type;
    PyObject **method_name;
    PyCFunction func;
    PyObject *method;
    int flag;
} __Pyx_CachedCFunction;

/* CallUnboundCMethod1.proto */
static PyObject* __Pyx__CallUnboundCMethod1(__Pyx_CachedCFunction* cfunc, PyObject* self, PyObject* arg);
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_CallUnboundCMethod1(__Pyx_CachedCFunction* cfunc, PyObject* self, PyObject* arg);
#else
#define __Pyx_CallUnboundCMethod1(cfunc, self, arg)  __Pyx__CallUnboundCMethod1(cfunc, self, arg)
#endif

/* PyIntBinop.proto */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyInt_AddObjC(PyObject *op1, PyObject *op2, long intval, int inplace, int zerodivision_check);
#else
#define __Pyx_PyInt_AddObjC(op1, op2, intval, inplace, zerodivision_check)\
    (inplace ? PyNumber_InPlaceAdd(op1, op2) : PyNumber_Add(op1, op2))
#endif

/* dict_getitem_default.proto */
static PyObject* __Pyx_PyDict_GetItemDefault(PyObject* d, PyObject* key, PyObject* default_value);

/* CallUnboundCMethod2.proto */
static PyObject* __Pyx__CallUnboundCMethod2(__Pyx_CachedCFunction* cfunc, PyObject* self, PyObject* arg1, PyObject* arg2);
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030600B1
static CYTHON_INLINE PyObject *__Pyx_CallUnboundCMethod2(__Pyx_CachedCFunction *cfunc, PyObject *self, PyObject *arg1, PyObject *arg2);
#else
#define __Pyx_CallUnboundCMethod2(cfunc, self, arg1, arg2)  __Pyx__CallUnboundCMethod2(cfunc, self, arg1, arg2)
#endif

/* py_abs.proto */
#if CYTHON_USE_PYLONG_INTERNALS
static PyObject *__Pyx_PyLong_AbsNeg(PyObject *num);
#define __Pyx_PyNumber_Absolute(x)\
    ((likely(PyLong_CheckExact(x))) ?\
         (likely(__Pyx_PyLong_IsNonNeg(x)) ? (Py_INCREF(x), (x)) : __Pyx_PyLong_AbsNeg(x)) :\
         PyNumber_Absolute(x))
#else
#define __Pyx_PyNumber_Absolute(x)  PyNumber_Absolute(x)
#endif

/* py_dict_pop.proto */
static CYTHON_INLINE PyObject *__Pyx_PyDict_Pop(PyObject *d, PyObject *key, PyObject *default_value);

/* SwapException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_ExceptionSwap(type, value, tb)  __Pyx__ExceptionSwap(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionSwap(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#else
static CYTHON_INLINE void __Pyx_ExceptionSwap(PyObject **type, PyObject **value, PyObject **tb);
#endif

/* PyNumberPow2.proto */
#define __Pyx_PyNumber_InPlacePowerOf2(a, b, c) __Pyx__PyNumber_PowerOf2(a, b, c, 1)
#define __Pyx_PyNumber_PowerOf2(a, b, c) __Pyx__PyNumber_PowerOf2(a, b, c, 0)
static PyObject* __Pyx__PyNumber_PowerOf2(PyObject *two, PyObject *exp, PyObject *none, int inplace);

/* PyObjectFormatSimple.proto */
#if CYTHON_COMPILING_IN_PYPY
    #define __Pyx_PyObject_FormatSimple(s, f) (\
        likely(PyUnicode_CheckExact(s)) ? (Py_INCREF(s), s) :\
        PyObject_Format(s, f))
#elif PY_MAJOR_VERSION < 3
    #define __Pyx_PyObject_FormatSimple(s, f) (\
        likely(PyUnicode_CheckExact(s)) ? (Py_INCREF(s), s) :\
        likely(PyString_CheckExact(s)) ? PyUnicode_FromEncodedObject(s, NULL, "strict") :\
        PyObject_Format(s, f))
#elif CYTHON_USE_TYPE_SLOTS
    #define __Pyx_PyObject_FormatSimple(s, f) (\
        likely(PyUnicode_CheckExact(s)) ? (Py_INCREF(s), s) :\
        likely(PyLong_CheckExact(s)) ? PyLong_Type.tp_repr(s) :\
        likely(PyFloat_CheckExact(s)) ? PyFloat_Type.tp_repr(s) :\
        PyObject_Format(s, f))
#else
    #define __Pyx_PyObject_FormatSimple(s, f) (\
        likely(PyUnicode_CheckExact(s)) ? (Py_INCREF(s), s) :\
        PyObject_Format(s, f))
#endif

/* SliceTupleAndList.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyList_GetSlice(PyObject* src, Py_ssize_t start, Py_ssize_t stop);
static CYTHON_INLINE PyObject* __Pyx_PyTuple_GetSlice(PyObject* src, Py_ssize_t start, Py_ssize_t stop);
#else
#define __Pyx_PyList_GetSlice(seq, start, stop)   PySequence_GetSlice(seq, start, stop)
#define __Pyx_PyTuple_GetSlice(seq, start, stop)  PySequence_GetSlice(seq, start, stop)
#endif

/* RaiseUnexpectedTypeError.proto */
static int __Pyx_RaiseUnexpectedTypeError(const char *expected, PyObject *obj);

/* IncludeStructmemberH.proto */
#include <structmember.h>

/* FixUpExtensionType.proto */
#if CYTHON_USE_TYPE_SPECS
static int __Pyx_fix_up_extension_type_from_spec(PyType_Spec *spec, PyTypeObject *type);
#endif

/* ValidateBasesTuple.proto */
#if CYTHON_COMPILING_IN_CPYTHON || CYTHON_COMPILING_IN_LIMITED_API || CYTHON_USE_TYPE_SPECS
static int __Pyx_validate_bases_tuple(const char *type_name, Py_ssize_t dictoffset, PyObject *bases);
#endif

/* PyType_Ready.proto */
CYTHON_UNUSED static int __Pyx_PyType_Ready(PyTypeObject *t);

/* PyObject_GenericGetAttrNoDict.proto */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static CYTHON_INLINE PyObject* __Pyx_PyObject_GenericGetAttrNoDict(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GenericGetAttrNoDict PyObject_GenericGetAttr
#endif

/* TypeImport.proto */
#ifndef __PYX_HAVE_RT_ImportType_proto_3_0_10
#define __PYX_HAVE_RT_ImportType_proto_3_0_10
#if defined (__STDC_VERSION__) && __STDC_VERSION__ >= 201112L
#include <stdalign.h>
#endif
#if (defined (__STDC_VERSION__) && __STDC_VERSION__ >= 201112L) || __cplusplus >= 201103L
#define __PYX_GET_STRUCT_ALIGNMENT_3_0_10(s) alignof(s)
#else
#define __PYX_GET_STRUCT_ALIGNMENT_3_0_10(s) sizeof(void*)
#endif
enum __Pyx_ImportType_CheckSize_3_0_10 {
   __Pyx_ImportType_CheckSize_Error_3_0_10 = 0,
   __Pyx_ImportType_CheckSize_Warn_3_0_10 = 1,
   __Pyx_ImportType_CheckSize_Ignore_3_0_10 = 2
};
static PyTypeObject *__Pyx_ImportType_3_0_10(PyObject* module, const char *module_name, const char *class_name, size_t size, size_t alignment, enum __Pyx_ImportType_CheckSize_3_0_10 check_size);
#endif

/* Import.proto */
static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level);

/* ImportFrom.proto */
static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name);

/* ImportDottedModule.proto */
static PyObject *__Pyx_ImportDottedModule(PyObject *name, PyObject *parts_tuple);
#if PY_MAJOR_VERSION >= 3
static PyObject *__Pyx_ImportDottedModule_WalkParts(PyObject *module, PyObject *name, PyObject *parts_tuple);
#endif

/* FetchSharedCythonModule.proto */
static PyObject *__Pyx_FetchSharedCythonABIModule(void);

/* FetchCommonType.proto */
#if !CYTHON_USE_TYPE_SPECS
static PyTypeObject* __Pyx_FetchCommonType(PyTypeObject* type);
#else
static PyTypeObject* __Pyx_FetchCommonTypeFromSpec(PyObject *module, PyType_Spec *spec, PyObject *bases);
#endif

/* PyMethodNew.proto */
#if CYTHON_COMPILING_IN_LIMITED_API
static PyObject *__Pyx_PyMethod_New(PyObject *func, PyObject *self, PyObject *typ) {
    PyObject *typesModule=NULL, *methodType=NULL, *result=NULL;
    CYTHON_UNUSED_VAR(typ);
    if (!self)
        return __Pyx_NewRef(func);
    typesModule = PyImport_ImportModule("types");
    if (!typesModule) return NULL;
    methodType = PyObject_GetAttrString(typesModule, "MethodType");
    Py_DECREF(typesModule);
    if (!methodType) return NULL;
    result = PyObject_CallFunctionObjArgs(methodType, func, self, NULL);
    Py_DECREF(methodType);
    return result;
}
#elif PY_MAJOR_VERSION >= 3
static PyObject *__Pyx_PyMethod_New(PyObject *func, PyObject *self, PyObject *typ) {
    CYTHON_UNUSED_VAR(typ);
    if (!self)
        return __Pyx_NewRef(func);
    return PyMethod_New(func, self);
}
#else
    #define __Pyx_PyMethod_New PyMethod_New
#endif

/* PyVectorcallFastCallDict.proto */
#if CYTHON_METH_FASTCALL
static CYTHON_INLINE PyObject *__Pyx_PyVectorcall_FastCallDict(PyObject *func, __pyx_vectorcallfunc vc, PyObject *const *args, size_t nargs, PyObject *kw);
#endif

/* CythonFunctionShared.proto */
#define __Pyx_CyFunction_USED
#define __Pyx_CYFUNCTION_STATICMETHOD  0x01
#define __Pyx_CYFUNCTION_CLASSMETHOD   0x02
#define __Pyx_CYFUNCTION_CCLASS        0x04
#define __Pyx_CYFUNCTION_COROUTINE     0x08
#define __Pyx_CyFunction_GetClosure(f)\
    (((__pyx_CyFunctionObject *) (f))->func_closure)
#if PY_VERSION_HEX < 0x030900B1 || CYTHON_COMPILING_IN_LIMITED_API
  #define __Pyx_CyFunction_GetClassObj(f)\
      (((__pyx_CyFunctionObject *) (f))->func_classobj)
#else
  #define __Pyx_CyFunction_GetClassObj(f)\
      ((PyObject*) ((PyCMethodObject *) (f))->mm_class)
#endif
#define __Pyx_CyFunction_SetClassObj(f, classobj)\
    __Pyx__CyFunction_SetClassObj((__pyx_CyFunctionObject *) (f), (classobj))
#define __Pyx_CyFunction_Defaults(type, f)\
    ((type *)(((__pyx_CyFunctionObject *) (f))->defaults))
#define __Pyx_CyFunction_SetDefaultsGetter(f, g)\
    ((__pyx_CyFunctionObject *) (f))->defaults_getter = (g)
typedef struct {
#if CYTHON_COMPILING_IN_LIMITED_API
    PyObject_HEAD
    PyObject *func;
#elif PY_VERSION_HEX < 0x030900B1
    PyCFunctionObject func;
#else
    PyCMethodObject func;
#endif
#if CYTHON_BACKPORT_VECTORCALL
    __pyx_vectorcallfunc func_vectorcall;
#endif
#if PY_VERSION_HEX < 0x030500A0 || CYTHON_COMPILING_IN_LIMITED_API
    PyObject *func_weakreflist;
#endif
    PyObject *func_dict;
    PyObject *func_name;
    PyObject *func_qualname;
    PyObject *func_doc;
    PyObject *func_globals;
    PyObject *func_code;
    PyObject *func_closure;
#if PY_VERSION_HEX < 0x030900B1 || CYTHON_COMPILING_IN_LIMITED_API
    PyObject *func_classobj;
#endif
    void *defaults;
    int defaults_pyobjects;
    size_t defaults_size;
    int flags;
    PyObject *defaults_tuple;
    PyObject *defaults_kwdict;
    PyObject *(*defaults_getter)(PyObject *);
    PyObject *func_annotations;
    PyObject *func_is_coroutine;
} __pyx_CyFunctionObject;
#undef __Pyx_CyOrPyCFunction_Check
#define __Pyx_CyFunction_Check(obj)  __Pyx_TypeCheck(obj, __pyx_CyFunctionType)
#define __Pyx_CyOrPyCFunction_Check(obj)  __Pyx_TypeCheck2(obj, __pyx_CyFunctionType, &PyCFunction_Type)
#define __Pyx_CyFunction_CheckExact(obj)  __Pyx_IS_TYPE(obj, __pyx_CyFunctionType)
static CYTHON_INLINE int __Pyx__IsSameCyOrCFunction(PyObject *func, void *cfunc);
#undef __Pyx_IsSameCFunction
#define __Pyx_IsSameCFunction(func, cfunc)   __Pyx__IsSameCyOrCFunction(func, cfunc)
static PyObject *__Pyx_CyFunction_Init(__pyx_CyFunctionObject* op, PyMethodDef *ml,
                                      int flags, PyObject* qualname,
                                      PyObject *closure,
                                      PyObject *module, PyObject *globals,
                                      PyObject* code);
static CYTHON_INLINE void __Pyx__CyFunction_SetClassObj(__pyx_CyFunctionObject* f, PyObject* classobj);
static CYTHON_INLINE void *__Pyx_CyFunction_InitDefaults(PyObject *m,
                                                         size_t size,
                                                         int pyobjects);
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsTuple(PyObject *m,
                                                            PyObject *tuple);
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsKwDict(PyObject *m,
                                                             PyObject *dict);
static CYTHON_INLINE void __Pyx_CyFunction_SetAnnotationsDict(PyObject *m,
                                                              PyObject *dict);
static int __pyx_CyFunction_init(PyObject *module);
#if CYTHON_METH_FASTCALL
static PyObject * __Pyx_CyFunction_Vectorcall_NOARGS(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames);
static PyObject * __Pyx_CyFunction_Vectorcall_O(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames);
static PyObject * __Pyx_CyFunction_Vectorcall_FASTCALL_KEYWORDS(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames);
static PyObject * __Pyx_CyFunction_Vectorcall_FASTCALL_KEYWORDS_METHOD(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames);
#if CYTHON_BACKPORT_VECTORCALL
#define __Pyx_CyFunction_func_vectorcall(f) (((__pyx_CyFunctionObject*)f)->func_vectorcall)
#else
#define __Pyx_CyFunction_func_vectorcall(f) (((PyCFunctionObject*)f)->vectorcall)
#endif
#endif

/* CythonFunction.proto */
static PyObject *__Pyx_CyFunction_New(PyMethodDef *ml,
                                      int flags, PyObject* qualname,
                                      PyObject *closure,
                                      PyObject *module, PyObject *globals,
                                      PyObject* code);

/* CLineInTraceback.proto */
#ifdef CYTHON_CLINE_IN_TRACEBACK
#define __Pyx_CLineForTraceback(tstate, c_line)  (((CYTHON_CLINE_IN_TRACEBACK)) ? c_line : 0)
#else
static int __Pyx_CLineForTraceback(PyThreadState *tstate, int c_line);
#endif

/* CodeObjectCache.proto */
#if !CYTHON_COMPILING_IN_LIMITED_API
typedef struct {
    PyCodeObject* code_object;
    int code_line;
} __Pyx_CodeObjectCacheEntry;
struct __Pyx_CodeObjectCache {
    int count;
    int max_count;
    __Pyx_CodeObjectCacheEntry* entries;
};
static struct __Pyx_CodeObjectCache __pyx_code_cache = {0,0,NULL};
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line);
static PyCodeObject *__pyx_find_code_object(int code_line);
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object);
#endif

/* AddTraceback.proto */
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename);

/* CppExceptionConversion.proto */
#ifndef __Pyx_CppExn2PyErr
#include <new>
#include <typeinfo>
#include <stdexcept>
#include <ios>
static void __Pyx_CppExn2PyErr() {
  try {
    if (PyErr_Occurred())
      ; // let the latest Python exn pass through and ignore the current one
    else
      throw;
  } catch (const std::bad_alloc& exn) {
    PyErr_SetString(PyExc_MemoryError, exn.what());
  } catch (const std::bad_cast& exn) {
    PyErr_SetString(PyExc_TypeError, exn.what());
  } catch (const std::bad_typeid& exn) {
    PyErr_SetString(PyExc_TypeError, exn.what());
  } catch (const std::domain_error& exn) {
    PyErr_SetString(PyExc_ValueError, exn.what());
  } catch (const std::invalid_argument& exn) {
    PyErr_SetString(PyExc_ValueError, exn.what());
  } catch (const std::ios_base::failure& exn) {
    PyErr_SetString(PyExc_IOError, exn.what());
  } catch (const std::out_of_range& exn) {
    PyErr_SetString(PyExc_IndexError, exn.what());
  } catch (const std::overflow_error& exn) {
    PyErr_SetString(PyExc_OverflowError, exn.what());
  } catch (const std::range_error& exn) {
    PyErr_SetString(PyExc_ArithmeticError, exn.what());
  } catch (const std::underflow_error& exn) {
    PyErr_SetString(PyExc_ArithmeticError, exn.what());
  } catch (const std::exception& exn) {
    PyErr_SetString(PyExc_RuntimeError, exn.what());
  }
  catch (...)
  {
    PyErr_SetString(PyExc_RuntimeError, "Unknown exception");
  }
}
#endif

/* CIntFromPy.proto */
static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int64_t(int64_t value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_uint64_t(uint64_t value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int(int value);

/* FormatTypeName.proto */
#if CYTHON_COMPILING_IN_LIMITED_API
typedef PyObject *__Pyx_TypeName;
#define __Pyx_FMT_TYPENAME "%U"
static __Pyx_TypeName __Pyx_PyType_GetName(PyTypeObject* tp);
#define __Pyx_DECREF_TypeName(obj) Py_XDECREF(obj)
#else
typedef const char *__Pyx_TypeName;
#define __Pyx_FMT_TYPENAME "%.200s"
#define __Pyx_PyType_GetName(tp) ((tp)->tp_name)
#define __Pyx_DECREF_TypeName(obj)
#endif

/* CIntFromPy.proto */
static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *);

/* CoroutineBase.proto */
struct __pyx_CoroutineObject;
typedef PyObject *(*__pyx_coroutine_body_t)(struct __pyx_CoroutineObject *, PyThreadState *, PyObject *);
#if CYTHON_USE_EXC_INFO_STACK
#define __Pyx_ExcInfoStruct  _PyErr_StackItem
#else
typedef struct {
    PyObject *exc_type;
    PyObject *exc_value;
    PyObject *exc_traceback;
} __Pyx_ExcInfoStruct;
#endif
typedef struct __pyx_CoroutineObject {
    PyObject_HEAD
    __pyx_coroutine_body_t body;
    PyObject *closure;
    __Pyx_ExcInfoStruct gi_exc_state;
    PyObject *gi_weakreflist;
    PyObject *classobj;
    PyObject *yieldfrom;
    PyObject *gi_name;
    PyObject *gi_qualname;
    PyObject *gi_modulename;
    PyObject *gi_code;
    PyObject *gi_frame;
    int resume_label;
    char is_running;
} __pyx_CoroutineObject;
static __pyx_CoroutineObject *__Pyx__Coroutine_New(
    PyTypeObject *type, __pyx_coroutine_body_t body, PyObject *code, PyObject *closure,
    PyObject *name, PyObject *qualname, PyObject *module_name);
static __pyx_CoroutineObject *__Pyx__Coroutine_NewInit(
            __pyx_CoroutineObject *gen, __pyx_coroutine_body_t body, PyObject *code, PyObject *closure,
            PyObject *name, PyObject *qualname, PyObject *module_name);
static CYTHON_INLINE void __Pyx_Coroutine_ExceptionClear(__Pyx_ExcInfoStruct *self);
static int __Pyx_Coroutine_clear(PyObject *self);
static PyObject *__Pyx_Coroutine_Send(PyObject *self, PyObject *value);
static PyObject *__Pyx_Coroutine_Close(PyObject *self);
static PyObject *__Pyx_Coroutine_Throw(PyObject *gen, PyObject *args);
#if CYTHON_USE_EXC_INFO_STACK
#define __Pyx_Coroutine_SwapException(self)
#define __Pyx_Coroutine_ResetAndClearException(self)  __Pyx_Coroutine_ExceptionClear(&(self)->gi_exc_state)
#else
#define __Pyx_Coroutine_SwapException(self) {\
    __Pyx_ExceptionSwap(&(self)->gi_exc_state.exc_type, &(self)->gi_exc_state.exc_value, &(self)->gi_exc_state.exc_traceback);\
    __Pyx_Coroutine_ResetFrameBackpointer(&(self)->gi_exc_state);\
    }
#define __Pyx_Coroutine_ResetAndClearException(self) {\
    __Pyx_ExceptionReset((self)->gi_exc_state.exc_type, (self)->gi_exc_state.exc_value, (self)->gi_exc_state.exc_traceback);\
    (self)->gi_exc_state.exc_type = (self)->gi_exc_state.exc_value = (self)->gi_exc_state.exc_traceback = NULL;\
    }
#endif
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyGen_FetchStopIterationValue(pvalue)\
    __Pyx_PyGen__FetchStopIterationValue(__pyx_tstate, pvalue)
#else
#define __Pyx_PyGen_FetchStopIterationValue(pvalue)\
    __Pyx_PyGen__FetchStopIterationValue(__Pyx_PyThreadState_Current, pvalue)
#endif
static int __Pyx_PyGen__FetchStopIterationValue(PyThreadState *tstate, PyObject **pvalue);
static CYTHON_INLINE void __Pyx_Coroutine_ResetFrameBackpointer(__Pyx_ExcInfoStruct *exc_state);

/* PatchModuleWithCoroutine.proto */
static PyObject* __Pyx_Coroutine_patch_module(PyObject* module, const char* py_code);

/* PatchGeneratorABC.proto */
static int __Pyx_patch_abc(void);

/* Generator.proto */
#define __Pyx_Generator_USED
#define __Pyx_Generator_CheckExact(obj) __Pyx_IS_TYPE(obj, __pyx_GeneratorType)
#define __Pyx_Generator_New(body, code, closure, name, qualname, module_name)\
    __Pyx__Coroutine_New(__pyx_GeneratorType, body, code, closure, name, qualname, module_name)
static PyObject *__Pyx_Generator_Next(PyObject *self);
static int __pyx_Generator_init(PyObject *module);

/* CStringEquals.proto */
static CYTHON_INLINE int __Pyx_StrEq(const char *, const char *);

/* CheckBinaryVersion.proto */
static unsigned long __Pyx_get_runtime_version(void);
static int __Pyx_check_binary_version(unsigned long ct_version, unsigned long rt_version, int allow_newer);

/* InitStrings.proto */
static int __Pyx_InitStrings(__Pyx_StringTabEntry *t);

/* #### Code section: module_declarations ### */

/* Module declarations from "libcpp" */

/* Module declarations from "libc.stdint" */

/* Module declarations from "libc.string" */

/* Module declarations from "libc.stdio" */

/* Module declarations from "__builtin__" */

/* Module declarations from "cpython.type" */

/* Module declarations from "cpython" */

/* Module declarations from "cpython.object" */

/* Module declarations from "cpython.bytes" */

/* Module declarations from "viur.datastore.transport" */
static CYTHON_INLINE PyObject *__pyx_f_4viur_9datastore_9transport_toPyStr(std::string_view); /*proto*/
static CYTHON_INLINE PyObject *__pyx_f_4viur_9datastore_9transport_parseKey(simdjson::dom::element); /*proto*/
static CYTHON_INLINE PyObject *__pyx_f_4viur_9datastore_9transport_toEntityStructure(simdjson::dom::element, struct __pyx_opt_args_4viur_9datastore_9transport_toEntityStructure *__pyx_optional_args); /*proto*/
static CYTHON_INLINE PyObject *__pyx_f_4viur_9datastore_9transport_toPythonStructure(simdjson::dom::element); /*proto*/
/* #### Code section: typeinfo ### */
/* #### Code section: before_global_var ### */
#define __Pyx_MODULE_NAME "viur.datastore.transport"
extern int __pyx_module_is_main_viur__datastore__transport;
int __pyx_module_is_main_viur__datastore__transport = 0;

/* Implementation of "viur.datastore.transport" */
/* #### Code section: global_var ### */
static PyObject *__pyx_builtin_range;
static PyObject *__pyx_builtin_AssertionError;
static PyObject *__pyx_builtin_ValueError;
static PyObject *__pyx_builtin_enumerate;
/* #### Code section: string_decls ### */
static const char __pyx_k_[] = " (";
static const char __pyx_k_i[] = "i";
static const char __pyx_k_k[] = "k";
static const char __pyx_k_v[] = "v";
static const char __pyx_k_x[] = "x";
static const char __pyx_k__3[] = ".";
static const char __pyx_k__5[] = " ";
static const char __pyx_k__6[] = "=";
static const char __pyx_k__7[] = "<";
static const char __pyx_k__8[] = "<=";
static const char __pyx_k__9[] = ">";
static const char __pyx_k_gc[] = "gc";
static const char __pyx_k_id[] = "id";
static const char __pyx_k_op[] = "op";
static const char __pyx_k_AND[] = "AND";
static const char __pyx_k_Any[] = "Any";
static const char __pyx_k_Get[] = "Get";
static const char __pyx_k_Key[] = "Key";
static const char __pyx_k_Put[] = "Put";
static const char __pyx_k__10[] = ">=";
static const char __pyx_k__11[] = "";
static const char __pyx_k__21[] = "*";
static const char __pyx_k__45[] = "?";
static const char __pyx_k_err[] = "err";
static const char __pyx_k_get[] = "get";
static const char __pyx_k_idx[] = "idx";
static const char __pyx_k_int[] = "int";
static const char __pyx_k_key[] = "key";
static const char __pyx_k_pop[] = "pop";
static const char __pyx_k_put[] = "put";
static const char __pyx_k_res[] = "res";
static const char __pyx_k_set[] = "set";
static const char __pyx_k_str[] = "str";
static const char __pyx_k_url[] = "url";
static const char __pyx_k_utc[] = "utc";
static const char __pyx_k_List[] = "List";
static const char __pyx_k_None[] = "None";
static const char __pyx_k_args[] = "args";
static const char __pyx_k_auth[] = "auth";
static const char __pyx_k_conf[] = "conf";
static const char __pyx_k_copy[] = "copy";
static const char __pyx_k_data[] = "data";
static const char __pyx_k_dict[] = "dict";
static const char __pyx_k_json[] = "json";
static const char __pyx_k_keys[] = "keys";
static const char __pyx_k_kind[] = "kind";
static const char __pyx_k_main[] = "__main__";
static const char __pyx_k_mode[] = "mode";
static const char __pyx_k_name[] = "name";
static const char __pyx_k_path[] = "path";
static const char __pyx_k_post[] = "post";
static const char __pyx_k_resp[] = "resp";
static const char __pyx_k_send[] = "send";
static const char __pyx_k_spec[] = "__spec__";
static const char __pyx_k_test[] = "__test__";
static const char __pyx_k_time[] = "time";
static const char __pyx_k_ASCII[] = "ASCII";
static const char __pyx_k_Count[] = "Count";
static const char __pyx_k_EQUAL[] = "EQUAL";
static const char __pyx_k_UTF_8[] = "UTF-8";
static const char __pyx_k_Union[] = "Union";
static const char __pyx_k_bytes[] = "bytes";
static const char __pyx_k_cache[] = "cache";
static const char __pyx_k_close[] = "close";
static const char __pyx_k_count[] = "count";
static const char __pyx_k_debug[] = "debug";
static const char __pyx_k_dumps[] = "dumps";
static const char __pyx_k_error[] = "error";
static const char __pyx_k_items[] = "items";
static const char __pyx_k_limit[] = "limit";
static const char __pyx_k_loads[] = "loads";
static const char __pyx_k_order[] = "order";
static const char __pyx_k_query[] = "query";
static const char __pyx_k_range[] = "range";
static const char __pyx_k_sleep[] = "sleep";
static const char __pyx_k_split[] = "split";
static const char __pyx_k_throw[] = "throw";
static const char __pyx_k_up_to[] = "up_to";
static const char __pyx_k_value[] = "value";
static const char __pyx_k_Delete[] = "Delete";
static const char __pyx_k_Entity[] = "Entity";
static const char __pyx_k_STRONG[] = "STRONG";
static const char __pyx_k_append[] = "append";
static const char __pyx_k_base64[] = "base64";
static const char __pyx_k_decode[] = "decode";
static const char __pyx_k_delete[] = "delete";
static const char __pyx_k_enable[] = "enable";
static const char __pyx_k_encode[] = "encode";
static const char __pyx_k_extend[] = "extend";
static const char __pyx_k_filter[] = "filter";
static const char __pyx_k_google[] = "google";
static const char __pyx_k_import[] = "__import__";
static const char __pyx_k_insert[] = "insert";
static const char __pyx_k_kwargs[] = "kwargs";
static const char __pyx_k_name_2[] = "__name__";
static const char __pyx_k_oldTxn[] = "oldTxn";
static const char __pyx_k_orders[] = "orders";
static const char __pyx_k_parent[] = "parent";
static const char __pyx_k_parser[] = "parser";
static const char __pyx_k_pysize[] = "pysize";
static const char __pyx_k_return[] = "return";
static const char __pyx_k_scopes[] = "scopes";
static const char __pyx_k_txnKey[] = "txnKey";
static const char __pyx_k_typing[] = "typing";
static const char __pyx_k_tzinfo[] = "tzinfo";
static const char __pyx_k_update[] = "update";
static const char __pyx_k_upsert[] = "upsert";
static const char __pyx_k_values[] = "values";
static const char __pyx_k_Queried[] = "Queried ";
static const char __pyx_k_arrayIt[] = "arrayIt";
static const char __pyx_k_content[] = "content";
static const char __pyx_k_default[] = "default";
static const char __pyx_k_dictKey[] = "dictKey";
static const char __pyx_k_disable[] = "disable";
static const char __pyx_k_element[] = "element";
static const char __pyx_k_filters[] = "filters";
static const char __pyx_k_genexpr[] = "genexpr";
static const char __pyx_k_isMulti[] = "isMulti";
static const char __pyx_k_keyList[] = "keyList";
static const char __pyx_k_logging[] = "logging";
static const char __pyx_k_replace[] = "replace";
static const char __pyx_k_resDict[] = "resDict";
static const char __pyx_k_resList[] = "resList";
static const char __pyx_k_results[] = " results";
static const char __pyx_k_seconds[] = " seconds";
static const char __pyx_k_version[] = "version";
static const char __pyx_k_warning[] = "warning";
static const char __pyx_k_Returned[] = ". Returned ";
static const char __pyx_k_callable[] = "callable";
static const char __pyx_k_callback[] = "callback";
static const char __pyx_k_data_ptr[] = "data_ptr";
static const char __pyx_k_datetime[] = "datetime";
static const char __pyx_k_distinct[] = "distinct";
static const char __pyx_k_entities[] = "entities";
static const char __pyx_k_keyValue[] = "keyValue";
static const char __pyx_k_postData[] = "postData";
static const char __pyx_k_property[] = "property";
static const char __pyx_k_requests[] = "requests";
static const char __pyx_k_strftime[] = "strftime";
static const char __pyx_k_strptime[] = "strptime";
static const char __pyx_k_timezone[] = "timezone";
static const char __pyx_k_ASCENDING[] = "ASCENDING";
static const char __pyx_k_LESS_THAN[] = "LESS_THAN";
static const char __pyx_k_List_dict[] = "List[dict]";
static const char __pyx_k_accessLog[] = "accessLog";
static const char __pyx_k_arrayElem[] = "arrayElem";
static const char __pyx_k_b64decode[] = "b64decode";
static const char __pyx_k_b64encode[] = "b64encode";
static const char __pyx_k_blobValue[] = "blobValue";
static const char __pyx_k_dictValue[] = "dictValue";
static const char __pyx_k_direction[] = "direction";
static const char __pyx_k_endCursor[] = "endCursor";
static const char __pyx_k_enumerate[] = "enumerate";
static const char __pyx_k_isenabled[] = "isenabled";
static const char __pyx_k_keyToPath[] = "keyToPath";
static const char __pyx_k_mutations[] = "mutations";
static const char __pyx_k_nullValue[] = "nullValue";
static const char __pyx_k_post_data[] = "post_data";
static const char __pyx_k_projectID[] = "projectID";
static const char __pyx_k_readWrite[] = "readWrite";
static const char __pyx_k_sortOrder[] = "sortOrder";
static const char __pyx_k_transport[] = "transport";
static const char __pyx_k_DESCENDING[] = "DESCENDING";
static const char __pyx_k_ValueError[] = "ValueError";
static const char __pyx_k_and_orders[] = " and orders ";
static const char __pyx_k_arrayValue[] = "arrayValue";
static const char __pyx_k_astimezone[] = "astimezone";
static const char __pyx_k_currentTxn[] = "currentTxn";
static const char __pyx_k_distinctOn[] = "distinctOn";
static const char __pyx_k_filterList[] = "filterList";
static const char __pyx_k_is_partial[] = "is_partial";
static const char __pyx_k_project_id[] = "project_id";
static const char __pyx_k_properties[] = "properties";
static const char __pyx_k_property_1[] = "property_1";
static const char __pyx_k_sleep_time[] = "sleep_time";
static const char __pyx_k_txnOptions[] = "txnOptions";
static const char __pyx_k_AllocateIDs[] = "AllocateIDs";
static const char __pyx_k_List_Entity[] = "List[Entity]";
static const char __pyx_k_credentials[] = "credentials";
static const char __pyx_k_distinctKey[] = "distinctKey";
static const char __pyx_k_doubleValue[] = "doubleValue";
static const char __pyx_k_entityValue[] = "entityValue";
static const char __pyx_k_flipResults[] = "flipResults";
static const char __pyx_k_google_auth[] = "google.auth";
static const char __pyx_k_partitionId[] = "partitionId";
static const char __pyx_k_readOptions[] = "readOptions";
static const char __pyx_k_res_from_db[] = "res_from_db";
static const char __pyx_k_rollbackTxn[] = "_rollbackTxn";
static const char __pyx_k_singleValue[] = "singleValue";
static const char __pyx_k_startCursor[] = "startCursor";
static const char __pyx_k_stringValue[] = "stringValue";
static const char __pyx_k_transaction[] = "transaction";
static const char __pyx_k_with_filter[] = " with filter ";
static const char __pyx_k_AbortedError[] = "AbortedError";
static const char __pyx_k_GREATER_THAN[] = "GREATER_THAN";
static const char __pyx_k_Invalid_op_s[] = "Invalid op %s";
static const char __pyx_k_NOT_FINISHED[] = "NOT_FINISHED";
static const char __pyx_k_Unknown_type[] = "Unknown type";
static const char __pyx_k_Y_m_dT_H_M_S[] = "%Y-%m-%dT%H:%M:%S";
static const char __pyx_k_aggregations[] = "aggregations";
static const char __pyx_k_booleanValue[] = "booleanValue";
static const char __pyx_k_deferred_key[] = "deferred_key";
static const char __pyx_k_initializing[] = "_initializing";
static const char __pyx_k_integerValue[] = "integerValue";
static const char __pyx_k_is_coroutine[] = "_is_coroutine";
static const char __pyx_k_missing_keys[] = "missing_keys";
static const char __pyx_k_nested_query[] = "nested_query";
static const char __pyx_k_resp_content[] = "resp.content=";
static const char __pyx_k_traceQueries[] = "traceQueries";
static const char __pyx_k_TRANSACTIONAL[] = "TRANSACTIONAL";
static const char __pyx_k_array_element[] = "array_element";
static const char __pyx_k_class_getitem[] = "__class_getitem__";
static const char __pyx_k_currentCursor[] = "currentCursor";
static const char __pyx_k_deferred_keys[] = "deferred_keys";
static const char __pyx_k_distinct_on_s[] = " distinct on %s";
static const char __pyx_k_element_inner[] = "element_inner";
static const char __pyx_k_http_internal[] = "_http_internal";
static const char __pyx_k_requestedKeys[] = "requestedKeys";
static const char __pyx_k_AssertionError[] = "AssertionError";
static const char __pyx_k_CollisionError[] = "CollisionError";
static const char __pyx_k_RecursionError[] = "RecursionError";
static const char __pyx_k_Y_m_dT_H_M_S_f[] = "%Y-%m-%dT%H:%M:%S.%f";
static const char __pyx_k_affectedEntity[] = "affectedEntity";
static const char __pyx_k_innerArrayElem[] = "innerArrayElem";
static const char __pyx_k_propertyFilter[] = "propertyFilter";
static const char __pyx_k_requested_keys[] = "requested_keys";
static const char __pyx_k_res_from_cache[] = "res_from_cache";
static const char __pyx_k_timestampValue[] = "timestampValue";
static const char __pyx_k_untouched_keys[] = "untouched_keys";
static const char __pyx_k_viur_datastore[] = "viur.datastore";
static const char __pyx_k_ConnectionError[] = "ConnectionError";
static const char __pyx_k_NO_MORE_RESULTS[] = "NO_MORE_RESULTS";
static const char __pyx_k_QueryDefinition[] = "QueryDefinition";
static const char __pyx_k_Y_m_dT_H_M_S_fZ[] = "%Y-%m-%dT%H:%M:%S.%fZ";
static const char __pyx_k_allowOverriding[] = "__allowOverriding__";
static const char __pyx_k_compositeFilter[] = "compositeFilter";
static const char __pyx_k_memcache_client[] = "memcache_client";
static const char __pyx_k_queryDefinition[] = "queryDefinition";
static const char __pyx_k_readConsistency[] = "readConsistency";
static const char __pyx_k_refresh_timeout[] = "refresh_timeout";
static const char __pyx_k_runSingleFilter[] = "runSingleFilter";
static const char __pyx_k_RunInTransaction[] = "RunInTransaction";
static const char __pyx_k_affectedEntities[] = "affectedEntities";
static const char __pyx_k_is_not_supported[] = ") is not supported";
static const char __pyx_k_keys_for_request[] = "keys_for_request";
static const char __pyx_k_pythonPropToJson[] = "pythonPropToJson";
static const char __pyx_k_AuthorizedSession[] = "AuthorizedSession";
static const char __pyx_k_NON_TRANSACTIONAL[] = "NON_TRANSACTIONAL";
static const char __pyx_k_aggregation_query[] = "aggregation_query";
static const char __pyx_k_allowOverriding_2[] = "allowOverriding";
static const char __pyx_k_requests_Response[] = "requests.Response";
static const char __pyx_k_Get_locals_genexpr[] = "Get.<locals>.genexpr";
static const char __pyx_k_LESS_THAN_OR_EQUAL[] = "LESS_THAN_OR_EQUAL";
static const char __pyx_k_Union_Key_List_Key[] = "Union[Key, List[Key]]";
static const char __pyx_k_ViurDatastoreError[] = "ViurDatastoreError";
static const char __pyx_k_aggregationResults[] = "aggregationResults";
static const char __pyx_k_asyncio_coroutines[] = "asyncio.coroutines";
static const char __pyx_k_cline_in_traceback[] = "cline_in_traceback";
static const char __pyx_k_currentDbAccessLog[] = "currentDbAccessLog";
static const char __pyx_k_currentTransaction[] = "currentTransaction";
static const char __pyx_k_excludeFromIndexes[] = "excludeFromIndexes";
static const char __pyx_k_excludedProperties[] = "excludedProperties";
static const char __pyx_k_missing_keys_paths[] = "missing_keys_paths";
static const char __pyx_k_transactionOptions[] = "transactionOptions";
static const char __pyx_k_aggregateProperties[] = "aggregateProperties";
static const char __pyx_k_array_element_inner[] = "array_element_inner";
static const char __pyx_k_exponential_backoff[] = "exponential_backoff";
static const char __pyx_k_from_legacy_urlsafe[] = "from_legacy_urlsafe";
static const char __pyx_k_internalStartCursor[] = "internalStartCursor";
static const char __pyx_k_previousTransaction[] = "previousTransaction";
static const char __pyx_k_requests_exceptions[] = "requests.exceptions";
static const char __pyx_k_InvalidArgumentError[] = "InvalidArgumentError";
static const char __pyx_k_exclude_from_indexes[] = "exclude_from_indexes";
static const char __pyx_k_viur_datastore_types[] = "viur.datastore.types";
static const char __pyx_k_GREATER_THAN_OR_EQUAL[] = "GREATER_THAN_OR_EQUAL";
static const char __pyx_k_authenticated_request[] = "authenticated_request";
static const char __pyx_k_viur_datastore_config[] = "viur.datastore.config";
static const char __pyx_k_viur_datastore_errors[] = "viur.datastore.errors";
static const char __pyx_k_NoMutationResultsError[] = "NoMutationResultsError";
static const char __pyx_k_RequestsConnectionError[] = "RequestsConnectionError";
static const char __pyx_k_Union_Entity_List_Entity[] = "Union[Entity, List[Entity]]";
static const char __pyx_k_viur_datastore_transport[] = "viur.datastore.transport";
static const char __pyx_k_INVALID_RESPONSE_RECEIVED[] = "INVALID RESPONSE RECEIVED";
static const char __pyx_k_Invalid_key_in_entity_json_s[] = "Invalid key in entity json: %s";
static const char __pyx_k_No_mutation_results_received[] = "No mutation results received";
static const char __pyx_k_is_viur_datastore_request_ok[] = "is_viur_datastore_request_ok";
static const char __pyx_k_Union_None_Entity_List_Entity[] = "Union[None, Entity, List[Entity]]";
static const char __pyx_k_No_mutation_results_received_2[] = "No mutation-results received";
static const char __pyx_k_Please_create_a_matching_index[] = ". - Please create a matching index";
static const char __pyx_k_All_retries_are_exhausted_for_th[] = "All retries are exhausted for this transaction";
static const char __pyx_k_Cannot_call_runInTransaction_whi[] = "Cannot call runInTransaction while inside a transaction!";
static const char __pyx_k_Empty_response_received_from_Dat[] = "Empty response received from Datastore API";
static const char __pyx_k_Invalid_data_received_from_Datas[] = "Invalid data received from Datastore API";
static const char __pyx_k_Invalid_number_of_mutation_resul[] = "Invalid number of mutation-results received";
static const char __pyx_k_Keys_must_not_be_partial_or_kind[] = "Keys must not be partial or kind-less";
static const char __pyx_k_Query_not_finished_Maybe_some_en[] = "Query not finished. Maybe some entries are missing.";
static const char __pyx_k_Received_an_unexpected_key_updat[] = "Received an unexpected key-update";
static const char __pyx_k_Retrying_http_post_request_to_da[] = "Retrying http post request to datastore";
static const char __pyx_k_The_Count_aggregation_query_is_a[] = "The 'Count() aggregation' query is a technical preview and cannot be guaranteed to work at this time!!!";
static const char __pyx_k_Union_Key_List_Key_Entity_List_E[] = "Union[Key, List[Key], Entity, List[Entity]]";
static const char __pyx_k_We_got_an_error_in_a_transaction[] = "We got an error in a transaction we try again in ";
static const char __pyx_k_We_read_an_incomplete_null_key_f[] = "We read an incomplete/null key from the datastore. This is likely a bug!";
static const char __pyx_k_https_datastore_googleapis_com_v[] = "https://datastore.googleapis.com/v1/projects/%s:runQuery";
static const char __pyx_k_https_www_googleapis_com_auth_da[] = "https://www.googleapis.com/auth/datastore";
static const char __pyx_k_src_viur_datastore_transport_pyx[] = "src/viur/datastore/transport.pyx";
static const char __pyx_k_https_datastore_googleapis_com_v_2[] = "https://datastore.googleapis.com/v1/projects/%s:lookup";
static const char __pyx_k_https_datastore_googleapis_com_v_3[] = "https://datastore.googleapis.com/v1/projects/%s:commit";
static const char __pyx_k_https_datastore_googleapis_com_v_4[] = "https://datastore.googleapis.com/v1/projects/%s:beginTransaction";
static const char __pyx_k_https_datastore_googleapis_com_v_5[] = "https://datastore.googleapis.com/v1/projects/%s:rollback";
static const char __pyx_k_https_datastore_googleapis_com_v_6[] = "https://datastore.googleapis.com/v1/projects/%s:allocateIds";
static const char __pyx_k_https_datastore_googleapis_com_v_7[] = "https://datastore.googleapis.com/v1/projects/%s:runAggregationQuery";
/* #### Code section: decls ### */
static PyObject *__pyx_pf_4viur_9datastore_9transport_authenticated_request(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_url, PyObject *__pyx_v_data); /* proto */
static PyObject *__pyx_pf_4viur_9datastore_9transport_2keyToPath(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_key); /* proto */
static PyObject *__pyx_pf_4viur_9datastore_9transport_4pythonPropToJson(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_v); /* proto */
static PyObject *__pyx_pf_4viur_9datastore_9transport_6runSingleFilter(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_queryDefinition, PyObject *__pyx_v_limit); /* proto */
static PyObject *__pyx_pf_4viur_9datastore_9transport_3Get_genexpr(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_genexpr_arg_0); /* proto */
static PyObject *__pyx_pf_4viur_9datastore_9transport_8Get(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_keys); /* proto */
static PyObject *__pyx_pf_4viur_9datastore_9transport_10Delete(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_keys); /* proto */
static PyObject *__pyx_pf_4viur_9datastore_9transport_12Put(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_entities); /* proto */
static PyObject *__pyx_pf_4viur_9datastore_9transport_14RunInTransaction(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_callback, PyObject *__pyx_v_args, PyObject *__pyx_v_kwargs); /* proto */
static PyObject *__pyx_pf_4viur_9datastore_9transport_16_rollbackTxn(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_txnKey); /* proto */
static PyObject *__pyx_pf_4viur_9datastore_9transport_18AllocateIDs(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_keys); /* proto */
static PyObject *__pyx_pf_4viur_9datastore_9transport_20Count(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_kind, PyObject *__pyx_v_up_to, PyObject *__pyx_v_queryDefinition); /* proto */
static PyObject *__pyx_tp_new_4viur_9datastore_9transport___pyx_scope_struct__genexpr(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static __Pyx_CachedCFunction __pyx_umethod_PyDict_Type_get = {0, 0, 0, 0, 0};
static __Pyx_CachedCFunction __pyx_umethod_PyDict_Type_pop = {0, 0, 0, 0, 0};
static __Pyx_CachedCFunction __pyx_umethod_PyDict_Type_update = {0, 0, 0, 0, 0};
/* #### Code section: late_includes ### */
/* #### Code section: module_state ### */
typedef struct {
  PyObject *__pyx_d;
  PyObject *__pyx_b;
  PyObject *__pyx_cython_runtime;
  PyObject *__pyx_empty_tuple;
  PyObject *__pyx_empty_bytes;
  PyObject *__pyx_empty_unicode;
  #ifdef __Pyx_CyFunction_USED
  PyTypeObject *__pyx_CyFunctionType;
  #endif
  #ifdef __Pyx_FusedFunction_USED
  PyTypeObject *__pyx_FusedFunctionType;
  #endif
  #ifdef __Pyx_Generator_USED
  PyTypeObject *__pyx_GeneratorType;
  #endif
  #ifdef __Pyx_IterableCoroutine_USED
  PyTypeObject *__pyx_IterableCoroutineType;
  #endif
  #ifdef __Pyx_Coroutine_USED
  PyTypeObject *__pyx_CoroutineAwaitType;
  #endif
  #ifdef __Pyx_Coroutine_USED
  PyTypeObject *__pyx_CoroutineType;
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  PyTypeObject *__pyx_ptype_7cpython_4type_type;
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  PyObject *__pyx_type_4viur_9datastore_9transport___pyx_scope_struct__genexpr;
  #endif
  PyTypeObject *__pyx_ptype_4viur_9datastore_9transport___pyx_scope_struct__genexpr;
  PyObject *__pyx_kp_u_;
  PyObject *__pyx_n_u_AND;
  PyObject *__pyx_n_u_ASCENDING;
  PyObject *__pyx_n_u_ASCII;
  PyObject *__pyx_n_s_AbortedError;
  PyObject *__pyx_kp_u_All_retries_are_exhausted_for_th;
  PyObject *__pyx_n_s_AllocateIDs;
  PyObject *__pyx_n_s_Any;
  PyObject *__pyx_n_s_AssertionError;
  PyObject *__pyx_n_s_AuthorizedSession;
  PyObject *__pyx_kp_u_Cannot_call_runInTransaction_whi;
  PyObject *__pyx_n_s_CollisionError;
  PyObject *__pyx_n_s_ConnectionError;
  PyObject *__pyx_n_s_Count;
  PyObject *__pyx_n_u_DESCENDING;
  PyObject *__pyx_n_s_Delete;
  PyObject *__pyx_n_u_EQUAL;
  PyObject *__pyx_kp_u_Empty_response_received_from_Dat;
  PyObject *__pyx_n_s_Entity;
  PyObject *__pyx_n_u_GREATER_THAN;
  PyObject *__pyx_n_u_GREATER_THAN_OR_EQUAL;
  PyObject *__pyx_n_s_Get;
  PyObject *__pyx_n_s_Get_locals_genexpr;
  PyObject *__pyx_kp_u_INVALID_RESPONSE_RECEIVED;
  PyObject *__pyx_n_s_InvalidArgumentError;
  PyObject *__pyx_kp_u_Invalid_data_received_from_Datas;
  PyObject *__pyx_kp_u_Invalid_key_in_entity_json_s;
  PyObject *__pyx_kp_u_Invalid_number_of_mutation_resul;
  PyObject *__pyx_kp_u_Invalid_op_s;
  PyObject *__pyx_n_s_Key;
  PyObject *__pyx_kp_u_Keys_must_not_be_partial_or_kind;
  PyObject *__pyx_n_u_LESS_THAN;
  PyObject *__pyx_n_u_LESS_THAN_OR_EQUAL;
  PyObject *__pyx_n_s_List;
  PyObject *__pyx_kp_s_List_Entity;
  PyObject *__pyx_kp_s_List_dict;
  PyObject *__pyx_n_u_NON_TRANSACTIONAL;
  PyObject *__pyx_n_u_NOT_FINISHED;
  PyObject *__pyx_n_u_NO_MORE_RESULTS;
  PyObject *__pyx_n_s_NoMutationResultsError;
  PyObject *__pyx_kp_u_No_mutation_results_received;
  PyObject *__pyx_kp_u_No_mutation_results_received_2;
  PyObject *__pyx_n_s_None;
  PyObject *__pyx_kp_u_None;
  PyObject *__pyx_kp_u_Please_create_a_matching_index;
  PyObject *__pyx_n_s_Put;
  PyObject *__pyx_kp_u_Queried;
  PyObject *__pyx_n_s_QueryDefinition;
  PyObject *__pyx_kp_u_Query_not_finished_Maybe_some_en;
  PyObject *__pyx_kp_u_Received_an_unexpected_key_updat;
  PyObject *__pyx_n_s_RecursionError;
  PyObject *__pyx_n_s_RequestsConnectionError;
  PyObject *__pyx_kp_u_Retrying_http_post_request_to_da;
  PyObject *__pyx_kp_u_Returned;
  PyObject *__pyx_n_s_RunInTransaction;
  PyObject *__pyx_n_u_STRONG;
  PyObject *__pyx_n_u_TRANSACTIONAL;
  PyObject *__pyx_kp_u_The_Count_aggregation_query_is_a;
  PyObject *__pyx_kp_u_UTF_8;
  PyObject *__pyx_n_s_Union;
  PyObject *__pyx_kp_s_Union_Entity_List_Entity;
  PyObject *__pyx_kp_s_Union_Key_List_Key;
  PyObject *__pyx_kp_s_Union_Key_List_Key_Entity_List_E;
  PyObject *__pyx_kp_s_Union_None_Entity_List_Entity;
  PyObject *__pyx_kp_u_Unknown_type;
  PyObject *__pyx_n_s_ValueError;
  PyObject *__pyx_n_s_ViurDatastoreError;
  PyObject *__pyx_kp_u_We_got_an_error_in_a_transaction;
  PyObject *__pyx_kp_u_We_read_an_incomplete_null_key_f;
  PyObject *__pyx_kp_u_Y_m_dT_H_M_S;
  PyObject *__pyx_kp_u_Y_m_dT_H_M_S_f;
  PyObject *__pyx_kp_u_Y_m_dT_H_M_S_fZ;
  PyObject *__pyx_kp_u__10;
  PyObject *__pyx_kp_u__11;
  PyObject *__pyx_n_s__21;
  PyObject *__pyx_kp_u__3;
  PyObject *__pyx_n_s__45;
  PyObject *__pyx_kp_u__5;
  PyObject *__pyx_kp_u__6;
  PyObject *__pyx_kp_u__7;
  PyObject *__pyx_kp_u__8;
  PyObject *__pyx_kp_u__9;
  PyObject *__pyx_n_s_accessLog;
  PyObject *__pyx_n_u_affectedEntities;
  PyObject *__pyx_n_s_affectedEntity;
  PyObject *__pyx_n_u_aggregateProperties;
  PyObject *__pyx_n_u_aggregationResults;
  PyObject *__pyx_n_u_aggregation_query;
  PyObject *__pyx_n_u_aggregations;
  PyObject *__pyx_n_u_allowOverriding;
  PyObject *__pyx_n_s_allowOverriding_2;
  PyObject *__pyx_kp_u_and_orders;
  PyObject *__pyx_n_s_append;
  PyObject *__pyx_n_s_args;
  PyObject *__pyx_n_s_arrayElem;
  PyObject *__pyx_n_s_arrayIt;
  PyObject *__pyx_n_u_arrayValue;
  PyObject *__pyx_n_s_array_element;
  PyObject *__pyx_n_s_array_element_inner;
  PyObject *__pyx_n_s_astimezone;
  PyObject *__pyx_n_s_asyncio_coroutines;
  PyObject *__pyx_n_s_auth;
  PyObject *__pyx_n_s_authenticated_request;
  PyObject *__pyx_n_s_b64decode;
  PyObject *__pyx_n_s_b64encode;
  PyObject *__pyx_n_s_base64;
  PyObject *__pyx_n_u_blobValue;
  PyObject *__pyx_n_u_booleanValue;
  PyObject *__pyx_n_s_bytes;
  PyObject *__pyx_n_s_cache;
  PyObject *__pyx_n_s_callable;
  PyObject *__pyx_n_s_callback;
  PyObject *__pyx_n_s_class_getitem;
  PyObject *__pyx_n_s_cline_in_traceback;
  PyObject *__pyx_n_s_close;
  PyObject *__pyx_n_u_compositeFilter;
  PyObject *__pyx_n_s_conf;
  PyObject *__pyx_n_s_content;
  PyObject *__pyx_n_s_copy;
  PyObject *__pyx_n_u_count;
  PyObject *__pyx_n_s_credentials;
  PyObject *__pyx_n_s_currentCursor;
  PyObject *__pyx_n_s_currentDbAccessLog;
  PyObject *__pyx_n_s_currentTransaction;
  PyObject *__pyx_n_s_currentTxn;
  PyObject *__pyx_n_s_data;
  PyObject *__pyx_n_s_data_ptr;
  PyObject *__pyx_n_s_datetime;
  PyObject *__pyx_n_s_debug;
  PyObject *__pyx_n_s_decode;
  PyObject *__pyx_n_s_default;
  PyObject *__pyx_n_s_deferred_key;
  PyObject *__pyx_n_s_deferred_keys;
  PyObject *__pyx_n_s_delete;
  PyObject *__pyx_n_u_delete;
  PyObject *__pyx_n_s_dict;
  PyObject *__pyx_n_s_dictKey;
  PyObject *__pyx_n_s_dictValue;
  PyObject *__pyx_n_u_direction;
  PyObject *__pyx_kp_u_disable;
  PyObject *__pyx_n_s_distinct;
  PyObject *__pyx_n_s_distinctKey;
  PyObject *__pyx_n_s_distinctOn;
  PyObject *__pyx_n_u_distinctOn;
  PyObject *__pyx_kp_u_distinct_on_s;
  PyObject *__pyx_n_u_doubleValue;
  PyObject *__pyx_n_s_dumps;
  PyObject *__pyx_n_s_element;
  PyObject *__pyx_n_s_element_inner;
  PyObject *__pyx_kp_u_enable;
  PyObject *__pyx_n_s_encode;
  PyObject *__pyx_n_s_endCursor;
  PyObject *__pyx_n_u_endCursor;
  PyObject *__pyx_n_s_entities;
  PyObject *__pyx_n_u_entityValue;
  PyObject *__pyx_n_s_enumerate;
  PyObject *__pyx_n_s_err;
  PyObject *__pyx_n_s_error;
  PyObject *__pyx_n_u_excludeFromIndexes;
  PyObject *__pyx_n_s_exclude_from_indexes;
  PyObject *__pyx_n_s_excludedProperties;
  PyObject *__pyx_n_s_exponential_backoff;
  PyObject *__pyx_n_s_extend;
  PyObject *__pyx_n_u_filter;
  PyObject *__pyx_n_s_filterList;
  PyObject *__pyx_n_s_filters;
  PyObject *__pyx_n_u_filters;
  PyObject *__pyx_n_s_flipResults;
  PyObject *__pyx_n_s_from_legacy_urlsafe;
  PyObject *__pyx_kp_u_gc;
  PyObject *__pyx_n_s_genexpr;
  PyObject *__pyx_n_s_get;
  PyObject *__pyx_n_s_google;
  PyObject *__pyx_n_s_google_auth;
  PyObject *__pyx_n_s_http_internal;
  PyObject *__pyx_kp_u_https_datastore_googleapis_com_v;
  PyObject *__pyx_kp_u_https_datastore_googleapis_com_v_2;
  PyObject *__pyx_kp_u_https_datastore_googleapis_com_v_3;
  PyObject *__pyx_kp_u_https_datastore_googleapis_com_v_4;
  PyObject *__pyx_kp_u_https_datastore_googleapis_com_v_5;
  PyObject *__pyx_kp_u_https_datastore_googleapis_com_v_6;
  PyObject *__pyx_kp_u_https_datastore_googleapis_com_v_7;
  PyObject *__pyx_kp_u_https_www_googleapis_com_auth_da;
  PyObject *__pyx_n_s_i;
  PyObject *__pyx_n_s_id;
  PyObject *__pyx_n_u_id;
  PyObject *__pyx_n_s_idx;
  PyObject *__pyx_n_s_import;
  PyObject *__pyx_n_s_initializing;
  PyObject *__pyx_n_s_innerArrayElem;
  PyObject *__pyx_n_s_insert;
  PyObject *__pyx_n_s_int;
  PyObject *__pyx_n_u_integerValue;
  PyObject *__pyx_n_s_internalStartCursor;
  PyObject *__pyx_n_s_isMulti;
  PyObject *__pyx_n_s_is_coroutine;
  PyObject *__pyx_kp_u_is_not_supported;
  PyObject *__pyx_n_s_is_partial;
  PyObject *__pyx_n_s_is_viur_datastore_request_ok;
  PyObject *__pyx_kp_u_isenabled;
  PyObject *__pyx_n_s_items;
  PyObject *__pyx_n_s_json;
  PyObject *__pyx_n_s_k;
  PyObject *__pyx_n_s_key;
  PyObject *__pyx_n_u_key;
  PyObject *__pyx_n_s_keyList;
  PyObject *__pyx_n_s_keyToPath;
  PyObject *__pyx_n_u_keyValue;
  PyObject *__pyx_n_s_keys;
  PyObject *__pyx_n_u_keys;
  PyObject *__pyx_n_s_keys_for_request;
  PyObject *__pyx_n_s_kind;
  PyObject *__pyx_n_u_kind;
  PyObject *__pyx_n_s_kwargs;
  PyObject *__pyx_n_s_limit;
  PyObject *__pyx_n_u_limit;
  PyObject *__pyx_n_s_loads;
  PyObject *__pyx_n_s_logging;
  PyObject *__pyx_n_s_main;
  PyObject *__pyx_n_u_memcache_client;
  PyObject *__pyx_n_s_missing_keys;
  PyObject *__pyx_n_s_missing_keys_paths;
  PyObject *__pyx_n_u_mode;
  PyObject *__pyx_n_u_mutations;
  PyObject *__pyx_n_s_name;
  PyObject *__pyx_n_u_name;
  PyObject *__pyx_n_s_name_2;
  PyObject *__pyx_n_u_nested_query;
  PyObject *__pyx_n_u_nullValue;
  PyObject *__pyx_n_s_oldTxn;
  PyObject *__pyx_n_s_op;
  PyObject *__pyx_n_u_op;
  PyObject *__pyx_n_u_order;
  PyObject *__pyx_n_s_orders;
  PyObject *__pyx_n_s_parent;
  PyObject *__pyx_n_s_parser;
  PyObject *__pyx_n_u_partitionId;
  PyObject *__pyx_n_u_path;
  PyObject *__pyx_n_s_pop;
  PyObject *__pyx_n_s_post;
  PyObject *__pyx_n_s_postData;
  PyObject *__pyx_n_s_post_data;
  PyObject *__pyx_n_u_previousTransaction;
  PyObject *__pyx_n_s_projectID;
  PyObject *__pyx_n_u_project_id;
  PyObject *__pyx_n_s_properties;
  PyObject *__pyx_n_u_properties;
  PyObject *__pyx_n_u_property;
  PyObject *__pyx_n_u_propertyFilter;
  PyObject *__pyx_n_u_property_1;
  PyObject *__pyx_n_s_put;
  PyObject *__pyx_n_s_pysize;
  PyObject *__pyx_n_s_pythonPropToJson;
  PyObject *__pyx_n_u_query;
  PyObject *__pyx_n_s_queryDefinition;
  PyObject *__pyx_n_s_range;
  PyObject *__pyx_n_u_readConsistency;
  PyObject *__pyx_n_s_readOptions;
  PyObject *__pyx_n_u_readOptions;
  PyObject *__pyx_n_u_readWrite;
  PyObject *__pyx_n_s_refresh_timeout;
  PyObject *__pyx_n_s_replace;
  PyObject *__pyx_n_s_requestedKeys;
  PyObject *__pyx_n_s_requested_keys;
  PyObject *__pyx_n_s_requests;
  PyObject *__pyx_kp_s_requests_Response;
  PyObject *__pyx_n_s_requests_exceptions;
  PyObject *__pyx_n_s_res;
  PyObject *__pyx_n_s_resDict;
  PyObject *__pyx_n_s_resList;
  PyObject *__pyx_n_s_res_from_cache;
  PyObject *__pyx_n_s_res_from_db;
  PyObject *__pyx_n_s_resp;
  PyObject *__pyx_kp_u_resp_content;
  PyObject *__pyx_kp_u_results;
  PyObject *__pyx_n_s_return;
  PyObject *__pyx_n_s_rollbackTxn;
  PyObject *__pyx_n_s_runSingleFilter;
  PyObject *__pyx_n_s_scopes;
  PyObject *__pyx_kp_u_seconds;
  PyObject *__pyx_n_s_send;
  PyObject *__pyx_n_s_set;
  PyObject *__pyx_n_s_singleValue;
  PyObject *__pyx_n_s_sleep;
  PyObject *__pyx_n_s_sleep_time;
  PyObject *__pyx_n_s_sortOrder;
  PyObject *__pyx_n_s_spec;
  PyObject *__pyx_n_s_split;
  PyObject *__pyx_kp_s_src_viur_datastore_transport_pyx;
  PyObject *__pyx_n_s_startCursor;
  PyObject *__pyx_n_u_startCursor;
  PyObject *__pyx_n_s_str;
  PyObject *__pyx_n_s_strftime;
  PyObject *__pyx_n_u_stringValue;
  PyObject *__pyx_n_s_strptime;
  PyObject *__pyx_n_s_test;
  PyObject *__pyx_n_s_throw;
  PyObject *__pyx_n_s_time;
  PyObject *__pyx_n_u_timestampValue;
  PyObject *__pyx_n_s_timezone;
  PyObject *__pyx_n_u_traceQueries;
  PyObject *__pyx_n_u_transaction;
  PyObject *__pyx_n_u_transactionOptions;
  PyObject *__pyx_n_s_transport;
  PyObject *__pyx_n_s_txnKey;
  PyObject *__pyx_n_s_txnOptions;
  PyObject *__pyx_n_s_typing;
  PyObject *__pyx_n_s_tzinfo;
  PyObject *__pyx_n_s_untouched_keys;
  PyObject *__pyx_n_s_up_to;
  PyObject *__pyx_n_u_up_to;
  PyObject *__pyx_n_s_update;
  PyObject *__pyx_n_u_upsert;
  PyObject *__pyx_n_s_url;
  PyObject *__pyx_n_s_utc;
  PyObject *__pyx_n_s_v;
  PyObject *__pyx_n_s_value;
  PyObject *__pyx_n_u_value;
  PyObject *__pyx_n_u_values;
  PyObject *__pyx_n_s_version;
  PyObject *__pyx_n_s_viur_datastore;
  PyObject *__pyx_n_s_viur_datastore_config;
  PyObject *__pyx_n_s_viur_datastore_errors;
  PyObject *__pyx_n_s_viur_datastore_transport;
  PyObject *__pyx_n_s_viur_datastore_types;
  PyObject *__pyx_n_s_warning;
  PyObject *__pyx_kp_u_with_filter;
  PyObject *__pyx_n_s_x;
  PyObject *__pyx_int_0;
  PyObject *__pyx_int_1;
  PyObject *__pyx_int_2;
  PyObject *__pyx_int_4;
  PyObject *__pyx_int_300;
  PyObject *__pyx_int_0x7fffffffffffffff;
  PyObject *__pyx_int_neg_1;
  PyObject *__pyx_slice__2;
  PyObject *__pyx_tuple__4;
  PyObject *__pyx_slice__12;
  PyObject *__pyx_slice__13;
  PyObject *__pyx_slice__14;
  PyObject *__pyx_tuple__15;
  PyObject *__pyx_tuple__16;
  PyObject *__pyx_tuple__17;
  PyObject *__pyx_tuple__18;
  PyObject *__pyx_tuple__19;
  PyObject *__pyx_tuple__20;
  PyObject *__pyx_tuple__22;
  PyObject *__pyx_tuple__24;
  PyObject *__pyx_tuple__26;
  PyObject *__pyx_tuple__28;
  PyObject *__pyx_tuple__30;
  PyObject *__pyx_tuple__32;
  PyObject *__pyx_tuple__34;
  PyObject *__pyx_tuple__36;
  PyObject *__pyx_tuple__38;
  PyObject *__pyx_tuple__40;
  PyObject *__pyx_tuple__42;
  PyObject *__pyx_tuple__44;
  PyObject *__pyx_codeobj__23;
  PyObject *__pyx_codeobj__25;
  PyObject *__pyx_codeobj__27;
  PyObject *__pyx_codeobj__29;
  PyObject *__pyx_codeobj__31;
  PyObject *__pyx_codeobj__33;
  PyObject *__pyx_codeobj__35;
  PyObject *__pyx_codeobj__37;
  PyObject *__pyx_codeobj__39;
  PyObject *__pyx_codeobj__41;
  PyObject *__pyx_codeobj__43;
} __pyx_mstate;

#if CYTHON_USE_MODULE_STATE
#ifdef __cplusplus
namespace {
  extern struct PyModuleDef __pyx_moduledef;
} /* anonymous namespace */
#else
static struct PyModuleDef __pyx_moduledef;
#endif

#define __pyx_mstate(o) ((__pyx_mstate *)__Pyx_PyModule_GetState(o))

#define __pyx_mstate_global (__pyx_mstate(PyState_FindModule(&__pyx_moduledef)))

#define __pyx_m (PyState_FindModule(&__pyx_moduledef))
#else
static __pyx_mstate __pyx_mstate_global_static =
#ifdef __cplusplus
    {};
#else
    {0};
#endif
static __pyx_mstate *__pyx_mstate_global = &__pyx_mstate_global_static;
#endif
/* #### Code section: module_state_clear ### */
#if CYTHON_USE_MODULE_STATE
static int __pyx_m_clear(PyObject *m) {
  __pyx_mstate *clear_module_state = __pyx_mstate(m);
  if (!clear_module_state) return 0;
  Py_CLEAR(clear_module_state->__pyx_d);
  Py_CLEAR(clear_module_state->__pyx_b);
  Py_CLEAR(clear_module_state->__pyx_cython_runtime);
  Py_CLEAR(clear_module_state->__pyx_empty_tuple);
  Py_CLEAR(clear_module_state->__pyx_empty_bytes);
  Py_CLEAR(clear_module_state->__pyx_empty_unicode);
  #ifdef __Pyx_CyFunction_USED
  Py_CLEAR(clear_module_state->__pyx_CyFunctionType);
  #endif
  #ifdef __Pyx_FusedFunction_USED
  Py_CLEAR(clear_module_state->__pyx_FusedFunctionType);
  #endif
  Py_CLEAR(clear_module_state->__pyx_ptype_7cpython_4type_type);
  Py_CLEAR(clear_module_state->__pyx_ptype_4viur_9datastore_9transport___pyx_scope_struct__genexpr);
  Py_CLEAR(clear_module_state->__pyx_type_4viur_9datastore_9transport___pyx_scope_struct__genexpr);
  Py_CLEAR(clear_module_state->__pyx_kp_u_);
  Py_CLEAR(clear_module_state->__pyx_n_u_AND);
  Py_CLEAR(clear_module_state->__pyx_n_u_ASCENDING);
  Py_CLEAR(clear_module_state->__pyx_n_u_ASCII);
  Py_CLEAR(clear_module_state->__pyx_n_s_AbortedError);
  Py_CLEAR(clear_module_state->__pyx_kp_u_All_retries_are_exhausted_for_th);
  Py_CLEAR(clear_module_state->__pyx_n_s_AllocateIDs);
  Py_CLEAR(clear_module_state->__pyx_n_s_Any);
  Py_CLEAR(clear_module_state->__pyx_n_s_AssertionError);
  Py_CLEAR(clear_module_state->__pyx_n_s_AuthorizedSession);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Cannot_call_runInTransaction_whi);
  Py_CLEAR(clear_module_state->__pyx_n_s_CollisionError);
  Py_CLEAR(clear_module_state->__pyx_n_s_ConnectionError);
  Py_CLEAR(clear_module_state->__pyx_n_s_Count);
  Py_CLEAR(clear_module_state->__pyx_n_u_DESCENDING);
  Py_CLEAR(clear_module_state->__pyx_n_s_Delete);
  Py_CLEAR(clear_module_state->__pyx_n_u_EQUAL);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Empty_response_received_from_Dat);
  Py_CLEAR(clear_module_state->__pyx_n_s_Entity);
  Py_CLEAR(clear_module_state->__pyx_n_u_GREATER_THAN);
  Py_CLEAR(clear_module_state->__pyx_n_u_GREATER_THAN_OR_EQUAL);
  Py_CLEAR(clear_module_state->__pyx_n_s_Get);
  Py_CLEAR(clear_module_state->__pyx_n_s_Get_locals_genexpr);
  Py_CLEAR(clear_module_state->__pyx_kp_u_INVALID_RESPONSE_RECEIVED);
  Py_CLEAR(clear_module_state->__pyx_n_s_InvalidArgumentError);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Invalid_data_received_from_Datas);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Invalid_key_in_entity_json_s);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Invalid_number_of_mutation_resul);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Invalid_op_s);
  Py_CLEAR(clear_module_state->__pyx_n_s_Key);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Keys_must_not_be_partial_or_kind);
  Py_CLEAR(clear_module_state->__pyx_n_u_LESS_THAN);
  Py_CLEAR(clear_module_state->__pyx_n_u_LESS_THAN_OR_EQUAL);
  Py_CLEAR(clear_module_state->__pyx_n_s_List);
  Py_CLEAR(clear_module_state->__pyx_kp_s_List_Entity);
  Py_CLEAR(clear_module_state->__pyx_kp_s_List_dict);
  Py_CLEAR(clear_module_state->__pyx_n_u_NON_TRANSACTIONAL);
  Py_CLEAR(clear_module_state->__pyx_n_u_NOT_FINISHED);
  Py_CLEAR(clear_module_state->__pyx_n_u_NO_MORE_RESULTS);
  Py_CLEAR(clear_module_state->__pyx_n_s_NoMutationResultsError);
  Py_CLEAR(clear_module_state->__pyx_kp_u_No_mutation_results_received);
  Py_CLEAR(clear_module_state->__pyx_kp_u_No_mutation_results_received_2);
  Py_CLEAR(clear_module_state->__pyx_n_s_None);
  Py_CLEAR(clear_module_state->__pyx_kp_u_None);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Please_create_a_matching_index);
  Py_CLEAR(clear_module_state->__pyx_n_s_Put);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Queried);
  Py_CLEAR(clear_module_state->__pyx_n_s_QueryDefinition);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Query_not_finished_Maybe_some_en);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Received_an_unexpected_key_updat);
  Py_CLEAR(clear_module_state->__pyx_n_s_RecursionError);
  Py_CLEAR(clear_module_state->__pyx_n_s_RequestsConnectionError);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Retrying_http_post_request_to_da);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Returned);
  Py_CLEAR(clear_module_state->__pyx_n_s_RunInTransaction);
  Py_CLEAR(clear_module_state->__pyx_n_u_STRONG);
  Py_CLEAR(clear_module_state->__pyx_n_u_TRANSACTIONAL);
  Py_CLEAR(clear_module_state->__pyx_kp_u_The_Count_aggregation_query_is_a);
  Py_CLEAR(clear_module_state->__pyx_kp_u_UTF_8);
  Py_CLEAR(clear_module_state->__pyx_n_s_Union);
  Py_CLEAR(clear_module_state->__pyx_kp_s_Union_Entity_List_Entity);
  Py_CLEAR(clear_module_state->__pyx_kp_s_Union_Key_List_Key);
  Py_CLEAR(clear_module_state->__pyx_kp_s_Union_Key_List_Key_Entity_List_E);
  Py_CLEAR(clear_module_state->__pyx_kp_s_Union_None_Entity_List_Entity);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Unknown_type);
  Py_CLEAR(clear_module_state->__pyx_n_s_ValueError);
  Py_CLEAR(clear_module_state->__pyx_n_s_ViurDatastoreError);
  Py_CLEAR(clear_module_state->__pyx_kp_u_We_got_an_error_in_a_transaction);
  Py_CLEAR(clear_module_state->__pyx_kp_u_We_read_an_incomplete_null_key_f);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Y_m_dT_H_M_S);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Y_m_dT_H_M_S_f);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Y_m_dT_H_M_S_fZ);
  Py_CLEAR(clear_module_state->__pyx_kp_u__10);
  Py_CLEAR(clear_module_state->__pyx_kp_u__11);
  Py_CLEAR(clear_module_state->__pyx_n_s__21);
  Py_CLEAR(clear_module_state->__pyx_kp_u__3);
  Py_CLEAR(clear_module_state->__pyx_n_s__45);
  Py_CLEAR(clear_module_state->__pyx_kp_u__5);
  Py_CLEAR(clear_module_state->__pyx_kp_u__6);
  Py_CLEAR(clear_module_state->__pyx_kp_u__7);
  Py_CLEAR(clear_module_state->__pyx_kp_u__8);
  Py_CLEAR(clear_module_state->__pyx_kp_u__9);
  Py_CLEAR(clear_module_state->__pyx_n_s_accessLog);
  Py_CLEAR(clear_module_state->__pyx_n_u_affectedEntities);
  Py_CLEAR(clear_module_state->__pyx_n_s_affectedEntity);
  Py_CLEAR(clear_module_state->__pyx_n_u_aggregateProperties);
  Py_CLEAR(clear_module_state->__pyx_n_u_aggregationResults);
  Py_CLEAR(clear_module_state->__pyx_n_u_aggregation_query);
  Py_CLEAR(clear_module_state->__pyx_n_u_aggregations);
  Py_CLEAR(clear_module_state->__pyx_n_u_allowOverriding);
  Py_CLEAR(clear_module_state->__pyx_n_s_allowOverriding_2);
  Py_CLEAR(clear_module_state->__pyx_kp_u_and_orders);
  Py_CLEAR(clear_module_state->__pyx_n_s_append);
  Py_CLEAR(clear_module_state->__pyx_n_s_args);
  Py_CLEAR(clear_module_state->__pyx_n_s_arrayElem);
  Py_CLEAR(clear_module_state->__pyx_n_s_arrayIt);
  Py_CLEAR(clear_module_state->__pyx_n_u_arrayValue);
  Py_CLEAR(clear_module_state->__pyx_n_s_array_element);
  Py_CLEAR(clear_module_state->__pyx_n_s_array_element_inner);
  Py_CLEAR(clear_module_state->__pyx_n_s_astimezone);
  Py_CLEAR(clear_module_state->__pyx_n_s_asyncio_coroutines);
  Py_CLEAR(clear_module_state->__pyx_n_s_auth);
  Py_CLEAR(clear_module_state->__pyx_n_s_authenticated_request);
  Py_CLEAR(clear_module_state->__pyx_n_s_b64decode);
  Py_CLEAR(clear_module_state->__pyx_n_s_b64encode);
  Py_CLEAR(clear_module_state->__pyx_n_s_base64);
  Py_CLEAR(clear_module_state->__pyx_n_u_blobValue);
  Py_CLEAR(clear_module_state->__pyx_n_u_booleanValue);
  Py_CLEAR(clear_module_state->__pyx_n_s_bytes);
  Py_CLEAR(clear_module_state->__pyx_n_s_cache);
  Py_CLEAR(clear_module_state->__pyx_n_s_callable);
  Py_CLEAR(clear_module_state->__pyx_n_s_callback);
  Py_CLEAR(clear_module_state->__pyx_n_s_class_getitem);
  Py_CLEAR(clear_module_state->__pyx_n_s_cline_in_traceback);
  Py_CLEAR(clear_module_state->__pyx_n_s_close);
  Py_CLEAR(clear_module_state->__pyx_n_u_compositeFilter);
  Py_CLEAR(clear_module_state->__pyx_n_s_conf);
  Py_CLEAR(clear_module_state->__pyx_n_s_content);
  Py_CLEAR(clear_module_state->__pyx_n_s_copy);
  Py_CLEAR(clear_module_state->__pyx_n_u_count);
  Py_CLEAR(clear_module_state->__pyx_n_s_credentials);
  Py_CLEAR(clear_module_state->__pyx_n_s_currentCursor);
  Py_CLEAR(clear_module_state->__pyx_n_s_currentDbAccessLog);
  Py_CLEAR(clear_module_state->__pyx_n_s_currentTransaction);
  Py_CLEAR(clear_module_state->__pyx_n_s_currentTxn);
  Py_CLEAR(clear_module_state->__pyx_n_s_data);
  Py_CLEAR(clear_module_state->__pyx_n_s_data_ptr);
  Py_CLEAR(clear_module_state->__pyx_n_s_datetime);
  Py_CLEAR(clear_module_state->__pyx_n_s_debug);
  Py_CLEAR(clear_module_state->__pyx_n_s_decode);
  Py_CLEAR(clear_module_state->__pyx_n_s_default);
  Py_CLEAR(clear_module_state->__pyx_n_s_deferred_key);
  Py_CLEAR(clear_module_state->__pyx_n_s_deferred_keys);
  Py_CLEAR(clear_module_state->__pyx_n_s_delete);
  Py_CLEAR(clear_module_state->__pyx_n_u_delete);
  Py_CLEAR(clear_module_state->__pyx_n_s_dict);
  Py_CLEAR(clear_module_state->__pyx_n_s_dictKey);
  Py_CLEAR(clear_module_state->__pyx_n_s_dictValue);
  Py_CLEAR(clear_module_state->__pyx_n_u_direction);
  Py_CLEAR(clear_module_state->__pyx_kp_u_disable);
  Py_CLEAR(clear_module_state->__pyx_n_s_distinct);
  Py_CLEAR(clear_module_state->__pyx_n_s_distinctKey);
  Py_CLEAR(clear_module_state->__pyx_n_s_distinctOn);
  Py_CLEAR(clear_module_state->__pyx_n_u_distinctOn);
  Py_CLEAR(clear_module_state->__pyx_kp_u_distinct_on_s);
  Py_CLEAR(clear_module_state->__pyx_n_u_doubleValue);
  Py_CLEAR(clear_module_state->__pyx_n_s_dumps);
  Py_CLEAR(clear_module_state->__pyx_n_s_element);
  Py_CLEAR(clear_module_state->__pyx_n_s_element_inner);
  Py_CLEAR(clear_module_state->__pyx_kp_u_enable);
  Py_CLEAR(clear_module_state->__pyx_n_s_encode);
  Py_CLEAR(clear_module_state->__pyx_n_s_endCursor);
  Py_CLEAR(clear_module_state->__pyx_n_u_endCursor);
  Py_CLEAR(clear_module_state->__pyx_n_s_entities);
  Py_CLEAR(clear_module_state->__pyx_n_u_entityValue);
  Py_CLEAR(clear_module_state->__pyx_n_s_enumerate);
  Py_CLEAR(clear_module_state->__pyx_n_s_err);
  Py_CLEAR(clear_module_state->__pyx_n_s_error);
  Py_CLEAR(clear_module_state->__pyx_n_u_excludeFromIndexes);
  Py_CLEAR(clear_module_state->__pyx_n_s_exclude_from_indexes);
  Py_CLEAR(clear_module_state->__pyx_n_s_excludedProperties);
  Py_CLEAR(clear_module_state->__pyx_n_s_exponential_backoff);
  Py_CLEAR(clear_module_state->__pyx_n_s_extend);
  Py_CLEAR(clear_module_state->__pyx_n_u_filter);
  Py_CLEAR(clear_module_state->__pyx_n_s_filterList);
  Py_CLEAR(clear_module_state->__pyx_n_s_filters);
  Py_CLEAR(clear_module_state->__pyx_n_u_filters);
  Py_CLEAR(clear_module_state->__pyx_n_s_flipResults);
  Py_CLEAR(clear_module_state->__pyx_n_s_from_legacy_urlsafe);
  Py_CLEAR(clear_module_state->__pyx_kp_u_gc);
  Py_CLEAR(clear_module_state->__pyx_n_s_genexpr);
  Py_CLEAR(clear_module_state->__pyx_n_s_get);
  Py_CLEAR(clear_module_state->__pyx_n_s_google);
  Py_CLEAR(clear_module_state->__pyx_n_s_google_auth);
  Py_CLEAR(clear_module_state->__pyx_n_s_http_internal);
  Py_CLEAR(clear_module_state->__pyx_kp_u_https_datastore_googleapis_com_v);
  Py_CLEAR(clear_module_state->__pyx_kp_u_https_datastore_googleapis_com_v_2);
  Py_CLEAR(clear_module_state->__pyx_kp_u_https_datastore_googleapis_com_v_3);
  Py_CLEAR(clear_module_state->__pyx_kp_u_https_datastore_googleapis_com_v_4);
  Py_CLEAR(clear_module_state->__pyx_kp_u_https_datastore_googleapis_com_v_5);
  Py_CLEAR(clear_module_state->__pyx_kp_u_https_datastore_googleapis_com_v_6);
  Py_CLEAR(clear_module_state->__pyx_kp_u_https_datastore_googleapis_com_v_7);
  Py_CLEAR(clear_module_state->__pyx_kp_u_https_www_googleapis_com_auth_da);
  Py_CLEAR(clear_module_state->__pyx_n_s_i);
  Py_CLEAR(clear_module_state->__pyx_n_s_id);
  Py_CLEAR(clear_module_state->__pyx_n_u_id);
  Py_CLEAR(clear_module_state->__pyx_n_s_idx);
  Py_CLEAR(clear_module_state->__pyx_n_s_import);
  Py_CLEAR(clear_module_state->__pyx_n_s_initializing);
  Py_CLEAR(clear_module_state->__pyx_n_s_innerArrayElem);
  Py_CLEAR(clear_module_state->__pyx_n_s_insert);
  Py_CLEAR(clear_module_state->__pyx_n_s_int);
  Py_CLEAR(clear_module_state->__pyx_n_u_integerValue);
  Py_CLEAR(clear_module_state->__pyx_n_s_internalStartCursor);
  Py_CLEAR(clear_module_state->__pyx_n_s_isMulti);
  Py_CLEAR(clear_module_state->__pyx_n_s_is_coroutine);
  Py_CLEAR(clear_module_state->__pyx_kp_u_is_not_supported);
  Py_CLEAR(clear_module_state->__pyx_n_s_is_partial);
  Py_CLEAR(clear_module_state->__pyx_n_s_is_viur_datastore_request_ok);
  Py_CLEAR(clear_module_state->__pyx_kp_u_isenabled);
  Py_CLEAR(clear_module_state->__pyx_n_s_items);
  Py_CLEAR(clear_module_state->__pyx_n_s_json);
  Py_CLEAR(clear_module_state->__pyx_n_s_k);
  Py_CLEAR(clear_module_state->__pyx_n_s_key);
  Py_CLEAR(clear_module_state->__pyx_n_u_key);
  Py_CLEAR(clear_module_state->__pyx_n_s_keyList);
  Py_CLEAR(clear_module_state->__pyx_n_s_keyToPath);
  Py_CLEAR(clear_module_state->__pyx_n_u_keyValue);
  Py_CLEAR(clear_module_state->__pyx_n_s_keys);
  Py_CLEAR(clear_module_state->__pyx_n_u_keys);
  Py_CLEAR(clear_module_state->__pyx_n_s_keys_for_request);
  Py_CLEAR(clear_module_state->__pyx_n_s_kind);
  Py_CLEAR(clear_module_state->__pyx_n_u_kind);
  Py_CLEAR(clear_module_state->__pyx_n_s_kwargs);
  Py_CLEAR(clear_module_state->__pyx_n_s_limit);
  Py_CLEAR(clear_module_state->__pyx_n_u_limit);
  Py_CLEAR(clear_module_state->__pyx_n_s_loads);
  Py_CLEAR(clear_module_state->__pyx_n_s_logging);
  Py_CLEAR(clear_module_state->__pyx_n_s_main);
  Py_CLEAR(clear_module_state->__pyx_n_u_memcache_client);
  Py_CLEAR(clear_module_state->__pyx_n_s_missing_keys);
  Py_CLEAR(clear_module_state->__pyx_n_s_missing_keys_paths);
  Py_CLEAR(clear_module_state->__pyx_n_u_mode);
  Py_CLEAR(clear_module_state->__pyx_n_u_mutations);
  Py_CLEAR(clear_module_state->__pyx_n_s_name);
  Py_CLEAR(clear_module_state->__pyx_n_u_name);
  Py_CLEAR(clear_module_state->__pyx_n_s_name_2);
  Py_CLEAR(clear_module_state->__pyx_n_u_nested_query);
  Py_CLEAR(clear_module_state->__pyx_n_u_nullValue);
  Py_CLEAR(clear_module_state->__pyx_n_s_oldTxn);
  Py_CLEAR(clear_module_state->__pyx_n_s_op);
  Py_CLEAR(clear_module_state->__pyx_n_u_op);
  Py_CLEAR(clear_module_state->__pyx_n_u_order);
  Py_CLEAR(clear_module_state->__pyx_n_s_orders);
  Py_CLEAR(clear_module_state->__pyx_n_s_parent);
  Py_CLEAR(clear_module_state->__pyx_n_s_parser);
  Py_CLEAR(clear_module_state->__pyx_n_u_partitionId);
  Py_CLEAR(clear_module_state->__pyx_n_u_path);
  Py_CLEAR(clear_module_state->__pyx_n_s_pop);
  Py_CLEAR(clear_module_state->__pyx_n_s_post);
  Py_CLEAR(clear_module_state->__pyx_n_s_postData);
  Py_CLEAR(clear_module_state->__pyx_n_s_post_data);
  Py_CLEAR(clear_module_state->__pyx_n_u_previousTransaction);
  Py_CLEAR(clear_module_state->__pyx_n_s_projectID);
  Py_CLEAR(clear_module_state->__pyx_n_u_project_id);
  Py_CLEAR(clear_module_state->__pyx_n_s_properties);
  Py_CLEAR(clear_module_state->__pyx_n_u_properties);
  Py_CLEAR(clear_module_state->__pyx_n_u_property);
  Py_CLEAR(clear_module_state->__pyx_n_u_propertyFilter);
  Py_CLEAR(clear_module_state->__pyx_n_u_property_1);
  Py_CLEAR(clear_module_state->__pyx_n_s_put);
  Py_CLEAR(clear_module_state->__pyx_n_s_pysize);
  Py_CLEAR(clear_module_state->__pyx_n_s_pythonPropToJson);
  Py_CLEAR(clear_module_state->__pyx_n_u_query);
  Py_CLEAR(clear_module_state->__pyx_n_s_queryDefinition);
  Py_CLEAR(clear_module_state->__pyx_n_s_range);
  Py_CLEAR(clear_module_state->__pyx_n_u_readConsistency);
  Py_CLEAR(clear_module_state->__pyx_n_s_readOptions);
  Py_CLEAR(clear_module_state->__pyx_n_u_readOptions);
  Py_CLEAR(clear_module_state->__pyx_n_u_readWrite);
  Py_CLEAR(clear_module_state->__pyx_n_s_refresh_timeout);
  Py_CLEAR(clear_module_state->__pyx_n_s_replace);
  Py_CLEAR(clear_module_state->__pyx_n_s_requestedKeys);
  Py_CLEAR(clear_module_state->__pyx_n_s_requested_keys);
  Py_CLEAR(clear_module_state->__pyx_n_s_requests);
  Py_CLEAR(clear_module_state->__pyx_kp_s_requests_Response);
  Py_CLEAR(clear_module_state->__pyx_n_s_requests_exceptions);
  Py_CLEAR(clear_module_state->__pyx_n_s_res);
  Py_CLEAR(clear_module_state->__pyx_n_s_resDict);
  Py_CLEAR(clear_module_state->__pyx_n_s_resList);
  Py_CLEAR(clear_module_state->__pyx_n_s_res_from_cache);
  Py_CLEAR(clear_module_state->__pyx_n_s_res_from_db);
  Py_CLEAR(clear_module_state->__pyx_n_s_resp);
  Py_CLEAR(clear_module_state->__pyx_kp_u_resp_content);
  Py_CLEAR(clear_module_state->__pyx_kp_u_results);
  Py_CLEAR(clear_module_state->__pyx_n_s_return);
  Py_CLEAR(clear_module_state->__pyx_n_s_rollbackTxn);
  Py_CLEAR(clear_module_state->__pyx_n_s_runSingleFilter);
  Py_CLEAR(clear_module_state->__pyx_n_s_scopes);
  Py_CLEAR(clear_module_state->__pyx_kp_u_seconds);
  Py_CLEAR(clear_module_state->__pyx_n_s_send);
  Py_CLEAR(clear_module_state->__pyx_n_s_set);
  Py_CLEAR(clear_module_state->__pyx_n_s_singleValue);
  Py_CLEAR(clear_module_state->__pyx_n_s_sleep);
  Py_CLEAR(clear_module_state->__pyx_n_s_sleep_time);
  Py_CLEAR(clear_module_state->__pyx_n_s_sortOrder);
  Py_CLEAR(clear_module_state->__pyx_n_s_spec);
  Py_CLEAR(clear_module_state->__pyx_n_s_split);
  Py_CLEAR(clear_module_state->__pyx_kp_s_src_viur_datastore_transport_pyx);
  Py_CLEAR(clear_module_state->__pyx_n_s_startCursor);
  Py_CLEAR(clear_module_state->__pyx_n_u_startCursor);
  Py_CLEAR(clear_module_state->__pyx_n_s_str);
  Py_CLEAR(clear_module_state->__pyx_n_s_strftime);
  Py_CLEAR(clear_module_state->__pyx_n_u_stringValue);
  Py_CLEAR(clear_module_state->__pyx_n_s_strptime);
  Py_CLEAR(clear_module_state->__pyx_n_s_test);
  Py_CLEAR(clear_module_state->__pyx_n_s_throw);
  Py_CLEAR(clear_module_state->__pyx_n_s_time);
  Py_CLEAR(clear_module_state->__pyx_n_u_timestampValue);
  Py_CLEAR(clear_module_state->__pyx_n_s_timezone);
  Py_CLEAR(clear_module_state->__pyx_n_u_traceQueries);
  Py_CLEAR(clear_module_state->__pyx_n_u_transaction);
  Py_CLEAR(clear_module_state->__pyx_n_u_transactionOptions);
  Py_CLEAR(clear_module_state->__pyx_n_s_transport);
  Py_CLEAR(clear_module_state->__pyx_n_s_txnKey);
  Py_CLEAR(clear_module_state->__pyx_n_s_txnOptions);
  Py_CLEAR(clear_module_state->__pyx_n_s_typing);
  Py_CLEAR(clear_module_state->__pyx_n_s_tzinfo);
  Py_CLEAR(clear_module_state->__pyx_n_s_untouched_keys);
  Py_CLEAR(clear_module_state->__pyx_n_s_up_to);
  Py_CLEAR(clear_module_state->__pyx_n_u_up_to);
  Py_CLEAR(clear_module_state->__pyx_n_s_update);
  Py_CLEAR(clear_module_state->__pyx_n_u_upsert);
  Py_CLEAR(clear_module_state->__pyx_n_s_url);
  Py_CLEAR(clear_module_state->__pyx_n_s_utc);
  Py_CLEAR(clear_module_state->__pyx_n_s_v);
  Py_CLEAR(clear_module_state->__pyx_n_s_value);
  Py_CLEAR(clear_module_state->__pyx_n_u_value);
  Py_CLEAR(clear_module_state->__pyx_n_u_values);
  Py_CLEAR(clear_module_state->__pyx_n_s_version);
  Py_CLEAR(clear_module_state->__pyx_n_s_viur_datastore);
  Py_CLEAR(clear_module_state->__pyx_n_s_viur_datastore_config);
  Py_CLEAR(clear_module_state->__pyx_n_s_viur_datastore_errors);
  Py_CLEAR(clear_module_state->__pyx_n_s_viur_datastore_transport);
  Py_CLEAR(clear_module_state->__pyx_n_s_viur_datastore_types);
  Py_CLEAR(clear_module_state->__pyx_n_s_warning);
  Py_CLEAR(clear_module_state->__pyx_kp_u_with_filter);
  Py_CLEAR(clear_module_state->__pyx_n_s_x);
  Py_CLEAR(clear_module_state->__pyx_int_0);
  Py_CLEAR(clear_module_state->__pyx_int_1);
  Py_CLEAR(clear_module_state->__pyx_int_2);
  Py_CLEAR(clear_module_state->__pyx_int_4);
  Py_CLEAR(clear_module_state->__pyx_int_300);
  Py_CLEAR(clear_module_state->__pyx_int_0x7fffffffffffffff);
  Py_CLEAR(clear_module_state->__pyx_int_neg_1);
  Py_CLEAR(clear_module_state->__pyx_slice__2);
  Py_CLEAR(clear_module_state->__pyx_tuple__4);
  Py_CLEAR(clear_module_state->__pyx_slice__12);
  Py_CLEAR(clear_module_state->__pyx_slice__13);
  Py_CLEAR(clear_module_state->__pyx_slice__14);
  Py_CLEAR(clear_module_state->__pyx_tuple__15);
  Py_CLEAR(clear_module_state->__pyx_tuple__16);
  Py_CLEAR(clear_module_state->__pyx_tuple__17);
  Py_CLEAR(clear_module_state->__pyx_tuple__18);
  Py_CLEAR(clear_module_state->__pyx_tuple__19);
  Py_CLEAR(clear_module_state->__pyx_tuple__20);
  Py_CLEAR(clear_module_state->__pyx_tuple__22);
  Py_CLEAR(clear_module_state->__pyx_tuple__24);
  Py_CLEAR(clear_module_state->__pyx_tuple__26);
  Py_CLEAR(clear_module_state->__pyx_tuple__28);
  Py_CLEAR(clear_module_state->__pyx_tuple__30);
  Py_CLEAR(clear_module_state->__pyx_tuple__32);
  Py_CLEAR(clear_module_state->__pyx_tuple__34);
  Py_CLEAR(clear_module_state->__pyx_tuple__36);
  Py_CLEAR(clear_module_state->__pyx_tuple__38);
  Py_CLEAR(clear_module_state->__pyx_tuple__40);
  Py_CLEAR(clear_module_state->__pyx_tuple__42);
  Py_CLEAR(clear_module_state->__pyx_tuple__44);
  Py_CLEAR(clear_module_state->__pyx_codeobj__23);
  Py_CLEAR(clear_module_state->__pyx_codeobj__25);
  Py_CLEAR(clear_module_state->__pyx_codeobj__27);
  Py_CLEAR(clear_module_state->__pyx_codeobj__29);
  Py_CLEAR(clear_module_state->__pyx_codeobj__31);
  Py_CLEAR(clear_module_state->__pyx_codeobj__33);
  Py_CLEAR(clear_module_state->__pyx_codeobj__35);
  Py_CLEAR(clear_module_state->__pyx_codeobj__37);
  Py_CLEAR(clear_module_state->__pyx_codeobj__39);
  Py_CLEAR(clear_module_state->__pyx_codeobj__41);
  Py_CLEAR(clear_module_state->__pyx_codeobj__43);
  return 0;
}
#endif
/* #### Code section: module_state_traverse ### */
#if CYTHON_USE_MODULE_STATE
static int __pyx_m_traverse(PyObject *m, visitproc visit, void *arg) {
  __pyx_mstate *traverse_module_state = __pyx_mstate(m);
  if (!traverse_module_state) return 0;
  Py_VISIT(traverse_module_state->__pyx_d);
  Py_VISIT(traverse_module_state->__pyx_b);
  Py_VISIT(traverse_module_state->__pyx_cython_runtime);
  Py_VISIT(traverse_module_state->__pyx_empty_tuple);
  Py_VISIT(traverse_module_state->__pyx_empty_bytes);
  Py_VISIT(traverse_module_state->__pyx_empty_unicode);
  #ifdef __Pyx_CyFunction_USED
  Py_VISIT(traverse_module_state->__pyx_CyFunctionType);
  #endif
  #ifdef __Pyx_FusedFunction_USED
  Py_VISIT(traverse_module_state->__pyx_FusedFunctionType);
  #endif
  Py_VISIT(traverse_module_state->__pyx_ptype_7cpython_4type_type);
  Py_VISIT(traverse_module_state->__pyx_ptype_4viur_9datastore_9transport___pyx_scope_struct__genexpr);
  Py_VISIT(traverse_module_state->__pyx_type_4viur_9datastore_9transport___pyx_scope_struct__genexpr);
  Py_VISIT(traverse_module_state->__pyx_kp_u_);
  Py_VISIT(traverse_module_state->__pyx_n_u_AND);
  Py_VISIT(traverse_module_state->__pyx_n_u_ASCENDING);
  Py_VISIT(traverse_module_state->__pyx_n_u_ASCII);
  Py_VISIT(traverse_module_state->__pyx_n_s_AbortedError);
  Py_VISIT(traverse_module_state->__pyx_kp_u_All_retries_are_exhausted_for_th);
  Py_VISIT(traverse_module_state->__pyx_n_s_AllocateIDs);
  Py_VISIT(traverse_module_state->__pyx_n_s_Any);
  Py_VISIT(traverse_module_state->__pyx_n_s_AssertionError);
  Py_VISIT(traverse_module_state->__pyx_n_s_AuthorizedSession);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Cannot_call_runInTransaction_whi);
  Py_VISIT(traverse_module_state->__pyx_n_s_CollisionError);
  Py_VISIT(traverse_module_state->__pyx_n_s_ConnectionError);
  Py_VISIT(traverse_module_state->__pyx_n_s_Count);
  Py_VISIT(traverse_module_state->__pyx_n_u_DESCENDING);
  Py_VISIT(traverse_module_state->__pyx_n_s_Delete);
  Py_VISIT(traverse_module_state->__pyx_n_u_EQUAL);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Empty_response_received_from_Dat);
  Py_VISIT(traverse_module_state->__pyx_n_s_Entity);
  Py_VISIT(traverse_module_state->__pyx_n_u_GREATER_THAN);
  Py_VISIT(traverse_module_state->__pyx_n_u_GREATER_THAN_OR_EQUAL);
  Py_VISIT(traverse_module_state->__pyx_n_s_Get);
  Py_VISIT(traverse_module_state->__pyx_n_s_Get_locals_genexpr);
  Py_VISIT(traverse_module_state->__pyx_kp_u_INVALID_RESPONSE_RECEIVED);
  Py_VISIT(traverse_module_state->__pyx_n_s_InvalidArgumentError);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Invalid_data_received_from_Datas);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Invalid_key_in_entity_json_s);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Invalid_number_of_mutation_resul);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Invalid_op_s);
  Py_VISIT(traverse_module_state->__pyx_n_s_Key);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Keys_must_not_be_partial_or_kind);
  Py_VISIT(traverse_module_state->__pyx_n_u_LESS_THAN);
  Py_VISIT(traverse_module_state->__pyx_n_u_LESS_THAN_OR_EQUAL);
  Py_VISIT(traverse_module_state->__pyx_n_s_List);
  Py_VISIT(traverse_module_state->__pyx_kp_s_List_Entity);
  Py_VISIT(traverse_module_state->__pyx_kp_s_List_dict);
  Py_VISIT(traverse_module_state->__pyx_n_u_NON_TRANSACTIONAL);
  Py_VISIT(traverse_module_state->__pyx_n_u_NOT_FINISHED);
  Py_VISIT(traverse_module_state->__pyx_n_u_NO_MORE_RESULTS);
  Py_VISIT(traverse_module_state->__pyx_n_s_NoMutationResultsError);
  Py_VISIT(traverse_module_state->__pyx_kp_u_No_mutation_results_received);
  Py_VISIT(traverse_module_state->__pyx_kp_u_No_mutation_results_received_2);
  Py_VISIT(traverse_module_state->__pyx_n_s_None);
  Py_VISIT(traverse_module_state->__pyx_kp_u_None);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Please_create_a_matching_index);
  Py_VISIT(traverse_module_state->__pyx_n_s_Put);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Queried);
  Py_VISIT(traverse_module_state->__pyx_n_s_QueryDefinition);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Query_not_finished_Maybe_some_en);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Received_an_unexpected_key_updat);
  Py_VISIT(traverse_module_state->__pyx_n_s_RecursionError);
  Py_VISIT(traverse_module_state->__pyx_n_s_RequestsConnectionError);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Retrying_http_post_request_to_da);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Returned);
  Py_VISIT(traverse_module_state->__pyx_n_s_RunInTransaction);
  Py_VISIT(traverse_module_state->__pyx_n_u_STRONG);
  Py_VISIT(traverse_module_state->__pyx_n_u_TRANSACTIONAL);
  Py_VISIT(traverse_module_state->__pyx_kp_u_The_Count_aggregation_query_is_a);
  Py_VISIT(traverse_module_state->__pyx_kp_u_UTF_8);
  Py_VISIT(traverse_module_state->__pyx_n_s_Union);
  Py_VISIT(traverse_module_state->__pyx_kp_s_Union_Entity_List_Entity);
  Py_VISIT(traverse_module_state->__pyx_kp_s_Union_Key_List_Key);
  Py_VISIT(traverse_module_state->__pyx_kp_s_Union_Key_List_Key_Entity_List_E);
  Py_VISIT(traverse_module_state->__pyx_kp_s_Union_None_Entity_List_Entity);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Unknown_type);
  Py_VISIT(traverse_module_state->__pyx_n_s_ValueError);
  Py_VISIT(traverse_module_state->__pyx_n_s_ViurDatastoreError);
  Py_VISIT(traverse_module_state->__pyx_kp_u_We_got_an_error_in_a_transaction);
  Py_VISIT(traverse_module_state->__pyx_kp_u_We_read_an_incomplete_null_key_f);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Y_m_dT_H_M_S);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Y_m_dT_H_M_S_f);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Y_m_dT_H_M_S_fZ);
  Py_VISIT(traverse_module_state->__pyx_kp_u__10);
  Py_VISIT(traverse_module_state->__pyx_kp_u__11);
  Py_VISIT(traverse_module_state->__pyx_n_s__21);
  Py_VISIT(traverse_module_state->__pyx_kp_u__3);
  Py_VISIT(traverse_module_state->__pyx_n_s__45);
  Py_VISIT(traverse_module_state->__pyx_kp_u__5);
  Py_VISIT(traverse_module_state->__pyx_kp_u__6);
  Py_VISIT(traverse_module_state->__pyx_kp_u__7);
  Py_VISIT(traverse_module_state->__pyx_kp_u__8);
  Py_VISIT(traverse_module_state->__pyx_kp_u__9);
  Py_VISIT(traverse_module_state->__pyx_n_s_accessLog);
  Py_VISIT(traverse_module_state->__pyx_n_u_affectedEntities);
  Py_VISIT(traverse_module_state->__pyx_n_s_affectedEntity);
  Py_VISIT(traverse_module_state->__pyx_n_u_aggregateProperties);
  Py_VISIT(traverse_module_state->__pyx_n_u_aggregationResults);
  Py_VISIT(traverse_module_state->__pyx_n_u_aggregation_query);
  Py_VISIT(traverse_module_state->__pyx_n_u_aggregations);
  Py_VISIT(traverse_module_state->__pyx_n_u_allowOverriding);
  Py_VISIT(traverse_module_state->__pyx_n_s_allowOverriding_2);
  Py_VISIT(traverse_module_state->__pyx_kp_u_and_orders);
  Py_VISIT(traverse_module_state->__pyx_n_s_append);
  Py_VISIT(traverse_module_state->__pyx_n_s_args);
  Py_VISIT(traverse_module_state->__pyx_n_s_arrayElem);
  Py_VISIT(traverse_module_state->__pyx_n_s_arrayIt);
  Py_VISIT(traverse_module_state->__pyx_n_u_arrayValue);
  Py_VISIT(traverse_module_state->__pyx_n_s_array_element);
  Py_VISIT(traverse_module_state->__pyx_n_s_array_element_inner);
  Py_VISIT(traverse_module_state->__pyx_n_s_astimezone);
  Py_VISIT(traverse_module_state->__pyx_n_s_asyncio_coroutines);
  Py_VISIT(traverse_module_state->__pyx_n_s_auth);
  Py_VISIT(traverse_module_state->__pyx_n_s_authenticated_request);
  Py_VISIT(traverse_module_state->__pyx_n_s_b64decode);
  Py_VISIT(traverse_module_state->__pyx_n_s_b64encode);
  Py_VISIT(traverse_module_state->__pyx_n_s_base64);
  Py_VISIT(traverse_module_state->__pyx_n_u_blobValue);
  Py_VISIT(traverse_module_state->__pyx_n_u_booleanValue);
  Py_VISIT(traverse_module_state->__pyx_n_s_bytes);
  Py_VISIT(traverse_module_state->__pyx_n_s_cache);
  Py_VISIT(traverse_module_state->__pyx_n_s_callable);
  Py_VISIT(traverse_module_state->__pyx_n_s_callback);
  Py_VISIT(traverse_module_state->__pyx_n_s_class_getitem);
  Py_VISIT(traverse_module_state->__pyx_n_s_cline_in_traceback);
  Py_VISIT(traverse_module_state->__pyx_n_s_close);
  Py_VISIT(traverse_module_state->__pyx_n_u_compositeFilter);
  Py_VISIT(traverse_module_state->__pyx_n_s_conf);
  Py_VISIT(traverse_module_state->__pyx_n_s_content);
  Py_VISIT(traverse_module_state->__pyx_n_s_copy);
  Py_VISIT(traverse_module_state->__pyx_n_u_count);
  Py_VISIT(traverse_module_state->__pyx_n_s_credentials);
  Py_VISIT(traverse_module_state->__pyx_n_s_currentCursor);
  Py_VISIT(traverse_module_state->__pyx_n_s_currentDbAccessLog);
  Py_VISIT(traverse_module_state->__pyx_n_s_currentTransaction);
  Py_VISIT(traverse_module_state->__pyx_n_s_currentTxn);
  Py_VISIT(traverse_module_state->__pyx_n_s_data);
  Py_VISIT(traverse_module_state->__pyx_n_s_data_ptr);
  Py_VISIT(traverse_module_state->__pyx_n_s_datetime);
  Py_VISIT(traverse_module_state->__pyx_n_s_debug);
  Py_VISIT(traverse_module_state->__pyx_n_s_decode);
  Py_VISIT(traverse_module_state->__pyx_n_s_default);
  Py_VISIT(traverse_module_state->__pyx_n_s_deferred_key);
  Py_VISIT(traverse_module_state->__pyx_n_s_deferred_keys);
  Py_VISIT(traverse_module_state->__pyx_n_s_delete);
  Py_VISIT(traverse_module_state->__pyx_n_u_delete);
  Py_VISIT(traverse_module_state->__pyx_n_s_dict);
  Py_VISIT(traverse_module_state->__pyx_n_s_dictKey);
  Py_VISIT(traverse_module_state->__pyx_n_s_dictValue);
  Py_VISIT(traverse_module_state->__pyx_n_u_direction);
  Py_VISIT(traverse_module_state->__pyx_kp_u_disable);
  Py_VISIT(traverse_module_state->__pyx_n_s_distinct);
  Py_VISIT(traverse_module_state->__pyx_n_s_distinctKey);
  Py_VISIT(traverse_module_state->__pyx_n_s_distinctOn);
  Py_VISIT(traverse_module_state->__pyx_n_u_distinctOn);
  Py_VISIT(traverse_module_state->__pyx_kp_u_distinct_on_s);
  Py_VISIT(traverse_module_state->__pyx_n_u_doubleValue);
  Py_VISIT(traverse_module_state->__pyx_n_s_dumps);
  Py_VISIT(traverse_module_state->__pyx_n_s_element);
  Py_VISIT(traverse_module_state->__pyx_n_s_element_inner);
  Py_VISIT(traverse_module_state->__pyx_kp_u_enable);
  Py_VISIT(traverse_module_state->__pyx_n_s_encode);
  Py_VISIT(traverse_module_state->__pyx_n_s_endCursor);
  Py_VISIT(traverse_module_state->__pyx_n_u_endCursor);
  Py_VISIT(traverse_module_state->__pyx_n_s_entities);
  Py_VISIT(traverse_module_state->__pyx_n_u_entityValue);
  Py_VISIT(traverse_module_state->__pyx_n_s_enumerate);
  Py_VISIT(traverse_module_state->__pyx_n_s_err);
  Py_VISIT(traverse_module_state->__pyx_n_s_error);
  Py_VISIT(traverse_module_state->__pyx_n_u_excludeFromIndexes);
  Py_VISIT(traverse_module_state->__pyx_n_s_exclude_from_indexes);
  Py_VISIT(traverse_module_state->__pyx_n_s_excludedProperties);
  Py_VISIT(traverse_module_state->__pyx_n_s_exponential_backoff);
  Py_VISIT(traverse_module_state->__pyx_n_s_extend);
  Py_VISIT(traverse_module_state->__pyx_n_u_filter);
  Py_VISIT(traverse_module_state->__pyx_n_s_filterList);
  Py_VISIT(traverse_module_state->__pyx_n_s_filters);
  Py_VISIT(traverse_module_state->__pyx_n_u_filters);
  Py_VISIT(traverse_module_state->__pyx_n_s_flipResults);
  Py_VISIT(traverse_module_state->__pyx_n_s_from_legacy_urlsafe);
  Py_VISIT(traverse_module_state->__pyx_kp_u_gc);
  Py_VISIT(traverse_module_state->__pyx_n_s_genexpr);
  Py_VISIT(traverse_module_state->__pyx_n_s_get);
  Py_VISIT(traverse_module_state->__pyx_n_s_google);
  Py_VISIT(traverse_module_state->__pyx_n_s_google_auth);
  Py_VISIT(traverse_module_state->__pyx_n_s_http_internal);
  Py_VISIT(traverse_module_state->__pyx_kp_u_https_datastore_googleapis_com_v);
  Py_VISIT(traverse_module_state->__pyx_kp_u_https_datastore_googleapis_com_v_2);
  Py_VISIT(traverse_module_state->__pyx_kp_u_https_datastore_googleapis_com_v_3);
  Py_VISIT(traverse_module_state->__pyx_kp_u_https_datastore_googleapis_com_v_4);
  Py_VISIT(traverse_module_state->__pyx_kp_u_https_datastore_googleapis_com_v_5);
  Py_VISIT(traverse_module_state->__pyx_kp_u_https_datastore_googleapis_com_v_6);
  Py_VISIT(traverse_module_state->__pyx_kp_u_https_datastore_googleapis_com_v_7);
  Py_VISIT(traverse_module_state->__pyx_kp_u_https_www_googleapis_com_auth_da);
  Py_VISIT(traverse_module_state->__pyx_n_s_i);
  Py_VISIT(traverse_module_state->__pyx_n_s_id);
  Py_VISIT(traverse_module_state->__pyx_n_u_id);
  Py_VISIT(traverse_module_state->__pyx_n_s_idx);
  Py_VISIT(traverse_module_state->__pyx_n_s_import);
  Py_VISIT(traverse_module_state->__pyx_n_s_initializing);
  Py_VISIT(traverse_module_state->__pyx_n_s_innerArrayElem);
  Py_VISIT(traverse_module_state->__pyx_n_s_insert);
  Py_VISIT(traverse_module_state->__pyx_n_s_int);
  Py_VISIT(traverse_module_state->__pyx_n_u_integerValue);
  Py_VISIT(traverse_module_state->__pyx_n_s_internalStartCursor);
  Py_VISIT(traverse_module_state->__pyx_n_s_isMulti);
  Py_VISIT(traverse_module_state->__pyx_n_s_is_coroutine);
  Py_VISIT(traverse_module_state->__pyx_kp_u_is_not_supported);
  Py_VISIT(traverse_module_state->__pyx_n_s_is_partial);
  Py_VISIT(traverse_module_state->__pyx_n_s_is_viur_datastore_request_ok);
  Py_VISIT(traverse_module_state->__pyx_kp_u_isenabled);
  Py_VISIT(traverse_module_state->__pyx_n_s_items);
  Py_VISIT(traverse_module_state->__pyx_n_s_json);
  Py_VISIT(traverse_module_state->__pyx_n_s_k);
  Py_VISIT(traverse_module_state->__pyx_n_s_key);
  Py_VISIT(traverse_module_state->__pyx_n_u_key);
  Py_VISIT(traverse_module_state->__pyx_n_s_keyList);
  Py_VISIT(traverse_module_state->__pyx_n_s_keyToPath);
  Py_VISIT(traverse_module_state->__pyx_n_u_keyValue);
  Py_VISIT(traverse_module_state->__pyx_n_s_keys);
  Py_VISIT(traverse_module_state->__pyx_n_u_keys);
  Py_VISIT(traverse_module_state->__pyx_n_s_keys_for_request);
  Py_VISIT(traverse_module_state->__pyx_n_s_kind);
  Py_VISIT(traverse_module_state->__pyx_n_u_kind);
  Py_VISIT(traverse_module_state->__pyx_n_s_kwargs);
  Py_VISIT(traverse_module_state->__pyx_n_s_limit);
  Py_VISIT(traverse_module_state->__pyx_n_u_limit);
  Py_VISIT(traverse_module_state->__pyx_n_s_loads);
  Py_VISIT(traverse_module_state->__pyx_n_s_logging);
  Py_VISIT(traverse_module_state->__pyx_n_s_main);
  Py_VISIT(traverse_module_state->__pyx_n_u_memcache_client);
  Py_VISIT(traverse_module_state->__pyx_n_s_missing_keys);
  Py_VISIT(traverse_module_state->__pyx_n_s_missing_keys_paths);
  Py_VISIT(traverse_module_state->__pyx_n_u_mode);
  Py_VISIT(traverse_module_state->__pyx_n_u_mutations);
  Py_VISIT(traverse_module_state->__pyx_n_s_name);
  Py_VISIT(traverse_module_state->__pyx_n_u_name);
  Py_VISIT(traverse_module_state->__pyx_n_s_name_2);
  Py_VISIT(traverse_module_state->__pyx_n_u_nested_query);
  Py_VISIT(traverse_module_state->__pyx_n_u_nullValue);
  Py_VISIT(traverse_module_state->__pyx_n_s_oldTxn);
  Py_VISIT(traverse_module_state->__pyx_n_s_op);
  Py_VISIT(traverse_module_state->__pyx_n_u_op);
  Py_VISIT(traverse_module_state->__pyx_n_u_order);
  Py_VISIT(traverse_module_state->__pyx_n_s_orders);
  Py_VISIT(traverse_module_state->__pyx_n_s_parent);
  Py_VISIT(traverse_module_state->__pyx_n_s_parser);
  Py_VISIT(traverse_module_state->__pyx_n_u_partitionId);
  Py_VISIT(traverse_module_state->__pyx_n_u_path);
  Py_VISIT(traverse_module_state->__pyx_n_s_pop);
  Py_VISIT(traverse_module_state->__pyx_n_s_post);
  Py_VISIT(traverse_module_state->__pyx_n_s_postData);
  Py_VISIT(traverse_module_state->__pyx_n_s_post_data);
  Py_VISIT(traverse_module_state->__pyx_n_u_previousTransaction);
  Py_VISIT(traverse_module_state->__pyx_n_s_projectID);
  Py_VISIT(traverse_module_state->__pyx_n_u_project_id);
  Py_VISIT(traverse_module_state->__pyx_n_s_properties);
  Py_VISIT(traverse_module_state->__pyx_n_u_properties);
  Py_VISIT(traverse_module_state->__pyx_n_u_property);
  Py_VISIT(traverse_module_state->__pyx_n_u_propertyFilter);
  Py_VISIT(traverse_module_state->__pyx_n_u_property_1);
  Py_VISIT(traverse_module_state->__pyx_n_s_put);
  Py_VISIT(traverse_module_state->__pyx_n_s_pysize);
  Py_VISIT(traverse_module_state->__pyx_n_s_pythonPropToJson);
  Py_VISIT(traverse_module_state->__pyx_n_u_query);
  Py_VISIT(traverse_module_state->__pyx_n_s_queryDefinition);
  Py_VISIT(traverse_module_state->__pyx_n_s_range);
  Py_VISIT(traverse_module_state->__pyx_n_u_readConsistency);
  Py_VISIT(traverse_module_state->__pyx_n_s_readOptions);
  Py_VISIT(traverse_module_state->__pyx_n_u_readOptions);
  Py_VISIT(traverse_module_state->__pyx_n_u_readWrite);
  Py_VISIT(traverse_module_state->__pyx_n_s_refresh_timeout);
  Py_VISIT(traverse_module_state->__pyx_n_s_replace);
  Py_VISIT(traverse_module_state->__pyx_n_s_requestedKeys);
  Py_VISIT(traverse_module_state->__pyx_n_s_requested_keys);
  Py_VISIT(traverse_module_state->__pyx_n_s_requests);
  Py_VISIT(traverse_module_state->__pyx_kp_s_requests_Response);
  Py_VISIT(traverse_module_state->__pyx_n_s_requests_exceptions);
  Py_VISIT(traverse_module_state->__pyx_n_s_res);
  Py_VISIT(traverse_module_state->__pyx_n_s_resDict);
  Py_VISIT(traverse_module_state->__pyx_n_s_resList);
  Py_VISIT(traverse_module_state->__pyx_n_s_res_from_cache);
  Py_VISIT(traverse_module_state->__pyx_n_s_res_from_db);
  Py_VISIT(traverse_module_state->__pyx_n_s_resp);
  Py_VISIT(traverse_module_state->__pyx_kp_u_resp_content);
  Py_VISIT(traverse_module_state->__pyx_kp_u_results);
  Py_VISIT(traverse_module_state->__pyx_n_s_return);
  Py_VISIT(traverse_module_state->__pyx_n_s_rollbackTxn);
  Py_VISIT(traverse_module_state->__pyx_n_s_runSingleFilter);
  Py_VISIT(traverse_module_state->__pyx_n_s_scopes);
  Py_VISIT(traverse_module_state->__pyx_kp_u_seconds);
  Py_VISIT(traverse_module_state->__pyx_n_s_send);
  Py_VISIT(traverse_module_state->__pyx_n_s_set);
  Py_VISIT(traverse_module_state->__pyx_n_s_singleValue);
  Py_VISIT(traverse_module_state->__pyx_n_s_sleep);
  Py_VISIT(traverse_module_state->__pyx_n_s_sleep_time);
  Py_VISIT(traverse_module_state->__pyx_n_s_sortOrder);
  Py_VISIT(traverse_module_state->__pyx_n_s_spec);
  Py_VISIT(traverse_module_state->__pyx_n_s_split);
  Py_VISIT(traverse_module_state->__pyx_kp_s_src_viur_datastore_transport_pyx);
  Py_VISIT(traverse_module_state->__pyx_n_s_startCursor);
  Py_VISIT(traverse_module_state->__pyx_n_u_startCursor);
  Py_VISIT(traverse_module_state->__pyx_n_s_str);
  Py_VISIT(traverse_module_state->__pyx_n_s_strftime);
  Py_VISIT(traverse_module_state->__pyx_n_u_stringValue);
  Py_VISIT(traverse_module_state->__pyx_n_s_strptime);
  Py_VISIT(traverse_module_state->__pyx_n_s_test);
  Py_VISIT(traverse_module_state->__pyx_n_s_throw);
  Py_VISIT(traverse_module_state->__pyx_n_s_time);
  Py_VISIT(traverse_module_state->__pyx_n_u_timestampValue);
  Py_VISIT(traverse_module_state->__pyx_n_s_timezone);
  Py_VISIT(traverse_module_state->__pyx_n_u_traceQueries);
  Py_VISIT(traverse_module_state->__pyx_n_u_transaction);
  Py_VISIT(traverse_module_state->__pyx_n_u_transactionOptions);
  Py_VISIT(traverse_module_state->__pyx_n_s_transport);
  Py_VISIT(traverse_module_state->__pyx_n_s_txnKey);
  Py_VISIT(traverse_module_state->__pyx_n_s_txnOptions);
  Py_VISIT(traverse_module_state->__pyx_n_s_typing);
  Py_VISIT(traverse_module_state->__pyx_n_s_tzinfo);
  Py_VISIT(traverse_module_state->__pyx_n_s_untouched_keys);
  Py_VISIT(traverse_module_state->__pyx_n_s_up_to);
  Py_VISIT(traverse_module_state->__pyx_n_u_up_to);
  Py_VISIT(traverse_module_state->__pyx_n_s_update);
  Py_VISIT(traverse_module_state->__pyx_n_u_upsert);
  Py_VISIT(traverse_module_state->__pyx_n_s_url);
  Py_VISIT(traverse_module_state->__pyx_n_s_utc);
  Py_VISIT(traverse_module_state->__pyx_n_s_v);
  Py_VISIT(traverse_module_state->__pyx_n_s_value);
  Py_VISIT(traverse_module_state->__pyx_n_u_value);
  Py_VISIT(traverse_module_state->__pyx_n_u_values);
  Py_VISIT(traverse_module_state->__pyx_n_s_version);
  Py_VISIT(traverse_module_state->__pyx_n_s_viur_datastore);
  Py_VISIT(traverse_module_state->__pyx_n_s_viur_datastore_config);
  Py_VISIT(traverse_module_state->__pyx_n_s_viur_datastore_errors);
  Py_VISIT(traverse_module_state->__pyx_n_s_viur_datastore_transport);
  Py_VISIT(traverse_module_state->__pyx_n_s_viur_datastore_types);
  Py_VISIT(traverse_module_state->__pyx_n_s_warning);
  Py_VISIT(traverse_module_state->__pyx_kp_u_with_filter);
  Py_VISIT(traverse_module_state->__pyx_n_s_x);
  Py_VISIT(traverse_module_state->__pyx_int_0);
  Py_VISIT(traverse_module_state->__pyx_int_1);
  Py_VISIT(traverse_module_state->__pyx_int_2);
  Py_VISIT(traverse_module_state->__pyx_int_4);
  Py_VISIT(traverse_module_state->__pyx_int_300);
  Py_VISIT(traverse_module_state->__pyx_int_0x7fffffffffffffff);
  Py_VISIT(traverse_module_state->__pyx_int_neg_1);
  Py_VISIT(traverse_module_state->__pyx_slice__2);
  Py_VISIT(traverse_module_state->__pyx_tuple__4);
  Py_VISIT(traverse_module_state->__pyx_slice__12);
  Py_VISIT(traverse_module_state->__pyx_slice__13);
  Py_VISIT(traverse_module_state->__pyx_slice__14);
  Py_VISIT(traverse_module_state->__pyx_tuple__15);
  Py_VISIT(traverse_module_state->__pyx_tuple__16);
  Py_VISIT(traverse_module_state->__pyx_tuple__17);
  Py_VISIT(traverse_module_state->__pyx_tuple__18);
  Py_VISIT(traverse_module_state->__pyx_tuple__19);
  Py_VISIT(traverse_module_state->__pyx_tuple__20);
  Py_VISIT(traverse_module_state->__pyx_tuple__22);
  Py_VISIT(traverse_module_state->__pyx_tuple__24);
  Py_VISIT(traverse_module_state->__pyx_tuple__26);
  Py_VISIT(traverse_module_state->__pyx_tuple__28);
  Py_VISIT(traverse_module_state->__pyx_tuple__30);
  Py_VISIT(traverse_module_state->__pyx_tuple__32);
  Py_VISIT(traverse_module_state->__pyx_tuple__34);
  Py_VISIT(traverse_module_state->__pyx_tuple__36);
  Py_VISIT(traverse_module_state->__pyx_tuple__38);
  Py_VISIT(traverse_module_state->__pyx_tuple__40);
  Py_VISIT(traverse_module_state->__pyx_tuple__42);
  Py_VISIT(traverse_module_state->__pyx_tuple__44);
  Py_VISIT(traverse_module_state->__pyx_codeobj__23);
  Py_VISIT(traverse_module_state->__pyx_codeobj__25);
  Py_VISIT(traverse_module_state->__pyx_codeobj__27);
  Py_VISIT(traverse_module_state->__pyx_codeobj__29);
  Py_VISIT(traverse_module_state->__pyx_codeobj__31);
  Py_VISIT(traverse_module_state->__pyx_codeobj__33);
  Py_VISIT(traverse_module_state->__pyx_codeobj__35);
  Py_VISIT(traverse_module_state->__pyx_codeobj__37);
  Py_VISIT(traverse_module_state->__pyx_codeobj__39);
  Py_VISIT(traverse_module_state->__pyx_codeobj__41);
  Py_VISIT(traverse_module_state->__pyx_codeobj__43);
  return 0;
}
#endif
/* #### Code section: module_state_defines ### */
#define __pyx_d __pyx_mstate_global->__pyx_d
#define __pyx_b __pyx_mstate_global->__pyx_b
#define __pyx_cython_runtime __pyx_mstate_global->__pyx_cython_runtime
#define __pyx_empty_tuple __pyx_mstate_global->__pyx_empty_tuple
#define __pyx_empty_bytes __pyx_mstate_global->__pyx_empty_bytes
#define __pyx_empty_unicode __pyx_mstate_global->__pyx_empty_unicode
#ifdef __Pyx_CyFunction_USED
#define __pyx_CyFunctionType __pyx_mstate_global->__pyx_CyFunctionType
#endif
#ifdef __Pyx_FusedFunction_USED
#define __pyx_FusedFunctionType __pyx_mstate_global->__pyx_FusedFunctionType
#endif
#ifdef __Pyx_Generator_USED
#define __pyx_GeneratorType __pyx_mstate_global->__pyx_GeneratorType
#endif
#ifdef __Pyx_IterableCoroutine_USED
#define __pyx_IterableCoroutineType __pyx_mstate_global->__pyx_IterableCoroutineType
#endif
#ifdef __Pyx_Coroutine_USED
#define __pyx_CoroutineAwaitType __pyx_mstate_global->__pyx_CoroutineAwaitType
#endif
#ifdef __Pyx_Coroutine_USED
#define __pyx_CoroutineType __pyx_mstate_global->__pyx_CoroutineType
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#define __pyx_ptype_7cpython_4type_type __pyx_mstate_global->__pyx_ptype_7cpython_4type_type
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#define __pyx_type_4viur_9datastore_9transport___pyx_scope_struct__genexpr __pyx_mstate_global->__pyx_type_4viur_9datastore_9transport___pyx_scope_struct__genexpr
#endif
#define __pyx_ptype_4viur_9datastore_9transport___pyx_scope_struct__genexpr __pyx_mstate_global->__pyx_ptype_4viur_9datastore_9transport___pyx_scope_struct__genexpr
#define __pyx_kp_u_ __pyx_mstate_global->__pyx_kp_u_
#define __pyx_n_u_AND __pyx_mstate_global->__pyx_n_u_AND
#define __pyx_n_u_ASCENDING __pyx_mstate_global->__pyx_n_u_ASCENDING
#define __pyx_n_u_ASCII __pyx_mstate_global->__pyx_n_u_ASCII
#define __pyx_n_s_AbortedError __pyx_mstate_global->__pyx_n_s_AbortedError
#define __pyx_kp_u_All_retries_are_exhausted_for_th __pyx_mstate_global->__pyx_kp_u_All_retries_are_exhausted_for_th
#define __pyx_n_s_AllocateIDs __pyx_mstate_global->__pyx_n_s_AllocateIDs
#define __pyx_n_s_Any __pyx_mstate_global->__pyx_n_s_Any
#define __pyx_n_s_AssertionError __pyx_mstate_global->__pyx_n_s_AssertionError
#define __pyx_n_s_AuthorizedSession __pyx_mstate_global->__pyx_n_s_AuthorizedSession
#define __pyx_kp_u_Cannot_call_runInTransaction_whi __pyx_mstate_global->__pyx_kp_u_Cannot_call_runInTransaction_whi
#define __pyx_n_s_CollisionError __pyx_mstate_global->__pyx_n_s_CollisionError
#define __pyx_n_s_ConnectionError __pyx_mstate_global->__pyx_n_s_ConnectionError
#define __pyx_n_s_Count __pyx_mstate_global->__pyx_n_s_Count
#define __pyx_n_u_DESCENDING __pyx_mstate_global->__pyx_n_u_DESCENDING
#define __pyx_n_s_Delete __pyx_mstate_global->__pyx_n_s_Delete
#define __pyx_n_u_EQUAL __pyx_mstate_global->__pyx_n_u_EQUAL
#define __pyx_kp_u_Empty_response_received_from_Dat __pyx_mstate_global->__pyx_kp_u_Empty_response_received_from_Dat
#define __pyx_n_s_Entity __pyx_mstate_global->__pyx_n_s_Entity
#define __pyx_n_u_GREATER_THAN __pyx_mstate_global->__pyx_n_u_GREATER_THAN
#define __pyx_n_u_GREATER_THAN_OR_EQUAL __pyx_mstate_global->__pyx_n_u_GREATER_THAN_OR_EQUAL
#define __pyx_n_s_Get __pyx_mstate_global->__pyx_n_s_Get
#define __pyx_n_s_Get_locals_genexpr __pyx_mstate_global->__pyx_n_s_Get_locals_genexpr
#define __pyx_kp_u_INVALID_RESPONSE_RECEIVED __pyx_mstate_global->__pyx_kp_u_INVALID_RESPONSE_RECEIVED
#define __pyx_n_s_InvalidArgumentError __pyx_mstate_global->__pyx_n_s_InvalidArgumentError
#define __pyx_kp_u_Invalid_data_received_from_Datas __pyx_mstate_global->__pyx_kp_u_Invalid_data_received_from_Datas
#define __pyx_kp_u_Invalid_key_in_entity_json_s __pyx_mstate_global->__pyx_kp_u_Invalid_key_in_entity_json_s
#define __pyx_kp_u_Invalid_number_of_mutation_resul __pyx_mstate_global->__pyx_kp_u_Invalid_number_of_mutation_resul
#define __pyx_kp_u_Invalid_op_s __pyx_mstate_global->__pyx_kp_u_Invalid_op_s
#define __pyx_n_s_Key __pyx_mstate_global->__pyx_n_s_Key
#define __pyx_kp_u_Keys_must_not_be_partial_or_kind __pyx_mstate_global->__pyx_kp_u_Keys_must_not_be_partial_or_kind
#define __pyx_n_u_LESS_THAN __pyx_mstate_global->__pyx_n_u_LESS_THAN
#define __pyx_n_u_LESS_THAN_OR_EQUAL __pyx_mstate_global->__pyx_n_u_LESS_THAN_OR_EQUAL
#define __pyx_n_s_List __pyx_mstate_global->__pyx_n_s_List
#define __pyx_kp_s_List_Entity __pyx_mstate_global->__pyx_kp_s_List_Entity
#define __pyx_kp_s_List_dict __pyx_mstate_global->__pyx_kp_s_List_dict
#define __pyx_n_u_NON_TRANSACTIONAL __pyx_mstate_global->__pyx_n_u_NON_TRANSACTIONAL
#define __pyx_n_u_NOT_FINISHED __pyx_mstate_global->__pyx_n_u_NOT_FINISHED
#define __pyx_n_u_NO_MORE_RESULTS __pyx_mstate_global->__pyx_n_u_NO_MORE_RESULTS
#define __pyx_n_s_NoMutationResultsError __pyx_mstate_global->__pyx_n_s_NoMutationResultsError
#define __pyx_kp_u_No_mutation_results_received __pyx_mstate_global->__pyx_kp_u_No_mutation_results_received
#define __pyx_kp_u_No_mutation_results_received_2 __pyx_mstate_global->__pyx_kp_u_No_mutation_results_received_2
#define __pyx_n_s_None __pyx_mstate_global->__pyx_n_s_None
#define __pyx_kp_u_None __pyx_mstate_global->__pyx_kp_u_None
#define __pyx_kp_u_Please_create_a_matching_index __pyx_mstate_global->__pyx_kp_u_Please_create_a_matching_index
#define __pyx_n_s_Put __pyx_mstate_global->__pyx_n_s_Put
#define __pyx_kp_u_Queried __pyx_mstate_global->__pyx_kp_u_Queried
#define __pyx_n_s_QueryDefinition __pyx_mstate_global->__pyx_n_s_QueryDefinition
#define __pyx_kp_u_Query_not_finished_Maybe_some_en __pyx_mstate_global->__pyx_kp_u_Query_not_finished_Maybe_some_en
#define __pyx_kp_u_Received_an_unexpected_key_updat __pyx_mstate_global->__pyx_kp_u_Received_an_unexpected_key_updat
#define __pyx_n_s_RecursionError __pyx_mstate_global->__pyx_n_s_RecursionError
#define __pyx_n_s_RequestsConnectionError __pyx_mstate_global->__pyx_n_s_RequestsConnectionError
#define __pyx_kp_u_Retrying_http_post_request_to_da __pyx_mstate_global->__pyx_kp_u_Retrying_http_post_request_to_da
#define __pyx_kp_u_Returned __pyx_mstate_global->__pyx_kp_u_Returned
#define __pyx_n_s_RunInTransaction __pyx_mstate_global->__pyx_n_s_RunInTransaction
#define __pyx_n_u_STRONG __pyx_mstate_global->__pyx_n_u_STRONG
#define __pyx_n_u_TRANSACTIONAL __pyx_mstate_global->__pyx_n_u_TRANSACTIONAL
#define __pyx_kp_u_The_Count_aggregation_query_is_a __pyx_mstate_global->__pyx_kp_u_The_Count_aggregation_query_is_a
#define __pyx_kp_u_UTF_8 __pyx_mstate_global->__pyx_kp_u_UTF_8
#define __pyx_n_s_Union __pyx_mstate_global->__pyx_n_s_Union
#define __pyx_kp_s_Union_Entity_List_Entity __pyx_mstate_global->__pyx_kp_s_Union_Entity_List_Entity
#define __pyx_kp_s_Union_Key_List_Key __pyx_mstate_global->__pyx_kp_s_Union_Key_List_Key
#define __pyx_kp_s_Union_Key_List_Key_Entity_List_E __pyx_mstate_global->__pyx_kp_s_Union_Key_List_Key_Entity_List_E
#define __pyx_kp_s_Union_None_Entity_List_Entity __pyx_mstate_global->__pyx_kp_s_Union_None_Entity_List_Entity
#define __pyx_kp_u_Unknown_type __pyx_mstate_global->__pyx_kp_u_Unknown_type
#define __pyx_n_s_ValueError __pyx_mstate_global->__pyx_n_s_ValueError
#define __pyx_n_s_ViurDatastoreError __pyx_mstate_global->__pyx_n_s_ViurDatastoreError
#define __pyx_kp_u_We_got_an_error_in_a_transaction __pyx_mstate_global->__pyx_kp_u_We_got_an_error_in_a_transaction
#define __pyx_kp_u_We_read_an_incomplete_null_key_f __pyx_mstate_global->__pyx_kp_u_We_read_an_incomplete_null_key_f
#define __pyx_kp_u_Y_m_dT_H_M_S __pyx_mstate_global->__pyx_kp_u_Y_m_dT_H_M_S
#define __pyx_kp_u_Y_m_dT_H_M_S_f __pyx_mstate_global->__pyx_kp_u_Y_m_dT_H_M_S_f
#define __pyx_kp_u_Y_m_dT_H_M_S_fZ __pyx_mstate_global->__pyx_kp_u_Y_m_dT_H_M_S_fZ
#define __pyx_kp_u__10 __pyx_mstate_global->__pyx_kp_u__10
#define __pyx_kp_u__11 __pyx_mstate_global->__pyx_kp_u__11
#define __pyx_n_s__21 __pyx_mstate_global->__pyx_n_s__21
#define __pyx_kp_u__3 __pyx_mstate_global->__pyx_kp_u__3
#define __pyx_n_s__45 __pyx_mstate_global->__pyx_n_s__45
#define __pyx_kp_u__5 __pyx_mstate_global->__pyx_kp_u__5
#define __pyx_kp_u__6 __pyx_mstate_global->__pyx_kp_u__6
#define __pyx_kp_u__7 __pyx_mstate_global->__pyx_kp_u__7
#define __pyx_kp_u__8 __pyx_mstate_global->__pyx_kp_u__8
#define __pyx_kp_u__9 __pyx_mstate_global->__pyx_kp_u__9
#define __pyx_n_s_accessLog __pyx_mstate_global->__pyx_n_s_accessLog
#define __pyx_n_u_affectedEntities __pyx_mstate_global->__pyx_n_u_affectedEntities
#define __pyx_n_s_affectedEntity __pyx_mstate_global->__pyx_n_s_affectedEntity
#define __pyx_n_u_aggregateProperties __pyx_mstate_global->__pyx_n_u_aggregateProperties
#define __pyx_n_u_aggregationResults __pyx_mstate_global->__pyx_n_u_aggregationResults
#define __pyx_n_u_aggregation_query __pyx_mstate_global->__pyx_n_u_aggregation_query
#define __pyx_n_u_aggregations __pyx_mstate_global->__pyx_n_u_aggregations
#define __pyx_n_u_allowOverriding __pyx_mstate_global->__pyx_n_u_allowOverriding
#define __pyx_n_s_allowOverriding_2 __pyx_mstate_global->__pyx_n_s_allowOverriding_2
#define __pyx_kp_u_and_orders __pyx_mstate_global->__pyx_kp_u_and_orders
#define __pyx_n_s_append __pyx_mstate_global->__pyx_n_s_append
#define __pyx_n_s_args __pyx_mstate_global->__pyx_n_s_args
#define __pyx_n_s_arrayElem __pyx_mstate_global->__pyx_n_s_arrayElem
#define __pyx_n_s_arrayIt __pyx_mstate_global->__pyx_n_s_arrayIt
#define __pyx_n_u_arrayValue __pyx_mstate_global->__pyx_n_u_arrayValue
#define __pyx_n_s_array_element __pyx_mstate_global->__pyx_n_s_array_element
#define __pyx_n_s_array_element_inner __pyx_mstate_global->__pyx_n_s_array_element_inner
#define __pyx_n_s_astimezone __pyx_mstate_global->__pyx_n_s_astimezone
#define __pyx_n_s_asyncio_coroutines __pyx_mstate_global->__pyx_n_s_asyncio_coroutines
#define __pyx_n_s_auth __pyx_mstate_global->__pyx_n_s_auth
#define __pyx_n_s_authenticated_request __pyx_mstate_global->__pyx_n_s_authenticated_request
#define __pyx_n_s_b64decode __pyx_mstate_global->__pyx_n_s_b64decode
#define __pyx_n_s_b64encode __pyx_mstate_global->__pyx_n_s_b64encode
#define __pyx_n_s_base64 __pyx_mstate_global->__pyx_n_s_base64
#define __pyx_n_u_blobValue __pyx_mstate_global->__pyx_n_u_blobValue
#define __pyx_n_u_booleanValue __pyx_mstate_global->__pyx_n_u_booleanValue
#define __pyx_n_s_bytes __pyx_mstate_global->__pyx_n_s_bytes
#define __pyx_n_s_cache __pyx_mstate_global->__pyx_n_s_cache
#define __pyx_n_s_callable __pyx_mstate_global->__pyx_n_s_callable
#define __pyx_n_s_callback __pyx_mstate_global->__pyx_n_s_callback
#define __pyx_n_s_class_getitem __pyx_mstate_global->__pyx_n_s_class_getitem
#define __pyx_n_s_cline_in_traceback __pyx_mstate_global->__pyx_n_s_cline_in_traceback
#define __pyx_n_s_close __pyx_mstate_global->__pyx_n_s_close
#define __pyx_n_u_compositeFilter __pyx_mstate_global->__pyx_n_u_compositeFilter
#define __pyx_n_s_conf __pyx_mstate_global->__pyx_n_s_conf
#define __pyx_n_s_content __pyx_mstate_global->__pyx_n_s_content
#define __pyx_n_s_copy __pyx_mstate_global->__pyx_n_s_copy
#define __pyx_n_u_count __pyx_mstate_global->__pyx_n_u_count
#define __pyx_n_s_credentials __pyx_mstate_global->__pyx_n_s_credentials
#define __pyx_n_s_currentCursor __pyx_mstate_global->__pyx_n_s_currentCursor
#define __pyx_n_s_currentDbAccessLog __pyx_mstate_global->__pyx_n_s_currentDbAccessLog
#define __pyx_n_s_currentTransaction __pyx_mstate_global->__pyx_n_s_currentTransaction
#define __pyx_n_s_currentTxn __pyx_mstate_global->__pyx_n_s_currentTxn
#define __pyx_n_s_data __pyx_mstate_global->__pyx_n_s_data
#define __pyx_n_s_data_ptr __pyx_mstate_global->__pyx_n_s_data_ptr
#define __pyx_n_s_datetime __pyx_mstate_global->__pyx_n_s_datetime
#define __pyx_n_s_debug __pyx_mstate_global->__pyx_n_s_debug
#define __pyx_n_s_decode __pyx_mstate_global->__pyx_n_s_decode
#define __pyx_n_s_default __pyx_mstate_global->__pyx_n_s_default
#define __pyx_n_s_deferred_key __pyx_mstate_global->__pyx_n_s_deferred_key
#define __pyx_n_s_deferred_keys __pyx_mstate_global->__pyx_n_s_deferred_keys
#define __pyx_n_s_delete __pyx_mstate_global->__pyx_n_s_delete
#define __pyx_n_u_delete __pyx_mstate_global->__pyx_n_u_delete
#define __pyx_n_s_dict __pyx_mstate_global->__pyx_n_s_dict
#define __pyx_n_s_dictKey __pyx_mstate_global->__pyx_n_s_dictKey
#define __pyx_n_s_dictValue __pyx_mstate_global->__pyx_n_s_dictValue
#define __pyx_n_u_direction __pyx_mstate_global->__pyx_n_u_direction
#define __pyx_kp_u_disable __pyx_mstate_global->__pyx_kp_u_disable
#define __pyx_n_s_distinct __pyx_mstate_global->__pyx_n_s_distinct
#define __pyx_n_s_distinctKey __pyx_mstate_global->__pyx_n_s_distinctKey
#define __pyx_n_s_distinctOn __pyx_mstate_global->__pyx_n_s_distinctOn
#define __pyx_n_u_distinctOn __pyx_mstate_global->__pyx_n_u_distinctOn
#define __pyx_kp_u_distinct_on_s __pyx_mstate_global->__pyx_kp_u_distinct_on_s
#define __pyx_n_u_doubleValue __pyx_mstate_global->__pyx_n_u_doubleValue
#define __pyx_n_s_dumps __pyx_mstate_global->__pyx_n_s_dumps
#define __pyx_n_s_element __pyx_mstate_global->__pyx_n_s_element
#define __pyx_n_s_element_inner __pyx_mstate_global->__pyx_n_s_element_inner
#define __pyx_kp_u_enable __pyx_mstate_global->__pyx_kp_u_enable
#define __pyx_n_s_encode __pyx_mstate_global->__pyx_n_s_encode
#define __pyx_n_s_endCursor __pyx_mstate_global->__pyx_n_s_endCursor
#define __pyx_n_u_endCursor __pyx_mstate_global->__pyx_n_u_endCursor
#define __pyx_n_s_entities __pyx_mstate_global->__pyx_n_s_entities
#define __pyx_n_u_entityValue __pyx_mstate_global->__pyx_n_u_entityValue
#define __pyx_n_s_enumerate __pyx_mstate_global->__pyx_n_s_enumerate
#define __pyx_n_s_err __pyx_mstate_global->__pyx_n_s_err
#define __pyx_n_s_error __pyx_mstate_global->__pyx_n_s_error
#define __pyx_n_u_excludeFromIndexes __pyx_mstate_global->__pyx_n_u_excludeFromIndexes
#define __pyx_n_s_exclude_from_indexes __pyx_mstate_global->__pyx_n_s_exclude_from_indexes
#define __pyx_n_s_excludedProperties __pyx_mstate_global->__pyx_n_s_excludedProperties
#define __pyx_n_s_exponential_backoff __pyx_mstate_global->__pyx_n_s_exponential_backoff
#define __pyx_n_s_extend __pyx_mstate_global->__pyx_n_s_extend
#define __pyx_n_u_filter __pyx_mstate_global->__pyx_n_u_filter
#define __pyx_n_s_filterList __pyx_mstate_global->__pyx_n_s_filterList
#define __pyx_n_s_filters __pyx_mstate_global->__pyx_n_s_filters
#define __pyx_n_u_filters __pyx_mstate_global->__pyx_n_u_filters
#define __pyx_n_s_flipResults __pyx_mstate_global->__pyx_n_s_flipResults
#define __pyx_n_s_from_legacy_urlsafe __pyx_mstate_global->__pyx_n_s_from_legacy_urlsafe
#define __pyx_kp_u_gc __pyx_mstate_global->__pyx_kp_u_gc
#define __pyx_n_s_genexpr __pyx_mstate_global->__pyx_n_s_genexpr
#define __pyx_n_s_get __pyx_mstate_global->__pyx_n_s_get
#define __pyx_n_s_google __pyx_mstate_global->__pyx_n_s_google
#define __pyx_n_s_google_auth __pyx_mstate_global->__pyx_n_s_google_auth
#define __pyx_n_s_http_internal __pyx_mstate_global->__pyx_n_s_http_internal
#define __pyx_kp_u_https_datastore_googleapis_com_v __pyx_mstate_global->__pyx_kp_u_https_datastore_googleapis_com_v
#define __pyx_kp_u_https_datastore_googleapis_com_v_2 __pyx_mstate_global->__pyx_kp_u_https_datastore_googleapis_com_v_2
#define __pyx_kp_u_https_datastore_googleapis_com_v_3 __pyx_mstate_global->__pyx_kp_u_https_datastore_googleapis_com_v_3
#define __pyx_kp_u_https_datastore_googleapis_com_v_4 __pyx_mstate_global->__pyx_kp_u_https_datastore_googleapis_com_v_4
#define __pyx_kp_u_https_datastore_googleapis_com_v_5 __pyx_mstate_global->__pyx_kp_u_https_datastore_googleapis_com_v_5
#define __pyx_kp_u_https_datastore_googleapis_com_v_6 __pyx_mstate_global->__pyx_kp_u_https_datastore_googleapis_com_v_6
#define __pyx_kp_u_https_datastore_googleapis_com_v_7 __pyx_mstate_global->__pyx_kp_u_https_datastore_googleapis_com_v_7
#define __pyx_kp_u_https_www_googleapis_com_auth_da __pyx_mstate_global->__pyx_kp_u_https_www_googleapis_com_auth_da
#define __pyx_n_s_i __pyx_mstate_global->__pyx_n_s_i
#define __pyx_n_s_id __pyx_mstate_global->__pyx_n_s_id
#define __pyx_n_u_id __pyx_mstate_global->__pyx_n_u_id
#define __pyx_n_s_idx __pyx_mstate_global->__pyx_n_s_idx
#define __pyx_n_s_import __pyx_mstate_global->__pyx_n_s_import
#define __pyx_n_s_initializing __pyx_mstate_global->__pyx_n_s_initializing
#define __pyx_n_s_innerArrayElem __pyx_mstate_global->__pyx_n_s_innerArrayElem
#define __pyx_n_s_insert __pyx_mstate_global->__pyx_n_s_insert
#define __pyx_n_s_int __pyx_mstate_global->__pyx_n_s_int
#define __pyx_n_u_integerValue __pyx_mstate_global->__pyx_n_u_integerValue
#define __pyx_n_s_internalStartCursor __pyx_mstate_global->__pyx_n_s_internalStartCursor
#define __pyx_n_s_isMulti __pyx_mstate_global->__pyx_n_s_isMulti
#define __pyx_n_s_is_coroutine __pyx_mstate_global->__pyx_n_s_is_coroutine
#define __pyx_kp_u_is_not_supported __pyx_mstate_global->__pyx_kp_u_is_not_supported
#define __pyx_n_s_is_partial __pyx_mstate_global->__pyx_n_s_is_partial
#define __pyx_n_s_is_viur_datastore_request_ok __pyx_mstate_global->__pyx_n_s_is_viur_datastore_request_ok
#define __pyx_kp_u_isenabled __pyx_mstate_global->__pyx_kp_u_isenabled
#define __pyx_n_s_items __pyx_mstate_global->__pyx_n_s_items
#define __pyx_n_s_json __pyx_mstate_global->__pyx_n_s_json
#define __pyx_n_s_k __pyx_mstate_global->__pyx_n_s_k
#define __pyx_n_s_key __pyx_mstate_global->__pyx_n_s_key
#define __pyx_n_u_key __pyx_mstate_global->__pyx_n_u_key
#define __pyx_n_s_keyList __pyx_mstate_global->__pyx_n_s_keyList
#define __pyx_n_s_keyToPath __pyx_mstate_global->__pyx_n_s_keyToPath
#define __pyx_n_u_keyValue __pyx_mstate_global->__pyx_n_u_keyValue
#define __pyx_n_s_keys __pyx_mstate_global->__pyx_n_s_keys
#define __pyx_n_u_keys __pyx_mstate_global->__pyx_n_u_keys
#define __pyx_n_s_keys_for_request __pyx_mstate_global->__pyx_n_s_keys_for_request
#define __pyx_n_s_kind __pyx_mstate_global->__pyx_n_s_kind
#define __pyx_n_u_kind __pyx_mstate_global->__pyx_n_u_kind
#define __pyx_n_s_kwargs __pyx_mstate_global->__pyx_n_s_kwargs
#define __pyx_n_s_limit __pyx_mstate_global->__pyx_n_s_limit
#define __pyx_n_u_limit __pyx_mstate_global->__pyx_n_u_limit
#define __pyx_n_s_loads __pyx_mstate_global->__pyx_n_s_loads
#define __pyx_n_s_logging __pyx_mstate_global->__pyx_n_s_logging
#define __pyx_n_s_main __pyx_mstate_global->__pyx_n_s_main
#define __pyx_n_u_memcache_client __pyx_mstate_global->__pyx_n_u_memcache_client
#define __pyx_n_s_missing_keys __pyx_mstate_global->__pyx_n_s_missing_keys
#define __pyx_n_s_missing_keys_paths __pyx_mstate_global->__pyx_n_s_missing_keys_paths
#define __pyx_n_u_mode __pyx_mstate_global->__pyx_n_u_mode
#define __pyx_n_u_mutations __pyx_mstate_global->__pyx_n_u_mutations
#define __pyx_n_s_name __pyx_mstate_global->__pyx_n_s_name
#define __pyx_n_u_name __pyx_mstate_global->__pyx_n_u_name
#define __pyx_n_s_name_2 __pyx_mstate_global->__pyx_n_s_name_2
#define __pyx_n_u_nested_query __pyx_mstate_global->__pyx_n_u_nested_query
#define __pyx_n_u_nullValue __pyx_mstate_global->__pyx_n_u_nullValue
#define __pyx_n_s_oldTxn __pyx_mstate_global->__pyx_n_s_oldTxn
#define __pyx_n_s_op __pyx_mstate_global->__pyx_n_s_op
#define __pyx_n_u_op __pyx_mstate_global->__pyx_n_u_op
#define __pyx_n_u_order __pyx_mstate_global->__pyx_n_u_order
#define __pyx_n_s_orders __pyx_mstate_global->__pyx_n_s_orders
#define __pyx_n_s_parent __pyx_mstate_global->__pyx_n_s_parent
#define __pyx_n_s_parser __pyx_mstate_global->__pyx_n_s_parser
#define __pyx_n_u_partitionId __pyx_mstate_global->__pyx_n_u_partitionId
#define __pyx_n_u_path __pyx_mstate_global->__pyx_n_u_path
#define __pyx_n_s_pop __pyx_mstate_global->__pyx_n_s_pop
#define __pyx_n_s_post __pyx_mstate_global->__pyx_n_s_post
#define __pyx_n_s_postData __pyx_mstate_global->__pyx_n_s_postData
#define __pyx_n_s_post_data __pyx_mstate_global->__pyx_n_s_post_data
#define __pyx_n_u_previousTransaction __pyx_mstate_global->__pyx_n_u_previousTransaction
#define __pyx_n_s_projectID __pyx_mstate_global->__pyx_n_s_projectID
#define __pyx_n_u_project_id __pyx_mstate_global->__pyx_n_u_project_id
#define __pyx_n_s_properties __pyx_mstate_global->__pyx_n_s_properties
#define __pyx_n_u_properties __pyx_mstate_global->__pyx_n_u_properties
#define __pyx_n_u_property __pyx_mstate_global->__pyx_n_u_property
#define __pyx_n_u_propertyFilter __pyx_mstate_global->__pyx_n_u_propertyFilter
#define __pyx_n_u_property_1 __pyx_mstate_global->__pyx_n_u_property_1
#define __pyx_n_s_put __pyx_mstate_global->__pyx_n_s_put
#define __pyx_n_s_pysize __pyx_mstate_global->__pyx_n_s_pysize
#define __pyx_n_s_pythonPropToJson __pyx_mstate_global->__pyx_n_s_pythonPropToJson
#define __pyx_n_u_query __pyx_mstate_global->__pyx_n_u_query
#define __pyx_n_s_queryDefinition __pyx_mstate_global->__pyx_n_s_queryDefinition
#define __pyx_n_s_range __pyx_mstate_global->__pyx_n_s_range
#define __pyx_n_u_readConsistency __pyx_mstate_global->__pyx_n_u_readConsistency
#define __pyx_n_s_readOptions __pyx_mstate_global->__pyx_n_s_readOptions
#define __pyx_n_u_readOptions __pyx_mstate_global->__pyx_n_u_readOptions
#define __pyx_n_u_readWrite __pyx_mstate_global->__pyx_n_u_readWrite
#define __pyx_n_s_refresh_timeout __pyx_mstate_global->__pyx_n_s_refresh_timeout
#define __pyx_n_s_replace __pyx_mstate_global->__pyx_n_s_replace
#define __pyx_n_s_requestedKeys __pyx_mstate_global->__pyx_n_s_requestedKeys
#define __pyx_n_s_requested_keys __pyx_mstate_global->__pyx_n_s_requested_keys
#define __pyx_n_s_requests __pyx_mstate_global->__pyx_n_s_requests
#define __pyx_kp_s_requests_Response __pyx_mstate_global->__pyx_kp_s_requests_Response
#define __pyx_n_s_requests_exceptions __pyx_mstate_global->__pyx_n_s_requests_exceptions
#define __pyx_n_s_res __pyx_mstate_global->__pyx_n_s_res
#define __pyx_n_s_resDict __pyx_mstate_global->__pyx_n_s_resDict
#define __pyx_n_s_resList __pyx_mstate_global->__pyx_n_s_resList
#define __pyx_n_s_res_from_cache __pyx_mstate_global->__pyx_n_s_res_from_cache
#define __pyx_n_s_res_from_db __pyx_mstate_global->__pyx_n_s_res_from_db
#define __pyx_n_s_resp __pyx_mstate_global->__pyx_n_s_resp
#define __pyx_kp_u_resp_content __pyx_mstate_global->__pyx_kp_u_resp_content
#define __pyx_kp_u_results __pyx_mstate_global->__pyx_kp_u_results
#define __pyx_n_s_return __pyx_mstate_global->__pyx_n_s_return
#define __pyx_n_s_rollbackTxn __pyx_mstate_global->__pyx_n_s_rollbackTxn
#define __pyx_n_s_runSingleFilter __pyx_mstate_global->__pyx_n_s_runSingleFilter
#define __pyx_n_s_scopes __pyx_mstate_global->__pyx_n_s_scopes
#define __pyx_kp_u_seconds __pyx_mstate_global->__pyx_kp_u_seconds
#define __pyx_n_s_send __pyx_mstate_global->__pyx_n_s_send
#define __pyx_n_s_set __pyx_mstate_global->__pyx_n_s_set
#define __pyx_n_s_singleValue __pyx_mstate_global->__pyx_n_s_singleValue
#define __pyx_n_s_sleep __pyx_mstate_global->__pyx_n_s_sleep
#define __pyx_n_s_sleep_time __pyx_mstate_global->__pyx_n_s_sleep_time
#define __pyx_n_s_sortOrder __pyx_mstate_global->__pyx_n_s_sortOrder
#define __pyx_n_s_spec __pyx_mstate_global->__pyx_n_s_spec
#define __pyx_n_s_split __pyx_mstate_global->__pyx_n_s_split
#define __pyx_kp_s_src_viur_datastore_transport_pyx __pyx_mstate_global->__pyx_kp_s_src_viur_datastore_transport_pyx
#define __pyx_n_s_startCursor __pyx_mstate_global->__pyx_n_s_startCursor
#define __pyx_n_u_startCursor __pyx_mstate_global->__pyx_n_u_startCursor
#define __pyx_n_s_str __pyx_mstate_global->__pyx_n_s_str
#define __pyx_n_s_strftime __pyx_mstate_global->__pyx_n_s_strftime
#define __pyx_n_u_stringValue __pyx_mstate_global->__pyx_n_u_stringValue
#define __pyx_n_s_strptime __pyx_mstate_global->__pyx_n_s_strptime
#define __pyx_n_s_test __pyx_mstate_global->__pyx_n_s_test
#define __pyx_n_s_throw __pyx_mstate_global->__pyx_n_s_throw
#define __pyx_n_s_time __pyx_mstate_global->__pyx_n_s_time
#define __pyx_n_u_timestampValue __pyx_mstate_global->__pyx_n_u_timestampValue
#define __pyx_n_s_timezone __pyx_mstate_global->__pyx_n_s_timezone
#define __pyx_n_u_traceQueries __pyx_mstate_global->__pyx_n_u_traceQueries
#define __pyx_n_u_transaction __pyx_mstate_global->__pyx_n_u_transaction
#define __pyx_n_u_transactionOptions __pyx_mstate_global->__pyx_n_u_transactionOptions
#define __pyx_n_s_transport __pyx_mstate_global->__pyx_n_s_transport
#define __pyx_n_s_txnKey __pyx_mstate_global->__pyx_n_s_txnKey
#define __pyx_n_s_txnOptions __pyx_mstate_global->__pyx_n_s_txnOptions
#define __pyx_n_s_typing __pyx_mstate_global->__pyx_n_s_typing
#define __pyx_n_s_tzinfo __pyx_mstate_global->__pyx_n_s_tzinfo
#define __pyx_n_s_untouched_keys __pyx_mstate_global->__pyx_n_s_untouched_keys
#define __pyx_n_s_up_to __pyx_mstate_global->__pyx_n_s_up_to
#define __pyx_n_u_up_to __pyx_mstate_global->__pyx_n_u_up_to
#define __pyx_n_s_update __pyx_mstate_global->__pyx_n_s_update
#define __pyx_n_u_upsert __pyx_mstate_global->__pyx_n_u_upsert
#define __pyx_n_s_url __pyx_mstate_global->__pyx_n_s_url
#define __pyx_n_s_utc __pyx_mstate_global->__pyx_n_s_utc
#define __pyx_n_s_v __pyx_mstate_global->__pyx_n_s_v
#define __pyx_n_s_value __pyx_mstate_global->__pyx_n_s_value
#define __pyx_n_u_value __pyx_mstate_global->__pyx_n_u_value
#define __pyx_n_u_values __pyx_mstate_global->__pyx_n_u_values
#define __pyx_n_s_version __pyx_mstate_global->__pyx_n_s_version
#define __pyx_n_s_viur_datastore __pyx_mstate_global->__pyx_n_s_viur_datastore
#define __pyx_n_s_viur_datastore_config __pyx_mstate_global->__pyx_n_s_viur_datastore_config
#define __pyx_n_s_viur_datastore_errors __pyx_mstate_global->__pyx_n_s_viur_datastore_errors
#define __pyx_n_s_viur_datastore_transport __pyx_mstate_global->__pyx_n_s_viur_datastore_transport
#define __pyx_n_s_viur_datastore_types __pyx_mstate_global->__pyx_n_s_viur_datastore_types
#define __pyx_n_s_warning __pyx_mstate_global->__pyx_n_s_warning
#define __pyx_kp_u_with_filter __pyx_mstate_global->__pyx_kp_u_with_filter
#define __pyx_n_s_x __pyx_mstate_global->__pyx_n_s_x
#define __pyx_int_0 __pyx_mstate_global->__pyx_int_0
#define __pyx_int_1 __pyx_mstate_global->__pyx_int_1
#define __pyx_int_2 __pyx_mstate_global->__pyx_int_2
#define __pyx_int_4 __pyx_mstate_global->__pyx_int_4
#define __pyx_int_300 __pyx_mstate_global->__pyx_int_300
#define __pyx_int_0x7fffffffffffffff __pyx_mstate_global->__pyx_int_0x7fffffffffffffff
#define __pyx_int_neg_1 __pyx_mstate_global->__pyx_int_neg_1
#define __pyx_slice__2 __pyx_mstate_global->__pyx_slice__2
#define __pyx_tuple__4 __pyx_mstate_global->__pyx_tuple__4
#define __pyx_slice__12 __pyx_mstate_global->__pyx_slice__12
#define __pyx_slice__13 __pyx_mstate_global->__pyx_slice__13
#define __pyx_slice__14 __pyx_mstate_global->__pyx_slice__14
#define __pyx_tuple__15 __pyx_mstate_global->__pyx_tuple__15
#define __pyx_tuple__16 __pyx_mstate_global->__pyx_tuple__16
#define __pyx_tuple__17 __pyx_mstate_global->__pyx_tuple__17
#define __pyx_tuple__18 __pyx_mstate_global->__pyx_tuple__18
#define __pyx_tuple__19 __pyx_mstate_global->__pyx_tuple__19
#define __pyx_tuple__20 __pyx_mstate_global->__pyx_tuple__20
#define __pyx_tuple__22 __pyx_mstate_global->__pyx_tuple__22
#define __pyx_tuple__24 __pyx_mstate_global->__pyx_tuple__24
#define __pyx_tuple__26 __pyx_mstate_global->__pyx_tuple__26
#define __pyx_tuple__28 __pyx_mstate_global->__pyx_tuple__28
#define __pyx_tuple__30 __pyx_mstate_global->__pyx_tuple__30
#define __pyx_tuple__32 __pyx_mstate_global->__pyx_tuple__32
#define __pyx_tuple__34 __pyx_mstate_global->__pyx_tuple__34
#define __pyx_tuple__36 __pyx_mstate_global->__pyx_tuple__36
#define __pyx_tuple__38 __pyx_mstate_global->__pyx_tuple__38
#define __pyx_tuple__40 __pyx_mstate_global->__pyx_tuple__40
#define __pyx_tuple__42 __pyx_mstate_global->__pyx_tuple__42
#define __pyx_tuple__44 __pyx_mstate_global->__pyx_tuple__44
#define __pyx_codeobj__23 __pyx_mstate_global->__pyx_codeobj__23
#define __pyx_codeobj__25 __pyx_mstate_global->__pyx_codeobj__25
#define __pyx_codeobj__27 __pyx_mstate_global->__pyx_codeobj__27
#define __pyx_codeobj__29 __pyx_mstate_global->__pyx_codeobj__29
#define __pyx_codeobj__31 __pyx_mstate_global->__pyx_codeobj__31
#define __pyx_codeobj__33 __pyx_mstate_global->__pyx_codeobj__33
#define __pyx_codeobj__35 __pyx_mstate_global->__pyx_codeobj__35
#define __pyx_codeobj__37 __pyx_mstate_global->__pyx_codeobj__37
#define __pyx_codeobj__39 __pyx_mstate_global->__pyx_codeobj__39
#define __pyx_codeobj__41 __pyx_mstate_global->__pyx_codeobj__41
#define __pyx_codeobj__43 __pyx_mstate_global->__pyx_codeobj__43
/* #### Code section: module_code ### */

/* "viur/datastore/transport.pyx":105
 * )
 * 
 * def authenticated_request(url: str, data: bytes) -> requests.Response:             # <<<<<<<<<<<<<<
 *     """
 *         Runs one http request to the datastore rest api, authenticated with the current projects service account.
 */

/* Python wrapper */
static PyObject *__pyx_pw_4viur_9datastore_9transport_1authenticated_request(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_4viur_9datastore_9transport_authenticated_request, "\n        Runs one http request to the datastore rest api, authenticated with the current projects service account.\n        Will retry up to three times in case the connection cannot be established the first time.\n\n        :param url: The url to post the data to\n        :param data:: The data to include in the post request\n        :return: The Response object\n    ");
static PyMethodDef __pyx_mdef_4viur_9datastore_9transport_1authenticated_request = {"authenticated_request", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_4viur_9datastore_9transport_1authenticated_request, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_4viur_9datastore_9transport_authenticated_request};
static PyObject *__pyx_pw_4viur_9datastore_9transport_1authenticated_request(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_url = 0;
  PyObject *__pyx_v_data = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[2] = {0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("authenticated_request (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_url,&__pyx_n_s_data,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_url)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 105, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_data)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 105, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("authenticated_request", 1, 2, 2, 1); __PYX_ERR(0, 105, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "authenticated_request") < 0)) __PYX_ERR(0, 105, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 2)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
    }
    __pyx_v_url = ((PyObject*)values[0]);
    __pyx_v_data = ((PyObject*)values[1]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("authenticated_request", 1, 2, 2, __pyx_nargs); __PYX_ERR(0, 105, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("viur.datastore.transport.authenticated_request", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_url), (&PyUnicode_Type), 0, "url", 1))) __PYX_ERR(0, 105, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_data), (&PyBytes_Type), 0, "data", 1))) __PYX_ERR(0, 105, __pyx_L1_error)
  __pyx_r = __pyx_pf_4viur_9datastore_9transport_authenticated_request(__pyx_self, __pyx_v_url, __pyx_v_data);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4viur_9datastore_9transport_authenticated_request(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_url, PyObject *__pyx_v_data) {
  long __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  long __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  int __pyx_t_12;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("authenticated_request", 1);

  /* "viur/datastore/transport.pyx":114
 *         :return: The Response object
 *     """
 *     for i in range(0, 3):             # <<<<<<<<<<<<<<
 *         try:
 *             return _http_internal.post(
 */
  for (__pyx_t_1 = 0; __pyx_t_1 < 3; __pyx_t_1+=1) {
    __pyx_v_i = __pyx_t_1;

    /* "viur/datastore/transport.pyx":115
 *     """
 *     for i in range(0, 3):
 *         try:             # <<<<<<<<<<<<<<
 *             return _http_internal.post(
 *                 url=url,
 */
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_2, &__pyx_t_3, &__pyx_t_4);
      __Pyx_XGOTREF(__pyx_t_2);
      __Pyx_XGOTREF(__pyx_t_3);
      __Pyx_XGOTREF(__pyx_t_4);
      /*try:*/ {

        /* "viur/datastore/transport.pyx":116
 *     for i in range(0, 3):
 *         try:
 *             return _http_internal.post(             # <<<<<<<<<<<<<<
 *                 url=url,
 *                 data=data,
 */
        __Pyx_XDECREF(__pyx_r);
        __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_http_internal); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 116, __pyx_L5_error)
        __Pyx_GOTREF(__pyx_t_5);
        __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_post); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 116, __pyx_L5_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

        /* "viur/datastore/transport.pyx":117
 *         try:
 *             return _http_internal.post(
 *                 url=url,             # <<<<<<<<<<<<<<
 *                 data=data,
 *             )
 */
        __pyx_t_5 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 117, __pyx_L5_error)
        __Pyx_GOTREF(__pyx_t_5);
        if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_url, __pyx_v_url) < 0) __PYX_ERR(0, 117, __pyx_L5_error)

        /* "viur/datastore/transport.pyx":118
 *             return _http_internal.post(
 *                 url=url,
 *                 data=data,             # <<<<<<<<<<<<<<
 *             )
 *         except RequestsConnectionError:
 */
        if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_data, __pyx_v_data) < 0) __PYX_ERR(0, 117, __pyx_L5_error)

        /* "viur/datastore/transport.pyx":116
 *     for i in range(0, 3):
 *         try:
 *             return _http_internal.post(             # <<<<<<<<<<<<<<
 *                 url=url,
 *                 data=data,
 */
        __pyx_t_7 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_empty_tuple, __pyx_t_5); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 116, __pyx_L5_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        __pyx_r = __pyx_t_7;
        __pyx_t_7 = 0;
        goto __pyx_L9_try_return;

        /* "viur/datastore/transport.pyx":115
 *     """
 *     for i in range(0, 3):
 *         try:             # <<<<<<<<<<<<<<
 *             return _http_internal.post(
 *                 url=url,
 */
      }
      __pyx_L5_error:;
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;

      /* "viur/datastore/transport.pyx":120
 *                 data=data,
 *             )
 *         except RequestsConnectionError:             # <<<<<<<<<<<<<<
 *             logging.debug("Retrying http post request to datastore")
 *             if i == 2:
 */
      __Pyx_ErrFetch(&__pyx_t_7, &__pyx_t_5, &__pyx_t_6);
      __Pyx_GetModuleGlobalName(__pyx_t_8, __pyx_n_s_RequestsConnectionError); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 120, __pyx_L7_except_error)
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_t_9 = __Pyx_PyErr_GivenExceptionMatches(__pyx_t_7, __pyx_t_8);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_ErrRestore(__pyx_t_7, __pyx_t_5, __pyx_t_6);
      __pyx_t_7 = 0; __pyx_t_5 = 0; __pyx_t_6 = 0;
      if (__pyx_t_9) {
        __Pyx_AddTraceback("viur.datastore.transport.authenticated_request", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_6, &__pyx_t_5, &__pyx_t_7) < 0) __PYX_ERR(0, 120, __pyx_L7_except_error)
        __Pyx_XGOTREF(__pyx_t_6);
        __Pyx_XGOTREF(__pyx_t_5);
        __Pyx_XGOTREF(__pyx_t_7);

        /* "viur/datastore/transport.pyx":121
 *             )
 *         except RequestsConnectionError:
 *             logging.debug("Retrying http post request to datastore")             # <<<<<<<<<<<<<<
 *             if i == 2:
 *                 raise
 */
        __Pyx_GetModuleGlobalName(__pyx_t_10, __pyx_n_s_logging); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 121, __pyx_L7_except_error)
        __Pyx_GOTREF(__pyx_t_10);
        __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_t_10, __pyx_n_s_debug); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 121, __pyx_L7_except_error)
        __Pyx_GOTREF(__pyx_t_11);
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
        __pyx_t_10 = NULL;
        __pyx_t_9 = 0;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_11))) {
          __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_11);
          if (likely(__pyx_t_10)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_11);
            __Pyx_INCREF(__pyx_t_10);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_11, function);
            __pyx_t_9 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_10, __pyx_kp_u_Retrying_http_post_request_to_da};
          __pyx_t_8 = __Pyx_PyObject_FastCall(__pyx_t_11, __pyx_callargs+1-__pyx_t_9, 1+__pyx_t_9);
          __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
          if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 121, __pyx_L7_except_error)
          __Pyx_GOTREF(__pyx_t_8);
          __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
        }
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

        /* "viur/datastore/transport.pyx":122
 *         except RequestsConnectionError:
 *             logging.debug("Retrying http post request to datastore")
 *             if i == 2:             # <<<<<<<<<<<<<<
 *                 raise
 *             continue
 */
        __pyx_t_12 = (__pyx_v_i == 2);
        if (unlikely(__pyx_t_12)) {

          /* "viur/datastore/transport.pyx":123
 *             logging.debug("Retrying http post request to datastore")
 *             if i == 2:
 *                 raise             # <<<<<<<<<<<<<<
 *             continue
 * 
 */
          __Pyx_GIVEREF(__pyx_t_6);
          __Pyx_GIVEREF(__pyx_t_5);
          __Pyx_XGIVEREF(__pyx_t_7);
          __Pyx_ErrRestoreWithState(__pyx_t_6, __pyx_t_5, __pyx_t_7);
          __pyx_t_6 = 0; __pyx_t_5 = 0; __pyx_t_7 = 0; 
          __PYX_ERR(0, 123, __pyx_L7_except_error)

          /* "viur/datastore/transport.pyx":122
 *         except RequestsConnectionError:
 *             logging.debug("Retrying http post request to datastore")
 *             if i == 2:             # <<<<<<<<<<<<<<
 *                 raise
 *             continue
 */
        }

        /* "viur/datastore/transport.pyx":124
 *             if i == 2:
 *                 raise
 *             continue             # <<<<<<<<<<<<<<
 * 
 * def keyToPath(key: Key) -> List[dict]:
 */
        goto __pyx_L13_except_continue;
        __pyx_L13_except_continue:;
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
        goto __pyx_L11_try_continue;
      }
      goto __pyx_L7_except_error;

      /* "viur/datastore/transport.pyx":115
 *     """
 *     for i in range(0, 3):
 *         try:             # <<<<<<<<<<<<<<
 *             return _http_internal.post(
 *                 url=url,
 */
      __pyx_L7_except_error:;
      __Pyx_XGIVEREF(__pyx_t_2);
      __Pyx_XGIVEREF(__pyx_t_3);
      __Pyx_XGIVEREF(__pyx_t_4);
      __Pyx_ExceptionReset(__pyx_t_2, __pyx_t_3, __pyx_t_4);
      goto __pyx_L1_error;
      __pyx_L11_try_continue:;
      __Pyx_XGIVEREF(__pyx_t_2);
      __Pyx_XGIVEREF(__pyx_t_3);
      __Pyx_XGIVEREF(__pyx_t_4);
      __Pyx_ExceptionReset(__pyx_t_2, __pyx_t_3, __pyx_t_4);
      goto __pyx_L3_continue;
      __pyx_L9_try_return:;
      __Pyx_XGIVEREF(__pyx_t_2);
      __Pyx_XGIVEREF(__pyx_t_3);
      __Pyx_XGIVEREF(__pyx_t_4);
      __Pyx_ExceptionReset(__pyx_t_2, __pyx_t_3, __pyx_t_4);
      goto __pyx_L0;
    }
    __pyx_L3_continue:;
  }

  /* "viur/datastore/transport.pyx":105
 * )
 * 
 * def authenticated_request(url: str, data: bytes) -> requests.Response:             # <<<<<<<<<<<<<<
 *     """
 *         Runs one http request to the datastore rest api, authenticated with the current projects service account.
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_AddTraceback("viur.datastore.transport.authenticated_request", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "viur/datastore/transport.pyx":126
 *             continue
 * 
 * def keyToPath(key: Key) -> List[dict]:             # <<<<<<<<<<<<<<
 *     """
 *         Converts a Key object to the PathElements expected by the rest API.
 */

/* Python wrapper */
static PyObject *__pyx_pw_4viur_9datastore_9transport_3keyToPath(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_4viur_9datastore_9transport_2keyToPath, "\n        Converts a Key object to the PathElements expected by the rest API.\n        See https://cloud.google.com/datastore/docs/reference/data/rest/v1/Key#PathElement\n\n        :param key: The key object to convert\n        :return: The list of path elements corresponding to this key\n    ");
static PyMethodDef __pyx_mdef_4viur_9datastore_9transport_3keyToPath = {"keyToPath", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_4viur_9datastore_9transport_3keyToPath, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_4viur_9datastore_9transport_2keyToPath};
static PyObject *__pyx_pw_4viur_9datastore_9transport_3keyToPath(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_key = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("keyToPath (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_key,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_key)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 126, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "keyToPath") < 0)) __PYX_ERR(0, 126, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_key = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("keyToPath", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 126, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("viur.datastore.transport.keyToPath", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4viur_9datastore_9transport_2keyToPath(__pyx_self, __pyx_v_key);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4viur_9datastore_9transport_2keyToPath(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_key) {
  PyObject *__pyx_v_res = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("keyToPath", 0);
  __Pyx_INCREF(__pyx_v_key);

  /* "viur/datastore/transport.pyx":134
 *         :return: The list of path elements corresponding to this key
 *     """
 *     res = []             # <<<<<<<<<<<<<<
 *     while key:
 *         res.insert(0,
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 134, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_res = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "viur/datastore/transport.pyx":135
 *     """
 *     res = []
 *     while key:             # <<<<<<<<<<<<<<
 *         res.insert(0,
 *                    {
 */
  while (1) {
    __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_v_key); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(0, 135, __pyx_L1_error)
    if (!__pyx_t_2) break;

    /* "viur/datastore/transport.pyx":140
 *                        "kind": key.kind,
 *                        "id": key.id,
 *                    } if key.id else {             # <<<<<<<<<<<<<<
 *                        "kind": key.kind,
 *                        "name": key.name
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_key, __pyx_n_s_id); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 140, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(0, 140, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (__pyx_t_2) {

      /* "viur/datastore/transport.pyx":138
 *         res.insert(0,
 *                    {
 *                        "kind": key.kind,             # <<<<<<<<<<<<<<
 *                        "id": key.id,
 *                    } if key.id else {
 */
      __pyx_t_3 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 138, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_key, __pyx_n_s_kind); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 138, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      if (PyDict_SetItem(__pyx_t_3, __pyx_n_u_kind, __pyx_t_4) < 0) __PYX_ERR(0, 138, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

      /* "viur/datastore/transport.pyx":139
 *                    {
 *                        "kind": key.kind,
 *                        "id": key.id,             # <<<<<<<<<<<<<<
 *                    } if key.id else {
 *                        "kind": key.kind,
 */
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_key, __pyx_n_s_id); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 139, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      if (PyDict_SetItem(__pyx_t_3, __pyx_n_u_id, __pyx_t_4) < 0) __PYX_ERR(0, 138, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_1 = __pyx_t_3;
      __pyx_t_3 = 0;
    } else {

      /* "viur/datastore/transport.pyx":143
 *                        "kind": key.kind,
 *                        "name": key.name
 *                    } if key.name else {             # <<<<<<<<<<<<<<
 *                        "kind": key.kind,
 *                    }
 */
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_key, __pyx_n_s_name); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 143, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely((__pyx_t_5 < 0))) __PYX_ERR(0, 143, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (__pyx_t_5) {

        /* "viur/datastore/transport.pyx":141
 *                        "id": key.id,
 *                    } if key.id else {
 *                        "kind": key.kind,             # <<<<<<<<<<<<<<
 *                        "name": key.name
 *                    } if key.name else {
 */
        __pyx_t_4 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 141, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_key, __pyx_n_s_kind); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 141, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        if (PyDict_SetItem(__pyx_t_4, __pyx_n_u_kind, __pyx_t_6) < 0) __PYX_ERR(0, 141, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

        /* "viur/datastore/transport.pyx":142
 *                    } if key.id else {
 *                        "kind": key.kind,
 *                        "name": key.name             # <<<<<<<<<<<<<<
 *                    } if key.name else {
 *                        "kind": key.kind,
 */
        __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_key, __pyx_n_s_name); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 142, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        if (PyDict_SetItem(__pyx_t_4, __pyx_n_u_name, __pyx_t_6) < 0) __PYX_ERR(0, 141, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __pyx_t_3 = __pyx_t_4;
        __pyx_t_4 = 0;
      } else {

        /* "viur/datastore/transport.pyx":144
 *                        "name": key.name
 *                    } if key.name else {
 *                        "kind": key.kind,             # <<<<<<<<<<<<<<
 *                    }
 *                    )
 */
        __pyx_t_4 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 144, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_key, __pyx_n_s_kind); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 144, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        if (PyDict_SetItem(__pyx_t_4, __pyx_n_u_kind, __pyx_t_6) < 0) __PYX_ERR(0, 144, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __pyx_t_3 = __pyx_t_4;
        __pyx_t_4 = 0;
      }
      __pyx_t_1 = __pyx_t_3;
      __pyx_t_3 = 0;
    }

    /* "viur/datastore/transport.pyx":136
 *     res = []
 *     while key:
 *         res.insert(0,             # <<<<<<<<<<<<<<
 *                    {
 *                        "kind": key.kind,
 */
    __pyx_t_7 = PyList_Insert(__pyx_v_res, 0, __pyx_t_1); if (unlikely(__pyx_t_7 == ((int)-1))) __PYX_ERR(0, 136, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "viur/datastore/transport.pyx":147
 *                    }
 *                    )
 *         key = key.parent             # <<<<<<<<<<<<<<
 *     return res
 * 
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_key, __pyx_n_s_parent); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 147, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF_SET(__pyx_v_key, __pyx_t_1);
    __pyx_t_1 = 0;
  }

  /* "viur/datastore/transport.pyx":148
 *                    )
 *         key = key.parent
 *     return res             # <<<<<<<<<<<<<<
 * 
 * def pythonPropToJson(v) -> dict:
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_res);
  __pyx_r = __pyx_v_res;
  goto __pyx_L0;

  /* "viur/datastore/transport.pyx":126
 *             continue
 * 
 * def keyToPath(key: Key) -> List[dict]:             # <<<<<<<<<<<<<<
 *     """
 *         Converts a Key object to the PathElements expected by the rest API.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("viur.datastore.transport.keyToPath", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_res);
  __Pyx_XDECREF(__pyx_v_key);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "viur/datastore/transport.pyx":150
 *     return res
 * 
 * def pythonPropToJson(v) -> dict:             # <<<<<<<<<<<<<<
 *     """
 *         Convert a python type to the object representation expected by the datastore rest API.
 */

/* Python wrapper */
static PyObject *__pyx_pw_4viur_9datastore_9transport_5pythonPropToJson(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_4viur_9datastore_9transport_4pythonPropToJson, "\n        Convert a python type to the object representation expected by the datastore rest API.\n        See https://cloud.google.com/datastore/docs/reference/data/rest/v1/projects/runQuery#Value\n\n        :param v: The python object to convert\n        :return: The representation of that object as expected by the rest api.\n    ");
static PyMethodDef __pyx_mdef_4viur_9datastore_9transport_5pythonPropToJson = {"pythonPropToJson", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_4viur_9datastore_9transport_5pythonPropToJson, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_4viur_9datastore_9transport_4pythonPropToJson};
static PyObject *__pyx_pw_4viur_9datastore_9transport_5pythonPropToJson(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_v = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("pythonPropToJson (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_v,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_v)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 150, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "pythonPropToJson") < 0)) __PYX_ERR(0, 150, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_v = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("pythonPropToJson", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 150, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("viur.datastore.transport.pythonPropToJson", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4viur_9datastore_9transport_4pythonPropToJson(__pyx_self, __pyx_v_v);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4viur_9datastore_9transport_4pythonPropToJson(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_v) {
  PyObject *__pyx_v_excludedProperties = 0;
  PyObject *__pyx_v_properties = NULL;
  PyObject *__pyx_v_dictKey = NULL;
  PyObject *__pyx_v_dictValue = NULL;
  PyObject *__pyx_v_resList = NULL;
  PyObject *__pyx_v_value = NULL;
  PyObject *__pyx_v_resDict = NULL;
  PyObject *__pyx_7genexpr__pyx_v_dictKey = NULL;
  PyObject *__pyx_7genexpr__pyx_v_dictValue = NULL;
  PyObject *__pyx_8genexpr1__pyx_v_x = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  PyObject *__pyx_t_9 = NULL;
  Py_ssize_t __pyx_t_10;
  Py_ssize_t __pyx_t_11;
  int __pyx_t_12;
  Py_ssize_t __pyx_t_13;
  PyObject *(*__pyx_t_14)(PyObject *);
  int __pyx_t_15;
  PyObject *__pyx_t_16 = NULL;
  Py_UCS4 __pyx_t_17;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("pythonPropToJson", 1);

  /* "viur/datastore/transport.pyx":158
 *         :return: The representation of that object as expected by the rest api.
 *     """
 *     if v is True or v is False:             # <<<<<<<<<<<<<<
 *         return {
 *             "booleanValue": v
 */
  __pyx_t_2 = (__pyx_v_v == Py_True);
  if (!__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_2 = (__pyx_v_v == Py_False);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_1) {

    /* "viur/datastore/transport.pyx":159
 *     """
 *     if v is True or v is False:
 *         return {             # <<<<<<<<<<<<<<
 *             "booleanValue": v
 *         }
 */
    __Pyx_XDECREF(__pyx_r);

    /* "viur/datastore/transport.pyx":160
 *     if v is True or v is False:
 *         return {
 *             "booleanValue": v             # <<<<<<<<<<<<<<
 *         }
 *     elif v is None:
 */
    __pyx_t_3 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 160, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (PyDict_SetItem(__pyx_t_3, __pyx_n_u_booleanValue, __pyx_v_v) < 0) __PYX_ERR(0, 160, __pyx_L1_error)
    __pyx_r = ((PyObject*)__pyx_t_3);
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "viur/datastore/transport.pyx":158
 *         :return: The representation of that object as expected by the rest api.
 *     """
 *     if v is True or v is False:             # <<<<<<<<<<<<<<
 *         return {
 *             "booleanValue": v
 */
  }

  /* "viur/datastore/transport.pyx":162
 *             "booleanValue": v
 *         }
 *     elif v is None:             # <<<<<<<<<<<<<<
 *         return {"nullValue": None}
 *     elif isinstance(v, int):
 */
  __pyx_t_1 = (__pyx_v_v == Py_None);
  if (__pyx_t_1) {

    /* "viur/datastore/transport.pyx":163
 *         }
 *     elif v is None:
 *         return {"nullValue": None}             # <<<<<<<<<<<<<<
 *     elif isinstance(v, int):
 *         return {
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_3 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 163, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (PyDict_SetItem(__pyx_t_3, __pyx_n_u_nullValue, Py_None) < 0) __PYX_ERR(0, 163, __pyx_L1_error)
    __pyx_r = ((PyObject*)__pyx_t_3);
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "viur/datastore/transport.pyx":162
 *             "booleanValue": v
 *         }
 *     elif v is None:             # <<<<<<<<<<<<<<
 *         return {"nullValue": None}
 *     elif isinstance(v, int):
 */
  }

  /* "viur/datastore/transport.pyx":164
 *     elif v is None:
 *         return {"nullValue": None}
 *     elif isinstance(v, int):             # <<<<<<<<<<<<<<
 *         return {
 *             "integerValue": str(v)
 */
  __pyx_t_1 = PyInt_Check(__pyx_v_v); 
  if (__pyx_t_1) {

    /* "viur/datastore/transport.pyx":165
 *         return {"nullValue": None}
 *     elif isinstance(v, int):
 *         return {             # <<<<<<<<<<<<<<
 *             "integerValue": str(v)
 *         }
 */
    __Pyx_XDECREF(__pyx_r);

    /* "viur/datastore/transport.pyx":166
 *     elif isinstance(v, int):
 *         return {
 *             "integerValue": str(v)             # <<<<<<<<<<<<<<
 *         }
 *     elif isinstance(v, float):
 */
    __pyx_t_3 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 166, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = __Pyx_PyObject_Str(__pyx_v_v); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 166, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    if (PyDict_SetItem(__pyx_t_3, __pyx_n_u_integerValue, __pyx_t_4) < 0) __PYX_ERR(0, 166, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = ((PyObject*)__pyx_t_3);
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "viur/datastore/transport.pyx":164
 *     elif v is None:
 *         return {"nullValue": None}
 *     elif isinstance(v, int):             # <<<<<<<<<<<<<<
 *         return {
 *             "integerValue": str(v)
 */
  }

  /* "viur/datastore/transport.pyx":168
 *             "integerValue": str(v)
 *         }
 *     elif isinstance(v, float):             # <<<<<<<<<<<<<<
 *         return {
 *             "doubleValue": v
 */
  __pyx_t_1 = PyFloat_Check(__pyx_v_v); 
  if (__pyx_t_1) {

    /* "viur/datastore/transport.pyx":169
 *         }
 *     elif isinstance(v, float):
 *         return {             # <<<<<<<<<<<<<<
 *             "doubleValue": v
 *         }
 */
    __Pyx_XDECREF(__pyx_r);

    /* "viur/datastore/transport.pyx":170
 *     elif isinstance(v, float):
 *         return {
 *             "doubleValue": v             # <<<<<<<<<<<<<<
 *         }
 *     elif isinstance(v, str):
 */
    __pyx_t_3 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 170, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (PyDict_SetItem(__pyx_t_3, __pyx_n_u_doubleValue, __pyx_v_v) < 0) __PYX_ERR(0, 170, __pyx_L1_error)
    __pyx_r = ((PyObject*)__pyx_t_3);
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "viur/datastore/transport.pyx":168
 *             "integerValue": str(v)
 *         }
 *     elif isinstance(v, float):             # <<<<<<<<<<<<<<
 *         return {
 *             "doubleValue": v
 */
  }

  /* "viur/datastore/transport.pyx":172
 *             "doubleValue": v
 *         }
 *     elif isinstance(v, str):             # <<<<<<<<<<<<<<
 *         return {
 *             "stringValue": v
 */
  __pyx_t_1 = PyUnicode_Check(__pyx_v_v); 
  if (__pyx_t_1) {

    /* "viur/datastore/transport.pyx":173
 *         }
 *     elif isinstance(v, str):
 *         return {             # <<<<<<<<<<<<<<
 *             "stringValue": v
 *         }
 */
    __Pyx_XDECREF(__pyx_r);

    /* "viur/datastore/transport.pyx":174
 *     elif isinstance(v, str):
 *         return {
 *             "stringValue": v             # <<<<<<<<<<<<<<
 *         }
 *     elif isinstance(v, Key):
 */
    __pyx_t_3 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 174, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (PyDict_SetItem(__pyx_t_3, __pyx_n_u_stringValue, __pyx_v_v) < 0) __PYX_ERR(0, 174, __pyx_L1_error)
    __pyx_r = ((PyObject*)__pyx_t_3);
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "viur/datastore/transport.pyx":172
 *             "doubleValue": v
 *         }
 *     elif isinstance(v, str):             # <<<<<<<<<<<<<<
 *         return {
 *             "stringValue": v
 */
  }

  /* "viur/datastore/transport.pyx":176
 *             "stringValue": v
 *         }
 *     elif isinstance(v, Key):             # <<<<<<<<<<<<<<
 *         return {
 *             "keyValue": {
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_Key); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 176, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_1 = PyObject_IsInstance(__pyx_v_v, __pyx_t_3); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 176, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (__pyx_t_1) {

    /* "viur/datastore/transport.pyx":177
 *         }
 *     elif isinstance(v, Key):
 *         return {             # <<<<<<<<<<<<<<
 *             "keyValue": {
 *                 "partitionId": {
 */
    __Pyx_XDECREF(__pyx_r);

    /* "viur/datastore/transport.pyx":178
 *     elif isinstance(v, Key):
 *         return {
 *             "keyValue": {             # <<<<<<<<<<<<<<
 *                 "partitionId": {
 *                     "project_id": projectID,
 */
    __pyx_t_3 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 178, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);

    /* "viur/datastore/transport.pyx":179
 *         return {
 *             "keyValue": {
 *                 "partitionId": {             # <<<<<<<<<<<<<<
 *                     "project_id": projectID,
 *                 },
 */
    __pyx_t_4 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 179, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);

    /* "viur/datastore/transport.pyx":180
 *             "keyValue": {
 *                 "partitionId": {
 *                     "project_id": projectID,             # <<<<<<<<<<<<<<
 *                 },
 *                 "path": keyToPath(v)
 */
    __pyx_t_5 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 180, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_projectID); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 180, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (PyDict_SetItem(__pyx_t_5, __pyx_n_u_project_id, __pyx_t_6) < 0) __PYX_ERR(0, 180, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (PyDict_SetItem(__pyx_t_4, __pyx_n_u_partitionId, __pyx_t_5) < 0) __PYX_ERR(0, 179, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "viur/datastore/transport.pyx":182
 *                     "project_id": projectID,
 *                 },
 *                 "path": keyToPath(v)             # <<<<<<<<<<<<<<
 *             }
 *         }
 */
    __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_keyToPath); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 182, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_7 = NULL;
    __pyx_t_8 = 0;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
        __pyx_t_8 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_v_v};
      __pyx_t_5 = __Pyx_PyObject_FastCall(__pyx_t_6, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 182, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    if (PyDict_SetItem(__pyx_t_4, __pyx_n_u_path, __pyx_t_5) < 0) __PYX_ERR(0, 179, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (PyDict_SetItem(__pyx_t_3, __pyx_n_u_keyValue, __pyx_t_4) < 0) __PYX_ERR(0, 178, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = ((PyObject*)__pyx_t_3);
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "viur/datastore/transport.pyx":176
 *             "stringValue": v
 *         }
 *     elif isinstance(v, Key):             # <<<<<<<<<<<<<<
 *         return {
 *             "keyValue": {
 */
  }

  /* "viur/datastore/transport.pyx":185
 *             }
 *         }
 *     elif isinstance(v, datetime):             # <<<<<<<<<<<<<<
 *         return {
 *             "timestampValue": v.astimezone(timezone.utc).strftime("%Y-%m-%dT%H:%M:%S.%fZ")
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_datetime); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 185, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_1 = PyObject_IsInstance(__pyx_v_v, __pyx_t_3); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 185, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (__pyx_t_1) {

    /* "viur/datastore/transport.pyx":186
 *         }
 *     elif isinstance(v, datetime):
 *         return {             # <<<<<<<<<<<<<<
 *             "timestampValue": v.astimezone(timezone.utc).strftime("%Y-%m-%dT%H:%M:%S.%fZ")
 *         }
 */
    __Pyx_XDECREF(__pyx_r);

    /* "viur/datastore/transport.pyx":187
 *     elif isinstance(v, datetime):
 *         return {
 *             "timestampValue": v.astimezone(timezone.utc).strftime("%Y-%m-%dT%H:%M:%S.%fZ")             # <<<<<<<<<<<<<<
 *         }
 *     elif isinstance(v, Entity):
 */
    __pyx_t_3 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 187, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_v, __pyx_n_s_astimezone); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 187, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_timezone); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 187, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_utc); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 187, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_7 = NULL;
    __pyx_t_8 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
        __pyx_t_8 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_t_9};
      __pyx_t_5 = __Pyx_PyObject_FastCall(__pyx_t_6, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 187, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_strftime); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 187, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = NULL;
    __pyx_t_8 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
        __pyx_t_8 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_5, __pyx_kp_u_Y_m_dT_H_M_S_fZ};
      __pyx_t_4 = __Pyx_PyObject_FastCall(__pyx_t_6, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 187, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    if (PyDict_SetItem(__pyx_t_3, __pyx_n_u_timestampValue, __pyx_t_4) < 0) __PYX_ERR(0, 187, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = ((PyObject*)__pyx_t_3);
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "viur/datastore/transport.pyx":185
 *             }
 *         }
 *     elif isinstance(v, datetime):             # <<<<<<<<<<<<<<
 *         return {
 *             "timestampValue": v.astimezone(timezone.utc).strftime("%Y-%m-%dT%H:%M:%S.%fZ")
 */
  }

  /* "viur/datastore/transport.pyx":189
 *             "timestampValue": v.astimezone(timezone.utc).strftime("%Y-%m-%dT%H:%M:%S.%fZ")
 *         }
 *     elif isinstance(v, Entity):             # <<<<<<<<<<<<<<
 *         excludedProperties: set = set(v.exclude_from_indexes)
 *         properties = {}
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_Entity); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 189, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_1 = PyObject_IsInstance(__pyx_v_v, __pyx_t_3); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 189, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (__pyx_t_1) {

    /* "viur/datastore/transport.pyx":190
 *         }
 *     elif isinstance(v, Entity):
 *         excludedProperties: set = set(v.exclude_from_indexes)             # <<<<<<<<<<<<<<
 *         properties = {}
 *         for dictKey, dictValue in v.items():
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_v, __pyx_n_s_exclude_from_indexes); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 190, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = PySet_New(__pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 190, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_excludedProperties = ((PyObject*)__pyx_t_4);
    __pyx_t_4 = 0;

    /* "viur/datastore/transport.pyx":191
 *     elif isinstance(v, Entity):
 *         excludedProperties: set = set(v.exclude_from_indexes)
 *         properties = {}             # <<<<<<<<<<<<<<
 *         for dictKey, dictValue in v.items():
 *             if isinstance(dictValue, list) and dictKey in excludedProperties:
 */
    __pyx_t_4 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 191, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_v_properties = ((PyObject*)__pyx_t_4);
    __pyx_t_4 = 0;

    /* "viur/datastore/transport.pyx":192
 *         excludedProperties: set = set(v.exclude_from_indexes)
 *         properties = {}
 *         for dictKey, dictValue in v.items():             # <<<<<<<<<<<<<<
 *             if isinstance(dictValue, list) and dictKey in excludedProperties:
 *                 # Lists cannot be not indexed (but each of its value can be). So we have to forward our not-indexed
 */
    __pyx_t_10 = 0;
    if (unlikely(__pyx_v_v == Py_None)) {
      PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "items");
      __PYX_ERR(0, 192, __pyx_L1_error)
    }
    __pyx_t_3 = __Pyx_dict_iterator(__pyx_v_v, 0, __pyx_n_s_items, (&__pyx_t_11), (&__pyx_t_8)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 192, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_XDECREF(__pyx_t_4);
    __pyx_t_4 = __pyx_t_3;
    __pyx_t_3 = 0;
    while (1) {
      __pyx_t_12 = __Pyx_dict_iter_next(__pyx_t_4, __pyx_t_11, &__pyx_t_10, &__pyx_t_3, &__pyx_t_6, NULL, __pyx_t_8);
      if (unlikely(__pyx_t_12 == 0)) break;
      if (unlikely(__pyx_t_12 == -1)) __PYX_ERR(0, 192, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_XDECREF_SET(__pyx_v_dictKey, __pyx_t_3);
      __pyx_t_3 = 0;
      __Pyx_XDECREF_SET(__pyx_v_dictValue, __pyx_t_6);
      __pyx_t_6 = 0;

      /* "viur/datastore/transport.pyx":193
 *         properties = {}
 *         for dictKey, dictValue in v.items():
 *             if isinstance(dictValue, list) and dictKey in excludedProperties:             # <<<<<<<<<<<<<<
 *                 # Lists cannot be not indexed (but each of its value can be). So we have to forward our not-indexed
 *                 # flag to our children. See https://github.com/googleapis/google-cloud-node/issues/2615.
 */
      __pyx_t_2 = PyList_Check(__pyx_v_dictValue); 
      if (__pyx_t_2) {
      } else {
        __pyx_t_1 = __pyx_t_2;
        goto __pyx_L9_bool_binop_done;
      }
      __pyx_t_2 = (__Pyx_PySet_ContainsTF(__pyx_v_dictKey, __pyx_v_excludedProperties, Py_EQ)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(0, 193, __pyx_L1_error)
      __pyx_t_1 = __pyx_t_2;
      __pyx_L9_bool_binop_done:;
      if (__pyx_t_1) {

        /* "viur/datastore/transport.pyx":196
 *                 # Lists cannot be not indexed (but each of its value can be). So we have to forward our not-indexed
 *                 # flag to our children. See https://github.com/googleapis/google-cloud-node/issues/2615.
 *                 resList = []             # <<<<<<<<<<<<<<
 *                 for value in dictValue:
 *                     resDict = pythonPropToJson(value)
 */
        __pyx_t_6 = PyList_New(0); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 196, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_XDECREF_SET(__pyx_v_resList, ((PyObject*)__pyx_t_6));
        __pyx_t_6 = 0;

        /* "viur/datastore/transport.pyx":197
 *                 # flag to our children. See https://github.com/googleapis/google-cloud-node/issues/2615.
 *                 resList = []
 *                 for value in dictValue:             # <<<<<<<<<<<<<<
 *                     resDict = pythonPropToJson(value)
 *                     resDict["excludeFromIndexes"] = True
 */
        if (likely(PyList_CheckExact(__pyx_v_dictValue)) || PyTuple_CheckExact(__pyx_v_dictValue)) {
          __pyx_t_6 = __pyx_v_dictValue; __Pyx_INCREF(__pyx_t_6);
          __pyx_t_13 = 0;
          __pyx_t_14 = NULL;
        } else {
          __pyx_t_13 = -1; __pyx_t_6 = PyObject_GetIter(__pyx_v_dictValue); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 197, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_6);
          __pyx_t_14 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_6); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 197, __pyx_L1_error)
        }
        for (;;) {
          if (likely(!__pyx_t_14)) {
            if (likely(PyList_CheckExact(__pyx_t_6))) {
              {
                Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_6);
                #if !CYTHON_ASSUME_SAFE_MACROS
                if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 197, __pyx_L1_error)
                #endif
                if (__pyx_t_13 >= __pyx_temp) break;
              }
              #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
              __pyx_t_3 = PyList_GET_ITEM(__pyx_t_6, __pyx_t_13); __Pyx_INCREF(__pyx_t_3); __pyx_t_13++; if (unlikely((0 < 0))) __PYX_ERR(0, 197, __pyx_L1_error)
              #else
              __pyx_t_3 = __Pyx_PySequence_ITEM(__pyx_t_6, __pyx_t_13); __pyx_t_13++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 197, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_3);
              #endif
            } else {
              {
                Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_6);
                #if !CYTHON_ASSUME_SAFE_MACROS
                if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 197, __pyx_L1_error)
                #endif
                if (__pyx_t_13 >= __pyx_temp) break;
              }
              #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
              __pyx_t_3 = PyTuple_GET_ITEM(__pyx_t_6, __pyx_t_13); __Pyx_INCREF(__pyx_t_3); __pyx_t_13++; if (unlikely((0 < 0))) __PYX_ERR(0, 197, __pyx_L1_error)
              #else
              __pyx_t_3 = __Pyx_PySequence_ITEM(__pyx_t_6, __pyx_t_13); __pyx_t_13++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 197, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_3);
              #endif
            }
          } else {
            __pyx_t_3 = __pyx_t_14(__pyx_t_6);
            if (unlikely(!__pyx_t_3)) {
              PyObject* exc_type = PyErr_Occurred();
              if (exc_type) {
                if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
                else __PYX_ERR(0, 197, __pyx_L1_error)
              }
              break;
            }
            __Pyx_GOTREF(__pyx_t_3);
          }
          __Pyx_XDECREF_SET(__pyx_v_value, __pyx_t_3);
          __pyx_t_3 = 0;

          /* "viur/datastore/transport.pyx":198
 *                 resList = []
 *                 for value in dictValue:
 *                     resDict = pythonPropToJson(value)             # <<<<<<<<<<<<<<
 *                     resDict["excludeFromIndexes"] = True
 *                     resList.append(resDict)
 */
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_pythonPropToJson); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 198, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_9 = NULL;
          __pyx_t_12 = 0;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_5);
            if (likely(__pyx_t_9)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
              __Pyx_INCREF(__pyx_t_9);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_5, function);
              __pyx_t_12 = 1;
            }
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_9, __pyx_v_value};
            __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_5, __pyx_callargs+1-__pyx_t_12, 1+__pyx_t_12);
            __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 198, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_3);
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          }
          __Pyx_XDECREF_SET(__pyx_v_resDict, __pyx_t_3);
          __pyx_t_3 = 0;

          /* "viur/datastore/transport.pyx":199
 *                 for value in dictValue:
 *                     resDict = pythonPropToJson(value)
 *                     resDict["excludeFromIndexes"] = True             # <<<<<<<<<<<<<<
 *                     resList.append(resDict)
 *                 properties[dictKey] = {
 */
          if (unlikely((PyObject_SetItem(__pyx_v_resDict, __pyx_n_u_excludeFromIndexes, Py_True) < 0))) __PYX_ERR(0, 199, __pyx_L1_error)

          /* "viur/datastore/transport.pyx":200
 *                     resDict = pythonPropToJson(value)
 *                     resDict["excludeFromIndexes"] = True
 *                     resList.append(resDict)             # <<<<<<<<<<<<<<
 *                 properties[dictKey] = {
 *                     "arrayValue": {
 */
          __pyx_t_15 = __Pyx_PyList_Append(__pyx_v_resList, __pyx_v_resDict); if (unlikely(__pyx_t_15 == ((int)-1))) __PYX_ERR(0, 200, __pyx_L1_error)

          /* "viur/datastore/transport.pyx":197
 *                 # flag to our children. See https://github.com/googleapis/google-cloud-node/issues/2615.
 *                 resList = []
 *                 for value in dictValue:             # <<<<<<<<<<<<<<
 *                     resDict = pythonPropToJson(value)
 *                     resDict["excludeFromIndexes"] = True
 */
        }
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

        /* "viur/datastore/transport.pyx":202
 *                     resList.append(resDict)
 *                 properties[dictKey] = {
 *                     "arrayValue": {             # <<<<<<<<<<<<<<
 *                         "values": resList
 *                     }
 */
        __pyx_t_6 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 202, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);

        /* "viur/datastore/transport.pyx":203
 *                 properties[dictKey] = {
 *                     "arrayValue": {
 *                         "values": resList             # <<<<<<<<<<<<<<
 *                     }
 *                 }
 */
        __pyx_t_3 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 203, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        if (PyDict_SetItem(__pyx_t_3, __pyx_n_u_values, __pyx_v_resList) < 0) __PYX_ERR(0, 203, __pyx_L1_error)
        if (PyDict_SetItem(__pyx_t_6, __pyx_n_u_arrayValue, __pyx_t_3) < 0) __PYX_ERR(0, 202, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

        /* "viur/datastore/transport.pyx":201
 *                     resDict["excludeFromIndexes"] = True
 *                     resList.append(resDict)
 *                 properties[dictKey] = {             # <<<<<<<<<<<<<<
 *                     "arrayValue": {
 *                         "values": resList
 */
        if (unlikely((PyDict_SetItem(__pyx_v_properties, __pyx_v_dictKey, __pyx_t_6) < 0))) __PYX_ERR(0, 201, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

        /* "viur/datastore/transport.pyx":193
 *         properties = {}
 *         for dictKey, dictValue in v.items():
 *             if isinstance(dictValue, list) and dictKey in excludedProperties:             # <<<<<<<<<<<<<<
 *                 # Lists cannot be not indexed (but each of its value can be). So we have to forward our not-indexed
 *                 # flag to our children. See https://github.com/googleapis/google-cloud-node/issues/2615.
 */
        goto __pyx_L8;
      }

      /* "viur/datastore/transport.pyx":207
 *                 }
 *             else:
 *                 resDict = pythonPropToJson(dictValue)             # <<<<<<<<<<<<<<
 *                 if dictKey in excludedProperties:
 *                     resDict["excludeFromIndexes"] = True
 */
      /*else*/ {
        __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_pythonPropToJson); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 207, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_5 = NULL;
        __pyx_t_12 = 0;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_5)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_5);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
            __pyx_t_12 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_5, __pyx_v_dictValue};
          __pyx_t_6 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_12, 1+__pyx_t_12);
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 207, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_6);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        }
        __Pyx_XDECREF_SET(__pyx_v_resDict, __pyx_t_6);
        __pyx_t_6 = 0;

        /* "viur/datastore/transport.pyx":208
 *             else:
 *                 resDict = pythonPropToJson(dictValue)
 *                 if dictKey in excludedProperties:             # <<<<<<<<<<<<<<
 *                     resDict["excludeFromIndexes"] = True
 *                 properties[dictKey] = resDict
 */
        __pyx_t_1 = (__Pyx_PySet_ContainsTF(__pyx_v_dictKey, __pyx_v_excludedProperties, Py_EQ)); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(0, 208, __pyx_L1_error)
        if (__pyx_t_1) {

          /* "viur/datastore/transport.pyx":209
 *                 resDict = pythonPropToJson(dictValue)
 *                 if dictKey in excludedProperties:
 *                     resDict["excludeFromIndexes"] = True             # <<<<<<<<<<<<<<
 *                 properties[dictKey] = resDict
 *         return {
 */
          if (unlikely((PyObject_SetItem(__pyx_v_resDict, __pyx_n_u_excludeFromIndexes, Py_True) < 0))) __PYX_ERR(0, 209, __pyx_L1_error)

          /* "viur/datastore/transport.pyx":208
 *             else:
 *                 resDict = pythonPropToJson(dictValue)
 *                 if dictKey in excludedProperties:             # <<<<<<<<<<<<<<
 *                     resDict["excludeFromIndexes"] = True
 *                 properties[dictKey] = resDict
 */
        }

        /* "viur/datastore/transport.pyx":210
 *                 if dictKey in excludedProperties:
 *                     resDict["excludeFromIndexes"] = True
 *                 properties[dictKey] = resDict             # <<<<<<<<<<<<<<
 *         return {
 *             "entityValue": {
 */
        if (unlikely((PyDict_SetItem(__pyx_v_properties, __pyx_v_dictKey, __pyx_v_resDict) < 0))) __PYX_ERR(0, 210, __pyx_L1_error)
      }
      __pyx_L8:;
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "viur/datastore/transport.pyx":211
 *                     resDict["excludeFromIndexes"] = True
 *                 properties[dictKey] = resDict
 *         return {             # <<<<<<<<<<<<<<
 *             "entityValue": {
 *                 "key": pythonPropToJson(v.key)["keyValue"] if isinstance(v, Entity) and v.key else None,
 */
    __Pyx_XDECREF(__pyx_r);

    /* "viur/datastore/transport.pyx":212
 *                 properties[dictKey] = resDict
 *         return {
 *             "entityValue": {             # <<<<<<<<<<<<<<
 *                 "key": pythonPropToJson(v.key)["keyValue"] if isinstance(v, Entity) and v.key else None,
 *                 "properties": properties
 */
    __pyx_t_4 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 212, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);

    /* "viur/datastore/transport.pyx":213
 *         return {
 *             "entityValue": {
 *                 "key": pythonPropToJson(v.key)["keyValue"] if isinstance(v, Entity) and v.key else None,             # <<<<<<<<<<<<<<
 *                 "properties": properties
 *             }
 */
    __pyx_t_6 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 213, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_Entity); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 213, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_2 = PyObject_IsInstance(__pyx_v_v, __pyx_t_5); if (unlikely(__pyx_t_2 == ((int)-1))) __PYX_ERR(0, 213, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (__pyx_t_2) {
    } else {
      __pyx_t_1 = __pyx_t_2;
      goto __pyx_L15_bool_binop_done;
    }
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_v, __pyx_n_s_key); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 213, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(0, 213, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_1 = __pyx_t_2;
    __pyx_L15_bool_binop_done:;
    if (__pyx_t_1) {
      __Pyx_GetModuleGlobalName(__pyx_t_9, __pyx_n_s_pythonPropToJson); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 213, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_v, __pyx_n_s_key); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 213, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_16 = NULL;
      __pyx_t_8 = 0;
      #if CYTHON_UNPACK_METHODS
      if (unlikely(PyMethod_Check(__pyx_t_9))) {
        __pyx_t_16 = PyMethod_GET_SELF(__pyx_t_9);
        if (likely(__pyx_t_16)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_9);
          __Pyx_INCREF(__pyx_t_16);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_9, function);
          __pyx_t_8 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_16, __pyx_t_7};
        __pyx_t_5 = __Pyx_PyObject_FastCall(__pyx_t_9, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
        __Pyx_XDECREF(__pyx_t_16); __pyx_t_16 = 0;
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 213, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      }
      __pyx_t_9 = __Pyx_PyObject_Dict_GetItem(__pyx_t_5, __pyx_n_u_keyValue); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 213, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_3 = __pyx_t_9;
      __pyx_t_9 = 0;
    } else {
      __Pyx_INCREF(Py_None);
      __pyx_t_3 = Py_None;
    }
    if (PyDict_SetItem(__pyx_t_6, __pyx_n_u_key, __pyx_t_3) < 0) __PYX_ERR(0, 213, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "viur/datastore/transport.pyx":214
 *             "entityValue": {
 *                 "key": pythonPropToJson(v.key)["keyValue"] if isinstance(v, Entity) and v.key else None,
 *                 "properties": properties             # <<<<<<<<<<<<<<
 *             }
 *         }
 */
    if (PyDict_SetItem(__pyx_t_6, __pyx_n_u_properties, __pyx_v_properties) < 0) __PYX_ERR(0, 213, __pyx_L1_error)
    if (PyDict_SetItem(__pyx_t_4, __pyx_n_u_entityValue, __pyx_t_6) < 0) __PYX_ERR(0, 212, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_r = ((PyObject*)__pyx_t_4);
    __pyx_t_4 = 0;
    goto __pyx_L0;

    /* "viur/datastore/transport.pyx":189
 *             "timestampValue": v.astimezone(timezone.utc).strftime("%Y-%m-%dT%H:%M:%S.%fZ")
 *         }
 *     elif isinstance(v, Entity):             # <<<<<<<<<<<<<<
 *         excludedProperties: set = set(v.exclude_from_indexes)
 *         properties = {}
 */
  }

  /* "viur/datastore/transport.pyx":217
 *             }
 *         }
 *     elif isinstance(v, dict):             # <<<<<<<<<<<<<<
 *         # We hande dicts separately as they don't have keys nor can keys be excluded from being indexed
 *         return {
 */
  __pyx_t_1 = PyDict_Check(__pyx_v_v); 
  if (__pyx_t_1) {

    /* "viur/datastore/transport.pyx":219
 *     elif isinstance(v, dict):
 *         # We hande dicts separately as they don't have keys nor can keys be excluded from being indexed
 *         return {             # <<<<<<<<<<<<<<
 *             "entityValue": {
 *                 "key": None,
 */
    __Pyx_XDECREF(__pyx_r);

    /* "viur/datastore/transport.pyx":220
 *         # We hande dicts separately as they don't have keys nor can keys be excluded from being indexed
 *         return {
 *             "entityValue": {             # <<<<<<<<<<<<<<
 *                 "key": None,
 *                 "properties": {
 */
    __pyx_t_4 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 220, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);

    /* "viur/datastore/transport.pyx":221
 *         return {
 *             "entityValue": {
 *                 "key": None,             # <<<<<<<<<<<<<<
 *                 "properties": {
 *                     dictKey: pythonPropToJson(dictValue) for dictKey, dictValue in v.items()
 */
    __pyx_t_6 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 221, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (PyDict_SetItem(__pyx_t_6, __pyx_n_u_key, Py_None) < 0) __PYX_ERR(0, 221, __pyx_L1_error)
    { /* enter inner scope */

      /* "viur/datastore/transport.pyx":222
 *             "entityValue": {
 *                 "key": None,
 *                 "properties": {             # <<<<<<<<<<<<<<
 *                     dictKey: pythonPropToJson(dictValue) for dictKey, dictValue in v.items()
 *                 }
 */
      __pyx_t_3 = PyDict_New(); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 222, __pyx_L19_error)
      __Pyx_GOTREF(__pyx_t_3);

      /* "viur/datastore/transport.pyx":223
 *                 "key": None,
 *                 "properties": {
 *                     dictKey: pythonPropToJson(dictValue) for dictKey, dictValue in v.items()             # <<<<<<<<<<<<<<
 *                 }
 *             }
 */
      __pyx_t_11 = 0;
      if (unlikely(__pyx_v_v == Py_None)) {
        PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "items");
        __PYX_ERR(0, 223, __pyx_L19_error)
      }
      __pyx_t_5 = __Pyx_dict_iterator(__pyx_v_v, 0, __pyx_n_s_items, (&__pyx_t_10), (&__pyx_t_8)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 223, __pyx_L19_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_XDECREF(__pyx_t_9);
      __pyx_t_9 = __pyx_t_5;
      __pyx_t_5 = 0;
      while (1) {
        __pyx_t_12 = __Pyx_dict_iter_next(__pyx_t_9, __pyx_t_10, &__pyx_t_11, &__pyx_t_5, &__pyx_t_7, NULL, __pyx_t_8);
        if (unlikely(__pyx_t_12 == 0)) break;
        if (unlikely(__pyx_t_12 == -1)) __PYX_ERR(0, 223, __pyx_L19_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_XDECREF_SET(__pyx_7genexpr__pyx_v_dictKey, __pyx_t_5);
        __pyx_t_5 = 0;
        __Pyx_XDECREF_SET(__pyx_7genexpr__pyx_v_dictValue, __pyx_t_7);
        __pyx_t_7 = 0;
        __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_pythonPropToJson); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 223, __pyx_L19_error)
        __Pyx_GOTREF(__pyx_t_5);
        __pyx_t_16 = NULL;
        __pyx_t_12 = 0;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_5))) {
          __pyx_t_16 = PyMethod_GET_SELF(__pyx_t_5);
          if (likely(__pyx_t_16)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_16);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_5, function);
            __pyx_t_12 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_16, __pyx_7genexpr__pyx_v_dictValue};
          __pyx_t_7 = __Pyx_PyObject_FastCall(__pyx_t_5, __pyx_callargs+1-__pyx_t_12, 1+__pyx_t_12);
          __Pyx_XDECREF(__pyx_t_16); __pyx_t_16 = 0;
          if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 223, __pyx_L19_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        }
        if (unlikely(PyDict_SetItem(__pyx_t_3, (PyObject*)__pyx_7genexpr__pyx_v_dictKey, (PyObject*)__pyx_t_7))) __PYX_ERR(0, 223, __pyx_L19_error)
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      }
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_XDECREF(__pyx_7genexpr__pyx_v_dictKey); __pyx_7genexpr__pyx_v_dictKey = 0;
      __Pyx_XDECREF(__pyx_7genexpr__pyx_v_dictValue); __pyx_7genexpr__pyx_v_dictValue = 0;
      goto __pyx_L22_exit_scope;
      __pyx_L19_error:;
      __Pyx_XDECREF(__pyx_7genexpr__pyx_v_dictKey); __pyx_7genexpr__pyx_v_dictKey = 0;
      __Pyx_XDECREF(__pyx_7genexpr__pyx_v_dictValue); __pyx_7genexpr__pyx_v_dictValue = 0;
      goto __pyx_L1_error;
      __pyx_L22_exit_scope:;
    } /* exit inner scope */
    if (PyDict_SetItem(__pyx_t_6, __pyx_n_u_properties, __pyx_t_3) < 0) __PYX_ERR(0, 221, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (PyDict_SetItem(__pyx_t_4, __pyx_n_u_entityValue, __pyx_t_6) < 0) __PYX_ERR(0, 220, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_r = ((PyObject*)__pyx_t_4);
    __pyx_t_4 = 0;
    goto __pyx_L0;

    /* "viur/datastore/transport.pyx":217
 *             }
 *         }
 *     elif isinstance(v, dict):             # <<<<<<<<<<<<<<
 *         # We hande dicts separately as they don't have keys nor can keys be excluded from being indexed
 *         return {
 */
  }

  /* "viur/datastore/transport.pyx":227
 *             }
 *         }
 *     elif isinstance(v, list):             # <<<<<<<<<<<<<<
 *         return {
 *             "arrayValue": {
 */
  __pyx_t_1 = PyList_Check(__pyx_v_v); 
  if (__pyx_t_1) {

    /* "viur/datastore/transport.pyx":228
 *         }
 *     elif isinstance(v, list):
 *         return {             # <<<<<<<<<<<<<<
 *             "arrayValue": {
 *                 "values": [pythonPropToJson(x) for x in v]
 */
    __Pyx_XDECREF(__pyx_r);

    /* "viur/datastore/transport.pyx":229
 *     elif isinstance(v, list):
 *         return {
 *             "arrayValue": {             # <<<<<<<<<<<<<<
 *                 "values": [pythonPropToJson(x) for x in v]
 *             }
 */
    __pyx_t_4 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 229, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);

    /* "viur/datastore/transport.pyx":230
 *         return {
 *             "arrayValue": {
 *                 "values": [pythonPropToJson(x) for x in v]             # <<<<<<<<<<<<<<
 *             }
 *         }
 */
    __pyx_t_6 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 230, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    { /* enter inner scope */
      __pyx_t_3 = PyList_New(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 230, __pyx_L25_error)
      __Pyx_GOTREF(__pyx_t_3);
      if (likely(PyList_CheckExact(__pyx_v_v)) || PyTuple_CheckExact(__pyx_v_v)) {
        __pyx_t_9 = __pyx_v_v; __Pyx_INCREF(__pyx_t_9);
        __pyx_t_10 = 0;
        __pyx_t_14 = NULL;
      } else {
        __pyx_t_10 = -1; __pyx_t_9 = PyObject_GetIter(__pyx_v_v); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 230, __pyx_L25_error)
        __Pyx_GOTREF(__pyx_t_9);
        __pyx_t_14 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_9); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 230, __pyx_L25_error)
      }
      for (;;) {
        if (likely(!__pyx_t_14)) {
          if (likely(PyList_CheckExact(__pyx_t_9))) {
            {
              Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_9);
              #if !CYTHON_ASSUME_SAFE_MACROS
              if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 230, __pyx_L25_error)
              #endif
              if (__pyx_t_10 >= __pyx_temp) break;
            }
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_7 = PyList_GET_ITEM(__pyx_t_9, __pyx_t_10); __Pyx_INCREF(__pyx_t_7); __pyx_t_10++; if (unlikely((0 < 0))) __PYX_ERR(0, 230, __pyx_L25_error)
            #else
            __pyx_t_7 = __Pyx_PySequence_ITEM(__pyx_t_9, __pyx_t_10); __pyx_t_10++; if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 230, __pyx_L25_error)
            __Pyx_GOTREF(__pyx_t_7);
            #endif
          } else {
            {
              Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_9);
              #if !CYTHON_ASSUME_SAFE_MACROS
              if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 230, __pyx_L25_error)
              #endif
              if (__pyx_t_10 >= __pyx_temp) break;
            }
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_7 = PyTuple_GET_ITEM(__pyx_t_9, __pyx_t_10); __Pyx_INCREF(__pyx_t_7); __pyx_t_10++; if (unlikely((0 < 0))) __PYX_ERR(0, 230, __pyx_L25_error)
            #else
            __pyx_t_7 = __Pyx_PySequence_ITEM(__pyx_t_9, __pyx_t_10); __pyx_t_10++; if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 230, __pyx_L25_error)
            __Pyx_GOTREF(__pyx_t_7);
            #endif
          }
        } else {
          __pyx_t_7 = __pyx_t_14(__pyx_t_9);
          if (unlikely(!__pyx_t_7)) {
            PyObject* exc_type = PyErr_Occurred();
            if (exc_type) {
              if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
              else __PYX_ERR(0, 230, __pyx_L25_error)
            }
            break;
          }
          __Pyx_GOTREF(__pyx_t_7);
        }
        __Pyx_XDECREF_SET(__pyx_8genexpr1__pyx_v_x, __pyx_t_7);
        __pyx_t_7 = 0;
        __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_pythonPropToJson); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 230, __pyx_L25_error)
        __Pyx_GOTREF(__pyx_t_5);
        __pyx_t_16 = NULL;
        __pyx_t_8 = 0;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_5))) {
          __pyx_t_16 = PyMethod_GET_SELF(__pyx_t_5);
          if (likely(__pyx_t_16)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_16);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_5, function);
            __pyx_t_8 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_16, __pyx_8genexpr1__pyx_v_x};
          __pyx_t_7 = __Pyx_PyObject_FastCall(__pyx_t_5, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
          __Pyx_XDECREF(__pyx_t_16); __pyx_t_16 = 0;
          if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 230, __pyx_L25_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        }
        if (unlikely(__Pyx_ListComp_Append(__pyx_t_3, (PyObject*)__pyx_t_7))) __PYX_ERR(0, 230, __pyx_L25_error)
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      }
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_XDECREF(__pyx_8genexpr1__pyx_v_x); __pyx_8genexpr1__pyx_v_x = 0;
      goto __pyx_L29_exit_scope;
      __pyx_L25_error:;
      __Pyx_XDECREF(__pyx_8genexpr1__pyx_v_x); __pyx_8genexpr1__pyx_v_x = 0;
      goto __pyx_L1_error;
      __pyx_L29_exit_scope:;
    } /* exit inner scope */
    if (PyDict_SetItem(__pyx_t_6, __pyx_n_u_values, __pyx_t_3) < 0) __PYX_ERR(0, 230, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (PyDict_SetItem(__pyx_t_4, __pyx_n_u_arrayValue, __pyx_t_6) < 0) __PYX_ERR(0, 229, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_r = ((PyObject*)__pyx_t_4);
    __pyx_t_4 = 0;
    goto __pyx_L0;

    /* "viur/datastore/transport.pyx":227
 *             }
 *         }
 *     elif isinstance(v, list):             # <<<<<<<<<<<<<<
 *         return {
 *             "arrayValue": {
 */
  }

  /* "viur/datastore/transport.pyx":233
 *             }
 *         }
 *     elif isinstance(v, bytes):             # <<<<<<<<<<<<<<
 *         return {
 *             "blobValue": b64encode(v).decode("ASCII")
 */
  __pyx_t_1 = PyBytes_Check(__pyx_v_v); 
  if (__pyx_t_1) {

    /* "viur/datastore/transport.pyx":234
 *         }
 *     elif isinstance(v, bytes):
 *         return {             # <<<<<<<<<<<<<<
 *             "blobValue": b64encode(v).decode("ASCII")
 *         }
 */
    __Pyx_XDECREF(__pyx_r);

    /* "viur/datastore/transport.pyx":235
 *     elif isinstance(v, bytes):
 *         return {
 *             "blobValue": b64encode(v).decode("ASCII")             # <<<<<<<<<<<<<<
 *         }
 *     assert False, "%s (%s) is not supported" % (v, type(v))
 */
    __pyx_t_4 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 235, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_GetModuleGlobalName(__pyx_t_9, __pyx_n_s_b64encode); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 235, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_7 = NULL;
    __pyx_t_8 = 0;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_9))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_9);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_9);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_9, function);
        __pyx_t_8 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_v_v};
      __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_9, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 235, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    }
    __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_decode); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 235, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = NULL;
    __pyx_t_8 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_9))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_9);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_9);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_9, function);
        __pyx_t_8 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_n_u_ASCII};
      __pyx_t_6 = __Pyx_PyObject_FastCall(__pyx_t_9, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 235, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    }
    if (PyDict_SetItem(__pyx_t_4, __pyx_n_u_blobValue, __pyx_t_6) < 0) __PYX_ERR(0, 235, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_r = ((PyObject*)__pyx_t_4);
    __pyx_t_4 = 0;
    goto __pyx_L0;

    /* "viur/datastore/transport.pyx":233
 *             }
 *         }
 *     elif isinstance(v, bytes):             # <<<<<<<<<<<<<<
 *         return {
 *             "blobValue": b64encode(v).decode("ASCII")
 */
  }

  /* "viur/datastore/transport.pyx":237
 *             "blobValue": b64encode(v).decode("ASCII")
 *         }
 *     assert False, "%s (%s) is not supported" % (v, type(v))             # <<<<<<<<<<<<<<
 * 
 * cdef inline object toPyStr(stringView strView):
 */
  #ifndef CYTHON_WITHOUT_ASSERTIONS
  if (unlikely(__pyx_assertions_enabled())) {
    if (unlikely(!0)) {
      __pyx_t_4 = PyTuple_New(4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 237, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_10 = 0;
      __pyx_t_17 = 127;
      __pyx_t_6 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Unicode(__pyx_v_v), __pyx_empty_unicode); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 237, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_17 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_6) > __pyx_t_17) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_6) : __pyx_t_17;
      __pyx_t_10 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_6);
      __Pyx_GIVEREF(__pyx_t_6);
      PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_6);
      __pyx_t_6 = 0;
      __Pyx_INCREF(__pyx_kp_u_);
      __pyx_t_10 += 2;
      __Pyx_GIVEREF(__pyx_kp_u_);
      PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_kp_u_);
      __pyx_t_6 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Unicode(((PyObject *)Py_TYPE(__pyx_v_v))), __pyx_empty_unicode); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 237, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_17 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_6) > __pyx_t_17) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_6) : __pyx_t_17;
      __pyx_t_10 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_6);
      __Pyx_GIVEREF(__pyx_t_6);
      PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_t_6);
      __pyx_t_6 = 0;
      __Pyx_INCREF(__pyx_kp_u_is_not_supported);
      __pyx_t_10 += 18;
      __Pyx_GIVEREF(__pyx_kp_u_is_not_supported);
      PyTuple_SET_ITEM(__pyx_t_4, 3, __pyx_kp_u_is_not_supported);
      __pyx_t_6 = __Pyx_PyUnicode_Join(__pyx_t_4, 4, __pyx_t_10, __pyx_t_17); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 237, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_Raise(__pyx_builtin_AssertionError, __pyx_t_6, 0, 0);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __PYX_ERR(0, 237, __pyx_L1_error)
    }
  }
  #else
  if ((1)); else __PYX_ERR(0, 237, __pyx_L1_error)
  #endif

  /* "viur/datastore/transport.pyx":150
 *     return res
 * 
 * def pythonPropToJson(v) -> dict:             # <<<<<<<<<<<<<<
 *     """
 *         Convert a python type to the object representation expected by the datastore rest API.
 */

  /* function exit code */
  __pyx_r = ((PyObject*)Py_None); __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_16);
  __Pyx_AddTraceback("viur.datastore.transport.pythonPropToJson", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_excludedProperties);
  __Pyx_XDECREF(__pyx_v_properties);
  __Pyx_XDECREF(__pyx_v_dictKey);
  __Pyx_XDECREF(__pyx_v_dictValue);
  __Pyx_XDECREF(__pyx_v_resList);
  __Pyx_XDECREF(__pyx_v_value);
  __Pyx_XDECREF(__pyx_v_resDict);
  __Pyx_XDECREF(__pyx_7genexpr__pyx_v_dictKey);
  __Pyx_XDECREF(__pyx_7genexpr__pyx_v_dictValue);
  __Pyx_XDECREF(__pyx_8genexpr1__pyx_v_x);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "viur/datastore/transport.pyx":239
 *     assert False, "%s (%s) is not supported" % (v, type(v))
 * 
 * cdef inline object toPyStr(stringView strView):             # <<<<<<<<<<<<<<
 *     """
 *         Converts a cpp stringview to a python str object
 */

static CYTHON_INLINE PyObject *__pyx_f_4viur_9datastore_9transport_toPyStr(std::string_view __pyx_v_strView) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("toPyStr", 1);

  /* "viur/datastore/transport.pyx":245
 *         :return: The corresponding python string object
 *     """
 *     return PyUnicode_FromStringAndSize(             # <<<<<<<<<<<<<<
 *         strView.data(),
 *         strView.length()
 */
  __Pyx_XDECREF(__pyx_r);

  /* "viur/datastore/transport.pyx":247
 *     return PyUnicode_FromStringAndSize(
 *         strView.data(),
 *         strView.length()             # <<<<<<<<<<<<<<
 *     )
 * 
 */
  __pyx_t_1 = PyUnicode_FromStringAndSize(__pyx_v_strView.data(), __pyx_v_strView.length()); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 245, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "viur/datastore/transport.pyx":239
 *     assert False, "%s (%s) is not supported" % (v, type(v))
 * 
 * cdef inline object toPyStr(stringView strView):             # <<<<<<<<<<<<<<
 *     """
 *         Converts a cpp stringview to a python str object
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("viur.datastore.transport.toPyStr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "viur/datastore/transport.pyx":250
 *     )
 * 
 * cdef inline object parseKey(simdjsonElement v):             # <<<<<<<<<<<<<<
 *     """
 *         Parses a simdJsonObject representing a key to a datastore.Key instance.
 */

static CYTHON_INLINE PyObject *__pyx_f_4viur_9datastore_9transport_parseKey(simdjson::dom::element __pyx_v_v) {
  simdjson::dom::array __pyx_v_arr;
  simdjson::dom::array::iterator __pyx_v_arrayIt;
  simdjson::dom::array::iterator __pyx_v_arrayItEnd;
  PyObject *__pyx_v_pathArgs = NULL;
  simdjson::dom::element __pyx_v_element;
  PyObject *__pyx_v_key = NULL;
  PyObject *__pyx_v_pathElem = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  simdjson::dom::element __pyx_t_2;
  simdjson::dom::array __pyx_t_3;
  int __pyx_t_4;
  std::string_view __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  int __pyx_t_9;
  Py_ssize_t __pyx_t_10;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("parseKey", 1);

  /* "viur/datastore/transport.pyx":259
 *     cdef simdjsonArray arr
 *     cdef simdjsonArray.iterator arrayIt, arrayItEnd
 *     pathArgs = []             # <<<<<<<<<<<<<<
 *     arr = v.at_key("path").get_array()
 *     arrayIt = arr.begin()
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 259, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_pathArgs = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "viur/datastore/transport.pyx":260
 *     cdef simdjsonArray.iterator arrayIt, arrayItEnd
 *     pathArgs = []
 *     arr = v.at_key("path").get_array()             # <<<<<<<<<<<<<<
 *     arrayIt = arr.begin()
 *     arrayItEnd = arr.end()
 */
  try {
    __pyx_t_2 = __pyx_v_v.at_key(((char const *)"path"));
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 260, __pyx_L1_error)
  }
  try {
    __pyx_t_3 = __pyx_t_2.get_array();
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 260, __pyx_L1_error)
  }
  __pyx_v_arr = __PYX_STD_MOVE_IF_SUPPORTED(__pyx_t_3);

  /* "viur/datastore/transport.pyx":261
 *     pathArgs = []
 *     arr = v.at_key("path").get_array()
 *     arrayIt = arr.begin()             # <<<<<<<<<<<<<<
 *     arrayItEnd = arr.end()
 *     while arrayIt != arrayItEnd:
 */
  __pyx_v_arrayIt = __pyx_v_arr.begin();

  /* "viur/datastore/transport.pyx":262
 *     arr = v.at_key("path").get_array()
 *     arrayIt = arr.begin()
 *     arrayItEnd = arr.end()             # <<<<<<<<<<<<<<
 *     while arrayIt != arrayItEnd:
 *         element = dereference(arrayIt)
 */
  __pyx_v_arrayItEnd = __pyx_v_arr.end();

  /* "viur/datastore/transport.pyx":263
 *     arrayIt = arr.begin()
 *     arrayItEnd = arr.end()
 *     while arrayIt != arrayItEnd:             # <<<<<<<<<<<<<<
 *         element = dereference(arrayIt)
 *         if element.at_pointer("/id").error() == SUCCESS:
 */
  while (1) {
    __pyx_t_4 = (__pyx_v_arrayIt != __pyx_v_arrayItEnd);
    if (!__pyx_t_4) break;

    /* "viur/datastore/transport.pyx":264
 *     arrayItEnd = arr.end()
 *     while arrayIt != arrayItEnd:
 *         element = dereference(arrayIt)             # <<<<<<<<<<<<<<
 *         if element.at_pointer("/id").error() == SUCCESS:
 *             pathArgs.append(
 */
    __pyx_v_element = (*__pyx_v_arrayIt);

    /* "viur/datastore/transport.pyx":265
 *     while arrayIt != arrayItEnd:
 *         element = dereference(arrayIt)
 *         if element.at_pointer("/id").error() == SUCCESS:             # <<<<<<<<<<<<<<
 *             pathArgs.append(
 *                 (toPyStr(element.at_key("kind").get_string()), int(toPyStr(element.at_key("id").get_string())))
 */
    __pyx_t_4 = (__pyx_v_element.at_pointer(((char const *)"/id")).error() == simdjson::error_code::SUCCESS);
    if (__pyx_t_4) {

      /* "viur/datastore/transport.pyx":267
 *         if element.at_pointer("/id").error() == SUCCESS:
 *             pathArgs.append(
 *                 (toPyStr(element.at_key("kind").get_string()), int(toPyStr(element.at_key("id").get_string())))             # <<<<<<<<<<<<<<
 *             )
 *         elif element.at_pointer("/name").error() == SUCCESS:
 */
      try {
        __pyx_t_2 = __pyx_v_element.at_key(((char const *)"kind"));
      } catch(...) {
        __Pyx_CppExn2PyErr();
        __PYX_ERR(0, 267, __pyx_L1_error)
      }
      try {
        __pyx_t_5 = __pyx_t_2.get_string();
      } catch(...) {
        __Pyx_CppExn2PyErr();
        __PYX_ERR(0, 267, __pyx_L1_error)
      }
      __pyx_t_1 = __pyx_f_4viur_9datastore_9transport_toPyStr(__PYX_STD_MOVE_IF_SUPPORTED(__pyx_t_5)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 267, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      try {
        __pyx_t_2 = __pyx_v_element.at_key(((char const *)"id"));
      } catch(...) {
        __Pyx_CppExn2PyErr();
        __PYX_ERR(0, 267, __pyx_L1_error)
      }
      try {
        __pyx_t_5 = __pyx_t_2.get_string();
      } catch(...) {
        __Pyx_CppExn2PyErr();
        __PYX_ERR(0, 267, __pyx_L1_error)
      }
      __pyx_t_6 = __pyx_f_4viur_9datastore_9transport_toPyStr(__PYX_STD_MOVE_IF_SUPPORTED(__pyx_t_5)); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 267, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_7 = __Pyx_PyNumber_Int(__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 267, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_6 = PyTuple_New(2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 267, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_GIVEREF(__pyx_t_1);
      if (__Pyx_PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_1)) __PYX_ERR(0, 267, __pyx_L1_error);
      __Pyx_GIVEREF(__pyx_t_7);
      if (__Pyx_PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_t_7)) __PYX_ERR(0, 267, __pyx_L1_error);
      __pyx_t_1 = 0;
      __pyx_t_7 = 0;

      /* "viur/datastore/transport.pyx":266
 *         element = dereference(arrayIt)
 *         if element.at_pointer("/id").error() == SUCCESS:
 *             pathArgs.append(             # <<<<<<<<<<<<<<
 *                 (toPyStr(element.at_key("kind").get_string()), int(toPyStr(element.at_key("id").get_string())))
 *             )
 */
      __pyx_t_8 = __Pyx_PyList_Append(__pyx_v_pathArgs, __pyx_t_6); if (unlikely(__pyx_t_8 == ((int)-1))) __PYX_ERR(0, 266, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

      /* "viur/datastore/transport.pyx":265
 *     while arrayIt != arrayItEnd:
 *         element = dereference(arrayIt)
 *         if element.at_pointer("/id").error() == SUCCESS:             # <<<<<<<<<<<<<<
 *             pathArgs.append(
 *                 (toPyStr(element.at_key("kind").get_string()), int(toPyStr(element.at_key("id").get_string())))
 */
      goto __pyx_L5;
    }

    /* "viur/datastore/transport.pyx":269
 *                 (toPyStr(element.at_key("kind").get_string()), int(toPyStr(element.at_key("id").get_string())))
 *             )
 *         elif element.at_pointer("/name").error() == SUCCESS:             # <<<<<<<<<<<<<<
 *             pathArgs.append(
 *                 (toPyStr(element.at_key("kind").get_string()), toPyStr(element.at_key("name").get_string()))
 */
    __pyx_t_4 = (__pyx_v_element.at_pointer(((char const *)"/name")).error() == simdjson::error_code::SUCCESS);
    if (__pyx_t_4) {

      /* "viur/datastore/transport.pyx":271
 *         elif element.at_pointer("/name").error() == SUCCESS:
 *             pathArgs.append(
 *                 (toPyStr(element.at_key("kind").get_string()), toPyStr(element.at_key("name").get_string()))             # <<<<<<<<<<<<<<
 *             )
 *         else:
 */
      try {
        __pyx_t_2 = __pyx_v_element.at_key(((char const *)"kind"));
      } catch(...) {
        __Pyx_CppExn2PyErr();
        __PYX_ERR(0, 271, __pyx_L1_error)
      }
      try {
        __pyx_t_5 = __pyx_t_2.get_string();
      } catch(...) {
        __Pyx_CppExn2PyErr();
        __PYX_ERR(0, 271, __pyx_L1_error)
      }
      __pyx_t_6 = __pyx_f_4viur_9datastore_9transport_toPyStr(__PYX_STD_MOVE_IF_SUPPORTED(__pyx_t_5)); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 271, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      try {
        __pyx_t_2 = __pyx_v_element.at_key(((char const *)"name"));
      } catch(...) {
        __Pyx_CppExn2PyErr();
        __PYX_ERR(0, 271, __pyx_L1_error)
      }
      try {
        __pyx_t_5 = __pyx_t_2.get_string();
      } catch(...) {
        __Pyx_CppExn2PyErr();
        __PYX_ERR(0, 271, __pyx_L1_error)
      }
      __pyx_t_7 = __pyx_f_4viur_9datastore_9transport_toPyStr(__PYX_STD_MOVE_IF_SUPPORTED(__pyx_t_5)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 271, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 271, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_GIVEREF(__pyx_t_6);
      if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_6)) __PYX_ERR(0, 271, __pyx_L1_error);
      __Pyx_GIVEREF(__pyx_t_7);
      if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_7)) __PYX_ERR(0, 271, __pyx_L1_error);
      __pyx_t_6 = 0;
      __pyx_t_7 = 0;

      /* "viur/datastore/transport.pyx":270
 *             )
 *         elif element.at_pointer("/name").error() == SUCCESS:
 *             pathArgs.append(             # <<<<<<<<<<<<<<
 *                 (toPyStr(element.at_key("kind").get_string()), toPyStr(element.at_key("name").get_string()))
 *             )
 */
      __pyx_t_8 = __Pyx_PyList_Append(__pyx_v_pathArgs, __pyx_t_1); if (unlikely(__pyx_t_8 == ((int)-1))) __PYX_ERR(0, 270, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "viur/datastore/transport.pyx":269
 *                 (toPyStr(element.at_key("kind").get_string()), int(toPyStr(element.at_key("id").get_string())))
 *             )
 *         elif element.at_pointer("/name").error() == SUCCESS:             # <<<<<<<<<<<<<<
 *             pathArgs.append(
 *                 (toPyStr(element.at_key("kind").get_string()), toPyStr(element.at_key("name").get_string()))
 */
      goto __pyx_L5;
    }

    /* "viur/datastore/transport.pyx":276
 *             # We read an incomplete/null key from the datastore. This is likely a bug.
 *             # Can happen if you manually created a key like datastore.Key("kind", 0).
 *             logging.error("We read an incomplete/null key from the datastore. This is likely a bug!")             # <<<<<<<<<<<<<<
 *             pathArgs.append((toPyStr(element.at_key("kind").get_string()), 0))
 *         preincrement(arrayIt)
 */
    /*else*/ {
      __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_logging); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 276, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_error); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 276, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __pyx_t_7 = NULL;
      __pyx_t_9 = 0;
      #if CYTHON_UNPACK_METHODS
      if (unlikely(PyMethod_Check(__pyx_t_6))) {
        __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
        if (likely(__pyx_t_7)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
          __Pyx_INCREF(__pyx_t_7);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_6, function);
          __pyx_t_9 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_kp_u_We_read_an_incomplete_null_key_f};
        __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_6, __pyx_callargs+1-__pyx_t_9, 1+__pyx_t_9);
        __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
        if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 276, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "viur/datastore/transport.pyx":277
 *             # Can happen if you manually created a key like datastore.Key("kind", 0).
 *             logging.error("We read an incomplete/null key from the datastore. This is likely a bug!")
 *             pathArgs.append((toPyStr(element.at_key("kind").get_string()), 0))             # <<<<<<<<<<<<<<
 *         preincrement(arrayIt)
 *     key = None
 */
      try {
        __pyx_t_2 = __pyx_v_element.at_key(((char const *)"kind"));
      } catch(...) {
        __Pyx_CppExn2PyErr();
        __PYX_ERR(0, 277, __pyx_L1_error)
      }
      try {
        __pyx_t_5 = __pyx_t_2.get_string();
      } catch(...) {
        __Pyx_CppExn2PyErr();
        __PYX_ERR(0, 277, __pyx_L1_error)
      }
      __pyx_t_1 = __pyx_f_4viur_9datastore_9transport_toPyStr(__PYX_STD_MOVE_IF_SUPPORTED(__pyx_t_5)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 277, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_6 = PyTuple_New(2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 277, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_GIVEREF(__pyx_t_1);
      if (__Pyx_PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_1)) __PYX_ERR(0, 277, __pyx_L1_error);
      __Pyx_INCREF(__pyx_int_0);
      __Pyx_GIVEREF(__pyx_int_0);
      if (__Pyx_PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_int_0)) __PYX_ERR(0, 277, __pyx_L1_error);
      __pyx_t_1 = 0;
      __pyx_t_8 = __Pyx_PyList_Append(__pyx_v_pathArgs, __pyx_t_6); if (unlikely(__pyx_t_8 == ((int)-1))) __PYX_ERR(0, 277, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __pyx_L5:;

    /* "viur/datastore/transport.pyx":278
 *             logging.error("We read an incomplete/null key from the datastore. This is likely a bug!")
 *             pathArgs.append((toPyStr(element.at_key("kind").get_string()), 0))
 *         preincrement(arrayIt)             # <<<<<<<<<<<<<<
 *     key = None
 *     for pathElem in pathArgs:
 */
    (void)((++__pyx_v_arrayIt));
  }

  /* "viur/datastore/transport.pyx":279
 *             pathArgs.append((toPyStr(element.at_key("kind").get_string()), 0))
 *         preincrement(arrayIt)
 *     key = None             # <<<<<<<<<<<<<<
 *     for pathElem in pathArgs:
 *         key = Key(*pathElem, parent=key)
 */
  __Pyx_INCREF(Py_None);
  __pyx_v_key = Py_None;

  /* "viur/datastore/transport.pyx":280
 *         preincrement(arrayIt)
 *     key = None
 *     for pathElem in pathArgs:             # <<<<<<<<<<<<<<
 *         key = Key(*pathElem, parent=key)
 *     return key
 */
  __pyx_t_6 = __pyx_v_pathArgs; __Pyx_INCREF(__pyx_t_6);
  __pyx_t_10 = 0;
  for (;;) {
    {
      Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_6);
      #if !CYTHON_ASSUME_SAFE_MACROS
      if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 280, __pyx_L1_error)
      #endif
      if (__pyx_t_10 >= __pyx_temp) break;
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_1 = PyList_GET_ITEM(__pyx_t_6, __pyx_t_10); __Pyx_INCREF(__pyx_t_1); __pyx_t_10++; if (unlikely((0 < 0))) __PYX_ERR(0, 280, __pyx_L1_error)
    #else
    __pyx_t_1 = __Pyx_PySequence_ITEM(__pyx_t_6, __pyx_t_10); __pyx_t_10++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 280, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    #endif
    __Pyx_XDECREF_SET(__pyx_v_pathElem, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "viur/datastore/transport.pyx":281
 *     key = None
 *     for pathElem in pathArgs:
 *         key = Key(*pathElem, parent=key)             # <<<<<<<<<<<<<<
 *     return key
 * 
 */
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_Key); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 281, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_7 = __Pyx_PySequence_Tuple(__pyx_v_pathElem); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 281, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_11 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 281, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    if (PyDict_SetItem(__pyx_t_11, __pyx_n_s_parent, __pyx_v_key) < 0) __PYX_ERR(0, 281, __pyx_L1_error)
    __pyx_t_12 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_7, __pyx_t_11); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 281, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_12);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __Pyx_DECREF_SET(__pyx_v_key, __pyx_t_12);
    __pyx_t_12 = 0;

    /* "viur/datastore/transport.pyx":280
 *         preincrement(arrayIt)
 *     key = None
 *     for pathElem in pathArgs:             # <<<<<<<<<<<<<<
 *         key = Key(*pathElem, parent=key)
 *     return key
 */
  }
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "viur/datastore/transport.pyx":282
 *     for pathElem in pathArgs:
 *         key = Key(*pathElem, parent=key)
 *     return key             # <<<<<<<<<<<<<<
 * 
 * cdef inline object toEntityStructure(simdjsonElement v, boolean_type isInitial = False):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_key);
  __pyx_r = __pyx_v_key;
  goto __pyx_L0;

  /* "viur/datastore/transport.pyx":250
 *     )
 * 
 * cdef inline object parseKey(simdjsonElement v):             # <<<<<<<<<<<<<<
 *     """
 *         Parses a simdJsonObject representing a key to a datastore.Key instance.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_AddTraceback("viur.datastore.transport.parseKey", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_pathArgs);
  __Pyx_XDECREF(__pyx_v_key);
  __Pyx_XDECREF(__pyx_v_pathElem);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "viur/datastore/transport.pyx":284
 *     return key
 * 
 * cdef inline object toEntityStructure(simdjsonElement v, boolean_type isInitial = False):             # <<<<<<<<<<<<<<
 *     """
 *         Parses a simdJsonElement into the corresponding python datatypes.
 */

static CYTHON_INLINE PyObject *__pyx_f_4viur_9datastore_9transport_toEntityStructure(simdjson::dom::element __pyx_v_v, struct __pyx_opt_args_4viur_9datastore_9transport_toEntityStructure *__pyx_optional_args) {
  bool __pyx_v_isInitial = ((bool)0);
  enum simdjson::dom::element_type __pyx_v_et;
  simdjson::dom::array::iterator __pyx_v_arrayIt;
  simdjson::dom::array::iterator __pyx_v_arrayItEnd;
  simdjson::dom::object::iterator __pyx_v_objIterStart;
  simdjson::dom::object::iterator __pyx_v_objIterStartInner;
  simdjson::dom::object::iterator __pyx_v_objIterEnd;
  simdjson::dom::object::iterator __pyx_v_objIterEndInner;
  simdjson::dom::element __pyx_v_element;
  simdjson::dom::object __pyx_v_outerObject;
  simdjson::dom::object __pyx_v_innerObject;
  std::string_view __pyx_v_strView;
  simdjson::simdjson_result<simdjson::dom::element> __pyx_v_tmpResult;
  PyObject *__pyx_v_e = NULL;
  PyObject *__pyx_v_excludeList = NULL;
  simdjson::dom::array __pyx_v_arr;
  int __pyx_v_allExcluded;
  PyObject *__pyx_v_dateStr = NULL;
  PyObject *__pyx_v_res = NULL;
  PyObject *__pyx_v_entity = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  enum simdjson::dom::element_type __pyx_t_1;
  simdjson::dom::object __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  simdjson::dom::array __pyx_t_9;
  bool __pyx_t_10;
  int __pyx_t_11;
  std::string_view __pyx_t_12;
  double __pyx_t_13;
  PyObject *__pyx_t_14 = NULL;
  simdjson::dom::element __pyx_t_15;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("toEntityStructure", 1);
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_isInitial = __pyx_optional_args->isInitial;
    }
  }

  /* "viur/datastore/transport.pyx":294
 *             to avoid additional loops over objects.
 *     """
 *     cdef simdjsonElementType et = v.type()             # <<<<<<<<<<<<<<
 *     cdef simdjsonArray.iterator arrayIt, arrayItEnd
 *     cdef simdjsonObject.iterator objIterStart, objIterStartInner, objIterEnd, objIterEndInner
 */
  try {
    __pyx_t_1 = __pyx_v_v.type();
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 294, __pyx_L1_error)
  }
  __pyx_v_et = __pyx_t_1;

  /* "viur/datastore/transport.pyx":301
 *     cdef stringView strView, bytesView
 *     cdef simdjsonResult tmpResult
 *     if et == OBJECT:             # <<<<<<<<<<<<<<
 *         outerObject = v.get_object()
 *         objIterStart = outerObject.begin()
 */
  switch (__pyx_v_et) {
    case simdjson::dom::element_type::OBJECT:

    /* "viur/datastore/transport.pyx":302
 *     cdef simdjsonResult tmpResult
 *     if et == OBJECT:
 *         outerObject = v.get_object()             # <<<<<<<<<<<<<<
 *         objIterStart = outerObject.begin()
 *         objIterEnd = outerObject.end()
 */
    try {
      __pyx_t_2 = __pyx_v_v.get_object();
    } catch(...) {
      __Pyx_CppExn2PyErr();
      __PYX_ERR(0, 302, __pyx_L1_error)
    }
    __pyx_v_outerObject = __PYX_STD_MOVE_IF_SUPPORTED(__pyx_t_2);

    /* "viur/datastore/transport.pyx":303
 *     if et == OBJECT:
 *         outerObject = v.get_object()
 *         objIterStart = outerObject.begin()             # <<<<<<<<<<<<<<
 *         objIterEnd = outerObject.end()
 *         while objIterStart != objIterEnd:
 */
    __pyx_v_objIterStart = __pyx_v_outerObject.begin();

    /* "viur/datastore/transport.pyx":304
 *         outerObject = v.get_object()
 *         objIterStart = outerObject.begin()
 *         objIterEnd = outerObject.end()             # <<<<<<<<<<<<<<
 *         while objIterStart != objIterEnd:
 *             strView = objIterStart.key()
 */
    __pyx_v_objIterEnd = __pyx_v_outerObject.end();

    /* "viur/datastore/transport.pyx":305
 *         objIterStart = outerObject.begin()
 *         objIterEnd = outerObject.end()
 *         while objIterStart != objIterEnd:             # <<<<<<<<<<<<<<
 *             strView = objIterStart.key()
 *             if strView.compare("entity") == 0 or strView.compare("entityValue") == 0:
 */
    while (1) {
      __pyx_t_3 = (__pyx_v_objIterStart != __pyx_v_objIterEnd);
      if (!__pyx_t_3) break;

      /* "viur/datastore/transport.pyx":306
 *         objIterEnd = outerObject.end()
 *         while objIterStart != objIterEnd:
 *             strView = objIterStart.key()             # <<<<<<<<<<<<<<
 *             if strView.compare("entity") == 0 or strView.compare("entityValue") == 0:
 *                 e = Entity()
 */
      __pyx_v_strView = __pyx_v_objIterStart.key();

      /* "viur/datastore/transport.pyx":307
 *         while objIterStart != objIterEnd:
 *             strView = objIterStart.key()
 *             if strView.compare("entity") == 0 or strView.compare("entityValue") == 0:             # <<<<<<<<<<<<<<
 *                 e = Entity()
 *                 excludeList = set()
 */
      __pyx_t_4 = (__pyx_v_strView.compare(((char *)"entity")) == 0);
      if (!__pyx_t_4) {
      } else {
        __pyx_t_3 = __pyx_t_4;
        goto __pyx_L6_bool_binop_done;
      }
      __pyx_t_4 = (__pyx_v_strView.compare(((char *)"entityValue")) == 0);
      __pyx_t_3 = __pyx_t_4;
      __pyx_L6_bool_binop_done:;
      if (__pyx_t_3) {

        /* "viur/datastore/transport.pyx":308
 *             strView = objIterStart.key()
 *             if strView.compare("entity") == 0 or strView.compare("entityValue") == 0:
 *                 e = Entity()             # <<<<<<<<<<<<<<
 *                 excludeList = set()
 *                 tmpResult = objIterStart.value().at_pointer("/key")
 */
        __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_Entity); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 308, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __pyx_t_7 = NULL;
        __pyx_t_8 = 0;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_6))) {
          __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
          if (likely(__pyx_t_7)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
            __Pyx_INCREF(__pyx_t_7);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_6, function);
            __pyx_t_8 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_7, NULL};
          __pyx_t_5 = __Pyx_PyObject_FastCall(__pyx_t_6, __pyx_callargs+1-__pyx_t_8, 0+__pyx_t_8);
          __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
          if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 308, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        }
        __pyx_v_e = __pyx_t_5;
        __pyx_t_5 = 0;

        /* "viur/datastore/transport.pyx":309
 *             if strView.compare("entity") == 0 or strView.compare("entityValue") == 0:
 *                 e = Entity()
 *                 excludeList = set()             # <<<<<<<<<<<<<<
 *                 tmpResult = objIterStart.value().at_pointer("/key")
 *                 if tmpResult.error() == SUCCESS:
 */
        __pyx_t_5 = PySet_New(0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 309, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __pyx_v_excludeList = ((PyObject*)__pyx_t_5);
        __pyx_t_5 = 0;

        /* "viur/datastore/transport.pyx":310
 *                 e = Entity()
 *                 excludeList = set()
 *                 tmpResult = objIterStart.value().at_pointer("/key")             # <<<<<<<<<<<<<<
 *                 if tmpResult.error() == SUCCESS:
 *                     e.key = parseKey(tmpResult.value())
 */
        __pyx_v_tmpResult = __pyx_v_objIterStart.value().at_pointer(((char const *)"/key"));

        /* "viur/datastore/transport.pyx":311
 *                 excludeList = set()
 *                 tmpResult = objIterStart.value().at_pointer("/key")
 *                 if tmpResult.error() == SUCCESS:             # <<<<<<<<<<<<<<
 *                     e.key = parseKey(tmpResult.value())
 *                 tmpResult = objIterStart.value().at_pointer("/properties")
 */
        __pyx_t_3 = (__pyx_v_tmpResult.error() == simdjson::error_code::SUCCESS);
        if (__pyx_t_3) {

          /* "viur/datastore/transport.pyx":312
 *                 tmpResult = objIterStart.value().at_pointer("/key")
 *                 if tmpResult.error() == SUCCESS:
 *                     e.key = parseKey(tmpResult.value())             # <<<<<<<<<<<<<<
 *                 tmpResult = objIterStart.value().at_pointer("/properties")
 *                 if tmpResult.error() == SUCCESS:
 */
          __pyx_t_5 = __pyx_f_4viur_9datastore_9transport_parseKey(__pyx_v_tmpResult.value()); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 312, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_e, __pyx_n_s_key, __pyx_t_5) < 0) __PYX_ERR(0, 312, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

          /* "viur/datastore/transport.pyx":311
 *                 excludeList = set()
 *                 tmpResult = objIterStart.value().at_pointer("/key")
 *                 if tmpResult.error() == SUCCESS:             # <<<<<<<<<<<<<<
 *                     e.key = parseKey(tmpResult.value())
 *                 tmpResult = objIterStart.value().at_pointer("/properties")
 */
        }

        /* "viur/datastore/transport.pyx":313
 *                 if tmpResult.error() == SUCCESS:
 *                     e.key = parseKey(tmpResult.value())
 *                 tmpResult = objIterStart.value().at_pointer("/properties")             # <<<<<<<<<<<<<<
 *                 if tmpResult.error() == SUCCESS:
 *                     innerObject = tmpResult.value().get_object()
 */
        __pyx_v_tmpResult = __pyx_v_objIterStart.value().at_pointer(((char const *)"/properties"));

        /* "viur/datastore/transport.pyx":314
 *                     e.key = parseKey(tmpResult.value())
 *                 tmpResult = objIterStart.value().at_pointer("/properties")
 *                 if tmpResult.error() == SUCCESS:             # <<<<<<<<<<<<<<
 *                     innerObject = tmpResult.value().get_object()
 *                     objIterStartInner = innerObject.begin()
 */
        __pyx_t_3 = (__pyx_v_tmpResult.error() == simdjson::error_code::SUCCESS);
        if (__pyx_t_3) {

          /* "viur/datastore/transport.pyx":315
 *                 tmpResult = objIterStart.value().at_pointer("/properties")
 *                 if tmpResult.error() == SUCCESS:
 *                     innerObject = tmpResult.value().get_object()             # <<<<<<<<<<<<<<
 *                     objIterStartInner = innerObject.begin()
 *                     objIterEndInner = innerObject.end()
 */
          try {
            __pyx_t_2 = __pyx_v_tmpResult.value().get_object();
          } catch(...) {
            __Pyx_CppExn2PyErr();
            __PYX_ERR(0, 315, __pyx_L1_error)
          }
          __pyx_v_innerObject = __PYX_STD_MOVE_IF_SUPPORTED(__pyx_t_2);

          /* "viur/datastore/transport.pyx":316
 *                 if tmpResult.error() == SUCCESS:
 *                     innerObject = tmpResult.value().get_object()
 *                     objIterStartInner = innerObject.begin()             # <<<<<<<<<<<<<<
 *                     objIterEndInner = innerObject.end()
 *                     while objIterStartInner != objIterEndInner:
 */
          __pyx_v_objIterStartInner = __pyx_v_innerObject.begin();

          /* "viur/datastore/transport.pyx":317
 *                     innerObject = tmpResult.value().get_object()
 *                     objIterStartInner = innerObject.begin()
 *                     objIterEndInner = innerObject.end()             # <<<<<<<<<<<<<<
 *                     while objIterStartInner != objIterEndInner:
 *                         e[toPyStr(objIterStartInner.key())] = toEntityStructure(objIterStartInner.value())
 */
          __pyx_v_objIterEndInner = __pyx_v_innerObject.end();

          /* "viur/datastore/transport.pyx":318
 *                     objIterStartInner = innerObject.begin()
 *                     objIterEndInner = innerObject.end()
 *                     while objIterStartInner != objIterEndInner:             # <<<<<<<<<<<<<<
 *                         e[toPyStr(objIterStartInner.key())] = toEntityStructure(objIterStartInner.value())
 *                         if objIterStartInner.value().at_pointer("/arrayValue/values").error() == SUCCESS:
 */
          while (1) {
            __pyx_t_3 = (__pyx_v_objIterStartInner != __pyx_v_objIterEndInner);
            if (!__pyx_t_3) break;

            /* "viur/datastore/transport.pyx":319
 *                     objIterEndInner = innerObject.end()
 *                     while objIterStartInner != objIterEndInner:
 *                         e[toPyStr(objIterStartInner.key())] = toEntityStructure(objIterStartInner.value())             # <<<<<<<<<<<<<<
 *                         if objIterStartInner.value().at_pointer("/arrayValue/values").error() == SUCCESS:
 *                             # We have to collect the non-indexed flag from the children of lists
 */
            __pyx_t_5 = __pyx_f_4viur_9datastore_9transport_toEntityStructure(__pyx_v_objIterStartInner.value(), NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 319, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_5);
            __pyx_t_6 = __pyx_f_4viur_9datastore_9transport_toPyStr(__pyx_v_objIterStartInner.key()); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 319, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_6);
            if (unlikely((PyObject_SetItem(__pyx_v_e, __pyx_t_6, __pyx_t_5) < 0))) __PYX_ERR(0, 319, __pyx_L1_error)
            __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

            /* "viur/datastore/transport.pyx":320
 *                     while objIterStartInner != objIterEndInner:
 *                         e[toPyStr(objIterStartInner.key())] = toEntityStructure(objIterStartInner.value())
 *                         if objIterStartInner.value().at_pointer("/arrayValue/values").error() == SUCCESS:             # <<<<<<<<<<<<<<
 *                             # We have to collect the non-indexed flag from the children of lists
 *                             arr = objIterStartInner.value().at_pointer("/arrayValue/values").value().get_array()
 */
            __pyx_t_3 = (__pyx_v_objIterStartInner.value().at_pointer(((char const *)"/arrayValue/values")).error() == simdjson::error_code::SUCCESS);
            if (__pyx_t_3) {

              /* "viur/datastore/transport.pyx":322
 *                         if objIterStartInner.value().at_pointer("/arrayValue/values").error() == SUCCESS:
 *                             # We have to collect the non-indexed flag from the children of lists
 *                             arr = objIterStartInner.value().at_pointer("/arrayValue/values").value().get_array()             # <<<<<<<<<<<<<<
 *                             arrayIt = arr.begin()
 *                             arrayItEnd = arr.end()
 */
              try {
                __pyx_t_9 = __pyx_v_objIterStartInner.value().at_pointer(((char const *)"/arrayValue/values")).value().get_array();
              } catch(...) {
                __Pyx_CppExn2PyErr();
                __PYX_ERR(0, 322, __pyx_L1_error)
              }
              __pyx_v_arr = __PYX_STD_MOVE_IF_SUPPORTED(__pyx_t_9);

              /* "viur/datastore/transport.pyx":323
 *                             # We have to collect the non-indexed flag from the children of lists
 *                             arr = objIterStartInner.value().at_pointer("/arrayValue/values").value().get_array()
 *                             arrayIt = arr.begin()             # <<<<<<<<<<<<<<
 *                             arrayItEnd = arr.end()
 *                             allExcluded = True
 */
              __pyx_v_arrayIt = __pyx_v_arr.begin();

              /* "viur/datastore/transport.pyx":324
 *                             arr = objIterStartInner.value().at_pointer("/arrayValue/values").value().get_array()
 *                             arrayIt = arr.begin()
 *                             arrayItEnd = arr.end()             # <<<<<<<<<<<<<<
 *                             allExcluded = True
 *                             while arrayIt != arrayItEnd:
 */
              __pyx_v_arrayItEnd = __pyx_v_arr.end();

              /* "viur/datastore/transport.pyx":325
 *                             arrayIt = arr.begin()
 *                             arrayItEnd = arr.end()
 *                             allExcluded = True             # <<<<<<<<<<<<<<
 *                             while arrayIt != arrayItEnd:
 *                                 element = dereference(arrayIt)
 */
              __pyx_v_allExcluded = 1;

              /* "viur/datastore/transport.pyx":326
 *                             arrayItEnd = arr.end()
 *                             allExcluded = True
 *                             while arrayIt != arrayItEnd:             # <<<<<<<<<<<<<<
 *                                 element = dereference(arrayIt)
 *                                 tmpResult = element.at_pointer("/excludeFromIndexes")
 */
              while (1) {
                __pyx_t_3 = (__pyx_v_arrayIt != __pyx_v_arrayItEnd);
                if (!__pyx_t_3) break;

                /* "viur/datastore/transport.pyx":327
 *                             allExcluded = True
 *                             while arrayIt != arrayItEnd:
 *                                 element = dereference(arrayIt)             # <<<<<<<<<<<<<<
 *                                 tmpResult = element.at_pointer("/excludeFromIndexes")
 *                                 if tmpResult.error() != SUCCESS or not tmpResult.value().get_bool():
 */
                __pyx_v_element = (*__pyx_v_arrayIt);

                /* "viur/datastore/transport.pyx":328
 *                             while arrayIt != arrayItEnd:
 *                                 element = dereference(arrayIt)
 *                                 tmpResult = element.at_pointer("/excludeFromIndexes")             # <<<<<<<<<<<<<<
 *                                 if tmpResult.error() != SUCCESS or not tmpResult.value().get_bool():
 *                                     allExcluded = False
 */
                __pyx_v_tmpResult = __pyx_v_element.at_pointer(((char const *)"/excludeFromIndexes"));

                /* "viur/datastore/transport.pyx":329
 *                                 element = dereference(arrayIt)
 *                                 tmpResult = element.at_pointer("/excludeFromIndexes")
 *                                 if tmpResult.error() != SUCCESS or not tmpResult.value().get_bool():             # <<<<<<<<<<<<<<
 *                                     allExcluded = False
 *                                     break
 */
                __pyx_t_4 = (__pyx_v_tmpResult.error() != simdjson::error_code::SUCCESS);
                if (!__pyx_t_4) {
                } else {
                  __pyx_t_3 = __pyx_t_4;
                  goto __pyx_L16_bool_binop_done;
                }
                try {
                  __pyx_t_10 = __pyx_v_tmpResult.value().get_bool();
                } catch(...) {
                  __Pyx_CppExn2PyErr();
                  __PYX_ERR(0, 329, __pyx_L1_error)
                }
                __pyx_t_4 = (!(__pyx_t_10 != 0));
                __pyx_t_3 = __pyx_t_4;
                __pyx_L16_bool_binop_done:;
                if (__pyx_t_3) {

                  /* "viur/datastore/transport.pyx":330
 *                                 tmpResult = element.at_pointer("/excludeFromIndexes")
 *                                 if tmpResult.error() != SUCCESS or not tmpResult.value().get_bool():
 *                                     allExcluded = False             # <<<<<<<<<<<<<<
 *                                     break
 *                                 preincrement(arrayIt)
 */
                  __pyx_v_allExcluded = 0;

                  /* "viur/datastore/transport.pyx":331
 *                                 if tmpResult.error() != SUCCESS or not tmpResult.value().get_bool():
 *                                     allExcluded = False
 *                                     break             # <<<<<<<<<<<<<<
 *                                 preincrement(arrayIt)
 *                             if allExcluded:
 */
                  goto __pyx_L14_break;

                  /* "viur/datastore/transport.pyx":329
 *                                 element = dereference(arrayIt)
 *                                 tmpResult = element.at_pointer("/excludeFromIndexes")
 *                                 if tmpResult.error() != SUCCESS or not tmpResult.value().get_bool():             # <<<<<<<<<<<<<<
 *                                     allExcluded = False
 *                                     break
 */
                }

                /* "viur/datastore/transport.pyx":332
 *                                     allExcluded = False
 *                                     break
 *                                 preincrement(arrayIt)             # <<<<<<<<<<<<<<
 *                             if allExcluded:
 *                                 excludeList.add(toPyStr(objIterStartInner.key()))
 */
                (void)((++__pyx_v_arrayIt));
              }
              __pyx_L14_break:;

              /* "viur/datastore/transport.pyx":333
 *                                     break
 *                                 preincrement(arrayIt)
 *                             if allExcluded:             # <<<<<<<<<<<<<<
 *                                 excludeList.add(toPyStr(objIterStartInner.key()))
 *                         else:
 */
              if (__pyx_v_allExcluded) {

                /* "viur/datastore/transport.pyx":334
 *                                 preincrement(arrayIt)
 *                             if allExcluded:
 *                                 excludeList.add(toPyStr(objIterStartInner.key()))             # <<<<<<<<<<<<<<
 *                         else:
 *                             # For *all* other datatypes, we can simply check the dict it's defined in
 */
                __pyx_t_5 = __pyx_f_4viur_9datastore_9transport_toPyStr(__pyx_v_objIterStartInner.key()); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 334, __pyx_L1_error)
                __Pyx_GOTREF(__pyx_t_5);
                __pyx_t_11 = PySet_Add(__pyx_v_excludeList, __pyx_t_5); if (unlikely(__pyx_t_11 == ((int)-1))) __PYX_ERR(0, 334, __pyx_L1_error)
                __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

                /* "viur/datastore/transport.pyx":333
 *                                     break
 *                                 preincrement(arrayIt)
 *                             if allExcluded:             # <<<<<<<<<<<<<<
 *                                 excludeList.add(toPyStr(objIterStartInner.key()))
 *                         else:
 */
              }

              /* "viur/datastore/transport.pyx":320
 *                     while objIterStartInner != objIterEndInner:
 *                         e[toPyStr(objIterStartInner.key())] = toEntityStructure(objIterStartInner.value())
 *                         if objIterStartInner.value().at_pointer("/arrayValue/values").error() == SUCCESS:             # <<<<<<<<<<<<<<
 *                             # We have to collect the non-indexed flag from the children of lists
 *                             arr = objIterStartInner.value().at_pointer("/arrayValue/values").value().get_array()
 */
              goto __pyx_L12;
            }

            /* "viur/datastore/transport.pyx":337
 *                         else:
 *                             # For *all* other datatypes, we can simply check the dict it's defined in
 *                             tmpResult = objIterStartInner.value().at_pointer("/excludeFromIndexes")             # <<<<<<<<<<<<<<
 *                             if tmpResult.error() == SUCCESS and tmpResult.value().get_bool():
 *                                 excludeList.add(toPyStr(objIterStartInner.key()))
 */
            /*else*/ {
              __pyx_v_tmpResult = __pyx_v_objIterStartInner.value().at_pointer(((char const *)"/excludeFromIndexes"));

              /* "viur/datastore/transport.pyx":338
 *                             # For *all* other datatypes, we can simply check the dict it's defined in
 *                             tmpResult = objIterStartInner.value().at_pointer("/excludeFromIndexes")
 *                             if tmpResult.error() == SUCCESS and tmpResult.value().get_bool():             # <<<<<<<<<<<<<<
 *                                 excludeList.add(toPyStr(objIterStartInner.key()))
 *                         preincrement(objIterStartInner)
 */
              __pyx_t_4 = (__pyx_v_tmpResult.error() == simdjson::error_code::SUCCESS);
              if (__pyx_t_4) {
              } else {
                __pyx_t_3 = __pyx_t_4;
                goto __pyx_L20_bool_binop_done;
              }
              try {
                __pyx_t_10 = __pyx_v_tmpResult.value().get_bool();
              } catch(...) {
                __Pyx_CppExn2PyErr();
                __PYX_ERR(0, 338, __pyx_L1_error)
              }
              __pyx_t_4 = (__pyx_t_10 != 0);
              __pyx_t_3 = __pyx_t_4;
              __pyx_L20_bool_binop_done:;
              if (__pyx_t_3) {

                /* "viur/datastore/transport.pyx":339
 *                             tmpResult = objIterStartInner.value().at_pointer("/excludeFromIndexes")
 *                             if tmpResult.error() == SUCCESS and tmpResult.value().get_bool():
 *                                 excludeList.add(toPyStr(objIterStartInner.key()))             # <<<<<<<<<<<<<<
 *                         preincrement(objIterStartInner)
 *                 e.exclude_from_indexes = excludeList
 */
                __pyx_t_5 = __pyx_f_4viur_9datastore_9transport_toPyStr(__pyx_v_objIterStartInner.key()); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 339, __pyx_L1_error)
                __Pyx_GOTREF(__pyx_t_5);
                __pyx_t_11 = PySet_Add(__pyx_v_excludeList, __pyx_t_5); if (unlikely(__pyx_t_11 == ((int)-1))) __PYX_ERR(0, 339, __pyx_L1_error)
                __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

                /* "viur/datastore/transport.pyx":338
 *                             # For *all* other datatypes, we can simply check the dict it's defined in
 *                             tmpResult = objIterStartInner.value().at_pointer("/excludeFromIndexes")
 *                             if tmpResult.error() == SUCCESS and tmpResult.value().get_bool():             # <<<<<<<<<<<<<<
 *                                 excludeList.add(toPyStr(objIterStartInner.key()))
 *                         preincrement(objIterStartInner)
 */
              }
            }
            __pyx_L12:;

            /* "viur/datastore/transport.pyx":340
 *                             if tmpResult.error() == SUCCESS and tmpResult.value().get_bool():
 *                                 excludeList.add(toPyStr(objIterStartInner.key()))
 *                         preincrement(objIterStartInner)             # <<<<<<<<<<<<<<
 *                 e.exclude_from_indexes = excludeList
 *                 return e
 */
            (void)((++__pyx_v_objIterStartInner));
          }

          /* "viur/datastore/transport.pyx":314
 *                     e.key = parseKey(tmpResult.value())
 *                 tmpResult = objIterStart.value().at_pointer("/properties")
 *                 if tmpResult.error() == SUCCESS:             # <<<<<<<<<<<<<<
 *                     innerObject = tmpResult.value().get_object()
 *                     objIterStartInner = innerObject.begin()
 */
        }

        /* "viur/datastore/transport.pyx":341
 *                                 excludeList.add(toPyStr(objIterStartInner.key()))
 *                         preincrement(objIterStartInner)
 *                 e.exclude_from_indexes = excludeList             # <<<<<<<<<<<<<<
 *                 return e
 *             elif (strView.compare("nullValue") == 0):
 */
        if (__Pyx_PyObject_SetAttrStr(__pyx_v_e, __pyx_n_s_exclude_from_indexes, __pyx_v_excludeList) < 0) __PYX_ERR(0, 341, __pyx_L1_error)

        /* "viur/datastore/transport.pyx":342
 *                         preincrement(objIterStartInner)
 *                 e.exclude_from_indexes = excludeList
 *                 return e             # <<<<<<<<<<<<<<
 *             elif (strView.compare("nullValue") == 0):
 *                 return None
 */
        __Pyx_XDECREF(__pyx_r);
        __Pyx_INCREF(__pyx_v_e);
        __pyx_r = __pyx_v_e;
        goto __pyx_L0;

        /* "viur/datastore/transport.pyx":307
 *         while objIterStart != objIterEnd:
 *             strView = objIterStart.key()
 *             if strView.compare("entity") == 0 or strView.compare("entityValue") == 0:             # <<<<<<<<<<<<<<
 *                 e = Entity()
 *                 excludeList = set()
 */
      }

      /* "viur/datastore/transport.pyx":343
 *                 e.exclude_from_indexes = excludeList
 *                 return e
 *             elif (strView.compare("nullValue") == 0):             # <<<<<<<<<<<<<<
 *                 return None
 *             elif (strView.compare("stringValue") == 0):
 */
      __pyx_t_3 = (__pyx_v_strView.compare(((char *)"nullValue")) == 0);
      if (__pyx_t_3) {

        /* "viur/datastore/transport.pyx":344
 *                 return e
 *             elif (strView.compare("nullValue") == 0):
 *                 return None             # <<<<<<<<<<<<<<
 *             elif (strView.compare("stringValue") == 0):
 *                 return toPyStr(objIterStart.value().get_string())
 */
        __Pyx_XDECREF(__pyx_r);
        __pyx_r = Py_None; __Pyx_INCREF(Py_None);
        goto __pyx_L0;

        /* "viur/datastore/transport.pyx":343
 *                 e.exclude_from_indexes = excludeList
 *                 return e
 *             elif (strView.compare("nullValue") == 0):             # <<<<<<<<<<<<<<
 *                 return None
 *             elif (strView.compare("stringValue") == 0):
 */
      }

      /* "viur/datastore/transport.pyx":345
 *             elif (strView.compare("nullValue") == 0):
 *                 return None
 *             elif (strView.compare("stringValue") == 0):             # <<<<<<<<<<<<<<
 *                 return toPyStr(objIterStart.value().get_string())
 *             elif (strView.compare("doubleValue") == 0):
 */
      __pyx_t_3 = (__pyx_v_strView.compare(((char *)"stringValue")) == 0);
      if (__pyx_t_3) {

        /* "viur/datastore/transport.pyx":346
 *                 return None
 *             elif (strView.compare("stringValue") == 0):
 *                 return toPyStr(objIterStart.value().get_string())             # <<<<<<<<<<<<<<
 *             elif (strView.compare("doubleValue") == 0):
 *                 return objIterStart.value().get_double()
 */
        __Pyx_XDECREF(__pyx_r);
        try {
          __pyx_t_12 = __pyx_v_objIterStart.value().get_string();
        } catch(...) {
          __Pyx_CppExn2PyErr();
          __PYX_ERR(0, 346, __pyx_L1_error)
        }
        __pyx_t_5 = __pyx_f_4viur_9datastore_9transport_toPyStr(__PYX_STD_MOVE_IF_SUPPORTED(__pyx_t_12)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 346, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __pyx_r = __pyx_t_5;
        __pyx_t_5 = 0;
        goto __pyx_L0;

        /* "viur/datastore/transport.pyx":345
 *             elif (strView.compare("nullValue") == 0):
 *                 return None
 *             elif (strView.compare("stringValue") == 0):             # <<<<<<<<<<<<<<
 *                 return toPyStr(objIterStart.value().get_string())
 *             elif (strView.compare("doubleValue") == 0):
 */
      }

      /* "viur/datastore/transport.pyx":347
 *             elif (strView.compare("stringValue") == 0):
 *                 return toPyStr(objIterStart.value().get_string())
 *             elif (strView.compare("doubleValue") == 0):             # <<<<<<<<<<<<<<
 *                 return objIterStart.value().get_double()
 *             elif (strView.compare("integerValue") == 0):
 */
      __pyx_t_3 = (__pyx_v_strView.compare(((char *)"doubleValue")) == 0);
      if (__pyx_t_3) {

        /* "viur/datastore/transport.pyx":348
 *                 return toPyStr(objIterStart.value().get_string())
 *             elif (strView.compare("doubleValue") == 0):
 *                 return objIterStart.value().get_double()             # <<<<<<<<<<<<<<
 *             elif (strView.compare("integerValue") == 0):
 *                 return int(toPyStr(objIterStart.value().get_string()))
 */
        __Pyx_XDECREF(__pyx_r);
        try {
          __pyx_t_13 = __pyx_v_objIterStart.value().get_double();
        } catch(...) {
          __Pyx_CppExn2PyErr();
          __PYX_ERR(0, 348, __pyx_L1_error)
        }
        __pyx_t_5 = PyFloat_FromDouble(__pyx_t_13); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 348, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __pyx_r = __pyx_t_5;
        __pyx_t_5 = 0;
        goto __pyx_L0;

        /* "viur/datastore/transport.pyx":347
 *             elif (strView.compare("stringValue") == 0):
 *                 return toPyStr(objIterStart.value().get_string())
 *             elif (strView.compare("doubleValue") == 0):             # <<<<<<<<<<<<<<
 *                 return objIterStart.value().get_double()
 *             elif (strView.compare("integerValue") == 0):
 */
      }

      /* "viur/datastore/transport.pyx":349
 *             elif (strView.compare("doubleValue") == 0):
 *                 return objIterStart.value().get_double()
 *             elif (strView.compare("integerValue") == 0):             # <<<<<<<<<<<<<<
 *                 return int(toPyStr(objIterStart.value().get_string()))
 *             elif (strView.compare("arrayValue") == 0):
 */
      __pyx_t_3 = (__pyx_v_strView.compare(((char *)"integerValue")) == 0);
      if (__pyx_t_3) {

        /* "viur/datastore/transport.pyx":350
 *                 return objIterStart.value().get_double()
 *             elif (strView.compare("integerValue") == 0):
 *                 return int(toPyStr(objIterStart.value().get_string()))             # <<<<<<<<<<<<<<
 *             elif (strView.compare("arrayValue") == 0):
 *                 tmpResult = objIterStart.value().at_pointer("/values")
 */
        __Pyx_XDECREF(__pyx_r);
        try {
          __pyx_t_12 = __pyx_v_objIterStart.value().get_string();
        } catch(...) {
          __Pyx_CppExn2PyErr();
          __PYX_ERR(0, 350, __pyx_L1_error)
        }
        __pyx_t_5 = __pyx_f_4viur_9datastore_9transport_toPyStr(__PYX_STD_MOVE_IF_SUPPORTED(__pyx_t_12)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 350, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __pyx_t_6 = __Pyx_PyNumber_Int(__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 350, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        __pyx_r = __pyx_t_6;
        __pyx_t_6 = 0;
        goto __pyx_L0;

        /* "viur/datastore/transport.pyx":349
 *             elif (strView.compare("doubleValue") == 0):
 *                 return objIterStart.value().get_double()
 *             elif (strView.compare("integerValue") == 0):             # <<<<<<<<<<<<<<
 *                 return int(toPyStr(objIterStart.value().get_string()))
 *             elif (strView.compare("arrayValue") == 0):
 */
      }

      /* "viur/datastore/transport.pyx":351
 *             elif (strView.compare("integerValue") == 0):
 *                 return int(toPyStr(objIterStart.value().get_string()))
 *             elif (strView.compare("arrayValue") == 0):             # <<<<<<<<<<<<<<
 *                 tmpResult = objIterStart.value().at_pointer("/values")
 *                 if tmpResult.error() == SUCCESS:
 */
      __pyx_t_3 = (__pyx_v_strView.compare(((char *)"arrayValue")) == 0);
      if (__pyx_t_3) {

        /* "viur/datastore/transport.pyx":352
 *                 return int(toPyStr(objIterStart.value().get_string()))
 *             elif (strView.compare("arrayValue") == 0):
 *                 tmpResult = objIterStart.value().at_pointer("/values")             # <<<<<<<<<<<<<<
 *                 if tmpResult.error() == SUCCESS:
 *                     return toEntityStructure(tmpResult.value())
 */
        __pyx_v_tmpResult = __pyx_v_objIterStart.value().at_pointer(((char const *)"/values"));

        /* "viur/datastore/transport.pyx":353
 *             elif (strView.compare("arrayValue") == 0):
 *                 tmpResult = objIterStart.value().at_pointer("/values")
 *                 if tmpResult.error() == SUCCESS:             # <<<<<<<<<<<<<<
 *                     return toEntityStructure(tmpResult.value())
 *                 else:
 */
        __pyx_t_3 = (__pyx_v_tmpResult.error() == simdjson::error_code::SUCCESS);
        if (__pyx_t_3) {

          /* "viur/datastore/transport.pyx":354
 *                 tmpResult = objIterStart.value().at_pointer("/values")
 *                 if tmpResult.error() == SUCCESS:
 *                     return toEntityStructure(tmpResult.value())             # <<<<<<<<<<<<<<
 *                 else:
 *                     return []
 */
          __Pyx_XDECREF(__pyx_r);
          __pyx_t_6 = __pyx_f_4viur_9datastore_9transport_toEntityStructure(__pyx_v_tmpResult.value(), NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 354, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_6);
          __pyx_r = __pyx_t_6;
          __pyx_t_6 = 0;
          goto __pyx_L0;

          /* "viur/datastore/transport.pyx":353
 *             elif (strView.compare("arrayValue") == 0):
 *                 tmpResult = objIterStart.value().at_pointer("/values")
 *                 if tmpResult.error() == SUCCESS:             # <<<<<<<<<<<<<<
 *                     return toEntityStructure(tmpResult.value())
 *                 else:
 */
        }

        /* "viur/datastore/transport.pyx":356
 *                     return toEntityStructure(tmpResult.value())
 *                 else:
 *                     return []             # <<<<<<<<<<<<<<
 *             elif (strView.compare("booleanValue") == 0):
 *                 return objIterStart.value().get_bool()
 */
        /*else*/ {
          __Pyx_XDECREF(__pyx_r);
          __pyx_t_6 = PyList_New(0); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 356, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_6);
          __pyx_r = __pyx_t_6;
          __pyx_t_6 = 0;
          goto __pyx_L0;
        }

        /* "viur/datastore/transport.pyx":351
 *             elif (strView.compare("integerValue") == 0):
 *                 return int(toPyStr(objIterStart.value().get_string()))
 *             elif (strView.compare("arrayValue") == 0):             # <<<<<<<<<<<<<<
 *                 tmpResult = objIterStart.value().at_pointer("/values")
 *                 if tmpResult.error() == SUCCESS:
 */
      }

      /* "viur/datastore/transport.pyx":357
 *                 else:
 *                     return []
 *             elif (strView.compare("booleanValue") == 0):             # <<<<<<<<<<<<<<
 *                 return objIterStart.value().get_bool()
 *             elif (strView.compare("timestampValue") == 0):
 */
      __pyx_t_3 = (__pyx_v_strView.compare(((char *)"booleanValue")) == 0);
      if (__pyx_t_3) {

        /* "viur/datastore/transport.pyx":358
 *                     return []
 *             elif (strView.compare("booleanValue") == 0):
 *                 return objIterStart.value().get_bool()             # <<<<<<<<<<<<<<
 *             elif (strView.compare("timestampValue") == 0):
 *                 dateStr = toPyStr(objIterStart.value().get_string())[:-1]  # Strip "Z" at the end
 */
        __Pyx_XDECREF(__pyx_r);
        try {
          __pyx_t_10 = __pyx_v_objIterStart.value().get_bool();
        } catch(...) {
          __Pyx_CppExn2PyErr();
          __PYX_ERR(0, 358, __pyx_L1_error)
        }
        __pyx_t_6 = __Pyx_PyBool_FromLong(__pyx_t_10); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 358, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __pyx_r = __pyx_t_6;
        __pyx_t_6 = 0;
        goto __pyx_L0;

        /* "viur/datastore/transport.pyx":357
 *                 else:
 *                     return []
 *             elif (strView.compare("booleanValue") == 0):             # <<<<<<<<<<<<<<
 *                 return objIterStart.value().get_bool()
 *             elif (strView.compare("timestampValue") == 0):
 */
      }

      /* "viur/datastore/transport.pyx":359
 *             elif (strView.compare("booleanValue") == 0):
 *                 return objIterStart.value().get_bool()
 *             elif (strView.compare("timestampValue") == 0):             # <<<<<<<<<<<<<<
 *                 dateStr = toPyStr(objIterStart.value().get_string())[:-1]  # Strip "Z" at the end
 *                 if "." in dateStr:  # With milli-seconds
 */
      __pyx_t_3 = (__pyx_v_strView.compare(((char *)"timestampValue")) == 0);
      if (__pyx_t_3) {

        /* "viur/datastore/transport.pyx":360
 *                 return objIterStart.value().get_bool()
 *             elif (strView.compare("timestampValue") == 0):
 *                 dateStr = toPyStr(objIterStart.value().get_string())[:-1]  # Strip "Z" at the end             # <<<<<<<<<<<<<<
 *                 if "." in dateStr:  # With milli-seconds
 *                     return datetime.strptime(dateStr, "%Y-%m-%dT%H:%M:%S.%f").replace(tzinfo=timezone.utc)
 */
        try {
          __pyx_t_12 = __pyx_v_objIterStart.value().get_string();
        } catch(...) {
          __Pyx_CppExn2PyErr();
          __PYX_ERR(0, 360, __pyx_L1_error)
        }
        __pyx_t_6 = __pyx_f_4viur_9datastore_9transport_toPyStr(__PYX_STD_MOVE_IF_SUPPORTED(__pyx_t_12)); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 360, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __pyx_t_5 = __Pyx_PyObject_GetSlice(__pyx_t_6, 0, -1L, NULL, NULL, &__pyx_slice__2, 0, 1, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 360, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __pyx_v_dateStr = __pyx_t_5;
        __pyx_t_5 = 0;

        /* "viur/datastore/transport.pyx":361
 *             elif (strView.compare("timestampValue") == 0):
 *                 dateStr = toPyStr(objIterStart.value().get_string())[:-1]  # Strip "Z" at the end
 *                 if "." in dateStr:  # With milli-seconds             # <<<<<<<<<<<<<<
 *                     return datetime.strptime(dateStr, "%Y-%m-%dT%H:%M:%S.%f").replace(tzinfo=timezone.utc)
 *                 else:
 */
        __pyx_t_3 = (__Pyx_PySequence_ContainsTF(__pyx_kp_u__3, __pyx_v_dateStr, Py_EQ)); if (unlikely((__pyx_t_3 < 0))) __PYX_ERR(0, 361, __pyx_L1_error)
        if (__pyx_t_3) {

          /* "viur/datastore/transport.pyx":362
 *                 dateStr = toPyStr(objIterStart.value().get_string())[:-1]  # Strip "Z" at the end
 *                 if "." in dateStr:  # With milli-seconds
 *                     return datetime.strptime(dateStr, "%Y-%m-%dT%H:%M:%S.%f").replace(tzinfo=timezone.utc)             # <<<<<<<<<<<<<<
 *                 else:
 *                     return datetime.strptime(dateStr, "%Y-%m-%dT%H:%M:%S").replace(tzinfo=timezone.utc)
 */
          __Pyx_XDECREF(__pyx_r);
          __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_datetime); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 362, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_6);
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_strptime); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 362, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
          __pyx_t_6 = NULL;
          __pyx_t_8 = 0;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_7))) {
            __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_7);
            if (likely(__pyx_t_6)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
              __Pyx_INCREF(__pyx_t_6);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_7, function);
              __pyx_t_8 = 1;
            }
          }
          #endif
          {
            PyObject *__pyx_callargs[3] = {__pyx_t_6, __pyx_v_dateStr, __pyx_kp_u_Y_m_dT_H_M_S_f};
            __pyx_t_5 = __Pyx_PyObject_FastCall(__pyx_t_7, __pyx_callargs+1-__pyx_t_8, 2+__pyx_t_8);
            __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
            if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 362, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_5);
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          }
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_replace); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 362, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __pyx_t_5 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 362, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_timezone); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 362, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_6);
          __pyx_t_14 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_utc); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 362, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_14);
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
          if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_tzinfo, __pyx_t_14) < 0) __PYX_ERR(0, 362, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
          __pyx_t_14 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_empty_tuple, __pyx_t_5); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 362, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_14);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __pyx_r = __pyx_t_14;
          __pyx_t_14 = 0;
          goto __pyx_L0;

          /* "viur/datastore/transport.pyx":361
 *             elif (strView.compare("timestampValue") == 0):
 *                 dateStr = toPyStr(objIterStart.value().get_string())[:-1]  # Strip "Z" at the end
 *                 if "." in dateStr:  # With milli-seconds             # <<<<<<<<<<<<<<
 *                     return datetime.strptime(dateStr, "%Y-%m-%dT%H:%M:%S.%f").replace(tzinfo=timezone.utc)
 *                 else:
 */
        }

        /* "viur/datastore/transport.pyx":364
 *                     return datetime.strptime(dateStr, "%Y-%m-%dT%H:%M:%S.%f").replace(tzinfo=timezone.utc)
 *                 else:
 *                     return datetime.strptime(dateStr, "%Y-%m-%dT%H:%M:%S").replace(tzinfo=timezone.utc)             # <<<<<<<<<<<<<<
 *             elif (strView.compare("blobValue") == 0):
 *                 return b64decode(toPyStr(objIterStart.value().get_string()))
 */
        /*else*/ {
          __Pyx_XDECREF(__pyx_r);
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_datetime); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 364, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_strptime); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 364, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __pyx_t_5 = NULL;
          __pyx_t_8 = 0;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_7))) {
            __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_7);
            if (likely(__pyx_t_5)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
              __Pyx_INCREF(__pyx_t_5);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_7, function);
              __pyx_t_8 = 1;
            }
          }
          #endif
          {
            PyObject *__pyx_callargs[3] = {__pyx_t_5, __pyx_v_dateStr, __pyx_kp_u_Y_m_dT_H_M_S};
            __pyx_t_14 = __Pyx_PyObject_FastCall(__pyx_t_7, __pyx_callargs+1-__pyx_t_8, 2+__pyx_t_8);
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 364, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_14);
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          }
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_14, __pyx_n_s_replace); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 364, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
          __pyx_t_14 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 364, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_14);
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_timezone); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 364, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_utc); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 364, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_6);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          if (PyDict_SetItem(__pyx_t_14, __pyx_n_s_tzinfo, __pyx_t_6) < 0) __PYX_ERR(0, 364, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
          __pyx_t_6 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_empty_tuple, __pyx_t_14); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 364, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_6);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
          __pyx_r = __pyx_t_6;
          __pyx_t_6 = 0;
          goto __pyx_L0;
        }

        /* "viur/datastore/transport.pyx":359
 *             elif (strView.compare("booleanValue") == 0):
 *                 return objIterStart.value().get_bool()
 *             elif (strView.compare("timestampValue") == 0):             # <<<<<<<<<<<<<<
 *                 dateStr = toPyStr(objIterStart.value().get_string())[:-1]  # Strip "Z" at the end
 *                 if "." in dateStr:  # With milli-seconds
 */
      }

      /* "viur/datastore/transport.pyx":365
 *                 else:
 *                     return datetime.strptime(dateStr, "%Y-%m-%dT%H:%M:%S").replace(tzinfo=timezone.utc)
 *             elif (strView.compare("blobValue") == 0):             # <<<<<<<<<<<<<<
 *                 return b64decode(toPyStr(objIterStart.value().get_string()))
 *             elif (strView.compare("keyValue") == 0):
 */
      __pyx_t_3 = (__pyx_v_strView.compare(((char *)"blobValue")) == 0);
      if (__pyx_t_3) {

        /* "viur/datastore/transport.pyx":366
 *                     return datetime.strptime(dateStr, "%Y-%m-%dT%H:%M:%S").replace(tzinfo=timezone.utc)
 *             elif (strView.compare("blobValue") == 0):
 *                 return b64decode(toPyStr(objIterStart.value().get_string()))             # <<<<<<<<<<<<<<
 *             elif (strView.compare("keyValue") == 0):
 *                 return parseKey(objIterStart.value())
 */
        __Pyx_XDECREF(__pyx_r);
        __Pyx_GetModuleGlobalName(__pyx_t_14, __pyx_n_s_b64decode); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 366, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_14);
        try {
          __pyx_t_12 = __pyx_v_objIterStart.value().get_string();
        } catch(...) {
          __Pyx_CppExn2PyErr();
          __PYX_ERR(0, 366, __pyx_L1_error)
        }
        __pyx_t_7 = __pyx_f_4viur_9datastore_9transport_toPyStr(__PYX_STD_MOVE_IF_SUPPORTED(__pyx_t_12)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 366, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_5 = NULL;
        __pyx_t_8 = 0;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_14))) {
          __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_14);
          if (likely(__pyx_t_5)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_14);
            __Pyx_INCREF(__pyx_t_5);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_14, function);
            __pyx_t_8 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_5, __pyx_t_7};
          __pyx_t_6 = __Pyx_PyObject_FastCall(__pyx_t_14, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 366, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_6);
          __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
        }
        __pyx_r = __pyx_t_6;
        __pyx_t_6 = 0;
        goto __pyx_L0;

        /* "viur/datastore/transport.pyx":365
 *                 else:
 *                     return datetime.strptime(dateStr, "%Y-%m-%dT%H:%M:%S").replace(tzinfo=timezone.utc)
 *             elif (strView.compare("blobValue") == 0):             # <<<<<<<<<<<<<<
 *                 return b64decode(toPyStr(objIterStart.value().get_string()))
 *             elif (strView.compare("keyValue") == 0):
 */
      }

      /* "viur/datastore/transport.pyx":367
 *             elif (strView.compare("blobValue") == 0):
 *                 return b64decode(toPyStr(objIterStart.value().get_string()))
 *             elif (strView.compare("keyValue") == 0):             # <<<<<<<<<<<<<<
 *                 return parseKey(objIterStart.value())
 *             elif (strView.compare("geoPointValue") == 0):
 */
      __pyx_t_3 = (__pyx_v_strView.compare(((char *)"keyValue")) == 0);
      if (__pyx_t_3) {

        /* "viur/datastore/transport.pyx":368
 *                 return b64decode(toPyStr(objIterStart.value().get_string()))
 *             elif (strView.compare("keyValue") == 0):
 *                 return parseKey(objIterStart.value())             # <<<<<<<<<<<<<<
 *             elif (strView.compare("geoPointValue") == 0):
 *                 return objIterStart.value().at_key("latitude").get_double(), \
 */
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_6 = __pyx_f_4viur_9datastore_9transport_parseKey(__pyx_v_objIterStart.value()); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 368, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __pyx_r = __pyx_t_6;
        __pyx_t_6 = 0;
        goto __pyx_L0;

        /* "viur/datastore/transport.pyx":367
 *             elif (strView.compare("blobValue") == 0):
 *                 return b64decode(toPyStr(objIterStart.value().get_string()))
 *             elif (strView.compare("keyValue") == 0):             # <<<<<<<<<<<<<<
 *                 return parseKey(objIterStart.value())
 *             elif (strView.compare("geoPointValue") == 0):
 */
      }

      /* "viur/datastore/transport.pyx":369
 *             elif (strView.compare("keyValue") == 0):
 *                 return parseKey(objIterStart.value())
 *             elif (strView.compare("geoPointValue") == 0):             # <<<<<<<<<<<<<<
 *                 return objIterStart.value().at_key("latitude").get_double(), \
 *                     objIterStart.value().at_key("longitude").get_double()
 */
      __pyx_t_3 = (__pyx_v_strView.compare(((char *)"geoPointValue")) == 0);
      if (__pyx_t_3) {

        /* "viur/datastore/transport.pyx":370
 *                 return parseKey(objIterStart.value())
 *             elif (strView.compare("geoPointValue") == 0):
 *                 return objIterStart.value().at_key("latitude").get_double(), \             # <<<<<<<<<<<<<<
 *                     objIterStart.value().at_key("longitude").get_double()
 *             elif (strView.compare("excludeFromIndexes") == 0 or strView.compare("version") == 0):
 */
        __Pyx_XDECREF(__pyx_r);
        try {
          __pyx_t_15 = __pyx_v_objIterStart.value().at_key(((char const *)"latitude"));
        } catch(...) {
          __Pyx_CppExn2PyErr();
          __PYX_ERR(0, 370, __pyx_L1_error)
        }
        try {
          __pyx_t_13 = __pyx_t_15.get_double();
        } catch(...) {
          __Pyx_CppExn2PyErr();
          __PYX_ERR(0, 370, __pyx_L1_error)
        }
        __pyx_t_6 = PyFloat_FromDouble(__pyx_t_13); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 370, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);

        /* "viur/datastore/transport.pyx":371
 *             elif (strView.compare("geoPointValue") == 0):
 *                 return objIterStart.value().at_key("latitude").get_double(), \
 *                     objIterStart.value().at_key("longitude").get_double()             # <<<<<<<<<<<<<<
 *             elif (strView.compare("excludeFromIndexes") == 0 or strView.compare("version") == 0):
 *                 pass
 */
        try {
          __pyx_t_15 = __pyx_v_objIterStart.value().at_key(((char const *)"longitude"));
        } catch(...) {
          __Pyx_CppExn2PyErr();
          __PYX_ERR(0, 371, __pyx_L1_error)
        }
        try {
          __pyx_t_13 = __pyx_t_15.get_double();
        } catch(...) {
          __Pyx_CppExn2PyErr();
          __PYX_ERR(0, 371, __pyx_L1_error)
        }
        __pyx_t_14 = PyFloat_FromDouble(__pyx_t_13); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 371, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_14);

        /* "viur/datastore/transport.pyx":370
 *                 return parseKey(objIterStart.value())
 *             elif (strView.compare("geoPointValue") == 0):
 *                 return objIterStart.value().at_key("latitude").get_double(), \             # <<<<<<<<<<<<<<
 *                     objIterStart.value().at_key("longitude").get_double()
 *             elif (strView.compare("excludeFromIndexes") == 0 or strView.compare("version") == 0):
 */
        __pyx_t_7 = PyTuple_New(2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 370, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_GIVEREF(__pyx_t_6);
        if (__Pyx_PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_6)) __PYX_ERR(0, 370, __pyx_L1_error);
        __Pyx_GIVEREF(__pyx_t_14);
        if (__Pyx_PyTuple_SET_ITEM(__pyx_t_7, 1, __pyx_t_14)) __PYX_ERR(0, 370, __pyx_L1_error);
        __pyx_t_6 = 0;
        __pyx_t_14 = 0;
        __pyx_r = __pyx_t_7;
        __pyx_t_7 = 0;
        goto __pyx_L0;

        /* "viur/datastore/transport.pyx":369
 *             elif (strView.compare("keyValue") == 0):
 *                 return parseKey(objIterStart.value())
 *             elif (strView.compare("geoPointValue") == 0):             # <<<<<<<<<<<<<<
 *                 return objIterStart.value().at_key("latitude").get_double(), \
 *                     objIterStart.value().at_key("longitude").get_double()
 */
      }

      /* "viur/datastore/transport.pyx":372
 *                 return objIterStart.value().at_key("latitude").get_double(), \
 *                     objIterStart.value().at_key("longitude").get_double()
 *             elif (strView.compare("excludeFromIndexes") == 0 or strView.compare("version") == 0):             # <<<<<<<<<<<<<<
 *                 pass
 *             else:
 */
      __pyx_t_4 = (__pyx_v_strView.compare(((char *)"excludeFromIndexes")) == 0);
      if (!__pyx_t_4) {
      } else {
        __pyx_t_3 = __pyx_t_4;
        goto __pyx_L24_bool_binop_done;
      }
      __pyx_t_4 = (__pyx_v_strView.compare(((char *)"version")) == 0);
      __pyx_t_3 = __pyx_t_4;
      __pyx_L24_bool_binop_done:;
      if (likely(__pyx_t_3)) {
        goto __pyx_L5;
      }

      /* "viur/datastore/transport.pyx":375
 *                 pass
 *             else:
 *                 raise ValueError("Invalid key in entity json: %s" % toPyStr(strView))             # <<<<<<<<<<<<<<
 *     elif et == ARRAY:
 *         arr = v.get_array()
 */
      /*else*/ {
        __pyx_t_7 = __pyx_f_4viur_9datastore_9transport_toPyStr(__pyx_v_strView); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 375, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_14 = __Pyx_PyUnicode_FormatSafe(__pyx_kp_u_Invalid_key_in_entity_json_s, __pyx_t_7); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 375, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_14);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_7 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_14); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 375, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
        __Pyx_Raise(__pyx_t_7, 0, 0, 0);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __PYX_ERR(0, 375, __pyx_L1_error)
      }
      __pyx_L5:;
    }

    /* "viur/datastore/transport.pyx":301
 *     cdef stringView strView, bytesView
 *     cdef simdjsonResult tmpResult
 *     if et == OBJECT:             # <<<<<<<<<<<<<<
 *         outerObject = v.get_object()
 *         objIterStart = outerObject.begin()
 */
    break;
    case simdjson::dom::element_type::ARRAY:

    /* "viur/datastore/transport.pyx":377
 *                 raise ValueError("Invalid key in entity json: %s" % toPyStr(strView))
 *     elif et == ARRAY:
 *         arr = v.get_array()             # <<<<<<<<<<<<<<
 *         arrayIt = arr.begin()
 *         arrayItEnd = arr.end()
 */
    try {
      __pyx_t_9 = __pyx_v_v.get_array();
    } catch(...) {
      __Pyx_CppExn2PyErr();
      __PYX_ERR(0, 377, __pyx_L1_error)
    }
    __pyx_v_arr = __PYX_STD_MOVE_IF_SUPPORTED(__pyx_t_9);

    /* "viur/datastore/transport.pyx":378
 *     elif et == ARRAY:
 *         arr = v.get_array()
 *         arrayIt = arr.begin()             # <<<<<<<<<<<<<<
 *         arrayItEnd = arr.end()
 *         if isInitial:
 */
    __pyx_v_arrayIt = __pyx_v_arr.begin();

    /* "viur/datastore/transport.pyx":379
 *         arr = v.get_array()
 *         arrayIt = arr.begin()
 *         arrayItEnd = arr.end()             # <<<<<<<<<<<<<<
 *         if isInitial:
 *             res = {}
 */
    __pyx_v_arrayItEnd = __pyx_v_arr.end();

    /* "viur/datastore/transport.pyx":380
 *         arrayIt = arr.begin()
 *         arrayItEnd = arr.end()
 *         if isInitial:             # <<<<<<<<<<<<<<
 *             res = {}
 *         else:
 */
    __pyx_t_3 = (__pyx_v_isInitial != 0);
    if (__pyx_t_3) {

      /* "viur/datastore/transport.pyx":381
 *         arrayItEnd = arr.end()
 *         if isInitial:
 *             res = {}             # <<<<<<<<<<<<<<
 *         else:
 *             res = []
 */
      __pyx_t_7 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 381, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_v_res = __pyx_t_7;
      __pyx_t_7 = 0;

      /* "viur/datastore/transport.pyx":380
 *         arrayIt = arr.begin()
 *         arrayItEnd = arr.end()
 *         if isInitial:             # <<<<<<<<<<<<<<
 *             res = {}
 *         else:
 */
      goto __pyx_L26;
    }

    /* "viur/datastore/transport.pyx":383
 *             res = {}
 *         else:
 *             res = []             # <<<<<<<<<<<<<<
 *         while arrayIt != arrayItEnd:
 *             element = dereference(arrayIt)
 */
    /*else*/ {
      __pyx_t_7 = PyList_New(0); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 383, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_v_res = __pyx_t_7;
      __pyx_t_7 = 0;
    }
    __pyx_L26:;

    /* "viur/datastore/transport.pyx":384
 *         else:
 *             res = []
 *         while arrayIt != arrayItEnd:             # <<<<<<<<<<<<<<
 *             element = dereference(arrayIt)
 *             entity = toEntityStructure(element)
 */
    while (1) {
      __pyx_t_3 = (__pyx_v_arrayIt != __pyx_v_arrayItEnd);
      if (!__pyx_t_3) break;

      /* "viur/datastore/transport.pyx":385
 *             res = []
 *         while arrayIt != arrayItEnd:
 *             element = dereference(arrayIt)             # <<<<<<<<<<<<<<
 *             entity = toEntityStructure(element)
 *             if isInitial:
 */
      __pyx_v_element = (*__pyx_v_arrayIt);

      /* "viur/datastore/transport.pyx":386
 *         while arrayIt != arrayItEnd:
 *             element = dereference(arrayIt)
 *             entity = toEntityStructure(element)             # <<<<<<<<<<<<<<
 *             if isInitial:
 *                 res[entity.key] = entity
 */
      __pyx_t_7 = __pyx_f_4viur_9datastore_9transport_toEntityStructure(__pyx_v_element, NULL); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 386, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_XDECREF_SET(__pyx_v_entity, __pyx_t_7);
      __pyx_t_7 = 0;

      /* "viur/datastore/transport.pyx":387
 *             element = dereference(arrayIt)
 *             entity = toEntityStructure(element)
 *             if isInitial:             # <<<<<<<<<<<<<<
 *                 res[entity.key] = entity
 *             else:
 */
      __pyx_t_3 = (__pyx_v_isInitial != 0);
      if (__pyx_t_3) {

        /* "viur/datastore/transport.pyx":388
 *             entity = toEntityStructure(element)
 *             if isInitial:
 *                 res[entity.key] = entity             # <<<<<<<<<<<<<<
 *             else:
 *                 res.append(entity)
 */
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_entity, __pyx_n_s_key); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 388, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        if (unlikely((PyObject_SetItem(__pyx_v_res, __pyx_t_7, __pyx_v_entity) < 0))) __PYX_ERR(0, 388, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

        /* "viur/datastore/transport.pyx":387
 *             element = dereference(arrayIt)
 *             entity = toEntityStructure(element)
 *             if isInitial:             # <<<<<<<<<<<<<<
 *                 res[entity.key] = entity
 *             else:
 */
        goto __pyx_L29;
      }

      /* "viur/datastore/transport.pyx":390
 *                 res[entity.key] = entity
 *             else:
 *                 res.append(entity)             # <<<<<<<<<<<<<<
 *             preincrement(arrayIt)
 *         return res
 */
      /*else*/ {
        __pyx_t_11 = __Pyx_PyObject_Append(__pyx_v_res, __pyx_v_entity); if (unlikely(__pyx_t_11 == ((int)-1))) __PYX_ERR(0, 390, __pyx_L1_error)
      }
      __pyx_L29:;

      /* "viur/datastore/transport.pyx":391
 *             else:
 *                 res.append(entity)
 *             preincrement(arrayIt)             # <<<<<<<<<<<<<<
 *         return res
 * 
 */
      (void)((++__pyx_v_arrayIt));
    }

    /* "viur/datastore/transport.pyx":392
 *                 res.append(entity)
 *             preincrement(arrayIt)
 *         return res             # <<<<<<<<<<<<<<
 * 
 * cdef inline object toPythonStructure(simdjsonElement v):
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_v_res);
    __pyx_r = __pyx_v_res;
    goto __pyx_L0;

    /* "viur/datastore/transport.pyx":376
 *             else:
 *                 raise ValueError("Invalid key in entity json: %s" % toPyStr(strView))
 *     elif et == ARRAY:             # <<<<<<<<<<<<<<
 *         arr = v.get_array()
 *         arrayIt = arr.begin()
 */
    break;
    default: break;
  }

  /* "viur/datastore/transport.pyx":284
 *     return key
 * 
 * cdef inline object toEntityStructure(simdjsonElement v, boolean_type isInitial = False):             # <<<<<<<<<<<<<<
 *     """
 *         Parses a simdJsonElement into the corresponding python datatypes.
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_14);
  __Pyx_AddTraceback("viur.datastore.transport.toEntityStructure", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_e);
  __Pyx_XDECREF(__pyx_v_excludeList);
  __Pyx_XDECREF(__pyx_v_dateStr);
  __Pyx_XDECREF(__pyx_v_res);
  __Pyx_XDECREF(__pyx_v_entity);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "viur/datastore/transport.pyx":394
 *         return res
 * 
 * cdef inline object toPythonStructure(simdjsonElement v):             # <<<<<<<<<<<<<<
 *     # Convert json (sub-)tree to python objects
 *     cdef simdjsonElementType et = v.type()
 */

static CYTHON_INLINE PyObject *__pyx_f_4viur_9datastore_9transport_toPythonStructure(simdjson::dom::element __pyx_v_v) {
  enum simdjson::dom::element_type __pyx_v_et;
  simdjson::dom::array::iterator __pyx_v_arrayIt;
  simdjson::dom::array::iterator __pyx_v_arrayItEnd;
  simdjson::dom::object::iterator __pyx_v_objIterStart;
  simdjson::dom::object::iterator __pyx_v_objIterEnd;
  simdjson::dom::element __pyx_v_element;
  simdjson::dom::object __pyx_v_Object;
  std::string_view __pyx_v_strView;
  PyObject *__pyx_v_res = NULL;
  simdjson::dom::array __pyx_v_arr;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  enum simdjson::dom::element_type __pyx_t_1;
  simdjson::dom::object __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  simdjson::dom::array __pyx_t_6;
  int __pyx_t_7;
  std::string_view __pyx_t_8;
  int64_t __pyx_t_9;
  uint64_t __pyx_t_10;
  double __pyx_t_11;
  bool __pyx_t_12;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("toPythonStructure", 1);

  /* "viur/datastore/transport.pyx":396
 * cdef inline object toPythonStructure(simdjsonElement v):
 *     # Convert json (sub-)tree to python objects
 *     cdef simdjsonElementType et = v.type()             # <<<<<<<<<<<<<<
 *     cdef simdjsonArray.iterator arrayIt, arrayItEnd
 *     cdef simdjsonObject.iterator objIterStart, objIterEnd
 */
  try {
    __pyx_t_1 = __pyx_v_v.type();
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 396, __pyx_L1_error)
  }
  __pyx_v_et = __pyx_t_1;

  /* "viur/datastore/transport.pyx":402
 *     cdef simdjsonObject Object
 *     cdef stringView strView
 *     if et == OBJECT:             # <<<<<<<<<<<<<<
 *         Object = v.get_object()
 *         objIterStart = Object.begin()
 */
  switch (__pyx_v_et) {
    case simdjson::dom::element_type::OBJECT:

    /* "viur/datastore/transport.pyx":403
 *     cdef stringView strView
 *     if et == OBJECT:
 *         Object = v.get_object()             # <<<<<<<<<<<<<<
 *         objIterStart = Object.begin()
 *         objIterEnd = Object.end()
 */
    try {
      __pyx_t_2 = __pyx_v_v.get_object();
    } catch(...) {
      __Pyx_CppExn2PyErr();
      __PYX_ERR(0, 403, __pyx_L1_error)
    }
    __pyx_v_Object = __PYX_STD_MOVE_IF_SUPPORTED(__pyx_t_2);

    /* "viur/datastore/transport.pyx":404
 *     if et == OBJECT:
 *         Object = v.get_object()
 *         objIterStart = Object.begin()             # <<<<<<<<<<<<<<
 *         objIterEnd = Object.end()
 *         res = {}
 */
    __pyx_v_objIterStart = __pyx_v_Object.begin();

    /* "viur/datastore/transport.pyx":405
 *         Object = v.get_object()
 *         objIterStart = Object.begin()
 *         objIterEnd = Object.end()             # <<<<<<<<<<<<<<
 *         res = {}
 *         while objIterStart != objIterEnd:
 */
    __pyx_v_objIterEnd = __pyx_v_Object.end();

    /* "viur/datastore/transport.pyx":406
 *         objIterStart = Object.begin()
 *         objIterEnd = Object.end()
 *         res = {}             # <<<<<<<<<<<<<<
 *         while objIterStart != objIterEnd:
 *             strView = objIterStart.key()
 */
    __pyx_t_3 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 406, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_v_res = __pyx_t_3;
    __pyx_t_3 = 0;

    /* "viur/datastore/transport.pyx":407
 *         objIterEnd = Object.end()
 *         res = {}
 *         while objIterStart != objIterEnd:             # <<<<<<<<<<<<<<
 *             strView = objIterStart.key()
 *             res[toPyStr(strView)] = toPythonStructure(objIterStart.value())
 */
    while (1) {
      __pyx_t_4 = (__pyx_v_objIterStart != __pyx_v_objIterEnd);
      if (!__pyx_t_4) break;

      /* "viur/datastore/transport.pyx":408
 *         res = {}
 *         while objIterStart != objIterEnd:
 *             strView = objIterStart.key()             # <<<<<<<<<<<<<<
 *             res[toPyStr(strView)] = toPythonStructure(objIterStart.value())
 *             preincrement(objIterStart)
 */
      __pyx_v_strView = __pyx_v_objIterStart.key();

      /* "viur/datastore/transport.pyx":409
 *         while objIterStart != objIterEnd:
 *             strView = objIterStart.key()
 *             res[toPyStr(strView)] = toPythonStructure(objIterStart.value())             # <<<<<<<<<<<<<<
 *             preincrement(objIterStart)
 *         return res
 */
      __pyx_t_3 = __pyx_f_4viur_9datastore_9transport_toPythonStructure(__pyx_v_objIterStart.value()); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 409, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_5 = __pyx_f_4viur_9datastore_9transport_toPyStr(__pyx_v_strView); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 409, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      if (unlikely((PyObject_SetItem(__pyx_v_res, __pyx_t_5, __pyx_t_3) < 0))) __PYX_ERR(0, 409, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

      /* "viur/datastore/transport.pyx":410
 *             strView = objIterStart.key()
 *             res[toPyStr(strView)] = toPythonStructure(objIterStart.value())
 *             preincrement(objIterStart)             # <<<<<<<<<<<<<<
 *         return res
 *     elif et == ARRAY:
 */
      (void)((++__pyx_v_objIterStart));
    }

    /* "viur/datastore/transport.pyx":411
 *             res[toPyStr(strView)] = toPythonStructure(objIterStart.value())
 *             preincrement(objIterStart)
 *         return res             # <<<<<<<<<<<<<<
 *     elif et == ARRAY:
 *         arr = v.get_array()
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_v_res);
    __pyx_r = __pyx_v_res;
    goto __pyx_L0;

    /* "viur/datastore/transport.pyx":402
 *     cdef simdjsonObject Object
 *     cdef stringView strView
 *     if et == OBJECT:             # <<<<<<<<<<<<<<
 *         Object = v.get_object()
 *         objIterStart = Object.begin()
 */
    break;
    case simdjson::dom::element_type::ARRAY:

    /* "viur/datastore/transport.pyx":413
 *         return res
 *     elif et == ARRAY:
 *         arr = v.get_array()             # <<<<<<<<<<<<<<
 *         arrayIt = arr.begin()
 *         arrayItEnd = arr.end()
 */
    try {
      __pyx_t_6 = __pyx_v_v.get_array();
    } catch(...) {
      __Pyx_CppExn2PyErr();
      __PYX_ERR(0, 413, __pyx_L1_error)
    }
    __pyx_v_arr = __PYX_STD_MOVE_IF_SUPPORTED(__pyx_t_6);

    /* "viur/datastore/transport.pyx":414
 *     elif et == ARRAY:
 *         arr = v.get_array()
 *         arrayIt = arr.begin()             # <<<<<<<<<<<<<<
 *         arrayItEnd = arr.end()
 *         res = []
 */
    __pyx_v_arrayIt = __pyx_v_arr.begin();

    /* "viur/datastore/transport.pyx":415
 *         arr = v.get_array()
 *         arrayIt = arr.begin()
 *         arrayItEnd = arr.end()             # <<<<<<<<<<<<<<
 *         res = []
 *         while arrayIt != arrayItEnd:
 */
    __pyx_v_arrayItEnd = __pyx_v_arr.end();

    /* "viur/datastore/transport.pyx":416
 *         arrayIt = arr.begin()
 *         arrayItEnd = arr.end()
 *         res = []             # <<<<<<<<<<<<<<
 *         while arrayIt != arrayItEnd:
 *             element = dereference(arrayIt)
 */
    __pyx_t_3 = PyList_New(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 416, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_v_res = __pyx_t_3;
    __pyx_t_3 = 0;

    /* "viur/datastore/transport.pyx":417
 *         arrayItEnd = arr.end()
 *         res = []
 *         while arrayIt != arrayItEnd:             # <<<<<<<<<<<<<<
 *             element = dereference(arrayIt)
 *             res.append(toPythonStructure(element))
 */
    while (1) {
      __pyx_t_4 = (__pyx_v_arrayIt != __pyx_v_arrayItEnd);
      if (!__pyx_t_4) break;

      /* "viur/datastore/transport.pyx":418
 *         res = []
 *         while arrayIt != arrayItEnd:
 *             element = dereference(arrayIt)             # <<<<<<<<<<<<<<
 *             res.append(toPythonStructure(element))
 *             preincrement(arrayIt)
 */
      __pyx_v_element = (*__pyx_v_arrayIt);

      /* "viur/datastore/transport.pyx":419
 *         while arrayIt != arrayItEnd:
 *             element = dereference(arrayIt)
 *             res.append(toPythonStructure(element))             # <<<<<<<<<<<<<<
 *             preincrement(arrayIt)
 *         return res
 */
      __pyx_t_3 = __pyx_f_4viur_9datastore_9transport_toPythonStructure(__pyx_v_element); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 419, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_7 = __Pyx_PyObject_Append(__pyx_v_res, __pyx_t_3); if (unlikely(__pyx_t_7 == ((int)-1))) __PYX_ERR(0, 419, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

      /* "viur/datastore/transport.pyx":420
 *             element = dereference(arrayIt)
 *             res.append(toPythonStructure(element))
 *             preincrement(arrayIt)             # <<<<<<<<<<<<<<
 *         return res
 *     elif et == STRING:
 */
      (void)((++__pyx_v_arrayIt));
    }

    /* "viur/datastore/transport.pyx":421
 *             res.append(toPythonStructure(element))
 *             preincrement(arrayIt)
 *         return res             # <<<<<<<<<<<<<<
 *     elif et == STRING:
 *         return toPyStr(v.get_string())
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_v_res);
    __pyx_r = __pyx_v_res;
    goto __pyx_L0;

    /* "viur/datastore/transport.pyx":412
 *             preincrement(objIterStart)
 *         return res
 *     elif et == ARRAY:             # <<<<<<<<<<<<<<
 *         arr = v.get_array()
 *         arrayIt = arr.begin()
 */
    break;
    case simdjson::dom::element_type::STRING:

    /* "viur/datastore/transport.pyx":423
 *         return res
 *     elif et == STRING:
 *         return toPyStr(v.get_string())             # <<<<<<<<<<<<<<
 *     elif et == INT64:
 *         return v.get_int64()
 */
    __Pyx_XDECREF(__pyx_r);
    try {
      __pyx_t_8 = __pyx_v_v.get_string();
    } catch(...) {
      __Pyx_CppExn2PyErr();
      __PYX_ERR(0, 423, __pyx_L1_error)
    }
    __pyx_t_3 = __pyx_f_4viur_9datastore_9transport_toPyStr(__PYX_STD_MOVE_IF_SUPPORTED(__pyx_t_8)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 423, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "viur/datastore/transport.pyx":422
 *             preincrement(arrayIt)
 *         return res
 *     elif et == STRING:             # <<<<<<<<<<<<<<
 *         return toPyStr(v.get_string())
 *     elif et == INT64:
 */
    break;
    case simdjson::dom::element_type::INT64:

    /* "viur/datastore/transport.pyx":425
 *         return toPyStr(v.get_string())
 *     elif et == INT64:
 *         return v.get_int64()             # <<<<<<<<<<<<<<
 *     elif et == UINT64:
 *         return v.get_uint64()
 */
    __Pyx_XDECREF(__pyx_r);
    try {
      __pyx_t_9 = __pyx_v_v.get_int64();
    } catch(...) {
      __Pyx_CppExn2PyErr();
      __PYX_ERR(0, 425, __pyx_L1_error)
    }
    __pyx_t_3 = __Pyx_PyInt_From_int64_t(__pyx_t_9); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 425, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "viur/datastore/transport.pyx":424
 *     elif et == STRING:
 *         return toPyStr(v.get_string())
 *     elif et == INT64:             # <<<<<<<<<<<<<<
 *         return v.get_int64()
 *     elif et == UINT64:
 */
    break;
    case simdjson::dom::element_type::UINT64:

    /* "viur/datastore/transport.pyx":427
 *         return v.get_int64()
 *     elif et == UINT64:
 *         return v.get_uint64()             # <<<<<<<<<<<<<<
 *     elif et == DOUBLE:
 *         return v.get_double()
 */
    __Pyx_XDECREF(__pyx_r);
    try {
      __pyx_t_10 = __pyx_v_v.get_uint64();
    } catch(...) {
      __Pyx_CppExn2PyErr();
      __PYX_ERR(0, 427, __pyx_L1_error)
    }
    __pyx_t_3 = __Pyx_PyInt_From_uint64_t(__pyx_t_10); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 427, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "viur/datastore/transport.pyx":426
 *     elif et == INT64:
 *         return v.get_int64()
 *     elif et == UINT64:             # <<<<<<<<<<<<<<
 *         return v.get_uint64()
 *     elif et == DOUBLE:
 */
    break;
    case simdjson::dom::element_type::DOUBLE:

    /* "viur/datastore/transport.pyx":429
 *         return v.get_uint64()
 *     elif et == DOUBLE:
 *         return v.get_double()             # <<<<<<<<<<<<<<
 *     elif et == NULL_VALUE:
 *         return None
 */
    __Pyx_XDECREF(__pyx_r);
    try {
      __pyx_t_11 = __pyx_v_v.get_double();
    } catch(...) {
      __Pyx_CppExn2PyErr();
      __PYX_ERR(0, 429, __pyx_L1_error)
    }
    __pyx_t_3 = PyFloat_FromDouble(__pyx_t_11); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 429, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "viur/datastore/transport.pyx":428
 *     elif et == UINT64:
 *         return v.get_uint64()
 *     elif et == DOUBLE:             # <<<<<<<<<<<<<<
 *         return v.get_double()
 *     elif et == NULL_VALUE:
 */
    break;
    case simdjson::dom::element_type::NULL_VALUE:

    /* "viur/datastore/transport.pyx":431
 *         return v.get_double()
 *     elif et == NULL_VALUE:
 *         return None             # <<<<<<<<<<<<<<
 *     elif et == BOOL:
 *         return v.get_bool()
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "viur/datastore/transport.pyx":430
 *     elif et == DOUBLE:
 *         return v.get_double()
 *     elif et == NULL_VALUE:             # <<<<<<<<<<<<<<
 *         return None
 *     elif et == BOOL:
 */
    break;
    case simdjson::dom::element_type::BOOL:

    /* "viur/datastore/transport.pyx":433
 *         return None
 *     elif et == BOOL:
 *         return v.get_bool()             # <<<<<<<<<<<<<<
 *     else:
 *         raise ValueError("Unknown type")
 */
    __Pyx_XDECREF(__pyx_r);
    try {
      __pyx_t_12 = __pyx_v_v.get_bool();
    } catch(...) {
      __Pyx_CppExn2PyErr();
      __PYX_ERR(0, 433, __pyx_L1_error)
    }
    __pyx_t_3 = __Pyx_PyBool_FromLong(__pyx_t_12); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 433, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "viur/datastore/transport.pyx":432
 *     elif et == NULL_VALUE:
 *         return None
 *     elif et == BOOL:             # <<<<<<<<<<<<<<
 *         return v.get_bool()
 *     else:
 */
    break;
    default:

    /* "viur/datastore/transport.pyx":435
 *         return v.get_bool()
 *     else:
 *         raise ValueError("Unknown type")             # <<<<<<<<<<<<<<
 * 
 * def runSingleFilter(queryDefinition: QueryDefinition, limit: int) -> List[Entity]:
 */
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__4, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 435, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(0, 435, __pyx_L1_error)
    break;
  }

  /* "viur/datastore/transport.pyx":394
 *         return res
 * 
 * cdef inline object toPythonStructure(simdjsonElement v):             # <<<<<<<<<<<<<<
 *     # Convert json (sub-)tree to python objects
 *     cdef simdjsonElementType et = v.type()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("viur.datastore.transport.toPythonStructure", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_res);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "viur/datastore/transport.pyx":437
 *         raise ValueError("Unknown type")
 * 
 * def runSingleFilter(queryDefinition: QueryDefinition, limit: int) -> List[Entity]:             # <<<<<<<<<<<<<<
 *     """
 *         Runs a single Query as defined by queryDefinition. The limit of the queryDefinition is ignored and must
 */

/* Python wrapper */
static PyObject *__pyx_pw_4viur_9datastore_9transport_7runSingleFilter(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_4viur_9datastore_9transport_6runSingleFilter, "\n        Runs a single Query as defined by queryDefinition. The limit of the queryDefinition is ignored and must\n        be specified separately to prevent _calculateInternalMultiQueryLimit from modifying the queryDefinition.\n\n        :param queryDefinition: The query to run\n        :param limit:  How many entities to return at maximum\n        :return: The list of entities fetched from the datastore\n    ");
static PyMethodDef __pyx_mdef_4viur_9datastore_9transport_7runSingleFilter = {"runSingleFilter", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_4viur_9datastore_9transport_7runSingleFilter, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_4viur_9datastore_9transport_6runSingleFilter};
static PyObject *__pyx_pw_4viur_9datastore_9transport_7runSingleFilter(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_queryDefinition = 0;
  PyObject *__pyx_v_limit = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[2] = {0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("runSingleFilter (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_queryDefinition,&__pyx_n_s_limit,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_queryDefinition)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 437, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_limit)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 437, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("runSingleFilter", 1, 2, 2, 1); __PYX_ERR(0, 437, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "runSingleFilter") < 0)) __PYX_ERR(0, 437, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 2)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
    }
    __pyx_v_queryDefinition = values[0];
    __pyx_v_limit = ((PyObject*)values[1]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("runSingleFilter", 1, 2, 2, __pyx_nargs); __PYX_ERR(0, 437, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("viur.datastore.transport.runSingleFilter", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_limit), (&PyInt_Type), 0, "limit", 1))) __PYX_ERR(0, 437, __pyx_L1_error)
  __pyx_r = __pyx_pf_4viur_9datastore_9transport_6runSingleFilter(__pyx_self, __pyx_v_queryDefinition, __pyx_v_limit);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4viur_9datastore_9transport_6runSingleFilter(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_queryDefinition, PyObject *__pyx_v_limit) {
  simdjson::dom::parser __pyx_v_parser;
  Py_ssize_t __pyx_v_pysize;
  char *__pyx_v_data_ptr;
  simdjson::dom::element __pyx_v_element;
  PyObject *__pyx_v_res = NULL;
  PyObject *__pyx_v_internalStartCursor = NULL;
  PyObject *__pyx_v_flipResults = NULL;
  PyObject *__pyx_v_currentTxn = NULL;
  PyObject *__pyx_v_readOptions = NULL;
  PyObject *__pyx_v_postData = NULL;
  PyObject *__pyx_v_filterList = NULL;
  PyObject *__pyx_v_k = NULL;
  PyObject *__pyx_v_v = NULL;
  PyObject *__pyx_v_key = NULL;
  PyObject *__pyx_v_op = NULL;
  PyObject *__pyx_v_singleValue = NULL;
  PyObject *__pyx_v_resp = NULL;
  PyObject *__pyx_v_orders = NULL;
  PyObject *__pyx_v_filters = NULL;
  PyObject *__pyx_v_distinctOn = NULL;
  PyObject *__pyx_8genexpr2__pyx_v_sortOrder = NULL;
  PyObject *__pyx_8genexpr3__pyx_v_distinctKey = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  Py_ssize_t __pyx_t_7;
  Py_ssize_t __pyx_t_8;
  int __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  PyObject *(*__pyx_t_11)(PyObject *);
  int __pyx_t_12;
  Py_ssize_t __pyx_t_13;
  PyObject *(*__pyx_t_14)(PyObject *);
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  int __pyx_t_17;
  simdjson::dom::element __pyx_t_18;
  struct __pyx_opt_args_4viur_9datastore_9transport_toEntityStructure __pyx_t_19;
  std::string_view __pyx_t_20;
  Py_UCS4 __pyx_t_21;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("runSingleFilter", 1);

  /* "viur/datastore/transport.pyx":446
 *         :return: The list of entities fetched from the datastore
 *     """
 *     cdef simdjsonParser parser = simdjsonParser()             # <<<<<<<<<<<<<<
 *     cdef Py_ssize_t pysize
 *     cdef char * data_ptr
 */
  __pyx_v_parser = simdjson::dom::parser();

  /* "viur/datastore/transport.pyx":450
 *     cdef char * data_ptr
 *     cdef simdjsonElement element
 *     res = []             # <<<<<<<<<<<<<<
 *     internalStartCursor = None  # Will be set if we need to fetch more than one batch
 *     flipResults = False  # If set, we'll reverse the list returned (Sortorder was Inverted*)
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 450, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_res = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "viur/datastore/transport.pyx":451
 *     cdef simdjsonElement element
 *     res = []
 *     internalStartCursor = None  # Will be set if we need to fetch more than one batch             # <<<<<<<<<<<<<<
 *     flipResults = False  # If set, we'll reverse the list returned (Sortorder was Inverted*)
 *     currentTxn = currentTransaction.get()
 */
  __Pyx_INCREF(Py_None);
  __pyx_v_internalStartCursor = Py_None;

  /* "viur/datastore/transport.pyx":452
 *     res = []
 *     internalStartCursor = None  # Will be set if we need to fetch more than one batch
 *     flipResults = False  # If set, we'll reverse the list returned (Sortorder was Inverted*)             # <<<<<<<<<<<<<<
 *     currentTxn = currentTransaction.get()
 *     if currentTxn:
 */
  __Pyx_INCREF(Py_False);
  __pyx_v_flipResults = Py_False;

  /* "viur/datastore/transport.pyx":453
 *     internalStartCursor = None  # Will be set if we need to fetch more than one batch
 *     flipResults = False  # If set, we'll reverse the list returned (Sortorder was Inverted*)
 *     currentTxn = currentTransaction.get()             # <<<<<<<<<<<<<<
 *     if currentTxn:
 *         readOptions = {"transaction": currentTxn["key"]}
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_currentTransaction); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 453, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_get); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 453, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  __pyx_t_4 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_4 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_4, 0+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 453, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_v_currentTxn = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "viur/datastore/transport.pyx":454
 *     flipResults = False  # If set, we'll reverse the list returned (Sortorder was Inverted*)
 *     currentTxn = currentTransaction.get()
 *     if currentTxn:             # <<<<<<<<<<<<<<
 *         readOptions = {"transaction": currentTxn["key"]}
 *     else:
 */
  __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_v_currentTxn); if (unlikely((__pyx_t_5 < 0))) __PYX_ERR(0, 454, __pyx_L1_error)
  if (__pyx_t_5) {

    /* "viur/datastore/transport.pyx":455
 *     currentTxn = currentTransaction.get()
 *     if currentTxn:
 *         readOptions = {"transaction": currentTxn["key"]}             # <<<<<<<<<<<<<<
 *     else:
 *         readOptions = {"readConsistency": "STRONG"}
 */
    __pyx_t_1 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 455, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = __Pyx_PyObject_Dict_GetItem(__pyx_v_currentTxn, __pyx_n_u_key); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 455, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (PyDict_SetItem(__pyx_t_1, __pyx_n_u_transaction, __pyx_t_3) < 0) __PYX_ERR(0, 455, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_readOptions = ((PyObject*)__pyx_t_1);
    __pyx_t_1 = 0;

    /* "viur/datastore/transport.pyx":454
 *     flipResults = False  # If set, we'll reverse the list returned (Sortorder was Inverted*)
 *     currentTxn = currentTransaction.get()
 *     if currentTxn:             # <<<<<<<<<<<<<<
 *         readOptions = {"transaction": currentTxn["key"]}
 *     else:
 */
    goto __pyx_L3;
  }

  /* "viur/datastore/transport.pyx":457
 *         readOptions = {"transaction": currentTxn["key"]}
 *     else:
 *         readOptions = {"readConsistency": "STRONG"}             # <<<<<<<<<<<<<<
 *     while True:  # We might need to fetch more than one batch
 *         postData = {
 */
  /*else*/ {
    __pyx_t_1 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 457, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (PyDict_SetItem(__pyx_t_1, __pyx_n_u_readConsistency, __pyx_n_u_STRONG) < 0) __PYX_ERR(0, 457, __pyx_L1_error)
    __pyx_v_readOptions = ((PyObject*)__pyx_t_1);
    __pyx_t_1 = 0;
  }
  __pyx_L3:;

  /* "viur/datastore/transport.pyx":458
 *     else:
 *         readOptions = {"readConsistency": "STRONG"}
 *     while True:  # We might need to fetch more than one batch             # <<<<<<<<<<<<<<
 *         postData = {
 *             "partitionId": {
 */
  while (1) {

    /* "viur/datastore/transport.pyx":460
 *     while True:  # We might need to fetch more than one batch
 *         postData = {
 *             "partitionId": {             # <<<<<<<<<<<<<<
 *                 "project_id": projectID,
 *             },
 */
    __pyx_t_1 = __Pyx_PyDict_NewPresized(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 460, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);

    /* "viur/datastore/transport.pyx":461
 *         postData = {
 *             "partitionId": {
 *                 "project_id": projectID,             # <<<<<<<<<<<<<<
 *             },
 *             "readOptions": readOptions,
 */
    __pyx_t_3 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 461, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_projectID); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 461, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (PyDict_SetItem(__pyx_t_3, __pyx_n_u_project_id, __pyx_t_2) < 0) __PYX_ERR(0, 461, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (PyDict_SetItem(__pyx_t_1, __pyx_n_u_partitionId, __pyx_t_3) < 0) __PYX_ERR(0, 460, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "viur/datastore/transport.pyx":463
 *                 "project_id": projectID,
 *             },
 *             "readOptions": readOptions,             # <<<<<<<<<<<<<<
 *             "query": {
 *                 "kind": [
 */
    if (PyDict_SetItem(__pyx_t_1, __pyx_n_u_readOptions, __pyx_v_readOptions) < 0) __PYX_ERR(0, 460, __pyx_L1_error)

    /* "viur/datastore/transport.pyx":465
 *             "readOptions": readOptions,
 *             "query": {
 *                 "kind": [             # <<<<<<<<<<<<<<
 *                     {
 *                         "name": queryDefinition.kind,
 */
    __pyx_t_3 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 465, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);

    /* "viur/datastore/transport.pyx":467
 *                 "kind": [
 *                     {
 *                         "name": queryDefinition.kind,             # <<<<<<<<<<<<<<
 *                     }
 *                 ],
 */
    __pyx_t_2 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 467, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_queryDefinition, __pyx_n_s_kind); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 467, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (PyDict_SetItem(__pyx_t_2, __pyx_n_u_name, __pyx_t_6) < 0) __PYX_ERR(0, 467, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

    /* "viur/datastore/transport.pyx":465
 *             "readOptions": readOptions,
 *             "query": {
 *                 "kind": [             # <<<<<<<<<<<<<<
 *                     {
 *                         "name": queryDefinition.kind,
 */
    __pyx_t_6 = PyList_New(1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 465, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_GIVEREF(__pyx_t_2);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_6, 0, __pyx_t_2)) __PYX_ERR(0, 465, __pyx_L1_error);
    __pyx_t_2 = 0;
    if (PyDict_SetItem(__pyx_t_3, __pyx_n_u_kind, __pyx_t_6) < 0) __PYX_ERR(0, 465, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

    /* "viur/datastore/transport.pyx":470
 *                     }
 *                 ],
 *                 "limit": limit - len(res)             # <<<<<<<<<<<<<<
 *             },
 *         }
 */
    __pyx_t_7 = __Pyx_PyList_GET_SIZE(__pyx_v_res); if (unlikely(__pyx_t_7 == ((Py_ssize_t)-1))) __PYX_ERR(0, 470, __pyx_L1_error)
    __pyx_t_6 = PyInt_FromSsize_t(__pyx_t_7); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 470, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_2 = PyNumber_Subtract(__pyx_v_limit, __pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 470, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (PyDict_SetItem(__pyx_t_3, __pyx_n_u_limit, __pyx_t_2) < 0) __PYX_ERR(0, 465, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (PyDict_SetItem(__pyx_t_1, __pyx_n_u_query, __pyx_t_3) < 0) __PYX_ERR(0, 460, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_XDECREF_SET(__pyx_v_postData, ((PyObject*)__pyx_t_1));
    __pyx_t_1 = 0;

    /* "viur/datastore/transport.pyx":473
 *             },
 *         }
 *         if queryDefinition.filters:             # <<<<<<<<<<<<<<
 *             filterList = []
 *             for k, v in queryDefinition.filters.items():
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_queryDefinition, __pyx_n_s_filters); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 473, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_5 < 0))) __PYX_ERR(0, 473, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (__pyx_t_5) {

      /* "viur/datastore/transport.pyx":474
 *         }
 *         if queryDefinition.filters:
 *             filterList = []             # <<<<<<<<<<<<<<
 *             for k, v in queryDefinition.filters.items():
 *                 key, op = k.split(" ")
 */
      __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 474, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_XDECREF_SET(__pyx_v_filterList, ((PyObject*)__pyx_t_1));
      __pyx_t_1 = 0;

      /* "viur/datastore/transport.pyx":475
 *         if queryDefinition.filters:
 *             filterList = []
 *             for k, v in queryDefinition.filters.items():             # <<<<<<<<<<<<<<
 *                 key, op = k.split(" ")
 *                 if op == "=":
 */
      __pyx_t_7 = 0;
      __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_queryDefinition, __pyx_n_s_filters); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 475, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      if (unlikely(__pyx_t_3 == Py_None)) {
        PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "items");
        __PYX_ERR(0, 475, __pyx_L1_error)
      }
      __pyx_t_2 = __Pyx_dict_iterator(__pyx_t_3, 0, __pyx_n_s_items, (&__pyx_t_8), (&__pyx_t_4)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 475, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_1);
      __pyx_t_1 = __pyx_t_2;
      __pyx_t_2 = 0;
      while (1) {
        __pyx_t_9 = __Pyx_dict_iter_next(__pyx_t_1, __pyx_t_8, &__pyx_t_7, &__pyx_t_2, &__pyx_t_3, NULL, __pyx_t_4);
        if (unlikely(__pyx_t_9 == 0)) break;
        if (unlikely(__pyx_t_9 == -1)) __PYX_ERR(0, 475, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_XDECREF_SET(__pyx_v_k, __pyx_t_2);
        __pyx_t_2 = 0;
        __Pyx_XDECREF_SET(__pyx_v_v, __pyx_t_3);
        __pyx_t_3 = 0;

        /* "viur/datastore/transport.pyx":476
 *             filterList = []
 *             for k, v in queryDefinition.filters.items():
 *                 key, op = k.split(" ")             # <<<<<<<<<<<<<<
 *                 if op == "=":
 *                     op = "EQUAL"
 */
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_k, __pyx_n_s_split); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 476, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_6 = NULL;
        __pyx_t_9 = 0;
        #if CYTHON_UNPACK_METHODS
        if (likely(PyMethod_Check(__pyx_t_2))) {
          __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_2);
          if (likely(__pyx_t_6)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
            __Pyx_INCREF(__pyx_t_6);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_2, function);
            __pyx_t_9 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_6, __pyx_kp_u__5};
          __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_9, 1+__pyx_t_9);
          __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
          if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 476, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        }
        if ((likely(PyTuple_CheckExact(__pyx_t_3))) || (PyList_CheckExact(__pyx_t_3))) {
          PyObject* sequence = __pyx_t_3;
          Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
          if (unlikely(size != 2)) {
            if (size > 2) __Pyx_RaiseTooManyValuesError(2);
            else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
            __PYX_ERR(0, 476, __pyx_L1_error)
          }
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          if (likely(PyTuple_CheckExact(sequence))) {
            __pyx_t_2 = PyTuple_GET_ITEM(sequence, 0); 
            __pyx_t_6 = PyTuple_GET_ITEM(sequence, 1); 
          } else {
            __pyx_t_2 = PyList_GET_ITEM(sequence, 0); 
            __pyx_t_6 = PyList_GET_ITEM(sequence, 1); 
          }
          __Pyx_INCREF(__pyx_t_2);
          __Pyx_INCREF(__pyx_t_6);
          #else
          __pyx_t_2 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 476, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_6 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 476, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_6);
          #endif
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else {
          Py_ssize_t index = -1;
          __pyx_t_10 = PyObject_GetIter(__pyx_t_3); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 476, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_10);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __pyx_t_11 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_10);
          index = 0; __pyx_t_2 = __pyx_t_11(__pyx_t_10); if (unlikely(!__pyx_t_2)) goto __pyx_L9_unpacking_failed;
          __Pyx_GOTREF(__pyx_t_2);
          index = 1; __pyx_t_6 = __pyx_t_11(__pyx_t_10); if (unlikely(!__pyx_t_6)) goto __pyx_L9_unpacking_failed;
          __Pyx_GOTREF(__pyx_t_6);
          if (__Pyx_IternextUnpackEndCheck(__pyx_t_11(__pyx_t_10), 2) < 0) __PYX_ERR(0, 476, __pyx_L1_error)
          __pyx_t_11 = NULL;
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
          goto __pyx_L10_unpacking_done;
          __pyx_L9_unpacking_failed:;
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
          __pyx_t_11 = NULL;
          if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
          __PYX_ERR(0, 476, __pyx_L1_error)
          __pyx_L10_unpacking_done:;
        }
        __Pyx_XDECREF_SET(__pyx_v_key, __pyx_t_2);
        __pyx_t_2 = 0;
        __Pyx_XDECREF_SET(__pyx_v_op, __pyx_t_6);
        __pyx_t_6 = 0;

        /* "viur/datastore/transport.pyx":477
 *             for k, v in queryDefinition.filters.items():
 *                 key, op = k.split(" ")
 *                 if op == "=":             # <<<<<<<<<<<<<<
 *                     op = "EQUAL"
 *                 elif op == "<":
 */
        __pyx_t_5 = (__Pyx_PyUnicode_Equals(__pyx_v_op, __pyx_kp_u__6, Py_EQ)); if (unlikely((__pyx_t_5 < 0))) __PYX_ERR(0, 477, __pyx_L1_error)
        if (__pyx_t_5) {

          /* "viur/datastore/transport.pyx":478
 *                 key, op = k.split(" ")
 *                 if op == "=":
 *                     op = "EQUAL"             # <<<<<<<<<<<<<<
 *                 elif op == "<":
 *                     op = "LESS_THAN"
 */
          __Pyx_INCREF(__pyx_n_u_EQUAL);
          __Pyx_DECREF_SET(__pyx_v_op, __pyx_n_u_EQUAL);

          /* "viur/datastore/transport.pyx":477
 *             for k, v in queryDefinition.filters.items():
 *                 key, op = k.split(" ")
 *                 if op == "=":             # <<<<<<<<<<<<<<
 *                     op = "EQUAL"
 *                 elif op == "<":
 */
          goto __pyx_L11;
        }

        /* "viur/datastore/transport.pyx":479
 *                 if op == "=":
 *                     op = "EQUAL"
 *                 elif op == "<":             # <<<<<<<<<<<<<<
 *                     op = "LESS_THAN"
 *                 elif op == "<=":
 */
        __pyx_t_5 = (__Pyx_PyUnicode_Equals(__pyx_v_op, __pyx_kp_u__7, Py_EQ)); if (unlikely((__pyx_t_5 < 0))) __PYX_ERR(0, 479, __pyx_L1_error)
        if (__pyx_t_5) {

          /* "viur/datastore/transport.pyx":480
 *                     op = "EQUAL"
 *                 elif op == "<":
 *                     op = "LESS_THAN"             # <<<<<<<<<<<<<<
 *                 elif op == "<=":
 *                     op = "LESS_THAN_OR_EQUAL"
 */
          __Pyx_INCREF(__pyx_n_u_LESS_THAN);
          __Pyx_DECREF_SET(__pyx_v_op, __pyx_n_u_LESS_THAN);

          /* "viur/datastore/transport.pyx":479
 *                 if op == "=":
 *                     op = "EQUAL"
 *                 elif op == "<":             # <<<<<<<<<<<<<<
 *                     op = "LESS_THAN"
 *                 elif op == "<=":
 */
          goto __pyx_L11;
        }

        /* "viur/datastore/transport.pyx":481
 *                 elif op == "<":
 *                     op = "LESS_THAN"
 *                 elif op == "<=":             # <<<<<<<<<<<<<<
 *                     op = "LESS_THAN_OR_EQUAL"
 *                 elif op == ">":
 */
        __pyx_t_5 = (__Pyx_PyUnicode_Equals(__pyx_v_op, __pyx_kp_u__8, Py_EQ)); if (unlikely((__pyx_t_5 < 0))) __PYX_ERR(0, 481, __pyx_L1_error)
        if (__pyx_t_5) {

          /* "viur/datastore/transport.pyx":482
 *                     op = "LESS_THAN"
 *                 elif op == "<=":
 *                     op = "LESS_THAN_OR_EQUAL"             # <<<<<<<<<<<<<<
 *                 elif op == ">":
 *                     op = "GREATER_THAN"
 */
          __Pyx_INCREF(__pyx_n_u_LESS_THAN_OR_EQUAL);
          __Pyx_DECREF_SET(__pyx_v_op, __pyx_n_u_LESS_THAN_OR_EQUAL);

          /* "viur/datastore/transport.pyx":481
 *                 elif op == "<":
 *                     op = "LESS_THAN"
 *                 elif op == "<=":             # <<<<<<<<<<<<<<
 *                     op = "LESS_THAN_OR_EQUAL"
 *                 elif op == ">":
 */
          goto __pyx_L11;
        }

        /* "viur/datastore/transport.pyx":483
 *                 elif op == "<=":
 *                     op = "LESS_THAN_OR_EQUAL"
 *                 elif op == ">":             # <<<<<<<<<<<<<<
 *                     op = "GREATER_THAN"
 *                 elif op == ">=":
 */
        __pyx_t_5 = (__Pyx_PyUnicode_Equals(__pyx_v_op, __pyx_kp_u__9, Py_EQ)); if (unlikely((__pyx_t_5 < 0))) __PYX_ERR(0, 483, __pyx_L1_error)
        if (__pyx_t_5) {

          /* "viur/datastore/transport.pyx":484
 *                     op = "LESS_THAN_OR_EQUAL"
 *                 elif op == ">":
 *                     op = "GREATER_THAN"             # <<<<<<<<<<<<<<
 *                 elif op == ">=":
 *                     op = "GREATER_THAN_OR_EQUAL"
 */
          __Pyx_INCREF(__pyx_n_u_GREATER_THAN);
          __Pyx_DECREF_SET(__pyx_v_op, __pyx_n_u_GREATER_THAN);

          /* "viur/datastore/transport.pyx":483
 *                 elif op == "<=":
 *                     op = "LESS_THAN_OR_EQUAL"
 *                 elif op == ">":             # <<<<<<<<<<<<<<
 *                     op = "GREATER_THAN"
 *                 elif op == ">=":
 */
          goto __pyx_L11;
        }

        /* "viur/datastore/transport.pyx":485
 *                 elif op == ">":
 *                     op = "GREATER_THAN"
 *                 elif op == ">=":             # <<<<<<<<<<<<<<
 *                     op = "GREATER_THAN_OR_EQUAL"
 *                 else:
 */
        __pyx_t_5 = (__Pyx_PyUnicode_Equals(__pyx_v_op, __pyx_kp_u__10, Py_EQ)); if (unlikely((__pyx_t_5 < 0))) __PYX_ERR(0, 485, __pyx_L1_error)
        if (likely(__pyx_t_5)) {

          /* "viur/datastore/transport.pyx":486
 *                     op = "GREATER_THAN"
 *                 elif op == ">=":
 *                     op = "GREATER_THAN_OR_EQUAL"             # <<<<<<<<<<<<<<
 *                 else:
 *                     raise ValueError("Invalid op %s" % op)
 */
          __Pyx_INCREF(__pyx_n_u_GREATER_THAN_OR_EQUAL);
          __Pyx_DECREF_SET(__pyx_v_op, __pyx_n_u_GREATER_THAN_OR_EQUAL);

          /* "viur/datastore/transport.pyx":485
 *                 elif op == ">":
 *                     op = "GREATER_THAN"
 *                 elif op == ">=":             # <<<<<<<<<<<<<<
 *                     op = "GREATER_THAN_OR_EQUAL"
 *                 else:
 */
          goto __pyx_L11;
        }

        /* "viur/datastore/transport.pyx":488
 *                     op = "GREATER_THAN_OR_EQUAL"
 *                 else:
 *                     raise ValueError("Invalid op %s" % op)             # <<<<<<<<<<<<<<
 *                 if not isinstance(v, list):
 *                     # An entity can have a list of values for a single property, so it's possible to enforce
 */
        /*else*/ {
          __pyx_t_3 = __Pyx_PyUnicode_FormatSafe(__pyx_kp_u_Invalid_op_s, __pyx_v_op); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 488, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_6 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_3); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 488, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_6);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_Raise(__pyx_t_6, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
          __PYX_ERR(0, 488, __pyx_L1_error)
        }
        __pyx_L11:;

        /* "viur/datastore/transport.pyx":489
 *                 else:
 *                     raise ValueError("Invalid op %s" % op)
 *                 if not isinstance(v, list):             # <<<<<<<<<<<<<<
 *                     # An entity can have a list of values for a single property, so it's possible to enforce
 *                     # more an one constraint for a a single property (e.g. x==5 and x==7 can be true), so
 */
        __pyx_t_5 = PyList_Check(__pyx_v_v); 
        __pyx_t_12 = (!__pyx_t_5);
        if (__pyx_t_12) {

          /* "viur/datastore/transport.pyx":493
 *                     # more an one constraint for a a single property (e.g. x==5 and x==7 can be true), so
 *                     # enforce we always have a list here
 *                     v = [v]             # <<<<<<<<<<<<<<
 *                 for singleValue in v:
 *                     filterList.append({
 */
          __pyx_t_6 = PyList_New(1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 493, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_6);
          __Pyx_INCREF(__pyx_v_v);
          __Pyx_GIVEREF(__pyx_v_v);
          if (__Pyx_PyList_SET_ITEM(__pyx_t_6, 0, __pyx_v_v)) __PYX_ERR(0, 493, __pyx_L1_error);
          __Pyx_DECREF_SET(__pyx_v_v, __pyx_t_6);
          __pyx_t_6 = 0;

          /* "viur/datastore/transport.pyx":489
 *                 else:
 *                     raise ValueError("Invalid op %s" % op)
 *                 if not isinstance(v, list):             # <<<<<<<<<<<<<<
 *                     # An entity can have a list of values for a single property, so it's possible to enforce
 *                     # more an one constraint for a a single property (e.g. x==5 and x==7 can be true), so
 */
        }

        /* "viur/datastore/transport.pyx":494
 *                     # enforce we always have a list here
 *                     v = [v]
 *                 for singleValue in v:             # <<<<<<<<<<<<<<
 *                     filterList.append({
 *                         "propertyFilter": {
 */
        if (likely(PyList_CheckExact(__pyx_v_v)) || PyTuple_CheckExact(__pyx_v_v)) {
          __pyx_t_6 = __pyx_v_v; __Pyx_INCREF(__pyx_t_6);
          __pyx_t_13 = 0;
          __pyx_t_14 = NULL;
        } else {
          __pyx_t_13 = -1; __pyx_t_6 = PyObject_GetIter(__pyx_v_v); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 494, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_6);
          __pyx_t_14 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_6); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 494, __pyx_L1_error)
        }
        for (;;) {
          if (likely(!__pyx_t_14)) {
            if (likely(PyList_CheckExact(__pyx_t_6))) {
              {
                Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_6);
                #if !CYTHON_ASSUME_SAFE_MACROS
                if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 494, __pyx_L1_error)
                #endif
                if (__pyx_t_13 >= __pyx_temp) break;
              }
              #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
              __pyx_t_3 = PyList_GET_ITEM(__pyx_t_6, __pyx_t_13); __Pyx_INCREF(__pyx_t_3); __pyx_t_13++; if (unlikely((0 < 0))) __PYX_ERR(0, 494, __pyx_L1_error)
              #else
              __pyx_t_3 = __Pyx_PySequence_ITEM(__pyx_t_6, __pyx_t_13); __pyx_t_13++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 494, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_3);
              #endif
            } else {
              {
                Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_6);
                #if !CYTHON_ASSUME_SAFE_MACROS
                if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 494, __pyx_L1_error)
                #endif
                if (__pyx_t_13 >= __pyx_temp) break;
              }
              #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
              __pyx_t_3 = PyTuple_GET_ITEM(__pyx_t_6, __pyx_t_13); __Pyx_INCREF(__pyx_t_3); __pyx_t_13++; if (unlikely((0 < 0))) __PYX_ERR(0, 494, __pyx_L1_error)
              #else
              __pyx_t_3 = __Pyx_PySequence_ITEM(__pyx_t_6, __pyx_t_13); __pyx_t_13++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 494, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_3);
              #endif
            }
          } else {
            __pyx_t_3 = __pyx_t_14(__pyx_t_6);
            if (unlikely(!__pyx_t_3)) {
              PyObject* exc_type = PyErr_Occurred();
              if (exc_type) {
                if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
                else __PYX_ERR(0, 494, __pyx_L1_error)
              }
              break;
            }
            __Pyx_GOTREF(__pyx_t_3);
          }
          __Pyx_XDECREF_SET(__pyx_v_singleValue, __pyx_t_3);
          __pyx_t_3 = 0;

          /* "viur/datastore/transport.pyx":496
 *                 for singleValue in v:
 *                     filterList.append({
 *                         "propertyFilter": {             # <<<<<<<<<<<<<<
 *                             "property": {
 *                                 "name": key,
 */
          __pyx_t_3 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 496, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_3);

          /* "viur/datastore/transport.pyx":497
 *                     filterList.append({
 *                         "propertyFilter": {
 *                             "property": {             # <<<<<<<<<<<<<<
 *                                 "name": key,
 *                             },
 */
          __pyx_t_2 = __Pyx_PyDict_NewPresized(3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 497, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);

          /* "viur/datastore/transport.pyx":498
 *                         "propertyFilter": {
 *                             "property": {
 *                                 "name": key,             # <<<<<<<<<<<<<<
 *                             },
 *                             "op": op,
 */
          __pyx_t_10 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 498, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_10);
          if (PyDict_SetItem(__pyx_t_10, __pyx_n_u_name, __pyx_v_key) < 0) __PYX_ERR(0, 498, __pyx_L1_error)
          if (PyDict_SetItem(__pyx_t_2, __pyx_n_u_property, __pyx_t_10) < 0) __PYX_ERR(0, 497, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

          /* "viur/datastore/transport.pyx":500
 *                                 "name": key,
 *                             },
 *                             "op": op,             # <<<<<<<<<<<<<<
 *                             "value": pythonPropToJson(singleValue)
 *                         }
 */
          if (PyDict_SetItem(__pyx_t_2, __pyx_n_u_op, __pyx_v_op) < 0) __PYX_ERR(0, 497, __pyx_L1_error)

          /* "viur/datastore/transport.pyx":501
 *                             },
 *                             "op": op,
 *                             "value": pythonPropToJson(singleValue)             # <<<<<<<<<<<<<<
 *                         }
 *                     })
 */
          __Pyx_GetModuleGlobalName(__pyx_t_15, __pyx_n_s_pythonPropToJson); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 501, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_15);
          __pyx_t_16 = NULL;
          __pyx_t_9 = 0;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_15))) {
            __pyx_t_16 = PyMethod_GET_SELF(__pyx_t_15);
            if (likely(__pyx_t_16)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_15);
              __Pyx_INCREF(__pyx_t_16);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_15, function);
              __pyx_t_9 = 1;
            }
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_16, __pyx_v_singleValue};
            __pyx_t_10 = __Pyx_PyObject_FastCall(__pyx_t_15, __pyx_callargs+1-__pyx_t_9, 1+__pyx_t_9);
            __Pyx_XDECREF(__pyx_t_16); __pyx_t_16 = 0;
            if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 501, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_10);
            __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
          }
          if (PyDict_SetItem(__pyx_t_2, __pyx_n_u_value, __pyx_t_10) < 0) __PYX_ERR(0, 497, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
          if (PyDict_SetItem(__pyx_t_3, __pyx_n_u_propertyFilter, __pyx_t_2) < 0) __PYX_ERR(0, 496, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

          /* "viur/datastore/transport.pyx":495
 *                     v = [v]
 *                 for singleValue in v:
 *                     filterList.append({             # <<<<<<<<<<<<<<
 *                         "propertyFilter": {
 *                             "property": {
 */
          __pyx_t_17 = __Pyx_PyList_Append(__pyx_v_filterList, __pyx_t_3); if (unlikely(__pyx_t_17 == ((int)-1))) __PYX_ERR(0, 495, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "viur/datastore/transport.pyx":494
 *                     # enforce we always have a list here
 *                     v = [v]
 *                 for singleValue in v:             # <<<<<<<<<<<<<<
 *                     filterList.append({
 *                         "propertyFilter": {
 */
        }
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "viur/datastore/transport.pyx":504
 *                         }
 *                     })
 *             if len(filterList) == 1:  # Special, single filter             # <<<<<<<<<<<<<<
 *                 postData["query"]["filter"] = filterList[0]
 *             else:
 */
      __pyx_t_8 = __Pyx_PyList_GET_SIZE(__pyx_v_filterList); if (unlikely(__pyx_t_8 == ((Py_ssize_t)-1))) __PYX_ERR(0, 504, __pyx_L1_error)
      __pyx_t_12 = (__pyx_t_8 == 1);
      if (__pyx_t_12) {

        /* "viur/datastore/transport.pyx":505
 *                     })
 *             if len(filterList) == 1:  # Special, single filter
 *                 postData["query"]["filter"] = filterList[0]             # <<<<<<<<<<<<<<
 *             else:
 *                 postData["query"]["filter"] = {
 */
        __pyx_t_1 = __Pyx_GetItemInt_List(__pyx_v_filterList, 0, long, 1, __Pyx_PyInt_From_long, 1, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 505, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_6 = __Pyx_PyDict_GetItem(__pyx_v_postData, __pyx_n_u_query); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 505, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        if (unlikely((PyObject_SetItem(__pyx_t_6, __pyx_n_u_filter, __pyx_t_1) < 0))) __PYX_ERR(0, 505, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "viur/datastore/transport.pyx":504
 *                         }
 *                     })
 *             if len(filterList) == 1:  # Special, single filter             # <<<<<<<<<<<<<<
 *                 postData["query"]["filter"] = filterList[0]
 *             else:
 */
        goto __pyx_L16;
      }

      /* "viur/datastore/transport.pyx":508
 *             else:
 *                 postData["query"]["filter"] = {
 *                     "compositeFilter": {             # <<<<<<<<<<<<<<
 *                         "op": "AND",
 *                         "filters": filterList
 */
      /*else*/ {
        __pyx_t_1 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 508, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);

        /* "viur/datastore/transport.pyx":509
 *                 postData["query"]["filter"] = {
 *                     "compositeFilter": {
 *                         "op": "AND",             # <<<<<<<<<<<<<<
 *                         "filters": filterList
 *                     }
 */
        __pyx_t_6 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 509, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        if (PyDict_SetItem(__pyx_t_6, __pyx_n_u_op, __pyx_n_u_AND) < 0) __PYX_ERR(0, 509, __pyx_L1_error)

        /* "viur/datastore/transport.pyx":510
 *                     "compositeFilter": {
 *                         "op": "AND",
 *                         "filters": filterList             # <<<<<<<<<<<<<<
 *                     }
 *                 }
 */
        if (PyDict_SetItem(__pyx_t_6, __pyx_n_u_filters, __pyx_v_filterList) < 0) __PYX_ERR(0, 509, __pyx_L1_error)
        if (PyDict_SetItem(__pyx_t_1, __pyx_n_u_compositeFilter, __pyx_t_6) < 0) __PYX_ERR(0, 508, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

        /* "viur/datastore/transport.pyx":507
 *                 postData["query"]["filter"] = filterList[0]
 *             else:
 *                 postData["query"]["filter"] = {             # <<<<<<<<<<<<<<
 *                     "compositeFilter": {
 *                         "op": "AND",
 */
        __pyx_t_6 = __Pyx_PyDict_GetItem(__pyx_v_postData, __pyx_n_u_query); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 507, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        if (unlikely((PyObject_SetItem(__pyx_t_6, __pyx_n_u_filter, __pyx_t_1) < 0))) __PYX_ERR(0, 507, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      }
      __pyx_L16:;

      /* "viur/datastore/transport.pyx":473
 *             },
 *         }
 *         if queryDefinition.filters:             # <<<<<<<<<<<<<<
 *             filterList = []
 *             for k, v in queryDefinition.filters.items():
 */
    }

    /* "viur/datastore/transport.pyx":513
 *                     }
 *                 }
 *         if queryDefinition.orders:             # <<<<<<<<<<<<<<
 *             postData["query"]["order"] = [
 *                 {
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_queryDefinition, __pyx_n_s_orders); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 513, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_12 < 0))) __PYX_ERR(0, 513, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (__pyx_t_12) {

      /* "viur/datastore/transport.pyx":514
 *                 }
 *         if queryDefinition.orders:
 *             postData["query"]["order"] = [             # <<<<<<<<<<<<<<
 *                 {
 *                     "property": {"name": sortOrder[0]},
 */
      { /* enter inner scope */
        __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 514, __pyx_L20_error)
        __Pyx_GOTREF(__pyx_t_1);

        /* "viur/datastore/transport.pyx":518
 *                     "property": {"name": sortOrder[0]},
 *                     "direction": "ASCENDING" if sortOrder[1].value in [1, 4] else "DESCENDING"
 *                 } for sortOrder in queryDefinition.orders             # <<<<<<<<<<<<<<
 *             ]
 *             flipResults = queryDefinition.orders[0][1].value > 2  # Either InvertedAscending or InvertedDescending
 */
        __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_queryDefinition, __pyx_n_s_orders); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 518, __pyx_L20_error)
        __Pyx_GOTREF(__pyx_t_6);
        if (likely(PyList_CheckExact(__pyx_t_6)) || PyTuple_CheckExact(__pyx_t_6)) {
          __pyx_t_3 = __pyx_t_6; __Pyx_INCREF(__pyx_t_3);
          __pyx_t_8 = 0;
          __pyx_t_14 = NULL;
        } else {
          __pyx_t_8 = -1; __pyx_t_3 = PyObject_GetIter(__pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 518, __pyx_L20_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_14 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_3); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 518, __pyx_L20_error)
        }
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        for (;;) {
          if (likely(!__pyx_t_14)) {
            if (likely(PyList_CheckExact(__pyx_t_3))) {
              {
                Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_3);
                #if !CYTHON_ASSUME_SAFE_MACROS
                if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 518, __pyx_L20_error)
                #endif
                if (__pyx_t_8 >= __pyx_temp) break;
              }
              #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
              __pyx_t_6 = PyList_GET_ITEM(__pyx_t_3, __pyx_t_8); __Pyx_INCREF(__pyx_t_6); __pyx_t_8++; if (unlikely((0 < 0))) __PYX_ERR(0, 518, __pyx_L20_error)
              #else
              __pyx_t_6 = __Pyx_PySequence_ITEM(__pyx_t_3, __pyx_t_8); __pyx_t_8++; if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 518, __pyx_L20_error)
              __Pyx_GOTREF(__pyx_t_6);
              #endif
            } else {
              {
                Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_3);
                #if !CYTHON_ASSUME_SAFE_MACROS
                if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 518, __pyx_L20_error)
                #endif
                if (__pyx_t_8 >= __pyx_temp) break;
              }
              #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
              __pyx_t_6 = PyTuple_GET_ITEM(__pyx_t_3, __pyx_t_8); __Pyx_INCREF(__pyx_t_6); __pyx_t_8++; if (unlikely((0 < 0))) __PYX_ERR(0, 518, __pyx_L20_error)
              #else
              __pyx_t_6 = __Pyx_PySequence_ITEM(__pyx_t_3, __pyx_t_8); __pyx_t_8++; if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 518, __pyx_L20_error)
              __Pyx_GOTREF(__pyx_t_6);
              #endif
            }
          } else {
            __pyx_t_6 = __pyx_t_14(__pyx_t_3);
            if (unlikely(!__pyx_t_6)) {
              PyObject* exc_type = PyErr_Occurred();
              if (exc_type) {
                if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
                else __PYX_ERR(0, 518, __pyx_L20_error)
              }
              break;
            }
            __Pyx_GOTREF(__pyx_t_6);
          }
          __Pyx_XDECREF_SET(__pyx_8genexpr2__pyx_v_sortOrder, __pyx_t_6);
          __pyx_t_6 = 0;

          /* "viur/datastore/transport.pyx":516
 *             postData["query"]["order"] = [
 *                 {
 *                     "property": {"name": sortOrder[0]},             # <<<<<<<<<<<<<<
 *                     "direction": "ASCENDING" if sortOrder[1].value in [1, 4] else "DESCENDING"
 *                 } for sortOrder in queryDefinition.orders
 */
          __pyx_t_6 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 516, __pyx_L20_error)
          __Pyx_GOTREF(__pyx_t_6);
          __pyx_t_2 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 516, __pyx_L20_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_10 = __Pyx_GetItemInt(__pyx_8genexpr2__pyx_v_sortOrder, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 516, __pyx_L20_error)
          __Pyx_GOTREF(__pyx_t_10);
          if (PyDict_SetItem(__pyx_t_2, __pyx_n_u_name, __pyx_t_10) < 0) __PYX_ERR(0, 516, __pyx_L20_error)
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
          if (PyDict_SetItem(__pyx_t_6, __pyx_n_u_property, __pyx_t_2) < 0) __PYX_ERR(0, 516, __pyx_L20_error)
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

          /* "viur/datastore/transport.pyx":517
 *                 {
 *                     "property": {"name": sortOrder[0]},
 *                     "direction": "ASCENDING" if sortOrder[1].value in [1, 4] else "DESCENDING"             # <<<<<<<<<<<<<<
 *                 } for sortOrder in queryDefinition.orders
 *             ]
 */
          __pyx_t_10 = __Pyx_GetItemInt(__pyx_8genexpr2__pyx_v_sortOrder, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 517, __pyx_L20_error)
          __Pyx_GOTREF(__pyx_t_10);
          __pyx_t_15 = __Pyx_PyObject_GetAttrStr(__pyx_t_10, __pyx_n_s_value); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 517, __pyx_L20_error)
          __Pyx_GOTREF(__pyx_t_15);
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
          __pyx_t_5 = (__Pyx_PyInt_BoolEqObjC(__pyx_t_15, __pyx_int_1, 1, 0)); if (unlikely((__pyx_t_5 < 0))) __PYX_ERR(0, 517, __pyx_L20_error)
          if (!__pyx_t_5) {
          } else {
            __pyx_t_12 = __pyx_t_5;
            goto __pyx_L23_bool_binop_done;
          }
          __pyx_t_5 = (__Pyx_PyInt_BoolEqObjC(__pyx_t_15, __pyx_int_4, 4, 0)); if (unlikely((__pyx_t_5 < 0))) __PYX_ERR(0, 517, __pyx_L20_error)
          __pyx_t_12 = __pyx_t_5;
          __pyx_L23_bool_binop_done:;
          __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
          __pyx_t_5 = __pyx_t_12;
          if (__pyx_t_5) {
            __Pyx_INCREF(__pyx_n_u_ASCENDING);
            __pyx_t_2 = __pyx_n_u_ASCENDING;
          } else {
            __Pyx_INCREF(__pyx_n_u_DESCENDING);
            __pyx_t_2 = __pyx_n_u_DESCENDING;
          }
          if (PyDict_SetItem(__pyx_t_6, __pyx_n_u_direction, __pyx_t_2) < 0) __PYX_ERR(0, 516, __pyx_L20_error)
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          if (unlikely(__Pyx_ListComp_Append(__pyx_t_1, (PyObject*)__pyx_t_6))) __PYX_ERR(0, 514, __pyx_L20_error)
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

          /* "viur/datastore/transport.pyx":518
 *                     "property": {"name": sortOrder[0]},
 *                     "direction": "ASCENDING" if sortOrder[1].value in [1, 4] else "DESCENDING"
 *                 } for sortOrder in queryDefinition.orders             # <<<<<<<<<<<<<<
 *             ]
 *             flipResults = queryDefinition.orders[0][1].value > 2  # Either InvertedAscending or InvertedDescending
 */
        }
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_XDECREF(__pyx_8genexpr2__pyx_v_sortOrder); __pyx_8genexpr2__pyx_v_sortOrder = 0;
        goto __pyx_L26_exit_scope;
        __pyx_L20_error:;
        __Pyx_XDECREF(__pyx_8genexpr2__pyx_v_sortOrder); __pyx_8genexpr2__pyx_v_sortOrder = 0;
        goto __pyx_L1_error;
        __pyx_L26_exit_scope:;
      } /* exit inner scope */

      /* "viur/datastore/transport.pyx":514
 *                 }
 *         if queryDefinition.orders:
 *             postData["query"]["order"] = [             # <<<<<<<<<<<<<<
 *                 {
 *                     "property": {"name": sortOrder[0]},
 */
      __pyx_t_3 = __Pyx_PyDict_GetItem(__pyx_v_postData, __pyx_n_u_query); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 514, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      if (unlikely((PyObject_SetItem(__pyx_t_3, __pyx_n_u_order, __pyx_t_1) < 0))) __PYX_ERR(0, 514, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "viur/datastore/transport.pyx":520
 *                 } for sortOrder in queryDefinition.orders
 *             ]
 *             flipResults = queryDefinition.orders[0][1].value > 2  # Either InvertedAscending or InvertedDescending             # <<<<<<<<<<<<<<
 *         if queryDefinition.distinct:
 *             postData["query"]["distinctOn"] = [
 */
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_queryDefinition, __pyx_n_s_orders); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 520, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_3 = __Pyx_GetItemInt(__pyx_t_1, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 520, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_1 = __Pyx_GetItemInt(__pyx_t_3, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 520, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_value); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 520, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_1 = PyObject_RichCompare(__pyx_t_3, __pyx_int_2, Py_GT); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 520, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF_SET(__pyx_v_flipResults, __pyx_t_1);
      __pyx_t_1 = 0;

      /* "viur/datastore/transport.pyx":513
 *                     }
 *                 }
 *         if queryDefinition.orders:             # <<<<<<<<<<<<<<
 *             postData["query"]["order"] = [
 *                 {
 */
    }

    /* "viur/datastore/transport.pyx":521
 *             ]
 *             flipResults = queryDefinition.orders[0][1].value > 2  # Either InvertedAscending or InvertedDescending
 *         if queryDefinition.distinct:             # <<<<<<<<<<<<<<
 *             postData["query"]["distinctOn"] = [
 *                 {
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_queryDefinition, __pyx_n_s_distinct); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 521, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_5 < 0))) __PYX_ERR(0, 521, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (__pyx_t_5) {

      /* "viur/datastore/transport.pyx":522
 *             flipResults = queryDefinition.orders[0][1].value > 2  # Either InvertedAscending or InvertedDescending
 *         if queryDefinition.distinct:
 *             postData["query"]["distinctOn"] = [             # <<<<<<<<<<<<<<
 *                 {
 *                     "name": distinctKey
 */
      { /* enter inner scope */
        __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 522, __pyx_L30_error)
        __Pyx_GOTREF(__pyx_t_1);

        /* "viur/datastore/transport.pyx":525
 *                 {
 *                     "name": distinctKey
 *                 } for distinctKey in queryDefinition.distinct             # <<<<<<<<<<<<<<
 *             ]
 *         if queryDefinition.startCursor or internalStartCursor:
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_queryDefinition, __pyx_n_s_distinct); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 525, __pyx_L30_error)
        __Pyx_GOTREF(__pyx_t_3);
        if (likely(PyList_CheckExact(__pyx_t_3)) || PyTuple_CheckExact(__pyx_t_3)) {
          __pyx_t_6 = __pyx_t_3; __Pyx_INCREF(__pyx_t_6);
          __pyx_t_8 = 0;
          __pyx_t_14 = NULL;
        } else {
          __pyx_t_8 = -1; __pyx_t_6 = PyObject_GetIter(__pyx_t_3); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 525, __pyx_L30_error)
          __Pyx_GOTREF(__pyx_t_6);
          __pyx_t_14 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_6); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 525, __pyx_L30_error)
        }
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        for (;;) {
          if (likely(!__pyx_t_14)) {
            if (likely(PyList_CheckExact(__pyx_t_6))) {
              {
                Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_6);
                #if !CYTHON_ASSUME_SAFE_MACROS
                if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 525, __pyx_L30_error)
                #endif
                if (__pyx_t_8 >= __pyx_temp) break;
              }
              #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
              __pyx_t_3 = PyList_GET_ITEM(__pyx_t_6, __pyx_t_8); __Pyx_INCREF(__pyx_t_3); __pyx_t_8++; if (unlikely((0 < 0))) __PYX_ERR(0, 525, __pyx_L30_error)
              #else
              __pyx_t_3 = __Pyx_PySequence_ITEM(__pyx_t_6, __pyx_t_8); __pyx_t_8++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 525, __pyx_L30_error)
              __Pyx_GOTREF(__pyx_t_3);
              #endif
            } else {
              {
                Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_6);
                #if !CYTHON_ASSUME_SAFE_MACROS
                if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 525, __pyx_L30_error)
                #endif
                if (__pyx_t_8 >= __pyx_temp) break;
              }
              #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
              __pyx_t_3 = PyTuple_GET_ITEM(__pyx_t_6, __pyx_t_8); __Pyx_INCREF(__pyx_t_3); __pyx_t_8++; if (unlikely((0 < 0))) __PYX_ERR(0, 525, __pyx_L30_error)
              #else
              __pyx_t_3 = __Pyx_PySequence_ITEM(__pyx_t_6, __pyx_t_8); __pyx_t_8++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 525, __pyx_L30_error)
              __Pyx_GOTREF(__pyx_t_3);
              #endif
            }
          } else {
            __pyx_t_3 = __pyx_t_14(__pyx_t_6);
            if (unlikely(!__pyx_t_3)) {
              PyObject* exc_type = PyErr_Occurred();
              if (exc_type) {
                if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
                else __PYX_ERR(0, 525, __pyx_L30_error)
              }
              break;
            }
            __Pyx_GOTREF(__pyx_t_3);
          }
          __Pyx_XDECREF_SET(__pyx_8genexpr3__pyx_v_distinctKey, __pyx_t_3);
          __pyx_t_3 = 0;

          /* "viur/datastore/transport.pyx":524
 *             postData["query"]["distinctOn"] = [
 *                 {
 *                     "name": distinctKey             # <<<<<<<<<<<<<<
 *                 } for distinctKey in queryDefinition.distinct
 *             ]
 */
          __pyx_t_3 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 524, __pyx_L30_error)
          __Pyx_GOTREF(__pyx_t_3);
          if (PyDict_SetItem(__pyx_t_3, __pyx_n_u_name, __pyx_8genexpr3__pyx_v_distinctKey) < 0) __PYX_ERR(0, 524, __pyx_L30_error)
          if (unlikely(__Pyx_ListComp_Append(__pyx_t_1, (PyObject*)__pyx_t_3))) __PYX_ERR(0, 522, __pyx_L30_error)
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "viur/datastore/transport.pyx":525
 *                 {
 *                     "name": distinctKey
 *                 } for distinctKey in queryDefinition.distinct             # <<<<<<<<<<<<<<
 *             ]
 *         if queryDefinition.startCursor or internalStartCursor:
 */
        }
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_XDECREF(__pyx_8genexpr3__pyx_v_distinctKey); __pyx_8genexpr3__pyx_v_distinctKey = 0;
        goto __pyx_L34_exit_scope;
        __pyx_L30_error:;
        __Pyx_XDECREF(__pyx_8genexpr3__pyx_v_distinctKey); __pyx_8genexpr3__pyx_v_distinctKey = 0;
        goto __pyx_L1_error;
        __pyx_L34_exit_scope:;
      } /* exit inner scope */

      /* "viur/datastore/transport.pyx":522
 *             flipResults = queryDefinition.orders[0][1].value > 2  # Either InvertedAscending or InvertedDescending
 *         if queryDefinition.distinct:
 *             postData["query"]["distinctOn"] = [             # <<<<<<<<<<<<<<
 *                 {
 *                     "name": distinctKey
 */
      __pyx_t_6 = __Pyx_PyDict_GetItem(__pyx_v_postData, __pyx_n_u_query); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 522, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      if (unlikely((PyObject_SetItem(__pyx_t_6, __pyx_n_u_distinctOn, __pyx_t_1) < 0))) __PYX_ERR(0, 522, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "viur/datastore/transport.pyx":521
 *             ]
 *             flipResults = queryDefinition.orders[0][1].value > 2  # Either InvertedAscending or InvertedDescending
 *         if queryDefinition.distinct:             # <<<<<<<<<<<<<<
 *             postData["query"]["distinctOn"] = [
 *                 {
 */
    }

    /* "viur/datastore/transport.pyx":527
 *                 } for distinctKey in queryDefinition.distinct
 *             ]
 *         if queryDefinition.startCursor or internalStartCursor:             # <<<<<<<<<<<<<<
 *             postData["query"]["startCursor"] = internalStartCursor or queryDefinition.startCursor
 *         if queryDefinition.endCursor:
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_queryDefinition, __pyx_n_s_startCursor); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 527, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_12 < 0))) __PYX_ERR(0, 527, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (!__pyx_t_12) {
    } else {
      __pyx_t_5 = __pyx_t_12;
      goto __pyx_L36_bool_binop_done;
    }
    __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_v_internalStartCursor); if (unlikely((__pyx_t_12 < 0))) __PYX_ERR(0, 527, __pyx_L1_error)
    __pyx_t_5 = __pyx_t_12;
    __pyx_L36_bool_binop_done:;
    if (__pyx_t_5) {

      /* "viur/datastore/transport.pyx":528
 *             ]
 *         if queryDefinition.startCursor or internalStartCursor:
 *             postData["query"]["startCursor"] = internalStartCursor or queryDefinition.startCursor             # <<<<<<<<<<<<<<
 *         if queryDefinition.endCursor:
 *             postData["query"]["endCursor"] = queryDefinition.endCursor
 */
      __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_v_internalStartCursor); if (unlikely((__pyx_t_5 < 0))) __PYX_ERR(0, 528, __pyx_L1_error)
      if (!__pyx_t_5) {
      } else {
        __Pyx_INCREF(__pyx_v_internalStartCursor);
        __pyx_t_1 = __pyx_v_internalStartCursor;
        goto __pyx_L38_bool_binop_done;
      }
      __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_queryDefinition, __pyx_n_s_startCursor); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 528, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_INCREF(__pyx_t_6);
      __pyx_t_1 = __pyx_t_6;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_L38_bool_binop_done:;
      __pyx_t_6 = __Pyx_PyDict_GetItem(__pyx_v_postData, __pyx_n_u_query); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 528, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      if (unlikely((PyObject_SetItem(__pyx_t_6, __pyx_n_u_startCursor, __pyx_t_1) < 0))) __PYX_ERR(0, 528, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "viur/datastore/transport.pyx":527
 *                 } for distinctKey in queryDefinition.distinct
 *             ]
 *         if queryDefinition.startCursor or internalStartCursor:             # <<<<<<<<<<<<<<
 *             postData["query"]["startCursor"] = internalStartCursor or queryDefinition.startCursor
 *         if queryDefinition.endCursor:
 */
    }

    /* "viur/datastore/transport.pyx":529
 *         if queryDefinition.startCursor or internalStartCursor:
 *             postData["query"]["startCursor"] = internalStartCursor or queryDefinition.startCursor
 *         if queryDefinition.endCursor:             # <<<<<<<<<<<<<<
 *             postData["query"]["endCursor"] = queryDefinition.endCursor
 *         resp = authenticated_request(
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_queryDefinition, __pyx_n_s_endCursor); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 529, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_5 < 0))) __PYX_ERR(0, 529, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (__pyx_t_5) {

      /* "viur/datastore/transport.pyx":530
 *             postData["query"]["startCursor"] = internalStartCursor or queryDefinition.startCursor
 *         if queryDefinition.endCursor:
 *             postData["query"]["endCursor"] = queryDefinition.endCursor             # <<<<<<<<<<<<<<
 *         resp = authenticated_request(
 *             url="https://datastore.googleapis.com/v1/projects/%s:runQuery" % projectID,
 */
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_queryDefinition, __pyx_n_s_endCursor); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 530, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_6 = __Pyx_PyDict_GetItem(__pyx_v_postData, __pyx_n_u_query); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 530, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      if (unlikely((PyObject_SetItem(__pyx_t_6, __pyx_n_u_endCursor, __pyx_t_1) < 0))) __PYX_ERR(0, 530, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "viur/datastore/transport.pyx":529
 *         if queryDefinition.startCursor or internalStartCursor:
 *             postData["query"]["startCursor"] = internalStartCursor or queryDefinition.startCursor
 *         if queryDefinition.endCursor:             # <<<<<<<<<<<<<<
 *             postData["query"]["endCursor"] = queryDefinition.endCursor
 *         resp = authenticated_request(
 */
    }

    /* "viur/datastore/transport.pyx":531
 *         if queryDefinition.endCursor:
 *             postData["query"]["endCursor"] = queryDefinition.endCursor
 *         resp = authenticated_request(             # <<<<<<<<<<<<<<
 *             url="https://datastore.googleapis.com/v1/projects/%s:runQuery" % projectID,
 *             data=json.dumps(postData).encode("UTF-8"),
 */
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_authenticated_request); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 531, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);

    /* "viur/datastore/transport.pyx":532
 *             postData["query"]["endCursor"] = queryDefinition.endCursor
 *         resp = authenticated_request(
 *             url="https://datastore.googleapis.com/v1/projects/%s:runQuery" % projectID,             # <<<<<<<<<<<<<<
 *             data=json.dumps(postData).encode("UTF-8"),
 *         )
 */
    __pyx_t_6 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 532, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_projectID); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 532, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_2 = __Pyx_PyUnicode_FormatSafe(__pyx_kp_u_https_datastore_googleapis_com_v, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 532, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (PyDict_SetItem(__pyx_t_6, __pyx_n_s_url, __pyx_t_2) < 0) __PYX_ERR(0, 532, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "viur/datastore/transport.pyx":533
 *         resp = authenticated_request(
 *             url="https://datastore.googleapis.com/v1/projects/%s:runQuery" % projectID,
 *             data=json.dumps(postData).encode("UTF-8"),             # <<<<<<<<<<<<<<
 *         )
 * 
 */
    __Pyx_GetModuleGlobalName(__pyx_t_15, __pyx_n_s_json); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 533, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_15);
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_t_15, __pyx_n_s_dumps); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 533, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
    __pyx_t_15 = NULL;
    __pyx_t_4 = 0;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_10))) {
      __pyx_t_15 = PyMethod_GET_SELF(__pyx_t_10);
      if (likely(__pyx_t_15)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_10);
        __Pyx_INCREF(__pyx_t_15);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_10, function);
        __pyx_t_4 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_15, __pyx_v_postData};
      __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_10, __pyx_callargs+1-__pyx_t_4, 1+__pyx_t_4);
      __Pyx_XDECREF(__pyx_t_15); __pyx_t_15 = 0;
      if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 533, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    }
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_encode); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 533, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = NULL;
    __pyx_t_4 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_10))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_10);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_10);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_10, function);
        __pyx_t_4 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_kp_u_UTF_8};
      __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_10, __pyx_callargs+1-__pyx_t_4, 1+__pyx_t_4);
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 533, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    }
    if (PyDict_SetItem(__pyx_t_6, __pyx_n_s_data, __pyx_t_2) < 0) __PYX_ERR(0, 532, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "viur/datastore/transport.pyx":531
 *         if queryDefinition.endCursor:
 *             postData["query"]["endCursor"] = queryDefinition.endCursor
 *         resp = authenticated_request(             # <<<<<<<<<<<<<<
 *             url="https://datastore.googleapis.com/v1/projects/%s:runQuery" % projectID,
 *             data=json.dumps(postData).encode("UTF-8"),
 */
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_empty_tuple, __pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 531, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_XDECREF_SET(__pyx_v_resp, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "viur/datastore/transport.pyx":536
 *         )
 * 
 *         is_viur_datastore_request_ok(resp)             # <<<<<<<<<<<<<<
 *         assert PyBytes_AsStringAndSize(resp.content, &data_ptr, &pysize) != -1
 *         element = parser.parse(data_ptr, pysize, 1)
 */
    __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_is_viur_datastore_request_ok); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 536, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_1 = NULL;
    __pyx_t_4 = 0;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_1)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_1);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
        __pyx_t_4 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_1, __pyx_v_resp};
      __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_6, __pyx_callargs+1-__pyx_t_4, 1+__pyx_t_4);
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 536, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "viur/datastore/transport.pyx":537
 * 
 *         is_viur_datastore_request_ok(resp)
 *         assert PyBytes_AsStringAndSize(resp.content, &data_ptr, &pysize) != -1             # <<<<<<<<<<<<<<
 *         element = parser.parse(data_ptr, pysize, 1)
 *         if element.at_pointer("/batch").error() != SUCCESS:
 */
    #ifndef CYTHON_WITHOUT_ASSERTIONS
    if (unlikely(__pyx_assertions_enabled())) {
      __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_resp, __pyx_n_s_content); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 537, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_4 = PyBytes_AsStringAndSize(__pyx_t_2, (&__pyx_v_data_ptr), (&__pyx_v_pysize)); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(0, 537, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_5 = (__pyx_t_4 != -1L);
      if (unlikely(!__pyx_t_5)) {
        __Pyx_Raise(__pyx_builtin_AssertionError, 0, 0, 0);
        __PYX_ERR(0, 537, __pyx_L1_error)
      }
    }
    #else
    if ((1)); else __PYX_ERR(0, 537, __pyx_L1_error)
    #endif

    /* "viur/datastore/transport.pyx":538
 *         is_viur_datastore_request_ok(resp)
 *         assert PyBytes_AsStringAndSize(resp.content, &data_ptr, &pysize) != -1
 *         element = parser.parse(data_ptr, pysize, 1)             # <<<<<<<<<<<<<<
 *         if element.at_pointer("/batch").error() != SUCCESS:
 *             logging.error("INVALID RESPONSE RECEIVED")
 */
    try {
      __pyx_t_18 = __pyx_v_parser.parse(__pyx_v_data_ptr, __pyx_v_pysize, 1);
    } catch(...) {
      __Pyx_CppExn2PyErr();
      __PYX_ERR(0, 538, __pyx_L1_error)
    }
    __pyx_v_element = __PYX_STD_MOVE_IF_SUPPORTED(__pyx_t_18);

    /* "viur/datastore/transport.pyx":539
 *         assert PyBytes_AsStringAndSize(resp.content, &data_ptr, &pysize) != -1
 *         element = parser.parse(data_ptr, pysize, 1)
 *         if element.at_pointer("/batch").error() != SUCCESS:             # <<<<<<<<<<<<<<
 *             logging.error("INVALID RESPONSE RECEIVED")
 *             logging.error(json.loads(resp.content))
 */
    __pyx_t_5 = (__pyx_v_element.at_pointer(((char const *)"/batch")).error() != simdjson::error_code::SUCCESS);
    if (__pyx_t_5) {

      /* "viur/datastore/transport.pyx":540
 *         element = parser.parse(data_ptr, pysize, 1)
 *         if element.at_pointer("/batch").error() != SUCCESS:
 *             logging.error("INVALID RESPONSE RECEIVED")             # <<<<<<<<<<<<<<
 *             logging.error(json.loads(resp.content))
 *         #	res.update(toEntityStructure(element.at_key("batch"), isInitial=True))
 */
      __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_logging); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 540, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_error); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 540, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_6 = NULL;
      __pyx_t_4 = 0;
      #if CYTHON_UNPACK_METHODS
      if (unlikely(PyMethod_Check(__pyx_t_1))) {
        __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_1);
        if (likely(__pyx_t_6)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
          __Pyx_INCREF(__pyx_t_6);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_1, function);
          __pyx_t_4 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_6, __pyx_kp_u_INVALID_RESPONSE_RECEIVED};
        __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_1, __pyx_callargs+1-__pyx_t_4, 1+__pyx_t_4);
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 540, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      }
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

      /* "viur/datastore/transport.pyx":541
 *         if element.at_pointer("/batch").error() != SUCCESS:
 *             logging.error("INVALID RESPONSE RECEIVED")
 *             logging.error(json.loads(resp.content))             # <<<<<<<<<<<<<<
 *         #	res.update(toEntityStructure(element.at_key("batch"), isInitial=True))
 *         element = element.at_key("batch")
 */
      __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_logging); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 541, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_error); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 541, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_GetModuleGlobalName(__pyx_t_10, __pyx_n_s_json); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 541, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_10, __pyx_n_s_loads); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 541, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v_resp, __pyx_n_s_content); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 541, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __pyx_t_15 = NULL;
      __pyx_t_4 = 0;
      #if CYTHON_UNPACK_METHODS
      if (unlikely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_15 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_15)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_15);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
          __pyx_t_4 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_15, __pyx_t_10};
        __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_4, 1+__pyx_t_4);
        __Pyx_XDECREF(__pyx_t_15); __pyx_t_15 = 0;
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
        if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 541, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      }
      __pyx_t_3 = NULL;
      __pyx_t_4 = 0;
      #if CYTHON_UNPACK_METHODS
      if (unlikely(PyMethod_Check(__pyx_t_6))) {
        __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_6);
        if (likely(__pyx_t_3)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
          __Pyx_INCREF(__pyx_t_3);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_6, function);
          __pyx_t_4 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_t_1};
        __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_6, __pyx_callargs+1-__pyx_t_4, 1+__pyx_t_4);
        __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 541, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

      /* "viur/datastore/transport.pyx":539
 *         assert PyBytes_AsStringAndSize(resp.content, &data_ptr, &pysize) != -1
 *         element = parser.parse(data_ptr, pysize, 1)
 *         if element.at_pointer("/batch").error() != SUCCESS:             # <<<<<<<<<<<<<<
 *             logging.error("INVALID RESPONSE RECEIVED")
 *             logging.error(json.loads(resp.content))
 */
    }

    /* "viur/datastore/transport.pyx":543
 *             logging.error(json.loads(resp.content))
 *         #	res.update(toEntityStructure(element.at_key("batch"), isInitial=True))
 *         element = element.at_key("batch")             # <<<<<<<<<<<<<<
 *         if element.at_pointer("/entityResults").error() == SUCCESS:
 *             res.extend(toEntityStructure(element.at_key("entityResults"), isInitial=False))
 */
    try {
      __pyx_t_18 = __pyx_v_element.at_key(((char const *)"batch"));
    } catch(...) {
      __Pyx_CppExn2PyErr();
      __PYX_ERR(0, 543, __pyx_L1_error)
    }
    __pyx_v_element = __PYX_STD_MOVE_IF_SUPPORTED(__pyx_t_18);

    /* "viur/datastore/transport.pyx":544
 *         #	res.update(toEntityStructure(element.at_key("batch"), isInitial=True))
 *         element = element.at_key("batch")
 *         if element.at_pointer("/entityResults").error() == SUCCESS:             # <<<<<<<<<<<<<<
 *             res.extend(toEntityStructure(element.at_key("entityResults"), isInitial=False))
 *         else:  # No results received
 */
    __pyx_t_5 = (__pyx_v_element.at_pointer(((char const *)"/entityResults")).error() == simdjson::error_code::SUCCESS);
    if (__pyx_t_5) {

      /* "viur/datastore/transport.pyx":545
 *         element = element.at_key("batch")
 *         if element.at_pointer("/entityResults").error() == SUCCESS:
 *             res.extend(toEntityStructure(element.at_key("entityResults"), isInitial=False))             # <<<<<<<<<<<<<<
 *         else:  # No results received
 *             if toPyStr(element.at_key("moreResults").get_string()) == "NOT_FINISHED":
 */
      try {
        __pyx_t_18 = __pyx_v_element.at_key(((char const *)"entityResults"));
      } catch(...) {
        __Pyx_CppExn2PyErr();
        __PYX_ERR(0, 545, __pyx_L1_error)
      }
      __pyx_t_19.__pyx_n = 1;
      __pyx_t_19.isInitial = 0;
      __pyx_t_2 = __pyx_f_4viur_9datastore_9transport_toEntityStructure(__PYX_STD_MOVE_IF_SUPPORTED(__pyx_t_18), &__pyx_t_19); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 545, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_17 = __Pyx_PyList_Extend(__pyx_v_res, __pyx_t_2); if (unlikely(__pyx_t_17 == ((int)-1))) __PYX_ERR(0, 545, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

      /* "viur/datastore/transport.pyx":544
 *         #	res.update(toEntityStructure(element.at_key("batch"), isInitial=True))
 *         element = element.at_key("batch")
 *         if element.at_pointer("/entityResults").error() == SUCCESS:             # <<<<<<<<<<<<<<
 *             res.extend(toEntityStructure(element.at_key("entityResults"), isInitial=False))
 *         else:  # No results received
 */
      goto __pyx_L42;
    }

    /* "viur/datastore/transport.pyx":547
 *             res.extend(toEntityStructure(element.at_key("entityResults"), isInitial=False))
 *         else:  # No results received
 *             if toPyStr(element.at_key("moreResults").get_string()) == "NOT_FINISHED":             # <<<<<<<<<<<<<<
 *                 logging.warning("Query not finished. Maybe some entries are missing.")
 *                 logging.warning("Queried %s with filter %s and orders %s. - Please create a matching index" % (
 */
    /*else*/ {
      try {
        __pyx_t_18 = __pyx_v_element.at_key(((char const *)"moreResults"));
      } catch(...) {
        __Pyx_CppExn2PyErr();
        __PYX_ERR(0, 547, __pyx_L1_error)
      }
      try {
        __pyx_t_20 = __pyx_t_18.get_string();
      } catch(...) {
        __Pyx_CppExn2PyErr();
        __PYX_ERR(0, 547, __pyx_L1_error)
      }
      __pyx_t_2 = __pyx_f_4viur_9datastore_9transport_toPyStr(__PYX_STD_MOVE_IF_SUPPORTED(__pyx_t_20)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 547, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_5 = (__Pyx_PyUnicode_Equals(__pyx_t_2, __pyx_n_u_NOT_FINISHED, Py_EQ)); if (unlikely((__pyx_t_5 < 0))) __PYX_ERR(0, 547, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      if (__pyx_t_5) {

        /* "viur/datastore/transport.pyx":548
 *         else:  # No results received
 *             if toPyStr(element.at_key("moreResults").get_string()) == "NOT_FINISHED":
 *                 logging.warning("Query not finished. Maybe some entries are missing.")             # <<<<<<<<<<<<<<
 *                 logging.warning("Queried %s with filter %s and orders %s. - Please create a matching index" % (
 *                     queryDefinition.kind, queryDefinition.filters, queryDefinition.orders))
 */
        __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_logging); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 548, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_warning); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 548, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __pyx_t_6 = NULL;
        __pyx_t_4 = 0;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_1))) {
          __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_1);
          if (likely(__pyx_t_6)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
            __Pyx_INCREF(__pyx_t_6);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_1, function);
            __pyx_t_4 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_6, __pyx_kp_u_Query_not_finished_Maybe_some_en};
          __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_1, __pyx_callargs+1-__pyx_t_4, 1+__pyx_t_4);
          __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
          if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 548, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        }
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

        /* "viur/datastore/transport.pyx":549
 *             if toPyStr(element.at_key("moreResults").get_string()) == "NOT_FINISHED":
 *                 logging.warning("Query not finished. Maybe some entries are missing.")
 *                 logging.warning("Queried %s with filter %s and orders %s. - Please create a matching index" % (             # <<<<<<<<<<<<<<
 *                     queryDefinition.kind, queryDefinition.filters, queryDefinition.orders))
 *             else:
 */
        __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_logging); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 549, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_warning); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 549, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_1 = PyTuple_New(7); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 549, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_8 = 0;
        __pyx_t_21 = 127;
        __Pyx_INCREF(__pyx_kp_u_Queried);
        __pyx_t_8 += 8;
        __Pyx_GIVEREF(__pyx_kp_u_Queried);
        PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_kp_u_Queried);

        /* "viur/datastore/transport.pyx":550
 *                 logging.warning("Query not finished. Maybe some entries are missing.")
 *                 logging.warning("Queried %s with filter %s and orders %s. - Please create a matching index" % (
 *                     queryDefinition.kind, queryDefinition.filters, queryDefinition.orders))             # <<<<<<<<<<<<<<
 *             else:
 *                 break
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_queryDefinition, __pyx_n_s_kind); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 550, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_10 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Unicode(__pyx_t_3), __pyx_empty_unicode); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 550, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_10);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_21 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_10) > __pyx_t_21) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_10) : __pyx_t_21;
        __pyx_t_8 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_10);
        __Pyx_GIVEREF(__pyx_t_10);
        PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_10);
        __pyx_t_10 = 0;
        __Pyx_INCREF(__pyx_kp_u_with_filter);
        __pyx_t_8 += 13;
        __Pyx_GIVEREF(__pyx_kp_u_with_filter);
        PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_kp_u_with_filter);
        __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v_queryDefinition, __pyx_n_s_filters); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 550, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_10);
        __pyx_t_3 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Unicode(__pyx_t_10), __pyx_empty_unicode); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 550, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
        __pyx_t_21 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_3) > __pyx_t_21) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_3) : __pyx_t_21;
        __pyx_t_8 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_3);
        __Pyx_GIVEREF(__pyx_t_3);
        PyTuple_SET_ITEM(__pyx_t_1, 3, __pyx_t_3);
        __pyx_t_3 = 0;
        __Pyx_INCREF(__pyx_kp_u_and_orders);
        __pyx_t_8 += 12;
        __Pyx_GIVEREF(__pyx_kp_u_and_orders);
        PyTuple_SET_ITEM(__pyx_t_1, 4, __pyx_kp_u_and_orders);
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_queryDefinition, __pyx_n_s_orders); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 550, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_10 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Unicode(__pyx_t_3), __pyx_empty_unicode); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 550, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_10);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_21 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_10) > __pyx_t_21) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_10) : __pyx_t_21;
        __pyx_t_8 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_10);
        __Pyx_GIVEREF(__pyx_t_10);
        PyTuple_SET_ITEM(__pyx_t_1, 5, __pyx_t_10);
        __pyx_t_10 = 0;
        __Pyx_INCREF(__pyx_kp_u_Please_create_a_matching_index);
        __pyx_t_8 += 34;
        __Pyx_GIVEREF(__pyx_kp_u_Please_create_a_matching_index);
        PyTuple_SET_ITEM(__pyx_t_1, 6, __pyx_kp_u_Please_create_a_matching_index);

        /* "viur/datastore/transport.pyx":549
 *             if toPyStr(element.at_key("moreResults").get_string()) == "NOT_FINISHED":
 *                 logging.warning("Query not finished. Maybe some entries are missing.")
 *                 logging.warning("Queried %s with filter %s and orders %s. - Please create a matching index" % (             # <<<<<<<<<<<<<<
 *                     queryDefinition.kind, queryDefinition.filters, queryDefinition.orders))
 *             else:
 */
        __pyx_t_10 = __Pyx_PyUnicode_Join(__pyx_t_1, 7, __pyx_t_8, __pyx_t_21); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 549, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_10);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_1 = NULL;
        __pyx_t_4 = 0;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_6))) {
          __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_6);
          if (likely(__pyx_t_1)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
            __Pyx_INCREF(__pyx_t_1);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_6, function);
            __pyx_t_4 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_1, __pyx_t_10};
          __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_6, __pyx_callargs+1-__pyx_t_4, 1+__pyx_t_4);
          __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
          if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 549, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        }
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

        /* "viur/datastore/transport.pyx":547
 *             res.extend(toEntityStructure(element.at_key("entityResults"), isInitial=False))
 *         else:  # No results received
 *             if toPyStr(element.at_key("moreResults").get_string()) == "NOT_FINISHED":             # <<<<<<<<<<<<<<
 *                 logging.warning("Query not finished. Maybe some entries are missing.")
 *                 logging.warning("Queried %s with filter %s and orders %s. - Please create a matching index" % (
 */
        goto __pyx_L43;
      }

      /* "viur/datastore/transport.pyx":552
 *                     queryDefinition.kind, queryDefinition.filters, queryDefinition.orders))
 *             else:
 *                 break             # <<<<<<<<<<<<<<
 *         if toPyStr(element.at_key("moreResults").get_string()) == "NO_MORE_RESULTS":
 *             internalStartCursor = None
 */
      /*else*/ {
        goto __pyx_L5_break;
      }
      __pyx_L43:;
    }
    __pyx_L42:;

    /* "viur/datastore/transport.pyx":553
 *             else:
 *                 break
 *         if toPyStr(element.at_key("moreResults").get_string()) == "NO_MORE_RESULTS":             # <<<<<<<<<<<<<<
 *             internalStartCursor = None
 *         else:
 */
    try {
      __pyx_t_18 = __pyx_v_element.at_key(((char const *)"moreResults"));
    } catch(...) {
      __Pyx_CppExn2PyErr();
      __PYX_ERR(0, 553, __pyx_L1_error)
    }
    try {
      __pyx_t_20 = __pyx_t_18.get_string();
    } catch(...) {
      __Pyx_CppExn2PyErr();
      __PYX_ERR(0, 553, __pyx_L1_error)
    }
    __pyx_t_2 = __pyx_f_4viur_9datastore_9transport_toPyStr(__PYX_STD_MOVE_IF_SUPPORTED(__pyx_t_20)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 553, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = (__Pyx_PyUnicode_Equals(__pyx_t_2, __pyx_n_u_NO_MORE_RESULTS, Py_EQ)); if (unlikely((__pyx_t_5 < 0))) __PYX_ERR(0, 553, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (__pyx_t_5) {

      /* "viur/datastore/transport.pyx":554
 *                 break
 *         if toPyStr(element.at_key("moreResults").get_string()) == "NO_MORE_RESULTS":
 *             internalStartCursor = None             # <<<<<<<<<<<<<<
 *         else:
 *             internalStartCursor = toPyStr(element.at_key("endCursor").get_string())
 */
      __Pyx_INCREF(Py_None);
      __Pyx_DECREF_SET(__pyx_v_internalStartCursor, Py_None);

      /* "viur/datastore/transport.pyx":553
 *             else:
 *                 break
 *         if toPyStr(element.at_key("moreResults").get_string()) == "NO_MORE_RESULTS":             # <<<<<<<<<<<<<<
 *             internalStartCursor = None
 *         else:
 */
      goto __pyx_L44;
    }

    /* "viur/datastore/transport.pyx":556
 *             internalStartCursor = None
 *         else:
 *             internalStartCursor = toPyStr(element.at_key("endCursor").get_string())             # <<<<<<<<<<<<<<
 *         if toPyStr(element.at_key("moreResults").get_string()) != "NOT_FINISHED" or len(res) == limit:
 *             break
 */
    /*else*/ {
      try {
        __pyx_t_18 = __pyx_v_element.at_key(((char const *)"endCursor"));
      } catch(...) {
        __Pyx_CppExn2PyErr();
        __PYX_ERR(0, 556, __pyx_L1_error)
      }
      try {
        __pyx_t_20 = __pyx_t_18.get_string();
      } catch(...) {
        __Pyx_CppExn2PyErr();
        __PYX_ERR(0, 556, __pyx_L1_error)
      }
      __pyx_t_2 = __pyx_f_4viur_9datastore_9transport_toPyStr(__PYX_STD_MOVE_IF_SUPPORTED(__pyx_t_20)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 556, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF_SET(__pyx_v_internalStartCursor, __pyx_t_2);
      __pyx_t_2 = 0;
    }
    __pyx_L44:;

    /* "viur/datastore/transport.pyx":557
 *         else:
 *             internalStartCursor = toPyStr(element.at_key("endCursor").get_string())
 *         if toPyStr(element.at_key("moreResults").get_string()) != "NOT_FINISHED" or len(res) == limit:             # <<<<<<<<<<<<<<
 *             break
 *     queryDefinition.currentCursor = internalStartCursor
 */
    try {
      __pyx_t_18 = __pyx_v_element.at_key(((char const *)"moreResults"));
    } catch(...) {
      __Pyx_CppExn2PyErr();
      __PYX_ERR(0, 557, __pyx_L1_error)
    }
    try {
      __pyx_t_20 = __pyx_t_18.get_string();
    } catch(...) {
      __Pyx_CppExn2PyErr();
      __PYX_ERR(0, 557, __pyx_L1_error)
    }
    __pyx_t_2 = __pyx_f_4viur_9datastore_9transport_toPyStr(__PYX_STD_MOVE_IF_SUPPORTED(__pyx_t_20)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 557, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_12 = (__Pyx_PyUnicode_Equals(__pyx_t_2, __pyx_n_u_NOT_FINISHED, Py_NE)); if (unlikely((__pyx_t_12 < 0))) __PYX_ERR(0, 557, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (!__pyx_t_12) {
    } else {
      __pyx_t_5 = __pyx_t_12;
      goto __pyx_L46_bool_binop_done;
    }
    __pyx_t_8 = __Pyx_PyList_GET_SIZE(__pyx_v_res); if (unlikely(__pyx_t_8 == ((Py_ssize_t)-1))) __PYX_ERR(0, 557, __pyx_L1_error)
    __pyx_t_2 = PyInt_FromSsize_t(__pyx_t_8); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 557, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_6 = PyObject_RichCompare(__pyx_t_2, __pyx_v_limit, Py_EQ); __Pyx_XGOTREF(__pyx_t_6); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 557, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_6); if (unlikely((__pyx_t_12 < 0))) __PYX_ERR(0, 557, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_5 = __pyx_t_12;
    __pyx_L46_bool_binop_done:;
    if (__pyx_t_5) {

      /* "viur/datastore/transport.pyx":558
 *             internalStartCursor = toPyStr(element.at_key("endCursor").get_string())
 *         if toPyStr(element.at_key("moreResults").get_string()) != "NOT_FINISHED" or len(res) == limit:
 *             break             # <<<<<<<<<<<<<<
 *     queryDefinition.currentCursor = internalStartCursor
 *     if conf["traceQueries"]:
 */
      goto __pyx_L5_break;

      /* "viur/datastore/transport.pyx":557
 *         else:
 *             internalStartCursor = toPyStr(element.at_key("endCursor").get_string())
 *         if toPyStr(element.at_key("moreResults").get_string()) != "NOT_FINISHED" or len(res) == limit:             # <<<<<<<<<<<<<<
 *             break
 *     queryDefinition.currentCursor = internalStartCursor
 */
    }
  }
  __pyx_L5_break:;

  /* "viur/datastore/transport.pyx":559
 *         if toPyStr(element.at_key("moreResults").get_string()) != "NOT_FINISHED" or len(res) == limit:
 *             break
 *     queryDefinition.currentCursor = internalStartCursor             # <<<<<<<<<<<<<<
 *     if conf["traceQueries"]:
 *         orders = queryDefinition.orders
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_queryDefinition, __pyx_n_s_currentCursor, __pyx_v_internalStartCursor) < 0) __PYX_ERR(0, 559, __pyx_L1_error)

  /* "viur/datastore/transport.pyx":560
 *             break
 *     queryDefinition.currentCursor = internalStartCursor
 *     if conf["traceQueries"]:             # <<<<<<<<<<<<<<
 *         orders = queryDefinition.orders
 *         filters = queryDefinition.filters
 */
  __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_conf); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 560, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_2 = __Pyx_PyObject_Dict_GetItem(__pyx_t_6, __pyx_n_u_traceQueries); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 560, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely((__pyx_t_5 < 0))) __PYX_ERR(0, 560, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__pyx_t_5) {

    /* "viur/datastore/transport.pyx":561
 *     queryDefinition.currentCursor = internalStartCursor
 *     if conf["traceQueries"]:
 *         orders = queryDefinition.orders             # <<<<<<<<<<<<<<
 *         filters = queryDefinition.filters
 *         distinctOn = " distinct on %s" % str(queryDefinition.distinct) if queryDefinition.distinct else ""
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_queryDefinition, __pyx_n_s_orders); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 561, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_v_orders = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "viur/datastore/transport.pyx":562
 *     if conf["traceQueries"]:
 *         orders = queryDefinition.orders
 *         filters = queryDefinition.filters             # <<<<<<<<<<<<<<
 *         distinctOn = " distinct on %s" % str(queryDefinition.distinct) if queryDefinition.distinct else ""
 *         logging.debug("Queried %s with filter %s and orders %s%s. Returned %s results" % (
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_queryDefinition, __pyx_n_s_filters); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 562, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_v_filters = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "viur/datastore/transport.pyx":563
 *         orders = queryDefinition.orders
 *         filters = queryDefinition.filters
 *         distinctOn = " distinct on %s" % str(queryDefinition.distinct) if queryDefinition.distinct else ""             # <<<<<<<<<<<<<<
 *         logging.debug("Queried %s with filter %s and orders %s%s. Returned %s results" % (
 *             queryDefinition.kind, filters, orders, distinctOn, len(res)))
 */
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_queryDefinition, __pyx_n_s_distinct); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 563, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_6); if (unlikely((__pyx_t_5 < 0))) __PYX_ERR(0, 563, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (__pyx_t_5) {
      __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_queryDefinition, __pyx_n_s_distinct); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 563, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_10 = __Pyx_PyObject_Str(__pyx_t_6); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 563, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_6 = __Pyx_PyUnicode_FormatSafe(__pyx_kp_u_distinct_on_s, __pyx_t_10); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 563, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __pyx_t_2 = __pyx_t_6;
      __pyx_t_6 = 0;
    } else {
      __Pyx_INCREF(__pyx_kp_u__11);
      __pyx_t_2 = __pyx_kp_u__11;
    }
    __pyx_v_distinctOn = ((PyObject*)__pyx_t_2);
    __pyx_t_2 = 0;

    /* "viur/datastore/transport.pyx":564
 *         filters = queryDefinition.filters
 *         distinctOn = " distinct on %s" % str(queryDefinition.distinct) if queryDefinition.distinct else ""
 *         logging.debug("Queried %s with filter %s and orders %s%s. Returned %s results" % (             # <<<<<<<<<<<<<<
 *             queryDefinition.kind, filters, orders, distinctOn, len(res)))
 *     if flipResults:
 */
    __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_logging); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 564, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_debug); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 564, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_6 = PyTuple_New(10); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 564, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_8 = 0;
    __pyx_t_21 = 127;
    __Pyx_INCREF(__pyx_kp_u_Queried);
    __pyx_t_8 += 8;
    __Pyx_GIVEREF(__pyx_kp_u_Queried);
    PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_kp_u_Queried);

    /* "viur/datastore/transport.pyx":565
 *         distinctOn = " distinct on %s" % str(queryDefinition.distinct) if queryDefinition.distinct else ""
 *         logging.debug("Queried %s with filter %s and orders %s%s. Returned %s results" % (
 *             queryDefinition.kind, filters, orders, distinctOn, len(res)))             # <<<<<<<<<<<<<<
 *     if flipResults:
 *         return res[::-1]
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_queryDefinition, __pyx_n_s_kind); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 565, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Unicode(__pyx_t_1), __pyx_empty_unicode); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 565, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_21 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_3) > __pyx_t_21) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_3) : __pyx_t_21;
    __pyx_t_8 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_t_3);
    __pyx_t_3 = 0;
    __Pyx_INCREF(__pyx_kp_u_with_filter);
    __pyx_t_8 += 13;
    __Pyx_GIVEREF(__pyx_kp_u_with_filter);
    PyTuple_SET_ITEM(__pyx_t_6, 2, __pyx_kp_u_with_filter);
    __pyx_t_3 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Unicode(__pyx_v_filters), __pyx_empty_unicode); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 565, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_21 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_3) > __pyx_t_21) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_3) : __pyx_t_21;
    __pyx_t_8 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_6, 3, __pyx_t_3);
    __pyx_t_3 = 0;
    __Pyx_INCREF(__pyx_kp_u_and_orders);
    __pyx_t_8 += 12;
    __Pyx_GIVEREF(__pyx_kp_u_and_orders);
    PyTuple_SET_ITEM(__pyx_t_6, 4, __pyx_kp_u_and_orders);
    __pyx_t_3 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Unicode(__pyx_v_orders), __pyx_empty_unicode); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 565, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_21 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_3) > __pyx_t_21) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_3) : __pyx_t_21;
    __pyx_t_8 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_6, 5, __pyx_t_3);
    __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_PyUnicode_Unicode(__pyx_v_distinctOn); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 565, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_21 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_3) > __pyx_t_21) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_3) : __pyx_t_21;
    __pyx_t_8 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_6, 6, __pyx_t_3);
    __pyx_t_3 = 0;
    __Pyx_INCREF(__pyx_kp_u_Returned);
    __pyx_t_8 += 11;
    __Pyx_GIVEREF(__pyx_kp_u_Returned);
    PyTuple_SET_ITEM(__pyx_t_6, 7, __pyx_kp_u_Returned);
    __pyx_t_7 = __Pyx_PyList_GET_SIZE(__pyx_v_res); if (unlikely(__pyx_t_7 == ((Py_ssize_t)-1))) __PYX_ERR(0, 565, __pyx_L1_error)
    __pyx_t_3 = __Pyx_PyUnicode_From_Py_ssize_t(__pyx_t_7, 0, ' ', 'd'); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 565, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_8 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_6, 8, __pyx_t_3);
    __pyx_t_3 = 0;
    __Pyx_INCREF(__pyx_kp_u_results);
    __pyx_t_8 += 8;
    __Pyx_GIVEREF(__pyx_kp_u_results);
    PyTuple_SET_ITEM(__pyx_t_6, 9, __pyx_kp_u_results);

    /* "viur/datastore/transport.pyx":564
 *         filters = queryDefinition.filters
 *         distinctOn = " distinct on %s" % str(queryDefinition.distinct) if queryDefinition.distinct else ""
 *         logging.debug("Queried %s with filter %s and orders %s%s. Returned %s results" % (             # <<<<<<<<<<<<<<
 *             queryDefinition.kind, filters, orders, distinctOn, len(res)))
 *     if flipResults:
 */
    __pyx_t_3 = __Pyx_PyUnicode_Join(__pyx_t_6, 10, __pyx_t_8, __pyx_t_21); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 564, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_6 = NULL;
    __pyx_t_4 = 0;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_10))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_10);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_10);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_10, function);
        __pyx_t_4 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_6, __pyx_t_3};
      __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_10, __pyx_callargs+1-__pyx_t_4, 1+__pyx_t_4);
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 564, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "viur/datastore/transport.pyx":560
 *             break
 *     queryDefinition.currentCursor = internalStartCursor
 *     if conf["traceQueries"]:             # <<<<<<<<<<<<<<
 *         orders = queryDefinition.orders
 *         filters = queryDefinition.filters
 */
  }

  /* "viur/datastore/transport.pyx":566
 *         logging.debug("Queried %s with filter %s and orders %s%s. Returned %s results" % (
 *             queryDefinition.kind, filters, orders, distinctOn, len(res)))
 *     if flipResults:             # <<<<<<<<<<<<<<
 *         return res[::-1]
 *     return res
 */
  __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_v_flipResults); if (unlikely((__pyx_t_5 < 0))) __PYX_ERR(0, 566, __pyx_L1_error)
  if (__pyx_t_5) {

    /* "viur/datastore/transport.pyx":567
 *             queryDefinition.kind, filters, orders, distinctOn, len(res)))
 *     if flipResults:
 *         return res[::-1]             # <<<<<<<<<<<<<<
 *     return res
 * 
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = __Pyx_PyObject_GetItem(__pyx_v_res, __pyx_slice__12); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 567, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_r = ((PyObject*)__pyx_t_2);
    __pyx_t_2 = 0;
    goto __pyx_L0;

    /* "viur/datastore/transport.pyx":566
 *         logging.debug("Queried %s with filter %s and orders %s%s. Returned %s results" % (
 *             queryDefinition.kind, filters, orders, distinctOn, len(res)))
 *     if flipResults:             # <<<<<<<<<<<<<<
 *         return res[::-1]
 *     return res
 */
  }

  /* "viur/datastore/transport.pyx":568
 *     if flipResults:
 *         return res[::-1]
 *     return res             # <<<<<<<<<<<<<<
 * 
 * def Get(keys: Union[Key, List[Key]]) -> Union[None, Entity, List[Entity]]:
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_res);
  __pyx_r = __pyx_v_res;
  goto __pyx_L0;

  /* "viur/datastore/transport.pyx":437
 *         raise ValueError("Unknown type")
 * 
 * def runSingleFilter(queryDefinition: QueryDefinition, limit: int) -> List[Entity]:             # <<<<<<<<<<<<<<
 *     """
 *         Runs a single Query as defined by queryDefinition. The limit of the queryDefinition is ignored and must
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_15);
  __Pyx_XDECREF(__pyx_t_16);
  __Pyx_AddTraceback("viur.datastore.transport.runSingleFilter", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_res);
  __Pyx_XDECREF(__pyx_v_internalStartCursor);
  __Pyx_XDECREF(__pyx_v_flipResults);
  __Pyx_XDECREF(__pyx_v_currentTxn);
  __Pyx_XDECREF(__pyx_v_readOptions);
  __Pyx_XDECREF(__pyx_v_postData);
  __Pyx_XDECREF(__pyx_v_filterList);
  __Pyx_XDECREF(__pyx_v_k);
  __Pyx_XDECREF(__pyx_v_v);
  __Pyx_XDECREF(__pyx_v_key);
  __Pyx_XDECREF(__pyx_v_op);
  __Pyx_XDECREF(__pyx_v_singleValue);
  __Pyx_XDECREF(__pyx_v_resp);
  __Pyx_XDECREF(__pyx_v_orders);
  __Pyx_XDECREF(__pyx_v_filters);
  __Pyx_XDECREF(__pyx_v_distinctOn);
  __Pyx_XDECREF(__pyx_8genexpr2__pyx_v_sortOrder);
  __Pyx_XDECREF(__pyx_8genexpr3__pyx_v_distinctKey);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "viur/datastore/transport.pyx":570
 *     return res
 * 
 * def Get(keys: Union[Key, List[Key]]) -> Union[None, Entity, List[Entity]]:             # <<<<<<<<<<<<<<
 *     """
 *         Fetches the entities determined by keys from the datastore. Returns or inserts None if a key is not found.
 */

/* Python wrapper */
static PyObject *__pyx_pw_4viur_9datastore_9transport_9Get(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_4viur_9datastore_9transport_8Get, "\n        Fetches the entities determined by keys from the datastore. Returns or inserts None if a key is not found.\n        :param keys: A Key or a List of Keys to fetch\n        :return: The entity or None for the given key, a list of Entities/None if a list has been supplied\n    ");
static PyMethodDef __pyx_mdef_4viur_9datastore_9transport_9Get = {"Get", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_4viur_9datastore_9transport_9Get, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_4viur_9datastore_9transport_8Get};
static PyObject *__pyx_pw_4viur_9datastore_9transport_9Get(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_keys = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("Get (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_keys,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_keys)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 570, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "Get") < 0)) __PYX_ERR(0, 570, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_keys = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("Get", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 570, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("viur.datastore.transport.Get", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4viur_9datastore_9transport_8Get(__pyx_self, __pyx_v_keys);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_4viur_9datastore_9transport_3Get_2generator(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "viur/datastore/transport.pyx":584
 *         keys = [keys]
 *         isMulti = False
 *     if any(key.is_partial or key.kind is None for key in keys):             # <<<<<<<<<<<<<<
 *         raise InvalidArgumentError("Keys must not be partial or kind-less")
 *     accessLog = currentDbAccessLog.get()
 */

static PyObject *__pyx_pf_4viur_9datastore_9transport_3Get_genexpr(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_genexpr_arg_0) {
  struct __pyx_obj_4viur_9datastore_9transport___pyx_scope_struct__genexpr *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("genexpr", 0);
  __pyx_cur_scope = (struct __pyx_obj_4viur_9datastore_9transport___pyx_scope_struct__genexpr *)__pyx_tp_new_4viur_9datastore_9transport___pyx_scope_struct__genexpr(__pyx_ptype_4viur_9datastore_9transport___pyx_scope_struct__genexpr, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_4viur_9datastore_9transport___pyx_scope_struct__genexpr *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 584, __pyx_L1_error)
  } else {
    __Pyx_GOTREF((PyObject *)__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_genexpr_arg_0 = __pyx_genexpr_arg_0;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_genexpr_arg_0);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_genexpr_arg_0);
  {
    __pyx_CoroutineObject *gen = __Pyx_Generator_New((__pyx_coroutine_body_t) __pyx_gb_4viur_9datastore_9transport_3Get_2generator, NULL, (PyObject *) __pyx_cur_scope, __pyx_n_s_genexpr, __pyx_n_s_Get_locals_genexpr, __pyx_n_s_viur_datastore_transport); if (unlikely(!gen)) __PYX_ERR(0, 584, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("viur.datastore.transport.Get.genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF((PyObject *)__pyx_cur_scope);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_4viur_9datastore_9transport_3Get_2generator(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_4viur_9datastore_9transport___pyx_scope_struct__genexpr *__pyx_cur_scope = ((struct __pyx_obj_4viur_9datastore_9transport___pyx_scope_struct__genexpr *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  PyObject *(*__pyx_t_3)(PyObject *);
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("genexpr", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 584, __pyx_L1_error)
  if (unlikely(!__pyx_cur_scope->__pyx_genexpr_arg_0)) { __Pyx_RaiseUnboundLocalError(".0"); __PYX_ERR(0, 584, __pyx_L1_error) }
  if (likely(PyList_CheckExact(__pyx_cur_scope->__pyx_genexpr_arg_0)) || PyTuple_CheckExact(__pyx_cur_scope->__pyx_genexpr_arg_0)) {
    __pyx_t_1 = __pyx_cur_scope->__pyx_genexpr_arg_0; __Pyx_INCREF(__pyx_t_1);
    __pyx_t_2 = 0;
    __pyx_t_3 = NULL;
  } else {
    __pyx_t_2 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_cur_scope->__pyx_genexpr_arg_0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 584, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 584, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_3)) {
      if (likely(PyList_CheckExact(__pyx_t_1))) {
        {
          Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_1);
          #if !CYTHON_ASSUME_SAFE_MACROS
          if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 584, __pyx_L1_error)
          #endif
          if (__pyx_t_2 >= __pyx_temp) break;
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_4); __pyx_t_2++; if (unlikely((0 < 0))) __PYX_ERR(0, 584, __pyx_L1_error)
        #else
        __pyx_t_4 = __Pyx_PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 584, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      } else {
        {
          Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_1);
          #if !CYTHON_ASSUME_SAFE_MACROS
          if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 584, __pyx_L1_error)
          #endif
          if (__pyx_t_2 >= __pyx_temp) break;
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_4); __pyx_t_2++; if (unlikely((0 < 0))) __PYX_ERR(0, 584, __pyx_L1_error)
        #else
        __pyx_t_4 = __Pyx_PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 584, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      }
    } else {
      __pyx_t_4 = __pyx_t_3(__pyx_t_1);
      if (unlikely(!__pyx_t_4)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 584, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_4);
    }
    __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_key);
    __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_key, __pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_4);
    __pyx_t_4 = 0;
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_key, __pyx_n_s_is_partial); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 584, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely((__pyx_t_6 < 0))) __PYX_ERR(0, 584, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (!__pyx_t_6) {
    } else {
      __pyx_t_5 = __pyx_t_6;
      goto __pyx_L7_bool_binop_done;
    }
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_key, __pyx_n_s_kind); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 584, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_6 = (__pyx_t_4 == Py_None);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_5 = __pyx_t_6;
    __pyx_L7_bool_binop_done:;
    if (__pyx_t_5) {
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(Py_True);
      __pyx_r = Py_True;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      goto __pyx_L0;
    }
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(Py_False);
    __pyx_r = Py_False;
    goto __pyx_L0;
  }
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_Generator_Replace_StopIteration(0);
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "viur/datastore/transport.pyx":570
 *     return res
 * 
 * def Get(keys: Union[Key, List[Key]]) -> Union[None, Entity, List[Entity]]:             # <<<<<<<<<<<<<<
 *     """
 *         Fetches the entities determined by keys from the datastore. Returns or inserts None if a key is not found.
 */

static PyObject *__pyx_pf_4viur_9datastore_9transport_8Get(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_keys) {
  simdjson::dom::parser __pyx_v_parser;
  Py_ssize_t __pyx_v_pysize;
  char *__pyx_v_data_ptr;
  simdjson::dom::element __pyx_v_element;
  int __pyx_v_isMulti;
  PyObject *__pyx_v_accessLog = NULL;
  PyObject *__pyx_v_currentTxn = NULL;
  PyObject *__pyx_v_readOptions = NULL;
  PyObject *__pyx_v_res = NULL;
  PyObject *__pyx_v_res_from_cache = NULL;
  PyObject *__pyx_v_res_from_db = NULL;
  PyObject *__pyx_v_untouched_keys = NULL;
  PyObject *__pyx_v_keys_for_request = NULL;
  PyObject *__pyx_v_missing_keys = NULL;
  PyObject *__pyx_v_requested_keys = NULL;
  PyObject *__pyx_v_postData = NULL;
  PyObject *__pyx_v_resp = NULL;
  PyObject *__pyx_v_deferred_keys = NULL;
  PyObject *__pyx_v_missing_keys_paths = NULL;
  PyObject *__pyx_v_i = NULL;
  PyObject *__pyx_v_deferred_key = NULL;
  PyObject *__pyx_gb_4viur_9datastore_9transport_3Get_2generator = 0;
  PyObject *__pyx_8genexpr5__pyx_v_key = NULL;
  PyObject *__pyx_8genexpr5__pyx_v_value = NULL;
  PyObject *__pyx_8genexpr6__pyx_v_key = NULL;
  PyObject *__pyx_8genexpr7__pyx_v_x = NULL;
  PyObject *__pyx_8genexpr8__pyx_v_key = NULL;
  PyObject *__pyx_8genexpr9__pyx_v_key = NULL;
  PyObject *__pyx_8genexpr9__pyx_v_value = NULL;
  PyObject *__pyx_9genexpr10__pyx_v_key = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  Py_ssize_t __pyx_t_7;
  Py_ssize_t __pyx_t_8;
  int __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  PyObject *(*__pyx_t_11)(PyObject *);
  int __pyx_t_12;
  PyObject *__pyx_t_13 = NULL;
  simdjson::dom::element __pyx_t_14;
  struct __pyx_opt_args_4viur_9datastore_9transport_toEntityStructure __pyx_t_15;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("Get", 0);
  __Pyx_INCREF(__pyx_v_keys);

  /* "viur/datastore/transport.pyx":576
 *         :return: The entity or None for the given key, a list of Entities/None if a list has been supplied
 *     """
 *     cdef simdjsonParser parser = simdjsonParser()             # <<<<<<<<<<<<<<
 *     cdef Py_ssize_t pysize
 *     cdef char * data_ptr
 */
  __pyx_v_parser = simdjson::dom::parser();

  /* "viur/datastore/transport.pyx":580
 *     cdef char * data_ptr
 *     cdef simdjsonElement element
 *     isMulti = True             # <<<<<<<<<<<<<<
 *     if isinstance(keys, Key):
 *         keys = [keys]
 */
  __pyx_v_isMulti = 1;

  /* "viur/datastore/transport.pyx":581
 *     cdef simdjsonElement element
 *     isMulti = True
 *     if isinstance(keys, Key):             # <<<<<<<<<<<<<<
 *         keys = [keys]
 *         isMulti = False
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_Key); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 581, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyObject_IsInstance(__pyx_v_keys, __pyx_t_1); if (unlikely(__pyx_t_2 == ((int)-1))) __PYX_ERR(0, 581, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "viur/datastore/transport.pyx":582
 *     isMulti = True
 *     if isinstance(keys, Key):
 *         keys = [keys]             # <<<<<<<<<<<<<<
 *         isMulti = False
 *     if any(key.is_partial or key.kind is None for key in keys):
 */
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 582, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_v_keys);
    __Pyx_GIVEREF(__pyx_v_keys);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 0, __pyx_v_keys)) __PYX_ERR(0, 582, __pyx_L1_error);
    __Pyx_DECREF_SET(__pyx_v_keys, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "viur/datastore/transport.pyx":583
 *     if isinstance(keys, Key):
 *         keys = [keys]
 *         isMulti = False             # <<<<<<<<<<<<<<
 *     if any(key.is_partial or key.kind is None for key in keys):
 *         raise InvalidArgumentError("Keys must not be partial or kind-less")
 */
    __pyx_v_isMulti = 0;

    /* "viur/datastore/transport.pyx":581
 *     cdef simdjsonElement element
 *     isMulti = True
 *     if isinstance(keys, Key):             # <<<<<<<<<<<<<<
 *         keys = [keys]
 *         isMulti = False
 */
  }

  /* "viur/datastore/transport.pyx":584
 *         keys = [keys]
 *         isMulti = False
 *     if any(key.is_partial or key.kind is None for key in keys):             # <<<<<<<<<<<<<<
 *         raise InvalidArgumentError("Keys must not be partial or kind-less")
 *     accessLog = currentDbAccessLog.get()
 */
  __pyx_t_1 = __pyx_pf_4viur_9datastore_9transport_3Get_genexpr(NULL, __pyx_v_keys); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 584, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_Generator_Next(__pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 584, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(0, 584, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(__pyx_t_2)) {

    /* "viur/datastore/transport.pyx":585
 *         isMulti = False
 *     if any(key.is_partial or key.kind is None for key in keys):
 *         raise InvalidArgumentError("Keys must not be partial or kind-less")             # <<<<<<<<<<<<<<
 *     accessLog = currentDbAccessLog.get()
 *     if isinstance(accessLog, set):
 */
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_InvalidArgumentError); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 585, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_4 = NULL;
    __pyx_t_5 = 0;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
        __pyx_t_5 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_kp_u_Keys_must_not_be_partial_or_kind};
      __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_1, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 585, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    }
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(0, 585, __pyx_L1_error)

    /* "viur/datastore/transport.pyx":584
 *         keys = [keys]
 *         isMulti = False
 *     if any(key.is_partial or key.kind is None for key in keys):             # <<<<<<<<<<<<<<
 *         raise InvalidArgumentError("Keys must not be partial or kind-less")
 *     accessLog = currentDbAccessLog.get()
 */
  }

  /* "viur/datastore/transport.pyx":586
 *     if any(key.is_partial or key.kind is None for key in keys):
 *         raise InvalidArgumentError("Keys must not be partial or kind-less")
 *     accessLog = currentDbAccessLog.get()             # <<<<<<<<<<<<<<
 *     if isinstance(accessLog, set):
 *         accessLog.update(set(keys))
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_currentDbAccessLog); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 586, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_get); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 586, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_1)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_1, NULL};
    __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_5, 0+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 586, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  }
  __pyx_v_accessLog = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "viur/datastore/transport.pyx":587
 *         raise InvalidArgumentError("Keys must not be partial or kind-less")
 *     accessLog = currentDbAccessLog.get()
 *     if isinstance(accessLog, set):             # <<<<<<<<<<<<<<
 *         accessLog.update(set(keys))
 * 
 */
  __pyx_t_2 = PySet_Check(__pyx_v_accessLog); 
  if (__pyx_t_2) {

    /* "viur/datastore/transport.pyx":588
 *     accessLog = currentDbAccessLog.get()
 *     if isinstance(accessLog, set):
 *         accessLog.update(set(keys))             # <<<<<<<<<<<<<<
 * 
 *     currentTxn = currentTransaction.get()
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_accessLog, __pyx_n_s_update); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 588, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_1 = PySet_New(__pyx_v_keys); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 588, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_6 = NULL;
    __pyx_t_5 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
        __pyx_t_5 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_6, __pyx_t_1};
      __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 588, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "viur/datastore/transport.pyx":587
 *         raise InvalidArgumentError("Keys must not be partial or kind-less")
 *     accessLog = currentDbAccessLog.get()
 *     if isinstance(accessLog, set):             # <<<<<<<<<<<<<<
 *         accessLog.update(set(keys))
 * 
 */
  }

  /* "viur/datastore/transport.pyx":590
 *         accessLog.update(set(keys))
 * 
 *     currentTxn = currentTransaction.get()             # <<<<<<<<<<<<<<
 *     if currentTxn:
 *         readOptions = {"transaction": currentTxn["key"]}
 */
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_currentTransaction); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 590, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_get); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 590, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_1);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_1, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, NULL};
    __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_1, __pyx_callargs+1-__pyx_t_5, 0+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 590, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }
  __pyx_v_currentTxn = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "viur/datastore/transport.pyx":591
 * 
 *     currentTxn = currentTransaction.get()
 *     if currentTxn:             # <<<<<<<<<<<<<<
 *         readOptions = {"transaction": currentTxn["key"]}
 *     else:
 */
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_v_currentTxn); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(0, 591, __pyx_L1_error)
  if (__pyx_t_2) {

    /* "viur/datastore/transport.pyx":592
 *     currentTxn = currentTransaction.get()
 *     if currentTxn:
 *         readOptions = {"transaction": currentTxn["key"]}             # <<<<<<<<<<<<<<
 *     else:
 *         readOptions = {"readConsistency": "STRONG"}
 */
    __pyx_t_3 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 592, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyObject_Dict_GetItem(__pyx_v_currentTxn, __pyx_n_u_key); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 592, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (PyDict_SetItem(__pyx_t_3, __pyx_n_u_transaction, __pyx_t_1) < 0) __PYX_ERR(0, 592, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_v_readOptions = ((PyObject*)__pyx_t_3);
    __pyx_t_3 = 0;

    /* "viur/datastore/transport.pyx":591
 * 
 *     currentTxn = currentTransaction.get()
 *     if currentTxn:             # <<<<<<<<<<<<<<
 *         readOptions = {"transaction": currentTxn["key"]}
 *     else:
 */
    goto __pyx_L6;
  }

  /* "viur/datastore/transport.pyx":594
 *         readOptions = {"transaction": currentTxn["key"]}
 *     else:
 *         readOptions = {"readConsistency": "STRONG"}             # <<<<<<<<<<<<<<
 *     res = {}
 *     res_from_cache = {}
 */
  /*else*/ {
    __pyx_t_3 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 594, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (PyDict_SetItem(__pyx_t_3, __pyx_n_u_readConsistency, __pyx_n_u_STRONG) < 0) __PYX_ERR(0, 594, __pyx_L1_error)
    __pyx_v_readOptions = ((PyObject*)__pyx_t_3);
    __pyx_t_3 = 0;
  }
  __pyx_L6:;

  /* "viur/datastore/transport.pyx":595
 *     else:
 *         readOptions = {"readConsistency": "STRONG"}
 *     res = {}             # <<<<<<<<<<<<<<
 *     res_from_cache = {}
 *     res_from_db = {}
 */
  __pyx_t_3 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 595, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_v_res = ((PyObject*)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "viur/datastore/transport.pyx":596
 *         readOptions = {"readConsistency": "STRONG"}
 *     res = {}
 *     res_from_cache = {}             # <<<<<<<<<<<<<<
 *     res_from_db = {}
 *     untouched_keys = copy(keys)
 */
  __pyx_t_3 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 596, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_v_res_from_cache = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "viur/datastore/transport.pyx":597
 *     res = {}
 *     res_from_cache = {}
 *     res_from_db = {}             # <<<<<<<<<<<<<<
 *     untouched_keys = copy(keys)
 *     while keys:
 */
  __pyx_t_3 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 597, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_v_res_from_db = ((PyObject*)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "viur/datastore/transport.pyx":598
 *     res_from_cache = {}
 *     res_from_db = {}
 *     untouched_keys = copy(keys)             # <<<<<<<<<<<<<<
 *     while keys:
 *         keys_for_request = keys[:300]
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_copy); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 598, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_1);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_1, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v_keys};
    __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_1, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 598, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }
  __pyx_v_untouched_keys = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "viur/datastore/transport.pyx":599
 *     res_from_db = {}
 *     untouched_keys = copy(keys)
 *     while keys:             # <<<<<<<<<<<<<<
 *         keys_for_request = keys[:300]
 * 
 */
  while (1) {
    __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_v_keys); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(0, 599, __pyx_L1_error)
    if (!__pyx_t_2) break;

    /* "viur/datastore/transport.pyx":600
 *     untouched_keys = copy(keys)
 *     while keys:
 *         keys_for_request = keys[:300]             # <<<<<<<<<<<<<<
 * 
 *         if conf["memcache_client"] is not None:
 */
    __pyx_t_3 = __Pyx_PyObject_GetSlice(__pyx_v_keys, 0, 0x12C, NULL, NULL, &__pyx_slice__13, 0, 1, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 600, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_XDECREF_SET(__pyx_v_keys_for_request, __pyx_t_3);
    __pyx_t_3 = 0;

    /* "viur/datastore/transport.pyx":602
 *         keys_for_request = keys[:300]
 * 
 *         if conf["memcache_client"] is not None:             # <<<<<<<<<<<<<<
 *             res_from_cache = cache.get(keys_for_request)
 *             # Convert the keys back to "class" representation
 */
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_conf); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 602, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyObject_Dict_GetItem(__pyx_t_3, __pyx_n_u_memcache_client); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 602, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_2 = (__pyx_t_1 != Py_None);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (__pyx_t_2) {

      /* "viur/datastore/transport.pyx":603
 * 
 *         if conf["memcache_client"] is not None:
 *             res_from_cache = cache.get(keys_for_request)             # <<<<<<<<<<<<<<
 *             # Convert the keys back to "class" representation
 *             res_from_cache = {Key.from_legacy_urlsafe(key): value
 */
      __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_cache); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 603, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_get); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 603, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_3 = NULL;
      __pyx_t_5 = 0;
      #if CYTHON_UNPACK_METHODS
      if (unlikely(PyMethod_Check(__pyx_t_4))) {
        __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
        if (likely(__pyx_t_3)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
          __Pyx_INCREF(__pyx_t_3);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_4, function);
          __pyx_t_5 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_v_keys_for_request};
        __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
        __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
        if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 603, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      }
      __Pyx_DECREF_SET(__pyx_v_res_from_cache, __pyx_t_1);
      __pyx_t_1 = 0;

      /* "viur/datastore/transport.pyx":605
 *             res_from_cache = cache.get(keys_for_request)
 *             # Convert the keys back to "class" representation
 *             res_from_cache = {Key.from_legacy_urlsafe(key): value             # <<<<<<<<<<<<<<
 *                               for key, value in res_from_cache.items()}
 * 
 */
      { /* enter inner scope */
        __pyx_t_1 = PyDict_New(); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 605, __pyx_L12_error)
        __Pyx_GOTREF(__pyx_t_1);

        /* "viur/datastore/transport.pyx":606
 *             # Convert the keys back to "class" representation
 *             res_from_cache = {Key.from_legacy_urlsafe(key): value
 *                               for key, value in res_from_cache.items()}             # <<<<<<<<<<<<<<
 * 
 *         missing_keys = [key for key in keys_for_request if key not in res_from_cache.keys()]
 */
        __pyx_t_7 = 0;
        if (unlikely(__pyx_v_res_from_cache == Py_None)) {
          PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "items");
          __PYX_ERR(0, 606, __pyx_L12_error)
        }
        __pyx_t_3 = __Pyx_dict_iterator(__pyx_v_res_from_cache, 0, __pyx_n_s_items, (&__pyx_t_8), (&__pyx_t_5)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 606, __pyx_L12_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_XDECREF(__pyx_t_4);
        __pyx_t_4 = __pyx_t_3;
        __pyx_t_3 = 0;
        while (1) {
          __pyx_t_9 = __Pyx_dict_iter_next(__pyx_t_4, __pyx_t_8, &__pyx_t_7, &__pyx_t_3, &__pyx_t_6, NULL, __pyx_t_5);
          if (unlikely(__pyx_t_9 == 0)) break;
          if (unlikely(__pyx_t_9 == -1)) __PYX_ERR(0, 606, __pyx_L12_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_GOTREF(__pyx_t_6);
          __Pyx_XDECREF_SET(__pyx_8genexpr5__pyx_v_key, __pyx_t_3);
          __pyx_t_3 = 0;
          __Pyx_XDECREF_SET(__pyx_8genexpr5__pyx_v_value, __pyx_t_6);
          __pyx_t_6 = 0;

          /* "viur/datastore/transport.pyx":605
 *             res_from_cache = cache.get(keys_for_request)
 *             # Convert the keys back to "class" representation
 *             res_from_cache = {Key.from_legacy_urlsafe(key): value             # <<<<<<<<<<<<<<
 *                               for key, value in res_from_cache.items()}
 * 
 */
          __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_Key); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 605, __pyx_L12_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_from_legacy_urlsafe); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 605, __pyx_L12_error)
          __Pyx_GOTREF(__pyx_t_10);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __pyx_t_3 = NULL;
          __pyx_t_9 = 0;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_10))) {
            __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_10);
            if (likely(__pyx_t_3)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_10);
              __Pyx_INCREF(__pyx_t_3);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_10, function);
              __pyx_t_9 = 1;
            }
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_8genexpr5__pyx_v_key};
            __pyx_t_6 = __Pyx_PyObject_FastCall(__pyx_t_10, __pyx_callargs+1-__pyx_t_9, 1+__pyx_t_9);
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 605, __pyx_L12_error)
            __Pyx_GOTREF(__pyx_t_6);
            __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
          }
          if (unlikely(PyDict_SetItem(__pyx_t_1, (PyObject*)__pyx_t_6, (PyObject*)__pyx_8genexpr5__pyx_v_value))) __PYX_ERR(0, 605, __pyx_L12_error)
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        }
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_XDECREF(__pyx_8genexpr5__pyx_v_key); __pyx_8genexpr5__pyx_v_key = 0;
        __Pyx_XDECREF(__pyx_8genexpr5__pyx_v_value); __pyx_8genexpr5__pyx_v_value = 0;
        goto __pyx_L15_exit_scope;
        __pyx_L12_error:;
        __Pyx_XDECREF(__pyx_8genexpr5__pyx_v_key); __pyx_8genexpr5__pyx_v_key = 0;
        __Pyx_XDECREF(__pyx_8genexpr5__pyx_v_value); __pyx_8genexpr5__pyx_v_value = 0;
        goto __pyx_L1_error;
        __pyx_L15_exit_scope:;
      } /* exit inner scope */
      __Pyx_DECREF_SET(__pyx_v_res_from_cache, __pyx_t_1);
      __pyx_t_1 = 0;

      /* "viur/datastore/transport.pyx":602
 *         keys_for_request = keys[:300]
 * 
 *         if conf["memcache_client"] is not None:             # <<<<<<<<<<<<<<
 *             res_from_cache = cache.get(keys_for_request)
 *             # Convert the keys back to "class" representation
 */
    }

    /* "viur/datastore/transport.pyx":608
 *                               for key, value in res_from_cache.items()}
 * 
 *         missing_keys = [key for key in keys_for_request if key not in res_from_cache.keys()]             # <<<<<<<<<<<<<<
 *         if not missing_keys:
 *             # We had all keys in the memcache we can fetch the next batch now.
 */
    { /* enter inner scope */
      __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 608, __pyx_L18_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (likely(PyList_CheckExact(__pyx_v_keys_for_request)) || PyTuple_CheckExact(__pyx_v_keys_for_request)) {
        __pyx_t_4 = __pyx_v_keys_for_request; __Pyx_INCREF(__pyx_t_4);
        __pyx_t_8 = 0;
        __pyx_t_11 = NULL;
      } else {
        __pyx_t_8 = -1; __pyx_t_4 = PyObject_GetIter(__pyx_v_keys_for_request); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 608, __pyx_L18_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_11 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_4); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 608, __pyx_L18_error)
      }
      for (;;) {
        if (likely(!__pyx_t_11)) {
          if (likely(PyList_CheckExact(__pyx_t_4))) {
            {
              Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_4);
              #if !CYTHON_ASSUME_SAFE_MACROS
              if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 608, __pyx_L18_error)
              #endif
              if (__pyx_t_8 >= __pyx_temp) break;
            }
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_6 = PyList_GET_ITEM(__pyx_t_4, __pyx_t_8); __Pyx_INCREF(__pyx_t_6); __pyx_t_8++; if (unlikely((0 < 0))) __PYX_ERR(0, 608, __pyx_L18_error)
            #else
            __pyx_t_6 = __Pyx_PySequence_ITEM(__pyx_t_4, __pyx_t_8); __pyx_t_8++; if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 608, __pyx_L18_error)
            __Pyx_GOTREF(__pyx_t_6);
            #endif
          } else {
            {
              Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_4);
              #if !CYTHON_ASSUME_SAFE_MACROS
              if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 608, __pyx_L18_error)
              #endif
              if (__pyx_t_8 >= __pyx_temp) break;
            }
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_6 = PyTuple_GET_ITEM(__pyx_t_4, __pyx_t_8); __Pyx_INCREF(__pyx_t_6); __pyx_t_8++; if (unlikely((0 < 0))) __PYX_ERR(0, 608, __pyx_L18_error)
            #else
            __pyx_t_6 = __Pyx_PySequence_ITEM(__pyx_t_4, __pyx_t_8); __pyx_t_8++; if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 608, __pyx_L18_error)
            __Pyx_GOTREF(__pyx_t_6);
            #endif
          }
        } else {
          __pyx_t_6 = __pyx_t_11(__pyx_t_4);
          if (unlikely(!__pyx_t_6)) {
            PyObject* exc_type = PyErr_Occurred();
            if (exc_type) {
              if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
              else __PYX_ERR(0, 608, __pyx_L18_error)
            }
            break;
          }
          __Pyx_GOTREF(__pyx_t_6);
        }
        __Pyx_XDECREF_SET(__pyx_8genexpr6__pyx_v_key, __pyx_t_6);
        __pyx_t_6 = 0;
        __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v_res_from_cache, __pyx_n_s_keys); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 608, __pyx_L18_error)
        __Pyx_GOTREF(__pyx_t_10);
        __pyx_t_3 = NULL;
        __pyx_t_5 = 0;
        #if CYTHON_UNPACK_METHODS
        if (likely(PyMethod_Check(__pyx_t_10))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_10);
          if (likely(__pyx_t_3)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_10);
            __Pyx_INCREF(__pyx_t_3);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_10, function);
            __pyx_t_5 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_3, NULL};
          __pyx_t_6 = __Pyx_PyObject_FastCall(__pyx_t_10, __pyx_callargs+1-__pyx_t_5, 0+__pyx_t_5);
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 608, __pyx_L18_error)
          __Pyx_GOTREF(__pyx_t_6);
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
        }
        __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_8genexpr6__pyx_v_key, __pyx_t_6, Py_NE)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(0, 608, __pyx_L18_error)
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        if (__pyx_t_2) {
          if (unlikely(__Pyx_ListComp_Append(__pyx_t_1, (PyObject*)__pyx_8genexpr6__pyx_v_key))) __PYX_ERR(0, 608, __pyx_L18_error)
        }
      }
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_8genexpr6__pyx_v_key); __pyx_8genexpr6__pyx_v_key = 0;
      goto __pyx_L23_exit_scope;
      __pyx_L18_error:;
      __Pyx_XDECREF(__pyx_8genexpr6__pyx_v_key); __pyx_8genexpr6__pyx_v_key = 0;
      goto __pyx_L1_error;
      __pyx_L23_exit_scope:;
    } /* exit inner scope */
    __Pyx_XDECREF_SET(__pyx_v_missing_keys, ((PyObject*)__pyx_t_1));
    __pyx_t_1 = 0;

    /* "viur/datastore/transport.pyx":609
 * 
 *         missing_keys = [key for key in keys_for_request if key not in res_from_cache.keys()]
 *         if not missing_keys:             # <<<<<<<<<<<<<<
 *             # We had all keys in the memcache we can fetch the next batch now.
 *             keys = keys[300:]
 */
    __pyx_t_2 = (PyList_GET_SIZE(__pyx_v_missing_keys) != 0);
    __pyx_t_12 = (!__pyx_t_2);
    if (__pyx_t_12) {

      /* "viur/datastore/transport.pyx":611
 *         if not missing_keys:
 *             # We had all keys in the memcache we can fetch the next batch now.
 *             keys = keys[300:]             # <<<<<<<<<<<<<<
 *             continue
 *         requested_keys = [
 */
      __pyx_t_1 = __Pyx_PyObject_GetSlice(__pyx_v_keys, 0x12C, 0, NULL, NULL, &__pyx_slice__14, 1, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 611, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF_SET(__pyx_v_keys, __pyx_t_1);
      __pyx_t_1 = 0;

      /* "viur/datastore/transport.pyx":612
 *             # We had all keys in the memcache we can fetch the next batch now.
 *             keys = keys[300:]
 *             continue             # <<<<<<<<<<<<<<
 *         requested_keys = [
 *             {
 */
      goto __pyx_L7_continue;

      /* "viur/datastore/transport.pyx":609
 * 
 *         missing_keys = [key for key in keys_for_request if key not in res_from_cache.keys()]
 *         if not missing_keys:             # <<<<<<<<<<<<<<
 *             # We had all keys in the memcache we can fetch the next batch now.
 *             keys = keys[300:]
 */
    }

    /* "viur/datastore/transport.pyx":613
 *             keys = keys[300:]
 *             continue
 *         requested_keys = [             # <<<<<<<<<<<<<<
 *             {
 *                 "partitionId": {
 */
    { /* enter inner scope */
      __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 613, __pyx_L27_error)
      __Pyx_GOTREF(__pyx_t_1);

      /* "viur/datastore/transport.pyx":620
 *                 "path": keyToPath(x)
 *             }
 *             for x in missing_keys]             # <<<<<<<<<<<<<<
 *         postData = {
 *             "readOptions": readOptions,
 */
      __pyx_t_4 = __pyx_v_missing_keys; __Pyx_INCREF(__pyx_t_4);
      __pyx_t_8 = 0;
      for (;;) {
        {
          Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_4);
          #if !CYTHON_ASSUME_SAFE_MACROS
          if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 620, __pyx_L27_error)
          #endif
          if (__pyx_t_8 >= __pyx_temp) break;
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_6 = PyList_GET_ITEM(__pyx_t_4, __pyx_t_8); __Pyx_INCREF(__pyx_t_6); __pyx_t_8++; if (unlikely((0 < 0))) __PYX_ERR(0, 620, __pyx_L27_error)
        #else
        __pyx_t_6 = __Pyx_PySequence_ITEM(__pyx_t_4, __pyx_t_8); __pyx_t_8++; if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 620, __pyx_L27_error)
        __Pyx_GOTREF(__pyx_t_6);
        #endif
        __Pyx_XDECREF_SET(__pyx_8genexpr7__pyx_v_x, __pyx_t_6);
        __pyx_t_6 = 0;

        /* "viur/datastore/transport.pyx":615
 *         requested_keys = [
 *             {
 *                 "partitionId": {             # <<<<<<<<<<<<<<
 *                     "project_id": projectID,
 *                 },
 */
        __pyx_t_6 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 615, __pyx_L27_error)
        __Pyx_GOTREF(__pyx_t_6);

        /* "viur/datastore/transport.pyx":616
 *             {
 *                 "partitionId": {
 *                     "project_id": projectID,             # <<<<<<<<<<<<<<
 *                 },
 *                 "path": keyToPath(x)
 */
        __pyx_t_10 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 616, __pyx_L27_error)
        __Pyx_GOTREF(__pyx_t_10);
        __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_projectID); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 616, __pyx_L27_error)
        __Pyx_GOTREF(__pyx_t_3);
        if (PyDict_SetItem(__pyx_t_10, __pyx_n_u_project_id, __pyx_t_3) < 0) __PYX_ERR(0, 616, __pyx_L27_error)
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        if (PyDict_SetItem(__pyx_t_6, __pyx_n_u_partitionId, __pyx_t_10) < 0) __PYX_ERR(0, 615, __pyx_L27_error)
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

        /* "viur/datastore/transport.pyx":618
 *                     "project_id": projectID,
 *                 },
 *                 "path": keyToPath(x)             # <<<<<<<<<<<<<<
 *             }
 *             for x in missing_keys]
 */
        __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_keyToPath); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 618, __pyx_L27_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_13 = NULL;
        __pyx_t_5 = 0;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_13 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_13)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_13);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
            __pyx_t_5 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_13, __pyx_8genexpr7__pyx_v_x};
          __pyx_t_10 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
          __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
          if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 618, __pyx_L27_error)
          __Pyx_GOTREF(__pyx_t_10);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        }
        if (PyDict_SetItem(__pyx_t_6, __pyx_n_u_path, __pyx_t_10) < 0) __PYX_ERR(0, 615, __pyx_L27_error)
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
        if (unlikely(__Pyx_ListComp_Append(__pyx_t_1, (PyObject*)__pyx_t_6))) __PYX_ERR(0, 613, __pyx_L27_error)
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

        /* "viur/datastore/transport.pyx":620
 *                 "path": keyToPath(x)
 *             }
 *             for x in missing_keys]             # <<<<<<<<<<<<<<
 *         postData = {
 *             "readOptions": readOptions,
 */
      }
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_8genexpr7__pyx_v_x); __pyx_8genexpr7__pyx_v_x = 0;
      goto __pyx_L31_exit_scope;
      __pyx_L27_error:;
      __Pyx_XDECREF(__pyx_8genexpr7__pyx_v_x); __pyx_8genexpr7__pyx_v_x = 0;
      goto __pyx_L1_error;
      __pyx_L31_exit_scope:;
    } /* exit inner scope */
    __Pyx_XDECREF_SET(__pyx_v_requested_keys, ((PyObject*)__pyx_t_1));
    __pyx_t_1 = 0;

    /* "viur/datastore/transport.pyx":622
 *             for x in missing_keys]
 *         postData = {
 *             "readOptions": readOptions,             # <<<<<<<<<<<<<<
 *             "keys": requested_keys,
 *         }
 */
    __pyx_t_1 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 622, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (PyDict_SetItem(__pyx_t_1, __pyx_n_u_readOptions, __pyx_v_readOptions) < 0) __PYX_ERR(0, 622, __pyx_L1_error)

    /* "viur/datastore/transport.pyx":623
 *         postData = {
 *             "readOptions": readOptions,
 *             "keys": requested_keys,             # <<<<<<<<<<<<<<
 *         }
 *         resp = authenticated_request(
 */
    if (PyDict_SetItem(__pyx_t_1, __pyx_n_u_keys, __pyx_v_requested_keys) < 0) __PYX_ERR(0, 622, __pyx_L1_error)
    __Pyx_XDECREF_SET(__pyx_v_postData, ((PyObject*)__pyx_t_1));
    __pyx_t_1 = 0;

    /* "viur/datastore/transport.pyx":625
 *             "keys": requested_keys,
 *         }
 *         resp = authenticated_request(             # <<<<<<<<<<<<<<
 *             url="https://datastore.googleapis.com/v1/projects/%s:lookup" % projectID,
 *             data=json.dumps(postData).encode("UTF-8"),
 */
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_authenticated_request); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 625, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);

    /* "viur/datastore/transport.pyx":626
 *         }
 *         resp = authenticated_request(
 *             url="https://datastore.googleapis.com/v1/projects/%s:lookup" % projectID,             # <<<<<<<<<<<<<<
 *             data=json.dumps(postData).encode("UTF-8"),
 *         )
 */
    __pyx_t_4 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 626, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_projectID); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 626, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_10 = __Pyx_PyUnicode_FormatSafe(__pyx_kp_u_https_datastore_googleapis_com_v_2, __pyx_t_6); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 626, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_url, __pyx_t_10) < 0) __PYX_ERR(0, 626, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

    /* "viur/datastore/transport.pyx":627
 *         resp = authenticated_request(
 *             url="https://datastore.googleapis.com/v1/projects/%s:lookup" % projectID,
 *             data=json.dumps(postData).encode("UTF-8"),             # <<<<<<<<<<<<<<
 *         )
 * 
 */
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_json); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 627, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_13 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_dumps); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 627, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_13);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = NULL;
    __pyx_t_5 = 0;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_13))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_13);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_13);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_13, function);
        __pyx_t_5 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_v_postData};
      __pyx_t_6 = __Pyx_PyObject_FastCall(__pyx_t_13, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 627, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
    }
    __pyx_t_13 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_encode); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 627, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_13);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_6 = NULL;
    __pyx_t_5 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_13))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_13);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_13);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_13, function);
        __pyx_t_5 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_6, __pyx_kp_u_UTF_8};
      __pyx_t_10 = __Pyx_PyObject_FastCall(__pyx_t_13, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 627, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
    }
    if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_data, __pyx_t_10) < 0) __PYX_ERR(0, 626, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

    /* "viur/datastore/transport.pyx":625
 *             "keys": requested_keys,
 *         }
 *         resp = authenticated_request(             # <<<<<<<<<<<<<<
 *             url="https://datastore.googleapis.com/v1/projects/%s:lookup" % projectID,
 *             data=json.dumps(postData).encode("UTF-8"),
 */
    __pyx_t_10 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_empty_tuple, __pyx_t_4); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 625, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF_SET(__pyx_v_resp, __pyx_t_10);
    __pyx_t_10 = 0;

    /* "viur/datastore/transport.pyx":630
 *         )
 * 
 *         if is_viur_datastore_request_ok(resp):             # <<<<<<<<<<<<<<
 *             assert PyBytes_AsStringAndSize(resp.content, &data_ptr, &pysize) != -1
 *             element = parser.parse(data_ptr, pysize, 1)
 */
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_is_viur_datastore_request_ok); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 630, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_1 = NULL;
    __pyx_t_5 = 0;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_1)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_1);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
        __pyx_t_5 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_1, __pyx_v_resp};
      __pyx_t_10 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 630, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
    __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_10); if (unlikely((__pyx_t_12 < 0))) __PYX_ERR(0, 630, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    if (__pyx_t_12) {

      /* "viur/datastore/transport.pyx":631
 * 
 *         if is_viur_datastore_request_ok(resp):
 *             assert PyBytes_AsStringAndSize(resp.content, &data_ptr, &pysize) != -1             # <<<<<<<<<<<<<<
 *             element = parser.parse(data_ptr, pysize, 1)
 *             if (element.at_pointer("/found").error() == SUCCESS):
 */
      #ifndef CYTHON_WITHOUT_ASSERTIONS
      if (unlikely(__pyx_assertions_enabled())) {
        __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v_resp, __pyx_n_s_content); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 631, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_10);
        __pyx_t_5 = PyBytes_AsStringAndSize(__pyx_t_10, (&__pyx_v_data_ptr), (&__pyx_v_pysize)); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(0, 631, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
        __pyx_t_12 = (__pyx_t_5 != -1L);
        if (unlikely(!__pyx_t_12)) {
          __Pyx_Raise(__pyx_builtin_AssertionError, 0, 0, 0);
          __PYX_ERR(0, 631, __pyx_L1_error)
        }
      }
      #else
      if ((1)); else __PYX_ERR(0, 631, __pyx_L1_error)
      #endif

      /* "viur/datastore/transport.pyx":632
 *         if is_viur_datastore_request_ok(resp):
 *             assert PyBytes_AsStringAndSize(resp.content, &data_ptr, &pysize) != -1
 *             element = parser.parse(data_ptr, pysize, 1)             # <<<<<<<<<<<<<<
 *             if (element.at_pointer("/found").error() == SUCCESS):
 *                 res_from_db.update(toEntityStructure(element.at_key("found"), isInitial=True))
 */
      try {
        __pyx_t_14 = __pyx_v_parser.parse(__pyx_v_data_ptr, __pyx_v_pysize, 1);
      } catch(...) {
        __Pyx_CppExn2PyErr();
        __PYX_ERR(0, 632, __pyx_L1_error)
      }
      __pyx_v_element = __PYX_STD_MOVE_IF_SUPPORTED(__pyx_t_14);

      /* "viur/datastore/transport.pyx":633
 *             assert PyBytes_AsStringAndSize(resp.content, &data_ptr, &pysize) != -1
 *             element = parser.parse(data_ptr, pysize, 1)
 *             if (element.at_pointer("/found").error() == SUCCESS):             # <<<<<<<<<<<<<<
 *                 res_from_db.update(toEntityStructure(element.at_key("found"), isInitial=True))
 * 
 */
      __pyx_t_12 = (__pyx_v_element.at_pointer(((char const *)"/found")).error() == simdjson::error_code::SUCCESS);
      if (__pyx_t_12) {

        /* "viur/datastore/transport.pyx":634
 *             element = parser.parse(data_ptr, pysize, 1)
 *             if (element.at_pointer("/found").error() == SUCCESS):
 *                 res_from_db.update(toEntityStructure(element.at_key("found"), isInitial=True))             # <<<<<<<<<<<<<<
 * 
 *             if (element.at_pointer("/deferred").error() == SUCCESS):
 */
        try {
          __pyx_t_14 = __pyx_v_element.at_key(((char const *)"found"));
        } catch(...) {
          __Pyx_CppExn2PyErr();
          __PYX_ERR(0, 634, __pyx_L1_error)
        }
        __pyx_t_15.__pyx_n = 1;
        __pyx_t_15.isInitial = 1;
        __pyx_t_10 = __pyx_f_4viur_9datastore_9transport_toEntityStructure(__PYX_STD_MOVE_IF_SUPPORTED(__pyx_t_14), &__pyx_t_15); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 634, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_10);
        __pyx_t_4 = __Pyx_CallUnboundCMethod1(&__pyx_umethod_PyDict_Type_update, __pyx_v_res_from_db, __pyx_t_10); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 634, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

        /* "viur/datastore/transport.pyx":633
 *             assert PyBytes_AsStringAndSize(resp.content, &data_ptr, &pysize) != -1
 *             element = parser.parse(data_ptr, pysize, 1)
 *             if (element.at_pointer("/found").error() == SUCCESS):             # <<<<<<<<<<<<<<
 *                 res_from_db.update(toEntityStructure(element.at_key("found"), isInitial=True))
 * 
 */
      }

      /* "viur/datastore/transport.pyx":636
 *                 res_from_db.update(toEntityStructure(element.at_key("found"), isInitial=True))
 * 
 *             if (element.at_pointer("/deferred").error() == SUCCESS):             # <<<<<<<<<<<<<<
 *                 deferred_keys = toPythonStructure(element.at_key("deferred"))
 *                 missing_keys_paths = [keyToPath(key) for key in missing_keys]
 */
      __pyx_t_12 = (__pyx_v_element.at_pointer(((char const *)"/deferred")).error() == simdjson::error_code::SUCCESS);
      if (__pyx_t_12) {

        /* "viur/datastore/transport.pyx":637
 * 
 *             if (element.at_pointer("/deferred").error() == SUCCESS):
 *                 deferred_keys = toPythonStructure(element.at_key("deferred"))             # <<<<<<<<<<<<<<
 *                 missing_keys_paths = [keyToPath(key) for key in missing_keys]
 *                 keys = keys[300:]
 */
        try {
          __pyx_t_14 = __pyx_v_element.at_key(((char const *)"deferred"));
        } catch(...) {
          __Pyx_CppExn2PyErr();
          __PYX_ERR(0, 637, __pyx_L1_error)
        }
        __pyx_t_4 = __pyx_f_4viur_9datastore_9transport_toPythonStructure(__PYX_STD_MOVE_IF_SUPPORTED(__pyx_t_14)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 637, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_XDECREF_SET(__pyx_v_deferred_keys, __pyx_t_4);
        __pyx_t_4 = 0;

        /* "viur/datastore/transport.pyx":638
 *             if (element.at_pointer("/deferred").error() == SUCCESS):
 *                 deferred_keys = toPythonStructure(element.at_key("deferred"))
 *                 missing_keys_paths = [keyToPath(key) for key in missing_keys]             # <<<<<<<<<<<<<<
 *                 keys = keys[300:]
 *                 for i, deferred_key in enumerate(deferred_keys):
 */
        { /* enter inner scope */
          __pyx_t_4 = PyList_New(0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 638, __pyx_L37_error)
          __Pyx_GOTREF(__pyx_t_4);
          __pyx_t_10 = __pyx_v_missing_keys; __Pyx_INCREF(__pyx_t_10);
          __pyx_t_8 = 0;
          for (;;) {
            {
              Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_10);
              #if !CYTHON_ASSUME_SAFE_MACROS
              if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 638, __pyx_L37_error)
              #endif
              if (__pyx_t_8 >= __pyx_temp) break;
            }
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_1 = PyList_GET_ITEM(__pyx_t_10, __pyx_t_8); __Pyx_INCREF(__pyx_t_1); __pyx_t_8++; if (unlikely((0 < 0))) __PYX_ERR(0, 638, __pyx_L37_error)
            #else
            __pyx_t_1 = __Pyx_PySequence_ITEM(__pyx_t_10, __pyx_t_8); __pyx_t_8++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 638, __pyx_L37_error)
            __Pyx_GOTREF(__pyx_t_1);
            #endif
            __Pyx_XDECREF_SET(__pyx_8genexpr8__pyx_v_key, __pyx_t_1);
            __pyx_t_1 = 0;
            __Pyx_GetModuleGlobalName(__pyx_t_13, __pyx_n_s_keyToPath); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 638, __pyx_L37_error)
            __Pyx_GOTREF(__pyx_t_13);
            __pyx_t_6 = NULL;
            __pyx_t_5 = 0;
            #if CYTHON_UNPACK_METHODS
            if (unlikely(PyMethod_Check(__pyx_t_13))) {
              __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_13);
              if (likely(__pyx_t_6)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_13);
                __Pyx_INCREF(__pyx_t_6);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_13, function);
                __pyx_t_5 = 1;
              }
            }
            #endif
            {
              PyObject *__pyx_callargs[2] = {__pyx_t_6, __pyx_8genexpr8__pyx_v_key};
              __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_13, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
              __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
              if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 638, __pyx_L37_error)
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
            }
            if (unlikely(__Pyx_ListComp_Append(__pyx_t_4, (PyObject*)__pyx_t_1))) __PYX_ERR(0, 638, __pyx_L37_error)
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          }
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
          __Pyx_XDECREF(__pyx_8genexpr8__pyx_v_key); __pyx_8genexpr8__pyx_v_key = 0;
          goto __pyx_L41_exit_scope;
          __pyx_L37_error:;
          __Pyx_XDECREF(__pyx_8genexpr8__pyx_v_key); __pyx_8genexpr8__pyx_v_key = 0;
          goto __pyx_L1_error;
          __pyx_L41_exit_scope:;
        } /* exit inner scope */
        __Pyx_XDECREF_SET(__pyx_v_missing_keys_paths, ((PyObject*)__pyx_t_4));
        __pyx_t_4 = 0;

        /* "viur/datastore/transport.pyx":639
 *                 deferred_keys = toPythonStructure(element.at_key("deferred"))
 *                 missing_keys_paths = [keyToPath(key) for key in missing_keys]
 *                 keys = keys[300:]             # <<<<<<<<<<<<<<
 *                 for i, deferred_key in enumerate(deferred_keys):
 *                     if deferred_key in missing_keys_paths:
 */
        __pyx_t_4 = __Pyx_PyObject_GetSlice(__pyx_v_keys, 0x12C, 0, NULL, NULL, &__pyx_slice__14, 1, 0, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 639, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF_SET(__pyx_v_keys, __pyx_t_4);
        __pyx_t_4 = 0;

        /* "viur/datastore/transport.pyx":640
 *                 missing_keys_paths = [keyToPath(key) for key in missing_keys]
 *                 keys = keys[300:]
 *                 for i, deferred_key in enumerate(deferred_keys):             # <<<<<<<<<<<<<<
 *                     if deferred_key in missing_keys_paths:
 *                         keys.insert(0, missing_keys[i])
 */
        __Pyx_INCREF(__pyx_int_0);
        __pyx_t_4 = __pyx_int_0;
        if (likely(PyList_CheckExact(__pyx_v_deferred_keys)) || PyTuple_CheckExact(__pyx_v_deferred_keys)) {
          __pyx_t_10 = __pyx_v_deferred_keys; __Pyx_INCREF(__pyx_t_10);
          __pyx_t_8 = 0;
          __pyx_t_11 = NULL;
        } else {
          __pyx_t_8 = -1; __pyx_t_10 = PyObject_GetIter(__pyx_v_deferred_keys); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 640, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_10);
          __pyx_t_11 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_10); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 640, __pyx_L1_error)
        }
        for (;;) {
          if (likely(!__pyx_t_11)) {
            if (likely(PyList_CheckExact(__pyx_t_10))) {
              {
                Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_10);
                #if !CYTHON_ASSUME_SAFE_MACROS
                if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 640, __pyx_L1_error)
                #endif
                if (__pyx_t_8 >= __pyx_temp) break;
              }
              #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
              __pyx_t_1 = PyList_GET_ITEM(__pyx_t_10, __pyx_t_8); __Pyx_INCREF(__pyx_t_1); __pyx_t_8++; if (unlikely((0 < 0))) __PYX_ERR(0, 640, __pyx_L1_error)
              #else
              __pyx_t_1 = __Pyx_PySequence_ITEM(__pyx_t_10, __pyx_t_8); __pyx_t_8++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 640, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_1);
              #endif
            } else {
              {
                Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_10);
                #if !CYTHON_ASSUME_SAFE_MACROS
                if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 640, __pyx_L1_error)
                #endif
                if (__pyx_t_8 >= __pyx_temp) break;
              }
              #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
              __pyx_t_1 = PyTuple_GET_ITEM(__pyx_t_10, __pyx_t_8); __Pyx_INCREF(__pyx_t_1); __pyx_t_8++; if (unlikely((0 < 0))) __PYX_ERR(0, 640, __pyx_L1_error)
              #else
              __pyx_t_1 = __Pyx_PySequence_ITEM(__pyx_t_10, __pyx_t_8); __pyx_t_8++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 640, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_1);
              #endif
            }
          } else {
            __pyx_t_1 = __pyx_t_11(__pyx_t_10);
            if (unlikely(!__pyx_t_1)) {
              PyObject* exc_type = PyErr_Occurred();
              if (exc_type) {
                if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
                else __PYX_ERR(0, 640, __pyx_L1_error)
              }
              break;
            }
            __Pyx_GOTREF(__pyx_t_1);
          }
          __Pyx_XDECREF_SET(__pyx_v_deferred_key, __pyx_t_1);
          __pyx_t_1 = 0;
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_XDECREF_SET(__pyx_v_i, __pyx_t_4);
          __pyx_t_1 = __Pyx_PyInt_AddObjC(__pyx_t_4, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 640, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_4);
          __pyx_t_4 = __pyx_t_1;
          __pyx_t_1 = 0;

          /* "viur/datastore/transport.pyx":641
 *                 keys = keys[300:]
 *                 for i, deferred_key in enumerate(deferred_keys):
 *                     if deferred_key in missing_keys_paths:             # <<<<<<<<<<<<<<
 *                         keys.insert(0, missing_keys[i])
 * 
 */
          __pyx_t_12 = (__Pyx_PySequence_ContainsTF(__pyx_v_deferred_key, __pyx_v_missing_keys_paths, Py_EQ)); if (unlikely((__pyx_t_12 < 0))) __PYX_ERR(0, 641, __pyx_L1_error)
          if (__pyx_t_12) {

            /* "viur/datastore/transport.pyx":642
 *                 for i, deferred_key in enumerate(deferred_keys):
 *                     if deferred_key in missing_keys_paths:
 *                         keys.insert(0, missing_keys[i])             # <<<<<<<<<<<<<<
 * 
 *             else:
 */
            __pyx_t_13 = __Pyx_PyObject_GetAttrStr(__pyx_v_keys, __pyx_n_s_insert); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 642, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_13);
            __pyx_t_6 = __Pyx_PyObject_GetItem(__pyx_v_missing_keys, __pyx_v_i); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 642, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_6);
            __pyx_t_3 = NULL;
            __pyx_t_5 = 0;
            #if CYTHON_UNPACK_METHODS
            if (likely(PyMethod_Check(__pyx_t_13))) {
              __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_13);
              if (likely(__pyx_t_3)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_13);
                __Pyx_INCREF(__pyx_t_3);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_13, function);
                __pyx_t_5 = 1;
              }
            }
            #endif
            {
              PyObject *__pyx_callargs[3] = {__pyx_t_3, __pyx_int_0, __pyx_t_6};
              __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_13, __pyx_callargs+1-__pyx_t_5, 2+__pyx_t_5);
              __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
              __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
              if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 642, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
            }
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

            /* "viur/datastore/transport.pyx":641
 *                 keys = keys[300:]
 *                 for i, deferred_key in enumerate(deferred_keys):
 *                     if deferred_key in missing_keys_paths:             # <<<<<<<<<<<<<<
 *                         keys.insert(0, missing_keys[i])
 * 
 */
          }

          /* "viur/datastore/transport.pyx":640
 *                 missing_keys_paths = [keyToPath(key) for key in missing_keys]
 *                 keys = keys[300:]
 *                 for i, deferred_key in enumerate(deferred_keys):             # <<<<<<<<<<<<<<
 *                     if deferred_key in missing_keys_paths:
 *                         keys.insert(0, missing_keys[i])
 */
        }
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

        /* "viur/datastore/transport.pyx":636
 *                 res_from_db.update(toEntityStructure(element.at_key("found"), isInitial=True))
 * 
 *             if (element.at_pointer("/deferred").error() == SUCCESS):             # <<<<<<<<<<<<<<
 *                 deferred_keys = toPythonStructure(element.at_key("deferred"))
 *                 missing_keys_paths = [keyToPath(key) for key in missing_keys]
 */
        goto __pyx_L34;
      }

      /* "viur/datastore/transport.pyx":645
 * 
 *             else:
 *                 keys = keys[300:]             # <<<<<<<<<<<<<<
 *     res = res_from_db | res_from_cache
 *     if conf["memcache_client"] is not None:
 */
      /*else*/ {
        __pyx_t_4 = __Pyx_PyObject_GetSlice(__pyx_v_keys, 0x12C, 0, NULL, NULL, &__pyx_slice__14, 1, 0, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 645, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF_SET(__pyx_v_keys, __pyx_t_4);
        __pyx_t_4 = 0;
      }
      __pyx_L34:;

      /* "viur/datastore/transport.pyx":630
 *         )
 * 
 *         if is_viur_datastore_request_ok(resp):             # <<<<<<<<<<<<<<
 *             assert PyBytes_AsStringAndSize(resp.content, &data_ptr, &pysize) != -1
 *             element = parser.parse(data_ptr, pysize, 1)
 */
    }
    __pyx_L7_continue:;
  }

  /* "viur/datastore/transport.pyx":646
 *             else:
 *                 keys = keys[300:]
 *     res = res_from_db | res_from_cache             # <<<<<<<<<<<<<<
 *     if conf["memcache_client"] is not None:
 *         # Cache only the entities form db.
 */
  __pyx_t_4 = PyNumber_Or(__pyx_v_res_from_db, __pyx_v_res_from_cache); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 646, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF_SET(__pyx_v_res, ((PyObject*)__pyx_t_4));
  __pyx_t_4 = 0;

  /* "viur/datastore/transport.pyx":647
 *                 keys = keys[300:]
 *     res = res_from_db | res_from_cache
 *     if conf["memcache_client"] is not None:             # <<<<<<<<<<<<<<
 *         # Cache only the entities form db.
 *         cache.put({str(key): value for key, value in res_from_db.items()})
 */
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_conf); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 647, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_10 = __Pyx_PyObject_Dict_GetItem(__pyx_t_4, __pyx_n_u_memcache_client); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 647, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_12 = (__pyx_t_10 != Py_None);
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
  if (__pyx_t_12) {

    /* "viur/datastore/transport.pyx":649
 *     if conf["memcache_client"] is not None:
 *         # Cache only the entities form db.
 *         cache.put({str(key): value for key, value in res_from_db.items()})             # <<<<<<<<<<<<<<
 * 
 *     if not isMulti:
 */
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_cache); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 649, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_put); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 649, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    { /* enter inner scope */
      __pyx_t_4 = PyDict_New(); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 649, __pyx_L49_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_8 = 0;
      __pyx_t_6 = __Pyx_dict_iterator(__pyx_v_res_from_db, 1, __pyx_n_s_items, (&__pyx_t_7), (&__pyx_t_5)); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 649, __pyx_L49_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_XDECREF(__pyx_t_13);
      __pyx_t_13 = __pyx_t_6;
      __pyx_t_6 = 0;
      while (1) {
        __pyx_t_9 = __Pyx_dict_iter_next(__pyx_t_13, __pyx_t_7, &__pyx_t_8, &__pyx_t_6, &__pyx_t_3, NULL, __pyx_t_5);
        if (unlikely(__pyx_t_9 == 0)) break;
        if (unlikely(__pyx_t_9 == -1)) __PYX_ERR(0, 649, __pyx_L49_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_XDECREF_SET(__pyx_8genexpr9__pyx_v_key, __pyx_t_6);
        __pyx_t_6 = 0;
        __Pyx_XDECREF_SET(__pyx_8genexpr9__pyx_v_value, __pyx_t_3);
        __pyx_t_3 = 0;
        __pyx_t_3 = __Pyx_PyObject_Str(__pyx_8genexpr9__pyx_v_key); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 649, __pyx_L49_error)
        __Pyx_GOTREF(__pyx_t_3);
        if (unlikely(PyDict_SetItem(__pyx_t_4, (PyObject*)__pyx_t_3, (PyObject*)__pyx_8genexpr9__pyx_v_value))) __PYX_ERR(0, 649, __pyx_L49_error)
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      }
      __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
      __Pyx_XDECREF(__pyx_8genexpr9__pyx_v_key); __pyx_8genexpr9__pyx_v_key = 0;
      __Pyx_XDECREF(__pyx_8genexpr9__pyx_v_value); __pyx_8genexpr9__pyx_v_value = 0;
      goto __pyx_L52_exit_scope;
      __pyx_L49_error:;
      __Pyx_XDECREF(__pyx_8genexpr9__pyx_v_key); __pyx_8genexpr9__pyx_v_key = 0;
      __Pyx_XDECREF(__pyx_8genexpr9__pyx_v_value); __pyx_8genexpr9__pyx_v_value = 0;
      goto __pyx_L1_error;
      __pyx_L52_exit_scope:;
    } /* exit inner scope */
    __pyx_t_13 = NULL;
    __pyx_t_5 = 0;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_13 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_13)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_13);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
        __pyx_t_5 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_13, __pyx_t_4};
      __pyx_t_10 = __Pyx_PyObject_FastCall(__pyx_t_1, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
      __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 649, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    }
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

    /* "viur/datastore/transport.pyx":647
 *                 keys = keys[300:]
 *     res = res_from_db | res_from_cache
 *     if conf["memcache_client"] is not None:             # <<<<<<<<<<<<<<
 *         # Cache only the entities form db.
 *         cache.put({str(key): value for key, value in res_from_db.items()})
 */
  }

  /* "viur/datastore/transport.pyx":651
 *         cache.put({str(key): value for key, value in res_from_db.items()})
 * 
 *     if not isMulti:             # <<<<<<<<<<<<<<
 *         return res.get(untouched_keys[0])
 *     else:
 */
  __pyx_t_12 = (!__pyx_v_isMulti);
  if (__pyx_t_12) {

    /* "viur/datastore/transport.pyx":652
 * 
 *     if not isMulti:
 *         return res.get(untouched_keys[0])             # <<<<<<<<<<<<<<
 *     else:
 *         return [res.get(key) for key in untouched_keys]  # Sort by order of incoming keys
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_10 = __Pyx_GetItemInt(__pyx_v_untouched_keys, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 652, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_t_1 = __Pyx_PyDict_GetItemDefault(__pyx_v_res, __pyx_t_10, Py_None); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 652, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "viur/datastore/transport.pyx":651
 *         cache.put({str(key): value for key, value in res_from_db.items()})
 * 
 *     if not isMulti:             # <<<<<<<<<<<<<<
 *         return res.get(untouched_keys[0])
 *     else:
 */
  }

  /* "viur/datastore/transport.pyx":654
 *         return res.get(untouched_keys[0])
 *     else:
 *         return [res.get(key) for key in untouched_keys]  # Sort by order of incoming keys             # <<<<<<<<<<<<<<
 * 
 * def Delete(keys: Union[Key, List[Key], Entity, List[Entity]]) -> None:
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    { /* enter inner scope */
      __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 654, __pyx_L56_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (likely(PyList_CheckExact(__pyx_v_untouched_keys)) || PyTuple_CheckExact(__pyx_v_untouched_keys)) {
        __pyx_t_10 = __pyx_v_untouched_keys; __Pyx_INCREF(__pyx_t_10);
        __pyx_t_7 = 0;
        __pyx_t_11 = NULL;
      } else {
        __pyx_t_7 = -1; __pyx_t_10 = PyObject_GetIter(__pyx_v_untouched_keys); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 654, __pyx_L56_error)
        __Pyx_GOTREF(__pyx_t_10);
        __pyx_t_11 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_10); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 654, __pyx_L56_error)
      }
      for (;;) {
        if (likely(!__pyx_t_11)) {
          if (likely(PyList_CheckExact(__pyx_t_10))) {
            {
              Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_10);
              #if !CYTHON_ASSUME_SAFE_MACROS
              if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 654, __pyx_L56_error)
              #endif
              if (__pyx_t_7 >= __pyx_temp) break;
            }
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_4 = PyList_GET_ITEM(__pyx_t_10, __pyx_t_7); __Pyx_INCREF(__pyx_t_4); __pyx_t_7++; if (unlikely((0 < 0))) __PYX_ERR(0, 654, __pyx_L56_error)
            #else
            __pyx_t_4 = __Pyx_PySequence_ITEM(__pyx_t_10, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 654, __pyx_L56_error)
            __Pyx_GOTREF(__pyx_t_4);
            #endif
          } else {
            {
              Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_10);
              #if !CYTHON_ASSUME_SAFE_MACROS
              if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 654, __pyx_L56_error)
              #endif
              if (__pyx_t_7 >= __pyx_temp) break;
            }
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_4 = PyTuple_GET_ITEM(__pyx_t_10, __pyx_t_7); __Pyx_INCREF(__pyx_t_4); __pyx_t_7++; if (unlikely((0 < 0))) __PYX_ERR(0, 654, __pyx_L56_error)
            #else
            __pyx_t_4 = __Pyx_PySequence_ITEM(__pyx_t_10, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 654, __pyx_L56_error)
            __Pyx_GOTREF(__pyx_t_4);
            #endif
          }
        } else {
          __pyx_t_4 = __pyx_t_11(__pyx_t_10);
          if (unlikely(!__pyx_t_4)) {
            PyObject* exc_type = PyErr_Occurred();
            if (exc_type) {
              if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
              else __PYX_ERR(0, 654, __pyx_L56_error)
            }
            break;
          }
          __Pyx_GOTREF(__pyx_t_4);
        }
        __Pyx_XDECREF_SET(__pyx_9genexpr10__pyx_v_key, __pyx_t_4);
        __pyx_t_4 = 0;
        __pyx_t_4 = __Pyx_PyDict_GetItemDefault(__pyx_v_res, __pyx_9genexpr10__pyx_v_key, Py_None); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 654, __pyx_L56_error)
        __Pyx_GOTREF(__pyx_t_4);
        if (unlikely(__Pyx_ListComp_Append(__pyx_t_1, (PyObject*)__pyx_t_4))) __PYX_ERR(0, 654, __pyx_L56_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      }
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_XDECREF(__pyx_9genexpr10__pyx_v_key); __pyx_9genexpr10__pyx_v_key = 0;
      goto __pyx_L60_exit_scope;
      __pyx_L56_error:;
      __Pyx_XDECREF(__pyx_9genexpr10__pyx_v_key); __pyx_9genexpr10__pyx_v_key = 0;
      goto __pyx_L1_error;
      __pyx_L60_exit_scope:;
    } /* exit inner scope */
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;
  }

  /* "viur/datastore/transport.pyx":570
 *     return res
 * 
 * def Get(keys: Union[Key, List[Key]]) -> Union[None, Entity, List[Entity]]:             # <<<<<<<<<<<<<<
 *     """
 *         Fetches the entities determined by keys from the datastore. Returns or inserts None if a key is not found.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_AddTraceback("viur.datastore.transport.Get", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_accessLog);
  __Pyx_XDECREF(__pyx_v_currentTxn);
  __Pyx_XDECREF(__pyx_v_readOptions);
  __Pyx_XDECREF(__pyx_v_res);
  __Pyx_XDECREF(__pyx_v_res_from_cache);
  __Pyx_XDECREF(__pyx_v_res_from_db);
  __Pyx_XDECREF(__pyx_v_untouched_keys);
  __Pyx_XDECREF(__pyx_v_keys_for_request);
  __Pyx_XDECREF(__pyx_v_missing_keys);
  __Pyx_XDECREF(__pyx_v_requested_keys);
  __Pyx_XDECREF(__pyx_v_postData);
  __Pyx_XDECREF(__pyx_v_resp);
  __Pyx_XDECREF(__pyx_v_deferred_keys);
  __Pyx_XDECREF(__pyx_v_missing_keys_paths);
  __Pyx_XDECREF(__pyx_v_i);
  __Pyx_XDECREF(__pyx_v_deferred_key);
  __Pyx_XDECREF(__pyx_gb_4viur_9datastore_9transport_3Get_2generator);
  __Pyx_XDECREF(__pyx_8genexpr5__pyx_v_key);
  __Pyx_XDECREF(__pyx_8genexpr5__pyx_v_value);
  __Pyx_XDECREF(__pyx_8genexpr6__pyx_v_key);
  __Pyx_XDECREF(__pyx_8genexpr7__pyx_v_x);
  __Pyx_XDECREF(__pyx_8genexpr8__pyx_v_key);
  __Pyx_XDECREF(__pyx_8genexpr9__pyx_v_key);
  __Pyx_XDECREF(__pyx_8genexpr9__pyx_v_value);
  __Pyx_XDECREF(__pyx_9genexpr10__pyx_v_key);
  __Pyx_XDECREF(__pyx_v_keys);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "viur/datastore/transport.pyx":656
 *         return [res.get(key) for key in untouched_keys]  # Sort by order of incoming keys
 * 
 * def Delete(keys: Union[Key, List[Key], Entity, List[Entity]]) -> None:             # <<<<<<<<<<<<<<
 *     """
 *         Deletes the entities stored under the given key(s).
 */

/* Python wrapper */
static PyObject *__pyx_pw_4viur_9datastore_9transport_11Delete(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_4viur_9datastore_9transport_10Delete, "\n        Deletes the entities stored under the given key(s).\n        If a key is not found, it's silently ignored.\n        A maximum of 300 Keys can be deleted at once.\n\n        :param keys: A Key or a List of Keys\n    ");
static PyMethodDef __pyx_mdef_4viur_9datastore_9transport_11Delete = {"Delete", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_4viur_9datastore_9transport_11Delete, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_4viur_9datastore_9transport_10Delete};
static PyObject *__pyx_pw_4viur_9datastore_9transport_11Delete(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_keys = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("Delete (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_keys,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_keys)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 656, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "Delete") < 0)) __PYX_ERR(0, 656, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_keys = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("Delete", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 656, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("viur.datastore.transport.Delete", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4viur_9datastore_9transport_10Delete(__pyx_self, __pyx_v_keys);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4viur_9datastore_9transport_10Delete(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_keys) {
  simdjson::dom::parser __pyx_v_parser;
  Py_ssize_t __pyx_v_pysize;
  char *__pyx_v_data_ptr;
  simdjson::dom::element __pyx_v_element;
  simdjson::dom::array __pyx_v_arrayElem;
  PyObject *__pyx_v_accessLog = NULL;
  PyObject *__pyx_v_postData = NULL;
  PyObject *__pyx_v_currentTxn = NULL;
  PyObject *__pyx_v_resp = NULL;
  PyObject *__pyx_9genexpr11__pyx_v_x = NULL;
  PyObject *__pyx_9genexpr12__pyx_v_x = NULL;
  PyObject *__pyx_9genexpr13__pyx_v_key = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  Py_ssize_t __pyx_t_4;
  PyObject *(*__pyx_t_5)(PyObject *);
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  int __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  simdjson::dom::element __pyx_t_13;
  simdjson::dom::array __pyx_t_14;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("Delete", 0);
  __Pyx_INCREF(__pyx_v_keys);

  /* "viur/datastore/transport.pyx":664
 *         :param keys: A Key or a List of Keys
 *     """
 *     cdef simdjsonParser parser = simdjsonParser()             # <<<<<<<<<<<<<<
 *     cdef Py_ssize_t pysize
 *     cdef char * data_ptr
 */
  __pyx_v_parser = simdjson::dom::parser();

  /* "viur/datastore/transport.pyx":669
 *     cdef simdjsonElement element
 *     cdef simdjsonArray arrayElem
 *     if isinstance(keys, Key):             # <<<<<<<<<<<<<<
 *         keys = [keys]
 *     elif isinstance(keys, Entity):
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_Key); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 669, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyObject_IsInstance(__pyx_v_keys, __pyx_t_1); if (unlikely(__pyx_t_2 == ((int)-1))) __PYX_ERR(0, 669, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "viur/datastore/transport.pyx":670
 *     cdef simdjsonArray arrayElem
 *     if isinstance(keys, Key):
 *         keys = [keys]             # <<<<<<<<<<<<<<
 *     elif isinstance(keys, Entity):
 *         keys = [keys.key]
 */
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 670, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_v_keys);
    __Pyx_GIVEREF(__pyx_v_keys);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 0, __pyx_v_keys)) __PYX_ERR(0, 670, __pyx_L1_error);
    __Pyx_DECREF_SET(__pyx_v_keys, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "viur/datastore/transport.pyx":669
 *     cdef simdjsonElement element
 *     cdef simdjsonArray arrayElem
 *     if isinstance(keys, Key):             # <<<<<<<<<<<<<<
 *         keys = [keys]
 *     elif isinstance(keys, Entity):
 */
    goto __pyx_L3;
  }

  /* "viur/datastore/transport.pyx":671
 *     if isinstance(keys, Key):
 *         keys = [keys]
 *     elif isinstance(keys, Entity):             # <<<<<<<<<<<<<<
 *         keys = [keys.key]
 *     keys = [(x.key if isinstance(x, Entity) else x) for x in keys]
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_Entity); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 671, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyObject_IsInstance(__pyx_v_keys, __pyx_t_1); if (unlikely(__pyx_t_2 == ((int)-1))) __PYX_ERR(0, 671, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "viur/datastore/transport.pyx":672
 *         keys = [keys]
 *     elif isinstance(keys, Entity):
 *         keys = [keys.key]             # <<<<<<<<<<<<<<
 *     keys = [(x.key if isinstance(x, Entity) else x) for x in keys]
 *     accessLog = currentDbAccessLog.get()
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_keys, __pyx_n_s_key); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 672, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = PyList_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 672, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_3, 0, __pyx_t_1)) __PYX_ERR(0, 672, __pyx_L1_error);
    __pyx_t_1 = 0;
    __Pyx_DECREF_SET(__pyx_v_keys, __pyx_t_3);
    __pyx_t_3 = 0;

    /* "viur/datastore/transport.pyx":671
 *     if isinstance(keys, Key):
 *         keys = [keys]
 *     elif isinstance(keys, Entity):             # <<<<<<<<<<<<<<
 *         keys = [keys.key]
 *     keys = [(x.key if isinstance(x, Entity) else x) for x in keys]
 */
  }
  __pyx_L3:;

  /* "viur/datastore/transport.pyx":673
 *     elif isinstance(keys, Entity):
 *         keys = [keys.key]
 *     keys = [(x.key if isinstance(x, Entity) else x) for x in keys]             # <<<<<<<<<<<<<<
 *     accessLog = currentDbAccessLog.get()
 *     if isinstance(accessLog, set):
 */
  { /* enter inner scope */
    __pyx_t_3 = PyList_New(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 673, __pyx_L6_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (likely(PyList_CheckExact(__pyx_v_keys)) || PyTuple_CheckExact(__pyx_v_keys)) {
      __pyx_t_1 = __pyx_v_keys; __Pyx_INCREF(__pyx_t_1);
      __pyx_t_4 = 0;
      __pyx_t_5 = NULL;
    } else {
      __pyx_t_4 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_v_keys); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 673, __pyx_L6_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_5 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 673, __pyx_L6_error)
    }
    for (;;) {
      if (likely(!__pyx_t_5)) {
        if (likely(PyList_CheckExact(__pyx_t_1))) {
          {
            Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_1);
            #if !CYTHON_ASSUME_SAFE_MACROS
            if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 673, __pyx_L6_error)
            #endif
            if (__pyx_t_4 >= __pyx_temp) break;
          }
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_6 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_4); __Pyx_INCREF(__pyx_t_6); __pyx_t_4++; if (unlikely((0 < 0))) __PYX_ERR(0, 673, __pyx_L6_error)
          #else
          __pyx_t_6 = __Pyx_PySequence_ITEM(__pyx_t_1, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 673, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_6);
          #endif
        } else {
          {
            Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_1);
            #if !CYTHON_ASSUME_SAFE_MACROS
            if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 673, __pyx_L6_error)
            #endif
            if (__pyx_t_4 >= __pyx_temp) break;
          }
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_6 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_4); __Pyx_INCREF(__pyx_t_6); __pyx_t_4++; if (unlikely((0 < 0))) __PYX_ERR(0, 673, __pyx_L6_error)
          #else
          __pyx_t_6 = __Pyx_PySequence_ITEM(__pyx_t_1, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 673, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_6);
          #endif
        }
      } else {
        __pyx_t_6 = __pyx_t_5(__pyx_t_1);
        if (unlikely(!__pyx_t_6)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 673, __pyx_L6_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_6);
      }
      __Pyx_XDECREF_SET(__pyx_9genexpr11__pyx_v_x, __pyx_t_6);
      __pyx_t_6 = 0;
      __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_Entity); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 673, __pyx_L6_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_2 = PyObject_IsInstance(__pyx_9genexpr11__pyx_v_x, __pyx_t_7); if (unlikely(__pyx_t_2 == ((int)-1))) __PYX_ERR(0, 673, __pyx_L6_error)
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      if (__pyx_t_2) {
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_9genexpr11__pyx_v_x, __pyx_n_s_key); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 673, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_6 = __pyx_t_7;
        __pyx_t_7 = 0;
      } else {
        __Pyx_INCREF(__pyx_9genexpr11__pyx_v_x);
        __pyx_t_6 = __pyx_9genexpr11__pyx_v_x;
      }
      if (unlikely(__Pyx_ListComp_Append(__pyx_t_3, (PyObject*)__pyx_t_6))) __PYX_ERR(0, 673, __pyx_L6_error)
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_9genexpr11__pyx_v_x); __pyx_9genexpr11__pyx_v_x = 0;
    goto __pyx_L10_exit_scope;
    __pyx_L6_error:;
    __Pyx_XDECREF(__pyx_9genexpr11__pyx_v_x); __pyx_9genexpr11__pyx_v_x = 0;
    goto __pyx_L1_error;
    __pyx_L10_exit_scope:;
  } /* exit inner scope */
  __Pyx_DECREF_SET(__pyx_v_keys, __pyx_t_3);
  __pyx_t_3 = 0;

  /* "viur/datastore/transport.pyx":674
 *         keys = [keys.key]
 *     keys = [(x.key if isinstance(x, Entity) else x) for x in keys]
 *     accessLog = currentDbAccessLog.get()             # <<<<<<<<<<<<<<
 *     if isinstance(accessLog, set):
 *         accessLog.update(set(keys))
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_currentDbAccessLog); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 674, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_get); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 674, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = NULL;
  __pyx_t_8 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_6))) {
    __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_6);
    if (likely(__pyx_t_1)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_6, function);
      __pyx_t_8 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_1, NULL};
    __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_6, __pyx_callargs+1-__pyx_t_8, 0+__pyx_t_8);
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 674, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __pyx_v_accessLog = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "viur/datastore/transport.pyx":675
 *     keys = [(x.key if isinstance(x, Entity) else x) for x in keys]
 *     accessLog = currentDbAccessLog.get()
 *     if isinstance(accessLog, set):             # <<<<<<<<<<<<<<
 *         accessLog.update(set(keys))
 *     if not keys:  # We got an empty list (probably a query that returned no results), noting to do here
 */
  __pyx_t_2 = PySet_Check(__pyx_v_accessLog); 
  if (__pyx_t_2) {

    /* "viur/datastore/transport.pyx":676
 *     accessLog = currentDbAccessLog.get()
 *     if isinstance(accessLog, set):
 *         accessLog.update(set(keys))             # <<<<<<<<<<<<<<
 *     if not keys:  # We got an empty list (probably a query that returned no results), noting to do here
 *         return
 */
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_accessLog, __pyx_n_s_update); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 676, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_1 = PySet_New(__pyx_v_keys); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 676, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_7 = NULL;
    __pyx_t_8 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
        __pyx_t_8 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_t_1};
      __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_6, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 676, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "viur/datastore/transport.pyx":675
 *     keys = [(x.key if isinstance(x, Entity) else x) for x in keys]
 *     accessLog = currentDbAccessLog.get()
 *     if isinstance(accessLog, set):             # <<<<<<<<<<<<<<
 *         accessLog.update(set(keys))
 *     if not keys:  # We got an empty list (probably a query that returned no results), noting to do here
 */
  }

  /* "viur/datastore/transport.pyx":677
 *     if isinstance(accessLog, set):
 *         accessLog.update(set(keys))
 *     if not keys:  # We got an empty list (probably a query that returned no results), noting to do here             # <<<<<<<<<<<<<<
 *         return
 *     postData = {
 */
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_v_keys); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(0, 677, __pyx_L1_error)
  __pyx_t_9 = (!__pyx_t_2);
  if (__pyx_t_9) {

    /* "viur/datastore/transport.pyx":678
 *         accessLog.update(set(keys))
 *     if not keys:  # We got an empty list (probably a query that returned no results), noting to do here
 *         return             # <<<<<<<<<<<<<<
 *     postData = {
 *         "mode": "NON_TRANSACTIONAL",  #"TRANSACTIONAL", #
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "viur/datastore/transport.pyx":677
 *     if isinstance(accessLog, set):
 *         accessLog.update(set(keys))
 *     if not keys:  # We got an empty list (probably a query that returned no results), noting to do here             # <<<<<<<<<<<<<<
 *         return
 *     postData = {
 */
  }

  /* "viur/datastore/transport.pyx":680
 *         return
 *     postData = {
 *         "mode": "NON_TRANSACTIONAL",  #"TRANSACTIONAL", #             # <<<<<<<<<<<<<<
 *         "mutations": [
 *             {
 */
  __pyx_t_3 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 680, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_u_mode, __pyx_n_u_NON_TRANSACTIONAL) < 0) __PYX_ERR(0, 680, __pyx_L1_error)
  { /* enter inner scope */

    /* "viur/datastore/transport.pyx":681
 *     postData = {
 *         "mode": "NON_TRANSACTIONAL",  #"TRANSACTIONAL", #
 *         "mutations": [             # <<<<<<<<<<<<<<
 *             {
 *                 "delete": pythonPropToJson(x)["keyValue"]
 */
    __pyx_t_6 = PyList_New(0); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 681, __pyx_L15_error)
    __Pyx_GOTREF(__pyx_t_6);

    /* "viur/datastore/transport.pyx":685
 *                 "delete": pythonPropToJson(x)["keyValue"]
 *             }
 *             for x in keys             # <<<<<<<<<<<<<<
 *         ]
 *     }
 */
    if (likely(PyList_CheckExact(__pyx_v_keys)) || PyTuple_CheckExact(__pyx_v_keys)) {
      __pyx_t_1 = __pyx_v_keys; __Pyx_INCREF(__pyx_t_1);
      __pyx_t_4 = 0;
      __pyx_t_5 = NULL;
    } else {
      __pyx_t_4 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_v_keys); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 685, __pyx_L15_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_5 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 685, __pyx_L15_error)
    }
    for (;;) {
      if (likely(!__pyx_t_5)) {
        if (likely(PyList_CheckExact(__pyx_t_1))) {
          {
            Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_1);
            #if !CYTHON_ASSUME_SAFE_MACROS
            if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 685, __pyx_L15_error)
            #endif
            if (__pyx_t_4 >= __pyx_temp) break;
          }
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_7 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_4); __Pyx_INCREF(__pyx_t_7); __pyx_t_4++; if (unlikely((0 < 0))) __PYX_ERR(0, 685, __pyx_L15_error)
          #else
          __pyx_t_7 = __Pyx_PySequence_ITEM(__pyx_t_1, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 685, __pyx_L15_error)
          __Pyx_GOTREF(__pyx_t_7);
          #endif
        } else {
          {
            Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_1);
            #if !CYTHON_ASSUME_SAFE_MACROS
            if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 685, __pyx_L15_error)
            #endif
            if (__pyx_t_4 >= __pyx_temp) break;
          }
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_7 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_4); __Pyx_INCREF(__pyx_t_7); __pyx_t_4++; if (unlikely((0 < 0))) __PYX_ERR(0, 685, __pyx_L15_error)
          #else
          __pyx_t_7 = __Pyx_PySequence_ITEM(__pyx_t_1, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 685, __pyx_L15_error)
          __Pyx_GOTREF(__pyx_t_7);
          #endif
        }
      } else {
        __pyx_t_7 = __pyx_t_5(__pyx_t_1);
        if (unlikely(!__pyx_t_7)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 685, __pyx_L15_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_7);
      }
      __Pyx_XDECREF_SET(__pyx_9genexpr12__pyx_v_x, __pyx_t_7);
      __pyx_t_7 = 0;

      /* "viur/datastore/transport.pyx":683
 *         "mutations": [
 *             {
 *                 "delete": pythonPropToJson(x)["keyValue"]             # <<<<<<<<<<<<<<
 *             }
 *             for x in keys
 */
      __pyx_t_7 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 683, __pyx_L15_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_GetModuleGlobalName(__pyx_t_11, __pyx_n_s_pythonPropToJson); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 683, __pyx_L15_error)
      __Pyx_GOTREF(__pyx_t_11);
      __pyx_t_12 = NULL;
      __pyx_t_8 = 0;
      #if CYTHON_UNPACK_METHODS
      if (unlikely(PyMethod_Check(__pyx_t_11))) {
        __pyx_t_12 = PyMethod_GET_SELF(__pyx_t_11);
        if (likely(__pyx_t_12)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_11);
          __Pyx_INCREF(__pyx_t_12);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_11, function);
          __pyx_t_8 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_12, __pyx_9genexpr12__pyx_v_x};
        __pyx_t_10 = __Pyx_PyObject_FastCall(__pyx_t_11, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
        __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
        if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 683, __pyx_L15_error)
        __Pyx_GOTREF(__pyx_t_10);
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
      }
      __pyx_t_11 = __Pyx_PyObject_Dict_GetItem(__pyx_t_10, __pyx_n_u_keyValue); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 683, __pyx_L15_error)
      __Pyx_GOTREF(__pyx_t_11);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      if (PyDict_SetItem(__pyx_t_7, __pyx_n_u_delete, __pyx_t_11) < 0) __PYX_ERR(0, 683, __pyx_L15_error)
      __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
      if (unlikely(__Pyx_ListComp_Append(__pyx_t_6, (PyObject*)__pyx_t_7))) __PYX_ERR(0, 681, __pyx_L15_error)
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

      /* "viur/datastore/transport.pyx":685
 *                 "delete": pythonPropToJson(x)["keyValue"]
 *             }
 *             for x in keys             # <<<<<<<<<<<<<<
 *         ]
 *     }
 */
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_9genexpr12__pyx_v_x); __pyx_9genexpr12__pyx_v_x = 0;
    goto __pyx_L19_exit_scope;
    __pyx_L15_error:;
    __Pyx_XDECREF(__pyx_9genexpr12__pyx_v_x); __pyx_9genexpr12__pyx_v_x = 0;
    goto __pyx_L1_error;
    __pyx_L19_exit_scope:;
  } /* exit inner scope */
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_u_mutations, __pyx_t_6) < 0) __PYX_ERR(0, 680, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_v_postData = ((PyObject*)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "viur/datastore/transport.pyx":688
 *         ]
 *     }
 *     currentTxn = currentTransaction.get()             # <<<<<<<<<<<<<<
 *     if currentTxn:
 *         currentTxn["mutations"].extend(postData["mutations"])
 */
  __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_currentTransaction); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 688, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_get); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 688, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_6 = NULL;
  __pyx_t_8 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_1);
    if (likely(__pyx_t_6)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_6);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_1, function);
      __pyx_t_8 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_6, NULL};
    __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_1, __pyx_callargs+1-__pyx_t_8, 0+__pyx_t_8);
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 688, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }
  __pyx_v_currentTxn = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "viur/datastore/transport.pyx":689
 *     }
 *     currentTxn = currentTransaction.get()
 *     if currentTxn:             # <<<<<<<<<<<<<<
 *         currentTxn["mutations"].extend(postData["mutations"])
 *         # Insert placeholders into affectedEntities as we receive a mutation-result for each key deleted
 */
  __pyx_t_9 = __Pyx_PyObject_IsTrue(__pyx_v_currentTxn); if (unlikely((__pyx_t_9 < 0))) __PYX_ERR(0, 689, __pyx_L1_error)
  if (__pyx_t_9) {

    /* "viur/datastore/transport.pyx":690
 *     currentTxn = currentTransaction.get()
 *     if currentTxn:
 *         currentTxn["mutations"].extend(postData["mutations"])             # <<<<<<<<<<<<<<
 *         # Insert placeholders into affectedEntities as we receive a mutation-result for each key deleted
 *         currentTxn["affectedEntities"].extend([None] * len(keys))
 */
    __pyx_t_1 = __Pyx_PyObject_Dict_GetItem(__pyx_v_currentTxn, __pyx_n_u_mutations); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 690, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_extend); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 690, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyDict_GetItem(__pyx_v_postData, __pyx_n_u_mutations); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 690, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_7 = NULL;
    __pyx_t_8 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
        __pyx_t_8 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_t_1};
      __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_6, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 690, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "viur/datastore/transport.pyx":692
 *         currentTxn["mutations"].extend(postData["mutations"])
 *         # Insert placeholders into affectedEntities as we receive a mutation-result for each key deleted
 *         currentTxn["affectedEntities"].extend([None] * len(keys))             # <<<<<<<<<<<<<<
 *         return
 *     resp = authenticated_request(
 */
    __pyx_t_6 = __Pyx_PyObject_Dict_GetItem(__pyx_v_currentTxn, __pyx_n_u_affectedEntities); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 692, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_extend); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 692, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_4 = PyObject_Length(__pyx_v_keys); if (unlikely(__pyx_t_4 == ((Py_ssize_t)-1))) __PYX_ERR(0, 692, __pyx_L1_error)
    __pyx_t_6 = PyList_New(1 * ((__pyx_t_4<0) ? 0:__pyx_t_4)); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 692, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    { Py_ssize_t __pyx_temp;
      for (__pyx_temp=0; __pyx_temp < __pyx_t_4; __pyx_temp++) {
        __Pyx_INCREF(Py_None);
        __Pyx_GIVEREF(Py_None);
        if (__Pyx_PyList_SET_ITEM(__pyx_t_6, __pyx_temp, Py_None)) __PYX_ERR(0, 692, __pyx_L1_error);
      }
    }
    __pyx_t_7 = NULL;
    __pyx_t_8 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
        __pyx_t_8 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_t_6};
      __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_1, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 692, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "viur/datastore/transport.pyx":693
 *         # Insert placeholders into affectedEntities as we receive a mutation-result for each key deleted
 *         currentTxn["affectedEntities"].extend([None] * len(keys))
 *         return             # <<<<<<<<<<<<<<
 *     resp = authenticated_request(
 *         url="https://datastore.googleapis.com/v1/projects/%s:commit" % projectID,
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "viur/datastore/transport.pyx":689
 *     }
 *     currentTxn = currentTransaction.get()
 *     if currentTxn:             # <<<<<<<<<<<<<<
 *         currentTxn["mutations"].extend(postData["mutations"])
 *         # Insert placeholders into affectedEntities as we receive a mutation-result for each key deleted
 */
  }

  /* "viur/datastore/transport.pyx":694
 *         currentTxn["affectedEntities"].extend([None] * len(keys))
 *         return
 *     resp = authenticated_request(             # <<<<<<<<<<<<<<
 *         url="https://datastore.googleapis.com/v1/projects/%s:commit" % projectID,
 *         data=json.dumps(postData).encode("UTF-8"),
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_authenticated_request); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 694, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);

  /* "viur/datastore/transport.pyx":695
 *         return
 *     resp = authenticated_request(
 *         url="https://datastore.googleapis.com/v1/projects/%s:commit" % projectID,             # <<<<<<<<<<<<<<
 *         data=json.dumps(postData).encode("UTF-8"),
 *     )
 */
  __pyx_t_1 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 695, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_projectID); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 695, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_7 = __Pyx_PyUnicode_FormatSafe(__pyx_kp_u_https_datastore_googleapis_com_v_3, __pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 695, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_url, __pyx_t_7) < 0) __PYX_ERR(0, 695, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "viur/datastore/transport.pyx":696
 *     resp = authenticated_request(
 *         url="https://datastore.googleapis.com/v1/projects/%s:commit" % projectID,
 *         data=json.dumps(postData).encode("UTF-8"),             # <<<<<<<<<<<<<<
 *     )
 *     if is_viur_datastore_request_ok(resp):
 */
  __Pyx_GetModuleGlobalName(__pyx_t_11, __pyx_n_s_json); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 696, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_t_11, __pyx_n_s_dumps); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 696, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
  __pyx_t_11 = NULL;
  __pyx_t_8 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_10))) {
    __pyx_t_11 = PyMethod_GET_SELF(__pyx_t_10);
    if (likely(__pyx_t_11)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_10);
      __Pyx_INCREF(__pyx_t_11);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_10, function);
      __pyx_t_8 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_11, __pyx_v_postData};
    __pyx_t_6 = __Pyx_PyObject_FastCall(__pyx_t_10, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
    __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
    if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 696, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
  }
  __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_encode); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 696, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_6 = NULL;
  __pyx_t_8 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_10))) {
    __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_10);
    if (likely(__pyx_t_6)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_10);
      __Pyx_INCREF(__pyx_t_6);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_10, function);
      __pyx_t_8 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_6, __pyx_kp_u_UTF_8};
    __pyx_t_7 = __Pyx_PyObject_FastCall(__pyx_t_10, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 696, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
  }
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_data, __pyx_t_7) < 0) __PYX_ERR(0, 695, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "viur/datastore/transport.pyx":694
 *         currentTxn["affectedEntities"].extend([None] * len(keys))
 *         return
 *     resp = authenticated_request(             # <<<<<<<<<<<<<<
 *         url="https://datastore.googleapis.com/v1/projects/%s:commit" % projectID,
 *         data=json.dumps(postData).encode("UTF-8"),
 */
  __pyx_t_7 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_empty_tuple, __pyx_t_1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 694, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_resp = __pyx_t_7;
  __pyx_t_7 = 0;

  /* "viur/datastore/transport.pyx":698
 *         data=json.dumps(postData).encode("UTF-8"),
 *     )
 *     if is_viur_datastore_request_ok(resp):             # <<<<<<<<<<<<<<
 *         assert PyBytes_AsStringAndSize(resp.content, &data_ptr, &pysize) != -1
 *         element = parser.parse(data_ptr, pysize, 1)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_is_viur_datastore_request_ok); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 698, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = NULL;
  __pyx_t_8 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_1);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_1, function);
      __pyx_t_8 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_v_resp};
    __pyx_t_7 = __Pyx_PyObject_FastCall(__pyx_t_1, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 698, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }
  __pyx_t_9 = __Pyx_PyObject_IsTrue(__pyx_t_7); if (unlikely((__pyx_t_9 < 0))) __PYX_ERR(0, 698, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (__pyx_t_9) {

    /* "viur/datastore/transport.pyx":699
 *     )
 *     if is_viur_datastore_request_ok(resp):
 *         assert PyBytes_AsStringAndSize(resp.content, &data_ptr, &pysize) != -1             # <<<<<<<<<<<<<<
 *         element = parser.parse(data_ptr, pysize, 1)
 *         if (element.at_pointer("/mutationResults").error() != SUCCESS):
 */
    #ifndef CYTHON_WITHOUT_ASSERTIONS
    if (unlikely(__pyx_assertions_enabled())) {
      __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_resp, __pyx_n_s_content); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 699, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_8 = PyBytes_AsStringAndSize(__pyx_t_7, (&__pyx_v_data_ptr), (&__pyx_v_pysize)); if (unlikely(__pyx_t_8 == ((int)-1))) __PYX_ERR(0, 699, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __pyx_t_9 = (__pyx_t_8 != -1L);
      if (unlikely(!__pyx_t_9)) {
        __Pyx_Raise(__pyx_builtin_AssertionError, 0, 0, 0);
        __PYX_ERR(0, 699, __pyx_L1_error)
      }
    }
    #else
    if ((1)); else __PYX_ERR(0, 699, __pyx_L1_error)
    #endif

    /* "viur/datastore/transport.pyx":700
 *     if is_viur_datastore_request_ok(resp):
 *         assert PyBytes_AsStringAndSize(resp.content, &data_ptr, &pysize) != -1
 *         element = parser.parse(data_ptr, pysize, 1)             # <<<<<<<<<<<<<<
 *         if (element.at_pointer("/mutationResults").error() != SUCCESS):
 *             logging.error(resp.content)
 */
    try {
      __pyx_t_13 = __pyx_v_parser.parse(__pyx_v_data_ptr, __pyx_v_pysize, 1);
    } catch(...) {
      __Pyx_CppExn2PyErr();
      __PYX_ERR(0, 700, __pyx_L1_error)
    }
    __pyx_v_element = __PYX_STD_MOVE_IF_SUPPORTED(__pyx_t_13);

    /* "viur/datastore/transport.pyx":701
 *         assert PyBytes_AsStringAndSize(resp.content, &data_ptr, &pysize) != -1
 *         element = parser.parse(data_ptr, pysize, 1)
 *         if (element.at_pointer("/mutationResults").error() != SUCCESS):             # <<<<<<<<<<<<<<
 *             logging.error(resp.content)
 *             raise NoMutationResultsError("No mutation results received")
 */
    __pyx_t_9 = (__pyx_v_element.at_pointer(((char const *)"/mutationResults")).error() != simdjson::error_code::SUCCESS);
    if (unlikely(__pyx_t_9)) {

      /* "viur/datastore/transport.pyx":702
 *         element = parser.parse(data_ptr, pysize, 1)
 *         if (element.at_pointer("/mutationResults").error() != SUCCESS):
 *             logging.error(resp.content)             # <<<<<<<<<<<<<<
 *             raise NoMutationResultsError("No mutation results received")
 *         arrayElem = element.at_key("mutationResults").get_array()
 */
      __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_logging); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 702, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_error); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 702, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_resp, __pyx_n_s_content); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 702, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_10 = NULL;
      __pyx_t_8 = 0;
      #if CYTHON_UNPACK_METHODS
      if (unlikely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_10)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_10);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
          __pyx_t_8 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_10, __pyx_t_1};
        __pyx_t_7 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
        __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 702, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      }
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

      /* "viur/datastore/transport.pyx":703
 *         if (element.at_pointer("/mutationResults").error() != SUCCESS):
 *             logging.error(resp.content)
 *             raise NoMutationResultsError("No mutation results received")             # <<<<<<<<<<<<<<
 *         arrayElem = element.at_key("mutationResults").get_array()
 *         if arrayElem.size() != abs(len(keys)):
 */
      __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_NoMutationResultsError); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 703, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_1 = NULL;
      __pyx_t_8 = 0;
      #if CYTHON_UNPACK_METHODS
      if (unlikely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_1)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_1);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
          __pyx_t_8 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_1, __pyx_kp_u_No_mutation_results_received};
        __pyx_t_7 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
        __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
        if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 703, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      }
      __Pyx_Raise(__pyx_t_7, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __PYX_ERR(0, 703, __pyx_L1_error)

      /* "viur/datastore/transport.pyx":701
 *         assert PyBytes_AsStringAndSize(resp.content, &data_ptr, &pysize) != -1
 *         element = parser.parse(data_ptr, pysize, 1)
 *         if (element.at_pointer("/mutationResults").error() != SUCCESS):             # <<<<<<<<<<<<<<
 *             logging.error(resp.content)
 *             raise NoMutationResultsError("No mutation results received")
 */
    }

    /* "viur/datastore/transport.pyx":704
 *             logging.error(resp.content)
 *             raise NoMutationResultsError("No mutation results received")
 *         arrayElem = element.at_key("mutationResults").get_array()             # <<<<<<<<<<<<<<
 *         if arrayElem.size() != abs(len(keys)):
 *             logging.error(resp.content)
 */
    try {
      __pyx_t_13 = __pyx_v_element.at_key(((char const *)"mutationResults"));
    } catch(...) {
      __Pyx_CppExn2PyErr();
      __PYX_ERR(0, 704, __pyx_L1_error)
    }
    try {
      __pyx_t_14 = __pyx_t_13.get_array();
    } catch(...) {
      __Pyx_CppExn2PyErr();
      __PYX_ERR(0, 704, __pyx_L1_error)
    }
    __pyx_v_arrayElem = __PYX_STD_MOVE_IF_SUPPORTED(__pyx_t_14);

    /* "viur/datastore/transport.pyx":705
 *             raise NoMutationResultsError("No mutation results received")
 *         arrayElem = element.at_key("mutationResults").get_array()
 *         if arrayElem.size() != abs(len(keys)):             # <<<<<<<<<<<<<<
 *             logging.error(resp.content)
 *             raise ValueError("Invalid number of mutation-results received")
 */
    __pyx_t_7 = __Pyx_PyInt_From_int(__pyx_v_arrayElem.size()); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 705, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_4 = PyObject_Length(__pyx_v_keys); if (unlikely(__pyx_t_4 == ((Py_ssize_t)-1))) __PYX_ERR(0, 705, __pyx_L1_error)
    __pyx_t_3 = PyInt_FromSsize_t(__pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 705, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyNumber_Absolute(__pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 705, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = PyObject_RichCompare(__pyx_t_7, __pyx_t_1, Py_NE); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 705, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_9 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely((__pyx_t_9 < 0))) __PYX_ERR(0, 705, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(__pyx_t_9)) {

      /* "viur/datastore/transport.pyx":706
 *         arrayElem = element.at_key("mutationResults").get_array()
 *         if arrayElem.size() != abs(len(keys)):
 *             logging.error(resp.content)             # <<<<<<<<<<<<<<
 *             raise ValueError("Invalid number of mutation-results received")
 *     if conf["memcache_client"] is not None:
 */
      __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_logging); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 706, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_error); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 706, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_resp, __pyx_n_s_content); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 706, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_10 = NULL;
      __pyx_t_8 = 0;
      #if CYTHON_UNPACK_METHODS
      if (unlikely(PyMethod_Check(__pyx_t_7))) {
        __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_7);
        if (likely(__pyx_t_10)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
          __Pyx_INCREF(__pyx_t_10);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_7, function);
          __pyx_t_8 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_10, __pyx_t_1};
        __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_7, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
        __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 706, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      }
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

      /* "viur/datastore/transport.pyx":707
 *         if arrayElem.size() != abs(len(keys)):
 *             logging.error(resp.content)
 *             raise ValueError("Invalid number of mutation-results received")             # <<<<<<<<<<<<<<
 *     if conf["memcache_client"] is not None:
 *         cache.delete([str(key) for key in keys])
 */
      __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__15, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 707, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_Raise(__pyx_t_3, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __PYX_ERR(0, 707, __pyx_L1_error)

      /* "viur/datastore/transport.pyx":705
 *             raise NoMutationResultsError("No mutation results received")
 *         arrayElem = element.at_key("mutationResults").get_array()
 *         if arrayElem.size() != abs(len(keys)):             # <<<<<<<<<<<<<<
 *             logging.error(resp.content)
 *             raise ValueError("Invalid number of mutation-results received")
 */
    }

    /* "viur/datastore/transport.pyx":698
 *         data=json.dumps(postData).encode("UTF-8"),
 *     )
 *     if is_viur_datastore_request_ok(resp):             # <<<<<<<<<<<<<<
 *         assert PyBytes_AsStringAndSize(resp.content, &data_ptr, &pysize) != -1
 *         element = parser.parse(data_ptr, pysize, 1)
 */
  }

  /* "viur/datastore/transport.pyx":708
 *             logging.error(resp.content)
 *             raise ValueError("Invalid number of mutation-results received")
 *     if conf["memcache_client"] is not None:             # <<<<<<<<<<<<<<
 *         cache.delete([str(key) for key in keys])
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_conf); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 708, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_7 = __Pyx_PyObject_Dict_GetItem(__pyx_t_3, __pyx_n_u_memcache_client); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 708, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_9 = (__pyx_t_7 != Py_None);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (__pyx_t_9) {

    /* "viur/datastore/transport.pyx":709
 *             raise ValueError("Invalid number of mutation-results received")
 *     if conf["memcache_client"] is not None:
 *         cache.delete([str(key) for key in keys])             # <<<<<<<<<<<<<<
 * 
 * def Put(entities: Union[Entity, List[Entity]]) -> Union[Entity, List[Entity]]:
 */
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_cache); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 709, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_delete); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 709, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    { /* enter inner scope */
      __pyx_t_3 = PyList_New(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 709, __pyx_L27_error)
      __Pyx_GOTREF(__pyx_t_3);
      if (likely(PyList_CheckExact(__pyx_v_keys)) || PyTuple_CheckExact(__pyx_v_keys)) {
        __pyx_t_10 = __pyx_v_keys; __Pyx_INCREF(__pyx_t_10);
        __pyx_t_4 = 0;
        __pyx_t_5 = NULL;
      } else {
        __pyx_t_4 = -1; __pyx_t_10 = PyObject_GetIter(__pyx_v_keys); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 709, __pyx_L27_error)
        __Pyx_GOTREF(__pyx_t_10);
        __pyx_t_5 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_10); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 709, __pyx_L27_error)
      }
      for (;;) {
        if (likely(!__pyx_t_5)) {
          if (likely(PyList_CheckExact(__pyx_t_10))) {
            {
              Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_10);
              #if !CYTHON_ASSUME_SAFE_MACROS
              if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 709, __pyx_L27_error)
              #endif
              if (__pyx_t_4 >= __pyx_temp) break;
            }
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_6 = PyList_GET_ITEM(__pyx_t_10, __pyx_t_4); __Pyx_INCREF(__pyx_t_6); __pyx_t_4++; if (unlikely((0 < 0))) __PYX_ERR(0, 709, __pyx_L27_error)
            #else
            __pyx_t_6 = __Pyx_PySequence_ITEM(__pyx_t_10, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 709, __pyx_L27_error)
            __Pyx_GOTREF(__pyx_t_6);
            #endif
          } else {
            {
              Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_10);
              #if !CYTHON_ASSUME_SAFE_MACROS
              if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 709, __pyx_L27_error)
              #endif
              if (__pyx_t_4 >= __pyx_temp) break;
            }
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_6 = PyTuple_GET_ITEM(__pyx_t_10, __pyx_t_4); __Pyx_INCREF(__pyx_t_6); __pyx_t_4++; if (unlikely((0 < 0))) __PYX_ERR(0, 709, __pyx_L27_error)
            #else
            __pyx_t_6 = __Pyx_PySequence_ITEM(__pyx_t_10, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 709, __pyx_L27_error)
            __Pyx_GOTREF(__pyx_t_6);
            #endif
          }
        } else {
          __pyx_t_6 = __pyx_t_5(__pyx_t_10);
          if (unlikely(!__pyx_t_6)) {
            PyObject* exc_type = PyErr_Occurred();
            if (exc_type) {
              if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
              else __PYX_ERR(0, 709, __pyx_L27_error)
            }
            break;
          }
          __Pyx_GOTREF(__pyx_t_6);
        }
        __Pyx_XDECREF_SET(__pyx_9genexpr13__pyx_v_key, __pyx_t_6);
        __pyx_t_6 = 0;
        __pyx_t_6 = __Pyx_PyObject_Str(__pyx_9genexpr13__pyx_v_key); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 709, __pyx_L27_error)
        __Pyx_GOTREF(__pyx_t_6);
        if (unlikely(__Pyx_ListComp_Append(__pyx_t_3, (PyObject*)__pyx_t_6))) __PYX_ERR(0, 709, __pyx_L27_error)
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_XDECREF(__pyx_9genexpr13__pyx_v_key); __pyx_9genexpr13__pyx_v_key = 0;
      goto __pyx_L31_exit_scope;
      __pyx_L27_error:;
      __Pyx_XDECREF(__pyx_9genexpr13__pyx_v_key); __pyx_9genexpr13__pyx_v_key = 0;
      goto __pyx_L1_error;
      __pyx_L31_exit_scope:;
    } /* exit inner scope */
    __pyx_t_10 = NULL;
    __pyx_t_8 = 0;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_10)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_10);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
        __pyx_t_8 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_10, __pyx_t_3};
      __pyx_t_7 = __Pyx_PyObject_FastCall(__pyx_t_1, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 709, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    }
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

    /* "viur/datastore/transport.pyx":708
 *             logging.error(resp.content)
 *             raise ValueError("Invalid number of mutation-results received")
 *     if conf["memcache_client"] is not None:             # <<<<<<<<<<<<<<
 *         cache.delete([str(key) for key in keys])
 * 
 */
  }

  /* "viur/datastore/transport.pyx":656
 *         return [res.get(key) for key in untouched_keys]  # Sort by order of incoming keys
 * 
 * def Delete(keys: Union[Key, List[Key], Entity, List[Entity]]) -> None:             # <<<<<<<<<<<<<<
 *     """
 *         Deletes the entities stored under the given key(s).
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_AddTraceback("viur.datastore.transport.Delete", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_accessLog);
  __Pyx_XDECREF(__pyx_v_postData);
  __Pyx_XDECREF(__pyx_v_currentTxn);
  __Pyx_XDECREF(__pyx_v_resp);
  __Pyx_XDECREF(__pyx_9genexpr11__pyx_v_x);
  __Pyx_XDECREF(__pyx_9genexpr12__pyx_v_x);
  __Pyx_XDECREF(__pyx_9genexpr13__pyx_v_key);
  __Pyx_XDECREF(__pyx_v_keys);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "viur/datastore/transport.pyx":711
 *         cache.delete([str(key) for key in keys])
 * 
 * def Put(entities: Union[Entity, List[Entity]]) -> Union[Entity, List[Entity]]:             # <<<<<<<<<<<<<<
 *     """
 *         Writes the given entities into the datastore. The entities can be from different kinds. If an entity has an
 */

/* Python wrapper */
static PyObject *__pyx_pw_4viur_9datastore_9transport_13Put(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_4viur_9datastore_9transport_12Put, "\n        Writes the given entities into the datastore. The entities can be from different kinds. If an entity has an\n        complete key, and there's already an entity stored under the given key, it's overwritten.\n\n        :param entities: The entities to store\n        :return: The Entity or List of Entities as supplied, with partial keys replaced by full ones (unless called\n            inside a transaction, in which case we return None as no Keys have been determined yet)\n    ");
static PyMethodDef __pyx_mdef_4viur_9datastore_9transport_13Put = {"Put", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_4viur_9datastore_9transport_13Put, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_4viur_9datastore_9transport_12Put};
static PyObject *__pyx_pw_4viur_9datastore_9transport_13Put(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_entities = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("Put (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_entities,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_entities)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 711, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "Put") < 0)) __PYX_ERR(0, 711, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_entities = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("Put", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 711, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("viur.datastore.transport.Put", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4viur_9datastore_9transport_12Put(__pyx_self, __pyx_v_entities);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4viur_9datastore_9transport_12Put(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_entities) {
  simdjson::dom::parser __pyx_v_parser;
  Py_ssize_t __pyx_v_pysize;
  char *__pyx_v_data_ptr;
  simdjson::dom::element __pyx_v_element;
  simdjson::dom::element __pyx_v_innerArrayElem;
  simdjson::dom::array __pyx_v_arrayElem;
  simdjson::dom::array::iterator __pyx_v_arrayIt;
  PyObject *__pyx_v_accessLog = NULL;
  PyObject *__pyx_v_postData = NULL;
  PyObject *__pyx_v_currentTxn = NULL;
  PyObject *__pyx_v_resp = NULL;
  PyObject *__pyx_v_idx = NULL;
  PyObject *__pyx_9genexpr14__pyx_v_x = NULL;
  PyObject *__pyx_9genexpr15__pyx_v_x = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  Py_ssize_t __pyx_t_7;
  PyObject *(*__pyx_t_8)(PyObject *);
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_t_11;
  PyObject *__pyx_t_12 = NULL;
  simdjson::dom::element __pyx_t_13;
  simdjson::dom::array __pyx_t_14;
  std::string_view __pyx_t_15;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("Put", 0);
  __Pyx_INCREF(__pyx_v_entities);

  /* "viur/datastore/transport.pyx":720
 *             inside a transaction, in which case we return None as no Keys have been determined yet)
 *     """
 *     cdef simdjsonParser parser = simdjsonParser()             # <<<<<<<<<<<<<<
 *     cdef Py_ssize_t pysize
 *     cdef char * data_ptr
 */
  __pyx_v_parser = simdjson::dom::parser();

  /* "viur/datastore/transport.pyx":726
 *     cdef simdjsonArray arrayElem
 *     cdef simdjsonArray.iterator arrayIt
 *     if isinstance(entities, Entity):             # <<<<<<<<<<<<<<
 *         entities = [entities]
 *     accessLog = currentDbAccessLog.get()
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_Entity); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 726, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyObject_IsInstance(__pyx_v_entities, __pyx_t_1); if (unlikely(__pyx_t_2 == ((int)-1))) __PYX_ERR(0, 726, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "viur/datastore/transport.pyx":727
 *     cdef simdjsonArray.iterator arrayIt
 *     if isinstance(entities, Entity):
 *         entities = [entities]             # <<<<<<<<<<<<<<
 *     accessLog = currentDbAccessLog.get()
 *     if isinstance(accessLog, set):
 */
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 727, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_v_entities);
    __Pyx_GIVEREF(__pyx_v_entities);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 0, __pyx_v_entities)) __PYX_ERR(0, 727, __pyx_L1_error);
    __Pyx_DECREF_SET(__pyx_v_entities, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "viur/datastore/transport.pyx":726
 *     cdef simdjsonArray arrayElem
 *     cdef simdjsonArray.iterator arrayIt
 *     if isinstance(entities, Entity):             # <<<<<<<<<<<<<<
 *         entities = [entities]
 *     accessLog = currentDbAccessLog.get()
 */
  }

  /* "viur/datastore/transport.pyx":728
 *     if isinstance(entities, Entity):
 *         entities = [entities]
 *     accessLog = currentDbAccessLog.get()             # <<<<<<<<<<<<<<
 *     if isinstance(accessLog, set):
 *         accessLog.update(set([x.key for x in entities if not x.key.is_partial]))
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_currentDbAccessLog); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 728, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_get); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 728, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_5, 0+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 728, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  }
  __pyx_v_accessLog = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "viur/datastore/transport.pyx":729
 *         entities = [entities]
 *     accessLog = currentDbAccessLog.get()
 *     if isinstance(accessLog, set):             # <<<<<<<<<<<<<<
 *         accessLog.update(set([x.key for x in entities if not x.key.is_partial]))
 *     postData = {
 */
  __pyx_t_2 = PySet_Check(__pyx_v_accessLog); 
  if (__pyx_t_2) {

    /* "viur/datastore/transport.pyx":730
 *     accessLog = currentDbAccessLog.get()
 *     if isinstance(accessLog, set):
 *         accessLog.update(set([x.key for x in entities if not x.key.is_partial]))             # <<<<<<<<<<<<<<
 *     postData = {
 *         "mode": "NON_TRANSACTIONAL",  # Always NON_TRANSACTIONAL; if we're inside a transaction we'll abort below
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_accessLog, __pyx_n_s_update); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 730, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    { /* enter inner scope */
      __pyx_t_3 = PyList_New(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 730, __pyx_L7_error)
      __Pyx_GOTREF(__pyx_t_3);
      if (likely(PyList_CheckExact(__pyx_v_entities)) || PyTuple_CheckExact(__pyx_v_entities)) {
        __pyx_t_6 = __pyx_v_entities; __Pyx_INCREF(__pyx_t_6);
        __pyx_t_7 = 0;
        __pyx_t_8 = NULL;
      } else {
        __pyx_t_7 = -1; __pyx_t_6 = PyObject_GetIter(__pyx_v_entities); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 730, __pyx_L7_error)
        __Pyx_GOTREF(__pyx_t_6);
        __pyx_t_8 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_6); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 730, __pyx_L7_error)
      }
      for (;;) {
        if (likely(!__pyx_t_8)) {
          if (likely(PyList_CheckExact(__pyx_t_6))) {
            {
              Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_6);
              #if !CYTHON_ASSUME_SAFE_MACROS
              if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 730, __pyx_L7_error)
              #endif
              if (__pyx_t_7 >= __pyx_temp) break;
            }
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_9 = PyList_GET_ITEM(__pyx_t_6, __pyx_t_7); __Pyx_INCREF(__pyx_t_9); __pyx_t_7++; if (unlikely((0 < 0))) __PYX_ERR(0, 730, __pyx_L7_error)
            #else
            __pyx_t_9 = __Pyx_PySequence_ITEM(__pyx_t_6, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 730, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_9);
            #endif
          } else {
            {
              Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_6);
              #if !CYTHON_ASSUME_SAFE_MACROS
              if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 730, __pyx_L7_error)
              #endif
              if (__pyx_t_7 >= __pyx_temp) break;
            }
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_9 = PyTuple_GET_ITEM(__pyx_t_6, __pyx_t_7); __Pyx_INCREF(__pyx_t_9); __pyx_t_7++; if (unlikely((0 < 0))) __PYX_ERR(0, 730, __pyx_L7_error)
            #else
            __pyx_t_9 = __Pyx_PySequence_ITEM(__pyx_t_6, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 730, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_9);
            #endif
          }
        } else {
          __pyx_t_9 = __pyx_t_8(__pyx_t_6);
          if (unlikely(!__pyx_t_9)) {
            PyObject* exc_type = PyErr_Occurred();
            if (exc_type) {
              if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
              else __PYX_ERR(0, 730, __pyx_L7_error)
            }
            break;
          }
          __Pyx_GOTREF(__pyx_t_9);
        }
        __Pyx_XDECREF_SET(__pyx_9genexpr14__pyx_v_x, __pyx_t_9);
        __pyx_t_9 = 0;
        __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_9genexpr14__pyx_v_x, __pyx_n_s_key); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 730, __pyx_L7_error)
        __Pyx_GOTREF(__pyx_t_9);
        __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_is_partial); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 730, __pyx_L7_error)
        __Pyx_GOTREF(__pyx_t_10);
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_10); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(0, 730, __pyx_L7_error)
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
        __pyx_t_11 = (!__pyx_t_2);
        if (__pyx_t_11) {
          __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_9genexpr14__pyx_v_x, __pyx_n_s_key); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 730, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_10);
          if (unlikely(__Pyx_ListComp_Append(__pyx_t_3, (PyObject*)__pyx_t_10))) __PYX_ERR(0, 730, __pyx_L7_error)
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
        }
      }
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_XDECREF(__pyx_9genexpr14__pyx_v_x); __pyx_9genexpr14__pyx_v_x = 0;
      goto __pyx_L12_exit_scope;
      __pyx_L7_error:;
      __Pyx_XDECREF(__pyx_9genexpr14__pyx_v_x); __pyx_9genexpr14__pyx_v_x = 0;
      goto __pyx_L1_error;
      __pyx_L12_exit_scope:;
    } /* exit inner scope */
    __pyx_t_6 = PySet_New(__pyx_t_3); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 730, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = NULL;
    __pyx_t_5 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
        __pyx_t_5 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_t_6};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 730, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "viur/datastore/transport.pyx":729
 *         entities = [entities]
 *     accessLog = currentDbAccessLog.get()
 *     if isinstance(accessLog, set):             # <<<<<<<<<<<<<<
 *         accessLog.update(set([x.key for x in entities if not x.key.is_partial]))
 *     postData = {
 */
  }

  /* "viur/datastore/transport.pyx":732
 *         accessLog.update(set([x.key for x in entities if not x.key.is_partial]))
 *     postData = {
 *         "mode": "NON_TRANSACTIONAL",  # Always NON_TRANSACTIONAL; if we're inside a transaction we'll abort below             # <<<<<<<<<<<<<<
 *         "mutations": [
 *             {
 */
  __pyx_t_1 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 732, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_u_mode, __pyx_n_u_NON_TRANSACTIONAL) < 0) __PYX_ERR(0, 732, __pyx_L1_error)
  { /* enter inner scope */

    /* "viur/datastore/transport.pyx":733
 *     postData = {
 *         "mode": "NON_TRANSACTIONAL",  # Always NON_TRANSACTIONAL; if we're inside a transaction we'll abort below
 *         "mutations": [             # <<<<<<<<<<<<<<
 *             {
 *                 "upsert": pythonPropToJson(x)["entityValue"]
 */
    __pyx_t_4 = PyList_New(0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 733, __pyx_L15_error)
    __Pyx_GOTREF(__pyx_t_4);

    /* "viur/datastore/transport.pyx":737
 *                 "upsert": pythonPropToJson(x)["entityValue"]
 *             }
 *             for x in entities             # <<<<<<<<<<<<<<
 *         ]
 *     }
 */
    if (likely(PyList_CheckExact(__pyx_v_entities)) || PyTuple_CheckExact(__pyx_v_entities)) {
      __pyx_t_6 = __pyx_v_entities; __Pyx_INCREF(__pyx_t_6);
      __pyx_t_7 = 0;
      __pyx_t_8 = NULL;
    } else {
      __pyx_t_7 = -1; __pyx_t_6 = PyObject_GetIter(__pyx_v_entities); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 737, __pyx_L15_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_8 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_6); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 737, __pyx_L15_error)
    }
    for (;;) {
      if (likely(!__pyx_t_8)) {
        if (likely(PyList_CheckExact(__pyx_t_6))) {
          {
            Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_6);
            #if !CYTHON_ASSUME_SAFE_MACROS
            if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 737, __pyx_L15_error)
            #endif
            if (__pyx_t_7 >= __pyx_temp) break;
          }
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_3 = PyList_GET_ITEM(__pyx_t_6, __pyx_t_7); __Pyx_INCREF(__pyx_t_3); __pyx_t_7++; if (unlikely((0 < 0))) __PYX_ERR(0, 737, __pyx_L15_error)
          #else
          __pyx_t_3 = __Pyx_PySequence_ITEM(__pyx_t_6, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 737, __pyx_L15_error)
          __Pyx_GOTREF(__pyx_t_3);
          #endif
        } else {
          {
            Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_6);
            #if !CYTHON_ASSUME_SAFE_MACROS
            if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 737, __pyx_L15_error)
            #endif
            if (__pyx_t_7 >= __pyx_temp) break;
          }
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_3 = PyTuple_GET_ITEM(__pyx_t_6, __pyx_t_7); __Pyx_INCREF(__pyx_t_3); __pyx_t_7++; if (unlikely((0 < 0))) __PYX_ERR(0, 737, __pyx_L15_error)
          #else
          __pyx_t_3 = __Pyx_PySequence_ITEM(__pyx_t_6, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 737, __pyx_L15_error)
          __Pyx_GOTREF(__pyx_t_3);
          #endif
        }
      } else {
        __pyx_t_3 = __pyx_t_8(__pyx_t_6);
        if (unlikely(!__pyx_t_3)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 737, __pyx_L15_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_3);
      }
      __Pyx_XDECREF_SET(__pyx_9genexpr15__pyx_v_x, __pyx_t_3);
      __pyx_t_3 = 0;

      /* "viur/datastore/transport.pyx":735
 *         "mutations": [
 *             {
 *                 "upsert": pythonPropToJson(x)["entityValue"]             # <<<<<<<<<<<<<<
 *             }
 *             for x in entities
 */
      __pyx_t_3 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 735, __pyx_L15_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_GetModuleGlobalName(__pyx_t_9, __pyx_n_s_pythonPropToJson); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 735, __pyx_L15_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_12 = NULL;
      __pyx_t_5 = 0;
      #if CYTHON_UNPACK_METHODS
      if (unlikely(PyMethod_Check(__pyx_t_9))) {
        __pyx_t_12 = PyMethod_GET_SELF(__pyx_t_9);
        if (likely(__pyx_t_12)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_9);
          __Pyx_INCREF(__pyx_t_12);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_9, function);
          __pyx_t_5 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_12, __pyx_9genexpr15__pyx_v_x};
        __pyx_t_10 = __Pyx_PyObject_FastCall(__pyx_t_9, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
        __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
        if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 735, __pyx_L15_error)
        __Pyx_GOTREF(__pyx_t_10);
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      }
      __pyx_t_9 = __Pyx_PyObject_Dict_GetItem(__pyx_t_10, __pyx_n_u_entityValue); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 735, __pyx_L15_error)
      __Pyx_GOTREF(__pyx_t_9);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      if (PyDict_SetItem(__pyx_t_3, __pyx_n_u_upsert, __pyx_t_9) < 0) __PYX_ERR(0, 735, __pyx_L15_error)
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      if (unlikely(__Pyx_ListComp_Append(__pyx_t_4, (PyObject*)__pyx_t_3))) __PYX_ERR(0, 733, __pyx_L15_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

      /* "viur/datastore/transport.pyx":737
 *                 "upsert": pythonPropToJson(x)["entityValue"]
 *             }
 *             for x in entities             # <<<<<<<<<<<<<<
 *         ]
 *     }
 */
    }
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_XDECREF(__pyx_9genexpr15__pyx_v_x); __pyx_9genexpr15__pyx_v_x = 0;
    goto __pyx_L19_exit_scope;
    __pyx_L15_error:;
    __Pyx_XDECREF(__pyx_9genexpr15__pyx_v_x); __pyx_9genexpr15__pyx_v_x = 0;
    goto __pyx_L1_error;
    __pyx_L19_exit_scope:;
  } /* exit inner scope */
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_u_mutations, __pyx_t_4) < 0) __PYX_ERR(0, 732, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_v_postData = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "viur/datastore/transport.pyx":740
 *         ]
 *     }
 *     currentTxn = currentTransaction.get()             # <<<<<<<<<<<<<<
 *     if currentTxn:  # We're currently inside a transaction, just queue the changes
 *         currentTxn["mutations"].extend(postData["mutations"])
 */
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_currentTransaction); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 740, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_get); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 740, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_6))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_6);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_6, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_6, __pyx_callargs+1-__pyx_t_5, 0+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 740, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __pyx_v_currentTxn = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "viur/datastore/transport.pyx":741
 *     }
 *     currentTxn = currentTransaction.get()
 *     if currentTxn:  # We're currently inside a transaction, just queue the changes             # <<<<<<<<<<<<<<
 *         currentTxn["mutations"].extend(postData["mutations"])
 *         currentTxn["affectedEntities"].extend(entities)
 */
  __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_v_currentTxn); if (unlikely((__pyx_t_11 < 0))) __PYX_ERR(0, 741, __pyx_L1_error)
  if (__pyx_t_11) {

    /* "viur/datastore/transport.pyx":742
 *     currentTxn = currentTransaction.get()
 *     if currentTxn:  # We're currently inside a transaction, just queue the changes
 *         currentTxn["mutations"].extend(postData["mutations"])             # <<<<<<<<<<<<<<
 *         currentTxn["affectedEntities"].extend(entities)
 *         return
 */
    __pyx_t_6 = __Pyx_PyObject_Dict_GetItem(__pyx_v_currentTxn, __pyx_n_u_mutations); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 742, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_extend); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 742, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_6 = __Pyx_PyDict_GetItem(__pyx_v_postData, __pyx_n_u_mutations); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 742, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_3 = NULL;
    __pyx_t_5 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
        __pyx_t_5 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_t_6};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 742, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "viur/datastore/transport.pyx":743
 *     if currentTxn:  # We're currently inside a transaction, just queue the changes
 *         currentTxn["mutations"].extend(postData["mutations"])
 *         currentTxn["affectedEntities"].extend(entities)             # <<<<<<<<<<<<<<
 *         return
 *     resp = authenticated_request(
 */
    __pyx_t_4 = __Pyx_PyObject_Dict_GetItem(__pyx_v_currentTxn, __pyx_n_u_affectedEntities); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 743, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_extend); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 743, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = NULL;
    __pyx_t_5 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
        __pyx_t_5 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v_entities};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_6, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 743, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "viur/datastore/transport.pyx":744
 *         currentTxn["mutations"].extend(postData["mutations"])
 *         currentTxn["affectedEntities"].extend(entities)
 *         return             # <<<<<<<<<<<<<<
 *     resp = authenticated_request(
 *         url="https://datastore.googleapis.com/v1/projects/%s:commit" % projectID,
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "viur/datastore/transport.pyx":741
 *     }
 *     currentTxn = currentTransaction.get()
 *     if currentTxn:  # We're currently inside a transaction, just queue the changes             # <<<<<<<<<<<<<<
 *         currentTxn["mutations"].extend(postData["mutations"])
 *         currentTxn["affectedEntities"].extend(entities)
 */
  }

  /* "viur/datastore/transport.pyx":745
 *         currentTxn["affectedEntities"].extend(entities)
 *         return
 *     resp = authenticated_request(             # <<<<<<<<<<<<<<
 *         url="https://datastore.googleapis.com/v1/projects/%s:commit" % projectID,
 *         data=json.dumps(postData).encode("UTF-8"),
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_authenticated_request); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 745, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);

  /* "viur/datastore/transport.pyx":746
 *         return
 *     resp = authenticated_request(
 *         url="https://datastore.googleapis.com/v1/projects/%s:commit" % projectID,             # <<<<<<<<<<<<<<
 *         data=json.dumps(postData).encode("UTF-8"),
 *     )
 */
  __pyx_t_6 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 746, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_projectID); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 746, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = __Pyx_PyUnicode_FormatSafe(__pyx_kp_u_https_datastore_googleapis_com_v_3, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 746, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (PyDict_SetItem(__pyx_t_6, __pyx_n_s_url, __pyx_t_3) < 0) __PYX_ERR(0, 746, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "viur/datastore/transport.pyx":747
 *     resp = authenticated_request(
 *         url="https://datastore.googleapis.com/v1/projects/%s:commit" % projectID,
 *         data=json.dumps(postData).encode("UTF-8"),             # <<<<<<<<<<<<<<
 *     )
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_9, __pyx_n_s_json); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 747, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_dumps); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 747, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_t_9 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_10))) {
    __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_10);
    if (likely(__pyx_t_9)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_10);
      __Pyx_INCREF(__pyx_t_9);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_10, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_9, __pyx_v_postData};
    __pyx_t_4 = __Pyx_PyObject_FastCall(__pyx_t_10, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
    if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 747, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
  }
  __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_encode); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 747, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_10))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_10);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_10);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_10, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_kp_u_UTF_8};
    __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_10, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 747, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
  }
  if (PyDict_SetItem(__pyx_t_6, __pyx_n_s_data, __pyx_t_3) < 0) __PYX_ERR(0, 746, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "viur/datastore/transport.pyx":745
 *         currentTxn["affectedEntities"].extend(entities)
 *         return
 *     resp = authenticated_request(             # <<<<<<<<<<<<<<
 *         url="https://datastore.googleapis.com/v1/projects/%s:commit" % projectID,
 *         data=json.dumps(postData).encode("UTF-8"),
 */
  __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_empty_tuple, __pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 745, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_v_resp = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "viur/datastore/transport.pyx":750
 *     )
 * 
 *     if is_viur_datastore_request_ok(resp):             # <<<<<<<<<<<<<<
 *         assert PyBytes_AsStringAndSize(resp.content, &data_ptr, &pysize) != -1
 *         element = parser.parse(data_ptr, pysize, 1)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_is_viur_datastore_request_ok); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 750, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_1 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_6))) {
    __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_6);
    if (likely(__pyx_t_1)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_6, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_1, __pyx_v_resp};
    __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_6, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 750, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely((__pyx_t_11 < 0))) __PYX_ERR(0, 750, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (__pyx_t_11) {

    /* "viur/datastore/transport.pyx":751
 * 
 *     if is_viur_datastore_request_ok(resp):
 *         assert PyBytes_AsStringAndSize(resp.content, &data_ptr, &pysize) != -1             # <<<<<<<<<<<<<<
 *         element = parser.parse(data_ptr, pysize, 1)
 *         if (element.at_pointer("/mutationResults").error() != SUCCESS):
 */
    #ifndef CYTHON_WITHOUT_ASSERTIONS
    if (unlikely(__pyx_assertions_enabled())) {
      __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_resp, __pyx_n_s_content); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 751, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_5 = PyBytes_AsStringAndSize(__pyx_t_3, (&__pyx_v_data_ptr), (&__pyx_v_pysize)); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(0, 751, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_11 = (__pyx_t_5 != -1L);
      if (unlikely(!__pyx_t_11)) {
        __Pyx_Raise(__pyx_builtin_AssertionError, 0, 0, 0);
        __PYX_ERR(0, 751, __pyx_L1_error)
      }
    }
    #else
    if ((1)); else __PYX_ERR(0, 751, __pyx_L1_error)
    #endif

    /* "viur/datastore/transport.pyx":752
 *     if is_viur_datastore_request_ok(resp):
 *         assert PyBytes_AsStringAndSize(resp.content, &data_ptr, &pysize) != -1
 *         element = parser.parse(data_ptr, pysize, 1)             # <<<<<<<<<<<<<<
 *         if (element.at_pointer("/mutationResults").error() != SUCCESS):
 *             logging.error(resp.content)
 */
    try {
      __pyx_t_13 = __pyx_v_parser.parse(__pyx_v_data_ptr, __pyx_v_pysize, 1);
    } catch(...) {
      __Pyx_CppExn2PyErr();
      __PYX_ERR(0, 752, __pyx_L1_error)
    }
    __pyx_v_element = __PYX_STD_MOVE_IF_SUPPORTED(__pyx_t_13);

    /* "viur/datastore/transport.pyx":753
 *         assert PyBytes_AsStringAndSize(resp.content, &data_ptr, &pysize) != -1
 *         element = parser.parse(data_ptr, pysize, 1)
 *         if (element.at_pointer("/mutationResults").error() != SUCCESS):             # <<<<<<<<<<<<<<
 *             logging.error(resp.content)
 *             raise ValueError("No mutation-results received")
 */
    __pyx_t_11 = (__pyx_v_element.at_pointer(((char const *)"/mutationResults")).error() != simdjson::error_code::SUCCESS);
    if (unlikely(__pyx_t_11)) {

      /* "viur/datastore/transport.pyx":754
 *         element = parser.parse(data_ptr, pysize, 1)
 *         if (element.at_pointer("/mutationResults").error() != SUCCESS):
 *             logging.error(resp.content)             # <<<<<<<<<<<<<<
 *             raise ValueError("No mutation-results received")
 *         arrayElem = element.at_key("mutationResults").get_array()
 */
      __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_logging); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 754, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_error); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 754, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_resp, __pyx_n_s_content); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 754, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_10 = NULL;
      __pyx_t_5 = 0;
      #if CYTHON_UNPACK_METHODS
      if (unlikely(PyMethod_Check(__pyx_t_1))) {
        __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_1);
        if (likely(__pyx_t_10)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
          __Pyx_INCREF(__pyx_t_10);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_1, function);
          __pyx_t_5 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_10, __pyx_t_6};
        __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_1, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
        __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 754, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      }
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

      /* "viur/datastore/transport.pyx":755
 *         if (element.at_pointer("/mutationResults").error() != SUCCESS):
 *             logging.error(resp.content)
 *             raise ValueError("No mutation-results received")             # <<<<<<<<<<<<<<
 *         arrayElem = element.at_key("mutationResults").get_array()
 *         if arrayElem.size() != abs(len(entities)):
 */
      __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__16, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 755, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_Raise(__pyx_t_3, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __PYX_ERR(0, 755, __pyx_L1_error)

      /* "viur/datastore/transport.pyx":753
 *         assert PyBytes_AsStringAndSize(resp.content, &data_ptr, &pysize) != -1
 *         element = parser.parse(data_ptr, pysize, 1)
 *         if (element.at_pointer("/mutationResults").error() != SUCCESS):             # <<<<<<<<<<<<<<
 *             logging.error(resp.content)
 *             raise ValueError("No mutation-results received")
 */
    }

    /* "viur/datastore/transport.pyx":756
 *             logging.error(resp.content)
 *             raise ValueError("No mutation-results received")
 *         arrayElem = element.at_key("mutationResults").get_array()             # <<<<<<<<<<<<<<
 *         if arrayElem.size() != abs(len(entities)):
 *             logging.error(resp.content)
 */
    try {
      __pyx_t_13 = __pyx_v_element.at_key(((char const *)"mutationResults"));
    } catch(...) {
      __Pyx_CppExn2PyErr();
      __PYX_ERR(0, 756, __pyx_L1_error)
    }
    try {
      __pyx_t_14 = __pyx_t_13.get_array();
    } catch(...) {
      __Pyx_CppExn2PyErr();
      __PYX_ERR(0, 756, __pyx_L1_error)
    }
    __pyx_v_arrayElem = __PYX_STD_MOVE_IF_SUPPORTED(__pyx_t_14);

    /* "viur/datastore/transport.pyx":757
 *             raise ValueError("No mutation-results received")
 *         arrayElem = element.at_key("mutationResults").get_array()
 *         if arrayElem.size() != abs(len(entities)):             # <<<<<<<<<<<<<<
 *             logging.error(resp.content)
 *             raise NoMutationResultsError("Invalid number of mutation-results received")
 */
    __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_arrayElem.size()); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 757, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_7 = PyObject_Length(__pyx_v_entities); if (unlikely(__pyx_t_7 == ((Py_ssize_t)-1))) __PYX_ERR(0, 757, __pyx_L1_error)
    __pyx_t_1 = PyInt_FromSsize_t(__pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 757, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_6 = __Pyx_PyNumber_Absolute(__pyx_t_1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 757, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = PyObject_RichCompare(__pyx_t_3, __pyx_t_6, Py_NE); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 757, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_11 < 0))) __PYX_ERR(0, 757, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (unlikely(__pyx_t_11)) {

      /* "viur/datastore/transport.pyx":758
 *         arrayElem = element.at_key("mutationResults").get_array()
 *         if arrayElem.size() != abs(len(entities)):
 *             logging.error(resp.content)             # <<<<<<<<<<<<<<
 *             raise NoMutationResultsError("Invalid number of mutation-results received")
 *         arrayIt = arrayElem.begin()
 */
      __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_logging); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 758, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_error); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 758, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_resp, __pyx_n_s_content); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 758, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_10 = NULL;
      __pyx_t_5 = 0;
      #if CYTHON_UNPACK_METHODS
      if (unlikely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_10)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_10);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
          __pyx_t_5 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_10, __pyx_t_6};
        __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
        __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 758, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      }
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "viur/datastore/transport.pyx":759
 *         if arrayElem.size() != abs(len(entities)):
 *             logging.error(resp.content)
 *             raise NoMutationResultsError("Invalid number of mutation-results received")             # <<<<<<<<<<<<<<
 *         arrayIt = arrayElem.begin()
 *         idx = 0
 */
      __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_NoMutationResultsError); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 759, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_6 = NULL;
      __pyx_t_5 = 0;
      #if CYTHON_UNPACK_METHODS
      if (unlikely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_6)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_6);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
          __pyx_t_5 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_6, __pyx_kp_u_Invalid_number_of_mutation_resul};
        __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 759, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      }
      __Pyx_Raise(__pyx_t_1, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __PYX_ERR(0, 759, __pyx_L1_error)

      /* "viur/datastore/transport.pyx":757
 *             raise ValueError("No mutation-results received")
 *         arrayElem = element.at_key("mutationResults").get_array()
 *         if arrayElem.size() != abs(len(entities)):             # <<<<<<<<<<<<<<
 *             logging.error(resp.content)
 *             raise NoMutationResultsError("Invalid number of mutation-results received")
 */
    }

    /* "viur/datastore/transport.pyx":760
 *             logging.error(resp.content)
 *             raise NoMutationResultsError("Invalid number of mutation-results received")
 *         arrayIt = arrayElem.begin()             # <<<<<<<<<<<<<<
 *         idx = 0
 *         while arrayIt != arrayElem.end():
 */
    __pyx_v_arrayIt = __pyx_v_arrayElem.begin();

    /* "viur/datastore/transport.pyx":761
 *             raise NoMutationResultsError("Invalid number of mutation-results received")
 *         arrayIt = arrayElem.begin()
 *         idx = 0             # <<<<<<<<<<<<<<
 *         while arrayIt != arrayElem.end():
 *             innerArrayElem = dereference(arrayIt)
 */
    __Pyx_INCREF(__pyx_int_0);
    __pyx_v_idx = __pyx_int_0;

    /* "viur/datastore/transport.pyx":762
 *         arrayIt = arrayElem.begin()
 *         idx = 0
 *         while arrayIt != arrayElem.end():             # <<<<<<<<<<<<<<
 *             innerArrayElem = dereference(arrayIt)
 *             if innerArrayElem.at_pointer("/key").error() == SUCCESS:  # We got a new key assigned
 */
    while (1) {
      __pyx_t_11 = (__pyx_v_arrayIt != __pyx_v_arrayElem.end());
      if (!__pyx_t_11) break;

      /* "viur/datastore/transport.pyx":763
 *         idx = 0
 *         while arrayIt != arrayElem.end():
 *             innerArrayElem = dereference(arrayIt)             # <<<<<<<<<<<<<<
 *             if innerArrayElem.at_pointer("/key").error() == SUCCESS:  # We got a new key assigned
 *                 entities[idx].key = parseKey(innerArrayElem.at_key("key"))
 */
      __pyx_v_innerArrayElem = (*__pyx_v_arrayIt);

      /* "viur/datastore/transport.pyx":764
 *         while arrayIt != arrayElem.end():
 *             innerArrayElem = dereference(arrayIt)
 *             if innerArrayElem.at_pointer("/key").error() == SUCCESS:  # We got a new key assigned             # <<<<<<<<<<<<<<
 *                 entities[idx].key = parseKey(innerArrayElem.at_key("key"))
 *             entities[idx].version = toPyStr(innerArrayElem.at_key("version").get_string())
 */
      __pyx_t_11 = (__pyx_v_innerArrayElem.at_pointer(((char const *)"/key")).error() == simdjson::error_code::SUCCESS);
      if (__pyx_t_11) {

        /* "viur/datastore/transport.pyx":765
 *             innerArrayElem = dereference(arrayIt)
 *             if innerArrayElem.at_pointer("/key").error() == SUCCESS:  # We got a new key assigned
 *                 entities[idx].key = parseKey(innerArrayElem.at_key("key"))             # <<<<<<<<<<<<<<
 *             entities[idx].version = toPyStr(innerArrayElem.at_key("version").get_string())
 *             preincrement(arrayIt)
 */
        try {
          __pyx_t_13 = __pyx_v_innerArrayElem.at_key(((char const *)"key"));
        } catch(...) {
          __Pyx_CppExn2PyErr();
          __PYX_ERR(0, 765, __pyx_L1_error)
        }
        __pyx_t_1 = __pyx_f_4viur_9datastore_9transport_parseKey(__PYX_STD_MOVE_IF_SUPPORTED(__pyx_t_13)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 765, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_3 = __Pyx_PyObject_GetItem(__pyx_v_entities, __pyx_v_idx); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 765, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        if (__Pyx_PyObject_SetAttrStr(__pyx_t_3, __pyx_n_s_key, __pyx_t_1) < 0) __PYX_ERR(0, 765, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

        /* "viur/datastore/transport.pyx":764
 *         while arrayIt != arrayElem.end():
 *             innerArrayElem = dereference(arrayIt)
 *             if innerArrayElem.at_pointer("/key").error() == SUCCESS:  # We got a new key assigned             # <<<<<<<<<<<<<<
 *                 entities[idx].key = parseKey(innerArrayElem.at_key("key"))
 *             entities[idx].version = toPyStr(innerArrayElem.at_key("version").get_string())
 */
      }

      /* "viur/datastore/transport.pyx":766
 *             if innerArrayElem.at_pointer("/key").error() == SUCCESS:  # We got a new key assigned
 *                 entities[idx].key = parseKey(innerArrayElem.at_key("key"))
 *             entities[idx].version = toPyStr(innerArrayElem.at_key("version").get_string())             # <<<<<<<<<<<<<<
 *             preincrement(arrayIt)
 *             idx += 1
 */
      try {
        __pyx_t_13 = __pyx_v_innerArrayElem.at_key(((char const *)"version"));
      } catch(...) {
        __Pyx_CppExn2PyErr();
        __PYX_ERR(0, 766, __pyx_L1_error)
      }
      try {
        __pyx_t_15 = __pyx_t_13.get_string();
      } catch(...) {
        __Pyx_CppExn2PyErr();
        __PYX_ERR(0, 766, __pyx_L1_error)
      }
      __pyx_t_3 = __pyx_f_4viur_9datastore_9transport_toPyStr(__PYX_STD_MOVE_IF_SUPPORTED(__pyx_t_15)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 766, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_1 = __Pyx_PyObject_GetItem(__pyx_v_entities, __pyx_v_idx); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 766, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (__Pyx_PyObject_SetAttrStr(__pyx_t_1, __pyx_n_s_version, __pyx_t_3) < 0) __PYX_ERR(0, 766, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "viur/datastore/transport.pyx":767
 *                 entities[idx].key = parseKey(innerArrayElem.at_key("key"))
 *             entities[idx].version = toPyStr(innerArrayElem.at_key("version").get_string())
 *             preincrement(arrayIt)             # <<<<<<<<<<<<<<
 *             idx += 1
 *         if conf["memcache_client"] is not None:
 */
      (void)((++__pyx_v_arrayIt));

      /* "viur/datastore/transport.pyx":768
 *             entities[idx].version = toPyStr(innerArrayElem.at_key("version").get_string())
 *             preincrement(arrayIt)
 *             idx += 1             # <<<<<<<<<<<<<<
 *         if conf["memcache_client"] is not None:
 *             # iter over all entities and write them to the cache
 */
      __pyx_t_1 = __Pyx_PyInt_AddObjC(__pyx_v_idx, __pyx_int_1, 1, 1, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 768, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF_SET(__pyx_v_idx, __pyx_t_1);
      __pyx_t_1 = 0;
    }

    /* "viur/datastore/transport.pyx":769
 *             preincrement(arrayIt)
 *             idx += 1
 *         if conf["memcache_client"] is not None:             # <<<<<<<<<<<<<<
 *             # iter over all entities and write them to the cache
 *             cache.put(entities)
 */
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_conf); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 769, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = __Pyx_PyObject_Dict_GetItem(__pyx_t_1, __pyx_n_u_memcache_client); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 769, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_11 = (__pyx_t_3 != Py_None);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (__pyx_t_11) {

      /* "viur/datastore/transport.pyx":771
 *         if conf["memcache_client"] is not None:
 *             # iter over all entities and write them to the cache
 *             cache.put(entities)             # <<<<<<<<<<<<<<
 *     return entities
 * 
 */
      __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_cache); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 771, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_put); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 771, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_1 = NULL;
      __pyx_t_5 = 0;
      #if CYTHON_UNPACK_METHODS
      if (unlikely(PyMethod_Check(__pyx_t_6))) {
        __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_6);
        if (likely(__pyx_t_1)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
          __Pyx_INCREF(__pyx_t_1);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_6, function);
          __pyx_t_5 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_1, __pyx_v_entities};
        __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_6, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
        __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
        if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 771, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

      /* "viur/datastore/transport.pyx":769
 *             preincrement(arrayIt)
 *             idx += 1
 *         if conf["memcache_client"] is not None:             # <<<<<<<<<<<<<<
 *             # iter over all entities and write them to the cache
 *             cache.put(entities)
 */
    }

    /* "viur/datastore/transport.pyx":750
 *     )
 * 
 *     if is_viur_datastore_request_ok(resp):             # <<<<<<<<<<<<<<
 *         assert PyBytes_AsStringAndSize(resp.content, &data_ptr, &pysize) != -1
 *         element = parser.parse(data_ptr, pysize, 1)
 */
  }

  /* "viur/datastore/transport.pyx":772
 *             # iter over all entities and write them to the cache
 *             cache.put(entities)
 *     return entities             # <<<<<<<<<<<<<<
 * 
 * def RunInTransaction(callback: callable, *args, **kwargs) -> Any:
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_entities);
  __pyx_r = __pyx_v_entities;
  goto __pyx_L0;

  /* "viur/datastore/transport.pyx":711
 *         cache.delete([str(key) for key in keys])
 * 
 * def Put(entities: Union[Entity, List[Entity]]) -> Union[Entity, List[Entity]]:             # <<<<<<<<<<<<<<
 *     """
 *         Writes the given entities into the datastore. The entities can be from different kinds. If an entity has an
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_AddTraceback("viur.datastore.transport.Put", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_accessLog);
  __Pyx_XDECREF(__pyx_v_postData);
  __Pyx_XDECREF(__pyx_v_currentTxn);
  __Pyx_XDECREF(__pyx_v_resp);
  __Pyx_XDECREF(__pyx_v_idx);
  __Pyx_XDECREF(__pyx_9genexpr14__pyx_v_x);
  __Pyx_XDECREF(__pyx_9genexpr15__pyx_v_x);
  __Pyx_XDECREF(__pyx_v_entities);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "viur/datastore/transport.pyx":774
 *     return entities
 * 
 * def RunInTransaction(callback: callable, *args, **kwargs) -> Any:             # <<<<<<<<<<<<<<
 *     """
 *         Runs the given function inside a AID transaction.
 */

/* Python wrapper */
static PyObject *__pyx_pw_4viur_9datastore_9transport_15RunInTransaction(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_4viur_9datastore_9transport_14RunInTransaction, "\n        Runs the given function inside a AID transaction.\n\n        :param callback: The function to run inside a transaction\n        :param args: Args to pass to the function\n        :param kwargs: Kwargs to pass to the function\n        :return: The return-value of the callback function\n    ");
static PyMethodDef __pyx_mdef_4viur_9datastore_9transport_15RunInTransaction = {"RunInTransaction", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_4viur_9datastore_9transport_15RunInTransaction, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_4viur_9datastore_9transport_14RunInTransaction};
static PyObject *__pyx_pw_4viur_9datastore_9transport_15RunInTransaction(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_callback = 0;
  PyObject *__pyx_v_args = 0;
  PyObject *__pyx_v_kwargs = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("RunInTransaction (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  __pyx_v_kwargs = PyDict_New(); if (unlikely(!__pyx_v_kwargs)) return NULL;
  __Pyx_GOTREF(__pyx_v_kwargs);
  __pyx_v_args = __Pyx_ArgsSlice_FASTCALL(__pyx_args, 1, __pyx_nargs);
  if (unlikely(!__pyx_v_args)) {
    __Pyx_DECREF(__pyx_v_kwargs); __pyx_v_kwargs = 0;
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __Pyx_GOTREF(__pyx_v_args);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_callback,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        default:
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_callback)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 774, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        const Py_ssize_t used_pos_args = (kwd_pos_args < 1) ? kwd_pos_args : 1;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, __pyx_v_kwargs, values + 0, used_pos_args, "RunInTransaction") < 0)) __PYX_ERR(0, 774, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs < 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_callback = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("RunInTransaction", 0, 1, 1, __pyx_nargs); __PYX_ERR(0, 774, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_DECREF(__pyx_v_args); __pyx_v_args = 0;
  __Pyx_DECREF(__pyx_v_kwargs); __pyx_v_kwargs = 0;
  __Pyx_AddTraceback("viur.datastore.transport.RunInTransaction", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4viur_9datastore_9transport_14RunInTransaction(__pyx_self, __pyx_v_callback, __pyx_v_args, __pyx_v_kwargs);

  /* function exit code */
  __Pyx_DECREF(__pyx_v_args);
  __Pyx_DECREF(__pyx_v_kwargs);
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4viur_9datastore_9transport_14RunInTransaction(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_callback, PyObject *__pyx_v_args, PyObject *__pyx_v_kwargs) {
  simdjson::dom::parser __pyx_v_parser;
  Py_ssize_t __pyx_v_pysize;
  char *__pyx_v_data_ptr;
  simdjson::dom::element __pyx_v_element;
  simdjson::dom::element __pyx_v_innerArrayElem;
  simdjson::dom::array __pyx_v_arrayElem;
  simdjson::dom::array::iterator __pyx_v_arrayIt;
  PyObject *__pyx_v_exponential_backoff = NULL;
  PyObject *__pyx_v_oldTxn = NULL;
  PyObject *__pyx_v_allowOverriding = NULL;
  PyObject *__pyx_v_txnOptions = NULL;
  PyObject *__pyx_v_postData = NULL;
  PyObject *__pyx_v_resp = NULL;
  PyObject *__pyx_v_txnKey = NULL;
  PyObject *__pyx_v_currentTxn = NULL;
  PyObject *__pyx_v_res = NULL;
  PyObject *__pyx_v_idx = NULL;
  PyObject *__pyx_v_affectedEntity = NULL;
  CYTHON_UNUSED PyObject *__pyx_v_err = NULL;
  PyObject *__pyx_v_sleep_time = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  long __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  int __pyx_t_9;
  int __pyx_t_10;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  simdjson::dom::element __pyx_t_17;
  simdjson::dom::array __pyx_t_18;
  Py_ssize_t __pyx_t_19;
  std::string_view __pyx_t_20;
  int __pyx_t_21;
  char const *__pyx_t_22;
  PyObject *__pyx_t_23 = NULL;
  PyObject *__pyx_t_24 = NULL;
  PyObject *__pyx_t_25 = NULL;
  int __pyx_t_26;
  Py_UCS4 __pyx_t_27;
  PyObject *__pyx_t_28 = NULL;
  char const *__pyx_t_29;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("RunInTransaction", 1);

  /* "viur/datastore/transport.pyx":783
 *         :return: The return-value of the callback function
 *     """
 *     cdef simdjsonParser parser = simdjsonParser()             # <<<<<<<<<<<<<<
 *     cdef Py_ssize_t pysize
 *     cdef char * data_ptr
 */
  __pyx_v_parser = simdjson::dom::parser();

  /* "viur/datastore/transport.pyx":789
 *     cdef simdjsonArray arrayElem
 *     cdef simdjsonArray.iterator arrayIt
 *     for exponential_backoff in range(1, 4):             # <<<<<<<<<<<<<<
 *         try:
 *             oldTxn = currentTransaction.get()
 */
  for (__pyx_t_1 = 1; __pyx_t_1 < 4; __pyx_t_1+=1) {
    __pyx_t_2 = __Pyx_PyInt_From_long(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 789, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_XDECREF_SET(__pyx_v_exponential_backoff, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "viur/datastore/transport.pyx":790
 *     cdef simdjsonArray.iterator arrayIt
 *     for exponential_backoff in range(1, 4):
 *         try:             # <<<<<<<<<<<<<<
 *             oldTxn = currentTransaction.get()
 *             allowOverriding = kwargs.pop("__allowOverriding__", None)
 */
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_3, &__pyx_t_4, &__pyx_t_5);
      __Pyx_XGOTREF(__pyx_t_3);
      __Pyx_XGOTREF(__pyx_t_4);
      __Pyx_XGOTREF(__pyx_t_5);
      /*try:*/ {

        /* "viur/datastore/transport.pyx":791
 *     for exponential_backoff in range(1, 4):
 *         try:
 *             oldTxn = currentTransaction.get()             # <<<<<<<<<<<<<<
 *             allowOverriding = kwargs.pop("__allowOverriding__", None)
 *             if oldTxn:
 */
        __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_currentTransaction); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 791, __pyx_L5_error)
        __Pyx_GOTREF(__pyx_t_6);
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_get); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 791, __pyx_L5_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __pyx_t_6 = NULL;
        __pyx_t_8 = 0;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_7))) {
          __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_7);
          if (likely(__pyx_t_6)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
            __Pyx_INCREF(__pyx_t_6);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_7, function);
            __pyx_t_8 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_6, NULL};
          __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_7, __pyx_callargs+1-__pyx_t_8, 0+__pyx_t_8);
          __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
          if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 791, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        }
        __Pyx_XDECREF_SET(__pyx_v_oldTxn, __pyx_t_2);
        __pyx_t_2 = 0;

        /* "viur/datastore/transport.pyx":792
 *         try:
 *             oldTxn = currentTransaction.get()
 *             allowOverriding = kwargs.pop("__allowOverriding__", None)             # <<<<<<<<<<<<<<
 *             if oldTxn:
 *                 if not allowOverriding:
 */
        __pyx_t_2 = __Pyx_PyDict_Pop(__pyx_v_kwargs, __pyx_n_u_allowOverriding, Py_None); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 792, __pyx_L5_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_XDECREF_SET(__pyx_v_allowOverriding, __pyx_t_2);
        __pyx_t_2 = 0;

        /* "viur/datastore/transport.pyx":793
 *             oldTxn = currentTransaction.get()
 *             allowOverriding = kwargs.pop("__allowOverriding__", None)
 *             if oldTxn:             # <<<<<<<<<<<<<<
 *                 if not allowOverriding:
 *                     raise RecursionError("Cannot call runInTransaction while inside a transaction!")
 */
        __pyx_t_9 = __Pyx_PyObject_IsTrue(__pyx_v_oldTxn); if (unlikely((__pyx_t_9 < 0))) __PYX_ERR(0, 793, __pyx_L5_error)
        if (__pyx_t_9) {

          /* "viur/datastore/transport.pyx":794
 *             allowOverriding = kwargs.pop("__allowOverriding__", None)
 *             if oldTxn:
 *                 if not allowOverriding:             # <<<<<<<<<<<<<<
 *                     raise RecursionError("Cannot call runInTransaction while inside a transaction!")
 *                 txnOptions = {
 */
          __pyx_t_9 = __Pyx_PyObject_IsTrue(__pyx_v_allowOverriding); if (unlikely((__pyx_t_9 < 0))) __PYX_ERR(0, 794, __pyx_L5_error)
          __pyx_t_10 = (!__pyx_t_9);
          if (unlikely(__pyx_t_10)) {

            /* "viur/datastore/transport.pyx":795
 *             if oldTxn:
 *                 if not allowOverriding:
 *                     raise RecursionError("Cannot call runInTransaction while inside a transaction!")             # <<<<<<<<<<<<<<
 *                 txnOptions = {
 *                     "previousTransaction": oldTxn["key"]
 */
            __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_RecursionError); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 795, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_2);
            __pyx_t_7 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_tuple__17, NULL); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 795, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_7);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            __Pyx_Raise(__pyx_t_7, 0, 0, 0);
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            __PYX_ERR(0, 795, __pyx_L5_error)

            /* "viur/datastore/transport.pyx":794
 *             allowOverriding = kwargs.pop("__allowOverriding__", None)
 *             if oldTxn:
 *                 if not allowOverriding:             # <<<<<<<<<<<<<<
 *                     raise RecursionError("Cannot call runInTransaction while inside a transaction!")
 *                 txnOptions = {
 */
          }

          /* "viur/datastore/transport.pyx":797
 *                     raise RecursionError("Cannot call runInTransaction while inside a transaction!")
 *                 txnOptions = {
 *                     "previousTransaction": oldTxn["key"]             # <<<<<<<<<<<<<<
 *                 }
 *             else:
 */
          __pyx_t_7 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 797, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_2 = __Pyx_PyObject_Dict_GetItem(__pyx_v_oldTxn, __pyx_n_u_key); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 797, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_2);
          if (PyDict_SetItem(__pyx_t_7, __pyx_n_u_previousTransaction, __pyx_t_2) < 0) __PYX_ERR(0, 797, __pyx_L5_error)
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_XDECREF_SET(__pyx_v_txnOptions, ((PyObject*)__pyx_t_7));
          __pyx_t_7 = 0;

          /* "viur/datastore/transport.pyx":793
 *             oldTxn = currentTransaction.get()
 *             allowOverriding = kwargs.pop("__allowOverriding__", None)
 *             if oldTxn:             # <<<<<<<<<<<<<<
 *                 if not allowOverriding:
 *                     raise RecursionError("Cannot call runInTransaction while inside a transaction!")
 */
          goto __pyx_L13;
        }

        /* "viur/datastore/transport.pyx":800
 *                 }
 *             else:
 *                 txnOptions = {}             # <<<<<<<<<<<<<<
 *             postData = {
 *                 "transactionOptions": {
 */
        /*else*/ {
          __pyx_t_7 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 800, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_XDECREF_SET(__pyx_v_txnOptions, ((PyObject*)__pyx_t_7));
          __pyx_t_7 = 0;
        }
        __pyx_L13:;

        /* "viur/datastore/transport.pyx":802
 *                 txnOptions = {}
 *             postData = {
 *                 "transactionOptions": {             # <<<<<<<<<<<<<<
 *                     "readWrite": txnOptions
 *                 }
 */
        __pyx_t_7 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 802, __pyx_L5_error)
        __Pyx_GOTREF(__pyx_t_7);

        /* "viur/datastore/transport.pyx":803
 *             postData = {
 *                 "transactionOptions": {
 *                     "readWrite": txnOptions             # <<<<<<<<<<<<<<
 *                 }
 *             }
 */
        __pyx_t_2 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 803, __pyx_L5_error)
        __Pyx_GOTREF(__pyx_t_2);
        if (PyDict_SetItem(__pyx_t_2, __pyx_n_u_readWrite, __pyx_v_txnOptions) < 0) __PYX_ERR(0, 803, __pyx_L5_error)
        if (PyDict_SetItem(__pyx_t_7, __pyx_n_u_transactionOptions, __pyx_t_2) < 0) __PYX_ERR(0, 802, __pyx_L5_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_XDECREF_SET(__pyx_v_postData, ((PyObject*)__pyx_t_7));
        __pyx_t_7 = 0;

        /* "viur/datastore/transport.pyx":806
 *                 }
 *             }
 *             resp = authenticated_request(             # <<<<<<<<<<<<<<
 *                 url="https://datastore.googleapis.com/v1/projects/%s:beginTransaction" % projectID,
 *                 data=json.dumps(postData).encode("UTF-8"),
 */
        __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_authenticated_request); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 806, __pyx_L5_error)
        __Pyx_GOTREF(__pyx_t_7);

        /* "viur/datastore/transport.pyx":807
 *             }
 *             resp = authenticated_request(
 *                 url="https://datastore.googleapis.com/v1/projects/%s:beginTransaction" % projectID,             # <<<<<<<<<<<<<<
 *                 data=json.dumps(postData).encode("UTF-8"),
 *             )
 */
        __pyx_t_2 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 807, __pyx_L5_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_projectID); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 807, __pyx_L5_error)
        __Pyx_GOTREF(__pyx_t_6);
        __pyx_t_11 = __Pyx_PyUnicode_FormatSafe(__pyx_kp_u_https_datastore_googleapis_com_v_4, __pyx_t_6); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 807, __pyx_L5_error)
        __Pyx_GOTREF(__pyx_t_11);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_url, __pyx_t_11) < 0) __PYX_ERR(0, 807, __pyx_L5_error)
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

        /* "viur/datastore/transport.pyx":808
 *             resp = authenticated_request(
 *                 url="https://datastore.googleapis.com/v1/projects/%s:beginTransaction" % projectID,
 *                 data=json.dumps(postData).encode("UTF-8"),             # <<<<<<<<<<<<<<
 *             )
 *             if is_viur_datastore_request_ok(resp):
 */
        __Pyx_GetModuleGlobalName(__pyx_t_12, __pyx_n_s_json); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 808, __pyx_L5_error)
        __Pyx_GOTREF(__pyx_t_12);
        __pyx_t_13 = __Pyx_PyObject_GetAttrStr(__pyx_t_12, __pyx_n_s_dumps); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 808, __pyx_L5_error)
        __Pyx_GOTREF(__pyx_t_13);
        __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
        __pyx_t_12 = NULL;
        __pyx_t_8 = 0;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_13))) {
          __pyx_t_12 = PyMethod_GET_SELF(__pyx_t_13);
          if (likely(__pyx_t_12)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_13);
            __Pyx_INCREF(__pyx_t_12);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_13, function);
            __pyx_t_8 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_12, __pyx_v_postData};
          __pyx_t_6 = __Pyx_PyObject_FastCall(__pyx_t_13, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
          __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
          if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 808, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_6);
          __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
        }
        __pyx_t_13 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_encode); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 808, __pyx_L5_error)
        __Pyx_GOTREF(__pyx_t_13);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __pyx_t_6 = NULL;
        __pyx_t_8 = 0;
        #if CYTHON_UNPACK_METHODS
        if (likely(PyMethod_Check(__pyx_t_13))) {
          __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_13);
          if (likely(__pyx_t_6)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_13);
            __Pyx_INCREF(__pyx_t_6);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_13, function);
            __pyx_t_8 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_6, __pyx_kp_u_UTF_8};
          __pyx_t_11 = __Pyx_PyObject_FastCall(__pyx_t_13, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
          __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
          if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 808, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_11);
          __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
        }
        if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_data, __pyx_t_11) < 0) __PYX_ERR(0, 807, __pyx_L5_error)
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

        /* "viur/datastore/transport.pyx":806
 *                 }
 *             }
 *             resp = authenticated_request(             # <<<<<<<<<<<<<<
 *                 url="https://datastore.googleapis.com/v1/projects/%s:beginTransaction" % projectID,
 *                 data=json.dumps(postData).encode("UTF-8"),
 */
        __pyx_t_11 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_empty_tuple, __pyx_t_2); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 806, __pyx_L5_error)
        __Pyx_GOTREF(__pyx_t_11);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_XDECREF_SET(__pyx_v_resp, __pyx_t_11);
        __pyx_t_11 = 0;

        /* "viur/datastore/transport.pyx":810
 *                 data=json.dumps(postData).encode("UTF-8"),
 *             )
 *             if is_viur_datastore_request_ok(resp):             # <<<<<<<<<<<<<<
 *                 txnKey = json.loads(resp.content)["transaction"]
 *                 try:
 */
        __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_is_viur_datastore_request_ok); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 810, __pyx_L5_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_7 = NULL;
        __pyx_t_8 = 0;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_2))) {
          __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_2);
          if (likely(__pyx_t_7)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
            __Pyx_INCREF(__pyx_t_7);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_2, function);
            __pyx_t_8 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_v_resp};
          __pyx_t_11 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
          __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
          if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 810, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_11);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        }
        __pyx_t_10 = __Pyx_PyObject_IsTrue(__pyx_t_11); if (unlikely((__pyx_t_10 < 0))) __PYX_ERR(0, 810, __pyx_L5_error)
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
        if (__pyx_t_10) {

          /* "viur/datastore/transport.pyx":811
 *             )
 *             if is_viur_datastore_request_ok(resp):
 *                 txnKey = json.loads(resp.content)["transaction"]             # <<<<<<<<<<<<<<
 *                 try:
 *                     currentTxn = {"key": txnKey, "mutations": [], "affectedEntities": []}
 */
          __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_json); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 811, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_loads); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 811, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_resp, __pyx_n_s_content); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 811, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_13 = NULL;
          __pyx_t_8 = 0;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_7))) {
            __pyx_t_13 = PyMethod_GET_SELF(__pyx_t_7);
            if (likely(__pyx_t_13)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
              __Pyx_INCREF(__pyx_t_13);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_7, function);
              __pyx_t_8 = 1;
            }
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_13, __pyx_t_2};
            __pyx_t_11 = __Pyx_PyObject_FastCall(__pyx_t_7, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
            __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 811, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_11);
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          }
          __pyx_t_7 = __Pyx_PyObject_Dict_GetItem(__pyx_t_11, __pyx_n_u_transaction); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 811, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
          __Pyx_XDECREF_SET(__pyx_v_txnKey, __pyx_t_7);
          __pyx_t_7 = 0;

          /* "viur/datastore/transport.pyx":812
 *             if is_viur_datastore_request_ok(resp):
 *                 txnKey = json.loads(resp.content)["transaction"]
 *                 try:             # <<<<<<<<<<<<<<
 *                     currentTxn = {"key": txnKey, "mutations": [], "affectedEntities": []}
 *                     currentTransaction.set(currentTxn)
 */
          /*try:*/ {

            /* "viur/datastore/transport.pyx":813
 *                 txnKey = json.loads(resp.content)["transaction"]
 *                 try:
 *                     currentTxn = {"key": txnKey, "mutations": [], "affectedEntities": []}             # <<<<<<<<<<<<<<
 *                     currentTransaction.set(currentTxn)
 *                     try:
 */
            __pyx_t_7 = __Pyx_PyDict_NewPresized(3); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 813, __pyx_L19_error)
            __Pyx_GOTREF(__pyx_t_7);
            if (PyDict_SetItem(__pyx_t_7, __pyx_n_u_key, __pyx_v_txnKey) < 0) __PYX_ERR(0, 813, __pyx_L19_error)
            __pyx_t_11 = PyList_New(0); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 813, __pyx_L19_error)
            __Pyx_GOTREF(__pyx_t_11);
            if (PyDict_SetItem(__pyx_t_7, __pyx_n_u_mutations, __pyx_t_11) < 0) __PYX_ERR(0, 813, __pyx_L19_error)
            __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
            __pyx_t_11 = PyList_New(0); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 813, __pyx_L19_error)
            __Pyx_GOTREF(__pyx_t_11);
            if (PyDict_SetItem(__pyx_t_7, __pyx_n_u_affectedEntities, __pyx_t_11) < 0) __PYX_ERR(0, 813, __pyx_L19_error)
            __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
            __Pyx_XDECREF_SET(__pyx_v_currentTxn, ((PyObject*)__pyx_t_7));
            __pyx_t_7 = 0;

            /* "viur/datastore/transport.pyx":814
 *                 try:
 *                     currentTxn = {"key": txnKey, "mutations": [], "affectedEntities": []}
 *                     currentTransaction.set(currentTxn)             # <<<<<<<<<<<<<<
 *                     try:
 *                         res = callback(*args, **kwargs)
 */
            __Pyx_GetModuleGlobalName(__pyx_t_11, __pyx_n_s_currentTransaction); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 814, __pyx_L19_error)
            __Pyx_GOTREF(__pyx_t_11);
            __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_11, __pyx_n_s_set); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 814, __pyx_L19_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
            __pyx_t_11 = NULL;
            __pyx_t_8 = 0;
            #if CYTHON_UNPACK_METHODS
            if (unlikely(PyMethod_Check(__pyx_t_2))) {
              __pyx_t_11 = PyMethod_GET_SELF(__pyx_t_2);
              if (likely(__pyx_t_11)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
                __Pyx_INCREF(__pyx_t_11);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_2, function);
                __pyx_t_8 = 1;
              }
            }
            #endif
            {
              PyObject *__pyx_callargs[2] = {__pyx_t_11, __pyx_v_currentTxn};
              __pyx_t_7 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
              __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
              if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 814, __pyx_L19_error)
              __Pyx_GOTREF(__pyx_t_7);
              __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            }
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

            /* "viur/datastore/transport.pyx":815
 *                     currentTxn = {"key": txnKey, "mutations": [], "affectedEntities": []}
 *                     currentTransaction.set(currentTxn)
 *                     try:             # <<<<<<<<<<<<<<
 *                         res = callback(*args, **kwargs)
 *                     except:
 */
            {
              __Pyx_PyThreadState_declare
              __Pyx_PyThreadState_assign
              __Pyx_ExceptionSave(&__pyx_t_14, &__pyx_t_15, &__pyx_t_16);
              __Pyx_XGOTREF(__pyx_t_14);
              __Pyx_XGOTREF(__pyx_t_15);
              __Pyx_XGOTREF(__pyx_t_16);
              /*try:*/ {

                /* "viur/datastore/transport.pyx":816
 *                     currentTransaction.set(currentTxn)
 *                     try:
 *                         res = callback(*args, **kwargs)             # <<<<<<<<<<<<<<
 *                     except:
 *                         _rollbackTxn(txnKey)
 */
                __pyx_t_7 = PyDict_Copy(__pyx_v_kwargs); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 816, __pyx_L21_error)
                __Pyx_GOTREF(__pyx_t_7);
                __pyx_t_2 = __Pyx_PyObject_Call(__pyx_v_callback, __pyx_v_args, __pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 816, __pyx_L21_error)
                __Pyx_GOTREF(__pyx_t_2);
                __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
                __Pyx_XDECREF_SET(__pyx_v_res, __pyx_t_2);
                __pyx_t_2 = 0;

                /* "viur/datastore/transport.pyx":815
 *                     currentTxn = {"key": txnKey, "mutations": [], "affectedEntities": []}
 *                     currentTransaction.set(currentTxn)
 *                     try:             # <<<<<<<<<<<<<<
 *                         res = callback(*args, **kwargs)
 *                     except:
 */
              }
              __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
              __Pyx_XDECREF(__pyx_t_15); __pyx_t_15 = 0;
              __Pyx_XDECREF(__pyx_t_16); __pyx_t_16 = 0;
              goto __pyx_L28_try_end;
              __pyx_L21_error:;
              __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
              __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
              __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
              __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
              __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
              __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;

              /* "viur/datastore/transport.pyx":817
 *                     try:
 *                         res = callback(*args, **kwargs)
 *                     except:             # <<<<<<<<<<<<<<
 *                         _rollbackTxn(txnKey)
 *                         raise
 */
              /*except:*/ {
                __Pyx_AddTraceback("viur.datastore.transport.RunInTransaction", __pyx_clineno, __pyx_lineno, __pyx_filename);
                if (__Pyx_GetException(&__pyx_t_2, &__pyx_t_7, &__pyx_t_11) < 0) __PYX_ERR(0, 817, __pyx_L23_except_error)
                __Pyx_XGOTREF(__pyx_t_2);
                __Pyx_XGOTREF(__pyx_t_7);
                __Pyx_XGOTREF(__pyx_t_11);

                /* "viur/datastore/transport.pyx":818
 *                         res = callback(*args, **kwargs)
 *                     except:
 *                         _rollbackTxn(txnKey)             # <<<<<<<<<<<<<<
 *                         raise
 *                     if currentTxn["mutations"]:
 */
                __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_rollbackTxn); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 818, __pyx_L23_except_error)
                __Pyx_GOTREF(__pyx_t_6);
                __pyx_t_12 = NULL;
                __pyx_t_8 = 0;
                #if CYTHON_UNPACK_METHODS
                if (unlikely(PyMethod_Check(__pyx_t_6))) {
                  __pyx_t_12 = PyMethod_GET_SELF(__pyx_t_6);
                  if (likely(__pyx_t_12)) {
                    PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
                    __Pyx_INCREF(__pyx_t_12);
                    __Pyx_INCREF(function);
                    __Pyx_DECREF_SET(__pyx_t_6, function);
                    __pyx_t_8 = 1;
                  }
                }
                #endif
                {
                  PyObject *__pyx_callargs[2] = {__pyx_t_12, __pyx_v_txnKey};
                  __pyx_t_13 = __Pyx_PyObject_FastCall(__pyx_t_6, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
                  __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
                  if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 818, __pyx_L23_except_error)
                  __Pyx_GOTREF(__pyx_t_13);
                  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
                }
                __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;

                /* "viur/datastore/transport.pyx":819
 *                     except:
 *                         _rollbackTxn(txnKey)
 *                         raise             # <<<<<<<<<<<<<<
 *                     if currentTxn["mutations"]:
 *                         # Commit TXN
 */
                __Pyx_GIVEREF(__pyx_t_2);
                __Pyx_GIVEREF(__pyx_t_7);
                __Pyx_XGIVEREF(__pyx_t_11);
                __Pyx_ErrRestoreWithState(__pyx_t_2, __pyx_t_7, __pyx_t_11);
                __pyx_t_2 = 0; __pyx_t_7 = 0; __pyx_t_11 = 0; 
                __PYX_ERR(0, 819, __pyx_L23_except_error)
              }

              /* "viur/datastore/transport.pyx":815
 *                     currentTxn = {"key": txnKey, "mutations": [], "affectedEntities": []}
 *                     currentTransaction.set(currentTxn)
 *                     try:             # <<<<<<<<<<<<<<
 *                         res = callback(*args, **kwargs)
 *                     except:
 */
              __pyx_L23_except_error:;
              __Pyx_XGIVEREF(__pyx_t_14);
              __Pyx_XGIVEREF(__pyx_t_15);
              __Pyx_XGIVEREF(__pyx_t_16);
              __Pyx_ExceptionReset(__pyx_t_14, __pyx_t_15, __pyx_t_16);
              goto __pyx_L19_error;
              __pyx_L28_try_end:;
            }

            /* "viur/datastore/transport.pyx":820
 *                         _rollbackTxn(txnKey)
 *                         raise
 *                     if currentTxn["mutations"]:             # <<<<<<<<<<<<<<
 *                         # Commit TXN
 *                         postData = {
 */
            __pyx_t_11 = __Pyx_PyDict_GetItem(__pyx_v_currentTxn, __pyx_n_u_mutations); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 820, __pyx_L19_error)
            __Pyx_GOTREF(__pyx_t_11);
            __pyx_t_10 = __Pyx_PyObject_IsTrue(__pyx_t_11); if (unlikely((__pyx_t_10 < 0))) __PYX_ERR(0, 820, __pyx_L19_error)
            __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
            if (__pyx_t_10) {

              /* "viur/datastore/transport.pyx":823
 *                         # Commit TXN
 *                         postData = {
 *                             "mode": "TRANSACTIONAL",  #             # <<<<<<<<<<<<<<
 *                             "transaction": txnKey,
 *                             "mutations": currentTxn["mutations"]
 */
              __pyx_t_11 = __Pyx_PyDict_NewPresized(3); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 823, __pyx_L19_error)
              __Pyx_GOTREF(__pyx_t_11);
              if (PyDict_SetItem(__pyx_t_11, __pyx_n_u_mode, __pyx_n_u_TRANSACTIONAL) < 0) __PYX_ERR(0, 823, __pyx_L19_error)

              /* "viur/datastore/transport.pyx":824
 *                         postData = {
 *                             "mode": "TRANSACTIONAL",  #
 *                             "transaction": txnKey,             # <<<<<<<<<<<<<<
 *                             "mutations": currentTxn["mutations"]
 *                         }
 */
              if (PyDict_SetItem(__pyx_t_11, __pyx_n_u_transaction, __pyx_v_txnKey) < 0) __PYX_ERR(0, 823, __pyx_L19_error)

              /* "viur/datastore/transport.pyx":825
 *                             "mode": "TRANSACTIONAL",  #
 *                             "transaction": txnKey,
 *                             "mutations": currentTxn["mutations"]             # <<<<<<<<<<<<<<
 *                         }
 *                         resp = authenticated_request(
 */
              __pyx_t_7 = __Pyx_PyDict_GetItem(__pyx_v_currentTxn, __pyx_n_u_mutations); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 825, __pyx_L19_error)
              __Pyx_GOTREF(__pyx_t_7);
              if (PyDict_SetItem(__pyx_t_11, __pyx_n_u_mutations, __pyx_t_7) < 0) __PYX_ERR(0, 823, __pyx_L19_error)
              __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
              __Pyx_DECREF_SET(__pyx_v_postData, ((PyObject*)__pyx_t_11));
              __pyx_t_11 = 0;

              /* "viur/datastore/transport.pyx":827
 *                             "mutations": currentTxn["mutations"]
 *                         }
 *                         resp = authenticated_request(             # <<<<<<<<<<<<<<
 *                             url="https://datastore.googleapis.com/v1/projects/%s:commit" % projectID,
 *                             data=json.dumps(postData).encode("UTF-8"),
 */
              __Pyx_GetModuleGlobalName(__pyx_t_11, __pyx_n_s_authenticated_request); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 827, __pyx_L19_error)
              __Pyx_GOTREF(__pyx_t_11);

              /* "viur/datastore/transport.pyx":828
 *                         }
 *                         resp = authenticated_request(
 *                             url="https://datastore.googleapis.com/v1/projects/%s:commit" % projectID,             # <<<<<<<<<<<<<<
 *                             data=json.dumps(postData).encode("UTF-8"),
 *                         )
 */
              __pyx_t_7 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 828, __pyx_L19_error)
              __Pyx_GOTREF(__pyx_t_7);
              __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_projectID); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 828, __pyx_L19_error)
              __Pyx_GOTREF(__pyx_t_2);
              __pyx_t_13 = __Pyx_PyUnicode_FormatSafe(__pyx_kp_u_https_datastore_googleapis_com_v_3, __pyx_t_2); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 828, __pyx_L19_error)
              __Pyx_GOTREF(__pyx_t_13);
              __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
              if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_url, __pyx_t_13) < 0) __PYX_ERR(0, 828, __pyx_L19_error)
              __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;

              /* "viur/datastore/transport.pyx":829
 *                         resp = authenticated_request(
 *                             url="https://datastore.googleapis.com/v1/projects/%s:commit" % projectID,
 *                             data=json.dumps(postData).encode("UTF-8"),             # <<<<<<<<<<<<<<
 *                         )
 * 
 */
              __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_json); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 829, __pyx_L19_error)
              __Pyx_GOTREF(__pyx_t_6);
              __pyx_t_12 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_dumps); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 829, __pyx_L19_error)
              __Pyx_GOTREF(__pyx_t_12);
              __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
              __pyx_t_6 = NULL;
              __pyx_t_8 = 0;
              #if CYTHON_UNPACK_METHODS
              if (unlikely(PyMethod_Check(__pyx_t_12))) {
                __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_12);
                if (likely(__pyx_t_6)) {
                  PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_12);
                  __Pyx_INCREF(__pyx_t_6);
                  __Pyx_INCREF(function);
                  __Pyx_DECREF_SET(__pyx_t_12, function);
                  __pyx_t_8 = 1;
                }
              }
              #endif
              {
                PyObject *__pyx_callargs[2] = {__pyx_t_6, __pyx_v_postData};
                __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_12, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
                __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
                if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 829, __pyx_L19_error)
                __Pyx_GOTREF(__pyx_t_2);
                __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
              }
              __pyx_t_12 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_encode); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 829, __pyx_L19_error)
              __Pyx_GOTREF(__pyx_t_12);
              __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
              __pyx_t_2 = NULL;
              __pyx_t_8 = 0;
              #if CYTHON_UNPACK_METHODS
              if (likely(PyMethod_Check(__pyx_t_12))) {
                __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_12);
                if (likely(__pyx_t_2)) {
                  PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_12);
                  __Pyx_INCREF(__pyx_t_2);
                  __Pyx_INCREF(function);
                  __Pyx_DECREF_SET(__pyx_t_12, function);
                  __pyx_t_8 = 1;
                }
              }
              #endif
              {
                PyObject *__pyx_callargs[2] = {__pyx_t_2, __pyx_kp_u_UTF_8};
                __pyx_t_13 = __Pyx_PyObject_FastCall(__pyx_t_12, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
                __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
                if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 829, __pyx_L19_error)
                __Pyx_GOTREF(__pyx_t_13);
                __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
              }
              if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_data, __pyx_t_13) < 0) __PYX_ERR(0, 828, __pyx_L19_error)
              __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;

              /* "viur/datastore/transport.pyx":827
 *                             "mutations": currentTxn["mutations"]
 *                         }
 *                         resp = authenticated_request(             # <<<<<<<<<<<<<<
 *                             url="https://datastore.googleapis.com/v1/projects/%s:commit" % projectID,
 *                             data=json.dumps(postData).encode("UTF-8"),
 */
              __pyx_t_13 = __Pyx_PyObject_Call(__pyx_t_11, __pyx_empty_tuple, __pyx_t_7); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 827, __pyx_L19_error)
              __Pyx_GOTREF(__pyx_t_13);
              __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
              __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
              __Pyx_DECREF_SET(__pyx_v_resp, __pyx_t_13);
              __pyx_t_13 = 0;

              /* "viur/datastore/transport.pyx":832
 *                         )
 * 
 *                         is_viur_datastore_request_ok(resp)             # <<<<<<<<<<<<<<
 *                         assert PyBytes_AsStringAndSize(resp.content, &data_ptr, &pysize) != -1
 *                         element = parser.parse(data_ptr, pysize, 1)
 */
              __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_is_viur_datastore_request_ok); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 832, __pyx_L19_error)
              __Pyx_GOTREF(__pyx_t_7);
              __pyx_t_11 = NULL;
              __pyx_t_8 = 0;
              #if CYTHON_UNPACK_METHODS
              if (unlikely(PyMethod_Check(__pyx_t_7))) {
                __pyx_t_11 = PyMethod_GET_SELF(__pyx_t_7);
                if (likely(__pyx_t_11)) {
                  PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
                  __Pyx_INCREF(__pyx_t_11);
                  __Pyx_INCREF(function);
                  __Pyx_DECREF_SET(__pyx_t_7, function);
                  __pyx_t_8 = 1;
                }
              }
              #endif
              {
                PyObject *__pyx_callargs[2] = {__pyx_t_11, __pyx_v_resp};
                __pyx_t_13 = __Pyx_PyObject_FastCall(__pyx_t_7, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
                __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
                if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 832, __pyx_L19_error)
                __Pyx_GOTREF(__pyx_t_13);
                __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
              }
              __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;

              /* "viur/datastore/transport.pyx":833
 * 
 *                         is_viur_datastore_request_ok(resp)
 *                         assert PyBytes_AsStringAndSize(resp.content, &data_ptr, &pysize) != -1             # <<<<<<<<<<<<<<
 *                         element = parser.parse(data_ptr, pysize, 1)
 *                         if (element.at_pointer("/mutationResults").error() != SUCCESS):
 */
              #ifndef CYTHON_WITHOUT_ASSERTIONS
              if (unlikely(__pyx_assertions_enabled())) {
                __pyx_t_13 = __Pyx_PyObject_GetAttrStr(__pyx_v_resp, __pyx_n_s_content); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 833, __pyx_L19_error)
                __Pyx_GOTREF(__pyx_t_13);
                __pyx_t_8 = PyBytes_AsStringAndSize(__pyx_t_13, (&__pyx_v_data_ptr), (&__pyx_v_pysize)); if (unlikely(__pyx_t_8 == ((int)-1))) __PYX_ERR(0, 833, __pyx_L19_error)
                __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
                __pyx_t_10 = (__pyx_t_8 != -1L);
                if (unlikely(!__pyx_t_10)) {
                  __Pyx_Raise(__pyx_builtin_AssertionError, 0, 0, 0);
                  __PYX_ERR(0, 833, __pyx_L19_error)
                }
              }
              #else
              if ((1)); else __PYX_ERR(0, 833, __pyx_L19_error)
              #endif

              /* "viur/datastore/transport.pyx":834
 *                         is_viur_datastore_request_ok(resp)
 *                         assert PyBytes_AsStringAndSize(resp.content, &data_ptr, &pysize) != -1
 *                         element = parser.parse(data_ptr, pysize, 1)             # <<<<<<<<<<<<<<
 *                         if (element.at_pointer("/mutationResults").error() != SUCCESS):
 *                             logging.error(resp.content)
 */
              try {
                __pyx_t_17 = __pyx_v_parser.parse(__pyx_v_data_ptr, __pyx_v_pysize, 1);
              } catch(...) {
                __Pyx_CppExn2PyErr();
                __PYX_ERR(0, 834, __pyx_L19_error)
              }
              __pyx_v_element = __PYX_STD_MOVE_IF_SUPPORTED(__pyx_t_17);

              /* "viur/datastore/transport.pyx":835
 *                         assert PyBytes_AsStringAndSize(resp.content, &data_ptr, &pysize) != -1
 *                         element = parser.parse(data_ptr, pysize, 1)
 *                         if (element.at_pointer("/mutationResults").error() != SUCCESS):             # <<<<<<<<<<<<<<
 *                             logging.error(resp.content)
 *                             raise NoMutationResultsError("No mutation-results received")
 */
              __pyx_t_10 = (__pyx_v_element.at_pointer(((char const *)"/mutationResults")).error() != simdjson::error_code::SUCCESS);
              if (unlikely(__pyx_t_10)) {

                /* "viur/datastore/transport.pyx":836
 *                         element = parser.parse(data_ptr, pysize, 1)
 *                         if (element.at_pointer("/mutationResults").error() != SUCCESS):
 *                             logging.error(resp.content)             # <<<<<<<<<<<<<<
 *                             raise NoMutationResultsError("No mutation-results received")
 *                         arrayElem = element.at_key("mutationResults").get_array()
 */
                __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_logging); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 836, __pyx_L19_error)
                __Pyx_GOTREF(__pyx_t_7);
                __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_error); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 836, __pyx_L19_error)
                __Pyx_GOTREF(__pyx_t_11);
                __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
                __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_resp, __pyx_n_s_content); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 836, __pyx_L19_error)
                __Pyx_GOTREF(__pyx_t_7);
                __pyx_t_12 = NULL;
                __pyx_t_8 = 0;
                #if CYTHON_UNPACK_METHODS
                if (unlikely(PyMethod_Check(__pyx_t_11))) {
                  __pyx_t_12 = PyMethod_GET_SELF(__pyx_t_11);
                  if (likely(__pyx_t_12)) {
                    PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_11);
                    __Pyx_INCREF(__pyx_t_12);
                    __Pyx_INCREF(function);
                    __Pyx_DECREF_SET(__pyx_t_11, function);
                    __pyx_t_8 = 1;
                  }
                }
                #endif
                {
                  PyObject *__pyx_callargs[2] = {__pyx_t_12, __pyx_t_7};
                  __pyx_t_13 = __Pyx_PyObject_FastCall(__pyx_t_11, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
                  __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
                  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
                  if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 836, __pyx_L19_error)
                  __Pyx_GOTREF(__pyx_t_13);
                  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
                }
                __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;

                /* "viur/datastore/transport.pyx":837
 *                         if (element.at_pointer("/mutationResults").error() != SUCCESS):
 *                             logging.error(resp.content)
 *                             raise NoMutationResultsError("No mutation-results received")             # <<<<<<<<<<<<<<
 *                         arrayElem = element.at_key("mutationResults").get_array()
 *                         if arrayElem.size() != abs(len(currentTxn["affectedEntities"])):
 */
                __Pyx_GetModuleGlobalName(__pyx_t_11, __pyx_n_s_NoMutationResultsError); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 837, __pyx_L19_error)
                __Pyx_GOTREF(__pyx_t_11);
                __pyx_t_7 = NULL;
                __pyx_t_8 = 0;
                #if CYTHON_UNPACK_METHODS
                if (unlikely(PyMethod_Check(__pyx_t_11))) {
                  __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_11);
                  if (likely(__pyx_t_7)) {
                    PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_11);
                    __Pyx_INCREF(__pyx_t_7);
                    __Pyx_INCREF(function);
                    __Pyx_DECREF_SET(__pyx_t_11, function);
                    __pyx_t_8 = 1;
                  }
                }
                #endif
                {
                  PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_kp_u_No_mutation_results_received_2};
                  __pyx_t_13 = __Pyx_PyObject_FastCall(__pyx_t_11, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
                  __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
                  if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 837, __pyx_L19_error)
                  __Pyx_GOTREF(__pyx_t_13);
                  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
                }
                __Pyx_Raise(__pyx_t_13, 0, 0, 0);
                __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
                __PYX_ERR(0, 837, __pyx_L19_error)

                /* "viur/datastore/transport.pyx":835
 *                         assert PyBytes_AsStringAndSize(resp.content, &data_ptr, &pysize) != -1
 *                         element = parser.parse(data_ptr, pysize, 1)
 *                         if (element.at_pointer("/mutationResults").error() != SUCCESS):             # <<<<<<<<<<<<<<
 *                             logging.error(resp.content)
 *                             raise NoMutationResultsError("No mutation-results received")
 */
              }

              /* "viur/datastore/transport.pyx":838
 *                             logging.error(resp.content)
 *                             raise NoMutationResultsError("No mutation-results received")
 *                         arrayElem = element.at_key("mutationResults").get_array()             # <<<<<<<<<<<<<<
 *                         if arrayElem.size() != abs(len(currentTxn["affectedEntities"])):
 *                             logging.error(resp.content)
 */
              try {
                __pyx_t_17 = __pyx_v_element.at_key(((char const *)"mutationResults"));
              } catch(...) {
                __Pyx_CppExn2PyErr();
                __PYX_ERR(0, 838, __pyx_L19_error)
              }
              try {
                __pyx_t_18 = __pyx_t_17.get_array();
              } catch(...) {
                __Pyx_CppExn2PyErr();
                __PYX_ERR(0, 838, __pyx_L19_error)
              }
              __pyx_v_arrayElem = __PYX_STD_MOVE_IF_SUPPORTED(__pyx_t_18);

              /* "viur/datastore/transport.pyx":839
 *                             raise NoMutationResultsError("No mutation-results received")
 *                         arrayElem = element.at_key("mutationResults").get_array()
 *                         if arrayElem.size() != abs(len(currentTxn["affectedEntities"])):             # <<<<<<<<<<<<<<
 *                             logging.error(resp.content)
 *                             raise ViurDatastoreError("Invalid number of mutation-results received")
 */
              __pyx_t_13 = __Pyx_PyInt_From_int(__pyx_v_arrayElem.size()); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 839, __pyx_L19_error)
              __Pyx_GOTREF(__pyx_t_13);
              __pyx_t_11 = __Pyx_PyDict_GetItem(__pyx_v_currentTxn, __pyx_n_u_affectedEntities); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 839, __pyx_L19_error)
              __Pyx_GOTREF(__pyx_t_11);
              __pyx_t_19 = PyObject_Length(__pyx_t_11); if (unlikely(__pyx_t_19 == ((Py_ssize_t)-1))) __PYX_ERR(0, 839, __pyx_L19_error)
              __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
              __pyx_t_11 = PyInt_FromSsize_t(__pyx_t_19); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 839, __pyx_L19_error)
              __Pyx_GOTREF(__pyx_t_11);
              __pyx_t_7 = __Pyx_PyNumber_Absolute(__pyx_t_11); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 839, __pyx_L19_error)
              __Pyx_GOTREF(__pyx_t_7);
              __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
              __pyx_t_11 = PyObject_RichCompare(__pyx_t_13, __pyx_t_7, Py_NE); __Pyx_XGOTREF(__pyx_t_11); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 839, __pyx_L19_error)
              __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
              __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
              __pyx_t_10 = __Pyx_PyObject_IsTrue(__pyx_t_11); if (unlikely((__pyx_t_10 < 0))) __PYX_ERR(0, 839, __pyx_L19_error)
              __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
              if (unlikely(__pyx_t_10)) {

                /* "viur/datastore/transport.pyx":840
 *                         arrayElem = element.at_key("mutationResults").get_array()
 *                         if arrayElem.size() != abs(len(currentTxn["affectedEntities"])):
 *                             logging.error(resp.content)             # <<<<<<<<<<<<<<
 *                             raise ViurDatastoreError("Invalid number of mutation-results received")
 *                         arrayIt = arrayElem.begin()
 */
                __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_logging); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 840, __pyx_L19_error)
                __Pyx_GOTREF(__pyx_t_7);
                __pyx_t_13 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_error); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 840, __pyx_L19_error)
                __Pyx_GOTREF(__pyx_t_13);
                __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
                __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_resp, __pyx_n_s_content); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 840, __pyx_L19_error)
                __Pyx_GOTREF(__pyx_t_7);
                __pyx_t_12 = NULL;
                __pyx_t_8 = 0;
                #if CYTHON_UNPACK_METHODS
                if (unlikely(PyMethod_Check(__pyx_t_13))) {
                  __pyx_t_12 = PyMethod_GET_SELF(__pyx_t_13);
                  if (likely(__pyx_t_12)) {
                    PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_13);
                    __Pyx_INCREF(__pyx_t_12);
                    __Pyx_INCREF(function);
                    __Pyx_DECREF_SET(__pyx_t_13, function);
                    __pyx_t_8 = 1;
                  }
                }
                #endif
                {
                  PyObject *__pyx_callargs[2] = {__pyx_t_12, __pyx_t_7};
                  __pyx_t_11 = __Pyx_PyObject_FastCall(__pyx_t_13, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
                  __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
                  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
                  if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 840, __pyx_L19_error)
                  __Pyx_GOTREF(__pyx_t_11);
                  __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
                }
                __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

                /* "viur/datastore/transport.pyx":841
 *                         if arrayElem.size() != abs(len(currentTxn["affectedEntities"])):
 *                             logging.error(resp.content)
 *                             raise ViurDatastoreError("Invalid number of mutation-results received")             # <<<<<<<<<<<<<<
 *                         arrayIt = arrayElem.begin()
 *                         idx = 0
 */
                __Pyx_GetModuleGlobalName(__pyx_t_13, __pyx_n_s_ViurDatastoreError); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 841, __pyx_L19_error)
                __Pyx_GOTREF(__pyx_t_13);
                __pyx_t_7 = NULL;
                __pyx_t_8 = 0;
                #if CYTHON_UNPACK_METHODS
                if (unlikely(PyMethod_Check(__pyx_t_13))) {
                  __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_13);
                  if (likely(__pyx_t_7)) {
                    PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_13);
                    __Pyx_INCREF(__pyx_t_7);
                    __Pyx_INCREF(function);
                    __Pyx_DECREF_SET(__pyx_t_13, function);
                    __pyx_t_8 = 1;
                  }
                }
                #endif
                {
                  PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_kp_u_Invalid_number_of_mutation_resul};
                  __pyx_t_11 = __Pyx_PyObject_FastCall(__pyx_t_13, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
                  __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
                  if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 841, __pyx_L19_error)
                  __Pyx_GOTREF(__pyx_t_11);
                  __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
                }
                __Pyx_Raise(__pyx_t_11, 0, 0, 0);
                __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
                __PYX_ERR(0, 841, __pyx_L19_error)

                /* "viur/datastore/transport.pyx":839
 *                             raise NoMutationResultsError("No mutation-results received")
 *                         arrayElem = element.at_key("mutationResults").get_array()
 *                         if arrayElem.size() != abs(len(currentTxn["affectedEntities"])):             # <<<<<<<<<<<<<<
 *                             logging.error(resp.content)
 *                             raise ViurDatastoreError("Invalid number of mutation-results received")
 */
              }

              /* "viur/datastore/transport.pyx":842
 *                             logging.error(resp.content)
 *                             raise ViurDatastoreError("Invalid number of mutation-results received")
 *                         arrayIt = arrayElem.begin()             # <<<<<<<<<<<<<<
 *                         idx = 0
 *                         while arrayIt != arrayElem.end():
 */
              __pyx_v_arrayIt = __pyx_v_arrayElem.begin();

              /* "viur/datastore/transport.pyx":843
 *                             raise ViurDatastoreError("Invalid number of mutation-results received")
 *                         arrayIt = arrayElem.begin()
 *                         idx = 0             # <<<<<<<<<<<<<<
 *                         while arrayIt != arrayElem.end():
 *                             innerArrayElem = dereference(arrayIt)
 */
              __Pyx_INCREF(__pyx_int_0);
              __Pyx_XDECREF_SET(__pyx_v_idx, __pyx_int_0);

              /* "viur/datastore/transport.pyx":844
 *                         arrayIt = arrayElem.begin()
 *                         idx = 0
 *                         while arrayIt != arrayElem.end():             # <<<<<<<<<<<<<<
 *                             innerArrayElem = dereference(arrayIt)
 *                             if innerArrayElem.at_pointer("/key").error() == SUCCESS:  # We got a new key assigned
 */
              while (1) {
                __pyx_t_10 = (__pyx_v_arrayIt != __pyx_v_arrayElem.end());
                if (!__pyx_t_10) break;

                /* "viur/datastore/transport.pyx":845
 *                         idx = 0
 *                         while arrayIt != arrayElem.end():
 *                             innerArrayElem = dereference(arrayIt)             # <<<<<<<<<<<<<<
 *                             if innerArrayElem.at_pointer("/key").error() == SUCCESS:  # We got a new key assigned
 *                                 affectedEntity = currentTxn["affectedEntities"][idx]
 */
                __pyx_v_innerArrayElem = (*__pyx_v_arrayIt);

                /* "viur/datastore/transport.pyx":846
 *                         while arrayIt != arrayElem.end():
 *                             innerArrayElem = dereference(arrayIt)
 *                             if innerArrayElem.at_pointer("/key").error() == SUCCESS:  # We got a new key assigned             # <<<<<<<<<<<<<<
 *                                 affectedEntity = currentTxn["affectedEntities"][idx]
 *                                 if not affectedEntity:
 */
                __pyx_t_10 = (__pyx_v_innerArrayElem.at_pointer(((char const *)"/key")).error() == simdjson::error_code::SUCCESS);
                if (__pyx_t_10) {

                  /* "viur/datastore/transport.pyx":847
 *                             innerArrayElem = dereference(arrayIt)
 *                             if innerArrayElem.at_pointer("/key").error() == SUCCESS:  # We got a new key assigned
 *                                 affectedEntity = currentTxn["affectedEntities"][idx]             # <<<<<<<<<<<<<<
 *                                 if not affectedEntity:
 *                                     logging.error(f"{resp.content=}")
 */
                  __pyx_t_11 = __Pyx_PyDict_GetItem(__pyx_v_currentTxn, __pyx_n_u_affectedEntities); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 847, __pyx_L19_error)
                  __Pyx_GOTREF(__pyx_t_11);
                  __pyx_t_13 = __Pyx_PyObject_GetItem(__pyx_t_11, __pyx_v_idx); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 847, __pyx_L19_error)
                  __Pyx_GOTREF(__pyx_t_13);
                  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
                  __Pyx_XDECREF_SET(__pyx_v_affectedEntity, __pyx_t_13);
                  __pyx_t_13 = 0;

                  /* "viur/datastore/transport.pyx":848
 *                             if innerArrayElem.at_pointer("/key").error() == SUCCESS:  # We got a new key assigned
 *                                 affectedEntity = currentTxn["affectedEntities"][idx]
 *                                 if not affectedEntity:             # <<<<<<<<<<<<<<
 *                                     logging.error(f"{resp.content=}")
 *                                     raise ViurDatastoreError("Received an unexpected key-update")
 */
                  __pyx_t_10 = __Pyx_PyObject_IsTrue(__pyx_v_affectedEntity); if (unlikely((__pyx_t_10 < 0))) __PYX_ERR(0, 848, __pyx_L19_error)
                  __pyx_t_9 = (!__pyx_t_10);
                  if (unlikely(__pyx_t_9)) {

                    /* "viur/datastore/transport.pyx":849
 *                                 affectedEntity = currentTxn["affectedEntities"][idx]
 *                                 if not affectedEntity:
 *                                     logging.error(f"{resp.content=}")             # <<<<<<<<<<<<<<
 *                                     raise ViurDatastoreError("Received an unexpected key-update")
 *                                 affectedEntity.key = parseKey(innerArrayElem.at_key("key"))
 */
                    __Pyx_GetModuleGlobalName(__pyx_t_11, __pyx_n_s_logging); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 849, __pyx_L19_error)
                    __Pyx_GOTREF(__pyx_t_11);
                    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_11, __pyx_n_s_error); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 849, __pyx_L19_error)
                    __Pyx_GOTREF(__pyx_t_7);
                    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
                    __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_v_resp, __pyx_n_s_content); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 849, __pyx_L19_error)
                    __Pyx_GOTREF(__pyx_t_11);
                    __pyx_t_12 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Repr(__pyx_t_11), __pyx_empty_unicode); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 849, __pyx_L19_error)
                    __Pyx_GOTREF(__pyx_t_12);
                    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
                    __pyx_t_11 = __Pyx_PyUnicode_Concat(__pyx_kp_u_resp_content, __pyx_t_12); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 849, __pyx_L19_error)
                    __Pyx_GOTREF(__pyx_t_11);
                    __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
                    __pyx_t_12 = NULL;
                    __pyx_t_8 = 0;
                    #if CYTHON_UNPACK_METHODS
                    if (unlikely(PyMethod_Check(__pyx_t_7))) {
                      __pyx_t_12 = PyMethod_GET_SELF(__pyx_t_7);
                      if (likely(__pyx_t_12)) {
                        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
                        __Pyx_INCREF(__pyx_t_12);
                        __Pyx_INCREF(function);
                        __Pyx_DECREF_SET(__pyx_t_7, function);
                        __pyx_t_8 = 1;
                      }
                    }
                    #endif
                    {
                      PyObject *__pyx_callargs[2] = {__pyx_t_12, __pyx_t_11};
                      __pyx_t_13 = __Pyx_PyObject_FastCall(__pyx_t_7, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
                      __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
                      __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
                      if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 849, __pyx_L19_error)
                      __Pyx_GOTREF(__pyx_t_13);
                      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
                    }
                    __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;

                    /* "viur/datastore/transport.pyx":850
 *                                 if not affectedEntity:
 *                                     logging.error(f"{resp.content=}")
 *                                     raise ViurDatastoreError("Received an unexpected key-update")             # <<<<<<<<<<<<<<
 *                                 affectedEntity.key = parseKey(innerArrayElem.at_key("key"))
 *                                 affectedEntity.version = toPyStr(innerArrayElem.at_key("version").get_string())
 */
                    __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_ViurDatastoreError); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 850, __pyx_L19_error)
                    __Pyx_GOTREF(__pyx_t_7);
                    __pyx_t_11 = NULL;
                    __pyx_t_8 = 0;
                    #if CYTHON_UNPACK_METHODS
                    if (unlikely(PyMethod_Check(__pyx_t_7))) {
                      __pyx_t_11 = PyMethod_GET_SELF(__pyx_t_7);
                      if (likely(__pyx_t_11)) {
                        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
                        __Pyx_INCREF(__pyx_t_11);
                        __Pyx_INCREF(function);
                        __Pyx_DECREF_SET(__pyx_t_7, function);
                        __pyx_t_8 = 1;
                      }
                    }
                    #endif
                    {
                      PyObject *__pyx_callargs[2] = {__pyx_t_11, __pyx_kp_u_Received_an_unexpected_key_updat};
                      __pyx_t_13 = __Pyx_PyObject_FastCall(__pyx_t_7, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
                      __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
                      if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 850, __pyx_L19_error)
                      __Pyx_GOTREF(__pyx_t_13);
                      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
                    }
                    __Pyx_Raise(__pyx_t_13, 0, 0, 0);
                    __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
                    __PYX_ERR(0, 850, __pyx_L19_error)

                    /* "viur/datastore/transport.pyx":848
 *                             if innerArrayElem.at_pointer("/key").error() == SUCCESS:  # We got a new key assigned
 *                                 affectedEntity = currentTxn["affectedEntities"][idx]
 *                                 if not affectedEntity:             # <<<<<<<<<<<<<<
 *                                     logging.error(f"{resp.content=}")
 *                                     raise ViurDatastoreError("Received an unexpected key-update")
 */
                  }

                  /* "viur/datastore/transport.pyx":851
 *                                     logging.error(f"{resp.content=}")
 *                                     raise ViurDatastoreError("Received an unexpected key-update")
 *                                 affectedEntity.key = parseKey(innerArrayElem.at_key("key"))             # <<<<<<<<<<<<<<
 *                                 affectedEntity.version = toPyStr(innerArrayElem.at_key("version").get_string())
 *                             preincrement(arrayIt)
 */
                  try {
                    __pyx_t_17 = __pyx_v_innerArrayElem.at_key(((char const *)"key"));
                  } catch(...) {
                    __Pyx_CppExn2PyErr();
                    __PYX_ERR(0, 851, __pyx_L19_error)
                  }
                  __pyx_t_13 = __pyx_f_4viur_9datastore_9transport_parseKey(__PYX_STD_MOVE_IF_SUPPORTED(__pyx_t_17)); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 851, __pyx_L19_error)
                  __Pyx_GOTREF(__pyx_t_13);
                  if (__Pyx_PyObject_SetAttrStr(__pyx_v_affectedEntity, __pyx_n_s_key, __pyx_t_13) < 0) __PYX_ERR(0, 851, __pyx_L19_error)
                  __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;

                  /* "viur/datastore/transport.pyx":852
 *                                     raise ViurDatastoreError("Received an unexpected key-update")
 *                                 affectedEntity.key = parseKey(innerArrayElem.at_key("key"))
 *                                 affectedEntity.version = toPyStr(innerArrayElem.at_key("version").get_string())             # <<<<<<<<<<<<<<
 *                             preincrement(arrayIt)
 *                             idx += 1
 */
                  try {
                    __pyx_t_17 = __pyx_v_innerArrayElem.at_key(((char const *)"version"));
                  } catch(...) {
                    __Pyx_CppExn2PyErr();
                    __PYX_ERR(0, 852, __pyx_L19_error)
                  }
                  try {
                    __pyx_t_20 = __pyx_t_17.get_string();
                  } catch(...) {
                    __Pyx_CppExn2PyErr();
                    __PYX_ERR(0, 852, __pyx_L19_error)
                  }
                  __pyx_t_13 = __pyx_f_4viur_9datastore_9transport_toPyStr(__PYX_STD_MOVE_IF_SUPPORTED(__pyx_t_20)); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 852, __pyx_L19_error)
                  __Pyx_GOTREF(__pyx_t_13);
                  if (__Pyx_PyObject_SetAttrStr(__pyx_v_affectedEntity, __pyx_n_s_version, __pyx_t_13) < 0) __PYX_ERR(0, 852, __pyx_L19_error)
                  __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;

                  /* "viur/datastore/transport.pyx":846
 *                         while arrayIt != arrayElem.end():
 *                             innerArrayElem = dereference(arrayIt)
 *                             if innerArrayElem.at_pointer("/key").error() == SUCCESS:  # We got a new key assigned             # <<<<<<<<<<<<<<
 *                                 affectedEntity = currentTxn["affectedEntities"][idx]
 *                                 if not affectedEntity:
 */
                }

                /* "viur/datastore/transport.pyx":853
 *                                 affectedEntity.key = parseKey(innerArrayElem.at_key("key"))
 *                                 affectedEntity.version = toPyStr(innerArrayElem.at_key("version").get_string())
 *                             preincrement(arrayIt)             # <<<<<<<<<<<<<<
 *                             idx += 1
 *                         return res
 */
                (void)((++__pyx_v_arrayIt));

                /* "viur/datastore/transport.pyx":854
 *                                 affectedEntity.version = toPyStr(innerArrayElem.at_key("version").get_string())
 *                             preincrement(arrayIt)
 *                             idx += 1             # <<<<<<<<<<<<<<
 *                         return res
 *                     else:  # No changes have been made - free txn
 */
                __pyx_t_13 = __Pyx_PyInt_AddObjC(__pyx_v_idx, __pyx_int_1, 1, 1, 0); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 854, __pyx_L19_error)
                __Pyx_GOTREF(__pyx_t_13);
                __Pyx_DECREF_SET(__pyx_v_idx, __pyx_t_13);
                __pyx_t_13 = 0;
              }

              /* "viur/datastore/transport.pyx":855
 *                             preincrement(arrayIt)
 *                             idx += 1
 *                         return res             # <<<<<<<<<<<<<<
 *                     else:  # No changes have been made - free txn
 *                         _rollbackTxn(txnKey)
 */
              __Pyx_XDECREF(__pyx_r);
              __Pyx_INCREF(__pyx_v_res);
              __pyx_r = __pyx_v_res;
              goto __pyx_L18_return;

              /* "viur/datastore/transport.pyx":820
 *                         _rollbackTxn(txnKey)
 *                         raise
 *                     if currentTxn["mutations"]:             # <<<<<<<<<<<<<<
 *                         # Commit TXN
 *                         postData = {
 */
            }

            /* "viur/datastore/transport.pyx":857
 *                         return res
 *                     else:  # No changes have been made - free txn
 *                         _rollbackTxn(txnKey)             # <<<<<<<<<<<<<<
 *                         return res
 *                 finally:  # Ensure, currentTransaction is always set back to none
 */
            /*else*/ {
              __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_rollbackTxn); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 857, __pyx_L19_error)
              __Pyx_GOTREF(__pyx_t_7);
              __pyx_t_11 = NULL;
              __pyx_t_8 = 0;
              #if CYTHON_UNPACK_METHODS
              if (unlikely(PyMethod_Check(__pyx_t_7))) {
                __pyx_t_11 = PyMethod_GET_SELF(__pyx_t_7);
                if (likely(__pyx_t_11)) {
                  PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
                  __Pyx_INCREF(__pyx_t_11);
                  __Pyx_INCREF(function);
                  __Pyx_DECREF_SET(__pyx_t_7, function);
                  __pyx_t_8 = 1;
                }
              }
              #endif
              {
                PyObject *__pyx_callargs[2] = {__pyx_t_11, __pyx_v_txnKey};
                __pyx_t_13 = __Pyx_PyObject_FastCall(__pyx_t_7, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
                __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
                if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 857, __pyx_L19_error)
                __Pyx_GOTREF(__pyx_t_13);
                __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
              }
              __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;

              /* "viur/datastore/transport.pyx":858
 *                     else:  # No changes have been made - free txn
 *                         _rollbackTxn(txnKey)
 *                         return res             # <<<<<<<<<<<<<<
 *                 finally:  # Ensure, currentTransaction is always set back to none
 *                     currentTransaction.set(None)
 */
              __Pyx_XDECREF(__pyx_r);
              __Pyx_INCREF(__pyx_v_res);
              __pyx_r = __pyx_v_res;
              goto __pyx_L18_return;
            }
          }

          /* "viur/datastore/transport.pyx":860
 *                         return res
 *                 finally:  # Ensure, currentTransaction is always set back to none
 *                     currentTransaction.set(None)             # <<<<<<<<<<<<<<
 *         except (CollisionError, AbortedError) as err:  # Got a collision or is aborted; retry the entire transaction
 *             sleep_time = 2 ** exponential_backoff
 */
          /*finally:*/ {
            __pyx_L19_error:;
            /*exception exit:*/{
              __Pyx_PyThreadState_declare
              __Pyx_PyThreadState_assign
              __pyx_t_16 = 0; __pyx_t_15 = 0; __pyx_t_14 = 0; __pyx_t_23 = 0; __pyx_t_24 = 0; __pyx_t_25 = 0;
              __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
              __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
              __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
              __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
              __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
              __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
              if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_23, &__pyx_t_24, &__pyx_t_25);
              if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_16, &__pyx_t_15, &__pyx_t_14) < 0)) __Pyx_ErrFetch(&__pyx_t_16, &__pyx_t_15, &__pyx_t_14);
              __Pyx_XGOTREF(__pyx_t_16);
              __Pyx_XGOTREF(__pyx_t_15);
              __Pyx_XGOTREF(__pyx_t_14);
              __Pyx_XGOTREF(__pyx_t_23);
              __Pyx_XGOTREF(__pyx_t_24);
              __Pyx_XGOTREF(__pyx_t_25);
              __pyx_t_8 = __pyx_lineno; __pyx_t_21 = __pyx_clineno; __pyx_t_22 = __pyx_filename;
              {
                __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_currentTransaction); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 860, __pyx_L41_error)
                __Pyx_GOTREF(__pyx_t_7);
                __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_set); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 860, __pyx_L41_error)
                __Pyx_GOTREF(__pyx_t_11);
                __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
                __pyx_t_7 = NULL;
                __pyx_t_26 = 0;
                #if CYTHON_UNPACK_METHODS
                if (unlikely(PyMethod_Check(__pyx_t_11))) {
                  __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_11);
                  if (likely(__pyx_t_7)) {
                    PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_11);
                    __Pyx_INCREF(__pyx_t_7);
                    __Pyx_INCREF(function);
                    __Pyx_DECREF_SET(__pyx_t_11, function);
                    __pyx_t_26 = 1;
                  }
                }
                #endif
                {
                  PyObject *__pyx_callargs[2] = {__pyx_t_7, Py_None};
                  __pyx_t_13 = __Pyx_PyObject_FastCall(__pyx_t_11, __pyx_callargs+1-__pyx_t_26, 1+__pyx_t_26);
                  __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
                  if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 860, __pyx_L41_error)
                  __Pyx_GOTREF(__pyx_t_13);
                  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
                }
                __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
              }
              if (PY_MAJOR_VERSION >= 3) {
                __Pyx_XGIVEREF(__pyx_t_23);
                __Pyx_XGIVEREF(__pyx_t_24);
                __Pyx_XGIVEREF(__pyx_t_25);
                __Pyx_ExceptionReset(__pyx_t_23, __pyx_t_24, __pyx_t_25);
              }
              __Pyx_XGIVEREF(__pyx_t_16);
              __Pyx_XGIVEREF(__pyx_t_15);
              __Pyx_XGIVEREF(__pyx_t_14);
              __Pyx_ErrRestore(__pyx_t_16, __pyx_t_15, __pyx_t_14);
              __pyx_t_16 = 0; __pyx_t_15 = 0; __pyx_t_14 = 0; __pyx_t_23 = 0; __pyx_t_24 = 0; __pyx_t_25 = 0;
              __pyx_lineno = __pyx_t_8; __pyx_clineno = __pyx_t_21; __pyx_filename = __pyx_t_22;
              goto __pyx_L5_error;
              __pyx_L41_error:;
              if (PY_MAJOR_VERSION >= 3) {
                __Pyx_XGIVEREF(__pyx_t_23);
                __Pyx_XGIVEREF(__pyx_t_24);
                __Pyx_XGIVEREF(__pyx_t_25);
                __Pyx_ExceptionReset(__pyx_t_23, __pyx_t_24, __pyx_t_25);
              }
              __Pyx_XDECREF(__pyx_t_16); __pyx_t_16 = 0;
              __Pyx_XDECREF(__pyx_t_15); __pyx_t_15 = 0;
              __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
              __pyx_t_23 = 0; __pyx_t_24 = 0; __pyx_t_25 = 0;
              goto __pyx_L5_error;
            }
            __pyx_L18_return: {
              __pyx_t_25 = __pyx_r;
              __pyx_r = 0;
              __Pyx_GetModuleGlobalName(__pyx_t_11, __pyx_n_s_currentTransaction); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 860, __pyx_L5_error)
              __Pyx_GOTREF(__pyx_t_11);
              __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_11, __pyx_n_s_set); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 860, __pyx_L5_error)
              __Pyx_GOTREF(__pyx_t_7);
              __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
              __pyx_t_11 = NULL;
              __pyx_t_21 = 0;
              #if CYTHON_UNPACK_METHODS
              if (unlikely(PyMethod_Check(__pyx_t_7))) {
                __pyx_t_11 = PyMethod_GET_SELF(__pyx_t_7);
                if (likely(__pyx_t_11)) {
                  PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
                  __Pyx_INCREF(__pyx_t_11);
                  __Pyx_INCREF(function);
                  __Pyx_DECREF_SET(__pyx_t_7, function);
                  __pyx_t_21 = 1;
                }
              }
              #endif
              {
                PyObject *__pyx_callargs[2] = {__pyx_t_11, Py_None};
                __pyx_t_13 = __Pyx_PyObject_FastCall(__pyx_t_7, __pyx_callargs+1-__pyx_t_21, 1+__pyx_t_21);
                __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
                if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 860, __pyx_L5_error)
                __Pyx_GOTREF(__pyx_t_13);
                __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
              }
              __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
              __pyx_r = __pyx_t_25;
              __pyx_t_25 = 0;
              goto __pyx_L9_try_return;
            }
          }

          /* "viur/datastore/transport.pyx":810
 *                 data=json.dumps(postData).encode("UTF-8"),
 *             )
 *             if is_viur_datastore_request_ok(resp):             # <<<<<<<<<<<<<<
 *                 txnKey = json.loads(resp.content)["transaction"]
 *                 try:
 */
        }

        /* "viur/datastore/transport.pyx":790
 *     cdef simdjsonArray.iterator arrayIt
 *     for exponential_backoff in range(1, 4):
 *         try:             # <<<<<<<<<<<<<<
 *             oldTxn = currentTransaction.get()
 *             allowOverriding = kwargs.pop("__allowOverriding__", None)
 */
      }
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      goto __pyx_L12_try_end;
      __pyx_L5_error:;
      __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
      __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
      __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;

      /* "viur/datastore/transport.pyx":861
 *                 finally:  # Ensure, currentTransaction is always set back to none
 *                     currentTransaction.set(None)
 *         except (CollisionError, AbortedError) as err:  # Got a collision or is aborted; retry the entire transaction             # <<<<<<<<<<<<<<
 *             sleep_time = 2 ** exponential_backoff
 *             logging.error(f"We got an error in a transaction we try again in {sleep_time} seconds")
 */
      __Pyx_ErrFetch(&__pyx_t_13, &__pyx_t_7, &__pyx_t_11);
      __Pyx_GetModuleGlobalName(__pyx_t_12, __pyx_n_s_CollisionError); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 861, __pyx_L7_except_error)
      __Pyx_GOTREF(__pyx_t_12);
      __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_AbortedError); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 861, __pyx_L7_except_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_21 = __Pyx_PyErr_GivenExceptionMatches2(__pyx_t_13, __pyx_t_12, __pyx_t_2);
      __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_ErrRestore(__pyx_t_13, __pyx_t_7, __pyx_t_11);
      __pyx_t_13 = 0; __pyx_t_7 = 0; __pyx_t_11 = 0;
      if (__pyx_t_21) {
        __Pyx_AddTraceback("viur.datastore.transport.RunInTransaction", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_11, &__pyx_t_7, &__pyx_t_13) < 0) __PYX_ERR(0, 861, __pyx_L7_except_error)
        __Pyx_XGOTREF(__pyx_t_11);
        __Pyx_XGOTREF(__pyx_t_7);
        __Pyx_XGOTREF(__pyx_t_13);
        __Pyx_INCREF(__pyx_t_7);
        __pyx_v_err = __pyx_t_7;
        /*try:*/ {

          /* "viur/datastore/transport.pyx":862
 *                     currentTransaction.set(None)
 *         except (CollisionError, AbortedError) as err:  # Got a collision or is aborted; retry the entire transaction
 *             sleep_time = 2 ** exponential_backoff             # <<<<<<<<<<<<<<
 *             logging.error(f"We got an error in a transaction we try again in {sleep_time} seconds")
 *             sleep(sleep_time)
 */
          __pyx_t_2 = __Pyx_PyNumber_PowerOf2(__pyx_int_2, __pyx_v_exponential_backoff, Py_None); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 862, __pyx_L47_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_XDECREF_SET(__pyx_v_sleep_time, __pyx_t_2);
          __pyx_t_2 = 0;

          /* "viur/datastore/transport.pyx":863
 *         except (CollisionError, AbortedError) as err:  # Got a collision or is aborted; retry the entire transaction
 *             sleep_time = 2 ** exponential_backoff
 *             logging.error(f"We got an error in a transaction we try again in {sleep_time} seconds")             # <<<<<<<<<<<<<<
 *             sleep(sleep_time)
 *     raise CollisionError(
 */
          __Pyx_GetModuleGlobalName(__pyx_t_12, __pyx_n_s_logging); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 863, __pyx_L47_error)
          __Pyx_GOTREF(__pyx_t_12);
          __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_12, __pyx_n_s_error); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 863, __pyx_L47_error)
          __Pyx_GOTREF(__pyx_t_6);
          __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
          __pyx_t_12 = PyTuple_New(3); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 863, __pyx_L47_error)
          __Pyx_GOTREF(__pyx_t_12);
          __pyx_t_19 = 0;
          __pyx_t_27 = 127;
          __Pyx_INCREF(__pyx_kp_u_We_got_an_error_in_a_transaction);
          __pyx_t_19 += 49;
          __Pyx_GIVEREF(__pyx_kp_u_We_got_an_error_in_a_transaction);
          PyTuple_SET_ITEM(__pyx_t_12, 0, __pyx_kp_u_We_got_an_error_in_a_transaction);
          __pyx_t_28 = __Pyx_PyObject_FormatSimple(__pyx_v_sleep_time, __pyx_empty_unicode); if (unlikely(!__pyx_t_28)) __PYX_ERR(0, 863, __pyx_L47_error)
          __Pyx_GOTREF(__pyx_t_28);
          __pyx_t_27 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_28) > __pyx_t_27) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_28) : __pyx_t_27;
          __pyx_t_19 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_28);
          __Pyx_GIVEREF(__pyx_t_28);
          PyTuple_SET_ITEM(__pyx_t_12, 1, __pyx_t_28);
          __pyx_t_28 = 0;
          __Pyx_INCREF(__pyx_kp_u_seconds);
          __pyx_t_19 += 8;
          __Pyx_GIVEREF(__pyx_kp_u_seconds);
          PyTuple_SET_ITEM(__pyx_t_12, 2, __pyx_kp_u_seconds);
          __pyx_t_28 = __Pyx_PyUnicode_Join(__pyx_t_12, 3, __pyx_t_19, __pyx_t_27); if (unlikely(!__pyx_t_28)) __PYX_ERR(0, 863, __pyx_L47_error)
          __Pyx_GOTREF(__pyx_t_28);
          __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
          __pyx_t_12 = NULL;
          __pyx_t_21 = 0;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_6))) {
            __pyx_t_12 = PyMethod_GET_SELF(__pyx_t_6);
            if (likely(__pyx_t_12)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
              __Pyx_INCREF(__pyx_t_12);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_6, function);
              __pyx_t_21 = 1;
            }
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_12, __pyx_t_28};
            __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_6, __pyx_callargs+1-__pyx_t_21, 1+__pyx_t_21);
            __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
            __Pyx_DECREF(__pyx_t_28); __pyx_t_28 = 0;
            if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 863, __pyx_L47_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
          }
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

          /* "viur/datastore/transport.pyx":864
 *             sleep_time = 2 ** exponential_backoff
 *             logging.error(f"We got an error in a transaction we try again in {sleep_time} seconds")
 *             sleep(sleep_time)             # <<<<<<<<<<<<<<
 *     raise CollisionError(
 *         "All retries are exhausted for this transaction")  # If we made it here, all tries are exhausted
 */
          __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_sleep); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 864, __pyx_L47_error)
          __Pyx_GOTREF(__pyx_t_6);
          __pyx_t_28 = NULL;
          __pyx_t_21 = 0;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_6))) {
            __pyx_t_28 = PyMethod_GET_SELF(__pyx_t_6);
            if (likely(__pyx_t_28)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
              __Pyx_INCREF(__pyx_t_28);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_6, function);
              __pyx_t_21 = 1;
            }
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_28, __pyx_v_sleep_time};
            __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_6, __pyx_callargs+1-__pyx_t_21, 1+__pyx_t_21);
            __Pyx_XDECREF(__pyx_t_28); __pyx_t_28 = 0;
            if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 864, __pyx_L47_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
          }
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        }

        /* "viur/datastore/transport.pyx":861
 *                 finally:  # Ensure, currentTransaction is always set back to none
 *                     currentTransaction.set(None)
 *         except (CollisionError, AbortedError) as err:  # Got a collision or is aborted; retry the entire transaction             # <<<<<<<<<<<<<<
 *             sleep_time = 2 ** exponential_backoff
 *             logging.error(f"We got an error in a transaction we try again in {sleep_time} seconds")
 */
        /*finally:*/ {
          /*normal exit:*/{
            __Pyx_DECREF(__pyx_v_err); __pyx_v_err = 0;
            goto __pyx_L48;
          }
          __pyx_L47_error:;
          /*exception exit:*/{
            __Pyx_PyThreadState_declare
            __Pyx_PyThreadState_assign
            __pyx_t_25 = 0; __pyx_t_24 = 0; __pyx_t_23 = 0; __pyx_t_14 = 0; __pyx_t_15 = 0; __pyx_t_16 = 0;
            __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
            __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
            __Pyx_XDECREF(__pyx_t_28); __pyx_t_28 = 0;
            __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
            if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_14, &__pyx_t_15, &__pyx_t_16);
            if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_25, &__pyx_t_24, &__pyx_t_23) < 0)) __Pyx_ErrFetch(&__pyx_t_25, &__pyx_t_24, &__pyx_t_23);
            __Pyx_XGOTREF(__pyx_t_25);
            __Pyx_XGOTREF(__pyx_t_24);
            __Pyx_XGOTREF(__pyx_t_23);
            __Pyx_XGOTREF(__pyx_t_14);
            __Pyx_XGOTREF(__pyx_t_15);
            __Pyx_XGOTREF(__pyx_t_16);
            __pyx_t_21 = __pyx_lineno; __pyx_t_8 = __pyx_clineno; __pyx_t_29 = __pyx_filename;
            {
              __Pyx_DECREF(__pyx_v_err); __pyx_v_err = 0;
            }
            if (PY_MAJOR_VERSION >= 3) {
              __Pyx_XGIVEREF(__pyx_t_14);
              __Pyx_XGIVEREF(__pyx_t_15);
              __Pyx_XGIVEREF(__pyx_t_16);
              __Pyx_ExceptionReset(__pyx_t_14, __pyx_t_15, __pyx_t_16);
            }
            __Pyx_XGIVEREF(__pyx_t_25);
            __Pyx_XGIVEREF(__pyx_t_24);
            __Pyx_XGIVEREF(__pyx_t_23);
            __Pyx_ErrRestore(__pyx_t_25, __pyx_t_24, __pyx_t_23);
            __pyx_t_25 = 0; __pyx_t_24 = 0; __pyx_t_23 = 0; __pyx_t_14 = 0; __pyx_t_15 = 0; __pyx_t_16 = 0;
            __pyx_lineno = __pyx_t_21; __pyx_clineno = __pyx_t_8; __pyx_filename = __pyx_t_29;
            goto __pyx_L7_except_error;
          }
          __pyx_L48:;
        }
        __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
        __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
        goto __pyx_L6_exception_handled;
      }
      goto __pyx_L7_except_error;

      /* "viur/datastore/transport.pyx":790
 *     cdef simdjsonArray.iterator arrayIt
 *     for exponential_backoff in range(1, 4):
 *         try:             # <<<<<<<<<<<<<<
 *             oldTxn = currentTransaction.get()
 *             allowOverriding = kwargs.pop("__allowOverriding__", None)
 */
      __pyx_L7_except_error:;
      __Pyx_XGIVEREF(__pyx_t_3);
      __Pyx_XGIVEREF(__pyx_t_4);
      __Pyx_XGIVEREF(__pyx_t_5);
      __Pyx_ExceptionReset(__pyx_t_3, __pyx_t_4, __pyx_t_5);
      goto __pyx_L1_error;
      __pyx_L9_try_return:;
      __Pyx_XGIVEREF(__pyx_t_3);
      __Pyx_XGIVEREF(__pyx_t_4);
      __Pyx_XGIVEREF(__pyx_t_5);
      __Pyx_ExceptionReset(__pyx_t_3, __pyx_t_4, __pyx_t_5);
      goto __pyx_L0;
      __pyx_L6_exception_handled:;
      __Pyx_XGIVEREF(__pyx_t_3);
      __Pyx_XGIVEREF(__pyx_t_4);
      __Pyx_XGIVEREF(__pyx_t_5);
      __Pyx_ExceptionReset(__pyx_t_3, __pyx_t_4, __pyx_t_5);
      __pyx_L12_try_end:;
    }
  }

  /* "viur/datastore/transport.pyx":865
 *             logging.error(f"We got an error in a transaction we try again in {sleep_time} seconds")
 *             sleep(sleep_time)
 *     raise CollisionError(             # <<<<<<<<<<<<<<
 *         "All retries are exhausted for this transaction")  # If we made it here, all tries are exhausted
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_CollisionError); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 865, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_11 = NULL;
  __pyx_t_8 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_7))) {
    __pyx_t_11 = PyMethod_GET_SELF(__pyx_t_7);
    if (likely(__pyx_t_11)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
      __Pyx_INCREF(__pyx_t_11);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_7, function);
      __pyx_t_8 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_11, __pyx_kp_u_All_retries_are_exhausted_for_th};
    __pyx_t_13 = __Pyx_PyObject_FastCall(__pyx_t_7, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
    __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
    if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 865, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_13);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  }
  __Pyx_Raise(__pyx_t_13, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
  __PYX_ERR(0, 865, __pyx_L1_error)

  /* "viur/datastore/transport.pyx":774
 *     return entities
 * 
 * def RunInTransaction(callback: callable, *args, **kwargs) -> Any:             # <<<<<<<<<<<<<<
 *     """
 *         Runs the given function inside a AID transaction.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_XDECREF(__pyx_t_28);
  __Pyx_AddTraceback("viur.datastore.transport.RunInTransaction", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_exponential_backoff);
  __Pyx_XDECREF(__pyx_v_oldTxn);
  __Pyx_XDECREF(__pyx_v_allowOverriding);
  __Pyx_XDECREF(__pyx_v_txnOptions);
  __Pyx_XDECREF(__pyx_v_postData);
  __Pyx_XDECREF(__pyx_v_resp);
  __Pyx_XDECREF(__pyx_v_txnKey);
  __Pyx_XDECREF(__pyx_v_currentTxn);
  __Pyx_XDECREF(__pyx_v_res);
  __Pyx_XDECREF(__pyx_v_idx);
  __Pyx_XDECREF(__pyx_v_affectedEntity);
  __Pyx_XDECREF(__pyx_v_err);
  __Pyx_XDECREF(__pyx_v_sleep_time);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "viur/datastore/transport.pyx":868
 *         "All retries are exhausted for this transaction")  # If we made it here, all tries are exhausted
 * 
 * def _rollbackTxn(txnKey: str):             # <<<<<<<<<<<<<<
 *     """
 *         Internal helper that aborts the given transaction. It's important to abort pending transactions (instead
 */

/* Python wrapper */
static PyObject *__pyx_pw_4viur_9datastore_9transport_17_rollbackTxn(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_4viur_9datastore_9transport_16_rollbackTxn, "\n        Internal helper that aborts the given transaction. It's important to abort pending transactions (instead\n        of letting them expire server-side) so that other requests that are stalled waiting for this transaction to\n        complete can continue.\n\n        :param txnKey: The ID of the transaction that should be aborted\n    ");
static PyMethodDef __pyx_mdef_4viur_9datastore_9transport_17_rollbackTxn = {"_rollbackTxn", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_4viur_9datastore_9transport_17_rollbackTxn, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_4viur_9datastore_9transport_16_rollbackTxn};
static PyObject *__pyx_pw_4viur_9datastore_9transport_17_rollbackTxn(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_txnKey = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_rollbackTxn (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_txnKey,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_txnKey)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 868, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "_rollbackTxn") < 0)) __PYX_ERR(0, 868, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_txnKey = ((PyObject*)values[0]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_rollbackTxn", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 868, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("viur.datastore.transport._rollbackTxn", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_txnKey), (&PyUnicode_Type), 0, "txnKey", 1))) __PYX_ERR(0, 868, __pyx_L1_error)
  __pyx_r = __pyx_pf_4viur_9datastore_9transport_16_rollbackTxn(__pyx_self, __pyx_v_txnKey);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4viur_9datastore_9transport_16_rollbackTxn(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_txnKey) {
  PyObject *__pyx_v_postData = NULL;
  CYTHON_UNUSED PyObject *__pyx_v_resp = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_rollbackTxn", 1);

  /* "viur/datastore/transport.pyx":877
 *     """
 *     postData = {
 *         "transaction": txnKey             # <<<<<<<<<<<<<<
 *     }
 *     resp = authenticated_request(
 */
  __pyx_t_1 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 877, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_u_transaction, __pyx_v_txnKey) < 0) __PYX_ERR(0, 877, __pyx_L1_error)
  __pyx_v_postData = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "viur/datastore/transport.pyx":879
 *         "transaction": txnKey
 *     }
 *     resp = authenticated_request(             # <<<<<<<<<<<<<<
 *         url="https://datastore.googleapis.com/v1/projects/%s:rollback" % projectID,
 *         data=json.dumps(postData).encode("UTF-8"),
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_authenticated_request); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 879, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);

  /* "viur/datastore/transport.pyx":880
 *     }
 *     resp = authenticated_request(
 *         url="https://datastore.googleapis.com/v1/projects/%s:rollback" % projectID,             # <<<<<<<<<<<<<<
 *         data=json.dumps(postData).encode("UTF-8"),
 *     )
 */
  __pyx_t_2 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 880, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_projectID); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 880, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyUnicode_FormatSafe(__pyx_kp_u_https_datastore_googleapis_com_v_5, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 880, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_url, __pyx_t_4) < 0) __PYX_ERR(0, 880, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "viur/datastore/transport.pyx":881
 *     resp = authenticated_request(
 *         url="https://datastore.googleapis.com/v1/projects/%s:rollback" % projectID,
 *         data=json.dumps(postData).encode("UTF-8"),             # <<<<<<<<<<<<<<
 *     )
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_json); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 881, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_dumps); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 881, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = NULL;
  __pyx_t_7 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_6))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_6);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_6, function);
      __pyx_t_7 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_5, __pyx_v_postData};
    __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_6, __pyx_callargs+1-__pyx_t_7, 1+__pyx_t_7);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 881, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_encode); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 881, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  __pyx_t_7 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_6))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_6);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_6, function);
      __pyx_t_7 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_kp_u_UTF_8};
    __pyx_t_4 = __Pyx_PyObject_FastCall(__pyx_t_6, __pyx_callargs+1-__pyx_t_7, 1+__pyx_t_7);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 881, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_data, __pyx_t_4) < 0) __PYX_ERR(0, 880, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "viur/datastore/transport.pyx":879
 *         "transaction": txnKey
 *     }
 *     resp = authenticated_request(             # <<<<<<<<<<<<<<
 *         url="https://datastore.googleapis.com/v1/projects/%s:rollback" % projectID,
 *         data=json.dumps(postData).encode("UTF-8"),
 */
  __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_empty_tuple, __pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 879, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_resp = __pyx_t_4;
  __pyx_t_4 = 0;

  /* "viur/datastore/transport.pyx":868
 *         "All retries are exhausted for this transaction")  # If we made it here, all tries are exhausted
 * 
 * def _rollbackTxn(txnKey: str):             # <<<<<<<<<<<<<<
 *     """
 *         Internal helper that aborts the given transaction. It's important to abort pending transactions (instead
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("viur.datastore.transport._rollbackTxn", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_postData);
  __Pyx_XDECREF(__pyx_v_resp);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "viur/datastore/transport.pyx":884
 *     )
 * 
 * def AllocateIDs(keys: Union[Key, List[Key]]) -> Union[Key, List[Key]]:             # <<<<<<<<<<<<<<
 *     """
 *         Allocates numeric IDs for the keys given.
 */

/* Python wrapper */
static PyObject *__pyx_pw_4viur_9datastore_9transport_19AllocateIDs(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_4viur_9datastore_9transport_18AllocateIDs, "\n        Allocates numeric IDs for the keys given.\n        :return: The complete Key (or a list hereof)\n\n        .. warning: This function does not support transactions! Even if called inside transactions, the keys will\n            be allocated immediately, even if the transaction aborts.\n    ");
static PyMethodDef __pyx_mdef_4viur_9datastore_9transport_19AllocateIDs = {"AllocateIDs", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_4viur_9datastore_9transport_19AllocateIDs, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_4viur_9datastore_9transport_18AllocateIDs};
static PyObject *__pyx_pw_4viur_9datastore_9transport_19AllocateIDs(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_keys = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("AllocateIDs (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_keys,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_keys)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 884, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "AllocateIDs") < 0)) __PYX_ERR(0, 884, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_keys = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("AllocateIDs", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 884, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("viur.datastore.transport.AllocateIDs", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4viur_9datastore_9transport_18AllocateIDs(__pyx_self, __pyx_v_keys);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4viur_9datastore_9transport_18AllocateIDs(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_keys) {
  simdjson::dom::parser __pyx_v_parser;
  Py_ssize_t __pyx_v_pysize;
  char *__pyx_v_data_ptr;
  simdjson::dom::element __pyx_v_element;
  int __pyx_v_isMulti;
  PyObject *__pyx_v_keyList = NULL;
  PyObject *__pyx_v_requestedKeys = NULL;
  PyObject *__pyx_v_postData = NULL;
  PyObject *__pyx_v_resp = NULL;
  PyObject *__pyx_v_res = NULL;
  simdjson::dom::array __pyx_v_arrayElem;
  simdjson::dom::array::iterator __pyx_v_arrayIt;
  simdjson::dom::element __pyx_v_innerArrayElem;
  PyObject *__pyx_9genexpr16__pyx_v_x = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  Py_ssize_t __pyx_t_4;
  PyObject *(*__pyx_t_5)(PyObject *);
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  int __pyx_t_10;
  simdjson::dom::element __pyx_t_11;
  simdjson::dom::array __pyx_t_12;
  int __pyx_t_13;
  int __pyx_t_14;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("AllocateIDs", 0);
  __Pyx_INCREF(__pyx_v_keys);

  /* "viur/datastore/transport.pyx":892
 *             be allocated immediately, even if the transaction aborts.
 *     """
 *     cdef simdjsonParser parser = simdjsonParser()             # <<<<<<<<<<<<<<
 *     cdef Py_ssize_t pysize
 *     cdef char * data_ptr
 */
  __pyx_v_parser = simdjson::dom::parser();

  /* "viur/datastore/transport.pyx":896
 *     cdef char * data_ptr
 *     cdef simdjsonElement element
 *     isMulti = True             # <<<<<<<<<<<<<<
 *     if isinstance(keys, Key):
 *         keys = [keys]
 */
  __pyx_v_isMulti = 1;

  /* "viur/datastore/transport.pyx":897
 *     cdef simdjsonElement element
 *     isMulti = True
 *     if isinstance(keys, Key):             # <<<<<<<<<<<<<<
 *         keys = [keys]
 *         isMulti = False
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_Key); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 897, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyObject_IsInstance(__pyx_v_keys, __pyx_t_1); if (unlikely(__pyx_t_2 == ((int)-1))) __PYX_ERR(0, 897, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "viur/datastore/transport.pyx":898
 *     isMulti = True
 *     if isinstance(keys, Key):
 *         keys = [keys]             # <<<<<<<<<<<<<<
 *         isMulti = False
 *     keyList = [
 */
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 898, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_v_keys);
    __Pyx_GIVEREF(__pyx_v_keys);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 0, __pyx_v_keys)) __PYX_ERR(0, 898, __pyx_L1_error);
    __Pyx_DECREF_SET(__pyx_v_keys, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "viur/datastore/transport.pyx":899
 *     if isinstance(keys, Key):
 *         keys = [keys]
 *         isMulti = False             # <<<<<<<<<<<<<<
 *     keyList = [
 *         {
 */
    __pyx_v_isMulti = 0;

    /* "viur/datastore/transport.pyx":897
 *     cdef simdjsonElement element
 *     isMulti = True
 *     if isinstance(keys, Key):             # <<<<<<<<<<<<<<
 *         keys = [keys]
 *         isMulti = False
 */
  }

  /* "viur/datastore/transport.pyx":900
 *         keys = [keys]
 *         isMulti = False
 *     keyList = [             # <<<<<<<<<<<<<<
 *         {
 *             "partitionId": {
 */
  { /* enter inner scope */
    __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 900, __pyx_L6_error)
    __Pyx_GOTREF(__pyx_t_1);

    /* "viur/datastore/transport.pyx":907
 *             "path": keyToPath(x)
 *         }
 *         for x in keys]             # <<<<<<<<<<<<<<
 * 
 *     requestedKeys = keyList[:300]
 */
    if (likely(PyList_CheckExact(__pyx_v_keys)) || PyTuple_CheckExact(__pyx_v_keys)) {
      __pyx_t_3 = __pyx_v_keys; __Pyx_INCREF(__pyx_t_3);
      __pyx_t_4 = 0;
      __pyx_t_5 = NULL;
    } else {
      __pyx_t_4 = -1; __pyx_t_3 = PyObject_GetIter(__pyx_v_keys); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 907, __pyx_L6_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_5 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_3); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 907, __pyx_L6_error)
    }
    for (;;) {
      if (likely(!__pyx_t_5)) {
        if (likely(PyList_CheckExact(__pyx_t_3))) {
          {
            Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_3);
            #if !CYTHON_ASSUME_SAFE_MACROS
            if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 907, __pyx_L6_error)
            #endif
            if (__pyx_t_4 >= __pyx_temp) break;
          }
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_6 = PyList_GET_ITEM(__pyx_t_3, __pyx_t_4); __Pyx_INCREF(__pyx_t_6); __pyx_t_4++; if (unlikely((0 < 0))) __PYX_ERR(0, 907, __pyx_L6_error)
          #else
          __pyx_t_6 = __Pyx_PySequence_ITEM(__pyx_t_3, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 907, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_6);
          #endif
        } else {
          {
            Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_3);
            #if !CYTHON_ASSUME_SAFE_MACROS
            if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 907, __pyx_L6_error)
            #endif
            if (__pyx_t_4 >= __pyx_temp) break;
          }
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_6 = PyTuple_GET_ITEM(__pyx_t_3, __pyx_t_4); __Pyx_INCREF(__pyx_t_6); __pyx_t_4++; if (unlikely((0 < 0))) __PYX_ERR(0, 907, __pyx_L6_error)
          #else
          __pyx_t_6 = __Pyx_PySequence_ITEM(__pyx_t_3, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 907, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_6);
          #endif
        }
      } else {
        __pyx_t_6 = __pyx_t_5(__pyx_t_3);
        if (unlikely(!__pyx_t_6)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 907, __pyx_L6_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_6);
      }
      __Pyx_XDECREF_SET(__pyx_9genexpr16__pyx_v_x, __pyx_t_6);
      __pyx_t_6 = 0;

      /* "viur/datastore/transport.pyx":902
 *     keyList = [
 *         {
 *             "partitionId": {             # <<<<<<<<<<<<<<
 *                 "project_id": projectID,
 *             },
 */
      __pyx_t_6 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 902, __pyx_L6_error)
      __Pyx_GOTREF(__pyx_t_6);

      /* "viur/datastore/transport.pyx":903
 *         {
 *             "partitionId": {
 *                 "project_id": projectID,             # <<<<<<<<<<<<<<
 *             },
 *             "path": keyToPath(x)
 */
      __pyx_t_7 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 903, __pyx_L6_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_GetModuleGlobalName(__pyx_t_8, __pyx_n_s_projectID); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 903, __pyx_L6_error)
      __Pyx_GOTREF(__pyx_t_8);
      if (PyDict_SetItem(__pyx_t_7, __pyx_n_u_project_id, __pyx_t_8) < 0) __PYX_ERR(0, 903, __pyx_L6_error)
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      if (PyDict_SetItem(__pyx_t_6, __pyx_n_u_partitionId, __pyx_t_7) < 0) __PYX_ERR(0, 902, __pyx_L6_error)
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

      /* "viur/datastore/transport.pyx":905
 *                 "project_id": projectID,
 *             },
 *             "path": keyToPath(x)             # <<<<<<<<<<<<<<
 *         }
 *         for x in keys]
 */
      __Pyx_GetModuleGlobalName(__pyx_t_8, __pyx_n_s_keyToPath); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 905, __pyx_L6_error)
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_t_9 = NULL;
      __pyx_t_10 = 0;
      #if CYTHON_UNPACK_METHODS
      if (unlikely(PyMethod_Check(__pyx_t_8))) {
        __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_8);
        if (likely(__pyx_t_9)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
          __Pyx_INCREF(__pyx_t_9);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_8, function);
          __pyx_t_10 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_9, __pyx_9genexpr16__pyx_v_x};
        __pyx_t_7 = __Pyx_PyObject_FastCall(__pyx_t_8, __pyx_callargs+1-__pyx_t_10, 1+__pyx_t_10);
        __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
        if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 905, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      }
      if (PyDict_SetItem(__pyx_t_6, __pyx_n_u_path, __pyx_t_7) < 0) __PYX_ERR(0, 902, __pyx_L6_error)
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      if (unlikely(__Pyx_ListComp_Append(__pyx_t_1, (PyObject*)__pyx_t_6))) __PYX_ERR(0, 900, __pyx_L6_error)
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

      /* "viur/datastore/transport.pyx":907
 *             "path": keyToPath(x)
 *         }
 *         for x in keys]             # <<<<<<<<<<<<<<
 * 
 *     requestedKeys = keyList[:300]
 */
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_XDECREF(__pyx_9genexpr16__pyx_v_x); __pyx_9genexpr16__pyx_v_x = 0;
    goto __pyx_L10_exit_scope;
    __pyx_L6_error:;
    __Pyx_XDECREF(__pyx_9genexpr16__pyx_v_x); __pyx_9genexpr16__pyx_v_x = 0;
    goto __pyx_L1_error;
    __pyx_L10_exit_scope:;
  } /* exit inner scope */
  __pyx_v_keyList = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "viur/datastore/transport.pyx":909
 *         for x in keys]
 * 
 *     requestedKeys = keyList[:300]             # <<<<<<<<<<<<<<
 *     postData = {
 *         "keys": requestedKeys,
 */
  __pyx_t_1 = __Pyx_PyList_GetSlice(__pyx_v_keyList, 0, 0x12C); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 909, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_requestedKeys = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "viur/datastore/transport.pyx":911
 *     requestedKeys = keyList[:300]
 *     postData = {
 *         "keys": requestedKeys,             # <<<<<<<<<<<<<<
 *     }
 *     resp = authenticated_request(
 */
  __pyx_t_1 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 911, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_u_keys, __pyx_v_requestedKeys) < 0) __PYX_ERR(0, 911, __pyx_L1_error)
  __pyx_v_postData = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "viur/datastore/transport.pyx":913
 *         "keys": requestedKeys,
 *     }
 *     resp = authenticated_request(             # <<<<<<<<<<<<<<
 *         url="https://datastore.googleapis.com/v1/projects/%s:allocateIds" % projectID,
 *         data=json.dumps(postData).encode("UTF-8"),
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_authenticated_request); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 913, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);

  /* "viur/datastore/transport.pyx":914
 *     }
 *     resp = authenticated_request(
 *         url="https://datastore.googleapis.com/v1/projects/%s:allocateIds" % projectID,             # <<<<<<<<<<<<<<
 *         data=json.dumps(postData).encode("UTF-8"),
 *     )
 */
  __pyx_t_3 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 914, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_projectID); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 914, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_7 = __Pyx_PyUnicode_FormatSafe(__pyx_kp_u_https_datastore_googleapis_com_v_6, __pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 914, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_url, __pyx_t_7) < 0) __PYX_ERR(0, 914, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "viur/datastore/transport.pyx":915
 *     resp = authenticated_request(
 *         url="https://datastore.googleapis.com/v1/projects/%s:allocateIds" % projectID,
 *         data=json.dumps(postData).encode("UTF-8"),             # <<<<<<<<<<<<<<
 *     )
 *     if is_viur_datastore_request_ok(resp):
 */
  __Pyx_GetModuleGlobalName(__pyx_t_8, __pyx_n_s_json); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 915, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_dumps); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 915, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_t_8 = NULL;
  __pyx_t_10 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_9))) {
    __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_9);
    if (likely(__pyx_t_8)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_9);
      __Pyx_INCREF(__pyx_t_8);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_9, function);
      __pyx_t_10 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_8, __pyx_v_postData};
    __pyx_t_6 = __Pyx_PyObject_FastCall(__pyx_t_9, __pyx_callargs+1-__pyx_t_10, 1+__pyx_t_10);
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 915, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  }
  __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_encode); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 915, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_6 = NULL;
  __pyx_t_10 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_9))) {
    __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_9);
    if (likely(__pyx_t_6)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_9);
      __Pyx_INCREF(__pyx_t_6);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_9, function);
      __pyx_t_10 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_6, __pyx_kp_u_UTF_8};
    __pyx_t_7 = __Pyx_PyObject_FastCall(__pyx_t_9, __pyx_callargs+1-__pyx_t_10, 1+__pyx_t_10);
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 915, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  }
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_data, __pyx_t_7) < 0) __PYX_ERR(0, 914, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "viur/datastore/transport.pyx":913
 *         "keys": requestedKeys,
 *     }
 *     resp = authenticated_request(             # <<<<<<<<<<<<<<
 *         url="https://datastore.googleapis.com/v1/projects/%s:allocateIds" % projectID,
 *         data=json.dumps(postData).encode("UTF-8"),
 */
  __pyx_t_7 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_empty_tuple, __pyx_t_3); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 913, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_resp = __pyx_t_7;
  __pyx_t_7 = 0;

  /* "viur/datastore/transport.pyx":917
 *         data=json.dumps(postData).encode("UTF-8"),
 *     )
 *     if is_viur_datastore_request_ok(resp):             # <<<<<<<<<<<<<<
 *         assert PyBytes_AsStringAndSize(resp.content, &data_ptr, &pysize) != -1
 *         element = parser.parse(data_ptr, pysize, 1)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_is_viur_datastore_request_ok); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 917, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_1 = NULL;
  __pyx_t_10 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_1)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_10 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_1, __pyx_v_resp};
    __pyx_t_7 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_10, 1+__pyx_t_10);
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 917, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_7); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(0, 917, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (__pyx_t_2) {

    /* "viur/datastore/transport.pyx":918
 *     )
 *     if is_viur_datastore_request_ok(resp):
 *         assert PyBytes_AsStringAndSize(resp.content, &data_ptr, &pysize) != -1             # <<<<<<<<<<<<<<
 *         element = parser.parse(data_ptr, pysize, 1)
 *         res = []
 */
    #ifndef CYTHON_WITHOUT_ASSERTIONS
    if (unlikely(__pyx_assertions_enabled())) {
      __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_resp, __pyx_n_s_content); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 918, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_10 = PyBytes_AsStringAndSize(__pyx_t_7, (&__pyx_v_data_ptr), (&__pyx_v_pysize)); if (unlikely(__pyx_t_10 == ((int)-1))) __PYX_ERR(0, 918, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __pyx_t_2 = (__pyx_t_10 != -1L);
      if (unlikely(!__pyx_t_2)) {
        __Pyx_Raise(__pyx_builtin_AssertionError, 0, 0, 0);
        __PYX_ERR(0, 918, __pyx_L1_error)
      }
    }
    #else
    if ((1)); else __PYX_ERR(0, 918, __pyx_L1_error)
    #endif

    /* "viur/datastore/transport.pyx":919
 *     if is_viur_datastore_request_ok(resp):
 *         assert PyBytes_AsStringAndSize(resp.content, &data_ptr, &pysize) != -1
 *         element = parser.parse(data_ptr, pysize, 1)             # <<<<<<<<<<<<<<
 *         res = []
 *         if (element.at_pointer("/keys").error() == SUCCESS):
 */
    try {
      __pyx_t_11 = __pyx_v_parser.parse(__pyx_v_data_ptr, __pyx_v_pysize, 1);
    } catch(...) {
      __Pyx_CppExn2PyErr();
      __PYX_ERR(0, 919, __pyx_L1_error)
    }
    __pyx_v_element = __PYX_STD_MOVE_IF_SUPPORTED(__pyx_t_11);

    /* "viur/datastore/transport.pyx":920
 *         assert PyBytes_AsStringAndSize(resp.content, &data_ptr, &pysize) != -1
 *         element = parser.parse(data_ptr, pysize, 1)
 *         res = []             # <<<<<<<<<<<<<<
 *         if (element.at_pointer("/keys").error() == SUCCESS):
 *             arrayElem = element.at_key("keys").get_array()
 */
    __pyx_t_7 = PyList_New(0); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 920, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_v_res = ((PyObject*)__pyx_t_7);
    __pyx_t_7 = 0;

    /* "viur/datastore/transport.pyx":921
 *         element = parser.parse(data_ptr, pysize, 1)
 *         res = []
 *         if (element.at_pointer("/keys").error() == SUCCESS):             # <<<<<<<<<<<<<<
 *             arrayElem = element.at_key("keys").get_array()
 *             arrayIt = arrayElem.begin()
 */
    __pyx_t_2 = (__pyx_v_element.at_pointer(((char const *)"/keys")).error() == simdjson::error_code::SUCCESS);
    if (likely(__pyx_t_2)) {

      /* "viur/datastore/transport.pyx":922
 *         res = []
 *         if (element.at_pointer("/keys").error() == SUCCESS):
 *             arrayElem = element.at_key("keys").get_array()             # <<<<<<<<<<<<<<
 *             arrayIt = arrayElem.begin()
 *             while arrayIt != arrayElem.end():
 */
      try {
        __pyx_t_11 = __pyx_v_element.at_key(((char const *)"keys"));
      } catch(...) {
        __Pyx_CppExn2PyErr();
        __PYX_ERR(0, 922, __pyx_L1_error)
      }
      try {
        __pyx_t_12 = __pyx_t_11.get_array();
      } catch(...) {
        __Pyx_CppExn2PyErr();
        __PYX_ERR(0, 922, __pyx_L1_error)
      }
      __pyx_v_arrayElem = __PYX_STD_MOVE_IF_SUPPORTED(__pyx_t_12);

      /* "viur/datastore/transport.pyx":923
 *         if (element.at_pointer("/keys").error() == SUCCESS):
 *             arrayElem = element.at_key("keys").get_array()
 *             arrayIt = arrayElem.begin()             # <<<<<<<<<<<<<<
 *             while arrayIt != arrayElem.end():
 *                 innerArrayElem = dereference(arrayIt)
 */
      __pyx_v_arrayIt = __pyx_v_arrayElem.begin();

      /* "viur/datastore/transport.pyx":924
 *             arrayElem = element.at_key("keys").get_array()
 *             arrayIt = arrayElem.begin()
 *             while arrayIt != arrayElem.end():             # <<<<<<<<<<<<<<
 *                 innerArrayElem = dereference(arrayIt)
 *                 res.append(parseKey(innerArrayElem))
 */
      while (1) {
        __pyx_t_2 = (__pyx_v_arrayIt != __pyx_v_arrayElem.end());
        if (!__pyx_t_2) break;

        /* "viur/datastore/transport.pyx":925
 *             arrayIt = arrayElem.begin()
 *             while arrayIt != arrayElem.end():
 *                 innerArrayElem = dereference(arrayIt)             # <<<<<<<<<<<<<<
 *                 res.append(parseKey(innerArrayElem))
 *                 preincrement(arrayIt)
 */
        __pyx_v_innerArrayElem = (*__pyx_v_arrayIt);

        /* "viur/datastore/transport.pyx":926
 *             while arrayIt != arrayElem.end():
 *                 innerArrayElem = dereference(arrayIt)
 *                 res.append(parseKey(innerArrayElem))             # <<<<<<<<<<<<<<
 *                 preincrement(arrayIt)
 *             if not res:
 */
        __pyx_t_7 = __pyx_f_4viur_9datastore_9transport_parseKey(__pyx_v_innerArrayElem); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 926, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_13 = __Pyx_PyList_Append(__pyx_v_res, __pyx_t_7); if (unlikely(__pyx_t_13 == ((int)-1))) __PYX_ERR(0, 926, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

        /* "viur/datastore/transport.pyx":927
 *                 innerArrayElem = dereference(arrayIt)
 *                 res.append(parseKey(innerArrayElem))
 *                 preincrement(arrayIt)             # <<<<<<<<<<<<<<
 *             if not res:
 *                 raise ValueError("Empty response received from Datastore API")
 */
        (void)((++__pyx_v_arrayIt));
      }

      /* "viur/datastore/transport.pyx":928
 *                 res.append(parseKey(innerArrayElem))
 *                 preincrement(arrayIt)
 *             if not res:             # <<<<<<<<<<<<<<
 *                 raise ValueError("Empty response received from Datastore API")
 *             elif not isMulti:
 */
      __pyx_t_2 = (PyList_GET_SIZE(__pyx_v_res) != 0);
      __pyx_t_14 = (!__pyx_t_2);
      if (unlikely(__pyx_t_14)) {

        /* "viur/datastore/transport.pyx":929
 *                 preincrement(arrayIt)
 *             if not res:
 *                 raise ValueError("Empty response received from Datastore API")             # <<<<<<<<<<<<<<
 *             elif not isMulti:
 *                 return res[0]
 */
        __pyx_t_7 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__18, NULL); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 929, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_Raise(__pyx_t_7, 0, 0, 0);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __PYX_ERR(0, 929, __pyx_L1_error)

        /* "viur/datastore/transport.pyx":928
 *                 res.append(parseKey(innerArrayElem))
 *                 preincrement(arrayIt)
 *             if not res:             # <<<<<<<<<<<<<<
 *                 raise ValueError("Empty response received from Datastore API")
 *             elif not isMulti:
 */
      }

      /* "viur/datastore/transport.pyx":930
 *             if not res:
 *                 raise ValueError("Empty response received from Datastore API")
 *             elif not isMulti:             # <<<<<<<<<<<<<<
 *                 return res[0]
 *             else:
 */
      __pyx_t_14 = (!__pyx_v_isMulti);
      if (__pyx_t_14) {

        /* "viur/datastore/transport.pyx":931
 *                 raise ValueError("Empty response received from Datastore API")
 *             elif not isMulti:
 *                 return res[0]             # <<<<<<<<<<<<<<
 *             else:
 *                 return res
 */
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_7 = __Pyx_GetItemInt_List(__pyx_v_res, 0, long, 1, __Pyx_PyInt_From_long, 1, 0, 1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 931, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_r = __pyx_t_7;
        __pyx_t_7 = 0;
        goto __pyx_L0;

        /* "viur/datastore/transport.pyx":930
 *             if not res:
 *                 raise ValueError("Empty response received from Datastore API")
 *             elif not isMulti:             # <<<<<<<<<<<<<<
 *                 return res[0]
 *             else:
 */
      }

      /* "viur/datastore/transport.pyx":933
 *                 return res[0]
 *             else:
 *                 return res             # <<<<<<<<<<<<<<
 *         else:
 *             logging.error("Invalid data received from Datastore API")
 */
      /*else*/ {
        __Pyx_XDECREF(__pyx_r);
        __Pyx_INCREF(__pyx_v_res);
        __pyx_r = __pyx_v_res;
        goto __pyx_L0;
      }

      /* "viur/datastore/transport.pyx":921
 *         element = parser.parse(data_ptr, pysize, 1)
 *         res = []
 *         if (element.at_pointer("/keys").error() == SUCCESS):             # <<<<<<<<<<<<<<
 *             arrayElem = element.at_key("keys").get_array()
 *             arrayIt = arrayElem.begin()
 */
    }

    /* "viur/datastore/transport.pyx":935
 *                 return res
 *         else:
 *             logging.error("Invalid data received from Datastore API")             # <<<<<<<<<<<<<<
 *             logging.error(resp.content)
 *             raise ValueError("Invalid data received from Datastore API")
 */
    /*else*/ {
      __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_logging); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 935, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_error); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 935, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_3 = NULL;
      __pyx_t_10 = 0;
      #if CYTHON_UNPACK_METHODS
      if (unlikely(PyMethod_Check(__pyx_t_1))) {
        __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_1);
        if (likely(__pyx_t_3)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
          __Pyx_INCREF(__pyx_t_3);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_1, function);
          __pyx_t_10 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_kp_u_Invalid_data_received_from_Datas};
        __pyx_t_7 = __Pyx_PyObject_FastCall(__pyx_t_1, __pyx_callargs+1-__pyx_t_10, 1+__pyx_t_10);
        __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
        if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 935, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      }
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

      /* "viur/datastore/transport.pyx":936
 *         else:
 *             logging.error("Invalid data received from Datastore API")
 *             logging.error(resp.content)             # <<<<<<<<<<<<<<
 *             raise ValueError("Invalid data received from Datastore API")
 * 
 */
      __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_logging); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 936, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_error); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 936, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_resp, __pyx_n_s_content); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 936, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_9 = NULL;
      __pyx_t_10 = 0;
      #if CYTHON_UNPACK_METHODS
      if (unlikely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_9)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_9);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
          __pyx_t_10 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_9, __pyx_t_1};
        __pyx_t_7 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_10, 1+__pyx_t_10);
        __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 936, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      }
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

      /* "viur/datastore/transport.pyx":937
 *             logging.error("Invalid data received from Datastore API")
 *             logging.error(resp.content)
 *             raise ValueError("Invalid data received from Datastore API")             # <<<<<<<<<<<<<<
 * 
 * def Count(kind: str = None, up_to= 2 ** 63 - 1, queryDefinition: QueryDefinition = None) -> Union[Key, List[Key]]:
 */
      __pyx_t_7 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__19, NULL); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 937, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_Raise(__pyx_t_7, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __PYX_ERR(0, 937, __pyx_L1_error)
    }

    /* "viur/datastore/transport.pyx":917
 *         data=json.dumps(postData).encode("UTF-8"),
 *     )
 *     if is_viur_datastore_request_ok(resp):             # <<<<<<<<<<<<<<
 *         assert PyBytes_AsStringAndSize(resp.content, &data_ptr, &pysize) != -1
 *         element = parser.parse(data_ptr, pysize, 1)
 */
  }

  /* "viur/datastore/transport.pyx":884
 *     )
 * 
 * def AllocateIDs(keys: Union[Key, List[Key]]) -> Union[Key, List[Key]]:             # <<<<<<<<<<<<<<
 *     """
 *         Allocates numeric IDs for the keys given.
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("viur.datastore.transport.AllocateIDs", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_keyList);
  __Pyx_XDECREF(__pyx_v_requestedKeys);
  __Pyx_XDECREF(__pyx_v_postData);
  __Pyx_XDECREF(__pyx_v_resp);
  __Pyx_XDECREF(__pyx_v_res);
  __Pyx_XDECREF(__pyx_9genexpr16__pyx_v_x);
  __Pyx_XDECREF(__pyx_v_keys);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "viur/datastore/transport.pyx":939
 *             raise ValueError("Invalid data received from Datastore API")
 * 
 * def Count(kind: str = None, up_to= 2 ** 63 - 1, queryDefinition: QueryDefinition = None) -> Union[Key, List[Key]]:             # <<<<<<<<<<<<<<
 *     """
 *         Count all entries in a kind if there is only a kind is provided
 */

/* Python wrapper */
static PyObject *__pyx_pw_4viur_9datastore_9transport_21Count(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_4viur_9datastore_9transport_20Count, "\n        Count all entries in a kind if there is only a kind is provided\n        Count the entries for an given query.\n        :param kind name of the module (kind) we want to count\n        :param up_to can be sigend int 64 bit (max positive 2^31-1)\n        :param queryDefinition: The query to run\n        :return: The count as an int\n\n        .. warning: This function does not support transactions! Even if called inside transactions, the keys will\n            be allocated immediately, even if the transaction aborts.\n    ");
static PyMethodDef __pyx_mdef_4viur_9datastore_9transport_21Count = {"Count", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_4viur_9datastore_9transport_21Count, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_4viur_9datastore_9transport_20Count};
static PyObject *__pyx_pw_4viur_9datastore_9transport_21Count(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_kind = 0;
  PyObject *__pyx_v_up_to = 0;
  PyObject *__pyx_v_queryDefinition = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("Count (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_kind,&__pyx_n_s_up_to,&__pyx_n_s_queryDefinition,0};
    values[0] = __Pyx_Arg_NewRef_FASTCALL(((PyObject*)Py_None));
    values[1] = __Pyx_Arg_NewRef_FASTCALL(((PyObject *)((PyObject *)__pyx_int_0x7fffffffffffffff)));
    values[2] = __Pyx_Arg_NewRef_FASTCALL(((PyObject *)Py_None));
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_kind);
          if (value) { values[0] = __Pyx_Arg_NewRef_FASTCALL(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 939, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_up_to);
          if (value) { values[1] = __Pyx_Arg_NewRef_FASTCALL(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 939, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_queryDefinition);
          if (value) { values[2] = __Pyx_Arg_NewRef_FASTCALL(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 939, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "Count") < 0)) __PYX_ERR(0, 939, __pyx_L3_error)
      }
    } else {
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_kind = ((PyObject*)values[0]);
    __pyx_v_up_to = values[1];
    __pyx_v_queryDefinition = values[2];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("Count", 0, 0, 3, __pyx_nargs); __PYX_ERR(0, 939, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("viur.datastore.transport.Count", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_kind), (&PyUnicode_Type), 1, "kind", 1))) __PYX_ERR(0, 939, __pyx_L1_error)
  __pyx_r = __pyx_pf_4viur_9datastore_9transport_20Count(__pyx_self, __pyx_v_kind, __pyx_v_up_to, __pyx_v_queryDefinition);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4viur_9datastore_9transport_20Count(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_kind, PyObject *__pyx_v_up_to, PyObject *__pyx_v_queryDefinition) {
  simdjson::dom::parser __pyx_v_parser;
  Py_ssize_t __pyx_v_pysize;
  char *__pyx_v_data_ptr;
  simdjson::dom::element __pyx_v_element;
  PyObject *__pyx_v_post_data = NULL;
  PyObject *__pyx_v_filterList = NULL;
  PyObject *__pyx_v_k = NULL;
  PyObject *__pyx_v_v = NULL;
  PyObject *__pyx_v_key = NULL;
  PyObject *__pyx_v_op = NULL;
  PyObject *__pyx_v_singleValue = NULL;
  PyObject *__pyx_v_resp = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_t_5;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  Py_ssize_t __pyx_t_9;
  Py_ssize_t __pyx_t_10;
  int __pyx_t_11;
  PyObject *(*__pyx_t_12)(PyObject *);
  Py_ssize_t __pyx_t_13;
  PyObject *(*__pyx_t_14)(PyObject *);
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  int __pyx_t_17;
  simdjson::dom::element __pyx_t_18;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("Count", 0);
  __Pyx_INCREF(__pyx_v_kind);

  /* "viur/datastore/transport.pyx":951
 *             be allocated immediately, even if the transaction aborts.
 *     """
 *     cdef simdjsonParser parser = simdjsonParser()             # <<<<<<<<<<<<<<
 *     cdef Py_ssize_t pysize
 *     cdef char * data_ptr
 */
  __pyx_v_parser = simdjson::dom::parser();

  /* "viur/datastore/transport.pyx":957
 *     cdef simdjsonElement element_inner
 *     cdef simdjsonArray array_element, array_element_inner
 *     logging.warning(             # <<<<<<<<<<<<<<
 *         "The 'Count() aggregation' query is a technical preview and cannot be guaranteed to work at this time!!!")
 *     if not kind:
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_logging); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 957, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_warning); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 957, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  __pyx_t_4 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_4 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, __pyx_kp_u_The_Count_aggregation_query_is_a};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_4, 1+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 957, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "viur/datastore/transport.pyx":959
 *     logging.warning(
 *         "The 'Count() aggregation' query is a technical preview and cannot be guaranteed to work at this time!!!")
 *     if not kind:             # <<<<<<<<<<<<<<
 *         kind = queryDefinition.kind
 * 
 */
  __pyx_t_5 = (__pyx_v_kind != Py_None)&&(__Pyx_PyUnicode_IS_TRUE(__pyx_v_kind) != 0);
  __pyx_t_6 = (!__pyx_t_5);
  if (__pyx_t_6) {

    /* "viur/datastore/transport.pyx":960
 *         "The 'Count() aggregation' query is a technical preview and cannot be guaranteed to work at this time!!!")
 *     if not kind:
 *         kind = queryDefinition.kind             # <<<<<<<<<<<<<<
 * 
 *     post_data = {
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_queryDefinition, __pyx_n_s_kind); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 960, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (!(likely(PyUnicode_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("unicode", __pyx_t_1))) __PYX_ERR(0, 960, __pyx_L1_error)
    __Pyx_DECREF_SET(__pyx_v_kind, ((PyObject*)__pyx_t_1));
    __pyx_t_1 = 0;

    /* "viur/datastore/transport.pyx":959
 *     logging.warning(
 *         "The 'Count() aggregation' query is a technical preview and cannot be guaranteed to work at this time!!!")
 *     if not kind:             # <<<<<<<<<<<<<<
 *         kind = queryDefinition.kind
 * 
 */
  }

  /* "viur/datastore/transport.pyx":964
 *     post_data = {
 * 
 *         "partitionId": {             # <<<<<<<<<<<<<<
 *             "project_id": projectID,
 *         }, "aggregation_query": {
 */
  __pyx_t_1 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 964, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);

  /* "viur/datastore/transport.pyx":965
 * 
 *         "partitionId": {
 *             "project_id": projectID,             # <<<<<<<<<<<<<<
 *         }, "aggregation_query": {
 *             "aggregations": {"count": {"up_to": up_to}},
 */
  __pyx_t_3 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 965, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_projectID); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 965, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_u_project_id, __pyx_t_2) < 0) __PYX_ERR(0, 965, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_u_partitionId, __pyx_t_3) < 0) __PYX_ERR(0, 964, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "viur/datastore/transport.pyx":967
 *             "project_id": projectID,
 *         }, "aggregation_query": {
 *             "aggregations": {"count": {"up_to": up_to}},             # <<<<<<<<<<<<<<
 *             "nested_query": {"kind": [{"name": kind}]}
 *         },
 */
  __pyx_t_3 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 967, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 967, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_7 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 967, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_u_up_to, __pyx_v_up_to) < 0) __PYX_ERR(0, 967, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_u_count, __pyx_t_7) < 0) __PYX_ERR(0, 967, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_u_aggregations, __pyx_t_2) < 0) __PYX_ERR(0, 967, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "viur/datastore/transport.pyx":968
 *         }, "aggregation_query": {
 *             "aggregations": {"count": {"up_to": up_to}},
 *             "nested_query": {"kind": [{"name": kind}]}             # <<<<<<<<<<<<<<
 *         },
 * 
 */
  __pyx_t_2 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 968, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_7 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 968, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_u_name, __pyx_v_kind) < 0) __PYX_ERR(0, 968, __pyx_L1_error)
  __pyx_t_8 = PyList_New(1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 968, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_GIVEREF(__pyx_t_7);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_8, 0, __pyx_t_7)) __PYX_ERR(0, 968, __pyx_L1_error);
  __pyx_t_7 = 0;
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_u_kind, __pyx_t_8) < 0) __PYX_ERR(0, 968, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_u_nested_query, __pyx_t_2) < 0) __PYX_ERR(0, 967, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_u_aggregation_query, __pyx_t_3) < 0) __PYX_ERR(0, 964, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_post_data = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "viur/datastore/transport.pyx":972
 * 
 *     }
 *     if queryDefinition:             # <<<<<<<<<<<<<<
 *         if queryDefinition.filters:
 *             filterList = []
 */
  __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_v_queryDefinition); if (unlikely((__pyx_t_6 < 0))) __PYX_ERR(0, 972, __pyx_L1_error)
  if (__pyx_t_6) {

    /* "viur/datastore/transport.pyx":973
 *     }
 *     if queryDefinition:
 *         if queryDefinition.filters:             # <<<<<<<<<<<<<<
 *             filterList = []
 *             for k, v in queryDefinition.filters.items():
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_queryDefinition, __pyx_n_s_filters); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 973, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_6 < 0))) __PYX_ERR(0, 973, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (__pyx_t_6) {

      /* "viur/datastore/transport.pyx":974
 *     if queryDefinition:
 *         if queryDefinition.filters:
 *             filterList = []             # <<<<<<<<<<<<<<
 *             for k, v in queryDefinition.filters.items():
 *                 key, op = k.split(" ")
 */
      __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 974, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_v_filterList = ((PyObject*)__pyx_t_1);
      __pyx_t_1 = 0;

      /* "viur/datastore/transport.pyx":975
 *         if queryDefinition.filters:
 *             filterList = []
 *             for k, v in queryDefinition.filters.items():             # <<<<<<<<<<<<<<
 *                 key, op = k.split(" ")
 *                 if op == "=":
 */
      __pyx_t_9 = 0;
      __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_queryDefinition, __pyx_n_s_filters); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 975, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      if (unlikely(__pyx_t_3 == Py_None)) {
        PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "items");
        __PYX_ERR(0, 975, __pyx_L1_error)
      }
      __pyx_t_2 = __Pyx_dict_iterator(__pyx_t_3, 0, __pyx_n_s_items, (&__pyx_t_10), (&__pyx_t_4)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 975, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_1);
      __pyx_t_1 = __pyx_t_2;
      __pyx_t_2 = 0;
      while (1) {
        __pyx_t_11 = __Pyx_dict_iter_next(__pyx_t_1, __pyx_t_10, &__pyx_t_9, &__pyx_t_2, &__pyx_t_3, NULL, __pyx_t_4);
        if (unlikely(__pyx_t_11 == 0)) break;
        if (unlikely(__pyx_t_11 == -1)) __PYX_ERR(0, 975, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_XDECREF_SET(__pyx_v_k, __pyx_t_2);
        __pyx_t_2 = 0;
        __Pyx_XDECREF_SET(__pyx_v_v, __pyx_t_3);
        __pyx_t_3 = 0;

        /* "viur/datastore/transport.pyx":976
 *             filterList = []
 *             for k, v in queryDefinition.filters.items():
 *                 key, op = k.split(" ")             # <<<<<<<<<<<<<<
 *                 if op == "=":
 *                     op = "EQUAL"
 */
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_k, __pyx_n_s_split); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 976, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_8 = NULL;
        __pyx_t_11 = 0;
        #if CYTHON_UNPACK_METHODS
        if (likely(PyMethod_Check(__pyx_t_2))) {
          __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_2);
          if (likely(__pyx_t_8)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
            __Pyx_INCREF(__pyx_t_8);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_2, function);
            __pyx_t_11 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_8, __pyx_kp_u__5};
          __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_11, 1+__pyx_t_11);
          __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
          if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 976, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        }
        if ((likely(PyTuple_CheckExact(__pyx_t_3))) || (PyList_CheckExact(__pyx_t_3))) {
          PyObject* sequence = __pyx_t_3;
          Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
          if (unlikely(size != 2)) {
            if (size > 2) __Pyx_RaiseTooManyValuesError(2);
            else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
            __PYX_ERR(0, 976, __pyx_L1_error)
          }
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          if (likely(PyTuple_CheckExact(sequence))) {
            __pyx_t_2 = PyTuple_GET_ITEM(sequence, 0); 
            __pyx_t_8 = PyTuple_GET_ITEM(sequence, 1); 
          } else {
            __pyx_t_2 = PyList_GET_ITEM(sequence, 0); 
            __pyx_t_8 = PyList_GET_ITEM(sequence, 1); 
          }
          __Pyx_INCREF(__pyx_t_2);
          __Pyx_INCREF(__pyx_t_8);
          #else
          __pyx_t_2 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 976, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_8 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 976, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_8);
          #endif
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else {
          Py_ssize_t index = -1;
          __pyx_t_7 = PyObject_GetIter(__pyx_t_3); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 976, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __pyx_t_12 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_7);
          index = 0; __pyx_t_2 = __pyx_t_12(__pyx_t_7); if (unlikely(!__pyx_t_2)) goto __pyx_L8_unpacking_failed;
          __Pyx_GOTREF(__pyx_t_2);
          index = 1; __pyx_t_8 = __pyx_t_12(__pyx_t_7); if (unlikely(!__pyx_t_8)) goto __pyx_L8_unpacking_failed;
          __Pyx_GOTREF(__pyx_t_8);
          if (__Pyx_IternextUnpackEndCheck(__pyx_t_12(__pyx_t_7), 2) < 0) __PYX_ERR(0, 976, __pyx_L1_error)
          __pyx_t_12 = NULL;
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          goto __pyx_L9_unpacking_done;
          __pyx_L8_unpacking_failed:;
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_12 = NULL;
          if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
          __PYX_ERR(0, 976, __pyx_L1_error)
          __pyx_L9_unpacking_done:;
        }
        __Pyx_XDECREF_SET(__pyx_v_key, __pyx_t_2);
        __pyx_t_2 = 0;
        __Pyx_XDECREF_SET(__pyx_v_op, __pyx_t_8);
        __pyx_t_8 = 0;

        /* "viur/datastore/transport.pyx":977
 *             for k, v in queryDefinition.filters.items():
 *                 key, op = k.split(" ")
 *                 if op == "=":             # <<<<<<<<<<<<<<
 *                     op = "EQUAL"
 *                 elif op == "<":
 */
        __pyx_t_6 = (__Pyx_PyUnicode_Equals(__pyx_v_op, __pyx_kp_u__6, Py_EQ)); if (unlikely((__pyx_t_6 < 0))) __PYX_ERR(0, 977, __pyx_L1_error)
        if (__pyx_t_6) {

          /* "viur/datastore/transport.pyx":978
 *                 key, op = k.split(" ")
 *                 if op == "=":
 *                     op = "EQUAL"             # <<<<<<<<<<<<<<
 *                 elif op == "<":
 *                     op = "LESS_THAN"
 */
          __Pyx_INCREF(__pyx_n_u_EQUAL);
          __Pyx_DECREF_SET(__pyx_v_op, __pyx_n_u_EQUAL);

          /* "viur/datastore/transport.pyx":977
 *             for k, v in queryDefinition.filters.items():
 *                 key, op = k.split(" ")
 *                 if op == "=":             # <<<<<<<<<<<<<<
 *                     op = "EQUAL"
 *                 elif op == "<":
 */
          goto __pyx_L10;
        }

        /* "viur/datastore/transport.pyx":979
 *                 if op == "=":
 *                     op = "EQUAL"
 *                 elif op == "<":             # <<<<<<<<<<<<<<
 *                     op = "LESS_THAN"
 *                 elif op == "<=":
 */
        __pyx_t_6 = (__Pyx_PyUnicode_Equals(__pyx_v_op, __pyx_kp_u__7, Py_EQ)); if (unlikely((__pyx_t_6 < 0))) __PYX_ERR(0, 979, __pyx_L1_error)
        if (__pyx_t_6) {

          /* "viur/datastore/transport.pyx":980
 *                     op = "EQUAL"
 *                 elif op == "<":
 *                     op = "LESS_THAN"             # <<<<<<<<<<<<<<
 *                 elif op == "<=":
 *                     op = "LESS_THAN_OR_EQUAL"
 */
          __Pyx_INCREF(__pyx_n_u_LESS_THAN);
          __Pyx_DECREF_SET(__pyx_v_op, __pyx_n_u_LESS_THAN);

          /* "viur/datastore/transport.pyx":979
 *                 if op == "=":
 *                     op = "EQUAL"
 *                 elif op == "<":             # <<<<<<<<<<<<<<
 *                     op = "LESS_THAN"
 *                 elif op == "<=":
 */
          goto __pyx_L10;
        }

        /* "viur/datastore/transport.pyx":981
 *                 elif op == "<":
 *                     op = "LESS_THAN"
 *                 elif op == "<=":             # <<<<<<<<<<<<<<
 *                     op = "LESS_THAN_OR_EQUAL"
 *                 elif op == ">":
 */
        __pyx_t_6 = (__Pyx_PyUnicode_Equals(__pyx_v_op, __pyx_kp_u__8, Py_EQ)); if (unlikely((__pyx_t_6 < 0))) __PYX_ERR(0, 981, __pyx_L1_error)
        if (__pyx_t_6) {

          /* "viur/datastore/transport.pyx":982
 *                     op = "LESS_THAN"
 *                 elif op == "<=":
 *                     op = "LESS_THAN_OR_EQUAL"             # <<<<<<<<<<<<<<
 *                 elif op == ">":
 *                     op = "GREATER_THAN"
 */
          __Pyx_INCREF(__pyx_n_u_LESS_THAN_OR_EQUAL);
          __Pyx_DECREF_SET(__pyx_v_op, __pyx_n_u_LESS_THAN_OR_EQUAL);

          /* "viur/datastore/transport.pyx":981
 *                 elif op == "<":
 *                     op = "LESS_THAN"
 *                 elif op == "<=":             # <<<<<<<<<<<<<<
 *                     op = "LESS_THAN_OR_EQUAL"
 *                 elif op == ">":
 */
          goto __pyx_L10;
        }

        /* "viur/datastore/transport.pyx":983
 *                 elif op == "<=":
 *                     op = "LESS_THAN_OR_EQUAL"
 *                 elif op == ">":             # <<<<<<<<<<<<<<
 *                     op = "GREATER_THAN"
 *                 elif op == ">=":
 */
        __pyx_t_6 = (__Pyx_PyUnicode_Equals(__pyx_v_op, __pyx_kp_u__9, Py_EQ)); if (unlikely((__pyx_t_6 < 0))) __PYX_ERR(0, 983, __pyx_L1_error)
        if (__pyx_t_6) {

          /* "viur/datastore/transport.pyx":984
 *                     op = "LESS_THAN_OR_EQUAL"
 *                 elif op == ">":
 *                     op = "GREATER_THAN"             # <<<<<<<<<<<<<<
 *                 elif op == ">=":
 *                     op = "GREATER_THAN_OR_EQUAL"
 */
          __Pyx_INCREF(__pyx_n_u_GREATER_THAN);
          __Pyx_DECREF_SET(__pyx_v_op, __pyx_n_u_GREATER_THAN);

          /* "viur/datastore/transport.pyx":983
 *                 elif op == "<=":
 *                     op = "LESS_THAN_OR_EQUAL"
 *                 elif op == ">":             # <<<<<<<<<<<<<<
 *                     op = "GREATER_THAN"
 *                 elif op == ">=":
 */
          goto __pyx_L10;
        }

        /* "viur/datastore/transport.pyx":985
 *                 elif op == ">":
 *                     op = "GREATER_THAN"
 *                 elif op == ">=":             # <<<<<<<<<<<<<<
 *                     op = "GREATER_THAN_OR_EQUAL"
 *                 else:
 */
        __pyx_t_6 = (__Pyx_PyUnicode_Equals(__pyx_v_op, __pyx_kp_u__10, Py_EQ)); if (unlikely((__pyx_t_6 < 0))) __PYX_ERR(0, 985, __pyx_L1_error)
        if (likely(__pyx_t_6)) {

          /* "viur/datastore/transport.pyx":986
 *                     op = "GREATER_THAN"
 *                 elif op == ">=":
 *                     op = "GREATER_THAN_OR_EQUAL"             # <<<<<<<<<<<<<<
 *                 else:
 *                     raise ValueError("Invalid op %s" % op)
 */
          __Pyx_INCREF(__pyx_n_u_GREATER_THAN_OR_EQUAL);
          __Pyx_DECREF_SET(__pyx_v_op, __pyx_n_u_GREATER_THAN_OR_EQUAL);

          /* "viur/datastore/transport.pyx":985
 *                 elif op == ">":
 *                     op = "GREATER_THAN"
 *                 elif op == ">=":             # <<<<<<<<<<<<<<
 *                     op = "GREATER_THAN_OR_EQUAL"
 *                 else:
 */
          goto __pyx_L10;
        }

        /* "viur/datastore/transport.pyx":988
 *                     op = "GREATER_THAN_OR_EQUAL"
 *                 else:
 *                     raise ValueError("Invalid op %s" % op)             # <<<<<<<<<<<<<<
 *                 if not isinstance(v, list):
 *                     # An entity can have a list of values for a single property, so it's possible to enforce
 */
        /*else*/ {
          __pyx_t_3 = __Pyx_PyUnicode_FormatSafe(__pyx_kp_u_Invalid_op_s, __pyx_v_op); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 988, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_8 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_3); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 988, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_8);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_Raise(__pyx_t_8, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
          __PYX_ERR(0, 988, __pyx_L1_error)
        }
        __pyx_L10:;

        /* "viur/datastore/transport.pyx":989
 *                 else:
 *                     raise ValueError("Invalid op %s" % op)
 *                 if not isinstance(v, list):             # <<<<<<<<<<<<<<
 *                     # An entity can have a list of values for a single property, so it's possible to enforce
 *                     # more an one constraint for a a single property (e.g. x==5 and x==7 can be true), so
 */
        __pyx_t_6 = PyList_Check(__pyx_v_v); 
        __pyx_t_5 = (!__pyx_t_6);
        if (__pyx_t_5) {

          /* "viur/datastore/transport.pyx":993
 *                     # more an one constraint for a a single property (e.g. x==5 and x==7 can be true), so
 *                     # enforce we always have a list here
 *                     v = [v]             # <<<<<<<<<<<<<<
 *                 for singleValue in v:
 *                     filterList.append({
 */
          __pyx_t_8 = PyList_New(1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 993, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_8);
          __Pyx_INCREF(__pyx_v_v);
          __Pyx_GIVEREF(__pyx_v_v);
          if (__Pyx_PyList_SET_ITEM(__pyx_t_8, 0, __pyx_v_v)) __PYX_ERR(0, 993, __pyx_L1_error);
          __Pyx_DECREF_SET(__pyx_v_v, __pyx_t_8);
          __pyx_t_8 = 0;

          /* "viur/datastore/transport.pyx":989
 *                 else:
 *                     raise ValueError("Invalid op %s" % op)
 *                 if not isinstance(v, list):             # <<<<<<<<<<<<<<
 *                     # An entity can have a list of values for a single property, so it's possible to enforce
 *                     # more an one constraint for a a single property (e.g. x==5 and x==7 can be true), so
 */
        }

        /* "viur/datastore/transport.pyx":994
 *                     # enforce we always have a list here
 *                     v = [v]
 *                 for singleValue in v:             # <<<<<<<<<<<<<<
 *                     filterList.append({
 *                         "propertyFilter": {
 */
        if (likely(PyList_CheckExact(__pyx_v_v)) || PyTuple_CheckExact(__pyx_v_v)) {
          __pyx_t_8 = __pyx_v_v; __Pyx_INCREF(__pyx_t_8);
          __pyx_t_13 = 0;
          __pyx_t_14 = NULL;
        } else {
          __pyx_t_13 = -1; __pyx_t_8 = PyObject_GetIter(__pyx_v_v); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 994, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_8);
          __pyx_t_14 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_8); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 994, __pyx_L1_error)
        }
        for (;;) {
          if (likely(!__pyx_t_14)) {
            if (likely(PyList_CheckExact(__pyx_t_8))) {
              {
                Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_8);
                #if !CYTHON_ASSUME_SAFE_MACROS
                if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 994, __pyx_L1_error)
                #endif
                if (__pyx_t_13 >= __pyx_temp) break;
              }
              #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
              __pyx_t_3 = PyList_GET_ITEM(__pyx_t_8, __pyx_t_13); __Pyx_INCREF(__pyx_t_3); __pyx_t_13++; if (unlikely((0 < 0))) __PYX_ERR(0, 994, __pyx_L1_error)
              #else
              __pyx_t_3 = __Pyx_PySequence_ITEM(__pyx_t_8, __pyx_t_13); __pyx_t_13++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 994, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_3);
              #endif
            } else {
              {
                Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_8);
                #if !CYTHON_ASSUME_SAFE_MACROS
                if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 994, __pyx_L1_error)
                #endif
                if (__pyx_t_13 >= __pyx_temp) break;
              }
              #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
              __pyx_t_3 = PyTuple_GET_ITEM(__pyx_t_8, __pyx_t_13); __Pyx_INCREF(__pyx_t_3); __pyx_t_13++; if (unlikely((0 < 0))) __PYX_ERR(0, 994, __pyx_L1_error)
              #else
              __pyx_t_3 = __Pyx_PySequence_ITEM(__pyx_t_8, __pyx_t_13); __pyx_t_13++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 994, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_3);
              #endif
            }
          } else {
            __pyx_t_3 = __pyx_t_14(__pyx_t_8);
            if (unlikely(!__pyx_t_3)) {
              PyObject* exc_type = PyErr_Occurred();
              if (exc_type) {
                if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
                else __PYX_ERR(0, 994, __pyx_L1_error)
              }
              break;
            }
            __Pyx_GOTREF(__pyx_t_3);
          }
          __Pyx_XDECREF_SET(__pyx_v_singleValue, __pyx_t_3);
          __pyx_t_3 = 0;

          /* "viur/datastore/transport.pyx":996
 *                 for singleValue in v:
 *                     filterList.append({
 *                         "propertyFilter": {             # <<<<<<<<<<<<<<
 *                             "property": {
 *                                 "name": key,
 */
          __pyx_t_3 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 996, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_3);

          /* "viur/datastore/transport.pyx":997
 *                     filterList.append({
 *                         "propertyFilter": {
 *                             "property": {             # <<<<<<<<<<<<<<
 *                                 "name": key,
 *                             },
 */
          __pyx_t_2 = __Pyx_PyDict_NewPresized(3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 997, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);

          /* "viur/datastore/transport.pyx":998
 *                         "propertyFilter": {
 *                             "property": {
 *                                 "name": key,             # <<<<<<<<<<<<<<
 *                             },
 *                             "op": op,
 */
          __pyx_t_7 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 998, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          if (PyDict_SetItem(__pyx_t_7, __pyx_n_u_name, __pyx_v_key) < 0) __PYX_ERR(0, 998, __pyx_L1_error)
          if (PyDict_SetItem(__pyx_t_2, __pyx_n_u_property, __pyx_t_7) < 0) __PYX_ERR(0, 997, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

          /* "viur/datastore/transport.pyx":1000
 *                                 "name": key,
 *                             },
 *                             "op": op,             # <<<<<<<<<<<<<<
 *                             "value": pythonPropToJson(singleValue)
 *                         }
 */
          if (PyDict_SetItem(__pyx_t_2, __pyx_n_u_op, __pyx_v_op) < 0) __PYX_ERR(0, 997, __pyx_L1_error)

          /* "viur/datastore/transport.pyx":1001
 *                             },
 *                             "op": op,
 *                             "value": pythonPropToJson(singleValue)             # <<<<<<<<<<<<<<
 *                         }
 *                     })
 */
          __Pyx_GetModuleGlobalName(__pyx_t_15, __pyx_n_s_pythonPropToJson); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 1001, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_15);
          __pyx_t_16 = NULL;
          __pyx_t_11 = 0;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_15))) {
            __pyx_t_16 = PyMethod_GET_SELF(__pyx_t_15);
            if (likely(__pyx_t_16)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_15);
              __Pyx_INCREF(__pyx_t_16);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_15, function);
              __pyx_t_11 = 1;
            }
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_16, __pyx_v_singleValue};
            __pyx_t_7 = __Pyx_PyObject_FastCall(__pyx_t_15, __pyx_callargs+1-__pyx_t_11, 1+__pyx_t_11);
            __Pyx_XDECREF(__pyx_t_16); __pyx_t_16 = 0;
            if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1001, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_7);
            __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
          }
          if (PyDict_SetItem(__pyx_t_2, __pyx_n_u_value, __pyx_t_7) < 0) __PYX_ERR(0, 997, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          if (PyDict_SetItem(__pyx_t_3, __pyx_n_u_propertyFilter, __pyx_t_2) < 0) __PYX_ERR(0, 996, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

          /* "viur/datastore/transport.pyx":995
 *                     v = [v]
 *                 for singleValue in v:
 *                     filterList.append({             # <<<<<<<<<<<<<<
 *                         "propertyFilter": {
 *                             "property": {
 */
          __pyx_t_17 = __Pyx_PyList_Append(__pyx_v_filterList, __pyx_t_3); if (unlikely(__pyx_t_17 == ((int)-1))) __PYX_ERR(0, 995, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "viur/datastore/transport.pyx":994
 *                     # enforce we always have a list here
 *                     v = [v]
 *                 for singleValue in v:             # <<<<<<<<<<<<<<
 *                     filterList.append({
 *                         "propertyFilter": {
 */
        }
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      }
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "viur/datastore/transport.pyx":1004
 *                         }
 *                     })
 *             if len(filterList) == 1:  # Special, single filter             # <<<<<<<<<<<<<<
 *                 post_data["aggregation_query"]["nested_query"]["filter"] = filterList[0]
 *             else:
 */
      __pyx_t_10 = __Pyx_PyList_GET_SIZE(__pyx_v_filterList); if (unlikely(__pyx_t_10 == ((Py_ssize_t)-1))) __PYX_ERR(0, 1004, __pyx_L1_error)
      __pyx_t_5 = (__pyx_t_10 == 1);
      if (__pyx_t_5) {

        /* "viur/datastore/transport.pyx":1005
 *                     })
 *             if len(filterList) == 1:  # Special, single filter
 *                 post_data["aggregation_query"]["nested_query"]["filter"] = filterList[0]             # <<<<<<<<<<<<<<
 *             else:
 *                 post_data["aggregation_query"]["nested_query"]["filter"] = {
 */
        __pyx_t_1 = __Pyx_GetItemInt_List(__pyx_v_filterList, 0, long, 1, __Pyx_PyInt_From_long, 1, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1005, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_8 = __Pyx_PyDict_GetItem(__pyx_v_post_data, __pyx_n_u_aggregation_query); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1005, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_8);
        __pyx_t_3 = __Pyx_PyObject_Dict_GetItem(__pyx_t_8, __pyx_n_u_nested_query); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1005, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        if (unlikely((PyObject_SetItem(__pyx_t_3, __pyx_n_u_filter, __pyx_t_1) < 0))) __PYX_ERR(0, 1005, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "viur/datastore/transport.pyx":1004
 *                         }
 *                     })
 *             if len(filterList) == 1:  # Special, single filter             # <<<<<<<<<<<<<<
 *                 post_data["aggregation_query"]["nested_query"]["filter"] = filterList[0]
 *             else:
 */
        goto __pyx_L15;
      }

      /* "viur/datastore/transport.pyx":1008
 *             else:
 *                 post_data["aggregation_query"]["nested_query"]["filter"] = {
 *                     "compositeFilter": {             # <<<<<<<<<<<<<<
 *                         "op": "AND",
 *                         "filters": filterList
 */
      /*else*/ {
        __pyx_t_1 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1008, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);

        /* "viur/datastore/transport.pyx":1009
 *                 post_data["aggregation_query"]["nested_query"]["filter"] = {
 *                     "compositeFilter": {
 *                         "op": "AND",             # <<<<<<<<<<<<<<
 *                         "filters": filterList
 *                     }
 */
        __pyx_t_3 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1009, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        if (PyDict_SetItem(__pyx_t_3, __pyx_n_u_op, __pyx_n_u_AND) < 0) __PYX_ERR(0, 1009, __pyx_L1_error)

        /* "viur/datastore/transport.pyx":1010
 *                     "compositeFilter": {
 *                         "op": "AND",
 *                         "filters": filterList             # <<<<<<<<<<<<<<
 *                     }
 *                 }
 */
        if (PyDict_SetItem(__pyx_t_3, __pyx_n_u_filters, __pyx_v_filterList) < 0) __PYX_ERR(0, 1009, __pyx_L1_error)
        if (PyDict_SetItem(__pyx_t_1, __pyx_n_u_compositeFilter, __pyx_t_3) < 0) __PYX_ERR(0, 1008, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

        /* "viur/datastore/transport.pyx":1007
 *                 post_data["aggregation_query"]["nested_query"]["filter"] = filterList[0]
 *             else:
 *                 post_data["aggregation_query"]["nested_query"]["filter"] = {             # <<<<<<<<<<<<<<
 *                     "compositeFilter": {
 *                         "op": "AND",
 */
        __pyx_t_3 = __Pyx_PyDict_GetItem(__pyx_v_post_data, __pyx_n_u_aggregation_query); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1007, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_8 = __Pyx_PyObject_Dict_GetItem(__pyx_t_3, __pyx_n_u_nested_query); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1007, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_8);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        if (unlikely((PyObject_SetItem(__pyx_t_8, __pyx_n_u_filter, __pyx_t_1) < 0))) __PYX_ERR(0, 1007, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      }
      __pyx_L15:;

      /* "viur/datastore/transport.pyx":973
 *     }
 *     if queryDefinition:
 *         if queryDefinition.filters:             # <<<<<<<<<<<<<<
 *             filterList = []
 *             for k, v in queryDefinition.filters.items():
 */
    }

    /* "viur/datastore/transport.pyx":972
 * 
 *     }
 *     if queryDefinition:             # <<<<<<<<<<<<<<
 *         if queryDefinition.filters:
 *             filterList = []
 */
  }

  /* "viur/datastore/transport.pyx":1013
 *                     }
 *                 }
 *     resp = authenticated_request(             # <<<<<<<<<<<<<<
 *         url="https://datastore.googleapis.com/v1/projects/%s:runAggregationQuery" % projectID,
 *         data=json.dumps(post_data).encode("UTF-8"),
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_authenticated_request); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1013, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);

  /* "viur/datastore/transport.pyx":1014
 *                 }
 *     resp = authenticated_request(
 *         url="https://datastore.googleapis.com/v1/projects/%s:runAggregationQuery" % projectID,             # <<<<<<<<<<<<<<
 *         data=json.dumps(post_data).encode("UTF-8"),
 *     )
 */
  __pyx_t_8 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1014, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_projectID); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1014, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __Pyx_PyUnicode_FormatSafe(__pyx_kp_u_https_datastore_googleapis_com_v_7, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1014, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (PyDict_SetItem(__pyx_t_8, __pyx_n_s_url, __pyx_t_2) < 0) __PYX_ERR(0, 1014, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "viur/datastore/transport.pyx":1015
 *     resp = authenticated_request(
 *         url="https://datastore.googleapis.com/v1/projects/%s:runAggregationQuery" % projectID,
 *         data=json.dumps(post_data).encode("UTF-8"),             # <<<<<<<<<<<<<<
 *     )
 *     if is_viur_datastore_request_ok(resp):
 */
  __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_json); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1015, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_15 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_dumps); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 1015, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = NULL;
  __pyx_t_4 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_15))) {
    __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_15);
    if (likely(__pyx_t_7)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_15);
      __Pyx_INCREF(__pyx_t_7);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_15, function);
      __pyx_t_4 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_v_post_data};
    __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_15, __pyx_callargs+1-__pyx_t_4, 1+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1015, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
  }
  __pyx_t_15 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_encode); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 1015, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_15))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_15);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_15);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_15, function);
      __pyx_t_4 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_kp_u_UTF_8};
    __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_15, __pyx_callargs+1-__pyx_t_4, 1+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1015, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
  }
  if (PyDict_SetItem(__pyx_t_8, __pyx_n_s_data, __pyx_t_2) < 0) __PYX_ERR(0, 1014, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "viur/datastore/transport.pyx":1013
 *                     }
 *                 }
 *     resp = authenticated_request(             # <<<<<<<<<<<<<<
 *         url="https://datastore.googleapis.com/v1/projects/%s:runAggregationQuery" % projectID,
 *         data=json.dumps(post_data).encode("UTF-8"),
 */
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_empty_tuple, __pyx_t_8); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1013, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_v_resp = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "viur/datastore/transport.pyx":1017
 *         data=json.dumps(post_data).encode("UTF-8"),
 *     )
 *     if is_viur_datastore_request_ok(resp):             # <<<<<<<<<<<<<<
 *         assert PyBytes_AsStringAndSize(resp.content, &data_ptr, &pysize) != -1
 *         element = parser.parse(data_ptr, pysize, 1)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_8, __pyx_n_s_is_viur_datastore_request_ok); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1017, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_1 = NULL;
  __pyx_t_4 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_8))) {
    __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_8);
    if (likely(__pyx_t_1)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_8, function);
      __pyx_t_4 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_1, __pyx_v_resp};
    __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_8, __pyx_callargs+1-__pyx_t_4, 1+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1017, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  }
  __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely((__pyx_t_5 < 0))) __PYX_ERR(0, 1017, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__pyx_t_5) {

    /* "viur/datastore/transport.pyx":1018
 *     )
 *     if is_viur_datastore_request_ok(resp):
 *         assert PyBytes_AsStringAndSize(resp.content, &data_ptr, &pysize) != -1             # <<<<<<<<<<<<<<
 *         element = parser.parse(data_ptr, pysize, 1)
 *         if element.at_pointer("/batch").error() != SUCCESS:
 */
    #ifndef CYTHON_WITHOUT_ASSERTIONS
    if (unlikely(__pyx_assertions_enabled())) {
      __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_resp, __pyx_n_s_content); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1018, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_4 = PyBytes_AsStringAndSize(__pyx_t_2, (&__pyx_v_data_ptr), (&__pyx_v_pysize)); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(0, 1018, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_5 = (__pyx_t_4 != -1L);
      if (unlikely(!__pyx_t_5)) {
        __Pyx_Raise(__pyx_builtin_AssertionError, 0, 0, 0);
        __PYX_ERR(0, 1018, __pyx_L1_error)
      }
    }
    #else
    if ((1)); else __PYX_ERR(0, 1018, __pyx_L1_error)
    #endif

    /* "viur/datastore/transport.pyx":1019
 *     if is_viur_datastore_request_ok(resp):
 *         assert PyBytes_AsStringAndSize(resp.content, &data_ptr, &pysize) != -1
 *         element = parser.parse(data_ptr, pysize, 1)             # <<<<<<<<<<<<<<
 *         if element.at_pointer("/batch").error() != SUCCESS:
 *             logging.error("INVALID RESPONSE RECEIVED")
 */
    try {
      __pyx_t_18 = __pyx_v_parser.parse(__pyx_v_data_ptr, __pyx_v_pysize, 1);
    } catch(...) {
      __Pyx_CppExn2PyErr();
      __PYX_ERR(0, 1019, __pyx_L1_error)
    }
    __pyx_v_element = __PYX_STD_MOVE_IF_SUPPORTED(__pyx_t_18);

    /* "viur/datastore/transport.pyx":1020
 *         assert PyBytes_AsStringAndSize(resp.content, &data_ptr, &pysize) != -1
 *         element = parser.parse(data_ptr, pysize, 1)
 *         if element.at_pointer("/batch").error() != SUCCESS:             # <<<<<<<<<<<<<<
 *             logging.error("INVALID RESPONSE RECEIVED")
 *             logging.error(json.loads(resp.content))
 */
    __pyx_t_5 = (__pyx_v_element.at_pointer(((char const *)"/batch")).error() != simdjson::error_code::SUCCESS);
    if (__pyx_t_5) {

      /* "viur/datastore/transport.pyx":1021
 *         element = parser.parse(data_ptr, pysize, 1)
 *         if element.at_pointer("/batch").error() != SUCCESS:
 *             logging.error("INVALID RESPONSE RECEIVED")             # <<<<<<<<<<<<<<
 *             logging.error(json.loads(resp.content))
 *         element = element.at_key("batch")
 */
      __Pyx_GetModuleGlobalName(__pyx_t_8, __pyx_n_s_logging); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1021, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_error); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1021, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __pyx_t_8 = NULL;
      __pyx_t_4 = 0;
      #if CYTHON_UNPACK_METHODS
      if (unlikely(PyMethod_Check(__pyx_t_1))) {
        __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_1);
        if (likely(__pyx_t_8)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
          __Pyx_INCREF(__pyx_t_8);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_1, function);
          __pyx_t_4 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_8, __pyx_kp_u_INVALID_RESPONSE_RECEIVED};
        __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_1, __pyx_callargs+1-__pyx_t_4, 1+__pyx_t_4);
        __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1021, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      }
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

      /* "viur/datastore/transport.pyx":1022
 *         if element.at_pointer("/batch").error() != SUCCESS:
 *             logging.error("INVALID RESPONSE RECEIVED")
 *             logging.error(json.loads(resp.content))             # <<<<<<<<<<<<<<
 *         element = element.at_key("batch")
 *         # TODO  maybe this can be solved more elegant
 */
      __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_logging); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1022, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_error); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1022, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_GetModuleGlobalName(__pyx_t_15, __pyx_n_s_json); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 1022, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_15);
      __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_15, __pyx_n_s_loads); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1022, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
      __pyx_t_15 = __Pyx_PyObject_GetAttrStr(__pyx_v_resp, __pyx_n_s_content); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 1022, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_15);
      __pyx_t_7 = NULL;
      __pyx_t_4 = 0;
      #if CYTHON_UNPACK_METHODS
      if (unlikely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_7)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_7);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
          __pyx_t_4 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_t_15};
        __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_4, 1+__pyx_t_4);
        __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
        if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1022, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      }
      __pyx_t_3 = NULL;
      __pyx_t_4 = 0;
      #if CYTHON_UNPACK_METHODS
      if (unlikely(PyMethod_Check(__pyx_t_8))) {
        __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_8);
        if (likely(__pyx_t_3)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
          __Pyx_INCREF(__pyx_t_3);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_8, function);
          __pyx_t_4 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_t_1};
        __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_8, __pyx_callargs+1-__pyx_t_4, 1+__pyx_t_4);
        __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1022, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      }
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

      /* "viur/datastore/transport.pyx":1020
 *         assert PyBytes_AsStringAndSize(resp.content, &data_ptr, &pysize) != -1
 *         element = parser.parse(data_ptr, pysize, 1)
 *         if element.at_pointer("/batch").error() != SUCCESS:             # <<<<<<<<<<<<<<
 *             logging.error("INVALID RESPONSE RECEIVED")
 *             logging.error(json.loads(resp.content))
 */
    }

    /* "viur/datastore/transport.pyx":1023
 *             logging.error("INVALID RESPONSE RECEIVED")
 *             logging.error(json.loads(resp.content))
 *         element = element.at_key("batch")             # <<<<<<<<<<<<<<
 *         # TODO  maybe this can be solved more elegant
 *         return int(
 */
    try {
      __pyx_t_18 = __pyx_v_element.at_key(((char const *)"batch"));
    } catch(...) {
      __Pyx_CppExn2PyErr();
      __PYX_ERR(0, 1023, __pyx_L1_error)
    }
    __pyx_v_element = __PYX_STD_MOVE_IF_SUPPORTED(__pyx_t_18);

    /* "viur/datastore/transport.pyx":1025
 *         element = element.at_key("batch")
 *         # TODO  maybe this can be solved more elegant
 *         return int(             # <<<<<<<<<<<<<<
 *             toPythonStructure(element)["aggregationResults"][0]["aggregateProperties"]["property_1"]["integerValue"])
 */
    __Pyx_XDECREF(__pyx_r);

    /* "viur/datastore/transport.pyx":1026
 *         # TODO  maybe this can be solved more elegant
 *         return int(
 *             toPythonStructure(element)["aggregationResults"][0]["aggregateProperties"]["property_1"]["integerValue"])             # <<<<<<<<<<<<<<
 */
    __pyx_t_2 = __pyx_f_4viur_9datastore_9transport_toPythonStructure(__pyx_v_element); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1026, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_8 = __Pyx_PyObject_Dict_GetItem(__pyx_t_2, __pyx_n_u_aggregationResults); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1026, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_GetItemInt(__pyx_t_8, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1026, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __pyx_t_8 = __Pyx_PyObject_Dict_GetItem(__pyx_t_2, __pyx_n_u_aggregateProperties); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1026, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyObject_Dict_GetItem(__pyx_t_8, __pyx_n_u_property_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1026, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __pyx_t_8 = __Pyx_PyObject_Dict_GetItem(__pyx_t_2, __pyx_n_u_integerValue); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1026, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "viur/datastore/transport.pyx":1025
 *         element = element.at_key("batch")
 *         # TODO  maybe this can be solved more elegant
 *         return int(             # <<<<<<<<<<<<<<
 *             toPythonStructure(element)["aggregationResults"][0]["aggregateProperties"]["property_1"]["integerValue"])
 */
    __pyx_t_2 = __Pyx_PyNumber_Int(__pyx_t_8); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1025, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L0;

    /* "viur/datastore/transport.pyx":1017
 *         data=json.dumps(post_data).encode("UTF-8"),
 *     )
 *     if is_viur_datastore_request_ok(resp):             # <<<<<<<<<<<<<<
 *         assert PyBytes_AsStringAndSize(resp.content, &data_ptr, &pysize) != -1
 *         element = parser.parse(data_ptr, pysize, 1)
 */
  }

  /* "viur/datastore/transport.pyx":939
 *             raise ValueError("Invalid data received from Datastore API")
 * 
 * def Count(kind: str = None, up_to= 2 ** 63 - 1, queryDefinition: QueryDefinition = None) -> Union[Key, List[Key]]:             # <<<<<<<<<<<<<<
 *     """
 *         Count all entries in a kind if there is only a kind is provided
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_15);
  __Pyx_XDECREF(__pyx_t_16);
  __Pyx_AddTraceback("viur.datastore.transport.Count", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_post_data);
  __Pyx_XDECREF(__pyx_v_filterList);
  __Pyx_XDECREF(__pyx_v_k);
  __Pyx_XDECREF(__pyx_v_v);
  __Pyx_XDECREF(__pyx_v_key);
  __Pyx_XDECREF(__pyx_v_op);
  __Pyx_XDECREF(__pyx_v_singleValue);
  __Pyx_XDECREF(__pyx_v_resp);
  __Pyx_XDECREF(__pyx_v_kind);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

#if CYTHON_USE_FREELISTS
static struct __pyx_obj_4viur_9datastore_9transport___pyx_scope_struct__genexpr *__pyx_freelist_4viur_9datastore_9transport___pyx_scope_struct__genexpr[8];
static int __pyx_freecount_4viur_9datastore_9transport___pyx_scope_struct__genexpr = 0;
#endif

static PyObject *__pyx_tp_new_4viur_9datastore_9transport___pyx_scope_struct__genexpr(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  #if CYTHON_USE_FREELISTS
  if (likely((int)(__pyx_freecount_4viur_9datastore_9transport___pyx_scope_struct__genexpr > 0) & (int)(t->tp_basicsize == sizeof(struct __pyx_obj_4viur_9datastore_9transport___pyx_scope_struct__genexpr)))) {
    o = (PyObject*)__pyx_freelist_4viur_9datastore_9transport___pyx_scope_struct__genexpr[--__pyx_freecount_4viur_9datastore_9transport___pyx_scope_struct__genexpr];
    memset(o, 0, sizeof(struct __pyx_obj_4viur_9datastore_9transport___pyx_scope_struct__genexpr));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else
  #endif
  {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  #endif
  return o;
}

static void __pyx_tp_dealloc_4viur_9datastore_9transport___pyx_scope_struct__genexpr(PyObject *o) {
  struct __pyx_obj_4viur_9datastore_9transport___pyx_scope_struct__genexpr *p = (struct __pyx_obj_4viur_9datastore_9transport___pyx_scope_struct__genexpr *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_4viur_9datastore_9transport___pyx_scope_struct__genexpr) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_genexpr_arg_0);
  Py_CLEAR(p->__pyx_v_key);
  #if CYTHON_USE_FREELISTS
  if (((int)(__pyx_freecount_4viur_9datastore_9transport___pyx_scope_struct__genexpr < 8) & (int)(Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_4viur_9datastore_9transport___pyx_scope_struct__genexpr)))) {
    __pyx_freelist_4viur_9datastore_9transport___pyx_scope_struct__genexpr[__pyx_freecount_4viur_9datastore_9transport___pyx_scope_struct__genexpr++] = ((struct __pyx_obj_4viur_9datastore_9transport___pyx_scope_struct__genexpr *)o);
  } else
  #endif
  {
    #if CYTHON_USE_TYPE_SLOTS || CYTHON_COMPILING_IN_PYPY
    (*Py_TYPE(o)->tp_free)(o);
    #else
    {
      freefunc tp_free = (freefunc)PyType_GetSlot(Py_TYPE(o), Py_tp_free);
      if (tp_free) tp_free(o);
    }
    #endif
  }
}

static int __pyx_tp_traverse_4viur_9datastore_9transport___pyx_scope_struct__genexpr(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_4viur_9datastore_9transport___pyx_scope_struct__genexpr *p = (struct __pyx_obj_4viur_9datastore_9transport___pyx_scope_struct__genexpr *)o;
  if (p->__pyx_genexpr_arg_0) {
    e = (*v)(p->__pyx_genexpr_arg_0, a); if (e) return e;
  }
  if (p->__pyx_v_key) {
    e = (*v)(p->__pyx_v_key, a); if (e) return e;
  }
  return 0;
}
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_4viur_9datastore_9transport___pyx_scope_struct__genexpr_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_4viur_9datastore_9transport___pyx_scope_struct__genexpr},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_4viur_9datastore_9transport___pyx_scope_struct__genexpr},
  {Py_tp_new, (void *)__pyx_tp_new_4viur_9datastore_9transport___pyx_scope_struct__genexpr},
  {0, 0},
};
static PyType_Spec __pyx_type_4viur_9datastore_9transport___pyx_scope_struct__genexpr_spec = {
  "viur.datastore.transport.__pyx_scope_struct__genexpr",
  sizeof(struct __pyx_obj_4viur_9datastore_9transport___pyx_scope_struct__genexpr),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC|Py_TPFLAGS_HAVE_FINALIZE,
  __pyx_type_4viur_9datastore_9transport___pyx_scope_struct__genexpr_slots,
};
#else

static PyTypeObject __pyx_type_4viur_9datastore_9transport___pyx_scope_struct__genexpr = {
  PyVarObject_HEAD_INIT(0, 0)
  "viur.datastore.transport.""__pyx_scope_struct__genexpr", /*tp_name*/
  sizeof(struct __pyx_obj_4viur_9datastore_9transport___pyx_scope_struct__genexpr), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_4viur_9datastore_9transport___pyx_scope_struct__genexpr, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC|Py_TPFLAGS_HAVE_FINALIZE, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_4viur_9datastore_9transport___pyx_scope_struct__genexpr, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_4viur_9datastore_9transport___pyx_scope_struct__genexpr, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif

static PyMethodDef __pyx_methods[] = {
  {0, 0, 0, 0}
};

static int __pyx_import_star_set(PyObject *o, PyObject* py_name, char *name) {
  static const char* internal_type_names[] = {
    "__pyx_ctuple_double__and_double",
    "__pyx_ctuple_double__and_double_struct",
    "__pyx_ctuple_long__and_long",
    "__pyx_ctuple_long__and_long_struct",
    "__pyx_opt_args_4viur_9datastore_9transport_toEntityStructure",
    "__pyx_scope_struct__genexpr",
    "boolean_type",
    "int64_t",
    "simdjsonArray",
    "simdjsonElement",
    "simdjsonElementType",
    "simdjsonErrorCode",
    "simdjsonObject",
    "simdjsonParser",
    "simdjsonResult",
    "stringView",
    "uint64_t",
    0
  };
  const char** type_name = internal_type_names;
  while (*type_name) {
    if (__Pyx_StrEq(name, *type_name)) {
      PyErr_Format(PyExc_TypeError, "Cannot overwrite C type %s", name);
      goto bad;
    }
    type_name++;
  }
  if (0);
  else {
    if (PyObject_SetAttr(__pyx_m, py_name, o) < 0) goto bad;
  }
  return 0;
  bad:
  return -1;
}

static int
__Pyx_import_all_from(PyObject *locals, PyObject *v)
{
    PyObject *all = PyObject_GetAttrString(v, "__all__");
    PyObject *dict, *name, *value;
    int skip_leading_underscores = 0;
    int pos, err;
    if (all == NULL) {
        if (!PyErr_ExceptionMatches(PyExc_AttributeError))
            return -1;
        PyErr_Clear();
        dict = PyObject_GetAttrString(v, "__dict__");
        if (dict == NULL) {
            if (!PyErr_ExceptionMatches(PyExc_AttributeError))
                return -1;
            PyErr_SetString(PyExc_ImportError,
            "from-import-* object has no __dict__ and no __all__");
            return -1;
        }
#if PY_MAJOR_VERSION < 3
        all = PyObject_CallMethod(dict, (char *)"keys", NULL);
#else
        all = PyMapping_Keys(dict);
#endif
        Py_DECREF(dict);
        if (all == NULL)
            return -1;
        skip_leading_underscores = 1;
    }
    for (pos = 0, err = 0; ; pos++) {
        name = PySequence_GetItem(all, pos);
        if (name == NULL) {
            if (!PyErr_ExceptionMatches(PyExc_IndexError))
                err = -1;
            else
                PyErr_Clear();
            break;
        }
        if (skip_leading_underscores &&
#if PY_MAJOR_VERSION < 3
            likely(PyString_Check(name)) &&
            PyString_AS_STRING(name)[0] == '_')
#else
            likely(PyUnicode_Check(name)) &&
            likely(__Pyx_PyUnicode_GET_LENGTH(name)) &&
            __Pyx_PyUnicode_READ_CHAR(name, 0) == '_')
#endif
        {
            Py_DECREF(name);
            continue;
        }
        value = PyObject_GetAttr(v, name);
        if (value == NULL)
            err = -1;
        else if (PyDict_CheckExact(locals))
            err = PyDict_SetItem(locals, name, value);
        else
            err = PyObject_SetItem(locals, name, value);
        Py_DECREF(name);
        Py_XDECREF(value);
        if (err != 0)
            break;
    }
    Py_DECREF(all);
    return err;
}
static int __pyx_import_star(PyObject* m) {
    int i;
    int ret = -1;
    char* s;
    PyObject *locals = 0;
    PyObject *list = 0;
#if PY_MAJOR_VERSION >= 3
    PyObject *utf8_name = 0;
#endif
    PyObject *name;
    PyObject *item;
    locals = PyDict_New();              if (!locals) goto bad;
    if (__Pyx_import_all_from(locals, m) < 0) goto bad;
    list = PyDict_Items(locals);        if (!list) goto bad;
    for(i=0; i<PyList_GET_SIZE(list); i++) {
        name = PyTuple_GET_ITEM(PyList_GET_ITEM(list, i), 0);
        item = PyTuple_GET_ITEM(PyList_GET_ITEM(list, i), 1);
#if PY_MAJOR_VERSION >= 3
        utf8_name = PyUnicode_AsUTF8String(name);
        if (!utf8_name) goto bad;
        s = PyBytes_AS_STRING(utf8_name);
        if (__pyx_import_star_set(item, name, s) < 0) goto bad;
        Py_DECREF(utf8_name); utf8_name = 0;
#else
        s = PyString_AsString(name);
        if (!s) goto bad;
        if (__pyx_import_star_set(item, name, s) < 0) goto bad;
#endif
    }
    ret = 0;
bad:
    Py_XDECREF(locals);
    Py_XDECREF(list);
#if PY_MAJOR_VERSION >= 3
    Py_XDECREF(utf8_name);
#endif
    return ret;
}


#ifndef CYTHON_SMALL_CODE
#if defined(__clang__)
    #define CYTHON_SMALL_CODE
#elif defined(__GNUC__) && (__GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 3))
    #define CYTHON_SMALL_CODE __attribute__((cold))
#else
    #define CYTHON_SMALL_CODE
#endif
#endif
/* #### Code section: pystring_table ### */

static int __Pyx_CreateStringTabAndInitStrings(void) {
  __Pyx_StringTabEntry __pyx_string_tab[] = {
    {&__pyx_kp_u_, __pyx_k_, sizeof(__pyx_k_), 0, 1, 0, 0},
    {&__pyx_n_u_AND, __pyx_k_AND, sizeof(__pyx_k_AND), 0, 1, 0, 1},
    {&__pyx_n_u_ASCENDING, __pyx_k_ASCENDING, sizeof(__pyx_k_ASCENDING), 0, 1, 0, 1},
    {&__pyx_n_u_ASCII, __pyx_k_ASCII, sizeof(__pyx_k_ASCII), 0, 1, 0, 1},
    {&__pyx_n_s_AbortedError, __pyx_k_AbortedError, sizeof(__pyx_k_AbortedError), 0, 0, 1, 1},
    {&__pyx_kp_u_All_retries_are_exhausted_for_th, __pyx_k_All_retries_are_exhausted_for_th, sizeof(__pyx_k_All_retries_are_exhausted_for_th), 0, 1, 0, 0},
    {&__pyx_n_s_AllocateIDs, __pyx_k_AllocateIDs, sizeof(__pyx_k_AllocateIDs), 0, 0, 1, 1},
    {&__pyx_n_s_Any, __pyx_k_Any, sizeof(__pyx_k_Any), 0, 0, 1, 1},
    {&__pyx_n_s_AssertionError, __pyx_k_AssertionError, sizeof(__pyx_k_AssertionError), 0, 0, 1, 1},
    {&__pyx_n_s_AuthorizedSession, __pyx_k_AuthorizedSession, sizeof(__pyx_k_AuthorizedSession), 0, 0, 1, 1},
    {&__pyx_kp_u_Cannot_call_runInTransaction_whi, __pyx_k_Cannot_call_runInTransaction_whi, sizeof(__pyx_k_Cannot_call_runInTransaction_whi), 0, 1, 0, 0},
    {&__pyx_n_s_CollisionError, __pyx_k_CollisionError, sizeof(__pyx_k_CollisionError), 0, 0, 1, 1},
    {&__pyx_n_s_ConnectionError, __pyx_k_ConnectionError, sizeof(__pyx_k_ConnectionError), 0, 0, 1, 1},
    {&__pyx_n_s_Count, __pyx_k_Count, sizeof(__pyx_k_Count), 0, 0, 1, 1},
    {&__pyx_n_u_DESCENDING, __pyx_k_DESCENDING, sizeof(__pyx_k_DESCENDING), 0, 1, 0, 1},
    {&__pyx_n_s_Delete, __pyx_k_Delete, sizeof(__pyx_k_Delete), 0, 0, 1, 1},
    {&__pyx_n_u_EQUAL, __pyx_k_EQUAL, sizeof(__pyx_k_EQUAL), 0, 1, 0, 1},
    {&__pyx_kp_u_Empty_response_received_from_Dat, __pyx_k_Empty_response_received_from_Dat, sizeof(__pyx_k_Empty_response_received_from_Dat), 0, 1, 0, 0},
    {&__pyx_n_s_Entity, __pyx_k_Entity, sizeof(__pyx_k_Entity), 0, 0, 1, 1},
    {&__pyx_n_u_GREATER_THAN, __pyx_k_GREATER_THAN, sizeof(__pyx_k_GREATER_THAN), 0, 1, 0, 1},
    {&__pyx_n_u_GREATER_THAN_OR_EQUAL, __pyx_k_GREATER_THAN_OR_EQUAL, sizeof(__pyx_k_GREATER_THAN_OR_EQUAL), 0, 1, 0, 1},
    {&__pyx_n_s_Get, __pyx_k_Get, sizeof(__pyx_k_Get), 0, 0, 1, 1},
    {&__pyx_n_s_Get_locals_genexpr, __pyx_k_Get_locals_genexpr, sizeof(__pyx_k_Get_locals_genexpr), 0, 0, 1, 1},
    {&__pyx_kp_u_INVALID_RESPONSE_RECEIVED, __pyx_k_INVALID_RESPONSE_RECEIVED, sizeof(__pyx_k_INVALID_RESPONSE_RECEIVED), 0, 1, 0, 0},
    {&__pyx_n_s_InvalidArgumentError, __pyx_k_InvalidArgumentError, sizeof(__pyx_k_InvalidArgumentError), 0, 0, 1, 1},
    {&__pyx_kp_u_Invalid_data_received_from_Datas, __pyx_k_Invalid_data_received_from_Datas, sizeof(__pyx_k_Invalid_data_received_from_Datas), 0, 1, 0, 0},
    {&__pyx_kp_u_Invalid_key_in_entity_json_s, __pyx_k_Invalid_key_in_entity_json_s, sizeof(__pyx_k_Invalid_key_in_entity_json_s), 0, 1, 0, 0},
    {&__pyx_kp_u_Invalid_number_of_mutation_resul, __pyx_k_Invalid_number_of_mutation_resul, sizeof(__pyx_k_Invalid_number_of_mutation_resul), 0, 1, 0, 0},
    {&__pyx_kp_u_Invalid_op_s, __pyx_k_Invalid_op_s, sizeof(__pyx_k_Invalid_op_s), 0, 1, 0, 0},
    {&__pyx_n_s_Key, __pyx_k_Key, sizeof(__pyx_k_Key), 0, 0, 1, 1},
    {&__pyx_kp_u_Keys_must_not_be_partial_or_kind, __pyx_k_Keys_must_not_be_partial_or_kind, sizeof(__pyx_k_Keys_must_not_be_partial_or_kind), 0, 1, 0, 0},
    {&__pyx_n_u_LESS_THAN, __pyx_k_LESS_THAN, sizeof(__pyx_k_LESS_THAN), 0, 1, 0, 1},
    {&__pyx_n_u_LESS_THAN_OR_EQUAL, __pyx_k_LESS_THAN_OR_EQUAL, sizeof(__pyx_k_LESS_THAN_OR_EQUAL), 0, 1, 0, 1},
    {&__pyx_n_s_List, __pyx_k_List, sizeof(__pyx_k_List), 0, 0, 1, 1},
    {&__pyx_kp_s_List_Entity, __pyx_k_List_Entity, sizeof(__pyx_k_List_Entity), 0, 0, 1, 0},
    {&__pyx_kp_s_List_dict, __pyx_k_List_dict, sizeof(__pyx_k_List_dict), 0, 0, 1, 0},
    {&__pyx_n_u_NON_TRANSACTIONAL, __pyx_k_NON_TRANSACTIONAL, sizeof(__pyx_k_NON_TRANSACTIONAL), 0, 1, 0, 1},
    {&__pyx_n_u_NOT_FINISHED, __pyx_k_NOT_FINISHED, sizeof(__pyx_k_NOT_FINISHED), 0, 1, 0, 1},
    {&__pyx_n_u_NO_MORE_RESULTS, __pyx_k_NO_MORE_RESULTS, sizeof(__pyx_k_NO_MORE_RESULTS), 0, 1, 0, 1},
    {&__pyx_n_s_NoMutationResultsError, __pyx_k_NoMutationResultsError, sizeof(__pyx_k_NoMutationResultsError), 0, 0, 1, 1},
    {&__pyx_kp_u_No_mutation_results_received, __pyx_k_No_mutation_results_received, sizeof(__pyx_k_No_mutation_results_received), 0, 1, 0, 0},
    {&__pyx_kp_u_No_mutation_results_received_2, __pyx_k_No_mutation_results_received_2, sizeof(__pyx_k_No_mutation_results_received_2), 0, 1, 0, 0},
    {&__pyx_n_s_None, __pyx_k_None, sizeof(__pyx_k_None), 0, 0, 1, 1},
    {&__pyx_kp_u_None, __pyx_k_None, sizeof(__pyx_k_None), 0, 1, 0, 0},
    {&__pyx_kp_u_Please_create_a_matching_index, __pyx_k_Please_create_a_matching_index, sizeof(__pyx_k_Please_create_a_matching_index), 0, 1, 0, 0},
    {&__pyx_n_s_Put, __pyx_k_Put, sizeof(__pyx_k_Put), 0, 0, 1, 1},
    {&__pyx_kp_u_Queried, __pyx_k_Queried, sizeof(__pyx_k_Queried), 0, 1, 0, 0},
    {&__pyx_n_s_QueryDefinition, __pyx_k_QueryDefinition, sizeof(__pyx_k_QueryDefinition), 0, 0, 1, 1},
    {&__pyx_kp_u_Query_not_finished_Maybe_some_en, __pyx_k_Query_not_finished_Maybe_some_en, sizeof(__pyx_k_Query_not_finished_Maybe_some_en), 0, 1, 0, 0},
    {&__pyx_kp_u_Received_an_unexpected_key_updat, __pyx_k_Received_an_unexpected_key_updat, sizeof(__pyx_k_Received_an_unexpected_key_updat), 0, 1, 0, 0},
    {&__pyx_n_s_RecursionError, __pyx_k_RecursionError, sizeof(__pyx_k_RecursionError), 0, 0, 1, 1},
    {&__pyx_n_s_RequestsConnectionError, __pyx_k_RequestsConnectionError, sizeof(__pyx_k_RequestsConnectionError), 0, 0, 1, 1},
    {&__pyx_kp_u_Retrying_http_post_request_to_da, __pyx_k_Retrying_http_post_request_to_da, sizeof(__pyx_k_Retrying_http_post_request_to_da), 0, 1, 0, 0},
    {&__pyx_kp_u_Returned, __pyx_k_Returned, sizeof(__pyx_k_Returned), 0, 1, 0, 0},
    {&__pyx_n_s_RunInTransaction, __pyx_k_RunInTransaction, sizeof(__pyx_k_RunInTransaction), 0, 0, 1, 1},
    {&__pyx_n_u_STRONG, __pyx_k_STRONG, sizeof(__pyx_k_STRONG), 0, 1, 0, 1},
    {&__pyx_n_u_TRANSACTIONAL, __pyx_k_TRANSACTIONAL, sizeof(__pyx_k_TRANSACTIONAL), 0, 1, 0, 1},
    {&__pyx_kp_u_The_Count_aggregation_query_is_a, __pyx_k_The_Count_aggregation_query_is_a, sizeof(__pyx_k_The_Count_aggregation_query_is_a), 0, 1, 0, 0},
    {&__pyx_kp_u_UTF_8, __pyx_k_UTF_8, sizeof(__pyx_k_UTF_8), 0, 1, 0, 0},
    {&__pyx_n_s_Union, __pyx_k_Union, sizeof(__pyx_k_Union), 0, 0, 1, 1},
    {&__pyx_kp_s_Union_Entity_List_Entity, __pyx_k_Union_Entity_List_Entity, sizeof(__pyx_k_Union_Entity_List_Entity), 0, 0, 1, 0},
    {&__pyx_kp_s_Union_Key_List_Key, __pyx_k_Union_Key_List_Key, sizeof(__pyx_k_Union_Key_List_Key), 0, 0, 1, 0},
    {&__pyx_kp_s_Union_Key_List_Key_Entity_List_E, __pyx_k_Union_Key_List_Key_Entity_List_E, sizeof(__pyx_k_Union_Key_List_Key_Entity_List_E), 0, 0, 1, 0},
    {&__pyx_kp_s_Union_None_Entity_List_Entity, __pyx_k_Union_None_Entity_List_Entity, sizeof(__pyx_k_Union_None_Entity_List_Entity), 0, 0, 1, 0},
    {&__pyx_kp_u_Unknown_type, __pyx_k_Unknown_type, sizeof(__pyx_k_Unknown_type), 0, 1, 0, 0},
    {&__pyx_n_s_ValueError, __pyx_k_ValueError, sizeof(__pyx_k_ValueError), 0, 0, 1, 1},
    {&__pyx_n_s_ViurDatastoreError, __pyx_k_ViurDatastoreError, sizeof(__pyx_k_ViurDatastoreError), 0, 0, 1, 1},
    {&__pyx_kp_u_We_got_an_error_in_a_transaction, __pyx_k_We_got_an_error_in_a_transaction, sizeof(__pyx_k_We_got_an_error_in_a_transaction), 0, 1, 0, 0},
    {&__pyx_kp_u_We_read_an_incomplete_null_key_f, __pyx_k_We_read_an_incomplete_null_key_f, sizeof(__pyx_k_We_read_an_incomplete_null_key_f), 0, 1, 0, 0},
    {&__pyx_kp_u_Y_m_dT_H_M_S, __pyx_k_Y_m_dT_H_M_S, sizeof(__pyx_k_Y_m_dT_H_M_S), 0, 1, 0, 0},
    {&__pyx_kp_u_Y_m_dT_H_M_S_f, __pyx_k_Y_m_dT_H_M_S_f, sizeof(__pyx_k_Y_m_dT_H_M_S_f), 0, 1, 0, 0},
    {&__pyx_kp_u_Y_m_dT_H_M_S_fZ, __pyx_k_Y_m_dT_H_M_S_fZ, sizeof(__pyx_k_Y_m_dT_H_M_S_fZ), 0, 1, 0, 0},
    {&__pyx_kp_u__10, __pyx_k__10, sizeof(__pyx_k__10), 0, 1, 0, 0},
    {&__pyx_kp_u__11, __pyx_k__11, sizeof(__pyx_k__11), 0, 1, 0, 0},
    {&__pyx_n_s__21, __pyx_k__21, sizeof(__pyx_k__21), 0, 0, 1, 1},
    {&__pyx_kp_u__3, __pyx_k__3, sizeof(__pyx_k__3), 0, 1, 0, 0},
    {&__pyx_n_s__45, __pyx_k__45, sizeof(__pyx_k__45), 0, 0, 1, 1},
    {&__pyx_kp_u__5, __pyx_k__5, sizeof(__pyx_k__5), 0, 1, 0, 0},
    {&__pyx_kp_u__6, __pyx_k__6, sizeof(__pyx_k__6), 0, 1, 0, 0},
    {&__pyx_kp_u__7, __pyx_k__7, sizeof(__pyx_k__7), 0, 1, 0, 0},
    {&__pyx_kp_u__8, __pyx_k__8, sizeof(__pyx_k__8), 0, 1, 0, 0},
    {&__pyx_kp_u__9, __pyx_k__9, sizeof(__pyx_k__9), 0, 1, 0, 0},
    {&__pyx_n_s_accessLog, __pyx_k_accessLog, sizeof(__pyx_k_accessLog), 0, 0, 1, 1},
    {&__pyx_n_u_affectedEntities, __pyx_k_affectedEntities, sizeof(__pyx_k_affectedEntities), 0, 1, 0, 1},
    {&__pyx_n_s_affectedEntity, __pyx_k_affectedEntity, sizeof(__pyx_k_affectedEntity), 0, 0, 1, 1},
    {&__pyx_n_u_aggregateProperties, __pyx_k_aggregateProperties, sizeof(__pyx_k_aggregateProperties), 0, 1, 0, 1},
    {&__pyx_n_u_aggregationResults, __pyx_k_aggregationResults, sizeof(__pyx_k_aggregationResults), 0, 1, 0, 1},
    {&__pyx_n_u_aggregation_query, __pyx_k_aggregation_query, sizeof(__pyx_k_aggregation_query), 0, 1, 0, 1},
    {&__pyx_n_u_aggregations, __pyx_k_aggregations, sizeof(__pyx_k_aggregations), 0, 1, 0, 1},
    {&__pyx_n_u_allowOverriding, __pyx_k_allowOverriding, sizeof(__pyx_k_allowOverriding), 0, 1, 0, 1},
    {&__pyx_n_s_allowOverriding_2, __pyx_k_allowOverriding_2, sizeof(__pyx_k_allowOverriding_2), 0, 0, 1, 1},
    {&__pyx_kp_u_and_orders, __pyx_k_and_orders, sizeof(__pyx_k_and_orders), 0, 1, 0, 0},
    {&__pyx_n_s_append, __pyx_k_append, sizeof(__pyx_k_append), 0, 0, 1, 1},
    {&__pyx_n_s_args, __pyx_k_args, sizeof(__pyx_k_args), 0, 0, 1, 1},
    {&__pyx_n_s_arrayElem, __pyx_k_arrayElem, sizeof(__pyx_k_arrayElem), 0, 0, 1, 1},
    {&__pyx_n_s_arrayIt, __pyx_k_arrayIt, sizeof(__pyx_k_arrayIt), 0, 0, 1, 1},
    {&__pyx_n_u_arrayValue, __pyx_k_arrayValue, sizeof(__pyx_k_arrayValue), 0, 1, 0, 1},
    {&__pyx_n_s_array_element, __pyx_k_array_element, sizeof(__pyx_k_array_element), 0, 0, 1, 1},
    {&__pyx_n_s_array_element_inner, __pyx_k_array_element_inner, sizeof(__pyx_k_array_element_inner), 0, 0, 1, 1},
    {&__pyx_n_s_astimezone, __pyx_k_astimezone, sizeof(__pyx_k_astimezone), 0, 0, 1, 1},
    {&__pyx_n_s_asyncio_coroutines, __pyx_k_asyncio_coroutines, sizeof(__pyx_k_asyncio_coroutines), 0, 0, 1, 1},
    {&__pyx_n_s_auth, __pyx_k_auth, sizeof(__pyx_k_auth), 0, 0, 1, 1},
    {&__pyx_n_s_authenticated_request, __pyx_k_authenticated_request, sizeof(__pyx_k_authenticated_request), 0, 0, 1, 1},
    {&__pyx_n_s_b64decode, __pyx_k_b64decode, sizeof(__pyx_k_b64decode), 0, 0, 1, 1},
    {&__pyx_n_s_b64encode, __pyx_k_b64encode, sizeof(__pyx_k_b64encode), 0, 0, 1, 1},
    {&__pyx_n_s_base64, __pyx_k_base64, sizeof(__pyx_k_base64), 0, 0, 1, 1},
    {&__pyx_n_u_blobValue, __pyx_k_blobValue, sizeof(__pyx_k_blobValue), 0, 1, 0, 1},
    {&__pyx_n_u_booleanValue, __pyx_k_booleanValue, sizeof(__pyx_k_booleanValue), 0, 1, 0, 1},
    {&__pyx_n_s_bytes, __pyx_k_bytes, sizeof(__pyx_k_bytes), 0, 0, 1, 1},
    {&__pyx_n_s_cache, __pyx_k_cache, sizeof(__pyx_k_cache), 0, 0, 1, 1},
    {&__pyx_n_s_callable, __pyx_k_callable, sizeof(__pyx_k_callable), 0, 0, 1, 1},
    {&__pyx_n_s_callback, __pyx_k_callback, sizeof(__pyx_k_callback), 0, 0, 1, 1},
    {&__pyx_n_s_class_getitem, __pyx_k_class_getitem, sizeof(__pyx_k_class_getitem), 0, 0, 1, 1},
    {&__pyx_n_s_cline_in_traceback, __pyx_k_cline_in_traceback, sizeof(__pyx_k_cline_in_traceback), 0, 0, 1, 1},
    {&__pyx_n_s_close, __pyx_k_close, sizeof(__pyx_k_close), 0, 0, 1, 1},
    {&__pyx_n_u_compositeFilter, __pyx_k_compositeFilter, sizeof(__pyx_k_compositeFilter), 0, 1, 0, 1},
    {&__pyx_n_s_conf, __pyx_k_conf, sizeof(__pyx_k_conf), 0, 0, 1, 1},
    {&__pyx_n_s_content, __pyx_k_content, sizeof(__pyx_k_content), 0, 0, 1, 1},
    {&__pyx_n_s_copy, __pyx_k_copy, sizeof(__pyx_k_copy), 0, 0, 1, 1},
    {&__pyx_n_u_count, __pyx_k_count, sizeof(__pyx_k_count), 0, 1, 0, 1},
    {&__pyx_n_s_credentials, __pyx_k_credentials, sizeof(__pyx_k_credentials), 0, 0, 1, 1},
    {&__pyx_n_s_currentCursor, __pyx_k_currentCursor, sizeof(__pyx_k_currentCursor), 0, 0, 1, 1},
    {&__pyx_n_s_currentDbAccessLog, __pyx_k_currentDbAccessLog, sizeof(__pyx_k_currentDbAccessLog), 0, 0, 1, 1},
    {&__pyx_n_s_currentTransaction, __pyx_k_currentTransaction, sizeof(__pyx_k_currentTransaction), 0, 0, 1, 1},
    {&__pyx_n_s_currentTxn, __pyx_k_currentTxn, sizeof(__pyx_k_currentTxn), 0, 0, 1, 1},
    {&__pyx_n_s_data, __pyx_k_data, sizeof(__pyx_k_data), 0, 0, 1, 1},
    {&__pyx_n_s_data_ptr, __pyx_k_data_ptr, sizeof(__pyx_k_data_ptr), 0, 0, 1, 1},
    {&__pyx_n_s_datetime, __pyx_k_datetime, sizeof(__pyx_k_datetime), 0, 0, 1, 1},
    {&__pyx_n_s_debug, __pyx_k_debug, sizeof(__pyx_k_debug), 0, 0, 1, 1},
    {&__pyx_n_s_decode, __pyx_k_decode, sizeof(__pyx_k_decode), 0, 0, 1, 1},
    {&__pyx_n_s_default, __pyx_k_default, sizeof(__pyx_k_default), 0, 0, 1, 1},
    {&__pyx_n_s_deferred_key, __pyx_k_deferred_key, sizeof(__pyx_k_deferred_key), 0, 0, 1, 1},
    {&__pyx_n_s_deferred_keys, __pyx_k_deferred_keys, sizeof(__pyx_k_deferred_keys), 0, 0, 1, 1},
    {&__pyx_n_s_delete, __pyx_k_delete, sizeof(__pyx_k_delete), 0, 0, 1, 1},
    {&__pyx_n_u_delete, __pyx_k_delete, sizeof(__pyx_k_delete), 0, 1, 0, 1},
    {&__pyx_n_s_dict, __pyx_k_dict, sizeof(__pyx_k_dict), 0, 0, 1, 1},
    {&__pyx_n_s_dictKey, __pyx_k_dictKey, sizeof(__pyx_k_dictKey), 0, 0, 1, 1},
    {&__pyx_n_s_dictValue, __pyx_k_dictValue, sizeof(__pyx_k_dictValue), 0, 0, 1, 1},
    {&__pyx_n_u_direction, __pyx_k_direction, sizeof(__pyx_k_direction), 0, 1, 0, 1},
    {&__pyx_kp_u_disable, __pyx_k_disable, sizeof(__pyx_k_disable), 0, 1, 0, 0},
    {&__pyx_n_s_distinct, __pyx_k_distinct, sizeof(__pyx_k_distinct), 0, 0, 1, 1},
    {&__pyx_n_s_distinctKey, __pyx_k_distinctKey, sizeof(__pyx_k_distinctKey), 0, 0, 1, 1},
    {&__pyx_n_s_distinctOn, __pyx_k_distinctOn, sizeof(__pyx_k_distinctOn), 0, 0, 1, 1},
    {&__pyx_n_u_distinctOn, __pyx_k_distinctOn, sizeof(__pyx_k_distinctOn), 0, 1, 0, 1},
    {&__pyx_kp_u_distinct_on_s, __pyx_k_distinct_on_s, sizeof(__pyx_k_distinct_on_s), 0, 1, 0, 0},
    {&__pyx_n_u_doubleValue, __pyx_k_doubleValue, sizeof(__pyx_k_doubleValue), 0, 1, 0, 1},
    {&__pyx_n_s_dumps, __pyx_k_dumps, sizeof(__pyx_k_dumps), 0, 0, 1, 1},
    {&__pyx_n_s_element, __pyx_k_element, sizeof(__pyx_k_element), 0, 0, 1, 1},
    {&__pyx_n_s_element_inner, __pyx_k_element_inner, sizeof(__pyx_k_element_inner), 0, 0, 1, 1},
    {&__pyx_kp_u_enable, __pyx_k_enable, sizeof(__pyx_k_enable), 0, 1, 0, 0},
    {&__pyx_n_s_encode, __pyx_k_encode, sizeof(__pyx_k_encode), 0, 0, 1, 1},
    {&__pyx_n_s_endCursor, __pyx_k_endCursor, sizeof(__pyx_k_endCursor), 0, 0, 1, 1},
    {&__pyx_n_u_endCursor, __pyx_k_endCursor, sizeof(__pyx_k_endCursor), 0, 1, 0, 1},
    {&__pyx_n_s_entities, __pyx_k_entities, sizeof(__pyx_k_entities), 0, 0, 1, 1},
    {&__pyx_n_u_entityValue, __pyx_k_entityValue, sizeof(__pyx_k_entityValue), 0, 1, 0, 1},
    {&__pyx_n_s_enumerate, __pyx_k_enumerate, sizeof(__pyx_k_enumerate), 0, 0, 1, 1},
    {&__pyx_n_s_err, __pyx_k_err, sizeof(__pyx_k_err), 0, 0, 1, 1},
    {&__pyx_n_s_error, __pyx_k_error, sizeof(__pyx_k_error), 0, 0, 1, 1},
    {&__pyx_n_u_excludeFromIndexes, __pyx_k_excludeFromIndexes, sizeof(__pyx_k_excludeFromIndexes), 0, 1, 0, 1},
    {&__pyx_n_s_exclude_from_indexes, __pyx_k_exclude_from_indexes, sizeof(__pyx_k_exclude_from_indexes), 0, 0, 1, 1},
    {&__pyx_n_s_excludedProperties, __pyx_k_excludedProperties, sizeof(__pyx_k_excludedProperties), 0, 0, 1, 1},
    {&__pyx_n_s_exponential_backoff, __pyx_k_exponential_backoff, sizeof(__pyx_k_exponential_backoff), 0, 0, 1, 1},
    {&__pyx_n_s_extend, __pyx_k_extend, sizeof(__pyx_k_extend), 0, 0, 1, 1},
    {&__pyx_n_u_filter, __pyx_k_filter, sizeof(__pyx_k_filter), 0, 1, 0, 1},
    {&__pyx_n_s_filterList, __pyx_k_filterList, sizeof(__pyx_k_filterList), 0, 0, 1, 1},
    {&__pyx_n_s_filters, __pyx_k_filters, sizeof(__pyx_k_filters), 0, 0, 1, 1},
    {&__pyx_n_u_filters, __pyx_k_filters, sizeof(__pyx_k_filters), 0, 1, 0, 1},
    {&__pyx_n_s_flipResults, __pyx_k_flipResults, sizeof(__pyx_k_flipResults), 0, 0, 1, 1},
    {&__pyx_n_s_from_legacy_urlsafe, __pyx_k_from_legacy_urlsafe, sizeof(__pyx_k_from_legacy_urlsafe), 0, 0, 1, 1},
    {&__pyx_kp_u_gc, __pyx_k_gc, sizeof(__pyx_k_gc), 0, 1, 0, 0},
    {&__pyx_n_s_genexpr, __pyx_k_genexpr, sizeof(__pyx_k_genexpr), 0, 0, 1, 1},
    {&__pyx_n_s_get, __pyx_k_get, sizeof(__pyx_k_get), 0, 0, 1, 1},
    {&__pyx_n_s_google, __pyx_k_google, sizeof(__pyx_k_google), 0, 0, 1, 1},
    {&__pyx_n_s_google_auth, __pyx_k_google_auth, sizeof(__pyx_k_google_auth), 0, 0, 1, 1},
    {&__pyx_n_s_http_internal, __pyx_k_http_internal, sizeof(__pyx_k_http_internal), 0, 0, 1, 1},
    {&__pyx_kp_u_https_datastore_googleapis_com_v, __pyx_k_https_datastore_googleapis_com_v, sizeof(__pyx_k_https_datastore_googleapis_com_v), 0, 1, 0, 0},
    {&__pyx_kp_u_https_datastore_googleapis_com_v_2, __pyx_k_https_datastore_googleapis_com_v_2, sizeof(__pyx_k_https_datastore_googleapis_com_v_2), 0, 1, 0, 0},
    {&__pyx_kp_u_https_datastore_googleapis_com_v_3, __pyx_k_https_datastore_googleapis_com_v_3, sizeof(__pyx_k_https_datastore_googleapis_com_v_3), 0, 1, 0, 0},
    {&__pyx_kp_u_https_datastore_googleapis_com_v_4, __pyx_k_https_datastore_googleapis_com_v_4, sizeof(__pyx_k_https_datastore_googleapis_com_v_4), 0, 1, 0, 0},
    {&__pyx_kp_u_https_datastore_googleapis_com_v_5, __pyx_k_https_datastore_googleapis_com_v_5, sizeof(__pyx_k_https_datastore_googleapis_com_v_5), 0, 1, 0, 0},
    {&__pyx_kp_u_https_datastore_googleapis_com_v_6, __pyx_k_https_datastore_googleapis_com_v_6, sizeof(__pyx_k_https_datastore_googleapis_com_v_6), 0, 1, 0, 0},
    {&__pyx_kp_u_https_datastore_googleapis_com_v_7, __pyx_k_https_datastore_googleapis_com_v_7, sizeof(__pyx_k_https_datastore_googleapis_com_v_7), 0, 1, 0, 0},
    {&__pyx_kp_u_https_www_googleapis_com_auth_da, __pyx_k_https_www_googleapis_com_auth_da, sizeof(__pyx_k_https_www_googleapis_com_auth_da), 0, 1, 0, 0},
    {&__pyx_n_s_i, __pyx_k_i, sizeof(__pyx_k_i), 0, 0, 1, 1},
    {&__pyx_n_s_id, __pyx_k_id, sizeof(__pyx_k_id), 0, 0, 1, 1},
    {&__pyx_n_u_id, __pyx_k_id, sizeof(__pyx_k_id), 0, 1, 0, 1},
    {&__pyx_n_s_idx, __pyx_k_idx, sizeof(__pyx_k_idx), 0, 0, 1, 1},
    {&__pyx_n_s_import, __pyx_k_import, sizeof(__pyx_k_import), 0, 0, 1, 1},
    {&__pyx_n_s_initializing, __pyx_k_initializing, sizeof(__pyx_k_initializing), 0, 0, 1, 1},
    {&__pyx_n_s_innerArrayElem, __pyx_k_innerArrayElem, sizeof(__pyx_k_innerArrayElem), 0, 0, 1, 1},
    {&__pyx_n_s_insert, __pyx_k_insert, sizeof(__pyx_k_insert), 0, 0, 1, 1},
    {&__pyx_n_s_int, __pyx_k_int, sizeof(__pyx_k_int), 0, 0, 1, 1},
    {&__pyx_n_u_integerValue, __pyx_k_integerValue, sizeof(__pyx_k_integerValue), 0, 1, 0, 1},
    {&__pyx_n_s_internalStartCursor, __pyx_k_internalStartCursor, sizeof(__pyx_k_internalStartCursor), 0, 0, 1, 1},
    {&__pyx_n_s_isMulti, __pyx_k_isMulti, sizeof(__pyx_k_isMulti), 0, 0, 1, 1},
    {&__pyx_n_s_is_coroutine, __pyx_k_is_coroutine, sizeof(__pyx_k_is_coroutine), 0, 0, 1, 1},
    {&__pyx_kp_u_is_not_supported, __pyx_k_is_not_supported, sizeof(__pyx_k_is_not_supported), 0, 1, 0, 0},
    {&__pyx_n_s_is_partial, __pyx_k_is_partial, sizeof(__pyx_k_is_partial), 0, 0, 1, 1},
    {&__pyx_n_s_is_viur_datastore_request_ok, __pyx_k_is_viur_datastore_request_ok, sizeof(__pyx_k_is_viur_datastore_request_ok), 0, 0, 1, 1},
    {&__pyx_kp_u_isenabled, __pyx_k_isenabled, sizeof(__pyx_k_isenabled), 0, 1, 0, 0},
    {&__pyx_n_s_items, __pyx_k_items, sizeof(__pyx_k_items), 0, 0, 1, 1},
    {&__pyx_n_s_json, __pyx_k_json, sizeof(__pyx_k_json), 0, 0, 1, 1},
    {&__pyx_n_s_k, __pyx_k_k, sizeof(__pyx_k_k), 0, 0, 1, 1},
    {&__pyx_n_s_key, __pyx_k_key, sizeof(__pyx_k_key), 0, 0, 1, 1},
    {&__pyx_n_u_key, __pyx_k_key, sizeof(__pyx_k_key), 0, 1, 0, 1},
    {&__pyx_n_s_keyList, __pyx_k_keyList, sizeof(__pyx_k_keyList), 0, 0, 1, 1},
    {&__pyx_n_s_keyToPath, __pyx_k_keyToPath, sizeof(__pyx_k_keyToPath), 0, 0, 1, 1},
    {&__pyx_n_u_keyValue, __pyx_k_keyValue, sizeof(__pyx_k_keyValue), 0, 1, 0, 1},
    {&__pyx_n_s_keys, __pyx_k_keys, sizeof(__pyx_k_keys), 0, 0, 1, 1},
    {&__pyx_n_u_keys, __pyx_k_keys, sizeof(__pyx_k_keys), 0, 1, 0, 1},
    {&__pyx_n_s_keys_for_request, __pyx_k_keys_for_request, sizeof(__pyx_k_keys_for_request), 0, 0, 1, 1},
    {&__pyx_n_s_kind, __pyx_k_kind, sizeof(__pyx_k_kind), 0, 0, 1, 1},
    {&__pyx_n_u_kind, __pyx_k_kind, sizeof(__pyx_k_kind), 0, 1, 0, 1},
    {&__pyx_n_s_kwargs, __pyx_k_kwargs, sizeof(__pyx_k_kwargs), 0, 0, 1, 1},
    {&__pyx_n_s_limit, __pyx_k_limit, sizeof(__pyx_k_limit), 0, 0, 1, 1},
    {&__pyx_n_u_limit, __pyx_k_limit, sizeof(__pyx_k_limit), 0, 1, 0, 1},
    {&__pyx_n_s_loads, __pyx_k_loads, sizeof(__pyx_k_loads), 0, 0, 1, 1},
    {&__pyx_n_s_logging, __pyx_k_logging, sizeof(__pyx_k_logging), 0, 0, 1, 1},
    {&__pyx_n_s_main, __pyx_k_main, sizeof(__pyx_k_main), 0, 0, 1, 1},
    {&__pyx_n_u_memcache_client, __pyx_k_memcache_client, sizeof(__pyx_k_memcache_client), 0, 1, 0, 1},
    {&__pyx_n_s_missing_keys, __pyx_k_missing_keys, sizeof(__pyx_k_missing_keys), 0, 0, 1, 1},
    {&__pyx_n_s_missing_keys_paths, __pyx_k_missing_keys_paths, sizeof(__pyx_k_missing_keys_paths), 0, 0, 1, 1},
    {&__pyx_n_u_mode, __pyx_k_mode, sizeof(__pyx_k_mode), 0, 1, 0, 1},
    {&__pyx_n_u_mutations, __pyx_k_mutations, sizeof(__pyx_k_mutations), 0, 1, 0, 1},
    {&__pyx_n_s_name, __pyx_k_name, sizeof(__pyx_k_name), 0, 0, 1, 1},
    {&__pyx_n_u_name, __pyx_k_name, sizeof(__pyx_k_name), 0, 1, 0, 1},
    {&__pyx_n_s_name_2, __pyx_k_name_2, sizeof(__pyx_k_name_2), 0, 0, 1, 1},
    {&__pyx_n_u_nested_query, __pyx_k_nested_query, sizeof(__pyx_k_nested_query), 0, 1, 0, 1},
    {&__pyx_n_u_nullValue, __pyx_k_nullValue, sizeof(__pyx_k_nullValue), 0, 1, 0, 1},
    {&__pyx_n_s_oldTxn, __pyx_k_oldTxn, sizeof(__pyx_k_oldTxn), 0, 0, 1, 1},
    {&__pyx_n_s_op, __pyx_k_op, sizeof(__pyx_k_op), 0, 0, 1, 1},
    {&__pyx_n_u_op, __pyx_k_op, sizeof(__pyx_k_op), 0, 1, 0, 1},
    {&__pyx_n_u_order, __pyx_k_order, sizeof(__pyx_k_order), 0, 1, 0, 1},
    {&__pyx_n_s_orders, __pyx_k_orders, sizeof(__pyx_k_orders), 0, 0, 1, 1},
    {&__pyx_n_s_parent, __pyx_k_parent, sizeof(__pyx_k_parent), 0, 0, 1, 1},
    {&__pyx_n_s_parser, __pyx_k_parser, sizeof(__pyx_k_parser), 0, 0, 1, 1},
    {&__pyx_n_u_partitionId, __pyx_k_partitionId, sizeof(__pyx_k_partitionId), 0, 1, 0, 1},
    {&__pyx_n_u_path, __pyx_k_path, sizeof(__pyx_k_path), 0, 1, 0, 1},
    {&__pyx_n_s_pop, __pyx_k_pop, sizeof(__pyx_k_pop), 0, 0, 1, 1},
    {&__pyx_n_s_post, __pyx_k_post, sizeof(__pyx_k_post), 0, 0, 1, 1},
    {&__pyx_n_s_postData, __pyx_k_postData, sizeof(__pyx_k_postData), 0, 0, 1, 1},
    {&__pyx_n_s_post_data, __pyx_k_post_data, sizeof(__pyx_k_post_data), 0, 0, 1, 1},
    {&__pyx_n_u_previousTransaction, __pyx_k_previousTransaction, sizeof(__pyx_k_previousTransaction), 0, 1, 0, 1},
    {&__pyx_n_s_projectID, __pyx_k_projectID, sizeof(__pyx_k_projectID), 0, 0, 1, 1},
    {&__pyx_n_u_project_id, __pyx_k_project_id, sizeof(__pyx_k_project_id), 0, 1, 0, 1},
    {&__pyx_n_s_properties, __pyx_k_properties, sizeof(__pyx_k_properties), 0, 0, 1, 1},
    {&__pyx_n_u_properties, __pyx_k_properties, sizeof(__pyx_k_properties), 0, 1, 0, 1},
    {&__pyx_n_u_property, __pyx_k_property, sizeof(__pyx_k_property), 0, 1, 0, 1},
    {&__pyx_n_u_propertyFilter, __pyx_k_propertyFilter, sizeof(__pyx_k_propertyFilter), 0, 1, 0, 1},
    {&__pyx_n_u_property_1, __pyx_k_property_1, sizeof(__pyx_k_property_1), 0, 1, 0, 1},
    {&__pyx_n_s_put, __pyx_k_put, sizeof(__pyx_k_put), 0, 0, 1, 1},
    {&__pyx_n_s_pysize, __pyx_k_pysize, sizeof(__pyx_k_pysize), 0, 0, 1, 1},
    {&__pyx_n_s_pythonPropToJson, __pyx_k_pythonPropToJson, sizeof(__pyx_k_pythonPropToJson), 0, 0, 1, 1},
    {&__pyx_n_u_query, __pyx_k_query, sizeof(__pyx_k_query), 0, 1, 0, 1},
    {&__pyx_n_s_queryDefinition, __pyx_k_queryDefinition, sizeof(__pyx_k_queryDefinition), 0, 0, 1, 1},
    {&__pyx_n_s_range, __pyx_k_range, sizeof(__pyx_k_range), 0, 0, 1, 1},
    {&__pyx_n_u_readConsistency, __pyx_k_readConsistency, sizeof(__pyx_k_readConsistency), 0, 1, 0, 1},
    {&__pyx_n_s_readOptions, __pyx_k_readOptions, sizeof(__pyx_k_readOptions), 0, 0, 1, 1},
    {&__pyx_n_u_readOptions, __pyx_k_readOptions, sizeof(__pyx_k_readOptions), 0, 1, 0, 1},
    {&__pyx_n_u_readWrite, __pyx_k_readWrite, sizeof(__pyx_k_readWrite), 0, 1, 0, 1},
    {&__pyx_n_s_refresh_timeout, __pyx_k_refresh_timeout, sizeof(__pyx_k_refresh_timeout), 0, 0, 1, 1},
    {&__pyx_n_s_replace, __pyx_k_replace, sizeof(__pyx_k_replace), 0, 0, 1, 1},
    {&__pyx_n_s_requestedKeys, __pyx_k_requestedKeys, sizeof(__pyx_k_requestedKeys), 0, 0, 1, 1},
    {&__pyx_n_s_requested_keys, __pyx_k_requested_keys, sizeof(__pyx_k_requested_keys), 0, 0, 1, 1},
    {&__pyx_n_s_requests, __pyx_k_requests, sizeof(__pyx_k_requests), 0, 0, 1, 1},
    {&__pyx_kp_s_requests_Response, __pyx_k_requests_Response, sizeof(__pyx_k_requests_Response), 0, 0, 1, 0},
    {&__pyx_n_s_requests_exceptions, __pyx_k_requests_exceptions, sizeof(__pyx_k_requests_exceptions), 0, 0, 1, 1},
    {&__pyx_n_s_res, __pyx_k_res, sizeof(__pyx_k_res), 0, 0, 1, 1},
    {&__pyx_n_s_resDict, __pyx_k_resDict, sizeof(__pyx_k_resDict), 0, 0, 1, 1},
    {&__pyx_n_s_resList, __pyx_k_resList, sizeof(__pyx_k_resList), 0, 0, 1, 1},
    {&__pyx_n_s_res_from_cache, __pyx_k_res_from_cache, sizeof(__pyx_k_res_from_cache), 0, 0, 1, 1},
    {&__pyx_n_s_res_from_db, __pyx_k_res_from_db, sizeof(__pyx_k_res_from_db), 0, 0, 1, 1},
    {&__pyx_n_s_resp, __pyx_k_resp, sizeof(__pyx_k_resp), 0, 0, 1, 1},
    {&__pyx_kp_u_resp_content, __pyx_k_resp_content, sizeof(__pyx_k_resp_content), 0, 1, 0, 0},
    {&__pyx_kp_u_results, __pyx_k_results, sizeof(__pyx_k_results), 0, 1, 0, 0},
    {&__pyx_n_s_return, __pyx_k_return, sizeof(__pyx_k_return), 0, 0, 1, 1},
    {&__pyx_n_s_rollbackTxn, __pyx_k_rollbackTxn, sizeof(__pyx_k_rollbackTxn), 0, 0, 1, 1},
    {&__pyx_n_s_runSingleFilter, __pyx_k_runSingleFilter, sizeof(__pyx_k_runSingleFilter), 0, 0, 1, 1},
    {&__pyx_n_s_scopes, __pyx_k_scopes, sizeof(__pyx_k_scopes), 0, 0, 1, 1},
    {&__pyx_kp_u_seconds, __pyx_k_seconds, sizeof(__pyx_k_seconds), 0, 1, 0, 0},
    {&__pyx_n_s_send, __pyx_k_send, sizeof(__pyx_k_send), 0, 0, 1, 1},
    {&__pyx_n_s_set, __pyx_k_set, sizeof(__pyx_k_set), 0, 0, 1, 1},
    {&__pyx_n_s_singleValue, __pyx_k_singleValue, sizeof(__pyx_k_singleValue), 0, 0, 1, 1},
    {&__pyx_n_s_sleep, __pyx_k_sleep, sizeof(__pyx_k_sleep), 0, 0, 1, 1},
    {&__pyx_n_s_sleep_time, __pyx_k_sleep_time, sizeof(__pyx_k_sleep_time), 0, 0, 1, 1},
    {&__pyx_n_s_sortOrder, __pyx_k_sortOrder, sizeof(__pyx_k_sortOrder), 0, 0, 1, 1},
    {&__pyx_n_s_spec, __pyx_k_spec, sizeof(__pyx_k_spec), 0, 0, 1, 1},
    {&__pyx_n_s_split, __pyx_k_split, sizeof(__pyx_k_split), 0, 0, 1, 1},
    {&__pyx_kp_s_src_viur_datastore_transport_pyx, __pyx_k_src_viur_datastore_transport_pyx, sizeof(__pyx_k_src_viur_datastore_transport_pyx), 0, 0, 1, 0},
    {&__pyx_n_s_startCursor, __pyx_k_startCursor, sizeof(__pyx_k_startCursor), 0, 0, 1, 1},
    {&__pyx_n_u_startCursor, __pyx_k_startCursor, sizeof(__pyx_k_startCursor), 0, 1, 0, 1},
    {&__pyx_n_s_str, __pyx_k_str, sizeof(__pyx_k_str), 0, 0, 1, 1},
    {&__pyx_n_s_strftime, __pyx_k_strftime, sizeof(__pyx_k_strftime), 0, 0, 1, 1},
    {&__pyx_n_u_stringValue, __pyx_k_stringValue, sizeof(__pyx_k_stringValue), 0, 1, 0, 1},
    {&__pyx_n_s_strptime, __pyx_k_strptime, sizeof(__pyx_k_strptime), 0, 0, 1, 1},
    {&__pyx_n_s_test, __pyx_k_test, sizeof(__pyx_k_test), 0, 0, 1, 1},
    {&__pyx_n_s_throw, __pyx_k_throw, sizeof(__pyx_k_throw), 0, 0, 1, 1},
    {&__pyx_n_s_time, __pyx_k_time, sizeof(__pyx_k_time), 0, 0, 1, 1},
    {&__pyx_n_u_timestampValue, __pyx_k_timestampValue, sizeof(__pyx_k_timestampValue), 0, 1, 0, 1},
    {&__pyx_n_s_timezone, __pyx_k_timezone, sizeof(__pyx_k_timezone), 0, 0, 1, 1},
    {&__pyx_n_u_traceQueries, __pyx_k_traceQueries, sizeof(__pyx_k_traceQueries), 0, 1, 0, 1},
    {&__pyx_n_u_transaction, __pyx_k_transaction, sizeof(__pyx_k_transaction), 0, 1, 0, 1},
    {&__pyx_n_u_transactionOptions, __pyx_k_transactionOptions, sizeof(__pyx_k_transactionOptions), 0, 1, 0, 1},
    {&__pyx_n_s_transport, __pyx_k_transport, sizeof(__pyx_k_transport), 0, 0, 1, 1},
    {&__pyx_n_s_txnKey, __pyx_k_txnKey, sizeof(__pyx_k_txnKey), 0, 0, 1, 1},
    {&__pyx_n_s_txnOptions, __pyx_k_txnOptions, sizeof(__pyx_k_txnOptions), 0, 0, 1, 1},
    {&__pyx_n_s_typing, __pyx_k_typing, sizeof(__pyx_k_typing), 0, 0, 1, 1},
    {&__pyx_n_s_tzinfo, __pyx_k_tzinfo, sizeof(__pyx_k_tzinfo), 0, 0, 1, 1},
    {&__pyx_n_s_untouched_keys, __pyx_k_untouched_keys, sizeof(__pyx_k_untouched_keys), 0, 0, 1, 1},
    {&__pyx_n_s_up_to, __pyx_k_up_to, sizeof(__pyx_k_up_to), 0, 0, 1, 1},
    {&__pyx_n_u_up_to, __pyx_k_up_to, sizeof(__pyx_k_up_to), 0, 1, 0, 1},
    {&__pyx_n_s_update, __pyx_k_update, sizeof(__pyx_k_update), 0, 0, 1, 1},
    {&__pyx_n_u_upsert, __pyx_k_upsert, sizeof(__pyx_k_upsert), 0, 1, 0, 1},
    {&__pyx_n_s_url, __pyx_k_url, sizeof(__pyx_k_url), 0, 0, 1, 1},
    {&__pyx_n_s_utc, __pyx_k_utc, sizeof(__pyx_k_utc), 0, 0, 1, 1},
    {&__pyx_n_s_v, __pyx_k_v, sizeof(__pyx_k_v), 0, 0, 1, 1},
    {&__pyx_n_s_value, __pyx_k_value, sizeof(__pyx_k_value), 0, 0, 1, 1},
    {&__pyx_n_u_value, __pyx_k_value, sizeof(__pyx_k_value), 0, 1, 0, 1},
    {&__pyx_n_u_values, __pyx_k_values, sizeof(__pyx_k_values), 0, 1, 0, 1},
    {&__pyx_n_s_version, __pyx_k_version, sizeof(__pyx_k_version), 0, 0, 1, 1},
    {&__pyx_n_s_viur_datastore, __pyx_k_viur_datastore, sizeof(__pyx_k_viur_datastore), 0, 0, 1, 1},
    {&__pyx_n_s_viur_datastore_config, __pyx_k_viur_datastore_config, sizeof(__pyx_k_viur_datastore_config), 0, 0, 1, 1},
    {&__pyx_n_s_viur_datastore_errors, __pyx_k_viur_datastore_errors, sizeof(__pyx_k_viur_datastore_errors), 0, 0, 1, 1},
    {&__pyx_n_s_viur_datastore_transport, __pyx_k_viur_datastore_transport, sizeof(__pyx_k_viur_datastore_transport), 0, 0, 1, 1},
    {&__pyx_n_s_viur_datastore_types, __pyx_k_viur_datastore_types, sizeof(__pyx_k_viur_datastore_types), 0, 0, 1, 1},
    {&__pyx_n_s_warning, __pyx_k_warning, sizeof(__pyx_k_warning), 0, 0, 1, 1},
    {&__pyx_kp_u_with_filter, __pyx_k_with_filter, sizeof(__pyx_k_with_filter), 0, 1, 0, 0},
    {&__pyx_n_s_x, __pyx_k_x, sizeof(__pyx_k_x), 0, 0, 1, 1},
    {0, 0, 0, 0, 0, 0, 0}
  };
  return __Pyx_InitStrings(__pyx_string_tab);
}
/* #### Code section: cached_builtins ### */
static CYTHON_SMALL_CODE int __Pyx_InitCachedBuiltins(void) {
  __pyx_builtin_range = __Pyx_GetBuiltinName(__pyx_n_s_range); if (!__pyx_builtin_range) __PYX_ERR(0, 114, __pyx_L1_error)
  __pyx_builtin_AssertionError = __Pyx_GetBuiltinName(__pyx_n_s_AssertionError); if (!__pyx_builtin_AssertionError) __PYX_ERR(0, 237, __pyx_L1_error)
  __pyx_builtin_ValueError = __Pyx_GetBuiltinName(__pyx_n_s_ValueError); if (!__pyx_builtin_ValueError) __PYX_ERR(0, 375, __pyx_L1_error)
  __pyx_builtin_enumerate = __Pyx_GetBuiltinName(__pyx_n_s_enumerate); if (!__pyx_builtin_enumerate) __PYX_ERR(0, 640, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}
/* #### Code section: cached_constants ### */

static CYTHON_SMALL_CODE int __Pyx_InitCachedConstants(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_InitCachedConstants", 0);

  /* "viur/datastore/transport.pyx":360
 *                 return objIterStart.value().get_bool()
 *             elif (strView.compare("timestampValue") == 0):
 *                 dateStr = toPyStr(objIterStart.value().get_string())[:-1]  # Strip "Z" at the end             # <<<<<<<<<<<<<<
 *                 if "." in dateStr:  # With milli-seconds
 *                     return datetime.strptime(dateStr, "%Y-%m-%dT%H:%M:%S.%f").replace(tzinfo=timezone.utc)
 */
  __pyx_slice__2 = PySlice_New(Py_None, __pyx_int_neg_1, Py_None); if (unlikely(!__pyx_slice__2)) __PYX_ERR(0, 360, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_slice__2);
  __Pyx_GIVEREF(__pyx_slice__2);

  /* "viur/datastore/transport.pyx":435
 *         return v.get_bool()
 *     else:
 *         raise ValueError("Unknown type")             # <<<<<<<<<<<<<<
 * 
 * def runSingleFilter(queryDefinition: QueryDefinition, limit: int) -> List[Entity]:
 */
  __pyx_tuple__4 = PyTuple_Pack(1, __pyx_kp_u_Unknown_type); if (unlikely(!__pyx_tuple__4)) __PYX_ERR(0, 435, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__4);
  __Pyx_GIVEREF(__pyx_tuple__4);

  /* "viur/datastore/transport.pyx":567
 *             queryDefinition.kind, filters, orders, distinctOn, len(res)))
 *     if flipResults:
 *         return res[::-1]             # <<<<<<<<<<<<<<
 *     return res
 * 
 */
  __pyx_slice__12 = PySlice_New(Py_None, Py_None, __pyx_int_neg_1); if (unlikely(!__pyx_slice__12)) __PYX_ERR(0, 567, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_slice__12);
  __Pyx_GIVEREF(__pyx_slice__12);

  /* "viur/datastore/transport.pyx":600
 *     untouched_keys = copy(keys)
 *     while keys:
 *         keys_for_request = keys[:300]             # <<<<<<<<<<<<<<
 * 
 *         if conf["memcache_client"] is not None:
 */
  __pyx_slice__13 = PySlice_New(Py_None, __pyx_int_300, Py_None); if (unlikely(!__pyx_slice__13)) __PYX_ERR(0, 600, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_slice__13);
  __Pyx_GIVEREF(__pyx_slice__13);

  /* "viur/datastore/transport.pyx":611
 *         if not missing_keys:
 *             # We had all keys in the memcache we can fetch the next batch now.
 *             keys = keys[300:]             # <<<<<<<<<<<<<<
 *             continue
 *         requested_keys = [
 */
  __pyx_slice__14 = PySlice_New(__pyx_int_300, Py_None, Py_None); if (unlikely(!__pyx_slice__14)) __PYX_ERR(0, 611, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_slice__14);
  __Pyx_GIVEREF(__pyx_slice__14);

  /* "viur/datastore/transport.pyx":707
 *         if arrayElem.size() != abs(len(keys)):
 *             logging.error(resp.content)
 *             raise ValueError("Invalid number of mutation-results received")             # <<<<<<<<<<<<<<
 *     if conf["memcache_client"] is not None:
 *         cache.delete([str(key) for key in keys])
 */
  __pyx_tuple__15 = PyTuple_Pack(1, __pyx_kp_u_Invalid_number_of_mutation_resul); if (unlikely(!__pyx_tuple__15)) __PYX_ERR(0, 707, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__15);
  __Pyx_GIVEREF(__pyx_tuple__15);

  /* "viur/datastore/transport.pyx":755
 *         if (element.at_pointer("/mutationResults").error() != SUCCESS):
 *             logging.error(resp.content)
 *             raise ValueError("No mutation-results received")             # <<<<<<<<<<<<<<
 *         arrayElem = element.at_key("mutationResults").get_array()
 *         if arrayElem.size() != abs(len(entities)):
 */
  __pyx_tuple__16 = PyTuple_Pack(1, __pyx_kp_u_No_mutation_results_received_2); if (unlikely(!__pyx_tuple__16)) __PYX_ERR(0, 755, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__16);
  __Pyx_GIVEREF(__pyx_tuple__16);

  /* "viur/datastore/transport.pyx":795
 *             if oldTxn:
 *                 if not allowOverriding:
 *                     raise RecursionError("Cannot call runInTransaction while inside a transaction!")             # <<<<<<<<<<<<<<
 *                 txnOptions = {
 *                     "previousTransaction": oldTxn["key"]
 */
  __pyx_tuple__17 = PyTuple_Pack(1, __pyx_kp_u_Cannot_call_runInTransaction_whi); if (unlikely(!__pyx_tuple__17)) __PYX_ERR(0, 795, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__17);
  __Pyx_GIVEREF(__pyx_tuple__17);

  /* "viur/datastore/transport.pyx":929
 *                 preincrement(arrayIt)
 *             if not res:
 *                 raise ValueError("Empty response received from Datastore API")             # <<<<<<<<<<<<<<
 *             elif not isMulti:
 *                 return res[0]
 */
  __pyx_tuple__18 = PyTuple_Pack(1, __pyx_kp_u_Empty_response_received_from_Dat); if (unlikely(!__pyx_tuple__18)) __PYX_ERR(0, 929, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__18);
  __Pyx_GIVEREF(__pyx_tuple__18);

  /* "viur/datastore/transport.pyx":937
 *             logging.error("Invalid data received from Datastore API")
 *             logging.error(resp.content)
 *             raise ValueError("Invalid data received from Datastore API")             # <<<<<<<<<<<<<<
 * 
 * def Count(kind: str = None, up_to= 2 ** 63 - 1, queryDefinition: QueryDefinition = None) -> Union[Key, List[Key]]:
 */
  __pyx_tuple__19 = PyTuple_Pack(1, __pyx_kp_u_Invalid_data_received_from_Datas); if (unlikely(!__pyx_tuple__19)) __PYX_ERR(0, 937, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__19);
  __Pyx_GIVEREF(__pyx_tuple__19);

  /* "viur/datastore/transport.pyx":6
 * from copy import copy
 * from viur.datastore import cache
 * import google.auth             # <<<<<<<<<<<<<<
 * import requests
 * from libcpp cimport bool as boolean_type
 */
  __pyx_tuple__20 = PyTuple_Pack(2, __pyx_n_s_google, __pyx_n_s_auth); if (unlikely(!__pyx_tuple__20)) __PYX_ERR(0, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__20);
  __Pyx_GIVEREF(__pyx_tuple__20);

  /* "viur/datastore/transport.pyx":105
 * )
 * 
 * def authenticated_request(url: str, data: bytes) -> requests.Response:             # <<<<<<<<<<<<<<
 *     """
 *         Runs one http request to the datastore rest api, authenticated with the current projects service account.
 */
  __pyx_tuple__22 = PyTuple_Pack(3, __pyx_n_s_url, __pyx_n_s_data, __pyx_n_s_i); if (unlikely(!__pyx_tuple__22)) __PYX_ERR(0, 105, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__22);
  __Pyx_GIVEREF(__pyx_tuple__22);
  __pyx_codeobj__23 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__22, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_viur_datastore_transport_pyx, __pyx_n_s_authenticated_request, 105, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__23)) __PYX_ERR(0, 105, __pyx_L1_error)

  /* "viur/datastore/transport.pyx":126
 *             continue
 * 
 * def keyToPath(key: Key) -> List[dict]:             # <<<<<<<<<<<<<<
 *     """
 *         Converts a Key object to the PathElements expected by the rest API.
 */
  __pyx_tuple__24 = PyTuple_Pack(2, __pyx_n_s_key, __pyx_n_s_res); if (unlikely(!__pyx_tuple__24)) __PYX_ERR(0, 126, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__24);
  __Pyx_GIVEREF(__pyx_tuple__24);
  __pyx_codeobj__25 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__24, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_viur_datastore_transport_pyx, __pyx_n_s_keyToPath, 126, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__25)) __PYX_ERR(0, 126, __pyx_L1_error)

  /* "viur/datastore/transport.pyx":150
 *     return res
 * 
 * def pythonPropToJson(v) -> dict:             # <<<<<<<<<<<<<<
 *     """
 *         Convert a python type to the object representation expected by the datastore rest API.
 */
  __pyx_tuple__26 = PyTuple_Pack(11, __pyx_n_s_v, __pyx_n_s_excludedProperties, __pyx_n_s_properties, __pyx_n_s_dictKey, __pyx_n_s_dictValue, __pyx_n_s_resList, __pyx_n_s_value, __pyx_n_s_resDict, __pyx_n_s_dictKey, __pyx_n_s_dictValue, __pyx_n_s_x); if (unlikely(!__pyx_tuple__26)) __PYX_ERR(0, 150, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__26);
  __Pyx_GIVEREF(__pyx_tuple__26);
  __pyx_codeobj__27 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 11, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__26, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_viur_datastore_transport_pyx, __pyx_n_s_pythonPropToJson, 150, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__27)) __PYX_ERR(0, 150, __pyx_L1_error)

  /* "viur/datastore/transport.pyx":437
 *         raise ValueError("Unknown type")
 * 
 * def runSingleFilter(queryDefinition: QueryDefinition, limit: int) -> List[Entity]:             # <<<<<<<<<<<<<<
 *     """
 *         Runs a single Query as defined by queryDefinition. The limit of the queryDefinition is ignored and must
 */
  __pyx_tuple__28 = PyTuple_Pack(24, __pyx_n_s_queryDefinition, __pyx_n_s_limit, __pyx_n_s_parser, __pyx_n_s_pysize, __pyx_n_s_data_ptr, __pyx_n_s_element, __pyx_n_s_res, __pyx_n_s_internalStartCursor, __pyx_n_s_flipResults, __pyx_n_s_currentTxn, __pyx_n_s_readOptions, __pyx_n_s_postData, __pyx_n_s_filterList, __pyx_n_s_k, __pyx_n_s_v, __pyx_n_s_key, __pyx_n_s_op, __pyx_n_s_singleValue, __pyx_n_s_resp, __pyx_n_s_orders, __pyx_n_s_filters, __pyx_n_s_distinctOn, __pyx_n_s_sortOrder, __pyx_n_s_distinctKey); if (unlikely(!__pyx_tuple__28)) __PYX_ERR(0, 437, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__28);
  __Pyx_GIVEREF(__pyx_tuple__28);
  __pyx_codeobj__29 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 24, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__28, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_viur_datastore_transport_pyx, __pyx_n_s_runSingleFilter, 437, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__29)) __PYX_ERR(0, 437, __pyx_L1_error)

  /* "viur/datastore/transport.pyx":570
 *     return res
 * 
 * def Get(keys: Union[Key, List[Key]]) -> Union[None, Entity, List[Entity]]:             # <<<<<<<<<<<<<<
 *     """
 *         Fetches the entities determined by keys from the datastore. Returns or inserts None if a key is not found.
 */
  __pyx_tuple__30 = PyTuple_Pack(32, __pyx_n_s_keys, __pyx_n_s_parser, __pyx_n_s_pysize, __pyx_n_s_data_ptr, __pyx_n_s_element, __pyx_n_s_isMulti, __pyx_n_s_accessLog, __pyx_n_s_currentTxn, __pyx_n_s_readOptions, __pyx_n_s_res, __pyx_n_s_res_from_cache, __pyx_n_s_res_from_db, __pyx_n_s_untouched_keys, __pyx_n_s_keys_for_request, __pyx_n_s_missing_keys, __pyx_n_s_requested_keys, __pyx_n_s_postData, __pyx_n_s_resp, __pyx_n_s_deferred_keys, __pyx_n_s_missing_keys_paths, __pyx_n_s_i, __pyx_n_s_deferred_key, __pyx_n_s_genexpr, __pyx_n_s_genexpr, __pyx_n_s_key, __pyx_n_s_value, __pyx_n_s_key, __pyx_n_s_x, __pyx_n_s_key, __pyx_n_s_key, __pyx_n_s_value, __pyx_n_s_key); if (unlikely(!__pyx_tuple__30)) __PYX_ERR(0, 570, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__30);
  __Pyx_GIVEREF(__pyx_tuple__30);
  __pyx_codeobj__31 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 32, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__30, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_viur_datastore_transport_pyx, __pyx_n_s_Get, 570, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__31)) __PYX_ERR(0, 570, __pyx_L1_error)

  /* "viur/datastore/transport.pyx":656
 *         return [res.get(key) for key in untouched_keys]  # Sort by order of incoming keys
 * 
 * def Delete(keys: Union[Key, List[Key], Entity, List[Entity]]) -> None:             # <<<<<<<<<<<<<<
 *     """
 *         Deletes the entities stored under the given key(s).
 */
  __pyx_tuple__32 = PyTuple_Pack(13, __pyx_n_s_keys, __pyx_n_s_parser, __pyx_n_s_pysize, __pyx_n_s_data_ptr, __pyx_n_s_element, __pyx_n_s_arrayElem, __pyx_n_s_accessLog, __pyx_n_s_postData, __pyx_n_s_currentTxn, __pyx_n_s_resp, __pyx_n_s_x, __pyx_n_s_x, __pyx_n_s_key); if (unlikely(!__pyx_tuple__32)) __PYX_ERR(0, 656, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__32);
  __Pyx_GIVEREF(__pyx_tuple__32);
  __pyx_codeobj__33 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 13, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__32, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_viur_datastore_transport_pyx, __pyx_n_s_Delete, 656, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__33)) __PYX_ERR(0, 656, __pyx_L1_error)

  /* "viur/datastore/transport.pyx":711
 *         cache.delete([str(key) for key in keys])
 * 
 * def Put(entities: Union[Entity, List[Entity]]) -> Union[Entity, List[Entity]]:             # <<<<<<<<<<<<<<
 *     """
 *         Writes the given entities into the datastore. The entities can be from different kinds. If an entity has an
 */
  __pyx_tuple__34 = PyTuple_Pack(15, __pyx_n_s_entities, __pyx_n_s_parser, __pyx_n_s_pysize, __pyx_n_s_data_ptr, __pyx_n_s_element, __pyx_n_s_innerArrayElem, __pyx_n_s_arrayElem, __pyx_n_s_arrayIt, __pyx_n_s_accessLog, __pyx_n_s_postData, __pyx_n_s_currentTxn, __pyx_n_s_resp, __pyx_n_s_idx, __pyx_n_s_x, __pyx_n_s_x); if (unlikely(!__pyx_tuple__34)) __PYX_ERR(0, 711, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__34);
  __Pyx_GIVEREF(__pyx_tuple__34);
  __pyx_codeobj__35 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 15, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__34, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_viur_datastore_transport_pyx, __pyx_n_s_Put, 711, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__35)) __PYX_ERR(0, 711, __pyx_L1_error)

  /* "viur/datastore/transport.pyx":774
 *     return entities
 * 
 * def RunInTransaction(callback: callable, *args, **kwargs) -> Any:             # <<<<<<<<<<<<<<
 *     """
 *         Runs the given function inside a AID transaction.
 */
  __pyx_tuple__36 = PyTuple_Pack(23, __pyx_n_s_callback, __pyx_n_s_args, __pyx_n_s_kwargs, __pyx_n_s_parser, __pyx_n_s_pysize, __pyx_n_s_data_ptr, __pyx_n_s_element, __pyx_n_s_innerArrayElem, __pyx_n_s_arrayElem, __pyx_n_s_arrayIt, __pyx_n_s_exponential_backoff, __pyx_n_s_oldTxn, __pyx_n_s_allowOverriding_2, __pyx_n_s_txnOptions, __pyx_n_s_postData, __pyx_n_s_resp, __pyx_n_s_txnKey, __pyx_n_s_currentTxn, __pyx_n_s_res, __pyx_n_s_idx, __pyx_n_s_affectedEntity, __pyx_n_s_err, __pyx_n_s_sleep_time); if (unlikely(!__pyx_tuple__36)) __PYX_ERR(0, 774, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__36);
  __Pyx_GIVEREF(__pyx_tuple__36);
  __pyx_codeobj__37 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 23, 0, CO_OPTIMIZED|CO_NEWLOCALS|CO_VARARGS|CO_VARKEYWORDS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__36, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_viur_datastore_transport_pyx, __pyx_n_s_RunInTransaction, 774, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__37)) __PYX_ERR(0, 774, __pyx_L1_error)

  /* "viur/datastore/transport.pyx":868
 *         "All retries are exhausted for this transaction")  # If we made it here, all tries are exhausted
 * 
 * def _rollbackTxn(txnKey: str):             # <<<<<<<<<<<<<<
 *     """
 *         Internal helper that aborts the given transaction. It's important to abort pending transactions (instead
 */
  __pyx_tuple__38 = PyTuple_Pack(3, __pyx_n_s_txnKey, __pyx_n_s_postData, __pyx_n_s_resp); if (unlikely(!__pyx_tuple__38)) __PYX_ERR(0, 868, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__38);
  __Pyx_GIVEREF(__pyx_tuple__38);
  __pyx_codeobj__39 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__38, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_viur_datastore_transport_pyx, __pyx_n_s_rollbackTxn, 868, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__39)) __PYX_ERR(0, 868, __pyx_L1_error)

  /* "viur/datastore/transport.pyx":884
 *     )
 * 
 * def AllocateIDs(keys: Union[Key, List[Key]]) -> Union[Key, List[Key]]:             # <<<<<<<<<<<<<<
 *     """
 *         Allocates numeric IDs for the keys given.
 */
  __pyx_tuple__40 = PyTuple_Pack(15, __pyx_n_s_keys, __pyx_n_s_parser, __pyx_n_s_pysize, __pyx_n_s_data_ptr, __pyx_n_s_element, __pyx_n_s_isMulti, __pyx_n_s_keyList, __pyx_n_s_requestedKeys, __pyx_n_s_postData, __pyx_n_s_resp, __pyx_n_s_res, __pyx_n_s_arrayElem, __pyx_n_s_arrayIt, __pyx_n_s_innerArrayElem, __pyx_n_s_x); if (unlikely(!__pyx_tuple__40)) __PYX_ERR(0, 884, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__40);
  __Pyx_GIVEREF(__pyx_tuple__40);
  __pyx_codeobj__41 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 15, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__40, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_viur_datastore_transport_pyx, __pyx_n_s_AllocateIDs, 884, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__41)) __PYX_ERR(0, 884, __pyx_L1_error)

  /* "viur/datastore/transport.pyx":939
 *             raise ValueError("Invalid data received from Datastore API")
 * 
 * def Count(kind: str = None, up_to= 2 ** 63 - 1, queryDefinition: QueryDefinition = None) -> Union[Key, List[Key]]:             # <<<<<<<<<<<<<<
 *     """
 *         Count all entries in a kind if there is only a kind is provided
 */
  __pyx_tuple__42 = PyTuple_Pack(18, __pyx_n_s_kind, __pyx_n_s_up_to, __pyx_n_s_queryDefinition, __pyx_n_s_parser, __pyx_n_s_pysize, __pyx_n_s_data_ptr, __pyx_n_s_element, __pyx_n_s_element_inner, __pyx_n_s_array_element, __pyx_n_s_array_element_inner, __pyx_n_s_post_data, __pyx_n_s_filterList, __pyx_n_s_k, __pyx_n_s_v, __pyx_n_s_key, __pyx_n_s_op, __pyx_n_s_singleValue, __pyx_n_s_resp); if (unlikely(!__pyx_tuple__42)) __PYX_ERR(0, 939, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__42);
  __Pyx_GIVEREF(__pyx_tuple__42);
  __pyx_codeobj__43 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 18, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__42, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_viur_datastore_transport_pyx, __pyx_n_s_Count, 939, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__43)) __PYX_ERR(0, 939, __pyx_L1_error)
  __pyx_tuple__44 = PyTuple_Pack(3, Py_None, ((PyObject *)__pyx_int_0x7fffffffffffffff), Py_None); if (unlikely(!__pyx_tuple__44)) __PYX_ERR(0, 939, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__44);
  __Pyx_GIVEREF(__pyx_tuple__44);
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}
/* #### Code section: init_constants ### */

static CYTHON_SMALL_CODE int __Pyx_InitConstants(void) {
  __pyx_umethod_PyDict_Type_get.type = (PyObject*)&PyDict_Type;
  __pyx_umethod_PyDict_Type_get.method_name = &__pyx_n_s_get;
  __pyx_umethod_PyDict_Type_pop.type = (PyObject*)&PyDict_Type;
  __pyx_umethod_PyDict_Type_pop.method_name = &__pyx_n_s_pop;
  __pyx_umethod_PyDict_Type_update.type = (PyObject*)&PyDict_Type;
  __pyx_umethod_PyDict_Type_update.method_name = &__pyx_n_s_update;
  if (__Pyx_CreateStringTabAndInitStrings() < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  __pyx_int_0 = PyInt_FromLong(0); if (unlikely(!__pyx_int_0)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_1 = PyInt_FromLong(1); if (unlikely(!__pyx_int_1)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_2 = PyInt_FromLong(2); if (unlikely(!__pyx_int_2)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_4 = PyInt_FromLong(4); if (unlikely(!__pyx_int_4)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_300 = PyInt_FromLong(300); if (unlikely(!__pyx_int_300)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_0x7fffffffffffffff = PyInt_FromString((char *)"0x7fffffffffffffff", 0, 0); if (unlikely(!__pyx_int_0x7fffffffffffffff)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_neg_1 = PyInt_FromLong(-1); if (unlikely(!__pyx_int_neg_1)) __PYX_ERR(0, 1, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}
/* #### Code section: init_globals ### */

static CYTHON_SMALL_CODE int __Pyx_InitGlobals(void) {
  /* AssertionsEnabled.init */
  if (likely(__Pyx_init_assertions_enabled() == 0)); else

if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1, __pyx_L1_error)

  return 0;
  __pyx_L1_error:;
  return -1;
}
/* #### Code section: init_module ### */

static CYTHON_SMALL_CODE int __Pyx_modinit_global_init_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_export_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_export_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_init_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_import_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_import_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_import_code(void); /*proto*/

static int __Pyx_modinit_global_init_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_global_init_code", 0);
  /*--- Global init code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_variable_export_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_export_code", 0);
  /*--- Variable export code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_function_export_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_export_code", 0);
  /*--- Function export code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_type_init_code(void) {
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_init_code", 0);
  /*--- Type init code ---*/
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_4viur_9datastore_9transport___pyx_scope_struct__genexpr = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_4viur_9datastore_9transport___pyx_scope_struct__genexpr_spec, NULL); if (unlikely(!__pyx_ptype_4viur_9datastore_9transport___pyx_scope_struct__genexpr)) __PYX_ERR(0, 584, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_4viur_9datastore_9transport___pyx_scope_struct__genexpr_spec, __pyx_ptype_4viur_9datastore_9transport___pyx_scope_struct__genexpr) < 0) __PYX_ERR(0, 584, __pyx_L1_error)
  #else
  __pyx_ptype_4viur_9datastore_9transport___pyx_scope_struct__genexpr = &__pyx_type_4viur_9datastore_9transport___pyx_scope_struct__genexpr;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_4viur_9datastore_9transport___pyx_scope_struct__genexpr) < 0) __PYX_ERR(0, 584, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_4viur_9datastore_9transport___pyx_scope_struct__genexpr->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_4viur_9datastore_9transport___pyx_scope_struct__genexpr->tp_dictoffset && __pyx_ptype_4viur_9datastore_9transport___pyx_scope_struct__genexpr->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_4viur_9datastore_9transport___pyx_scope_struct__genexpr->tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  #endif
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_type_import_code(void) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_import_code", 0);
  /*--- Type import code ---*/
  __pyx_t_1 = PyImport_ImportModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 9, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7cpython_4type_type = __Pyx_ImportType_3_0_10(__pyx_t_1, __Pyx_BUILTIN_MODULE_NAME, "type", 
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(PyTypeObject), __PYX_GET_STRUCT_ALIGNMENT_3_0_10(PyTypeObject),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(PyTypeObject), __PYX_GET_STRUCT_ALIGNMENT_3_0_10(PyTypeObject),
  #else
  sizeof(PyHeapTypeObject), __PYX_GET_STRUCT_ALIGNMENT_3_0_10(PyHeapTypeObject),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_0_10); if (!__pyx_ptype_7cpython_4type_type) __PYX_ERR(1, 9, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_variable_import_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_import_code", 0);
  /*--- Variable import code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_function_import_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_import_code", 0);
  /*--- Function import code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}


#if PY_MAJOR_VERSION >= 3
#if CYTHON_PEP489_MULTI_PHASE_INIT
static PyObject* __pyx_pymod_create(PyObject *spec, PyModuleDef *def); /*proto*/
static int __pyx_pymod_exec_transport(PyObject* module); /*proto*/
static PyModuleDef_Slot __pyx_moduledef_slots[] = {
  {Py_mod_create, (void*)__pyx_pymod_create},
  {Py_mod_exec, (void*)__pyx_pymod_exec_transport},
  {0, NULL}
};
#endif

#ifdef __cplusplus
namespace {
  struct PyModuleDef __pyx_moduledef =
  #else
  static struct PyModuleDef __pyx_moduledef =
  #endif
  {
      PyModuleDef_HEAD_INIT,
      "transport",
      0, /* m_doc */
    #if CYTHON_PEP489_MULTI_PHASE_INIT
      0, /* m_size */
    #elif CYTHON_USE_MODULE_STATE
      sizeof(__pyx_mstate), /* m_size */
    #else
      -1, /* m_size */
    #endif
      __pyx_methods /* m_methods */,
    #if CYTHON_PEP489_MULTI_PHASE_INIT
      __pyx_moduledef_slots, /* m_slots */
    #else
      NULL, /* m_reload */
    #endif
    #if CYTHON_USE_MODULE_STATE
      __pyx_m_traverse, /* m_traverse */
      __pyx_m_clear, /* m_clear */
      NULL /* m_free */
    #else
      NULL, /* m_traverse */
      NULL, /* m_clear */
      NULL /* m_free */
    #endif
  };
  #ifdef __cplusplus
} /* anonymous namespace */
#endif
#endif

#ifndef CYTHON_NO_PYINIT_EXPORT
#define __Pyx_PyMODINIT_FUNC PyMODINIT_FUNC
#elif PY_MAJOR_VERSION < 3
#ifdef __cplusplus
#define __Pyx_PyMODINIT_FUNC extern "C" void
#else
#define __Pyx_PyMODINIT_FUNC void
#endif
#else
#ifdef __cplusplus
#define __Pyx_PyMODINIT_FUNC extern "C" PyObject *
#else
#define __Pyx_PyMODINIT_FUNC PyObject *
#endif
#endif


#if PY_MAJOR_VERSION < 3
__Pyx_PyMODINIT_FUNC inittransport(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC inittransport(void)
#else
__Pyx_PyMODINIT_FUNC PyInit_transport(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC PyInit_transport(void)
#if CYTHON_PEP489_MULTI_PHASE_INIT
{
  return PyModuleDef_Init(&__pyx_moduledef);
}
static CYTHON_SMALL_CODE int __Pyx_check_single_interpreter(void) {
    #if PY_VERSION_HEX >= 0x030700A1
    static PY_INT64_T main_interpreter_id = -1;
    PY_INT64_T current_id = PyInterpreterState_GetID(PyThreadState_Get()->interp);
    if (main_interpreter_id == -1) {
        main_interpreter_id = current_id;
        return (unlikely(current_id == -1)) ? -1 : 0;
    } else if (unlikely(main_interpreter_id != current_id))
    #else
    static PyInterpreterState *main_interpreter = NULL;
    PyInterpreterState *current_interpreter = PyThreadState_Get()->interp;
    if (!main_interpreter) {
        main_interpreter = current_interpreter;
    } else if (unlikely(main_interpreter != current_interpreter))
    #endif
    {
        PyErr_SetString(
            PyExc_ImportError,
            "Interpreter change detected - this module can only be loaded into one interpreter per process.");
        return -1;
    }
    return 0;
}
#if CYTHON_COMPILING_IN_LIMITED_API
static CYTHON_SMALL_CODE int __Pyx_copy_spec_to_module(PyObject *spec, PyObject *module, const char* from_name, const char* to_name, int allow_none)
#else
static CYTHON_SMALL_CODE int __Pyx_copy_spec_to_module(PyObject *spec, PyObject *moddict, const char* from_name, const char* to_name, int allow_none)
#endif
{
    PyObject *value = PyObject_GetAttrString(spec, from_name);
    int result = 0;
    if (likely(value)) {
        if (allow_none || value != Py_None) {
#if CYTHON_COMPILING_IN_LIMITED_API
            result = PyModule_AddObject(module, to_name, value);
#else
            result = PyDict_SetItemString(moddict, to_name, value);
#endif
        }
        Py_DECREF(value);
    } else if (PyErr_ExceptionMatches(PyExc_AttributeError)) {
        PyErr_Clear();
    } else {
        result = -1;
    }
    return result;
}
static CYTHON_SMALL_CODE PyObject* __pyx_pymod_create(PyObject *spec, PyModuleDef *def) {
    PyObject *module = NULL, *moddict, *modname;
    CYTHON_UNUSED_VAR(def);
    if (__Pyx_check_single_interpreter())
        return NULL;
    if (__pyx_m)
        return __Pyx_NewRef(__pyx_m);
    modname = PyObject_GetAttrString(spec, "name");
    if (unlikely(!modname)) goto bad;
    module = PyModule_NewObject(modname);
    Py_DECREF(modname);
    if (unlikely(!module)) goto bad;
#if CYTHON_COMPILING_IN_LIMITED_API
    moddict = module;
#else
    moddict = PyModule_GetDict(module);
    if (unlikely(!moddict)) goto bad;
#endif
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "loader", "__loader__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "origin", "__file__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "parent", "__package__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "submodule_search_locations", "__path__", 0) < 0)) goto bad;
    return module;
bad:
    Py_XDECREF(module);
    return NULL;
}


static CYTHON_SMALL_CODE int __pyx_pymod_exec_transport(PyObject *__pyx_pyinit_module)
#endif
#endif
{
  int stringtab_initialized = 0;
  #if CYTHON_USE_MODULE_STATE
  int pystate_addmodule_run = 0;
  #endif
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *(*__pyx_t_6)(PyObject *);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  if (__pyx_m) {
    if (__pyx_m == __pyx_pyinit_module) return 0;
    PyErr_SetString(PyExc_RuntimeError, "Module 'transport' has already been imported. Re-initialisation is not supported.");
    return -1;
  }
  #elif PY_MAJOR_VERSION >= 3
  if (__pyx_m) return __Pyx_NewRef(__pyx_m);
  #endif
  /*--- Module creation code ---*/
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  __pyx_m = __pyx_pyinit_module;
  Py_INCREF(__pyx_m);
  #else
  #if PY_MAJOR_VERSION < 3
  __pyx_m = Py_InitModule4("transport", __pyx_methods, 0, 0, PYTHON_API_VERSION); Py_XINCREF(__pyx_m);
  if (unlikely(!__pyx_m)) __PYX_ERR(0, 1, __pyx_L1_error)
  #elif CYTHON_USE_MODULE_STATE
  __pyx_t_1 = PyModule_Create(&__pyx_moduledef); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1, __pyx_L1_error)
  {
    int add_module_result = PyState_AddModule(__pyx_t_1, &__pyx_moduledef);
    __pyx_t_1 = 0; /* transfer ownership from __pyx_t_1 to "transport" pseudovariable */
    if (unlikely((add_module_result < 0))) __PYX_ERR(0, 1, __pyx_L1_error)
    pystate_addmodule_run = 1;
  }
  #else
  __pyx_m = PyModule_Create(&__pyx_moduledef);
  if (unlikely(!__pyx_m)) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #endif
  CYTHON_UNUSED_VAR(__pyx_t_1);
  __pyx_d = PyModule_GetDict(__pyx_m); if (unlikely(!__pyx_d)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_d);
  __pyx_b = __Pyx_PyImport_AddModuleRef(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_b)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_cython_runtime = __Pyx_PyImport_AddModuleRef((const char *) "cython_runtime"); if (unlikely(!__pyx_cython_runtime)) __PYX_ERR(0, 1, __pyx_L1_error)
  if (PyObject_SetAttrString(__pyx_m, "__builtins__", __pyx_b) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #if CYTHON_REFNANNY
__Pyx_RefNanny = __Pyx_RefNannyImportAPI("refnanny");
if (!__Pyx_RefNanny) {
  PyErr_Clear();
  __Pyx_RefNanny = __Pyx_RefNannyImportAPI("Cython.Runtime.refnanny");
  if (!__Pyx_RefNanny)
      Py_FatalError("failed to import 'refnanny' module");
}
#endif
  __Pyx_RefNannySetupContext("__Pyx_PyMODINIT_FUNC PyInit_transport(void)", 0);
  if (__Pyx_check_binary_version(__PYX_LIMITED_VERSION_HEX, __Pyx_get_runtime_version(), CYTHON_COMPILING_IN_LIMITED_API) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #ifdef __Pxy_PyFrame_Initialize_Offsets
  __Pxy_PyFrame_Initialize_Offsets();
  #endif
  __pyx_empty_tuple = PyTuple_New(0); if (unlikely(!__pyx_empty_tuple)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_bytes = PyBytes_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_bytes)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_unicode = PyUnicode_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_unicode)) __PYX_ERR(0, 1, __pyx_L1_error)
  #ifdef __Pyx_CyFunction_USED
  if (__pyx_CyFunction_init(__pyx_m) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_FusedFunction_USED
  if (__pyx_FusedFunction_init(__pyx_m) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Coroutine_USED
  if (__pyx_Coroutine_init(__pyx_m) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Generator_USED
  if (__pyx_Generator_init(__pyx_m) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_AsyncGen_USED
  if (__pyx_AsyncGen_init(__pyx_m) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_StopAsyncIteration_USED
  if (__pyx_StopAsyncIteration_init(__pyx_m) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  /*--- Library function declarations ---*/
  /*--- Threads initialization code ---*/
  #if defined(WITH_THREAD) && PY_VERSION_HEX < 0x030700F0 && defined(__PYX_FORCE_INIT_THREADS) && __PYX_FORCE_INIT_THREADS
  PyEval_InitThreads();
  #endif
  /*--- Initialize various global constants etc. ---*/
  if (__Pyx_InitConstants() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  stringtab_initialized = 1;
  if (__Pyx_InitGlobals() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #if PY_MAJOR_VERSION < 3 && (__PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT)
  if (__Pyx_init_sys_getdefaultencoding_params() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  if (__pyx_module_is_main_viur__datastore__transport) {
    if (PyObject_SetAttr(__pyx_m, __pyx_n_s_name_2, __pyx_n_s_main) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  }
  #if PY_MAJOR_VERSION >= 3
  {
    PyObject *modules = PyImport_GetModuleDict(); if (unlikely(!modules)) __PYX_ERR(0, 1, __pyx_L1_error)
    if (!PyDict_GetItemString(modules, "viur.datastore.transport")) {
      if (unlikely((PyDict_SetItemString(modules, "viur.datastore.transport", __pyx_m) < 0))) __PYX_ERR(0, 1, __pyx_L1_error)
    }
  }
  #endif
  /*--- Builtin init code ---*/
  if (__Pyx_InitCachedBuiltins() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Constants init code ---*/
  if (__Pyx_InitCachedConstants() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Global type/function init code ---*/
  (void)__Pyx_modinit_global_init_code();
  (void)__Pyx_modinit_variable_export_code();
  (void)__Pyx_modinit_function_export_code();
  if (unlikely((__Pyx_modinit_type_init_code() < 0))) __PYX_ERR(0, 1, __pyx_L1_error)
  if (unlikely((__Pyx_modinit_type_import_code() < 0))) __PYX_ERR(0, 1, __pyx_L1_error)
  (void)__Pyx_modinit_variable_import_code();
  (void)__Pyx_modinit_function_import_code();
  /*--- Execution code ---*/
  #if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
  if (__Pyx_patch_abc() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif

  /* "viur/datastore/transport.pyx":4
 * # distutils: language = c++
 * # cython: language_level=3
 * from copy import copy             # <<<<<<<<<<<<<<
 * from viur.datastore import cache
 * import google.auth
 */
  __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_n_s_copy);
  __Pyx_GIVEREF(__pyx_n_s_copy);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_copy)) __PYX_ERR(0, 4, __pyx_L1_error);
  __pyx_t_3 = __Pyx_Import(__pyx_n_s_copy, __pyx_t_2, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_copy); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_copy, __pyx_t_2) < 0) __PYX_ERR(0, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "viur/datastore/transport.pyx":5
 * # cython: language_level=3
 * from copy import copy
 * from viur.datastore import cache             # <<<<<<<<<<<<<<
 * import google.auth
 * import requests
 */
  __pyx_t_3 = PyList_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_n_s_cache);
  __Pyx_GIVEREF(__pyx_n_s_cache);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_3, 0, __pyx_n_s_cache)) __PYX_ERR(0, 5, __pyx_L1_error);
  __pyx_t_2 = __Pyx_Import(__pyx_n_s_viur_datastore, __pyx_t_3, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_cache); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_cache, __pyx_t_3) < 0) __PYX_ERR(0, 5, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "viur/datastore/transport.pyx":6
 * from copy import copy
 * from viur.datastore import cache
 * import google.auth             # <<<<<<<<<<<<<<
 * import requests
 * from libcpp cimport bool as boolean_type
 */
  __pyx_t_2 = __Pyx_Import(__pyx_n_s_google_auth, 0, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_google, __pyx_t_2) < 0) __PYX_ERR(0, 6, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "viur/datastore/transport.pyx":7
 * from viur.datastore import cache
 * import google.auth
 * import requests             # <<<<<<<<<<<<<<
 * from libcpp cimport bool as boolean_type
 * from viur.datastore.types import currentTransaction, Entity, Key, QueryDefinition, currentDbAccessLog
 */
  __pyx_t_2 = __Pyx_ImportDottedModule(__pyx_n_s_requests, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_requests, __pyx_t_2) < 0) __PYX_ERR(0, 7, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "viur/datastore/transport.pyx":9
 * import requests
 * from libcpp cimport bool as boolean_type
 * from viur.datastore.types import currentTransaction, Entity, Key, QueryDefinition, currentDbAccessLog             # <<<<<<<<<<<<<<
 * from viur.datastore.config import conf
 * from viur.datastore.errors import *
 */
  __pyx_t_2 = PyList_New(5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 9, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_n_s_currentTransaction);
  __Pyx_GIVEREF(__pyx_n_s_currentTransaction);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_currentTransaction)) __PYX_ERR(0, 9, __pyx_L1_error);
  __Pyx_INCREF(__pyx_n_s_Entity);
  __Pyx_GIVEREF(__pyx_n_s_Entity);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_2, 1, __pyx_n_s_Entity)) __PYX_ERR(0, 9, __pyx_L1_error);
  __Pyx_INCREF(__pyx_n_s_Key);
  __Pyx_GIVEREF(__pyx_n_s_Key);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_2, 2, __pyx_n_s_Key)) __PYX_ERR(0, 9, __pyx_L1_error);
  __Pyx_INCREF(__pyx_n_s_QueryDefinition);
  __Pyx_GIVEREF(__pyx_n_s_QueryDefinition);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_2, 3, __pyx_n_s_QueryDefinition)) __PYX_ERR(0, 9, __pyx_L1_error);
  __Pyx_INCREF(__pyx_n_s_currentDbAccessLog);
  __Pyx_GIVEREF(__pyx_n_s_currentDbAccessLog);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_2, 4, __pyx_n_s_currentDbAccessLog)) __PYX_ERR(0, 9, __pyx_L1_error);
  __pyx_t_3 = __Pyx_Import(__pyx_n_s_viur_datastore_types, __pyx_t_2, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 9, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_currentTransaction); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 9, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_currentTransaction, __pyx_t_2) < 0) __PYX_ERR(0, 9, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_Entity); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 9, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_Entity, __pyx_t_2) < 0) __PYX_ERR(0, 9, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_Key); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 9, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_Key, __pyx_t_2) < 0) __PYX_ERR(0, 9, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_QueryDefinition); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 9, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_QueryDefinition, __pyx_t_2) < 0) __PYX_ERR(0, 9, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_currentDbAccessLog); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 9, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_currentDbAccessLog, __pyx_t_2) < 0) __PYX_ERR(0, 9, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "viur/datastore/transport.pyx":10
 * from libcpp cimport bool as boolean_type
 * from viur.datastore.types import currentTransaction, Entity, Key, QueryDefinition, currentDbAccessLog
 * from viur.datastore.config import conf             # <<<<<<<<<<<<<<
 * from viur.datastore.errors import *
 * from cython.operator cimport preincrement, dereference
 */
  __pyx_t_3 = PyList_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 10, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_n_s_conf);
  __Pyx_GIVEREF(__pyx_n_s_conf);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_3, 0, __pyx_n_s_conf)) __PYX_ERR(0, 10, __pyx_L1_error);
  __pyx_t_2 = __Pyx_Import(__pyx_n_s_viur_datastore_config, __pyx_t_3, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 10, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_conf); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 10, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_conf, __pyx_t_3) < 0) __PYX_ERR(0, 10, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "viur/datastore/transport.pyx":11
 * from viur.datastore.types import currentTransaction, Entity, Key, QueryDefinition, currentDbAccessLog
 * from viur.datastore.config import conf
 * from viur.datastore.errors import *             # <<<<<<<<<<<<<<
 * from cython.operator cimport preincrement, dereference
 * from libc.stdint cimport int64_t, uint64_t
 */
  __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 11, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_n_s__21);
  __Pyx_GIVEREF(__pyx_n_s__21);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s__21)) __PYX_ERR(0, 11, __pyx_L1_error);
  __pyx_t_3 = __Pyx_Import(__pyx_n_s_viur_datastore_errors, __pyx_t_2, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 11, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__pyx_import_star(__pyx_t_3) < 0) __PYX_ERR(0, 11, __pyx_L1_error);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "viur/datastore/transport.pyx":14
 * from cython.operator cimport preincrement, dereference
 * from libc.stdint cimport int64_t, uint64_t
 * from datetime import datetime, timezone             # <<<<<<<<<<<<<<
 * from cpython.bytes cimport PyBytes_AsStringAndSize
 * import json
 */
  __pyx_t_3 = PyList_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 14, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_n_s_datetime);
  __Pyx_GIVEREF(__pyx_n_s_datetime);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_3, 0, __pyx_n_s_datetime)) __PYX_ERR(0, 14, __pyx_L1_error);
  __Pyx_INCREF(__pyx_n_s_timezone);
  __Pyx_GIVEREF(__pyx_n_s_timezone);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_3, 1, __pyx_n_s_timezone)) __PYX_ERR(0, 14, __pyx_L1_error);
  __pyx_t_2 = __Pyx_Import(__pyx_n_s_datetime, __pyx_t_3, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 14, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_datetime); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 14, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_datetime, __pyx_t_3) < 0) __PYX_ERR(0, 14, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_timezone); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 14, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_timezone, __pyx_t_3) < 0) __PYX_ERR(0, 14, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "viur/datastore/transport.pyx":16
 * from datetime import datetime, timezone
 * from cpython.bytes cimport PyBytes_AsStringAndSize
 * import json             # <<<<<<<<<<<<<<
 * from base64 import b64decode, b64encode
 * from typing import Union, List, Any
 */
  __pyx_t_2 = __Pyx_ImportDottedModule(__pyx_n_s_json, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_json, __pyx_t_2) < 0) __PYX_ERR(0, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "viur/datastore/transport.pyx":17
 * from cpython.bytes cimport PyBytes_AsStringAndSize
 * import json
 * from base64 import b64decode, b64encode             # <<<<<<<<<<<<<<
 * from typing import Union, List, Any
 * from requests.exceptions import ConnectionError as RequestsConnectionError
 */
  __pyx_t_2 = PyList_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_n_s_b64decode);
  __Pyx_GIVEREF(__pyx_n_s_b64decode);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_b64decode)) __PYX_ERR(0, 17, __pyx_L1_error);
  __Pyx_INCREF(__pyx_n_s_b64encode);
  __Pyx_GIVEREF(__pyx_n_s_b64encode);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_2, 1, __pyx_n_s_b64encode)) __PYX_ERR(0, 17, __pyx_L1_error);
  __pyx_t_3 = __Pyx_Import(__pyx_n_s_base64, __pyx_t_2, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_b64decode); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_b64decode, __pyx_t_2) < 0) __PYX_ERR(0, 17, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_b64encode); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_b64encode, __pyx_t_2) < 0) __PYX_ERR(0, 17, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "viur/datastore/transport.pyx":18
 * import json
 * from base64 import b64decode, b64encode
 * from typing import Union, List, Any             # <<<<<<<<<<<<<<
 * from requests.exceptions import ConnectionError as RequestsConnectionError
 * import logging
 */
  __pyx_t_3 = PyList_New(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 18, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_n_s_Union);
  __Pyx_GIVEREF(__pyx_n_s_Union);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_3, 0, __pyx_n_s_Union)) __PYX_ERR(0, 18, __pyx_L1_error);
  __Pyx_INCREF(__pyx_n_s_List);
  __Pyx_GIVEREF(__pyx_n_s_List);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_3, 1, __pyx_n_s_List)) __PYX_ERR(0, 18, __pyx_L1_error);
  __Pyx_INCREF(__pyx_n_s_Any);
  __Pyx_GIVEREF(__pyx_n_s_Any);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_3, 2, __pyx_n_s_Any)) __PYX_ERR(0, 18, __pyx_L1_error);
  __pyx_t_2 = __Pyx_Import(__pyx_n_s_typing, __pyx_t_3, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 18, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_Union); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 18, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_Union, __pyx_t_3) < 0) __PYX_ERR(0, 18, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_List); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 18, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_List, __pyx_t_3) < 0) __PYX_ERR(0, 18, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_Any); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 18, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_Any, __pyx_t_3) < 0) __PYX_ERR(0, 18, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "viur/datastore/transport.pyx":19
 * from base64 import b64decode, b64encode
 * from typing import Union, List, Any
 * from requests.exceptions import ConnectionError as RequestsConnectionError             # <<<<<<<<<<<<<<
 * import logging
 * from time import sleep
 */
  __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 19, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_n_s_ConnectionError);
  __Pyx_GIVEREF(__pyx_n_s_ConnectionError);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_ConnectionError)) __PYX_ERR(0, 19, __pyx_L1_error);
  __pyx_t_3 = __Pyx_Import(__pyx_n_s_requests_exceptions, __pyx_t_2, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 19, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_ConnectionError); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 19, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_RequestsConnectionError, __pyx_t_2) < 0) __PYX_ERR(0, 19, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "viur/datastore/transport.pyx":20
 * from typing import Union, List, Any
 * from requests.exceptions import ConnectionError as RequestsConnectionError
 * import logging             # <<<<<<<<<<<<<<
 * from time import sleep
 * ## Start of CPP-Imports required for the simdjson->python bridge
 */
  __pyx_t_3 = __Pyx_ImportDottedModule(__pyx_n_s_logging, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 20, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_logging, __pyx_t_3) < 0) __PYX_ERR(0, 20, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "viur/datastore/transport.pyx":21
 * from requests.exceptions import ConnectionError as RequestsConnectionError
 * import logging
 * from time import sleep             # <<<<<<<<<<<<<<
 * ## Start of CPP-Imports required for the simdjson->python bridge
 * 
 */
  __pyx_t_3 = PyList_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 21, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_n_s_sleep);
  __Pyx_GIVEREF(__pyx_n_s_sleep);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_3, 0, __pyx_n_s_sleep)) __PYX_ERR(0, 21, __pyx_L1_error);
  __pyx_t_2 = __Pyx_Import(__pyx_n_s_time, __pyx_t_3, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 21, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_sleep); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 21, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_sleep, __pyx_t_3) < 0) __PYX_ERR(0, 21, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "viur/datastore/transport.pyx":99
 * 
 * 
 * credentials, projectID = google.auth.default(scopes=["https://www.googleapis.com/auth/datastore"])             # <<<<<<<<<<<<<<
 * _http_internal = google.auth.transport.requests.AuthorizedSession(
 *     credentials,
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_google); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 99, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_auth); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 99, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_default); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 99, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 99, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = PyList_New(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 99, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_INCREF(__pyx_kp_u_https_www_googleapis_com_auth_da);
  __Pyx_GIVEREF(__pyx_kp_u_https_www_googleapis_com_auth_da);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_4, 0, __pyx_kp_u_https_www_googleapis_com_auth_da)) __PYX_ERR(0, 99, __pyx_L1_error);
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_scopes, __pyx_t_4) < 0) __PYX_ERR(0, 99, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_empty_tuple, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 99, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if ((likely(PyTuple_CheckExact(__pyx_t_4))) || (PyList_CheckExact(__pyx_t_4))) {
    PyObject* sequence = __pyx_t_4;
    Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
    if (unlikely(size != 2)) {
      if (size > 2) __Pyx_RaiseTooManyValuesError(2);
      else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
      __PYX_ERR(0, 99, __pyx_L1_error)
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    if (likely(PyTuple_CheckExact(sequence))) {
      __pyx_t_3 = PyTuple_GET_ITEM(sequence, 0); 
      __pyx_t_2 = PyTuple_GET_ITEM(sequence, 1); 
    } else {
      __pyx_t_3 = PyList_GET_ITEM(sequence, 0); 
      __pyx_t_2 = PyList_GET_ITEM(sequence, 1); 
    }
    __Pyx_INCREF(__pyx_t_3);
    __Pyx_INCREF(__pyx_t_2);
    #else
    __pyx_t_3 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 99, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_2 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 99, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    #endif
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else {
    Py_ssize_t index = -1;
    __pyx_t_5 = PyObject_GetIter(__pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 99, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_6 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_5);
    index = 0; __pyx_t_3 = __pyx_t_6(__pyx_t_5); if (unlikely(!__pyx_t_3)) goto __pyx_L2_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_3);
    index = 1; __pyx_t_2 = __pyx_t_6(__pyx_t_5); if (unlikely(!__pyx_t_2)) goto __pyx_L2_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_2);
    if (__Pyx_IternextUnpackEndCheck(__pyx_t_6(__pyx_t_5), 2) < 0) __PYX_ERR(0, 99, __pyx_L1_error)
    __pyx_t_6 = NULL;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    goto __pyx_L3_unpacking_done;
    __pyx_L2_unpacking_failed:;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_6 = NULL;
    if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
    __PYX_ERR(0, 99, __pyx_L1_error)
    __pyx_L3_unpacking_done:;
  }
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_credentials, __pyx_t_3) < 0) __PYX_ERR(0, 99, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_projectID, __pyx_t_2) < 0) __PYX_ERR(0, 99, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "viur/datastore/transport.pyx":100
 * 
 * credentials, projectID = google.auth.default(scopes=["https://www.googleapis.com/auth/datastore"])
 * _http_internal = google.auth.transport.requests.AuthorizedSession(             # <<<<<<<<<<<<<<
 *     credentials,
 *     refresh_timeout=300,
 */
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_google); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 100, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_auth); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 100, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_transport); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 100, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_requests); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 100, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_AuthorizedSession); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 100, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "viur/datastore/transport.pyx":101
 * credentials, projectID = google.auth.default(scopes=["https://www.googleapis.com/auth/datastore"])
 * _http_internal = google.auth.transport.requests.AuthorizedSession(
 *     credentials,             # <<<<<<<<<<<<<<
 *     refresh_timeout=300,
 * )
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_credentials); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 101, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);

  /* "viur/datastore/transport.pyx":100
 * 
 * credentials, projectID = google.auth.default(scopes=["https://www.googleapis.com/auth/datastore"])
 * _http_internal = google.auth.transport.requests.AuthorizedSession(             # <<<<<<<<<<<<<<
 *     credentials,
 *     refresh_timeout=300,
 */
  __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 100, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2)) __PYX_ERR(0, 100, __pyx_L1_error);
  __pyx_t_2 = 0;

  /* "viur/datastore/transport.pyx":102
 * _http_internal = google.auth.transport.requests.AuthorizedSession(
 *     credentials,
 *     refresh_timeout=300,             # <<<<<<<<<<<<<<
 * )
 * 
 */
  __pyx_t_2 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 102, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_refresh_timeout, __pyx_int_300) < 0) __PYX_ERR(0, 102, __pyx_L1_error)

  /* "viur/datastore/transport.pyx":100
 * 
 * credentials, projectID = google.auth.default(scopes=["https://www.googleapis.com/auth/datastore"])
 * _http_internal = google.auth.transport.requests.AuthorizedSession(             # <<<<<<<<<<<<<<
 *     credentials,
 *     refresh_timeout=300,
 */
  __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_3, __pyx_t_2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 100, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_http_internal, __pyx_t_5) < 0) __PYX_ERR(0, 100, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "viur/datastore/transport.pyx":105
 * )
 * 
 * def authenticated_request(url: str, data: bytes) -> requests.Response:             # <<<<<<<<<<<<<<
 *     """
 *         Runs one http request to the datastore rest api, authenticated with the current projects service account.
 */
  __pyx_t_5 = __Pyx_PyDict_NewPresized(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 105, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_url, __pyx_n_s_str) < 0) __PYX_ERR(0, 105, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_data, __pyx_n_s_bytes) < 0) __PYX_ERR(0, 105, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_return, __pyx_kp_s_requests_Response) < 0) __PYX_ERR(0, 105, __pyx_L1_error)
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_4viur_9datastore_9transport_1authenticated_request, 0, __pyx_n_s_authenticated_request, NULL, __pyx_n_s_viur_datastore_transport, __pyx_d, ((PyObject *)__pyx_codeobj__23)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 105, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_2, __pyx_t_5);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_authenticated_request, __pyx_t_2) < 0) __PYX_ERR(0, 105, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "viur/datastore/transport.pyx":126
 *             continue
 * 
 * def keyToPath(key: Key) -> List[dict]:             # <<<<<<<<<<<<<<
 *     """
 *         Converts a Key object to the PathElements expected by the rest API.
 */
  __pyx_t_2 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 126, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_key, __pyx_n_s_Key) < 0) __PYX_ERR(0, 126, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_return, __pyx_kp_s_List_dict) < 0) __PYX_ERR(0, 126, __pyx_L1_error)
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_4viur_9datastore_9transport_3keyToPath, 0, __pyx_n_s_keyToPath, NULL, __pyx_n_s_viur_datastore_transport, __pyx_d, ((PyObject *)__pyx_codeobj__25)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 126, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_5, __pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_keyToPath, __pyx_t_5) < 0) __PYX_ERR(0, 126, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "viur/datastore/transport.pyx":150
 *     return res
 * 
 * def pythonPropToJson(v) -> dict:             # <<<<<<<<<<<<<<
 *     """
 *         Convert a python type to the object representation expected by the datastore rest API.
 */
  __pyx_t_5 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 150, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_return, __pyx_n_s_dict) < 0) __PYX_ERR(0, 150, __pyx_L1_error)
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_4viur_9datastore_9transport_5pythonPropToJson, 0, __pyx_n_s_pythonPropToJson, NULL, __pyx_n_s_viur_datastore_transport, __pyx_d, ((PyObject *)__pyx_codeobj__27)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 150, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_2, __pyx_t_5);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pythonPropToJson, __pyx_t_2) < 0) __PYX_ERR(0, 150, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "viur/datastore/transport.pyx":437
 *         raise ValueError("Unknown type")
 * 
 * def runSingleFilter(queryDefinition: QueryDefinition, limit: int) -> List[Entity]:             # <<<<<<<<<<<<<<
 *     """
 *         Runs a single Query as defined by queryDefinition. The limit of the queryDefinition is ignored and must
 */
  __pyx_t_2 = __Pyx_PyDict_NewPresized(3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 437, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_queryDefinition, __pyx_n_s_QueryDefinition) < 0) __PYX_ERR(0, 437, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_limit, __pyx_n_s_int) < 0) __PYX_ERR(0, 437, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_return, __pyx_kp_s_List_Entity) < 0) __PYX_ERR(0, 437, __pyx_L1_error)
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_4viur_9datastore_9transport_7runSingleFilter, 0, __pyx_n_s_runSingleFilter, NULL, __pyx_n_s_viur_datastore_transport, __pyx_d, ((PyObject *)__pyx_codeobj__29)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 437, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_5, __pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_runSingleFilter, __pyx_t_5) < 0) __PYX_ERR(0, 437, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "viur/datastore/transport.pyx":570
 *     return res
 * 
 * def Get(keys: Union[Key, List[Key]]) -> Union[None, Entity, List[Entity]]:             # <<<<<<<<<<<<<<
 *     """
 *         Fetches the entities determined by keys from the datastore. Returns or inserts None if a key is not found.
 */
  __pyx_t_5 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 570, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_keys, __pyx_kp_s_Union_Key_List_Key) < 0) __PYX_ERR(0, 570, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_return, __pyx_kp_s_Union_None_Entity_List_Entity) < 0) __PYX_ERR(0, 570, __pyx_L1_error)
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_4viur_9datastore_9transport_9Get, 0, __pyx_n_s_Get, NULL, __pyx_n_s_viur_datastore_transport, __pyx_d, ((PyObject *)__pyx_codeobj__31)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 570, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_2, __pyx_t_5);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_Get, __pyx_t_2) < 0) __PYX_ERR(0, 570, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "viur/datastore/transport.pyx":656
 *         return [res.get(key) for key in untouched_keys]  # Sort by order of incoming keys
 * 
 * def Delete(keys: Union[Key, List[Key], Entity, List[Entity]]) -> None:             # <<<<<<<<<<<<<<
 *     """
 *         Deletes the entities stored under the given key(s).
 */
  __pyx_t_2 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 656, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_keys, __pyx_kp_s_Union_Key_List_Key_Entity_List_E) < 0) __PYX_ERR(0, 656, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_return, __pyx_n_s_None) < 0) __PYX_ERR(0, 656, __pyx_L1_error)
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_4viur_9datastore_9transport_11Delete, 0, __pyx_n_s_Delete, NULL, __pyx_n_s_viur_datastore_transport, __pyx_d, ((PyObject *)__pyx_codeobj__33)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 656, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_5, __pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_Delete, __pyx_t_5) < 0) __PYX_ERR(0, 656, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "viur/datastore/transport.pyx":711
 *         cache.delete([str(key) for key in keys])
 * 
 * def Put(entities: Union[Entity, List[Entity]]) -> Union[Entity, List[Entity]]:             # <<<<<<<<<<<<<<
 *     """
 *         Writes the given entities into the datastore. The entities can be from different kinds. If an entity has an
 */
  __pyx_t_5 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 711, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_entities, __pyx_kp_s_Union_Entity_List_Entity) < 0) __PYX_ERR(0, 711, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_return, __pyx_kp_s_Union_Entity_List_Entity) < 0) __PYX_ERR(0, 711, __pyx_L1_error)
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_4viur_9datastore_9transport_13Put, 0, __pyx_n_s_Put, NULL, __pyx_n_s_viur_datastore_transport, __pyx_d, ((PyObject *)__pyx_codeobj__35)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 711, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_2, __pyx_t_5);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_Put, __pyx_t_2) < 0) __PYX_ERR(0, 711, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "viur/datastore/transport.pyx":774
 *     return entities
 * 
 * def RunInTransaction(callback: callable, *args, **kwargs) -> Any:             # <<<<<<<<<<<<<<
 *     """
 *         Runs the given function inside a AID transaction.
 */
  __pyx_t_2 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 774, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_callback, __pyx_n_s_callable) < 0) __PYX_ERR(0, 774, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_return, __pyx_n_s_Any) < 0) __PYX_ERR(0, 774, __pyx_L1_error)
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_4viur_9datastore_9transport_15RunInTransaction, 0, __pyx_n_s_RunInTransaction, NULL, __pyx_n_s_viur_datastore_transport, __pyx_d, ((PyObject *)__pyx_codeobj__37)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 774, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_5, __pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_RunInTransaction, __pyx_t_5) < 0) __PYX_ERR(0, 774, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "viur/datastore/transport.pyx":868
 *         "All retries are exhausted for this transaction")  # If we made it here, all tries are exhausted
 * 
 * def _rollbackTxn(txnKey: str):             # <<<<<<<<<<<<<<
 *     """
 *         Internal helper that aborts the given transaction. It's important to abort pending transactions (instead
 */
  __pyx_t_5 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 868, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_txnKey, __pyx_n_s_str) < 0) __PYX_ERR(0, 868, __pyx_L1_error)
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_4viur_9datastore_9transport_17_rollbackTxn, 0, __pyx_n_s_rollbackTxn, NULL, __pyx_n_s_viur_datastore_transport, __pyx_d, ((PyObject *)__pyx_codeobj__39)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 868, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_2, __pyx_t_5);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_rollbackTxn, __pyx_t_2) < 0) __PYX_ERR(0, 868, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "viur/datastore/transport.pyx":884
 *     )
 * 
 * def AllocateIDs(keys: Union[Key, List[Key]]) -> Union[Key, List[Key]]:             # <<<<<<<<<<<<<<
 *     """
 *         Allocates numeric IDs for the keys given.
 */
  __pyx_t_2 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 884, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_keys, __pyx_kp_s_Union_Key_List_Key) < 0) __PYX_ERR(0, 884, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_return, __pyx_kp_s_Union_Key_List_Key) < 0) __PYX_ERR(0, 884, __pyx_L1_error)
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_4viur_9datastore_9transport_19AllocateIDs, 0, __pyx_n_s_AllocateIDs, NULL, __pyx_n_s_viur_datastore_transport, __pyx_d, ((PyObject *)__pyx_codeobj__41)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 884, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_5, __pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_AllocateIDs, __pyx_t_5) < 0) __PYX_ERR(0, 884, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "viur/datastore/transport.pyx":939
 *             raise ValueError("Invalid data received from Datastore API")
 * 
 * def Count(kind: str = None, up_to= 2 ** 63 - 1, queryDefinition: QueryDefinition = None) -> Union[Key, List[Key]]:             # <<<<<<<<<<<<<<
 *     """
 *         Count all entries in a kind if there is only a kind is provided
 */
  __pyx_t_5 = __Pyx_PyDict_NewPresized(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 939, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_kind, __pyx_n_s_str) < 0) __PYX_ERR(0, 939, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_queryDefinition, __pyx_n_s_QueryDefinition) < 0) __PYX_ERR(0, 939, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_return, __pyx_kp_s_Union_Key_List_Key) < 0) __PYX_ERR(0, 939, __pyx_L1_error)
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_4viur_9datastore_9transport_21Count, 0, __pyx_n_s_Count, NULL, __pyx_n_s_viur_datastore_transport, __pyx_d, ((PyObject *)__pyx_codeobj__43)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 939, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_2, __pyx_tuple__44);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_2, __pyx_t_5);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_Count, __pyx_t_2) < 0) __PYX_ERR(0, 939, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "viur/datastore/transport.pyx":1
 * # distutils: sources = src/viur/datastore/simdjson.cpp             # <<<<<<<<<<<<<<
 * # distutils: language = c++
 * # cython: language_level=3
 */
  __pyx_t_2 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_test, __pyx_t_2) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /*--- Wrapped vars code ---*/

  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  if (__pyx_m) {
    if (__pyx_d && stringtab_initialized) {
      __Pyx_AddTraceback("init viur.datastore.transport", __pyx_clineno, __pyx_lineno, __pyx_filename);
    }
    #if !CYTHON_USE_MODULE_STATE
    Py_CLEAR(__pyx_m);
    #else
    Py_DECREF(__pyx_m);
    if (pystate_addmodule_run) {
      PyObject *tp, *value, *tb;
      PyErr_Fetch(&tp, &value, &tb);
      PyState_RemoveModule(&__pyx_moduledef);
      PyErr_Restore(tp, value, tb);
    }
    #endif
  } else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_ImportError, "init viur.datastore.transport");
  }
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  return (__pyx_m != NULL) ? 0 : -1;
  #elif PY_MAJOR_VERSION >= 3
  return __pyx_m;
  #else
  return;
  #endif
}
/* #### Code section: cleanup_globals ### */
/* #### Code section: cleanup_module ### */
/* #### Code section: main_method ### */
/* #### Code section: utility_code_pragmas ### */
#ifdef _MSC_VER
#pragma warning( push )
/* Warning 4127: conditional expression is constant
 * Cython uses constant conditional expressions to allow in inline functions to be optimized at
 * compile-time, so this warning is not useful
 */
#pragma warning( disable : 4127 )
#endif



/* #### Code section: utility_code_def ### */

/* --- Runtime support code --- */
/* Refnanny */
#if CYTHON_REFNANNY
static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname) {
    PyObject *m = NULL, *p = NULL;
    void *r = NULL;
    m = PyImport_ImportModule(modname);
    if (!m) goto end;
    p = PyObject_GetAttrString(m, "RefNannyAPI");
    if (!p) goto end;
    r = PyLong_AsVoidPtr(p);
end:
    Py_XDECREF(p);
    Py_XDECREF(m);
    return (__Pyx_RefNannyAPIStruct *)r;
}
#endif

/* PyErrExceptionMatches */
#if CYTHON_FAST_THREAD_STATE
static int __Pyx_PyErr_ExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    n = PyTuple_GET_SIZE(tuple);
#if PY_MAJOR_VERSION >= 3
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
#endif
    for (i=0; i<n; i++) {
        if (__Pyx_PyErr_GivenExceptionMatches(exc_type, PyTuple_GET_ITEM(tuple, i))) return 1;
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err) {
    int result;
    PyObject *exc_type;
#if PY_VERSION_HEX >= 0x030C00A6
    PyObject *current_exception = tstate->current_exception;
    if (unlikely(!current_exception)) return 0;
    exc_type = (PyObject*) Py_TYPE(current_exception);
    if (exc_type == err) return 1;
#else
    exc_type = tstate->curexc_type;
    if (exc_type == err) return 1;
    if (unlikely(!exc_type)) return 0;
#endif
    #if CYTHON_AVOID_BORROWED_REFS
    Py_INCREF(exc_type);
    #endif
    if (unlikely(PyTuple_Check(err))) {
        result = __Pyx_PyErr_ExceptionMatchesTuple(exc_type, err);
    } else {
        result = __Pyx_PyErr_GivenExceptionMatches(exc_type, err);
    }
    #if CYTHON_AVOID_BORROWED_REFS
    Py_DECREF(exc_type);
    #endif
    return result;
}
#endif

/* PyErrFetchRestore */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
#if PY_VERSION_HEX >= 0x030C00A6
    PyObject *tmp_value;
    assert(type == NULL || (value != NULL && type == (PyObject*) Py_TYPE(value)));
    if (value) {
        #if CYTHON_COMPILING_IN_CPYTHON
        if (unlikely(((PyBaseExceptionObject*) value)->traceback != tb))
        #endif
            PyException_SetTraceback(value, tb);
    }
    tmp_value = tstate->current_exception;
    tstate->current_exception = value;
    Py_XDECREF(tmp_value);
    Py_XDECREF(type);
    Py_XDECREF(tb);
#else
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    tmp_type = tstate->curexc_type;
    tmp_value = tstate->curexc_value;
    tmp_tb = tstate->curexc_traceback;
    tstate->curexc_type = type;
    tstate->curexc_value = value;
    tstate->curexc_traceback = tb;
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
#endif
}
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
#if PY_VERSION_HEX >= 0x030C00A6
    PyObject* exc_value;
    exc_value = tstate->current_exception;
    tstate->current_exception = 0;
    *value = exc_value;
    *type = NULL;
    *tb = NULL;
    if (exc_value) {
        *type = (PyObject*) Py_TYPE(exc_value);
        Py_INCREF(*type);
        #if CYTHON_COMPILING_IN_CPYTHON
        *tb = ((PyBaseExceptionObject*) exc_value)->traceback;
        Py_XINCREF(*tb);
        #else
        *tb = PyException_GetTraceback(exc_value);
        #endif
    }
#else
    *type = tstate->curexc_type;
    *value = tstate->curexc_value;
    *tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
#endif
}
#endif

/* PyObjectGetAttrStr */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro))
        return tp->tp_getattro(obj, attr_name);
#if PY_MAJOR_VERSION < 3
    if (likely(tp->tp_getattr))
        return tp->tp_getattr(obj, PyString_AS_STRING(attr_name));
#endif
    return PyObject_GetAttr(obj, attr_name);
}
#endif

/* PyObjectGetAttrStrNoError */
#if __PYX_LIMITED_VERSION_HEX < 0x030d00A1
static void __Pyx_PyObject_GetAttrStr_ClearAttributeError(void) {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    if (likely(__Pyx_PyErr_ExceptionMatches(PyExc_AttributeError)))
        __Pyx_PyErr_Clear();
}
#endif
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStrNoError(PyObject* obj, PyObject* attr_name) {
    PyObject *result;
#if __PYX_LIMITED_VERSION_HEX >= 0x030d00A1
    (void) PyObject_GetOptionalAttr(obj, attr_name, &result);
    return result;
#else
#if CYTHON_COMPILING_IN_CPYTHON && CYTHON_USE_TYPE_SLOTS && PY_VERSION_HEX >= 0x030700B1
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro == PyObject_GenericGetAttr)) {
        return _PyObject_GenericGetAttrWithDict(obj, attr_name, NULL, 1);
    }
#endif
    result = __Pyx_PyObject_GetAttrStr(obj, attr_name);
    if (unlikely(!result)) {
        __Pyx_PyObject_GetAttrStr_ClearAttributeError();
    }
    return result;
#endif
}

/* GetBuiltinName */
static PyObject *__Pyx_GetBuiltinName(PyObject *name) {
    PyObject* result = __Pyx_PyObject_GetAttrStrNoError(__pyx_b, name);
    if (unlikely(!result) && !PyErr_Occurred()) {
        PyErr_Format(PyExc_NameError,
#if PY_MAJOR_VERSION >= 3
            "name '%U' is not defined", name);
#else
            "name '%.200s' is not defined", PyString_AS_STRING(name));
#endif
    }
    return result;
}

/* TupleAndListFromArray */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE void __Pyx_copy_object_array(PyObject *const *CYTHON_RESTRICT src, PyObject** CYTHON_RESTRICT dest, Py_ssize_t length) {
    PyObject *v;
    Py_ssize_t i;
    for (i = 0; i < length; i++) {
        v = dest[i] = src[i];
        Py_INCREF(v);
    }
}
static CYTHON_INLINE PyObject *
__Pyx_PyTuple_FromArray(PyObject *const *src, Py_ssize_t n)
{
    PyObject *res;
    if (n <= 0) {
        Py_INCREF(__pyx_empty_tuple);
        return __pyx_empty_tuple;
    }
    res = PyTuple_New(n);
    if (unlikely(res == NULL)) return NULL;
    __Pyx_copy_object_array(src, ((PyTupleObject*)res)->ob_item, n);
    return res;
}
static CYTHON_INLINE PyObject *
__Pyx_PyList_FromArray(PyObject *const *src, Py_ssize_t n)
{
    PyObject *res;
    if (n <= 0) {
        return PyList_New(0);
    }
    res = PyList_New(n);
    if (unlikely(res == NULL)) return NULL;
    __Pyx_copy_object_array(src, ((PyListObject*)res)->ob_item, n);
    return res;
}
#endif

/* BytesEquals */
static CYTHON_INLINE int __Pyx_PyBytes_Equals(PyObject* s1, PyObject* s2, int equals) {
#if CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API
    return PyObject_RichCompareBool(s1, s2, equals);
#else
    if (s1 == s2) {
        return (equals == Py_EQ);
    } else if (PyBytes_CheckExact(s1) & PyBytes_CheckExact(s2)) {
        const char *ps1, *ps2;
        Py_ssize_t length = PyBytes_GET_SIZE(s1);
        if (length != PyBytes_GET_SIZE(s2))
            return (equals == Py_NE);
        ps1 = PyBytes_AS_STRING(s1);
        ps2 = PyBytes_AS_STRING(s2);
        if (ps1[0] != ps2[0]) {
            return (equals == Py_NE);
        } else if (length == 1) {
            return (equals == Py_EQ);
        } else {
            int result;
#if CYTHON_USE_UNICODE_INTERNALS && (PY_VERSION_HEX < 0x030B0000)
            Py_hash_t hash1, hash2;
            hash1 = ((PyBytesObject*)s1)->ob_shash;
            hash2 = ((PyBytesObject*)s2)->ob_shash;
            if (hash1 != hash2 && hash1 != -1 && hash2 != -1) {
                return (equals == Py_NE);
            }
#endif
            result = memcmp(ps1, ps2, (size_t)length);
            return (equals == Py_EQ) ? (result == 0) : (result != 0);
        }
    } else if ((s1 == Py_None) & PyBytes_CheckExact(s2)) {
        return (equals == Py_NE);
    } else if ((s2 == Py_None) & PyBytes_CheckExact(s1)) {
        return (equals == Py_NE);
    } else {
        int result;
        PyObject* py_result = PyObject_RichCompare(s1, s2, equals);
        if (!py_result)
            return -1;
        result = __Pyx_PyObject_IsTrue(py_result);
        Py_DECREF(py_result);
        return result;
    }
#endif
}

/* UnicodeEquals */
static CYTHON_INLINE int __Pyx_PyUnicode_Equals(PyObject* s1, PyObject* s2, int equals) {
#if CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API
    return PyObject_RichCompareBool(s1, s2, equals);
#else
#if PY_MAJOR_VERSION < 3
    PyObject* owned_ref = NULL;
#endif
    int s1_is_unicode, s2_is_unicode;
    if (s1 == s2) {
        goto return_eq;
    }
    s1_is_unicode = PyUnicode_CheckExact(s1);
    s2_is_unicode = PyUnicode_CheckExact(s2);
#if PY_MAJOR_VERSION < 3
    if ((s1_is_unicode & (!s2_is_unicode)) && PyString_CheckExact(s2)) {
        owned_ref = PyUnicode_FromObject(s2);
        if (unlikely(!owned_ref))
            return -1;
        s2 = owned_ref;
        s2_is_unicode = 1;
    } else if ((s2_is_unicode & (!s1_is_unicode)) && PyString_CheckExact(s1)) {
        owned_ref = PyUnicode_FromObject(s1);
        if (unlikely(!owned_ref))
            return -1;
        s1 = owned_ref;
        s1_is_unicode = 1;
    } else if (((!s2_is_unicode) & (!s1_is_unicode))) {
        return __Pyx_PyBytes_Equals(s1, s2, equals);
    }
#endif
    if (s1_is_unicode & s2_is_unicode) {
        Py_ssize_t length;
        int kind;
        void *data1, *data2;
        if (unlikely(__Pyx_PyUnicode_READY(s1) < 0) || unlikely(__Pyx_PyUnicode_READY(s2) < 0))
            return -1;
        length = __Pyx_PyUnicode_GET_LENGTH(s1);
        if (length != __Pyx_PyUnicode_GET_LENGTH(s2)) {
            goto return_ne;
        }
#if CYTHON_USE_UNICODE_INTERNALS
        {
            Py_hash_t hash1, hash2;
        #if CYTHON_PEP393_ENABLED
            hash1 = ((PyASCIIObject*)s1)->hash;
            hash2 = ((PyASCIIObject*)s2)->hash;
        #else
            hash1 = ((PyUnicodeObject*)s1)->hash;
            hash2 = ((PyUnicodeObject*)s2)->hash;
        #endif
            if (hash1 != hash2 && hash1 != -1 && hash2 != -1) {
                goto return_ne;
            }
        }
#endif
        kind = __Pyx_PyUnicode_KIND(s1);
        if (kind != __Pyx_PyUnicode_KIND(s2)) {
            goto return_ne;
        }
        data1 = __Pyx_PyUnicode_DATA(s1);
        data2 = __Pyx_PyUnicode_DATA(s2);
        if (__Pyx_PyUnicode_READ(kind, data1, 0) != __Pyx_PyUnicode_READ(kind, data2, 0)) {
            goto return_ne;
        } else if (length == 1) {
            goto return_eq;
        } else {
            int result = memcmp(data1, data2, (size_t)(length * kind));
            #if PY_MAJOR_VERSION < 3
            Py_XDECREF(owned_ref);
            #endif
            return (equals == Py_EQ) ? (result == 0) : (result != 0);
        }
    } else if ((s1 == Py_None) & s2_is_unicode) {
        goto return_ne;
    } else if ((s2 == Py_None) & s1_is_unicode) {
        goto return_ne;
    } else {
        int result;
        PyObject* py_result = PyObject_RichCompare(s1, s2, equals);
        #if PY_MAJOR_VERSION < 3
        Py_XDECREF(owned_ref);
        #endif
        if (!py_result)
            return -1;
        result = __Pyx_PyObject_IsTrue(py_result);
        Py_DECREF(py_result);
        return result;
    }
return_eq:
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(owned_ref);
    #endif
    return (equals == Py_EQ);
return_ne:
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(owned_ref);
    #endif
    return (equals == Py_NE);
#endif
}

/* fastcall */
#if CYTHON_METH_FASTCALL
static CYTHON_INLINE PyObject * __Pyx_GetKwValue_FASTCALL(PyObject *kwnames, PyObject *const *kwvalues, PyObject *s)
{
    Py_ssize_t i, n = PyTuple_GET_SIZE(kwnames);
    for (i = 0; i < n; i++)
    {
        if (s == PyTuple_GET_ITEM(kwnames, i)) return kwvalues[i];
    }
    for (i = 0; i < n; i++)
    {
        int eq = __Pyx_PyUnicode_Equals(s, PyTuple_GET_ITEM(kwnames, i), Py_EQ);
        if (unlikely(eq != 0)) {
            if (unlikely(eq < 0)) return NULL;
            return kwvalues[i];
        }
    }
    return NULL;
}
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030d0000
CYTHON_UNUSED static PyObject *__Pyx_KwargsAsDict_FASTCALL(PyObject *kwnames, PyObject *const *kwvalues) {
    Py_ssize_t i, nkwargs = PyTuple_GET_SIZE(kwnames);
    PyObject *dict;
    dict = PyDict_New();
    if (unlikely(!dict))
        return NULL;
    for (i=0; i<nkwargs; i++) {
        PyObject *key = PyTuple_GET_ITEM(kwnames, i);
        if (unlikely(PyDict_SetItem(dict, key, kwvalues[i]) < 0))
            goto bad;
    }
    return dict;
bad:
    Py_DECREF(dict);
    return NULL;
}
#endif
#endif

/* RaiseArgTupleInvalid */
static void __Pyx_RaiseArgtupleInvalid(
    const char* func_name,
    int exact,
    Py_ssize_t num_min,
    Py_ssize_t num_max,
    Py_ssize_t num_found)
{
    Py_ssize_t num_expected;
    const char *more_or_less;
    if (num_found < num_min) {
        num_expected = num_min;
        more_or_less = "at least";
    } else {
        num_expected = num_max;
        more_or_less = "at most";
    }
    if (exact) {
        more_or_less = "exactly";
    }
    PyErr_Format(PyExc_TypeError,
                 "%.200s() takes %.8s %" CYTHON_FORMAT_SSIZE_T "d positional argument%.1s (%" CYTHON_FORMAT_SSIZE_T "d given)",
                 func_name, more_or_less, num_expected,
                 (num_expected == 1) ? "" : "s", num_found);
}

/* RaiseDoubleKeywords */
static void __Pyx_RaiseDoubleKeywordsError(
    const char* func_name,
    PyObject* kw_name)
{
    PyErr_Format(PyExc_TypeError,
        #if PY_MAJOR_VERSION >= 3
        "%s() got multiple values for keyword argument '%U'", func_name, kw_name);
        #else
        "%s() got multiple values for keyword argument '%s'", func_name,
        PyString_AsString(kw_name));
        #endif
}

/* ParseKeywords */
static int __Pyx_ParseOptionalKeywords(
    PyObject *kwds,
    PyObject *const *kwvalues,
    PyObject **argnames[],
    PyObject *kwds2,
    PyObject *values[],
    Py_ssize_t num_pos_args,
    const char* function_name)
{
    PyObject *key = 0, *value = 0;
    Py_ssize_t pos = 0;
    PyObject*** name;
    PyObject*** first_kw_arg = argnames + num_pos_args;
    int kwds_is_tuple = CYTHON_METH_FASTCALL && likely(PyTuple_Check(kwds));
    while (1) {
        Py_XDECREF(key); key = NULL;
        Py_XDECREF(value); value = NULL;
        if (kwds_is_tuple) {
            Py_ssize_t size;
#if CYTHON_ASSUME_SAFE_MACROS
            size = PyTuple_GET_SIZE(kwds);
#else
            size = PyTuple_Size(kwds);
            if (size < 0) goto bad;
#endif
            if (pos >= size) break;
#if CYTHON_AVOID_BORROWED_REFS
            key = __Pyx_PySequence_ITEM(kwds, pos);
            if (!key) goto bad;
#elif CYTHON_ASSUME_SAFE_MACROS
            key = PyTuple_GET_ITEM(kwds, pos);
#else
            key = PyTuple_GetItem(kwds, pos);
            if (!key) goto bad;
#endif
            value = kwvalues[pos];
            pos++;
        }
        else
        {
            if (!PyDict_Next(kwds, &pos, &key, &value)) break;
#if CYTHON_AVOID_BORROWED_REFS
            Py_INCREF(key);
#endif
        }
        name = first_kw_arg;
        while (*name && (**name != key)) name++;
        if (*name) {
            values[name-argnames] = value;
#if CYTHON_AVOID_BORROWED_REFS
            Py_INCREF(value);
            Py_DECREF(key);
#endif
            key = NULL;
            value = NULL;
            continue;
        }
#if !CYTHON_AVOID_BORROWED_REFS
        Py_INCREF(key);
#endif
        Py_INCREF(value);
        name = first_kw_arg;
        #if PY_MAJOR_VERSION < 3
        if (likely(PyString_Check(key))) {
            while (*name) {
                if ((CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**name) == PyString_GET_SIZE(key))
                        && _PyString_Eq(**name, key)) {
                    values[name-argnames] = value;
#if CYTHON_AVOID_BORROWED_REFS
                    value = NULL;
#endif
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    if ((**argname == key) || (
                            (CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**argname) == PyString_GET_SIZE(key))
                             && _PyString_Eq(**argname, key))) {
                        goto arg_passed_twice;
                    }
                    argname++;
                }
            }
        } else
        #endif
        if (likely(PyUnicode_Check(key))) {
            while (*name) {
                int cmp = (
                #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                    (__Pyx_PyUnicode_GET_LENGTH(**name) != __Pyx_PyUnicode_GET_LENGTH(key)) ? 1 :
                #endif
                    PyUnicode_Compare(**name, key)
                );
                if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                if (cmp == 0) {
                    values[name-argnames] = value;
#if CYTHON_AVOID_BORROWED_REFS
                    value = NULL;
#endif
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    int cmp = (**argname == key) ? 0 :
                    #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                        (__Pyx_PyUnicode_GET_LENGTH(**argname) != __Pyx_PyUnicode_GET_LENGTH(key)) ? 1 :
                    #endif
                        PyUnicode_Compare(**argname, key);
                    if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                    if (cmp == 0) goto arg_passed_twice;
                    argname++;
                }
            }
        } else
            goto invalid_keyword_type;
        if (kwds2) {
            if (unlikely(PyDict_SetItem(kwds2, key, value))) goto bad;
        } else {
            goto invalid_keyword;
        }
    }
    Py_XDECREF(key);
    Py_XDECREF(value);
    return 0;
arg_passed_twice:
    __Pyx_RaiseDoubleKeywordsError(function_name, key);
    goto bad;
invalid_keyword_type:
    PyErr_Format(PyExc_TypeError,
        "%.200s() keywords must be strings", function_name);
    goto bad;
invalid_keyword:
    #if PY_MAJOR_VERSION < 3
    PyErr_Format(PyExc_TypeError,
        "%.200s() got an unexpected keyword argument '%.200s'",
        function_name, PyString_AsString(key));
    #else
    PyErr_Format(PyExc_TypeError,
        "%s() got an unexpected keyword argument '%U'",
        function_name, key);
    #endif
bad:
    Py_XDECREF(key);
    Py_XDECREF(value);
    return -1;
}

/* ArgTypeTest */
static int __Pyx__ArgTypeTest(PyObject *obj, PyTypeObject *type, const char *name, int exact)
{
    __Pyx_TypeName type_name;
    __Pyx_TypeName obj_type_name;
    if (unlikely(!type)) {
        PyErr_SetString(PyExc_SystemError, "Missing type object");
        return 0;
    }
    else if (exact) {
        #if PY_MAJOR_VERSION == 2
        if ((type == &PyBaseString_Type) && likely(__Pyx_PyBaseString_CheckExact(obj))) return 1;
        #endif
    }
    else {
        if (likely(__Pyx_TypeCheck(obj, type))) return 1;
    }
    type_name = __Pyx_PyType_GetName(type);
    obj_type_name = __Pyx_PyType_GetName(Py_TYPE(obj));
    PyErr_Format(PyExc_TypeError,
        "Argument '%.200s' has incorrect type (expected " __Pyx_FMT_TYPENAME
        ", got " __Pyx_FMT_TYPENAME ")", name, type_name, obj_type_name);
    __Pyx_DECREF_TypeName(type_name);
    __Pyx_DECREF_TypeName(obj_type_name);
    return 0;
}

/* PyDictVersioning */
#if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    return likely(dict) ? __PYX_GET_DICT_VERSION(dict) : 0;
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj) {
    PyObject **dictptr = NULL;
    Py_ssize_t offset = Py_TYPE(obj)->tp_dictoffset;
    if (offset) {
#if CYTHON_COMPILING_IN_CPYTHON
        dictptr = (likely(offset > 0)) ? (PyObject **) ((char *)obj + offset) : _PyObject_GetDictPtr(obj);
#else
        dictptr = _PyObject_GetDictPtr(obj);
#endif
    }
    return (dictptr && *dictptr) ? __PYX_GET_DICT_VERSION(*dictptr) : 0;
}
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    if (unlikely(!dict) || unlikely(tp_dict_version != __PYX_GET_DICT_VERSION(dict)))
        return 0;
    return obj_dict_version == __Pyx_get_object_dict_version(obj);
}
#endif

/* GetModuleGlobalName */
#if CYTHON_USE_DICT_VERSIONS
static PyObject *__Pyx__GetModuleGlobalName(PyObject *name, PY_UINT64_T *dict_version, PyObject **dict_cached_value)
#else
static CYTHON_INLINE PyObject *__Pyx__GetModuleGlobalName(PyObject *name)
#endif
{
    PyObject *result;
#if !CYTHON_AVOID_BORROWED_REFS
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030500A1 && PY_VERSION_HEX < 0x030d0000
    result = _PyDict_GetItem_KnownHash(__pyx_d, name, ((PyASCIIObject *) name)->hash);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    } else if (unlikely(PyErr_Occurred())) {
        return NULL;
    }
#elif CYTHON_COMPILING_IN_LIMITED_API
    if (unlikely(!__pyx_m)) {
        return NULL;
    }
    result = PyObject_GetAttr(__pyx_m, name);
    if (likely(result)) {
        return result;
    }
#else
    result = PyDict_GetItem(__pyx_d, name);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    }
#endif
#else
    result = PyObject_GetItem(__pyx_d, name);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    }
    PyErr_Clear();
#endif
    return __Pyx_GetBuiltinName(name);
}

/* PyObjectCall */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw) {
    PyObject *result;
    ternaryfunc call = Py_TYPE(func)->tp_call;
    if (unlikely(!call))
        return PyObject_Call(func, arg, kw);
    #if PY_MAJOR_VERSION < 3
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    #else
    if (unlikely(Py_EnterRecursiveCall(" while calling a Python object")))
        return NULL;
    #endif
    result = (*call)(func, arg, kw);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* GetTopmostException */
#if CYTHON_USE_EXC_INFO_STACK && CYTHON_FAST_THREAD_STATE
static _PyErr_StackItem *
__Pyx_PyErr_GetTopmostException(PyThreadState *tstate)
{
    _PyErr_StackItem *exc_info = tstate->exc_info;
    while ((exc_info->exc_value == NULL || exc_info->exc_value == Py_None) &&
           exc_info->previous_item != NULL)
    {
        exc_info = exc_info->previous_item;
    }
    return exc_info;
}
#endif

/* SaveResetException */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
  #if CYTHON_USE_EXC_INFO_STACK && PY_VERSION_HEX >= 0x030B00a4
    _PyErr_StackItem *exc_info = __Pyx_PyErr_GetTopmostException(tstate);
    PyObject *exc_value = exc_info->exc_value;
    if (exc_value == NULL || exc_value == Py_None) {
        *value = NULL;
        *type = NULL;
        *tb = NULL;
    } else {
        *value = exc_value;
        Py_INCREF(*value);
        *type = (PyObject*) Py_TYPE(exc_value);
        Py_INCREF(*type);
        *tb = PyException_GetTraceback(exc_value);
    }
  #elif CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = __Pyx_PyErr_GetTopmostException(tstate);
    *type = exc_info->exc_type;
    *value = exc_info->exc_value;
    *tb = exc_info->exc_traceback;
    Py_XINCREF(*type);
    Py_XINCREF(*value);
    Py_XINCREF(*tb);
  #else
    *type = tstate->exc_type;
    *value = tstate->exc_value;
    *tb = tstate->exc_traceback;
    Py_XINCREF(*type);
    Py_XINCREF(*value);
    Py_XINCREF(*tb);
  #endif
}
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
  #if CYTHON_USE_EXC_INFO_STACK && PY_VERSION_HEX >= 0x030B00a4
    _PyErr_StackItem *exc_info = tstate->exc_info;
    PyObject *tmp_value = exc_info->exc_value;
    exc_info->exc_value = value;
    Py_XDECREF(tmp_value);
    Py_XDECREF(type);
    Py_XDECREF(tb);
  #else
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    #if CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = tstate->exc_info;
    tmp_type = exc_info->exc_type;
    tmp_value = exc_info->exc_value;
    tmp_tb = exc_info->exc_traceback;
    exc_info->exc_type = type;
    exc_info->exc_value = value;
    exc_info->exc_traceback = tb;
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = type;
    tstate->exc_value = value;
    tstate->exc_traceback = tb;
    #endif
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
  #endif
}
#endif

/* FastTypeChecks */
#if CYTHON_COMPILING_IN_CPYTHON
static int __Pyx_InBases(PyTypeObject *a, PyTypeObject *b) {
    while (a) {
        a = __Pyx_PyType_GetSlot(a, tp_base, PyTypeObject*);
        if (a == b)
            return 1;
    }
    return b == &PyBaseObject_Type;
}
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b) {
    PyObject *mro;
    if (a == b) return 1;
    mro = a->tp_mro;
    if (likely(mro)) {
        Py_ssize_t i, n;
        n = PyTuple_GET_SIZE(mro);
        for (i = 0; i < n; i++) {
            if (PyTuple_GET_ITEM(mro, i) == (PyObject *)b)
                return 1;
        }
        return 0;
    }
    return __Pyx_InBases(a, b);
}
static CYTHON_INLINE int __Pyx_IsAnySubtype2(PyTypeObject *cls, PyTypeObject *a, PyTypeObject *b) {
    PyObject *mro;
    if (cls == a || cls == b) return 1;
    mro = cls->tp_mro;
    if (likely(mro)) {
        Py_ssize_t i, n;
        n = PyTuple_GET_SIZE(mro);
        for (i = 0; i < n; i++) {
            PyObject *base = PyTuple_GET_ITEM(mro, i);
            if (base == (PyObject *)a || base == (PyObject *)b)
                return 1;
        }
        return 0;
    }
    return __Pyx_InBases(cls, a) || __Pyx_InBases(cls, b);
}
#if PY_MAJOR_VERSION == 2
static int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject* exc_type2) {
    PyObject *exception, *value, *tb;
    int res;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&exception, &value, &tb);
    res = exc_type1 ? PyObject_IsSubclass(err, exc_type1) : 0;
    if (unlikely(res == -1)) {
        PyErr_WriteUnraisable(err);
        res = 0;
    }
    if (!res) {
        res = PyObject_IsSubclass(err, exc_type2);
        if (unlikely(res == -1)) {
            PyErr_WriteUnraisable(err);
            res = 0;
        }
    }
    __Pyx_ErrRestore(exception, value, tb);
    return res;
}
#else
static CYTHON_INLINE int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject *exc_type2) {
    if (exc_type1) {
        return __Pyx_IsAnySubtype2((PyTypeObject*)err, (PyTypeObject*)exc_type1, (PyTypeObject*)exc_type2);
    } else {
        return __Pyx_IsSubtype((PyTypeObject*)err, (PyTypeObject*)exc_type2);
    }
}
#endif
static int __Pyx_PyErr_GivenExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    assert(PyExceptionClass_Check(exc_type));
    n = PyTuple_GET_SIZE(tuple);
#if PY_MAJOR_VERSION >= 3
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
#endif
    for (i=0; i<n; i++) {
        PyObject *t = PyTuple_GET_ITEM(tuple, i);
        #if PY_MAJOR_VERSION < 3
        if (likely(exc_type == t)) return 1;
        #endif
        if (likely(PyExceptionClass_Check(t))) {
            if (__Pyx_inner_PyErr_GivenExceptionMatches2(exc_type, NULL, t)) return 1;
        } else {
        }
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject* exc_type) {
    if (likely(err == exc_type)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        if (likely(PyExceptionClass_Check(exc_type))) {
            return __Pyx_inner_PyErr_GivenExceptionMatches2(err, NULL, exc_type);
        } else if (likely(PyTuple_Check(exc_type))) {
            return __Pyx_PyErr_GivenExceptionMatchesTuple(err, exc_type);
        } else {
        }
    }
    return PyErr_GivenExceptionMatches(err, exc_type);
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *exc_type1, PyObject *exc_type2) {
    assert(PyExceptionClass_Check(exc_type1));
    assert(PyExceptionClass_Check(exc_type2));
    if (likely(err == exc_type1 || err == exc_type2)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        return __Pyx_inner_PyErr_GivenExceptionMatches2(err, exc_type1, exc_type2);
    }
    return (PyErr_GivenExceptionMatches(err, exc_type1) || PyErr_GivenExceptionMatches(err, exc_type2));
}
#endif

/* GetException */
#if CYTHON_FAST_THREAD_STATE
static int __Pyx__GetException(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb)
#else
static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb)
#endif
{
    PyObject *local_type = NULL, *local_value, *local_tb = NULL;
#if CYTHON_FAST_THREAD_STATE
    PyObject *tmp_type, *tmp_value, *tmp_tb;
  #if PY_VERSION_HEX >= 0x030C00A6
    local_value = tstate->current_exception;
    tstate->current_exception = 0;
    if (likely(local_value)) {
        local_type = (PyObject*) Py_TYPE(local_value);
        Py_INCREF(local_type);
        local_tb = PyException_GetTraceback(local_value);
    }
  #else
    local_type = tstate->curexc_type;
    local_value = tstate->curexc_value;
    local_tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
  #endif
#else
    PyErr_Fetch(&local_type, &local_value, &local_tb);
#endif
    PyErr_NormalizeException(&local_type, &local_value, &local_tb);
#if CYTHON_FAST_THREAD_STATE && PY_VERSION_HEX >= 0x030C00A6
    if (unlikely(tstate->current_exception))
#elif CYTHON_FAST_THREAD_STATE
    if (unlikely(tstate->curexc_type))
#else
    if (unlikely(PyErr_Occurred()))
#endif
        goto bad;
    #if PY_MAJOR_VERSION >= 3
    if (local_tb) {
        if (unlikely(PyException_SetTraceback(local_value, local_tb) < 0))
            goto bad;
    }
    #endif
    Py_XINCREF(local_tb);
    Py_XINCREF(local_type);
    Py_XINCREF(local_value);
    *type = local_type;
    *value = local_value;
    *tb = local_tb;
#if CYTHON_FAST_THREAD_STATE
    #if CYTHON_USE_EXC_INFO_STACK
    {
        _PyErr_StackItem *exc_info = tstate->exc_info;
      #if PY_VERSION_HEX >= 0x030B00a4
        tmp_value = exc_info->exc_value;
        exc_info->exc_value = local_value;
        tmp_type = NULL;
        tmp_tb = NULL;
        Py_XDECREF(local_type);
        Py_XDECREF(local_tb);
      #else
        tmp_type = exc_info->exc_type;
        tmp_value = exc_info->exc_value;
        tmp_tb = exc_info->exc_traceback;
        exc_info->exc_type = local_type;
        exc_info->exc_value = local_value;
        exc_info->exc_traceback = local_tb;
      #endif
    }
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = local_type;
    tstate->exc_value = local_value;
    tstate->exc_traceback = local_tb;
    #endif
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
#else
    PyErr_SetExcInfo(local_type, local_value, local_tb);
#endif
    return 0;
bad:
    *type = 0;
    *value = 0;
    *tb = 0;
    Py_XDECREF(local_type);
    Py_XDECREF(local_value);
    Py_XDECREF(local_tb);
    return -1;
}

/* PyFunctionFastCall */
#if CYTHON_FAST_PYCALL && !CYTHON_VECTORCALL
static PyObject* __Pyx_PyFunction_FastCallNoKw(PyCodeObject *co, PyObject **args, Py_ssize_t na,
                                               PyObject *globals) {
    PyFrameObject *f;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    PyObject **fastlocals;
    Py_ssize_t i;
    PyObject *result;
    assert(globals != NULL);
    /* XXX Perhaps we should create a specialized
       PyFrame_New() that doesn't take locals, but does
       take builtins without sanity checking them.
       */
    assert(tstate != NULL);
    f = PyFrame_New(tstate, co, globals, NULL);
    if (f == NULL) {
        return NULL;
    }
    fastlocals = __Pyx_PyFrame_GetLocalsplus(f);
    for (i = 0; i < na; i++) {
        Py_INCREF(*args);
        fastlocals[i] = *args++;
    }
    result = PyEval_EvalFrameEx(f,0);
    ++tstate->recursion_depth;
    Py_DECREF(f);
    --tstate->recursion_depth;
    return result;
}
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject **args, Py_ssize_t nargs, PyObject *kwargs) {
    PyCodeObject *co = (PyCodeObject *)PyFunction_GET_CODE(func);
    PyObject *globals = PyFunction_GET_GLOBALS(func);
    PyObject *argdefs = PyFunction_GET_DEFAULTS(func);
    PyObject *closure;
#if PY_MAJOR_VERSION >= 3
    PyObject *kwdefs;
#endif
    PyObject *kwtuple, **k;
    PyObject **d;
    Py_ssize_t nd;
    Py_ssize_t nk;
    PyObject *result;
    assert(kwargs == NULL || PyDict_Check(kwargs));
    nk = kwargs ? PyDict_Size(kwargs) : 0;
    #if PY_MAJOR_VERSION < 3
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object"))) {
        return NULL;
    }
    #else
    if (unlikely(Py_EnterRecursiveCall(" while calling a Python object"))) {
        return NULL;
    }
    #endif
    if (
#if PY_MAJOR_VERSION >= 3
            co->co_kwonlyargcount == 0 &&
#endif
            likely(kwargs == NULL || nk == 0) &&
            co->co_flags == (CO_OPTIMIZED | CO_NEWLOCALS | CO_NOFREE)) {
        if (argdefs == NULL && co->co_argcount == nargs) {
            result = __Pyx_PyFunction_FastCallNoKw(co, args, nargs, globals);
            goto done;
        }
        else if (nargs == 0 && argdefs != NULL
                 && co->co_argcount == Py_SIZE(argdefs)) {
            /* function called with no arguments, but all parameters have
               a default value: use default values as arguments .*/
            args = &PyTuple_GET_ITEM(argdefs, 0);
            result =__Pyx_PyFunction_FastCallNoKw(co, args, Py_SIZE(argdefs), globals);
            goto done;
        }
    }
    if (kwargs != NULL) {
        Py_ssize_t pos, i;
        kwtuple = PyTuple_New(2 * nk);
        if (kwtuple == NULL) {
            result = NULL;
            goto done;
        }
        k = &PyTuple_GET_ITEM(kwtuple, 0);
        pos = i = 0;
        while (PyDict_Next(kwargs, &pos, &k[i], &k[i+1])) {
            Py_INCREF(k[i]);
            Py_INCREF(k[i+1]);
            i += 2;
        }
        nk = i / 2;
    }
    else {
        kwtuple = NULL;
        k = NULL;
    }
    closure = PyFunction_GET_CLOSURE(func);
#if PY_MAJOR_VERSION >= 3
    kwdefs = PyFunction_GET_KW_DEFAULTS(func);
#endif
    if (argdefs != NULL) {
        d = &PyTuple_GET_ITEM(argdefs, 0);
        nd = Py_SIZE(argdefs);
    }
    else {
        d = NULL;
        nd = 0;
    }
#if PY_MAJOR_VERSION >= 3
    result = PyEval_EvalCodeEx((PyObject*)co, globals, (PyObject *)NULL,
                               args, (int)nargs,
                               k, (int)nk,
                               d, (int)nd, kwdefs, closure);
#else
    result = PyEval_EvalCodeEx(co, globals, (PyObject *)NULL,
                               args, (int)nargs,
                               k, (int)nk,
                               d, (int)nd, closure);
#endif
    Py_XDECREF(kwtuple);
done:
    Py_LeaveRecursiveCall();
    return result;
}
#endif

/* PyObjectCallMethO */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg) {
    PyObject *self, *result;
    PyCFunction cfunc;
    cfunc = __Pyx_CyOrPyCFunction_GET_FUNCTION(func);
    self = __Pyx_CyOrPyCFunction_GET_SELF(func);
    #if PY_MAJOR_VERSION < 3
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    #else
    if (unlikely(Py_EnterRecursiveCall(" while calling a Python object")))
        return NULL;
    #endif
    result = cfunc(self, arg);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyObjectFastCall */
#if PY_VERSION_HEX < 0x03090000 || CYTHON_COMPILING_IN_LIMITED_API
static PyObject* __Pyx_PyObject_FastCall_fallback(PyObject *func, PyObject **args, size_t nargs, PyObject *kwargs) {
    PyObject *argstuple;
    PyObject *result = 0;
    size_t i;
    argstuple = PyTuple_New((Py_ssize_t)nargs);
    if (unlikely(!argstuple)) return NULL;
    for (i = 0; i < nargs; i++) {
        Py_INCREF(args[i]);
        if (__Pyx_PyTuple_SET_ITEM(argstuple, (Py_ssize_t)i, args[i]) < 0) goto bad;
    }
    result = __Pyx_PyObject_Call(func, argstuple, kwargs);
  bad:
    Py_DECREF(argstuple);
    return result;
}
#endif
static CYTHON_INLINE PyObject* __Pyx_PyObject_FastCallDict(PyObject *func, PyObject **args, size_t _nargs, PyObject *kwargs) {
    Py_ssize_t nargs = __Pyx_PyVectorcall_NARGS(_nargs);
#if CYTHON_COMPILING_IN_CPYTHON
    if (nargs == 0 && kwargs == NULL) {
        if (__Pyx_CyOrPyCFunction_Check(func) && likely( __Pyx_CyOrPyCFunction_GET_FLAGS(func) & METH_NOARGS))
            return __Pyx_PyObject_CallMethO(func, NULL);
    }
    else if (nargs == 1 && kwargs == NULL) {
        if (__Pyx_CyOrPyCFunction_Check(func) && likely( __Pyx_CyOrPyCFunction_GET_FLAGS(func) & METH_O))
            return __Pyx_PyObject_CallMethO(func, args[0]);
    }
#endif
    #if PY_VERSION_HEX < 0x030800B1
    #if CYTHON_FAST_PYCCALL
    if (PyCFunction_Check(func)) {
        if (kwargs) {
            return _PyCFunction_FastCallDict(func, args, nargs, kwargs);
        } else {
            return _PyCFunction_FastCallKeywords(func, args, nargs, NULL);
        }
    }
    #if PY_VERSION_HEX >= 0x030700A1
    if (!kwargs && __Pyx_IS_TYPE(func, &PyMethodDescr_Type)) {
        return _PyMethodDescr_FastCallKeywords(func, args, nargs, NULL);
    }
    #endif
    #endif
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(func)) {
        return __Pyx_PyFunction_FastCallDict(func, args, nargs, kwargs);
    }
    #endif
    #endif
    if (kwargs == NULL) {
        #if CYTHON_VECTORCALL
        #if PY_VERSION_HEX < 0x03090000
        vectorcallfunc f = _PyVectorcall_Function(func);
        #else
        vectorcallfunc f = PyVectorcall_Function(func);
        #endif
        if (f) {
            return f(func, args, (size_t)nargs, NULL);
        }
        #elif defined(__Pyx_CyFunction_USED) && CYTHON_BACKPORT_VECTORCALL
        if (__Pyx_CyFunction_CheckExact(func)) {
            __pyx_vectorcallfunc f = __Pyx_CyFunction_func_vectorcall(func);
            if (f) return f(func, args, (size_t)nargs, NULL);
        }
        #endif
    }
    if (nargs == 0) {
        return __Pyx_PyObject_Call(func, __pyx_empty_tuple, kwargs);
    }
    #if PY_VERSION_HEX >= 0x03090000 && !CYTHON_COMPILING_IN_LIMITED_API
    return PyObject_VectorcallDict(func, args, (size_t)nargs, kwargs);
    #else
    return __Pyx_PyObject_FastCall_fallback(func, args, (size_t)nargs, kwargs);
    #endif
}

/* IterFinish */
static CYTHON_INLINE int __Pyx_IterFinish(void) {
    PyObject* exc_type;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    exc_type = __Pyx_PyErr_CurrentExceptionType();
    if (unlikely(exc_type)) {
        if (unlikely(!__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration)))
            return -1;
        __Pyx_PyErr_Clear();
        return 0;
    }
    return 0;
}

/* PyObjectCallNoArg */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func) {
    PyObject *arg[2] = {NULL, NULL};
    return __Pyx_PyObject_FastCall(func, arg + 1, 0 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET);
}

/* PyObjectCallOneArg */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *args[2] = {NULL, arg};
    return __Pyx_PyObject_FastCall(func, args+1, 1 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET);
}

/* PyObjectGetMethod */
static int __Pyx_PyObject_GetMethod(PyObject *obj, PyObject *name, PyObject **method) {
    PyObject *attr;
#if CYTHON_UNPACK_METHODS && CYTHON_COMPILING_IN_CPYTHON && CYTHON_USE_PYTYPE_LOOKUP
    __Pyx_TypeName type_name;
    PyTypeObject *tp = Py_TYPE(obj);
    PyObject *descr;
    descrgetfunc f = NULL;
    PyObject **dictptr, *dict;
    int meth_found = 0;
    assert (*method == NULL);
    if (unlikely(tp->tp_getattro != PyObject_GenericGetAttr)) {
        attr = __Pyx_PyObject_GetAttrStr(obj, name);
        goto try_unpack;
    }
    if (unlikely(tp->tp_dict == NULL) && unlikely(PyType_Ready(tp) < 0)) {
        return 0;
    }
    descr = _PyType_Lookup(tp, name);
    if (likely(descr != NULL)) {
        Py_INCREF(descr);
#if defined(Py_TPFLAGS_METHOD_DESCRIPTOR) && Py_TPFLAGS_METHOD_DESCRIPTOR
        if (__Pyx_PyType_HasFeature(Py_TYPE(descr), Py_TPFLAGS_METHOD_DESCRIPTOR))
#elif PY_MAJOR_VERSION >= 3
        #ifdef __Pyx_CyFunction_USED
        if (likely(PyFunction_Check(descr) || __Pyx_IS_TYPE(descr, &PyMethodDescr_Type) || __Pyx_CyFunction_Check(descr)))
        #else
        if (likely(PyFunction_Check(descr) || __Pyx_IS_TYPE(descr, &PyMethodDescr_Type)))
        #endif
#else
        #ifdef __Pyx_CyFunction_USED
        if (likely(PyFunction_Check(descr) || __Pyx_CyFunction_Check(descr)))
        #else
        if (likely(PyFunction_Check(descr)))
        #endif
#endif
        {
            meth_found = 1;
        } else {
            f = Py_TYPE(descr)->tp_descr_get;
            if (f != NULL && PyDescr_IsData(descr)) {
                attr = f(descr, obj, (PyObject *)Py_TYPE(obj));
                Py_DECREF(descr);
                goto try_unpack;
            }
        }
    }
    dictptr = _PyObject_GetDictPtr(obj);
    if (dictptr != NULL && (dict = *dictptr) != NULL) {
        Py_INCREF(dict);
        attr = __Pyx_PyDict_GetItemStr(dict, name);
        if (attr != NULL) {
            Py_INCREF(attr);
            Py_DECREF(dict);
            Py_XDECREF(descr);
            goto try_unpack;
        }
        Py_DECREF(dict);
    }
    if (meth_found) {
        *method = descr;
        return 1;
    }
    if (f != NULL) {
        attr = f(descr, obj, (PyObject *)Py_TYPE(obj));
        Py_DECREF(descr);
        goto try_unpack;
    }
    if (likely(descr != NULL)) {
        *method = descr;
        return 0;
    }
    type_name = __Pyx_PyType_GetName(tp);
    PyErr_Format(PyExc_AttributeError,
#if PY_MAJOR_VERSION >= 3
                 "'" __Pyx_FMT_TYPENAME "' object has no attribute '%U'",
                 type_name, name);
#else
                 "'" __Pyx_FMT_TYPENAME "' object has no attribute '%.400s'",
                 type_name, PyString_AS_STRING(name));
#endif
    __Pyx_DECREF_TypeName(type_name);
    return 0;
#else
    attr = __Pyx_PyObject_GetAttrStr(obj, name);
    goto try_unpack;
#endif
try_unpack:
#if CYTHON_UNPACK_METHODS
    if (likely(attr) && PyMethod_Check(attr) && likely(PyMethod_GET_SELF(attr) == obj)) {
        PyObject *function = PyMethod_GET_FUNCTION(attr);
        Py_INCREF(function);
        Py_DECREF(attr);
        *method = function;
        return 1;
    }
#endif
    *method = attr;
    return 0;
}

/* PyObjectCallMethod0 */
static PyObject* __Pyx_PyObject_CallMethod0(PyObject* obj, PyObject* method_name) {
    PyObject *method = NULL, *result = NULL;
    int is_method = __Pyx_PyObject_GetMethod(obj, method_name, &method);
    if (likely(is_method)) {
        result = __Pyx_PyObject_CallOneArg(method, obj);
        Py_DECREF(method);
        return result;
    }
    if (unlikely(!method)) goto bad;
    result = __Pyx_PyObject_CallNoArg(method);
    Py_DECREF(method);
bad:
    return result;
}

/* RaiseNeedMoreValuesToUnpack */
static CYTHON_INLINE void __Pyx_RaiseNeedMoreValuesError(Py_ssize_t index) {
    PyErr_Format(PyExc_ValueError,
                 "need more than %" CYTHON_FORMAT_SSIZE_T "d value%.1s to unpack",
                 index, (index == 1) ? "" : "s");
}

/* RaiseTooManyValuesToUnpack */
static CYTHON_INLINE void __Pyx_RaiseTooManyValuesError(Py_ssize_t expected) {
    PyErr_Format(PyExc_ValueError,
                 "too many values to unpack (expected %" CYTHON_FORMAT_SSIZE_T "d)", expected);
}

/* UnpackItemEndCheck */
static int __Pyx_IternextUnpackEndCheck(PyObject *retval, Py_ssize_t expected) {
    if (unlikely(retval)) {
        Py_DECREF(retval);
        __Pyx_RaiseTooManyValuesError(expected);
        return -1;
    }
    return __Pyx_IterFinish();
}

/* RaiseNoneIterError */
static CYTHON_INLINE void __Pyx_RaiseNoneNotIterableError(void) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
}

/* UnpackTupleError */
static void __Pyx_UnpackTupleError(PyObject *t, Py_ssize_t index) {
    if (t == Py_None) {
      __Pyx_RaiseNoneNotIterableError();
    } else if (PyTuple_GET_SIZE(t) < index) {
      __Pyx_RaiseNeedMoreValuesError(PyTuple_GET_SIZE(t));
    } else {
      __Pyx_RaiseTooManyValuesError(index);
    }
}

/* UnpackTuple2 */
static CYTHON_INLINE int __Pyx_unpack_tuple2_exact(
        PyObject* tuple, PyObject** pvalue1, PyObject** pvalue2, int decref_tuple) {
    PyObject *value1 = NULL, *value2 = NULL;
#if CYTHON_COMPILING_IN_PYPY
    value1 = PySequence_ITEM(tuple, 0);  if (unlikely(!value1)) goto bad;
    value2 = PySequence_ITEM(tuple, 1);  if (unlikely(!value2)) goto bad;
#else
    value1 = PyTuple_GET_ITEM(tuple, 0);  Py_INCREF(value1);
    value2 = PyTuple_GET_ITEM(tuple, 1);  Py_INCREF(value2);
#endif
    if (decref_tuple) {
        Py_DECREF(tuple);
    }
    *pvalue1 = value1;
    *pvalue2 = value2;
    return 0;
#if CYTHON_COMPILING_IN_PYPY
bad:
    Py_XDECREF(value1);
    Py_XDECREF(value2);
    if (decref_tuple) { Py_XDECREF(tuple); }
    return -1;
#endif
}
static int __Pyx_unpack_tuple2_generic(PyObject* tuple, PyObject** pvalue1, PyObject** pvalue2,
                                       int has_known_size, int decref_tuple) {
    Py_ssize_t index;
    PyObject *value1 = NULL, *value2 = NULL, *iter = NULL;
    iternextfunc iternext;
    iter = PyObject_GetIter(tuple);
    if (unlikely(!iter)) goto bad;
    if (decref_tuple) { Py_DECREF(tuple); tuple = NULL; }
    iternext = __Pyx_PyObject_GetIterNextFunc(iter);
    value1 = iternext(iter); if (unlikely(!value1)) { index = 0; goto unpacking_failed; }
    value2 = iternext(iter); if (unlikely(!value2)) { index = 1; goto unpacking_failed; }
    if (!has_known_size && unlikely(__Pyx_IternextUnpackEndCheck(iternext(iter), 2))) goto bad;
    Py_DECREF(iter);
    *pvalue1 = value1;
    *pvalue2 = value2;
    return 0;
unpacking_failed:
    if (!has_known_size && __Pyx_IterFinish() == 0)
        __Pyx_RaiseNeedMoreValuesError(index);
bad:
    Py_XDECREF(iter);
    Py_XDECREF(value1);
    Py_XDECREF(value2);
    if (decref_tuple) { Py_XDECREF(tuple); }
    return -1;
}

/* dict_iter */
#if CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
#include <string.h>
#endif
static CYTHON_INLINE PyObject* __Pyx_dict_iterator(PyObject* iterable, int is_dict, PyObject* method_name,
                                                   Py_ssize_t* p_orig_length, int* p_source_is_dict) {
    is_dict = is_dict || likely(PyDict_CheckExact(iterable));
    *p_source_is_dict = is_dict;
    if (is_dict) {
#if !CYTHON_COMPILING_IN_PYPY
        *p_orig_length = PyDict_Size(iterable);
        Py_INCREF(iterable);
        return iterable;
#elif PY_MAJOR_VERSION >= 3
        static PyObject *py_items = NULL, *py_keys = NULL, *py_values = NULL;
        PyObject **pp = NULL;
        if (method_name) {
            const char *name = PyUnicode_AsUTF8(method_name);
            if (strcmp(name, "iteritems") == 0) pp = &py_items;
            else if (strcmp(name, "iterkeys") == 0) pp = &py_keys;
            else if (strcmp(name, "itervalues") == 0) pp = &py_values;
            if (pp) {
                if (!*pp) {
                    *pp = PyUnicode_FromString(name + 4);
                    if (!*pp)
                        return NULL;
                }
                method_name = *pp;
            }
        }
#endif
    }
    *p_orig_length = 0;
    if (method_name) {
        PyObject* iter;
        iterable = __Pyx_PyObject_CallMethod0(iterable, method_name);
        if (!iterable)
            return NULL;
#if !CYTHON_COMPILING_IN_PYPY
        if (PyTuple_CheckExact(iterable) || PyList_CheckExact(iterable))
            return iterable;
#endif
        iter = PyObject_GetIter(iterable);
        Py_DECREF(iterable);
        return iter;
    }
    return PyObject_GetIter(iterable);
}
static CYTHON_INLINE int __Pyx_dict_iter_next(
        PyObject* iter_obj, CYTHON_NCP_UNUSED Py_ssize_t orig_length, CYTHON_NCP_UNUSED Py_ssize_t* ppos,
        PyObject** pkey, PyObject** pvalue, PyObject** pitem, int source_is_dict) {
    PyObject* next_item;
#if !CYTHON_COMPILING_IN_PYPY
    if (source_is_dict) {
        PyObject *key, *value;
        if (unlikely(orig_length != PyDict_Size(iter_obj))) {
            PyErr_SetString(PyExc_RuntimeError, "dictionary changed size during iteration");
            return -1;
        }
        if (unlikely(!PyDict_Next(iter_obj, ppos, &key, &value))) {
            return 0;
        }
        if (pitem) {
            PyObject* tuple = PyTuple_New(2);
            if (unlikely(!tuple)) {
                return -1;
            }
            Py_INCREF(key);
            Py_INCREF(value);
            PyTuple_SET_ITEM(tuple, 0, key);
            PyTuple_SET_ITEM(tuple, 1, value);
            *pitem = tuple;
        } else {
            if (pkey) {
                Py_INCREF(key);
                *pkey = key;
            }
            if (pvalue) {
                Py_INCREF(value);
                *pvalue = value;
            }
        }
        return 1;
    } else if (PyTuple_CheckExact(iter_obj)) {
        Py_ssize_t pos = *ppos;
        if (unlikely(pos >= PyTuple_GET_SIZE(iter_obj))) return 0;
        *ppos = pos + 1;
        next_item = PyTuple_GET_ITEM(iter_obj, pos);
        Py_INCREF(next_item);
    } else if (PyList_CheckExact(iter_obj)) {
        Py_ssize_t pos = *ppos;
        if (unlikely(pos >= PyList_GET_SIZE(iter_obj))) return 0;
        *ppos = pos + 1;
        next_item = PyList_GET_ITEM(iter_obj, pos);
        Py_INCREF(next_item);
    } else
#endif
    {
        next_item = PyIter_Next(iter_obj);
        if (unlikely(!next_item)) {
            return __Pyx_IterFinish();
        }
    }
    if (pitem) {
        *pitem = next_item;
    } else if (pkey && pvalue) {
        if (__Pyx_unpack_tuple2(next_item, pkey, pvalue, source_is_dict, source_is_dict, 1))
            return -1;
    } else if (pkey) {
        *pkey = next_item;
    } else {
        *pvalue = next_item;
    }
    return 1;
}

/* pyfrozenset_new */
static CYTHON_INLINE PyObject* __Pyx_PyFrozenSet_New(PyObject* it) {
    if (it) {
        PyObject* result;
#if CYTHON_COMPILING_IN_PYPY
        PyObject* args;
        args = PyTuple_Pack(1, it);
        if (unlikely(!args))
            return NULL;
        result = PyObject_Call((PyObject*)&PyFrozenSet_Type, args, NULL);
        Py_DECREF(args);
        return result;
#else
        if (PyFrozenSet_CheckExact(it)) {
            Py_INCREF(it);
            return it;
        }
        result = PyFrozenSet_New(it);
        if (unlikely(!result))
            return NULL;
        if ((PY_VERSION_HEX >= 0x031000A1) || likely(PySet_GET_SIZE(result)))
            return result;
        Py_DECREF(result);
#endif
    }
#if CYTHON_USE_TYPE_SLOTS
    return PyFrozenSet_Type.tp_new(&PyFrozenSet_Type, __pyx_empty_tuple, NULL);
#else
    return PyObject_Call((PyObject*)&PyFrozenSet_Type, __pyx_empty_tuple, NULL);
#endif
}

/* PySetContains */
static int __Pyx_PySet_ContainsUnhashable(PyObject *set, PyObject *key) {
    int result = -1;
    if (PySet_Check(key) && PyErr_ExceptionMatches(PyExc_TypeError)) {
        PyObject *tmpkey;
        PyErr_Clear();
        tmpkey = __Pyx_PyFrozenSet_New(key);
        if (tmpkey != NULL) {
            result = PySet_Contains(set, tmpkey);
            Py_DECREF(tmpkey);
        }
    }
    return result;
}
static CYTHON_INLINE int __Pyx_PySet_ContainsTF(PyObject* key, PyObject* set, int eq) {
    int result = PySet_Contains(set, key);
    if (unlikely(result < 0)) {
        result = __Pyx_PySet_ContainsUnhashable(set, key);
    }
    return unlikely(result < 0) ? result : (result == (eq == Py_EQ));
}

/* DictGetItem */
#if PY_MAJOR_VERSION >= 3 && !CYTHON_COMPILING_IN_PYPY
static PyObject *__Pyx_PyDict_GetItem(PyObject *d, PyObject* key) {
    PyObject *value;
    value = PyDict_GetItemWithError(d, key);
    if (unlikely(!value)) {
        if (!PyErr_Occurred()) {
            if (unlikely(PyTuple_Check(key))) {
                PyObject* args = PyTuple_Pack(1, key);
                if (likely(args)) {
                    PyErr_SetObject(PyExc_KeyError, args);
                    Py_DECREF(args);
                }
            } else {
                PyErr_SetObject(PyExc_KeyError, key);
            }
        }
        return NULL;
    }
    Py_INCREF(value);
    return value;
}
#endif

/* PyObjectFormatAndDecref */
static CYTHON_INLINE PyObject* __Pyx_PyObject_FormatSimpleAndDecref(PyObject* s, PyObject* f) {
    if (unlikely(!s)) return NULL;
    if (likely(PyUnicode_CheckExact(s))) return s;
    #if PY_MAJOR_VERSION < 3
    if (likely(PyString_CheckExact(s))) {
        PyObject *result = PyUnicode_FromEncodedObject(s, NULL, "strict");
        Py_DECREF(s);
        return result;
    }
    #endif
    return __Pyx_PyObject_FormatAndDecref(s, f);
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_FormatAndDecref(PyObject* s, PyObject* f) {
    PyObject *result;
    if (unlikely(!s)) return NULL;
    result = PyObject_Format(s, f);
    Py_DECREF(s);
    return result;
}

/* JoinPyUnicode */
static PyObject* __Pyx_PyUnicode_Join(PyObject* value_tuple, Py_ssize_t value_count, Py_ssize_t result_ulength,
                                      Py_UCS4 max_char) {
#if CYTHON_USE_UNICODE_INTERNALS && CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    PyObject *result_uval;
    int result_ukind, kind_shift;
    Py_ssize_t i, char_pos;
    void *result_udata;
    CYTHON_MAYBE_UNUSED_VAR(max_char);
#if CYTHON_PEP393_ENABLED
    result_uval = PyUnicode_New(result_ulength, max_char);
    if (unlikely(!result_uval)) return NULL;
    result_ukind = (max_char <= 255) ? PyUnicode_1BYTE_KIND : (max_char <= 65535) ? PyUnicode_2BYTE_KIND : PyUnicode_4BYTE_KIND;
    kind_shift = (result_ukind == PyUnicode_4BYTE_KIND) ? 2 : result_ukind - 1;
    result_udata = PyUnicode_DATA(result_uval);
#else
    result_uval = PyUnicode_FromUnicode(NULL, result_ulength);
    if (unlikely(!result_uval)) return NULL;
    result_ukind = sizeof(Py_UNICODE);
    kind_shift = (result_ukind == 4) ? 2 : result_ukind - 1;
    result_udata = PyUnicode_AS_UNICODE(result_uval);
#endif
    assert(kind_shift == 2 || kind_shift == 1 || kind_shift == 0);
    char_pos = 0;
    for (i=0; i < value_count; i++) {
        int ukind;
        Py_ssize_t ulength;
        void *udata;
        PyObject *uval = PyTuple_GET_ITEM(value_tuple, i);
        if (unlikely(__Pyx_PyUnicode_READY(uval)))
            goto bad;
        ulength = __Pyx_PyUnicode_GET_LENGTH(uval);
        if (unlikely(!ulength))
            continue;
        if (unlikely((PY_SSIZE_T_MAX >> kind_shift) - ulength < char_pos))
            goto overflow;
        ukind = __Pyx_PyUnicode_KIND(uval);
        udata = __Pyx_PyUnicode_DATA(uval);
        if (!CYTHON_PEP393_ENABLED || ukind == result_ukind) {
            memcpy((char *)result_udata + (char_pos << kind_shift), udata, (size_t) (ulength << kind_shift));
        } else {
            #if PY_VERSION_HEX >= 0x030d0000
            if (unlikely(PyUnicode_CopyCharacters(result_uval, char_pos, uval, 0, ulength) < 0)) goto bad;
            #elif CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030300F0 || defined(_PyUnicode_FastCopyCharacters)
            _PyUnicode_FastCopyCharacters(result_uval, char_pos, uval, 0, ulength);
            #else
            Py_ssize_t j;
            for (j=0; j < ulength; j++) {
                Py_UCS4 uchar = __Pyx_PyUnicode_READ(ukind, udata, j);
                __Pyx_PyUnicode_WRITE(result_ukind, result_udata, char_pos+j, uchar);
            }
            #endif
        }
        char_pos += ulength;
    }
    return result_uval;
overflow:
    PyErr_SetString(PyExc_OverflowError, "join() result is too long for a Python string");
bad:
    Py_DECREF(result_uval);
    return NULL;
#else
    CYTHON_UNUSED_VAR(max_char);
    CYTHON_UNUSED_VAR(result_ulength);
    CYTHON_UNUSED_VAR(value_count);
    return PyUnicode_Join(__pyx_empty_unicode, value_tuple);
#endif
}

/* RaiseException */
#if PY_MAJOR_VERSION < 3
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause) {
    __Pyx_PyThreadState_declare
    CYTHON_UNUSED_VAR(cause);
    Py_XINCREF(type);
    if (!value || value == Py_None)
        value = NULL;
    else
        Py_INCREF(value);
    if (!tb || tb == Py_None)
        tb = NULL;
    else {
        Py_INCREF(tb);
        if (!PyTraceBack_Check(tb)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: arg 3 must be a traceback or None");
            goto raise_error;
        }
    }
    if (PyType_Check(type)) {
#if CYTHON_COMPILING_IN_PYPY
        if (!value) {
            Py_INCREF(Py_None);
            value = Py_None;
        }
#endif
        PyErr_NormalizeException(&type, &value, &tb);
    } else {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto raise_error;
        }
        value = type;
        type = (PyObject*) Py_TYPE(type);
        Py_INCREF(type);
        if (!PyType_IsSubtype((PyTypeObject *)type, (PyTypeObject *)PyExc_BaseException)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: exception class must be a subclass of BaseException");
            goto raise_error;
        }
    }
    __Pyx_PyThreadState_assign
    __Pyx_ErrRestore(type, value, tb);
    return;
raise_error:
    Py_XDECREF(value);
    Py_XDECREF(type);
    Py_XDECREF(tb);
    return;
}
#else
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause) {
    PyObject* owned_instance = NULL;
    if (tb == Py_None) {
        tb = 0;
    } else if (tb && !PyTraceBack_Check(tb)) {
        PyErr_SetString(PyExc_TypeError,
            "raise: arg 3 must be a traceback or None");
        goto bad;
    }
    if (value == Py_None)
        value = 0;
    if (PyExceptionInstance_Check(type)) {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto bad;
        }
        value = type;
        type = (PyObject*) Py_TYPE(value);
    } else if (PyExceptionClass_Check(type)) {
        PyObject *instance_class = NULL;
        if (value && PyExceptionInstance_Check(value)) {
            instance_class = (PyObject*) Py_TYPE(value);
            if (instance_class != type) {
                int is_subclass = PyObject_IsSubclass(instance_class, type);
                if (!is_subclass) {
                    instance_class = NULL;
                } else if (unlikely(is_subclass == -1)) {
                    goto bad;
                } else {
                    type = instance_class;
                }
            }
        }
        if (!instance_class) {
            PyObject *args;
            if (!value)
                args = PyTuple_New(0);
            else if (PyTuple_Check(value)) {
                Py_INCREF(value);
                args = value;
            } else
                args = PyTuple_Pack(1, value);
            if (!args)
                goto bad;
            owned_instance = PyObject_Call(type, args, NULL);
            Py_DECREF(args);
            if (!owned_instance)
                goto bad;
            value = owned_instance;
            if (!PyExceptionInstance_Check(value)) {
                PyErr_Format(PyExc_TypeError,
                             "calling %R should have returned an instance of "
                             "BaseException, not %R",
                             type, Py_TYPE(value));
                goto bad;
            }
        }
    } else {
        PyErr_SetString(PyExc_TypeError,
            "raise: exception class must be a subclass of BaseException");
        goto bad;
    }
    if (cause) {
        PyObject *fixed_cause;
        if (cause == Py_None) {
            fixed_cause = NULL;
        } else if (PyExceptionClass_Check(cause)) {
            fixed_cause = PyObject_CallObject(cause, NULL);
            if (fixed_cause == NULL)
                goto bad;
        } else if (PyExceptionInstance_Check(cause)) {
            fixed_cause = cause;
            Py_INCREF(fixed_cause);
        } else {
            PyErr_SetString(PyExc_TypeError,
                            "exception causes must derive from "
                            "BaseException");
            goto bad;
        }
        PyException_SetCause(value, fixed_cause);
    }
    PyErr_SetObject(type, value);
    if (tb) {
      #if PY_VERSION_HEX >= 0x030C00A6
        PyException_SetTraceback(value, tb);
      #elif CYTHON_FAST_THREAD_STATE
        PyThreadState *tstate = __Pyx_PyThreadState_Current;
        PyObject* tmp_tb = tstate->curexc_traceback;
        if (tb != tmp_tb) {
            Py_INCREF(tb);
            tstate->curexc_traceback = tb;
            Py_XDECREF(tmp_tb);
        }
#else
        PyObject *tmp_type, *tmp_value, *tmp_tb;
        PyErr_Fetch(&tmp_type, &tmp_value, &tmp_tb);
        Py_INCREF(tb);
        PyErr_Restore(tmp_type, tmp_value, tb);
        Py_XDECREF(tmp_tb);
#endif
    }
bad:
    Py_XDECREF(owned_instance);
    return;
}
#endif

/* PyObjectSetAttrStr */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE int __Pyx_PyObject_SetAttrStr(PyObject* obj, PyObject* attr_name, PyObject* value) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_setattro))
        return tp->tp_setattro(obj, attr_name, value);
#if PY_MAJOR_VERSION < 3
    if (likely(tp->tp_setattr))
        return tp->tp_setattr(obj, PyString_AS_STRING(attr_name), value);
#endif
    return PyObject_SetAttr(obj, attr_name, value);
}
#endif

/* SliceObject */
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetSlice(PyObject* obj,
        Py_ssize_t cstart, Py_ssize_t cstop,
        PyObject** _py_start, PyObject** _py_stop, PyObject** _py_slice,
        int has_cstart, int has_cstop, int wraparound) {
    __Pyx_TypeName obj_type_name;
#if CYTHON_USE_TYPE_SLOTS
    PyMappingMethods* mp;
#if PY_MAJOR_VERSION < 3
    PySequenceMethods* ms = Py_TYPE(obj)->tp_as_sequence;
    if (likely(ms && ms->sq_slice)) {
        if (!has_cstart) {
            if (_py_start && (*_py_start != Py_None)) {
                cstart = __Pyx_PyIndex_AsSsize_t(*_py_start);
                if ((cstart == (Py_ssize_t)-1) && PyErr_Occurred()) goto bad;
            } else
                cstart = 0;
        }
        if (!has_cstop) {
            if (_py_stop && (*_py_stop != Py_None)) {
                cstop = __Pyx_PyIndex_AsSsize_t(*_py_stop);
                if ((cstop == (Py_ssize_t)-1) && PyErr_Occurred()) goto bad;
            } else
                cstop = PY_SSIZE_T_MAX;
        }
        if (wraparound && unlikely((cstart < 0) | (cstop < 0)) && likely(ms->sq_length)) {
            Py_ssize_t l = ms->sq_length(obj);
            if (likely(l >= 0)) {
                if (cstop < 0) {
                    cstop += l;
                    if (cstop < 0) cstop = 0;
                }
                if (cstart < 0) {
                    cstart += l;
                    if (cstart < 0) cstart = 0;
                }
            } else {
                if (!PyErr_ExceptionMatches(PyExc_OverflowError))
                    goto bad;
                PyErr_Clear();
            }
        }
        return ms->sq_slice(obj, cstart, cstop);
    }
#else
    CYTHON_UNUSED_VAR(wraparound);
#endif
    mp = Py_TYPE(obj)->tp_as_mapping;
    if (likely(mp && mp->mp_subscript))
#else
    CYTHON_UNUSED_VAR(wraparound);
#endif
    {
        PyObject* result;
        PyObject *py_slice, *py_start, *py_stop;
        if (_py_slice) {
            py_slice = *_py_slice;
        } else {
            PyObject* owned_start = NULL;
            PyObject* owned_stop = NULL;
            if (_py_start) {
                py_start = *_py_start;
            } else {
                if (has_cstart) {
                    owned_start = py_start = PyInt_FromSsize_t(cstart);
                    if (unlikely(!py_start)) goto bad;
                } else
                    py_start = Py_None;
            }
            if (_py_stop) {
                py_stop = *_py_stop;
            } else {
                if (has_cstop) {
                    owned_stop = py_stop = PyInt_FromSsize_t(cstop);
                    if (unlikely(!py_stop)) {
                        Py_XDECREF(owned_start);
                        goto bad;
                    }
                } else
                    py_stop = Py_None;
            }
            py_slice = PySlice_New(py_start, py_stop, Py_None);
            Py_XDECREF(owned_start);
            Py_XDECREF(owned_stop);
            if (unlikely(!py_slice)) goto bad;
        }
#if CYTHON_USE_TYPE_SLOTS
        result = mp->mp_subscript(obj, py_slice);
#else
        result = PyObject_GetItem(obj, py_slice);
#endif
        if (!_py_slice) {
            Py_DECREF(py_slice);
        }
        return result;
    }
    obj_type_name = __Pyx_PyType_GetName(Py_TYPE(obj));
    PyErr_Format(PyExc_TypeError,
        "'" __Pyx_FMT_TYPENAME "' object is unsliceable", obj_type_name);
    __Pyx_DECREF_TypeName(obj_type_name);
bad:
    return NULL;
}

/* PyObjectCall2Args */
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call2Args(PyObject* function, PyObject* arg1, PyObject* arg2) {
    PyObject *args[3] = {NULL, arg1, arg2};
    return __Pyx_PyObject_FastCall(function, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET);
}

/* PyObjectCallMethod1 */
#if !(CYTHON_VECTORCALL && __PYX_LIMITED_VERSION_HEX >= 0x030C00A2)
static PyObject* __Pyx__PyObject_CallMethod1(PyObject* method, PyObject* arg) {
    PyObject *result = __Pyx_PyObject_CallOneArg(method, arg);
    Py_DECREF(method);
    return result;
}
#endif
static PyObject* __Pyx_PyObject_CallMethod1(PyObject* obj, PyObject* method_name, PyObject* arg) {
#if CYTHON_VECTORCALL && __PYX_LIMITED_VERSION_HEX >= 0x030C00A2
    PyObject *args[2] = {obj, arg};
    (void) __Pyx_PyObject_GetMethod;
    (void) __Pyx_PyObject_CallOneArg;
    (void) __Pyx_PyObject_Call2Args;
    return PyObject_VectorcallMethod(method_name, args, 2 | PY_VECTORCALL_ARGUMENTS_OFFSET, NULL);
#else
    PyObject *method = NULL, *result;
    int is_method = __Pyx_PyObject_GetMethod(obj, method_name, &method);
    if (likely(is_method)) {
        result = __Pyx_PyObject_Call2Args(method, obj, arg);
        Py_DECREF(method);
        return result;
    }
    if (unlikely(!method)) return NULL;
    return __Pyx__PyObject_CallMethod1(method, arg);
#endif
}

/* append */
static CYTHON_INLINE int __Pyx_PyObject_Append(PyObject* L, PyObject* x) {
    if (likely(PyList_CheckExact(L))) {
        if (unlikely(__Pyx_PyList_Append(L, x) < 0)) return -1;
    } else {
        PyObject* retval = __Pyx_PyObject_CallMethod1(L, __pyx_n_s_append, x);
        if (unlikely(!retval))
            return -1;
        Py_DECREF(retval);
    }
    return 0;
}

/* GetItemInt */
static PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j) {
    PyObject *r;
    if (unlikely(!j)) return NULL;
    r = PyObject_GetItem(o, j);
    Py_DECREF(j);
    return r;
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              CYTHON_NCP_UNUSED int wraparound,
                                                              CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    Py_ssize_t wrapped_i = i;
    if (wraparound & unlikely(i < 0)) {
        wrapped_i += PyList_GET_SIZE(o);
    }
    if ((!boundscheck) || likely(__Pyx_is_valid_index(wrapped_i, PyList_GET_SIZE(o)))) {
        PyObject *r = PyList_GET_ITEM(o, wrapped_i);
        Py_INCREF(r);
        return r;
    }
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
#else
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              CYTHON_NCP_UNUSED int wraparound,
                                                              CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    Py_ssize_t wrapped_i = i;
    if (wraparound & unlikely(i < 0)) {
        wrapped_i += PyTuple_GET_SIZE(o);
    }
    if ((!boundscheck) || likely(__Pyx_is_valid_index(wrapped_i, PyTuple_GET_SIZE(o)))) {
        PyObject *r = PyTuple_GET_ITEM(o, wrapped_i);
        Py_INCREF(r);
        return r;
    }
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
#else
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i, int is_list,
                                                     CYTHON_NCP_UNUSED int wraparound,
                                                     CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS && CYTHON_USE_TYPE_SLOTS
    if (is_list || PyList_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyList_GET_SIZE(o);
        if ((!boundscheck) || (likely(__Pyx_is_valid_index(n, PyList_GET_SIZE(o))))) {
            PyObject *r = PyList_GET_ITEM(o, n);
            Py_INCREF(r);
            return r;
        }
    }
    else if (PyTuple_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyTuple_GET_SIZE(o);
        if ((!boundscheck) || likely(__Pyx_is_valid_index(n, PyTuple_GET_SIZE(o)))) {
            PyObject *r = PyTuple_GET_ITEM(o, n);
            Py_INCREF(r);
            return r;
        }
    } else {
        PyMappingMethods *mm = Py_TYPE(o)->tp_as_mapping;
        PySequenceMethods *sm = Py_TYPE(o)->tp_as_sequence;
        if (mm && mm->mp_subscript) {
            PyObject *r, *key = PyInt_FromSsize_t(i);
            if (unlikely(!key)) return NULL;
            r = mm->mp_subscript(o, key);
            Py_DECREF(key);
            return r;
        }
        if (likely(sm && sm->sq_item)) {
            if (wraparound && unlikely(i < 0) && likely(sm->sq_length)) {
                Py_ssize_t l = sm->sq_length(o);
                if (likely(l >= 0)) {
                    i += l;
                } else {
                    if (!PyErr_ExceptionMatches(PyExc_OverflowError))
                        return NULL;
                    PyErr_Clear();
                }
            }
            return sm->sq_item(o, i);
        }
    }
#else
    if (is_list || !PyMapping_Check(o)) {
        return PySequence_GetItem(o, i);
    }
#endif
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
}

/* PyIntCompare */
static CYTHON_INLINE int __Pyx_PyInt_BoolEqObjC(PyObject *op1, PyObject *op2, long intval, long inplace) {
    CYTHON_MAYBE_UNUSED_VAR(intval);
    CYTHON_UNUSED_VAR(inplace);
    if (op1 == op2) {
        return 1;
    }
    #if PY_MAJOR_VERSION < 3
    if (likely(PyInt_CheckExact(op1))) {
        const long b = intval;
        long a = PyInt_AS_LONG(op1);
        return (a == b);
    }
    #endif
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(PyLong_CheckExact(op1))) {
        int unequal;
        unsigned long uintval;
        Py_ssize_t size = __Pyx_PyLong_DigitCount(op1);
        const digit* digits = __Pyx_PyLong_Digits(op1);
        if (intval == 0) {
            return (__Pyx_PyLong_IsZero(op1) == 1);
        } else if (intval < 0) {
            if (__Pyx_PyLong_IsNonNeg(op1))
                return 0;
            intval = -intval;
        } else {
            if (__Pyx_PyLong_IsNeg(op1))
                return 0;
        }
        uintval = (unsigned long) intval;
#if PyLong_SHIFT * 4 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 4)) {
            unequal = (size != 5) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[2] != ((uintval >> (2 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[3] != ((uintval >> (3 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[4] != ((uintval >> (4 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
#if PyLong_SHIFT * 3 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 3)) {
            unequal = (size != 4) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[2] != ((uintval >> (2 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[3] != ((uintval >> (3 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
#if PyLong_SHIFT * 2 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 2)) {
            unequal = (size != 3) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[2] != ((uintval >> (2 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
#if PyLong_SHIFT * 1 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 1)) {
            unequal = (size != 2) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
            unequal = (size != 1) || (((unsigned long) digits[0]) != (uintval & (unsigned long) PyLong_MASK));
        return (unequal == 0);
    }
    #endif
    if (PyFloat_CheckExact(op1)) {
        const long b = intval;
#if CYTHON_COMPILING_IN_LIMITED_API
        double a = __pyx_PyFloat_AsDouble(op1);
#else
        double a = PyFloat_AS_DOUBLE(op1);
#endif
        return ((double)a == (double)b);
    }
    return __Pyx_PyObject_IsTrueAndDecref(
        PyObject_RichCompare(op1, op2, Py_EQ));
}

/* PyUnicode_Unicode */
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_Unicode(PyObject *obj) {
    if (unlikely(obj == Py_None))
        obj = __pyx_kp_u_None;
    return __Pyx_NewRef(obj);
}

/* CIntToDigits */
static const char DIGIT_PAIRS_10[2*10*10+1] = {
    "00010203040506070809"
    "10111213141516171819"
    "20212223242526272829"
    "30313233343536373839"
    "40414243444546474849"
    "50515253545556575859"
    "60616263646566676869"
    "70717273747576777879"
    "80818283848586878889"
    "90919293949596979899"
};
static const char DIGIT_PAIRS_8[2*8*8+1] = {
    "0001020304050607"
    "1011121314151617"
    "2021222324252627"
    "3031323334353637"
    "4041424344454647"
    "5051525354555657"
    "6061626364656667"
    "7071727374757677"
};
static const char DIGITS_HEX[2*16+1] = {
    "0123456789abcdef"
    "0123456789ABCDEF"
};

/* BuildPyUnicode */
static PyObject* __Pyx_PyUnicode_BuildFromAscii(Py_ssize_t ulength, char* chars, int clength,
                                                int prepend_sign, char padding_char) {
    PyObject *uval;
    Py_ssize_t uoffset = ulength - clength;
#if CYTHON_USE_UNICODE_INTERNALS
    Py_ssize_t i;
#if CYTHON_PEP393_ENABLED
    void *udata;
    uval = PyUnicode_New(ulength, 127);
    if (unlikely(!uval)) return NULL;
    udata = PyUnicode_DATA(uval);
#else
    Py_UNICODE *udata;
    uval = PyUnicode_FromUnicode(NULL, ulength);
    if (unlikely(!uval)) return NULL;
    udata = PyUnicode_AS_UNICODE(uval);
#endif
    if (uoffset > 0) {
        i = 0;
        if (prepend_sign) {
            __Pyx_PyUnicode_WRITE(PyUnicode_1BYTE_KIND, udata, 0, '-');
            i++;
        }
        for (; i < uoffset; i++) {
            __Pyx_PyUnicode_WRITE(PyUnicode_1BYTE_KIND, udata, i, padding_char);
        }
    }
    for (i=0; i < clength; i++) {
        __Pyx_PyUnicode_WRITE(PyUnicode_1BYTE_KIND, udata, uoffset+i, chars[i]);
    }
#else
    {
        PyObject *sign = NULL, *padding = NULL;
        uval = NULL;
        if (uoffset > 0) {
            prepend_sign = !!prepend_sign;
            if (uoffset > prepend_sign) {
                padding = PyUnicode_FromOrdinal(padding_char);
                if (likely(padding) && uoffset > prepend_sign + 1) {
                    PyObject *tmp;
                    PyObject *repeat = PyInt_FromSsize_t(uoffset - prepend_sign);
                    if (unlikely(!repeat)) goto done_or_error;
                    tmp = PyNumber_Multiply(padding, repeat);
                    Py_DECREF(repeat);
                    Py_DECREF(padding);
                    padding = tmp;
                }
                if (unlikely(!padding)) goto done_or_error;
            }
            if (prepend_sign) {
                sign = PyUnicode_FromOrdinal('-');
                if (unlikely(!sign)) goto done_or_error;
            }
        }
        uval = PyUnicode_DecodeASCII(chars, clength, NULL);
        if (likely(uval) && padding) {
            PyObject *tmp = PyNumber_Add(padding, uval);
            Py_DECREF(uval);
            uval = tmp;
        }
        if (likely(uval) && sign) {
            PyObject *tmp = PyNumber_Add(sign, uval);
            Py_DECREF(uval);
            uval = tmp;
        }
done_or_error:
        Py_XDECREF(padding);
        Py_XDECREF(sign);
    }
#endif
    return uval;
}

/* CIntToPyUnicode */
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_From_Py_ssize_t(Py_ssize_t value, Py_ssize_t width, char padding_char, char format_char) {
    char digits[sizeof(Py_ssize_t)*3+2];
    char *dpos, *end = digits + sizeof(Py_ssize_t)*3+2;
    const char *hex_digits = DIGITS_HEX;
    Py_ssize_t length, ulength;
    int prepend_sign, last_one_off;
    Py_ssize_t remaining;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const Py_ssize_t neg_one = (Py_ssize_t) -1, const_zero = (Py_ssize_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (format_char == 'X') {
        hex_digits += 16;
        format_char = 'x';
    }
    remaining = value;
    last_one_off = 0;
    dpos = end;
    do {
        int digit_pos;
        switch (format_char) {
        case 'o':
            digit_pos = abs((int)(remaining % (8*8)));
            remaining = (Py_ssize_t) (remaining / (8*8));
            dpos -= 2;
            memcpy(dpos, DIGIT_PAIRS_8 + digit_pos * 2, 2);
            last_one_off = (digit_pos < 8);
            break;
        case 'd':
            digit_pos = abs((int)(remaining % (10*10)));
            remaining = (Py_ssize_t) (remaining / (10*10));
            dpos -= 2;
            memcpy(dpos, DIGIT_PAIRS_10 + digit_pos * 2, 2);
            last_one_off = (digit_pos < 10);
            break;
        case 'x':
            *(--dpos) = hex_digits[abs((int)(remaining % 16))];
            remaining = (Py_ssize_t) (remaining / 16);
            break;
        default:
            assert(0);
            break;
        }
    } while (unlikely(remaining != 0));
    assert(!last_one_off || *dpos == '0');
    dpos += last_one_off;
    length = end - dpos;
    ulength = length;
    prepend_sign = 0;
    if (!is_unsigned && value <= neg_one) {
        if (padding_char == ' ' || width <= length + 1) {
            *(--dpos) = '-';
            ++length;
        } else {
            prepend_sign = 1;
        }
        ++ulength;
    }
    if (width > ulength) {
        ulength = width;
    }
    if (ulength == 1) {
        return PyUnicode_FromOrdinal(*dpos);
    }
    return __Pyx_PyUnicode_BuildFromAscii(ulength, dpos, (int) length, prepend_sign, padding_char);
}

/* ObjectGetItem */
#if CYTHON_USE_TYPE_SLOTS
static PyObject *__Pyx_PyObject_GetIndex(PyObject *obj, PyObject *index) {
    PyObject *runerr = NULL;
    Py_ssize_t key_value;
    key_value = __Pyx_PyIndex_AsSsize_t(index);
    if (likely(key_value != -1 || !(runerr = PyErr_Occurred()))) {
        return __Pyx_GetItemInt_Fast(obj, key_value, 0, 1, 1);
    }
    if (PyErr_GivenExceptionMatches(runerr, PyExc_OverflowError)) {
        __Pyx_TypeName index_type_name = __Pyx_PyType_GetName(Py_TYPE(index));
        PyErr_Clear();
        PyErr_Format(PyExc_IndexError,
            "cannot fit '" __Pyx_FMT_TYPENAME "' into an index-sized integer", index_type_name);
        __Pyx_DECREF_TypeName(index_type_name);
    }
    return NULL;
}
static PyObject *__Pyx_PyObject_GetItem_Slow(PyObject *obj, PyObject *key) {
    __Pyx_TypeName obj_type_name;
    if (likely(PyType_Check(obj))) {
        PyObject *meth = __Pyx_PyObject_GetAttrStrNoError(obj, __pyx_n_s_class_getitem);
        if (!meth) {
            PyErr_Clear();
        } else {
            PyObject *result = __Pyx_PyObject_CallOneArg(meth, key);
            Py_DECREF(meth);
            return result;
        }
    }
    obj_type_name = __Pyx_PyType_GetName(Py_TYPE(obj));
    PyErr_Format(PyExc_TypeError,
        "'" __Pyx_FMT_TYPENAME "' object is not subscriptable", obj_type_name);
    __Pyx_DECREF_TypeName(obj_type_name);
    return NULL;
}
static PyObject *__Pyx_PyObject_GetItem(PyObject *obj, PyObject *key) {
    PyTypeObject *tp = Py_TYPE(obj);
    PyMappingMethods *mm = tp->tp_as_mapping;
    PySequenceMethods *sm = tp->tp_as_sequence;
    if (likely(mm && mm->mp_subscript)) {
        return mm->mp_subscript(obj, key);
    }
    if (likely(sm && sm->sq_item)) {
        return __Pyx_PyObject_GetIndex(obj, key);
    }
    return __Pyx_PyObject_GetItem_Slow(obj, key);
}
#endif

/* RaiseUnboundLocalError */
static CYTHON_INLINE void __Pyx_RaiseUnboundLocalError(const char *varname) {
    PyErr_Format(PyExc_UnboundLocalError, "local variable '%s' referenced before assignment", varname);
}

/* pep479 */
static void __Pyx_Generator_Replace_StopIteration(int in_async_gen) {
    PyObject *exc, *val, *tb, *cur_exc;
    __Pyx_PyThreadState_declare
    #ifdef __Pyx_StopAsyncIteration_USED
    int is_async_stopiteration = 0;
    #endif
    CYTHON_MAYBE_UNUSED_VAR(in_async_gen);
    cur_exc = PyErr_Occurred();
    if (likely(!__Pyx_PyErr_GivenExceptionMatches(cur_exc, PyExc_StopIteration))) {
        #ifdef __Pyx_StopAsyncIteration_USED
        if (in_async_gen && unlikely(__Pyx_PyErr_GivenExceptionMatches(cur_exc, __Pyx_PyExc_StopAsyncIteration))) {
            is_async_stopiteration = 1;
        } else
        #endif
            return;
    }
    __Pyx_PyThreadState_assign
    __Pyx_GetException(&exc, &val, &tb);
    Py_XDECREF(exc);
    Py_XDECREF(val);
    Py_XDECREF(tb);
    PyErr_SetString(PyExc_RuntimeError,
        #ifdef __Pyx_StopAsyncIteration_USED
        is_async_stopiteration ? "async generator raised StopAsyncIteration" :
        in_async_gen ? "async generator raised StopIteration" :
        #endif
        "generator raised StopIteration");
}

/* UnpackUnboundCMethod */
static PyObject *__Pyx_SelflessCall(PyObject *method, PyObject *args, PyObject *kwargs) {
    PyObject *result;
    PyObject *selfless_args = PyTuple_GetSlice(args, 1, PyTuple_Size(args));
    if (unlikely(!selfless_args)) return NULL;
    result = PyObject_Call(method, selfless_args, kwargs);
    Py_DECREF(selfless_args);
    return result;
}
static PyMethodDef __Pyx_UnboundCMethod_Def = {
     "CythonUnboundCMethod",
     __PYX_REINTERPRET_FUNCION(PyCFunction, __Pyx_SelflessCall),
     METH_VARARGS | METH_KEYWORDS,
     NULL
};
static int __Pyx_TryUnpackUnboundCMethod(__Pyx_CachedCFunction* target) {
    PyObject *method;
    method = __Pyx_PyObject_GetAttrStr(target->type, *target->method_name);
    if (unlikely(!method))
        return -1;
    target->method = method;
#if CYTHON_COMPILING_IN_CPYTHON
    #if PY_MAJOR_VERSION >= 3
    if (likely(__Pyx_TypeCheck(method, &PyMethodDescr_Type)))
    #else
    if (likely(!__Pyx_CyOrPyCFunction_Check(method)))
    #endif
    {
        PyMethodDescrObject *descr = (PyMethodDescrObject*) method;
        target->func = descr->d_method->ml_meth;
        target->flag = descr->d_method->ml_flags & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_STACKLESS);
    } else
#endif
#if CYTHON_COMPILING_IN_PYPY
#else
    if (PyCFunction_Check(method))
#endif
    {
        PyObject *self;
        int self_found;
#if CYTHON_COMPILING_IN_LIMITED_API || CYTHON_COMPILING_IN_PYPY
        self = PyObject_GetAttrString(method, "__self__");
        if (!self) {
            PyErr_Clear();
        }
#else
        self = PyCFunction_GET_SELF(method);
#endif
        self_found = (self && self != Py_None);
#if CYTHON_COMPILING_IN_LIMITED_API || CYTHON_COMPILING_IN_PYPY
        Py_XDECREF(self);
#endif
        if (self_found) {
            PyObject *unbound_method = PyCFunction_New(&__Pyx_UnboundCMethod_Def, method);
            if (unlikely(!unbound_method)) return -1;
            Py_DECREF(method);
            target->method = unbound_method;
        }
    }
    return 0;
}

/* CallUnboundCMethod1 */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_CallUnboundCMethod1(__Pyx_CachedCFunction* cfunc, PyObject* self, PyObject* arg) {
    if (likely(cfunc->func)) {
        int flag = cfunc->flag;
        if (flag == METH_O) {
            return (*(cfunc->func))(self, arg);
        } else if ((PY_VERSION_HEX >= 0x030600B1) && flag == METH_FASTCALL) {
            #if PY_VERSION_HEX >= 0x030700A0
                return (*(__Pyx_PyCFunctionFast)(void*)(PyCFunction)cfunc->func)(self, &arg, 1);
            #else
                return (*(__Pyx_PyCFunctionFastWithKeywords)(void*)(PyCFunction)cfunc->func)(self, &arg, 1, NULL);
            #endif
        } else if ((PY_VERSION_HEX >= 0x030700A0) && flag == (METH_FASTCALL | METH_KEYWORDS)) {
            return (*(__Pyx_PyCFunctionFastWithKeywords)(void*)(PyCFunction)cfunc->func)(self, &arg, 1, NULL);
        }
    }
    return __Pyx__CallUnboundCMethod1(cfunc, self, arg);
}
#endif
static PyObject* __Pyx__CallUnboundCMethod1(__Pyx_CachedCFunction* cfunc, PyObject* self, PyObject* arg){
    PyObject *args, *result = NULL;
    if (unlikely(!cfunc->func && !cfunc->method) && unlikely(__Pyx_TryUnpackUnboundCMethod(cfunc) < 0)) return NULL;
#if CYTHON_COMPILING_IN_CPYTHON
    if (cfunc->func && (cfunc->flag & METH_VARARGS)) {
        args = PyTuple_New(1);
        if (unlikely(!args)) goto bad;
        Py_INCREF(arg);
        PyTuple_SET_ITEM(args, 0, arg);
        if (cfunc->flag & METH_KEYWORDS)
            result = (*(PyCFunctionWithKeywords)(void*)(PyCFunction)cfunc->func)(self, args, NULL);
        else
            result = (*cfunc->func)(self, args);
    } else {
        args = PyTuple_New(2);
        if (unlikely(!args)) goto bad;
        Py_INCREF(self);
        PyTuple_SET_ITEM(args, 0, self);
        Py_INCREF(arg);
        PyTuple_SET_ITEM(args, 1, arg);
        result = __Pyx_PyObject_Call(cfunc->method, args, NULL);
    }
#else
    args = PyTuple_Pack(2, self, arg);
    if (unlikely(!args)) goto bad;
    result = __Pyx_PyObject_Call(cfunc->method, args, NULL);
#endif
bad:
    Py_XDECREF(args);
    return result;
}

/* PyIntBinop */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyInt_AddObjC(PyObject *op1, PyObject *op2, long intval, int inplace, int zerodivision_check) {
    CYTHON_MAYBE_UNUSED_VAR(intval);
    CYTHON_MAYBE_UNUSED_VAR(inplace);
    CYTHON_UNUSED_VAR(zerodivision_check);
    #if PY_MAJOR_VERSION < 3
    if (likely(PyInt_CheckExact(op1))) {
        const long b = intval;
        long x;
        long a = PyInt_AS_LONG(op1);
        
            x = (long)((unsigned long)a + (unsigned long)b);
            if (likely((x^a) >= 0 || (x^b) >= 0))
                return PyInt_FromLong(x);
            return PyLong_Type.tp_as_number->nb_add(op1, op2);
    }
    #endif
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(PyLong_CheckExact(op1))) {
        const long b = intval;
        long a, x;
#ifdef HAVE_LONG_LONG
        const PY_LONG_LONG llb = intval;
        PY_LONG_LONG lla, llx;
#endif
        if (unlikely(__Pyx_PyLong_IsZero(op1))) {
            return __Pyx_NewRef(op2);
        }
        if (likely(__Pyx_PyLong_IsCompact(op1))) {
            a = __Pyx_PyLong_CompactValue(op1);
        } else {
            const digit* digits = __Pyx_PyLong_Digits(op1);
            const Py_ssize_t size = __Pyx_PyLong_SignedDigitCount(op1);
            switch (size) {
                case -2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        a = -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    #ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
                    #endif
                    }
                    CYTHON_FALLTHROUGH;
                case 2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        a = (long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    #ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
                    #endif
                    }
                    CYTHON_FALLTHROUGH;
                case -3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        a = -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    #ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
                    #endif
                    }
                    CYTHON_FALLTHROUGH;
                case 3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        a = (long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    #ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
                    #endif
                    }
                    CYTHON_FALLTHROUGH;
                case -4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                        a = -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    #ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
                    #endif
                    }
                    CYTHON_FALLTHROUGH;
                case 4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                        a = (long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    #ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
                    #endif
                    }
                    CYTHON_FALLTHROUGH;
                default: return PyLong_Type.tp_as_number->nb_add(op1, op2);
            }
        }
                x = a + b;
            return PyLong_FromLong(x);
#ifdef HAVE_LONG_LONG
        long_long:
                llx = lla + llb;
            return PyLong_FromLongLong(llx);
#endif
        
        
    }
    #endif
    if (PyFloat_CheckExact(op1)) {
        const long b = intval;
#if CYTHON_COMPILING_IN_LIMITED_API
        double a = __pyx_PyFloat_AsDouble(op1);
#else
        double a = PyFloat_AS_DOUBLE(op1);
#endif
            double result;
            
            PyFPE_START_PROTECT("add", return NULL)
            result = ((double)a) + (double)b;
            PyFPE_END_PROTECT(result)
            return PyFloat_FromDouble(result);
    }
    return (inplace ? PyNumber_InPlaceAdd : PyNumber_Add)(op1, op2);
}
#endif

/* CallUnboundCMethod2 */
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030600B1
static CYTHON_INLINE PyObject *__Pyx_CallUnboundCMethod2(__Pyx_CachedCFunction *cfunc, PyObject *self, PyObject *arg1, PyObject *arg2) {
    if (likely(cfunc->func)) {
        PyObject *args[2] = {arg1, arg2};
        if (cfunc->flag == METH_FASTCALL) {
            #if PY_VERSION_HEX >= 0x030700A0
            return (*(__Pyx_PyCFunctionFast)(void*)(PyCFunction)cfunc->func)(self, args, 2);
            #else
            return (*(__Pyx_PyCFunctionFastWithKeywords)(void*)(PyCFunction)cfunc->func)(self, args, 2, NULL);
            #endif
        }
        #if PY_VERSION_HEX >= 0x030700A0
        if (cfunc->flag == (METH_FASTCALL | METH_KEYWORDS))
            return (*(__Pyx_PyCFunctionFastWithKeywords)(void*)(PyCFunction)cfunc->func)(self, args, 2, NULL);
        #endif
    }
    return __Pyx__CallUnboundCMethod2(cfunc, self, arg1, arg2);
}
#endif
static PyObject* __Pyx__CallUnboundCMethod2(__Pyx_CachedCFunction* cfunc, PyObject* self, PyObject* arg1, PyObject* arg2){
    PyObject *args, *result = NULL;
    if (unlikely(!cfunc->func && !cfunc->method) && unlikely(__Pyx_TryUnpackUnboundCMethod(cfunc) < 0)) return NULL;
#if CYTHON_COMPILING_IN_CPYTHON
    if (cfunc->func && (cfunc->flag & METH_VARARGS)) {
        args = PyTuple_New(2);
        if (unlikely(!args)) goto bad;
        Py_INCREF(arg1);
        PyTuple_SET_ITEM(args, 0, arg1);
        Py_INCREF(arg2);
        PyTuple_SET_ITEM(args, 1, arg2);
        if (cfunc->flag & METH_KEYWORDS)
            result = (*(PyCFunctionWithKeywords)(void*)(PyCFunction)cfunc->func)(self, args, NULL);
        else
            result = (*cfunc->func)(self, args);
    } else {
        args = PyTuple_New(3);
        if (unlikely(!args)) goto bad;
        Py_INCREF(self);
        PyTuple_SET_ITEM(args, 0, self);
        Py_INCREF(arg1);
        PyTuple_SET_ITEM(args, 1, arg1);
        Py_INCREF(arg2);
        PyTuple_SET_ITEM(args, 2, arg2);
        result = __Pyx_PyObject_Call(cfunc->method, args, NULL);
    }
#else
    args = PyTuple_Pack(3, self, arg1, arg2);
    if (unlikely(!args)) goto bad;
    result = __Pyx_PyObject_Call(cfunc->method, args, NULL);
#endif
bad:
    Py_XDECREF(args);
    return result;
}

/* dict_getitem_default */
static PyObject* __Pyx_PyDict_GetItemDefault(PyObject* d, PyObject* key, PyObject* default_value) {
    PyObject* value;
#if PY_MAJOR_VERSION >= 3 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07020000)
    value = PyDict_GetItemWithError(d, key);
    if (unlikely(!value)) {
        if (unlikely(PyErr_Occurred()))
            return NULL;
        value = default_value;
    }
    Py_INCREF(value);
    if ((1));
#else
    if (PyString_CheckExact(key) || PyUnicode_CheckExact(key) || PyInt_CheckExact(key)) {
        value = PyDict_GetItem(d, key);
        if (unlikely(!value)) {
            value = default_value;
        }
        Py_INCREF(value);
    }
#endif
    else {
        if (default_value == Py_None)
            value = __Pyx_CallUnboundCMethod1(&__pyx_umethod_PyDict_Type_get, d, key);
        else
            value = __Pyx_CallUnboundCMethod2(&__pyx_umethod_PyDict_Type_get, d, key, default_value);
    }
    return value;
}

/* py_abs */
#if CYTHON_USE_PYLONG_INTERNALS
static PyObject *__Pyx_PyLong_AbsNeg(PyObject *n) {
#if PY_VERSION_HEX >= 0x030C00A7
    if (likely(__Pyx_PyLong_IsCompact(n))) {
        return PyLong_FromSize_t(__Pyx_PyLong_CompactValueUnsigned(n));
    }
#else
    if (likely(Py_SIZE(n) == -1)) {
        return PyLong_FromUnsignedLong(__Pyx_PyLong_Digits(n)[0]);
    }
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030d0000
    {
        PyObject *copy = _PyLong_Copy((PyLongObject*)n);
        if (likely(copy)) {
            #if PY_VERSION_HEX >= 0x030C00A7
            ((PyLongObject*)copy)->long_value.lv_tag = ((PyLongObject*)copy)->long_value.lv_tag & ~_PyLong_SIGN_MASK;
            #else
            __Pyx_SET_SIZE(copy, -Py_SIZE(copy));
            #endif
        }
        return copy;
    }
#else
    return PyNumber_Negative(n);
#endif
}
#endif

/* py_dict_pop */
static CYTHON_INLINE PyObject *__Pyx_PyDict_Pop(PyObject *d, PyObject *key, PyObject *default_value) {
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX > 0x030600B3 & PY_VERSION_HEX < 0x030d0000
    if ((1)) {
        return _PyDict_Pop(d, key, default_value);
    } else
#endif
    if (default_value) {
        return __Pyx_CallUnboundCMethod2(&__pyx_umethod_PyDict_Type_pop, d, key, default_value);
    } else {
        return __Pyx_CallUnboundCMethod1(&__pyx_umethod_PyDict_Type_pop, d, key);
    }
}

/* SwapException */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx__ExceptionSwap(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
  #if CYTHON_USE_EXC_INFO_STACK && PY_VERSION_HEX >= 0x030B00a4
    _PyErr_StackItem *exc_info = tstate->exc_info;
    tmp_value = exc_info->exc_value;
    exc_info->exc_value = *value;
    if (tmp_value == NULL || tmp_value == Py_None) {
        Py_XDECREF(tmp_value);
        tmp_value = NULL;
        tmp_type = NULL;
        tmp_tb = NULL;
    } else {
        tmp_type = (PyObject*) Py_TYPE(tmp_value);
        Py_INCREF(tmp_type);
        #if CYTHON_COMPILING_IN_CPYTHON
        tmp_tb = ((PyBaseExceptionObject*) tmp_value)->traceback;
        Py_XINCREF(tmp_tb);
        #else
        tmp_tb = PyException_GetTraceback(tmp_value);
        #endif
    }
  #elif CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = tstate->exc_info;
    tmp_type = exc_info->exc_type;
    tmp_value = exc_info->exc_value;
    tmp_tb = exc_info->exc_traceback;
    exc_info->exc_type = *type;
    exc_info->exc_value = *value;
    exc_info->exc_traceback = *tb;
  #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = *type;
    tstate->exc_value = *value;
    tstate->exc_traceback = *tb;
  #endif
    *type = tmp_type;
    *value = tmp_value;
    *tb = tmp_tb;
}
#else
static CYTHON_INLINE void __Pyx_ExceptionSwap(PyObject **type, PyObject **value, PyObject **tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    PyErr_GetExcInfo(&tmp_type, &tmp_value, &tmp_tb);
    PyErr_SetExcInfo(*type, *value, *tb);
    *type = tmp_type;
    *value = tmp_value;
    *tb = tmp_tb;
}
#endif

/* PyNumberPow2 */
static PyObject* __Pyx__PyNumber_PowerOf2(PyObject *two, PyObject *exp, PyObject *none, int inplace) {
#if !CYTHON_COMPILING_IN_PYPY
    Py_ssize_t shiftby;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_CheckExact(exp))) {
        shiftby = PyInt_AS_LONG(exp);
    } else
#endif
    if (likely(PyLong_CheckExact(exp))) {
        #if CYTHON_USE_PYLONG_INTERNALS
        if (__Pyx_PyLong_IsZero(exp)) {
            return PyInt_FromLong(1L);
        } else if (__Pyx_PyLong_IsNeg(exp)) {
            goto fallback;
        } else if (__Pyx_PyLong_IsCompact(exp)) {
            shiftby = __Pyx_PyLong_CompactValueUnsigned(exp);
        } else {
            shiftby = PyLong_AsSsize_t(exp);
        }
        #else
        shiftby = PyLong_AsSsize_t(exp);
        #endif
    } else {
        goto fallback;
    }
    if (likely(shiftby >= 0)) {
        if ((size_t)shiftby <= sizeof(long) * 8 - 2) {
            long value = 1L << shiftby;
            return PyInt_FromLong(value);
#ifdef HAVE_LONG_LONG
        } else if ((size_t)shiftby <= sizeof(unsigned PY_LONG_LONG) * 8 - 1) {
            unsigned PY_LONG_LONG value = ((unsigned PY_LONG_LONG)1) << shiftby;
            return PyLong_FromUnsignedLongLong(value);
#endif
        } else {
            PyObject *result, *one = PyInt_FromLong(1L);
            if (unlikely(!one)) return NULL;
            result = PyNumber_Lshift(one, exp);
            Py_DECREF(one);
            return result;
        }
    } else if (shiftby == -1 && PyErr_Occurred()) {
        PyErr_Clear();
    }
fallback:
#endif
    return (inplace ? PyNumber_InPlacePower : PyNumber_Power)(two, exp, none);
}

/* SliceTupleAndList */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE void __Pyx_crop_slice(Py_ssize_t* _start, Py_ssize_t* _stop, Py_ssize_t* _length) {
    Py_ssize_t start = *_start, stop = *_stop, length = *_length;
    if (start < 0) {
        start += length;
        if (start < 0)
            start = 0;
    }
    if (stop < 0)
        stop += length;
    else if (stop > length)
        stop = length;
    *_length = stop - start;
    *_start = start;
    *_stop = stop;
}
static CYTHON_INLINE PyObject* __Pyx_PyList_GetSlice(
            PyObject* src, Py_ssize_t start, Py_ssize_t stop) {
    Py_ssize_t length = PyList_GET_SIZE(src);
    __Pyx_crop_slice(&start, &stop, &length);
    if (length <= 0) {
        return PyList_New(0);
    }
    return __Pyx_PyList_FromArray(((PyListObject*)src)->ob_item + start, length);
}
static CYTHON_INLINE PyObject* __Pyx_PyTuple_GetSlice(
            PyObject* src, Py_ssize_t start, Py_ssize_t stop) {
    Py_ssize_t length = PyTuple_GET_SIZE(src);
    __Pyx_crop_slice(&start, &stop, &length);
    return __Pyx_PyTuple_FromArray(((PyTupleObject*)src)->ob_item + start, length);
}
#endif

/* RaiseUnexpectedTypeError */
static int
__Pyx_RaiseUnexpectedTypeError(const char *expected, PyObject *obj)
{
    __Pyx_TypeName obj_type_name = __Pyx_PyType_GetName(Py_TYPE(obj));
    PyErr_Format(PyExc_TypeError, "Expected %s, got " __Pyx_FMT_TYPENAME,
                 expected, obj_type_name);
    __Pyx_DECREF_TypeName(obj_type_name);
    return 0;
}

/* FixUpExtensionType */
#if CYTHON_USE_TYPE_SPECS
static int __Pyx_fix_up_extension_type_from_spec(PyType_Spec *spec, PyTypeObject *type) {
#if PY_VERSION_HEX > 0x030900B1 || CYTHON_COMPILING_IN_LIMITED_API
    CYTHON_UNUSED_VAR(spec);
    CYTHON_UNUSED_VAR(type);
#else
    const PyType_Slot *slot = spec->slots;
    while (slot && slot->slot && slot->slot != Py_tp_members)
        slot++;
    if (slot && slot->slot == Py_tp_members) {
        int changed = 0;
#if !(PY_VERSION_HEX <= 0x030900b1 && CYTHON_COMPILING_IN_CPYTHON)
        const
#endif
            PyMemberDef *memb = (PyMemberDef*) slot->pfunc;
        while (memb && memb->name) {
            if (memb->name[0] == '_' && memb->name[1] == '_') {
#if PY_VERSION_HEX < 0x030900b1
                if (strcmp(memb->name, "__weaklistoffset__") == 0) {
                    assert(memb->type == T_PYSSIZET);
                    assert(memb->flags == READONLY);
                    type->tp_weaklistoffset = memb->offset;
                    changed = 1;
                }
                else if (strcmp(memb->name, "__dictoffset__") == 0) {
                    assert(memb->type == T_PYSSIZET);
                    assert(memb->flags == READONLY);
                    type->tp_dictoffset = memb->offset;
                    changed = 1;
                }
#if CYTHON_METH_FASTCALL
                else if (strcmp(memb->name, "__vectorcalloffset__") == 0) {
                    assert(memb->type == T_PYSSIZET);
                    assert(memb->flags == READONLY);
#if PY_VERSION_HEX >= 0x030800b4
                    type->tp_vectorcall_offset = memb->offset;
#else
                    type->tp_print = (printfunc) memb->offset;
#endif
                    changed = 1;
                }
#endif
#else
                if ((0));
#endif
#if PY_VERSION_HEX <= 0x030900b1 && CYTHON_COMPILING_IN_CPYTHON
                else if (strcmp(memb->name, "__module__") == 0) {
                    PyObject *descr;
                    assert(memb->type == T_OBJECT);
                    assert(memb->flags == 0 || memb->flags == READONLY);
                    descr = PyDescr_NewMember(type, memb);
                    if (unlikely(!descr))
                        return -1;
                    if (unlikely(PyDict_SetItem(type->tp_dict, PyDescr_NAME(descr), descr) < 0)) {
                        Py_DECREF(descr);
                        return -1;
                    }
                    Py_DECREF(descr);
                    changed = 1;
                }
#endif
            }
            memb++;
        }
        if (changed)
            PyType_Modified(type);
    }
#endif
    return 0;
}
#endif

/* ValidateBasesTuple */
#if CYTHON_COMPILING_IN_CPYTHON || CYTHON_COMPILING_IN_LIMITED_API || CYTHON_USE_TYPE_SPECS
static int __Pyx_validate_bases_tuple(const char *type_name, Py_ssize_t dictoffset, PyObject *bases) {
    Py_ssize_t i, n;
#if CYTHON_ASSUME_SAFE_MACROS
    n = PyTuple_GET_SIZE(bases);
#else
    n = PyTuple_Size(bases);
    if (n < 0) return -1;
#endif
    for (i = 1; i < n; i++)
    {
#if CYTHON_AVOID_BORROWED_REFS
        PyObject *b0 = PySequence_GetItem(bases, i);
        if (!b0) return -1;
#elif CYTHON_ASSUME_SAFE_MACROS
        PyObject *b0 = PyTuple_GET_ITEM(bases, i);
#else
        PyObject *b0 = PyTuple_GetItem(bases, i);
        if (!b0) return -1;
#endif
        PyTypeObject *b;
#if PY_MAJOR_VERSION < 3
        if (PyClass_Check(b0))
        {
            PyErr_Format(PyExc_TypeError, "base class '%.200s' is an old-style class",
                         PyString_AS_STRING(((PyClassObject*)b0)->cl_name));
#if CYTHON_AVOID_BORROWED_REFS
            Py_DECREF(b0);
#endif
            return -1;
        }
#endif
        b = (PyTypeObject*) b0;
        if (!__Pyx_PyType_HasFeature(b, Py_TPFLAGS_HEAPTYPE))
        {
            __Pyx_TypeName b_name = __Pyx_PyType_GetName(b);
            PyErr_Format(PyExc_TypeError,
                "base class '" __Pyx_FMT_TYPENAME "' is not a heap type", b_name);
            __Pyx_DECREF_TypeName(b_name);
#if CYTHON_AVOID_BORROWED_REFS
            Py_DECREF(b0);
#endif
            return -1;
        }
        if (dictoffset == 0)
        {
            Py_ssize_t b_dictoffset = 0;
#if CYTHON_USE_TYPE_SLOTS || CYTHON_COMPILING_IN_PYPY
            b_dictoffset = b->tp_dictoffset;
#else
            PyObject *py_b_dictoffset = PyObject_GetAttrString((PyObject*)b, "__dictoffset__");
            if (!py_b_dictoffset) goto dictoffset_return;
            b_dictoffset = PyLong_AsSsize_t(py_b_dictoffset);
            Py_DECREF(py_b_dictoffset);
            if (b_dictoffset == -1 && PyErr_Occurred()) goto dictoffset_return;
#endif
            if (b_dictoffset) {
                {
                    __Pyx_TypeName b_name = __Pyx_PyType_GetName(b);
                    PyErr_Format(PyExc_TypeError,
                        "extension type '%.200s' has no __dict__ slot, "
                        "but base type '" __Pyx_FMT_TYPENAME "' has: "
                        "either add 'cdef dict __dict__' to the extension type "
                        "or add '__slots__ = [...]' to the base type",
                        type_name, b_name);
                    __Pyx_DECREF_TypeName(b_name);
                }
#if !(CYTHON_USE_TYPE_SLOTS || CYTHON_COMPILING_IN_PYPY)
              dictoffset_return:
#endif
#if CYTHON_AVOID_BORROWED_REFS
                Py_DECREF(b0);
#endif
                return -1;
            }
        }
#if CYTHON_AVOID_BORROWED_REFS
        Py_DECREF(b0);
#endif
    }
    return 0;
}
#endif

/* PyType_Ready */
static int __Pyx_PyType_Ready(PyTypeObject *t) {
#if CYTHON_USE_TYPE_SPECS || !(CYTHON_COMPILING_IN_CPYTHON || CYTHON_COMPILING_IN_LIMITED_API) || defined(PYSTON_MAJOR_VERSION)
    (void)__Pyx_PyObject_CallMethod0;
#if CYTHON_USE_TYPE_SPECS
    (void)__Pyx_validate_bases_tuple;
#endif
    return PyType_Ready(t);
#else
    int r;
    PyObject *bases = __Pyx_PyType_GetSlot(t, tp_bases, PyObject*);
    if (bases && unlikely(__Pyx_validate_bases_tuple(t->tp_name, t->tp_dictoffset, bases) == -1))
        return -1;
#if PY_VERSION_HEX >= 0x03050000 && !defined(PYSTON_MAJOR_VERSION)
    {
        int gc_was_enabled;
    #if PY_VERSION_HEX >= 0x030A00b1
        gc_was_enabled = PyGC_Disable();
        (void)__Pyx_PyObject_CallMethod0;
    #else
        PyObject *ret, *py_status;
        PyObject *gc = NULL;
        #if PY_VERSION_HEX >= 0x030700a1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM+0 >= 0x07030400)
        gc = PyImport_GetModule(__pyx_kp_u_gc);
        #endif
        if (unlikely(!gc)) gc = PyImport_Import(__pyx_kp_u_gc);
        if (unlikely(!gc)) return -1;
        py_status = __Pyx_PyObject_CallMethod0(gc, __pyx_kp_u_isenabled);
        if (unlikely(!py_status)) {
            Py_DECREF(gc);
            return -1;
        }
        gc_was_enabled = __Pyx_PyObject_IsTrue(py_status);
        Py_DECREF(py_status);
        if (gc_was_enabled > 0) {
            ret = __Pyx_PyObject_CallMethod0(gc, __pyx_kp_u_disable);
            if (unlikely(!ret)) {
                Py_DECREF(gc);
                return -1;
            }
            Py_DECREF(ret);
        } else if (unlikely(gc_was_enabled == -1)) {
            Py_DECREF(gc);
            return -1;
        }
    #endif
        t->tp_flags |= Py_TPFLAGS_HEAPTYPE;
#if PY_VERSION_HEX >= 0x030A0000
        t->tp_flags |= Py_TPFLAGS_IMMUTABLETYPE;
#endif
#else
        (void)__Pyx_PyObject_CallMethod0;
#endif
    r = PyType_Ready(t);
#if PY_VERSION_HEX >= 0x03050000 && !defined(PYSTON_MAJOR_VERSION)
        t->tp_flags &= ~Py_TPFLAGS_HEAPTYPE;
    #if PY_VERSION_HEX >= 0x030A00b1
        if (gc_was_enabled)
            PyGC_Enable();
    #else
        if (gc_was_enabled) {
            PyObject *tp, *v, *tb;
            PyErr_Fetch(&tp, &v, &tb);
            ret = __Pyx_PyObject_CallMethod0(gc, __pyx_kp_u_enable);
            if (likely(ret || r == -1)) {
                Py_XDECREF(ret);
                PyErr_Restore(tp, v, tb);
            } else {
                Py_XDECREF(tp);
                Py_XDECREF(v);
                Py_XDECREF(tb);
                r = -1;
            }
        }
        Py_DECREF(gc);
    #endif
    }
#endif
    return r;
#endif
}

/* PyObject_GenericGetAttrNoDict */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject *__Pyx_RaiseGenericGetAttributeError(PyTypeObject *tp, PyObject *attr_name) {
    __Pyx_TypeName type_name = __Pyx_PyType_GetName(tp);
    PyErr_Format(PyExc_AttributeError,
#if PY_MAJOR_VERSION >= 3
                 "'" __Pyx_FMT_TYPENAME "' object has no attribute '%U'",
                 type_name, attr_name);
#else
                 "'" __Pyx_FMT_TYPENAME "' object has no attribute '%.400s'",
                 type_name, PyString_AS_STRING(attr_name));
#endif
    __Pyx_DECREF_TypeName(type_name);
    return NULL;
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_GenericGetAttrNoDict(PyObject* obj, PyObject* attr_name) {
    PyObject *descr;
    PyTypeObject *tp = Py_TYPE(obj);
    if (unlikely(!PyString_Check(attr_name))) {
        return PyObject_GenericGetAttr(obj, attr_name);
    }
    assert(!tp->tp_dictoffset);
    descr = _PyType_Lookup(tp, attr_name);
    if (unlikely(!descr)) {
        return __Pyx_RaiseGenericGetAttributeError(tp, attr_name);
    }
    Py_INCREF(descr);
    #if PY_MAJOR_VERSION < 3
    if (likely(PyType_HasFeature(Py_TYPE(descr), Py_TPFLAGS_HAVE_CLASS)))
    #endif
    {
        descrgetfunc f = Py_TYPE(descr)->tp_descr_get;
        if (unlikely(f)) {
            PyObject *res = f(descr, obj, (PyObject *)tp);
            Py_DECREF(descr);
            return res;
        }
    }
    return descr;
}
#endif

/* TypeImport */
#ifndef __PYX_HAVE_RT_ImportType_3_0_10
#define __PYX_HAVE_RT_ImportType_3_0_10
static PyTypeObject *__Pyx_ImportType_3_0_10(PyObject *module, const char *module_name, const char *class_name,
    size_t size, size_t alignment, enum __Pyx_ImportType_CheckSize_3_0_10 check_size)
{
    PyObject *result = 0;
    char warning[200];
    Py_ssize_t basicsize;
    Py_ssize_t itemsize;
#if CYTHON_COMPILING_IN_LIMITED_API
    PyObject *py_basicsize;
    PyObject *py_itemsize;
#endif
    result = PyObject_GetAttrString(module, class_name);
    if (!result)
        goto bad;
    if (!PyType_Check(result)) {
        PyErr_Format(PyExc_TypeError,
            "%.200s.%.200s is not a type object",
            module_name, class_name);
        goto bad;
    }
#if !CYTHON_COMPILING_IN_LIMITED_API
    basicsize = ((PyTypeObject *)result)->tp_basicsize;
    itemsize = ((PyTypeObject *)result)->tp_itemsize;
#else
    py_basicsize = PyObject_GetAttrString(result, "__basicsize__");
    if (!py_basicsize)
        goto bad;
    basicsize = PyLong_AsSsize_t(py_basicsize);
    Py_DECREF(py_basicsize);
    py_basicsize = 0;
    if (basicsize == (Py_ssize_t)-1 && PyErr_Occurred())
        goto bad;
    py_itemsize = PyObject_GetAttrString(result, "__itemsize__");
    if (!py_itemsize)
        goto bad;
    itemsize = PyLong_AsSsize_t(py_itemsize);
    Py_DECREF(py_itemsize);
    py_itemsize = 0;
    if (itemsize == (Py_ssize_t)-1 && PyErr_Occurred())
        goto bad;
#endif
    if (itemsize) {
        if (size % alignment) {
            alignment = size % alignment;
        }
        if (itemsize < (Py_ssize_t)alignment)
            itemsize = (Py_ssize_t)alignment;
    }
    if ((size_t)(basicsize + itemsize) < size) {
        PyErr_Format(PyExc_ValueError,
            "%.200s.%.200s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize+itemsize);
        goto bad;
    }
    if (check_size == __Pyx_ImportType_CheckSize_Error_3_0_10 &&
            ((size_t)basicsize > size || (size_t)(basicsize + itemsize) < size)) {
        PyErr_Format(PyExc_ValueError,
            "%.200s.%.200s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd-%zd from PyObject",
            module_name, class_name, size, basicsize, basicsize+itemsize);
        goto bad;
    }
    else if (check_size == __Pyx_ImportType_CheckSize_Warn_3_0_10 && (size_t)basicsize > size) {
        PyOS_snprintf(warning, sizeof(warning),
            "%s.%s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize);
        if (PyErr_WarnEx(NULL, warning, 0) < 0) goto bad;
    }
    return (PyTypeObject *)result;
bad:
    Py_XDECREF(result);
    return NULL;
}
#endif

/* Import */
static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level) {
    PyObject *module = 0;
    PyObject *empty_dict = 0;
    PyObject *empty_list = 0;
    #if PY_MAJOR_VERSION < 3
    PyObject *py_import;
    py_import = __Pyx_PyObject_GetAttrStr(__pyx_b, __pyx_n_s_import);
    if (unlikely(!py_import))
        goto bad;
    if (!from_list) {
        empty_list = PyList_New(0);
        if (unlikely(!empty_list))
            goto bad;
        from_list = empty_list;
    }
    #endif
    empty_dict = PyDict_New();
    if (unlikely(!empty_dict))
        goto bad;
    {
        #if PY_MAJOR_VERSION >= 3
        if (level == -1) {
            if (strchr(__Pyx_MODULE_NAME, '.') != NULL) {
                module = PyImport_ImportModuleLevelObject(
                    name, __pyx_d, empty_dict, from_list, 1);
                if (unlikely(!module)) {
                    if (unlikely(!PyErr_ExceptionMatches(PyExc_ImportError)))
                        goto bad;
                    PyErr_Clear();
                }
            }
            level = 0;
        }
        #endif
        if (!module) {
            #if PY_MAJOR_VERSION < 3
            PyObject *py_level = PyInt_FromLong(level);
            if (unlikely(!py_level))
                goto bad;
            module = PyObject_CallFunctionObjArgs(py_import,
                name, __pyx_d, empty_dict, from_list, py_level, (PyObject *)NULL);
            Py_DECREF(py_level);
            #else
            module = PyImport_ImportModuleLevelObject(
                name, __pyx_d, empty_dict, from_list, level);
            #endif
        }
    }
bad:
    Py_XDECREF(empty_dict);
    Py_XDECREF(empty_list);
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(py_import);
    #endif
    return module;
}

/* ImportFrom */
static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name) {
    PyObject* value = __Pyx_PyObject_GetAttrStr(module, name);
    if (unlikely(!value) && PyErr_ExceptionMatches(PyExc_AttributeError)) {
        const char* module_name_str = 0;
        PyObject* module_name = 0;
        PyObject* module_dot = 0;
        PyObject* full_name = 0;
        PyErr_Clear();
        module_name_str = PyModule_GetName(module);
        if (unlikely(!module_name_str)) { goto modbad; }
        module_name = PyUnicode_FromString(module_name_str);
        if (unlikely(!module_name)) { goto modbad; }
        module_dot = PyUnicode_Concat(module_name, __pyx_kp_u__3);
        if (unlikely(!module_dot)) { goto modbad; }
        full_name = PyUnicode_Concat(module_dot, name);
        if (unlikely(!full_name)) { goto modbad; }
        #if PY_VERSION_HEX < 0x030700A1 || (CYTHON_COMPILING_IN_PYPY && PYPY_VERSION_NUM  < 0x07030400)
        {
            PyObject *modules = PyImport_GetModuleDict();
            if (unlikely(!modules))
                goto modbad;
            value = PyObject_GetItem(modules, full_name);
        }
        #else
        value = PyImport_GetModule(full_name);
        #endif
      modbad:
        Py_XDECREF(full_name);
        Py_XDECREF(module_dot);
        Py_XDECREF(module_name);
    }
    if (unlikely(!value)) {
        PyErr_Format(PyExc_ImportError,
        #if PY_MAJOR_VERSION < 3
            "cannot import name %.230s", PyString_AS_STRING(name));
        #else
            "cannot import name %S", name);
        #endif
    }
    return value;
}

/* ImportDottedModule */
#if PY_MAJOR_VERSION >= 3
static PyObject *__Pyx__ImportDottedModule_Error(PyObject *name, PyObject *parts_tuple, Py_ssize_t count) {
    PyObject *partial_name = NULL, *slice = NULL, *sep = NULL;
    if (unlikely(PyErr_Occurred())) {
        PyErr_Clear();
    }
    if (likely(PyTuple_GET_SIZE(parts_tuple) == count)) {
        partial_name = name;
    } else {
        slice = PySequence_GetSlice(parts_tuple, 0, count);
        if (unlikely(!slice))
            goto bad;
        sep = PyUnicode_FromStringAndSize(".", 1);
        if (unlikely(!sep))
            goto bad;
        partial_name = PyUnicode_Join(sep, slice);
    }
    PyErr_Format(
#if PY_MAJOR_VERSION < 3
        PyExc_ImportError,
        "No module named '%s'", PyString_AS_STRING(partial_name));
#else
#if PY_VERSION_HEX >= 0x030600B1
        PyExc_ModuleNotFoundError,
#else
        PyExc_ImportError,
#endif
        "No module named '%U'", partial_name);
#endif
bad:
    Py_XDECREF(sep);
    Py_XDECREF(slice);
    Py_XDECREF(partial_name);
    return NULL;
}
#endif
#if PY_MAJOR_VERSION >= 3
static PyObject *__Pyx__ImportDottedModule_Lookup(PyObject *name) {
    PyObject *imported_module;
#if PY_VERSION_HEX < 0x030700A1 || (CYTHON_COMPILING_IN_PYPY && PYPY_VERSION_NUM  < 0x07030400)
    PyObject *modules = PyImport_GetModuleDict();
    if (unlikely(!modules))
        return NULL;
    imported_module = __Pyx_PyDict_GetItemStr(modules, name);
    Py_XINCREF(imported_module);
#else
    imported_module = PyImport_GetModule(name);
#endif
    return imported_module;
}
#endif
#if PY_MAJOR_VERSION >= 3
static PyObject *__Pyx_ImportDottedModule_WalkParts(PyObject *module, PyObject *name, PyObject *parts_tuple) {
    Py_ssize_t i, nparts;
    nparts = PyTuple_GET_SIZE(parts_tuple);
    for (i=1; i < nparts && module; i++) {
        PyObject *part, *submodule;
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        part = PyTuple_GET_ITEM(parts_tuple, i);
#else
        part = PySequence_ITEM(parts_tuple, i);
#endif
        submodule = __Pyx_PyObject_GetAttrStrNoError(module, part);
#if !(CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS)
        Py_DECREF(part);
#endif
        Py_DECREF(module);
        module = submodule;
    }
    if (unlikely(!module)) {
        return __Pyx__ImportDottedModule_Error(name, parts_tuple, i);
    }
    return module;
}
#endif
static PyObject *__Pyx__ImportDottedModule(PyObject *name, PyObject *parts_tuple) {
#if PY_MAJOR_VERSION < 3
    PyObject *module, *from_list, *star = __pyx_n_s__21;
    CYTHON_UNUSED_VAR(parts_tuple);
    from_list = PyList_New(1);
    if (unlikely(!from_list))
        return NULL;
    Py_INCREF(star);
    PyList_SET_ITEM(from_list, 0, star);
    module = __Pyx_Import(name, from_list, 0);
    Py_DECREF(from_list);
    return module;
#else
    PyObject *imported_module;
    PyObject *module = __Pyx_Import(name, NULL, 0);
    if (!parts_tuple || unlikely(!module))
        return module;
    imported_module = __Pyx__ImportDottedModule_Lookup(name);
    if (likely(imported_module)) {
        Py_DECREF(module);
        return imported_module;
    }
    PyErr_Clear();
    return __Pyx_ImportDottedModule_WalkParts(module, name, parts_tuple);
#endif
}
static PyObject *__Pyx_ImportDottedModule(PyObject *name, PyObject *parts_tuple) {
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030400B1
    PyObject *module = __Pyx__ImportDottedModule_Lookup(name);
    if (likely(module)) {
        PyObject *spec = __Pyx_PyObject_GetAttrStrNoError(module, __pyx_n_s_spec);
        if (likely(spec)) {
            PyObject *unsafe = __Pyx_PyObject_GetAttrStrNoError(spec, __pyx_n_s_initializing);
            if (likely(!unsafe || !__Pyx_PyObject_IsTrue(unsafe))) {
                Py_DECREF(spec);
                spec = NULL;
            }
            Py_XDECREF(unsafe);
        }
        if (likely(!spec)) {
            PyErr_Clear();
            return module;
        }
        Py_DECREF(spec);
        Py_DECREF(module);
    } else if (PyErr_Occurred()) {
        PyErr_Clear();
    }
#endif
    return __Pyx__ImportDottedModule(name, parts_tuple);
}

/* FetchSharedCythonModule */
static PyObject *__Pyx_FetchSharedCythonABIModule(void) {
    return __Pyx_PyImport_AddModuleRef((char*) __PYX_ABI_MODULE_NAME);
}

/* FetchCommonType */
static int __Pyx_VerifyCachedType(PyObject *cached_type,
                               const char *name,
                               Py_ssize_t basicsize,
                               Py_ssize_t expected_basicsize) {
    if (!PyType_Check(cached_type)) {
        PyErr_Format(PyExc_TypeError,
            "Shared Cython type %.200s is not a type object", name);
        return -1;
    }
    if (basicsize != expected_basicsize) {
        PyErr_Format(PyExc_TypeError,
            "Shared Cython type %.200s has the wrong size, try recompiling",
            name);
        return -1;
    }
    return 0;
}
#if !CYTHON_USE_TYPE_SPECS
static PyTypeObject* __Pyx_FetchCommonType(PyTypeObject* type) {
    PyObject* abi_module;
    const char* object_name;
    PyTypeObject *cached_type = NULL;
    abi_module = __Pyx_FetchSharedCythonABIModule();
    if (!abi_module) return NULL;
    object_name = strrchr(type->tp_name, '.');
    object_name = object_name ? object_name+1 : type->tp_name;
    cached_type = (PyTypeObject*) PyObject_GetAttrString(abi_module, object_name);
    if (cached_type) {
        if (__Pyx_VerifyCachedType(
              (PyObject *)cached_type,
              object_name,
              cached_type->tp_basicsize,
              type->tp_basicsize) < 0) {
            goto bad;
        }
        goto done;
    }
    if (!PyErr_ExceptionMatches(PyExc_AttributeError)) goto bad;
    PyErr_Clear();
    if (PyType_Ready(type) < 0) goto bad;
    if (PyObject_SetAttrString(abi_module, object_name, (PyObject *)type) < 0)
        goto bad;
    Py_INCREF(type);
    cached_type = type;
done:
    Py_DECREF(abi_module);
    return cached_type;
bad:
    Py_XDECREF(cached_type);
    cached_type = NULL;
    goto done;
}
#else
static PyTypeObject *__Pyx_FetchCommonTypeFromSpec(PyObject *module, PyType_Spec *spec, PyObject *bases) {
    PyObject *abi_module, *cached_type = NULL;
    const char* object_name = strrchr(spec->name, '.');
    object_name = object_name ? object_name+1 : spec->name;
    abi_module = __Pyx_FetchSharedCythonABIModule();
    if (!abi_module) return NULL;
    cached_type = PyObject_GetAttrString(abi_module, object_name);
    if (cached_type) {
        Py_ssize_t basicsize;
#if CYTHON_COMPILING_IN_LIMITED_API
        PyObject *py_basicsize;
        py_basicsize = PyObject_GetAttrString(cached_type, "__basicsize__");
        if (unlikely(!py_basicsize)) goto bad;
        basicsize = PyLong_AsSsize_t(py_basicsize);
        Py_DECREF(py_basicsize);
        py_basicsize = 0;
        if (unlikely(basicsize == (Py_ssize_t)-1) && PyErr_Occurred()) goto bad;
#else
        basicsize = likely(PyType_Check(cached_type)) ? ((PyTypeObject*) cached_type)->tp_basicsize : -1;
#endif
        if (__Pyx_VerifyCachedType(
              cached_type,
              object_name,
              basicsize,
              spec->basicsize) < 0) {
            goto bad;
        }
        goto done;
    }
    if (!PyErr_ExceptionMatches(PyExc_AttributeError)) goto bad;
    PyErr_Clear();
    CYTHON_UNUSED_VAR(module);
    cached_type = __Pyx_PyType_FromModuleAndSpec(abi_module, spec, bases);
    if (unlikely(!cached_type)) goto bad;
    if (unlikely(__Pyx_fix_up_extension_type_from_spec(spec, (PyTypeObject *) cached_type) < 0)) goto bad;
    if (PyObject_SetAttrString(abi_module, object_name, cached_type) < 0) goto bad;
done:
    Py_DECREF(abi_module);
    assert(cached_type == NULL || PyType_Check(cached_type));
    return (PyTypeObject *) cached_type;
bad:
    Py_XDECREF(cached_type);
    cached_type = NULL;
    goto done;
}
#endif

/* PyVectorcallFastCallDict */
#if CYTHON_METH_FASTCALL
static PyObject *__Pyx_PyVectorcall_FastCallDict_kw(PyObject *func, __pyx_vectorcallfunc vc, PyObject *const *args, size_t nargs, PyObject *kw)
{
    PyObject *res = NULL;
    PyObject *kwnames;
    PyObject **newargs;
    PyObject **kwvalues;
    Py_ssize_t i, pos;
    size_t j;
    PyObject *key, *value;
    unsigned long keys_are_strings;
    Py_ssize_t nkw = PyDict_GET_SIZE(kw);
    newargs = (PyObject **)PyMem_Malloc((nargs + (size_t)nkw) * sizeof(args[0]));
    if (unlikely(newargs == NULL)) {
        PyErr_NoMemory();
        return NULL;
    }
    for (j = 0; j < nargs; j++) newargs[j] = args[j];
    kwnames = PyTuple_New(nkw);
    if (unlikely(kwnames == NULL)) {
        PyMem_Free(newargs);
        return NULL;
    }
    kwvalues = newargs + nargs;
    pos = i = 0;
    keys_are_strings = Py_TPFLAGS_UNICODE_SUBCLASS;
    while (PyDict_Next(kw, &pos, &key, &value)) {
        keys_are_strings &= Py_TYPE(key)->tp_flags;
        Py_INCREF(key);
        Py_INCREF(value);
        PyTuple_SET_ITEM(kwnames, i, key);
        kwvalues[i] = value;
        i++;
    }
    if (unlikely(!keys_are_strings)) {
        PyErr_SetString(PyExc_TypeError, "keywords must be strings");
        goto cleanup;
    }
    res = vc(func, newargs, nargs, kwnames);
cleanup:
    Py_DECREF(kwnames);
    for (i = 0; i < nkw; i++)
        Py_DECREF(kwvalues[i]);
    PyMem_Free(newargs);
    return res;
}
static CYTHON_INLINE PyObject *__Pyx_PyVectorcall_FastCallDict(PyObject *func, __pyx_vectorcallfunc vc, PyObject *const *args, size_t nargs, PyObject *kw)
{
    if (likely(kw == NULL) || PyDict_GET_SIZE(kw) == 0) {
        return vc(func, args, nargs, NULL);
    }
    return __Pyx_PyVectorcall_FastCallDict_kw(func, vc, args, nargs, kw);
}
#endif

/* CythonFunctionShared */
#if CYTHON_COMPILING_IN_LIMITED_API
static CYTHON_INLINE int __Pyx__IsSameCyOrCFunction(PyObject *func, void *cfunc) {
    if (__Pyx_CyFunction_Check(func)) {
        return PyCFunction_GetFunction(((__pyx_CyFunctionObject*)func)->func) == (PyCFunction) cfunc;
    } else if (PyCFunction_Check(func)) {
        return PyCFunction_GetFunction(func) == (PyCFunction) cfunc;
    }
    return 0;
}
#else
static CYTHON_INLINE int __Pyx__IsSameCyOrCFunction(PyObject *func, void *cfunc) {
    return __Pyx_CyOrPyCFunction_Check(func) && __Pyx_CyOrPyCFunction_GET_FUNCTION(func) == (PyCFunction) cfunc;
}
#endif
static CYTHON_INLINE void __Pyx__CyFunction_SetClassObj(__pyx_CyFunctionObject* f, PyObject* classobj) {
#if PY_VERSION_HEX < 0x030900B1 || CYTHON_COMPILING_IN_LIMITED_API
    __Pyx_Py_XDECREF_SET(
        __Pyx_CyFunction_GetClassObj(f),
            ((classobj) ? __Pyx_NewRef(classobj) : NULL));
#else
    __Pyx_Py_XDECREF_SET(
        ((PyCMethodObject *) (f))->mm_class,
        (PyTypeObject*)((classobj) ? __Pyx_NewRef(classobj) : NULL));
#endif
}
static PyObject *
__Pyx_CyFunction_get_doc(__pyx_CyFunctionObject *op, void *closure)
{
    CYTHON_UNUSED_VAR(closure);
    if (unlikely(op->func_doc == NULL)) {
#if CYTHON_COMPILING_IN_LIMITED_API
        op->func_doc = PyObject_GetAttrString(op->func, "__doc__");
        if (unlikely(!op->func_doc)) return NULL;
#else
        if (((PyCFunctionObject*)op)->m_ml->ml_doc) {
#if PY_MAJOR_VERSION >= 3
            op->func_doc = PyUnicode_FromString(((PyCFunctionObject*)op)->m_ml->ml_doc);
#else
            op->func_doc = PyString_FromString(((PyCFunctionObject*)op)->m_ml->ml_doc);
#endif
            if (unlikely(op->func_doc == NULL))
                return NULL;
        } else {
            Py_INCREF(Py_None);
            return Py_None;
        }
#endif
    }
    Py_INCREF(op->func_doc);
    return op->func_doc;
}
static int
__Pyx_CyFunction_set_doc(__pyx_CyFunctionObject *op, PyObject *value, void *context)
{
    CYTHON_UNUSED_VAR(context);
    if (value == NULL) {
        value = Py_None;
    }
    Py_INCREF(value);
    __Pyx_Py_XDECREF_SET(op->func_doc, value);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_name(__pyx_CyFunctionObject *op, void *context)
{
    CYTHON_UNUSED_VAR(context);
    if (unlikely(op->func_name == NULL)) {
#if CYTHON_COMPILING_IN_LIMITED_API
        op->func_name = PyObject_GetAttrString(op->func, "__name__");
#elif PY_MAJOR_VERSION >= 3
        op->func_name = PyUnicode_InternFromString(((PyCFunctionObject*)op)->m_ml->ml_name);
#else
        op->func_name = PyString_InternFromString(((PyCFunctionObject*)op)->m_ml->ml_name);
#endif
        if (unlikely(op->func_name == NULL))
            return NULL;
    }
    Py_INCREF(op->func_name);
    return op->func_name;
}
static int
__Pyx_CyFunction_set_name(__pyx_CyFunctionObject *op, PyObject *value, void *context)
{
    CYTHON_UNUSED_VAR(context);
#if PY_MAJOR_VERSION >= 3
    if (unlikely(value == NULL || !PyUnicode_Check(value)))
#else
    if (unlikely(value == NULL || !PyString_Check(value)))
#endif
    {
        PyErr_SetString(PyExc_TypeError,
                        "__name__ must be set to a string object");
        return -1;
    }
    Py_INCREF(value);
    __Pyx_Py_XDECREF_SET(op->func_name, value);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_qualname(__pyx_CyFunctionObject *op, void *context)
{
    CYTHON_UNUSED_VAR(context);
    Py_INCREF(op->func_qualname);
    return op->func_qualname;
}
static int
__Pyx_CyFunction_set_qualname(__pyx_CyFunctionObject *op, PyObject *value, void *context)
{
    CYTHON_UNUSED_VAR(context);
#if PY_MAJOR_VERSION >= 3
    if (unlikely(value == NULL || !PyUnicode_Check(value)))
#else
    if (unlikely(value == NULL || !PyString_Check(value)))
#endif
    {
        PyErr_SetString(PyExc_TypeError,
                        "__qualname__ must be set to a string object");
        return -1;
    }
    Py_INCREF(value);
    __Pyx_Py_XDECREF_SET(op->func_qualname, value);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_dict(__pyx_CyFunctionObject *op, void *context)
{
    CYTHON_UNUSED_VAR(context);
    if (unlikely(op->func_dict == NULL)) {
        op->func_dict = PyDict_New();
        if (unlikely(op->func_dict == NULL))
            return NULL;
    }
    Py_INCREF(op->func_dict);
    return op->func_dict;
}
static int
__Pyx_CyFunction_set_dict(__pyx_CyFunctionObject *op, PyObject *value, void *context)
{
    CYTHON_UNUSED_VAR(context);
    if (unlikely(value == NULL)) {
        PyErr_SetString(PyExc_TypeError,
               "function's dictionary may not be deleted");
        return -1;
    }
    if (unlikely(!PyDict_Check(value))) {
        PyErr_SetString(PyExc_TypeError,
               "setting function's dictionary to a non-dict");
        return -1;
    }
    Py_INCREF(value);
    __Pyx_Py_XDECREF_SET(op->func_dict, value);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_globals(__pyx_CyFunctionObject *op, void *context)
{
    CYTHON_UNUSED_VAR(context);
    Py_INCREF(op->func_globals);
    return op->func_globals;
}
static PyObject *
__Pyx_CyFunction_get_closure(__pyx_CyFunctionObject *op, void *context)
{
    CYTHON_UNUSED_VAR(op);
    CYTHON_UNUSED_VAR(context);
    Py_INCREF(Py_None);
    return Py_None;
}
static PyObject *
__Pyx_CyFunction_get_code(__pyx_CyFunctionObject *op, void *context)
{
    PyObject* result = (op->func_code) ? op->func_code : Py_None;
    CYTHON_UNUSED_VAR(context);
    Py_INCREF(result);
    return result;
}
static int
__Pyx_CyFunction_init_defaults(__pyx_CyFunctionObject *op) {
    int result = 0;
    PyObject *res = op->defaults_getter((PyObject *) op);
    if (unlikely(!res))
        return -1;
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    op->defaults_tuple = PyTuple_GET_ITEM(res, 0);
    Py_INCREF(op->defaults_tuple);
    op->defaults_kwdict = PyTuple_GET_ITEM(res, 1);
    Py_INCREF(op->defaults_kwdict);
    #else
    op->defaults_tuple = __Pyx_PySequence_ITEM(res, 0);
    if (unlikely(!op->defaults_tuple)) result = -1;
    else {
        op->defaults_kwdict = __Pyx_PySequence_ITEM(res, 1);
        if (unlikely(!op->defaults_kwdict)) result = -1;
    }
    #endif
    Py_DECREF(res);
    return result;
}
static int
__Pyx_CyFunction_set_defaults(__pyx_CyFunctionObject *op, PyObject* value, void *context) {
    CYTHON_UNUSED_VAR(context);
    if (!value) {
        value = Py_None;
    } else if (unlikely(value != Py_None && !PyTuple_Check(value))) {
        PyErr_SetString(PyExc_TypeError,
                        "__defaults__ must be set to a tuple object");
        return -1;
    }
    PyErr_WarnEx(PyExc_RuntimeWarning, "changes to cyfunction.__defaults__ will not "
                 "currently affect the values used in function calls", 1);
    Py_INCREF(value);
    __Pyx_Py_XDECREF_SET(op->defaults_tuple, value);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_defaults(__pyx_CyFunctionObject *op, void *context) {
    PyObject* result = op->defaults_tuple;
    CYTHON_UNUSED_VAR(context);
    if (unlikely(!result)) {
        if (op->defaults_getter) {
            if (unlikely(__Pyx_CyFunction_init_defaults(op) < 0)) return NULL;
            result = op->defaults_tuple;
        } else {
            result = Py_None;
        }
    }
    Py_INCREF(result);
    return result;
}
static int
__Pyx_CyFunction_set_kwdefaults(__pyx_CyFunctionObject *op, PyObject* value, void *context) {
    CYTHON_UNUSED_VAR(context);
    if (!value) {
        value = Py_None;
    } else if (unlikely(value != Py_None && !PyDict_Check(value))) {
        PyErr_SetString(PyExc_TypeError,
                        "__kwdefaults__ must be set to a dict object");
        return -1;
    }
    PyErr_WarnEx(PyExc_RuntimeWarning, "changes to cyfunction.__kwdefaults__ will not "
                 "currently affect the values used in function calls", 1);
    Py_INCREF(value);
    __Pyx_Py_XDECREF_SET(op->defaults_kwdict, value);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_kwdefaults(__pyx_CyFunctionObject *op, void *context) {
    PyObject* result = op->defaults_kwdict;
    CYTHON_UNUSED_VAR(context);
    if (unlikely(!result)) {
        if (op->defaults_getter) {
            if (unlikely(__Pyx_CyFunction_init_defaults(op) < 0)) return NULL;
            result = op->defaults_kwdict;
        } else {
            result = Py_None;
        }
    }
    Py_INCREF(result);
    return result;
}
static int
__Pyx_CyFunction_set_annotations(__pyx_CyFunctionObject *op, PyObject* value, void *context) {
    CYTHON_UNUSED_VAR(context);
    if (!value || value == Py_None) {
        value = NULL;
    } else if (unlikely(!PyDict_Check(value))) {
        PyErr_SetString(PyExc_TypeError,
                        "__annotations__ must be set to a dict object");
        return -1;
    }
    Py_XINCREF(value);
    __Pyx_Py_XDECREF_SET(op->func_annotations, value);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_annotations(__pyx_CyFunctionObject *op, void *context) {
    PyObject* result = op->func_annotations;
    CYTHON_UNUSED_VAR(context);
    if (unlikely(!result)) {
        result = PyDict_New();
        if (unlikely(!result)) return NULL;
        op->func_annotations = result;
    }
    Py_INCREF(result);
    return result;
}
static PyObject *
__Pyx_CyFunction_get_is_coroutine(__pyx_CyFunctionObject *op, void *context) {
    int is_coroutine;
    CYTHON_UNUSED_VAR(context);
    if (op->func_is_coroutine) {
        return __Pyx_NewRef(op->func_is_coroutine);
    }
    is_coroutine = op->flags & __Pyx_CYFUNCTION_COROUTINE;
#if PY_VERSION_HEX >= 0x03050000
    if (is_coroutine) {
        PyObject *module, *fromlist, *marker = __pyx_n_s_is_coroutine;
        fromlist = PyList_New(1);
        if (unlikely(!fromlist)) return NULL;
        Py_INCREF(marker);
#if CYTHON_ASSUME_SAFE_MACROS
        PyList_SET_ITEM(fromlist, 0, marker);
#else
        if (unlikely(PyList_SetItem(fromlist, 0, marker) < 0)) {
            Py_DECREF(marker);
            Py_DECREF(fromlist);
            return NULL;
        }
#endif
        module = PyImport_ImportModuleLevelObject(__pyx_n_s_asyncio_coroutines, NULL, NULL, fromlist, 0);
        Py_DECREF(fromlist);
        if (unlikely(!module)) goto ignore;
        op->func_is_coroutine = __Pyx_PyObject_GetAttrStr(module, marker);
        Py_DECREF(module);
        if (likely(op->func_is_coroutine)) {
            return __Pyx_NewRef(op->func_is_coroutine);
        }
ignore:
        PyErr_Clear();
    }
#endif
    op->func_is_coroutine = __Pyx_PyBool_FromLong(is_coroutine);
    return __Pyx_NewRef(op->func_is_coroutine);
}
#if CYTHON_COMPILING_IN_LIMITED_API
static PyObject *
__Pyx_CyFunction_get_module(__pyx_CyFunctionObject *op, void *context) {
    CYTHON_UNUSED_VAR(context);
    return PyObject_GetAttrString(op->func, "__module__");
}
static int
__Pyx_CyFunction_set_module(__pyx_CyFunctionObject *op, PyObject* value, void *context) {
    CYTHON_UNUSED_VAR(context);
    return PyObject_SetAttrString(op->func, "__module__", value);
}
#endif
static PyGetSetDef __pyx_CyFunction_getsets[] = {
    {(char *) "func_doc", (getter)__Pyx_CyFunction_get_doc, (setter)__Pyx_CyFunction_set_doc, 0, 0},
    {(char *) "__doc__",  (getter)__Pyx_CyFunction_get_doc, (setter)__Pyx_CyFunction_set_doc, 0, 0},
    {(char *) "func_name", (getter)__Pyx_CyFunction_get_name, (setter)__Pyx_CyFunction_set_name, 0, 0},
    {(char *) "__name__", (getter)__Pyx_CyFunction_get_name, (setter)__Pyx_CyFunction_set_name, 0, 0},
    {(char *) "__qualname__", (getter)__Pyx_CyFunction_get_qualname, (setter)__Pyx_CyFunction_set_qualname, 0, 0},
    {(char *) "func_dict", (getter)__Pyx_CyFunction_get_dict, (setter)__Pyx_CyFunction_set_dict, 0, 0},
    {(char *) "__dict__", (getter)__Pyx_CyFunction_get_dict, (setter)__Pyx_CyFunction_set_dict, 0, 0},
    {(char *) "func_globals", (getter)__Pyx_CyFunction_get_globals, 0, 0, 0},
    {(char *) "__globals__", (getter)__Pyx_CyFunction_get_globals, 0, 0, 0},
    {(char *) "func_closure", (getter)__Pyx_CyFunction_get_closure, 0, 0, 0},
    {(char *) "__closure__", (getter)__Pyx_CyFunction_get_closure, 0, 0, 0},
    {(char *) "func_code", (getter)__Pyx_CyFunction_get_code, 0, 0, 0},
    {(char *) "__code__", (getter)__Pyx_CyFunction_get_code, 0, 0, 0},
    {(char *) "func_defaults", (getter)__Pyx_CyFunction_get_defaults, (setter)__Pyx_CyFunction_set_defaults, 0, 0},
    {(char *) "__defaults__", (getter)__Pyx_CyFunction_get_defaults, (setter)__Pyx_CyFunction_set_defaults, 0, 0},
    {(char *) "__kwdefaults__", (getter)__Pyx_CyFunction_get_kwdefaults, (setter)__Pyx_CyFunction_set_kwdefaults, 0, 0},
    {(char *) "__annotations__", (getter)__Pyx_CyFunction_get_annotations, (setter)__Pyx_CyFunction_set_annotations, 0, 0},
    {(char *) "_is_coroutine", (getter)__Pyx_CyFunction_get_is_coroutine, 0, 0, 0},
#if CYTHON_COMPILING_IN_LIMITED_API
    {"__module__", (getter)__Pyx_CyFunction_get_module, (setter)__Pyx_CyFunction_set_module, 0, 0},
#endif
    {0, 0, 0, 0, 0}
};
static PyMemberDef __pyx_CyFunction_members[] = {
#if !CYTHON_COMPILING_IN_LIMITED_API
    {(char *) "__module__", T_OBJECT, offsetof(PyCFunctionObject, m_module), 0, 0},
#endif
#if CYTHON_USE_TYPE_SPECS
    {(char *) "__dictoffset__", T_PYSSIZET, offsetof(__pyx_CyFunctionObject, func_dict), READONLY, 0},
#if CYTHON_METH_FASTCALL
#if CYTHON_BACKPORT_VECTORCALL
    {(char *) "__vectorcalloffset__", T_PYSSIZET, offsetof(__pyx_CyFunctionObject, func_vectorcall), READONLY, 0},
#else
#if !CYTHON_COMPILING_IN_LIMITED_API
    {(char *) "__vectorcalloffset__", T_PYSSIZET, offsetof(PyCFunctionObject, vectorcall), READONLY, 0},
#endif
#endif
#endif
#if PY_VERSION_HEX < 0x030500A0 || CYTHON_COMPILING_IN_LIMITED_API
    {(char *) "__weaklistoffset__", T_PYSSIZET, offsetof(__pyx_CyFunctionObject, func_weakreflist), READONLY, 0},
#else
    {(char *) "__weaklistoffset__", T_PYSSIZET, offsetof(PyCFunctionObject, m_weakreflist), READONLY, 0},
#endif
#endif
    {0, 0, 0,  0, 0}
};
static PyObject *
__Pyx_CyFunction_reduce(__pyx_CyFunctionObject *m, PyObject *args)
{
    CYTHON_UNUSED_VAR(args);
#if PY_MAJOR_VERSION >= 3
    Py_INCREF(m->func_qualname);
    return m->func_qualname;
#else
    return PyString_FromString(((PyCFunctionObject*)m)->m_ml->ml_name);
#endif
}
static PyMethodDef __pyx_CyFunction_methods[] = {
    {"__reduce__", (PyCFunction)__Pyx_CyFunction_reduce, METH_VARARGS, 0},
    {0, 0, 0, 0}
};
#if PY_VERSION_HEX < 0x030500A0 || CYTHON_COMPILING_IN_LIMITED_API
#define __Pyx_CyFunction_weakreflist(cyfunc) ((cyfunc)->func_weakreflist)
#else
#define __Pyx_CyFunction_weakreflist(cyfunc) (((PyCFunctionObject*)cyfunc)->m_weakreflist)
#endif
static PyObject *__Pyx_CyFunction_Init(__pyx_CyFunctionObject *op, PyMethodDef *ml, int flags, PyObject* qualname,
                                       PyObject *closure, PyObject *module, PyObject* globals, PyObject* code) {
#if !CYTHON_COMPILING_IN_LIMITED_API
    PyCFunctionObject *cf = (PyCFunctionObject*) op;
#endif
    if (unlikely(op == NULL))
        return NULL;
#if CYTHON_COMPILING_IN_LIMITED_API
    op->func = PyCFunction_NewEx(ml, (PyObject*)op, module);
    if (unlikely(!op->func)) return NULL;
#endif
    op->flags = flags;
    __Pyx_CyFunction_weakreflist(op) = NULL;
#if !CYTHON_COMPILING_IN_LIMITED_API
    cf->m_ml = ml;
    cf->m_self = (PyObject *) op;
#endif
    Py_XINCREF(closure);
    op->func_closure = closure;
#if !CYTHON_COMPILING_IN_LIMITED_API
    Py_XINCREF(module);
    cf->m_module = module;
#endif
    op->func_dict = NULL;
    op->func_name = NULL;
    Py_INCREF(qualname);
    op->func_qualname = qualname;
    op->func_doc = NULL;
#if PY_VERSION_HEX < 0x030900B1 || CYTHON_COMPILING_IN_LIMITED_API
    op->func_classobj = NULL;
#else
    ((PyCMethodObject*)op)->mm_class = NULL;
#endif
    op->func_globals = globals;
    Py_INCREF(op->func_globals);
    Py_XINCREF(code);
    op->func_code = code;
    op->defaults_pyobjects = 0;
    op->defaults_size = 0;
    op->defaults = NULL;
    op->defaults_tuple = NULL;
    op->defaults_kwdict = NULL;
    op->defaults_getter = NULL;
    op->func_annotations = NULL;
    op->func_is_coroutine = NULL;
#if CYTHON_METH_FASTCALL
    switch (ml->ml_flags & (METH_VARARGS | METH_FASTCALL | METH_NOARGS | METH_O | METH_KEYWORDS | METH_METHOD)) {
    case METH_NOARGS:
        __Pyx_CyFunction_func_vectorcall(op) = __Pyx_CyFunction_Vectorcall_NOARGS;
        break;
    case METH_O:
        __Pyx_CyFunction_func_vectorcall(op) = __Pyx_CyFunction_Vectorcall_O;
        break;
    case METH_METHOD | METH_FASTCALL | METH_KEYWORDS:
        __Pyx_CyFunction_func_vectorcall(op) = __Pyx_CyFunction_Vectorcall_FASTCALL_KEYWORDS_METHOD;
        break;
    case METH_FASTCALL | METH_KEYWORDS:
        __Pyx_CyFunction_func_vectorcall(op) = __Pyx_CyFunction_Vectorcall_FASTCALL_KEYWORDS;
        break;
    case METH_VARARGS | METH_KEYWORDS:
        __Pyx_CyFunction_func_vectorcall(op) = NULL;
        break;
    default:
        PyErr_SetString(PyExc_SystemError, "Bad call flags for CyFunction");
        Py_DECREF(op);
        return NULL;
    }
#endif
    return (PyObject *) op;
}
static int
__Pyx_CyFunction_clear(__pyx_CyFunctionObject *m)
{
    Py_CLEAR(m->func_closure);
#if CYTHON_COMPILING_IN_LIMITED_API
    Py_CLEAR(m->func);
#else
    Py_CLEAR(((PyCFunctionObject*)m)->m_module);
#endif
    Py_CLEAR(m->func_dict);
    Py_CLEAR(m->func_name);
    Py_CLEAR(m->func_qualname);
    Py_CLEAR(m->func_doc);
    Py_CLEAR(m->func_globals);
    Py_CLEAR(m->func_code);
#if !CYTHON_COMPILING_IN_LIMITED_API
#if PY_VERSION_HEX < 0x030900B1
    Py_CLEAR(__Pyx_CyFunction_GetClassObj(m));
#else
    {
        PyObject *cls = (PyObject*) ((PyCMethodObject *) (m))->mm_class;
        ((PyCMethodObject *) (m))->mm_class = NULL;
        Py_XDECREF(cls);
    }
#endif
#endif
    Py_CLEAR(m->defaults_tuple);
    Py_CLEAR(m->defaults_kwdict);
    Py_CLEAR(m->func_annotations);
    Py_CLEAR(m->func_is_coroutine);
    if (m->defaults) {
        PyObject **pydefaults = __Pyx_CyFunction_Defaults(PyObject *, m);
        int i;
        for (i = 0; i < m->defaults_pyobjects; i++)
            Py_XDECREF(pydefaults[i]);
        PyObject_Free(m->defaults);
        m->defaults = NULL;
    }
    return 0;
}
static void __Pyx__CyFunction_dealloc(__pyx_CyFunctionObject *m)
{
    if (__Pyx_CyFunction_weakreflist(m) != NULL)
        PyObject_ClearWeakRefs((PyObject *) m);
    __Pyx_CyFunction_clear(m);
    __Pyx_PyHeapTypeObject_GC_Del(m);
}
static void __Pyx_CyFunction_dealloc(__pyx_CyFunctionObject *m)
{
    PyObject_GC_UnTrack(m);
    __Pyx__CyFunction_dealloc(m);
}
static int __Pyx_CyFunction_traverse(__pyx_CyFunctionObject *m, visitproc visit, void *arg)
{
    Py_VISIT(m->func_closure);
#if CYTHON_COMPILING_IN_LIMITED_API
    Py_VISIT(m->func);
#else
    Py_VISIT(((PyCFunctionObject*)m)->m_module);
#endif
    Py_VISIT(m->func_dict);
    Py_VISIT(m->func_name);
    Py_VISIT(m->func_qualname);
    Py_VISIT(m->func_doc);
    Py_VISIT(m->func_globals);
    Py_VISIT(m->func_code);
#if !CYTHON_COMPILING_IN_LIMITED_API
    Py_VISIT(__Pyx_CyFunction_GetClassObj(m));
#endif
    Py_VISIT(m->defaults_tuple);
    Py_VISIT(m->defaults_kwdict);
    Py_VISIT(m->func_is_coroutine);
    if (m->defaults) {
        PyObject **pydefaults = __Pyx_CyFunction_Defaults(PyObject *, m);
        int i;
        for (i = 0; i < m->defaults_pyobjects; i++)
            Py_VISIT(pydefaults[i]);
    }
    return 0;
}
static PyObject*
__Pyx_CyFunction_repr(__pyx_CyFunctionObject *op)
{
#if PY_MAJOR_VERSION >= 3
    return PyUnicode_FromFormat("<cyfunction %U at %p>",
                                op->func_qualname, (void *)op);
#else
    return PyString_FromFormat("<cyfunction %s at %p>",
                               PyString_AsString(op->func_qualname), (void *)op);
#endif
}
static PyObject * __Pyx_CyFunction_CallMethod(PyObject *func, PyObject *self, PyObject *arg, PyObject *kw) {
#if CYTHON_COMPILING_IN_LIMITED_API
    PyObject *f = ((__pyx_CyFunctionObject*)func)->func;
    PyObject *py_name = NULL;
    PyCFunction meth;
    int flags;
    meth = PyCFunction_GetFunction(f);
    if (unlikely(!meth)) return NULL;
    flags = PyCFunction_GetFlags(f);
    if (unlikely(flags < 0)) return NULL;
#else
    PyCFunctionObject* f = (PyCFunctionObject*)func;
    PyCFunction meth = f->m_ml->ml_meth;
    int flags = f->m_ml->ml_flags;
#endif
    Py_ssize_t size;
    switch (flags & (METH_VARARGS | METH_KEYWORDS | METH_NOARGS | METH_O)) {
    case METH_VARARGS:
        if (likely(kw == NULL || PyDict_Size(kw) == 0))
            return (*meth)(self, arg);
        break;
    case METH_VARARGS | METH_KEYWORDS:
        return (*(PyCFunctionWithKeywords)(void*)meth)(self, arg, kw);
    case METH_NOARGS:
        if (likely(kw == NULL || PyDict_Size(kw) == 0)) {
#if CYTHON_ASSUME_SAFE_MACROS
            size = PyTuple_GET_SIZE(arg);
#else
            size = PyTuple_Size(arg);
            if (unlikely(size < 0)) return NULL;
#endif
            if (likely(size == 0))
                return (*meth)(self, NULL);
#if CYTHON_COMPILING_IN_LIMITED_API
            py_name = __Pyx_CyFunction_get_name((__pyx_CyFunctionObject*)func, NULL);
            if (!py_name) return NULL;
            PyErr_Format(PyExc_TypeError,
                "%.200S() takes no arguments (%" CYTHON_FORMAT_SSIZE_T "d given)",
                py_name, size);
            Py_DECREF(py_name);
#else
            PyErr_Format(PyExc_TypeError,
                "%.200s() takes no arguments (%" CYTHON_FORMAT_SSIZE_T "d given)",
                f->m_ml->ml_name, size);
#endif
            return NULL;
        }
        break;
    case METH_O:
        if (likely(kw == NULL || PyDict_Size(kw) == 0)) {
#if CYTHON_ASSUME_SAFE_MACROS
            size = PyTuple_GET_SIZE(arg);
#else
            size = PyTuple_Size(arg);
            if (unlikely(size < 0)) return NULL;
#endif
            if (likely(size == 1)) {
                PyObject *result, *arg0;
                #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                arg0 = PyTuple_GET_ITEM(arg, 0);
                #else
                arg0 = __Pyx_PySequence_ITEM(arg, 0); if (unlikely(!arg0)) return NULL;
                #endif
                result = (*meth)(self, arg0);
                #if !(CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS)
                Py_DECREF(arg0);
                #endif
                return result;
            }
#if CYTHON_COMPILING_IN_LIMITED_API
            py_name = __Pyx_CyFunction_get_name((__pyx_CyFunctionObject*)func, NULL);
            if (!py_name) return NULL;
            PyErr_Format(PyExc_TypeError,
                "%.200S() takes exactly one argument (%" CYTHON_FORMAT_SSIZE_T "d given)",
                py_name, size);
            Py_DECREF(py_name);
#else
            PyErr_Format(PyExc_TypeError,
                "%.200s() takes exactly one argument (%" CYTHON_FORMAT_SSIZE_T "d given)",
                f->m_ml->ml_name, size);
#endif
            return NULL;
        }
        break;
    default:
        PyErr_SetString(PyExc_SystemError, "Bad call flags for CyFunction");
        return NULL;
    }
#if CYTHON_COMPILING_IN_LIMITED_API
    py_name = __Pyx_CyFunction_get_name((__pyx_CyFunctionObject*)func, NULL);
    if (!py_name) return NULL;
    PyErr_Format(PyExc_TypeError, "%.200S() takes no keyword arguments",
                 py_name);
    Py_DECREF(py_name);
#else
    PyErr_Format(PyExc_TypeError, "%.200s() takes no keyword arguments",
                 f->m_ml->ml_name);
#endif
    return NULL;
}
static CYTHON_INLINE PyObject *__Pyx_CyFunction_Call(PyObject *func, PyObject *arg, PyObject *kw) {
    PyObject *self, *result;
#if CYTHON_COMPILING_IN_LIMITED_API
    self = PyCFunction_GetSelf(((__pyx_CyFunctionObject*)func)->func);
    if (unlikely(!self) && PyErr_Occurred()) return NULL;
#else
    self = ((PyCFunctionObject*)func)->m_self;
#endif
    result = __Pyx_CyFunction_CallMethod(func, self, arg, kw);
    return result;
}
static PyObject *__Pyx_CyFunction_CallAsMethod(PyObject *func, PyObject *args, PyObject *kw) {
    PyObject *result;
    __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *) func;
#if CYTHON_METH_FASTCALL
     __pyx_vectorcallfunc vc = __Pyx_CyFunction_func_vectorcall(cyfunc);
    if (vc) {
#if CYTHON_ASSUME_SAFE_MACROS
        return __Pyx_PyVectorcall_FastCallDict(func, vc, &PyTuple_GET_ITEM(args, 0), (size_t)PyTuple_GET_SIZE(args), kw);
#else
        (void) &__Pyx_PyVectorcall_FastCallDict;
        return PyVectorcall_Call(func, args, kw);
#endif
    }
#endif
    if ((cyfunc->flags & __Pyx_CYFUNCTION_CCLASS) && !(cyfunc->flags & __Pyx_CYFUNCTION_STATICMETHOD)) {
        Py_ssize_t argc;
        PyObject *new_args;
        PyObject *self;
#if CYTHON_ASSUME_SAFE_MACROS
        argc = PyTuple_GET_SIZE(args);
#else
        argc = PyTuple_Size(args);
        if (unlikely(!argc) < 0) return NULL;
#endif
        new_args = PyTuple_GetSlice(args, 1, argc);
        if (unlikely(!new_args))
            return NULL;
        self = PyTuple_GetItem(args, 0);
        if (unlikely(!self)) {
            Py_DECREF(new_args);
#if PY_MAJOR_VERSION > 2
            PyErr_Format(PyExc_TypeError,
                         "unbound method %.200S() needs an argument",
                         cyfunc->func_qualname);
#else
            PyErr_SetString(PyExc_TypeError,
                            "unbound method needs an argument");
#endif
            return NULL;
        }
        result = __Pyx_CyFunction_CallMethod(func, self, new_args, kw);
        Py_DECREF(new_args);
    } else {
        result = __Pyx_CyFunction_Call(func, args, kw);
    }
    return result;
}
#if CYTHON_METH_FASTCALL
static CYTHON_INLINE int __Pyx_CyFunction_Vectorcall_CheckArgs(__pyx_CyFunctionObject *cyfunc, Py_ssize_t nargs, PyObject *kwnames)
{
    int ret = 0;
    if ((cyfunc->flags & __Pyx_CYFUNCTION_CCLASS) && !(cyfunc->flags & __Pyx_CYFUNCTION_STATICMETHOD)) {
        if (unlikely(nargs < 1)) {
            PyErr_Format(PyExc_TypeError, "%.200s() needs an argument",
                         ((PyCFunctionObject*)cyfunc)->m_ml->ml_name);
            return -1;
        }
        ret = 1;
    }
    if (unlikely(kwnames) && unlikely(PyTuple_GET_SIZE(kwnames))) {
        PyErr_Format(PyExc_TypeError,
                     "%.200s() takes no keyword arguments", ((PyCFunctionObject*)cyfunc)->m_ml->ml_name);
        return -1;
    }
    return ret;
}
static PyObject * __Pyx_CyFunction_Vectorcall_NOARGS(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames)
{
    __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *)func;
    PyMethodDef* def = ((PyCFunctionObject*)cyfunc)->m_ml;
#if CYTHON_BACKPORT_VECTORCALL
    Py_ssize_t nargs = (Py_ssize_t)nargsf;
#else
    Py_ssize_t nargs = PyVectorcall_NARGS(nargsf);
#endif
    PyObject *self;
    switch (__Pyx_CyFunction_Vectorcall_CheckArgs(cyfunc, nargs, kwnames)) {
    case 1:
        self = args[0];
        args += 1;
        nargs -= 1;
        break;
    case 0:
        self = ((PyCFunctionObject*)cyfunc)->m_self;
        break;
    default:
        return NULL;
    }
    if (unlikely(nargs != 0)) {
        PyErr_Format(PyExc_TypeError,
            "%.200s() takes no arguments (%" CYTHON_FORMAT_SSIZE_T "d given)",
            def->ml_name, nargs);
        return NULL;
    }
    return def->ml_meth(self, NULL);
}
static PyObject * __Pyx_CyFunction_Vectorcall_O(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames)
{
    __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *)func;
    PyMethodDef* def = ((PyCFunctionObject*)cyfunc)->m_ml;
#if CYTHON_BACKPORT_VECTORCALL
    Py_ssize_t nargs = (Py_ssize_t)nargsf;
#else
    Py_ssize_t nargs = PyVectorcall_NARGS(nargsf);
#endif
    PyObject *self;
    switch (__Pyx_CyFunction_Vectorcall_CheckArgs(cyfunc, nargs, kwnames)) {
    case 1:
        self = args[0];
        args += 1;
        nargs -= 1;
        break;
    case 0:
        self = ((PyCFunctionObject*)cyfunc)->m_self;
        break;
    default:
        return NULL;
    }
    if (unlikely(nargs != 1)) {
        PyErr_Format(PyExc_TypeError,
            "%.200s() takes exactly one argument (%" CYTHON_FORMAT_SSIZE_T "d given)",
            def->ml_name, nargs);
        return NULL;
    }
    return def->ml_meth(self, args[0]);
}
static PyObject * __Pyx_CyFunction_Vectorcall_FASTCALL_KEYWORDS(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames)
{
    __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *)func;
    PyMethodDef* def = ((PyCFunctionObject*)cyfunc)->m_ml;
#if CYTHON_BACKPORT_VECTORCALL
    Py_ssize_t nargs = (Py_ssize_t)nargsf;
#else
    Py_ssize_t nargs = PyVectorcall_NARGS(nargsf);
#endif
    PyObject *self;
    switch (__Pyx_CyFunction_Vectorcall_CheckArgs(cyfunc, nargs, NULL)) {
    case 1:
        self = args[0];
        args += 1;
        nargs -= 1;
        break;
    case 0:
        self = ((PyCFunctionObject*)cyfunc)->m_self;
        break;
    default:
        return NULL;
    }
    return ((__Pyx_PyCFunctionFastWithKeywords)(void(*)(void))def->ml_meth)(self, args, nargs, kwnames);
}
static PyObject * __Pyx_CyFunction_Vectorcall_FASTCALL_KEYWORDS_METHOD(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames)
{
    __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *)func;
    PyMethodDef* def = ((PyCFunctionObject*)cyfunc)->m_ml;
    PyTypeObject *cls = (PyTypeObject *) __Pyx_CyFunction_GetClassObj(cyfunc);
#if CYTHON_BACKPORT_VECTORCALL
    Py_ssize_t nargs = (Py_ssize_t)nargsf;
#else
    Py_ssize_t nargs = PyVectorcall_NARGS(nargsf);
#endif
    PyObject *self;
    switch (__Pyx_CyFunction_Vectorcall_CheckArgs(cyfunc, nargs, NULL)) {
    case 1:
        self = args[0];
        args += 1;
        nargs -= 1;
        break;
    case 0:
        self = ((PyCFunctionObject*)cyfunc)->m_self;
        break;
    default:
        return NULL;
    }
    return ((__Pyx_PyCMethod)(void(*)(void))def->ml_meth)(self, cls, args, (size_t)nargs, kwnames);
}
#endif
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_CyFunctionType_slots[] = {
    {Py_tp_dealloc, (void *)__Pyx_CyFunction_dealloc},
    {Py_tp_repr, (void *)__Pyx_CyFunction_repr},
    {Py_tp_call, (void *)__Pyx_CyFunction_CallAsMethod},
    {Py_tp_traverse, (void *)__Pyx_CyFunction_traverse},
    {Py_tp_clear, (void *)__Pyx_CyFunction_clear},
    {Py_tp_methods, (void *)__pyx_CyFunction_methods},
    {Py_tp_members, (void *)__pyx_CyFunction_members},
    {Py_tp_getset, (void *)__pyx_CyFunction_getsets},
    {Py_tp_descr_get, (void *)__Pyx_PyMethod_New},
    {0, 0},
};
static PyType_Spec __pyx_CyFunctionType_spec = {
    __PYX_TYPE_MODULE_PREFIX "cython_function_or_method",
    sizeof(__pyx_CyFunctionObject),
    0,
#ifdef Py_TPFLAGS_METHOD_DESCRIPTOR
    Py_TPFLAGS_METHOD_DESCRIPTOR |
#endif
#if (defined(_Py_TPFLAGS_HAVE_VECTORCALL) && CYTHON_METH_FASTCALL)
    _Py_TPFLAGS_HAVE_VECTORCALL |
#endif
    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC | Py_TPFLAGS_BASETYPE,
    __pyx_CyFunctionType_slots
};
#else
static PyTypeObject __pyx_CyFunctionType_type = {
    PyVarObject_HEAD_INIT(0, 0)
    __PYX_TYPE_MODULE_PREFIX "cython_function_or_method",
    sizeof(__pyx_CyFunctionObject),
    0,
    (destructor) __Pyx_CyFunction_dealloc,
#if !CYTHON_METH_FASTCALL
    0,
#elif CYTHON_BACKPORT_VECTORCALL
    (printfunc)offsetof(__pyx_CyFunctionObject, func_vectorcall),
#else
    offsetof(PyCFunctionObject, vectorcall),
#endif
    0,
    0,
#if PY_MAJOR_VERSION < 3
    0,
#else
    0,
#endif
    (reprfunc) __Pyx_CyFunction_repr,
    0,
    0,
    0,
    0,
    __Pyx_CyFunction_CallAsMethod,
    0,
    0,
    0,
    0,
#ifdef Py_TPFLAGS_METHOD_DESCRIPTOR
    Py_TPFLAGS_METHOD_DESCRIPTOR |
#endif
#if defined(_Py_TPFLAGS_HAVE_VECTORCALL) && CYTHON_METH_FASTCALL
    _Py_TPFLAGS_HAVE_VECTORCALL |
#endif
    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC | Py_TPFLAGS_BASETYPE,
    0,
    (traverseproc) __Pyx_CyFunction_traverse,
    (inquiry) __Pyx_CyFunction_clear,
    0,
#if PY_VERSION_HEX < 0x030500A0
    offsetof(__pyx_CyFunctionObject, func_weakreflist),
#else
    offsetof(PyCFunctionObject, m_weakreflist),
#endif
    0,
    0,
    __pyx_CyFunction_methods,
    __pyx_CyFunction_members,
    __pyx_CyFunction_getsets,
    0,
    0,
    __Pyx_PyMethod_New,
    0,
    offsetof(__pyx_CyFunctionObject, func_dict),
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
#if PY_VERSION_HEX >= 0x030400a1
    0,
#endif
#if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
    0,
#endif
#if __PYX_NEED_TP_PRINT_SLOT
    0,
#endif
#if PY_VERSION_HEX >= 0x030C0000
    0,
#endif
#if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
    0,
#endif
};
#endif
static int __pyx_CyFunction_init(PyObject *module) {
#if CYTHON_USE_TYPE_SPECS
    __pyx_CyFunctionType = __Pyx_FetchCommonTypeFromSpec(module, &__pyx_CyFunctionType_spec, NULL);
#else
    CYTHON_UNUSED_VAR(module);
    __pyx_CyFunctionType = __Pyx_FetchCommonType(&__pyx_CyFunctionType_type);
#endif
    if (unlikely(__pyx_CyFunctionType == NULL)) {
        return -1;
    }
    return 0;
}
static CYTHON_INLINE void *__Pyx_CyFunction_InitDefaults(PyObject *func, size_t size, int pyobjects) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults = PyObject_Malloc(size);
    if (unlikely(!m->defaults))
        return PyErr_NoMemory();
    memset(m->defaults, 0, size);
    m->defaults_pyobjects = pyobjects;
    m->defaults_size = size;
    return m->defaults;
}
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsTuple(PyObject *func, PyObject *tuple) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults_tuple = tuple;
    Py_INCREF(tuple);
}
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsKwDict(PyObject *func, PyObject *dict) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults_kwdict = dict;
    Py_INCREF(dict);
}
static CYTHON_INLINE void __Pyx_CyFunction_SetAnnotationsDict(PyObject *func, PyObject *dict) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->func_annotations = dict;
    Py_INCREF(dict);
}

/* CythonFunction */
static PyObject *__Pyx_CyFunction_New(PyMethodDef *ml, int flags, PyObject* qualname,
                                      PyObject *closure, PyObject *module, PyObject* globals, PyObject* code) {
    PyObject *op = __Pyx_CyFunction_Init(
        PyObject_GC_New(__pyx_CyFunctionObject, __pyx_CyFunctionType),
        ml, flags, qualname, closure, module, globals, code
    );
    if (likely(op)) {
        PyObject_GC_Track(op);
    }
    return op;
}

/* CLineInTraceback */
#ifndef CYTHON_CLINE_IN_TRACEBACK
static int __Pyx_CLineForTraceback(PyThreadState *tstate, int c_line) {
    PyObject *use_cline;
    PyObject *ptype, *pvalue, *ptraceback;
#if CYTHON_COMPILING_IN_CPYTHON
    PyObject **cython_runtime_dict;
#endif
    CYTHON_MAYBE_UNUSED_VAR(tstate);
    if (unlikely(!__pyx_cython_runtime)) {
        return c_line;
    }
    __Pyx_ErrFetchInState(tstate, &ptype, &pvalue, &ptraceback);
#if CYTHON_COMPILING_IN_CPYTHON
    cython_runtime_dict = _PyObject_GetDictPtr(__pyx_cython_runtime);
    if (likely(cython_runtime_dict)) {
        __PYX_PY_DICT_LOOKUP_IF_MODIFIED(
            use_cline, *cython_runtime_dict,
            __Pyx_PyDict_GetItemStr(*cython_runtime_dict, __pyx_n_s_cline_in_traceback))
    } else
#endif
    {
      PyObject *use_cline_obj = __Pyx_PyObject_GetAttrStrNoError(__pyx_cython_runtime, __pyx_n_s_cline_in_traceback);
      if (use_cline_obj) {
        use_cline = PyObject_Not(use_cline_obj) ? Py_False : Py_True;
        Py_DECREF(use_cline_obj);
      } else {
        PyErr_Clear();
        use_cline = NULL;
      }
    }
    if (!use_cline) {
        c_line = 0;
        (void) PyObject_SetAttr(__pyx_cython_runtime, __pyx_n_s_cline_in_traceback, Py_False);
    }
    else if (use_cline == Py_False || (use_cline != Py_True && PyObject_Not(use_cline) != 0)) {
        c_line = 0;
    }
    __Pyx_ErrRestoreInState(tstate, ptype, pvalue, ptraceback);
    return c_line;
}
#endif

/* CodeObjectCache */
#if !CYTHON_COMPILING_IN_LIMITED_API
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line) {
    int start = 0, mid = 0, end = count - 1;
    if (end >= 0 && code_line > entries[end].code_line) {
        return count;
    }
    while (start < end) {
        mid = start + (end - start) / 2;
        if (code_line < entries[mid].code_line) {
            end = mid;
        } else if (code_line > entries[mid].code_line) {
             start = mid + 1;
        } else {
            return mid;
        }
    }
    if (code_line <= entries[mid].code_line) {
        return mid;
    } else {
        return mid + 1;
    }
}
static PyCodeObject *__pyx_find_code_object(int code_line) {
    PyCodeObject* code_object;
    int pos;
    if (unlikely(!code_line) || unlikely(!__pyx_code_cache.entries)) {
        return NULL;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if (unlikely(pos >= __pyx_code_cache.count) || unlikely(__pyx_code_cache.entries[pos].code_line != code_line)) {
        return NULL;
    }
    code_object = __pyx_code_cache.entries[pos].code_object;
    Py_INCREF(code_object);
    return code_object;
}
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object) {
    int pos, i;
    __Pyx_CodeObjectCacheEntry* entries = __pyx_code_cache.entries;
    if (unlikely(!code_line)) {
        return;
    }
    if (unlikely(!entries)) {
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Malloc(64*sizeof(__Pyx_CodeObjectCacheEntry));
        if (likely(entries)) {
            __pyx_code_cache.entries = entries;
            __pyx_code_cache.max_count = 64;
            __pyx_code_cache.count = 1;
            entries[0].code_line = code_line;
            entries[0].code_object = code_object;
            Py_INCREF(code_object);
        }
        return;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if ((pos < __pyx_code_cache.count) && unlikely(__pyx_code_cache.entries[pos].code_line == code_line)) {
        PyCodeObject* tmp = entries[pos].code_object;
        entries[pos].code_object = code_object;
        Py_DECREF(tmp);
        return;
    }
    if (__pyx_code_cache.count == __pyx_code_cache.max_count) {
        int new_max = __pyx_code_cache.max_count + 64;
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Realloc(
            __pyx_code_cache.entries, ((size_t)new_max) * sizeof(__Pyx_CodeObjectCacheEntry));
        if (unlikely(!entries)) {
            return;
        }
        __pyx_code_cache.entries = entries;
        __pyx_code_cache.max_count = new_max;
    }
    for (i=__pyx_code_cache.count; i>pos; i--) {
        entries[i] = entries[i-1];
    }
    entries[pos].code_line = code_line;
    entries[pos].code_object = code_object;
    __pyx_code_cache.count++;
    Py_INCREF(code_object);
}
#endif

/* AddTraceback */
#include "compile.h"
#include "frameobject.h"
#include "traceback.h"
#if PY_VERSION_HEX >= 0x030b00a6 && !CYTHON_COMPILING_IN_LIMITED_API
  #ifndef Py_BUILD_CORE
    #define Py_BUILD_CORE 1
  #endif
  #include "internal/pycore_frame.h"
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
static PyObject *__Pyx_PyCode_Replace_For_AddTraceback(PyObject *code, PyObject *scratch_dict,
                                                       PyObject *firstlineno, PyObject *name) {
    PyObject *replace = NULL;
    if (unlikely(PyDict_SetItemString(scratch_dict, "co_firstlineno", firstlineno))) return NULL;
    if (unlikely(PyDict_SetItemString(scratch_dict, "co_name", name))) return NULL;
    replace = PyObject_GetAttrString(code, "replace");
    if (likely(replace)) {
        PyObject *result;
        result = PyObject_Call(replace, __pyx_empty_tuple, scratch_dict);
        Py_DECREF(replace);
        return result;
    }
    PyErr_Clear();
    #if __PYX_LIMITED_VERSION_HEX < 0x030780000
    {
        PyObject *compiled = NULL, *result = NULL;
        if (unlikely(PyDict_SetItemString(scratch_dict, "code", code))) return NULL;
        if (unlikely(PyDict_SetItemString(scratch_dict, "type", (PyObject*)(&PyType_Type)))) return NULL;
        compiled = Py_CompileString(
            "out = type(code)(\n"
            "  code.co_argcount, code.co_kwonlyargcount, code.co_nlocals, code.co_stacksize,\n"
            "  code.co_flags, code.co_code, code.co_consts, code.co_names,\n"
            "  code.co_varnames, code.co_filename, co_name, co_firstlineno,\n"
            "  code.co_lnotab)\n", "<dummy>", Py_file_input);
        if (!compiled) return NULL;
        result = PyEval_EvalCode(compiled, scratch_dict, scratch_dict);
        Py_DECREF(compiled);
        if (!result) PyErr_Print();
        Py_DECREF(result);
        result = PyDict_GetItemString(scratch_dict, "out");
        if (result) Py_INCREF(result);
        return result;
    }
    #else
    return NULL;
    #endif
}
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename) {
    PyObject *code_object = NULL, *py_py_line = NULL, *py_funcname = NULL, *dict = NULL;
    PyObject *replace = NULL, *getframe = NULL, *frame = NULL;
    PyObject *exc_type, *exc_value, *exc_traceback;
    int success = 0;
    if (c_line) {
        (void) __pyx_cfilenm;
        (void) __Pyx_CLineForTraceback(__Pyx_PyThreadState_Current, c_line);
    }
    PyErr_Fetch(&exc_type, &exc_value, &exc_traceback);
    code_object = Py_CompileString("_getframe()", filename, Py_eval_input);
    if (unlikely(!code_object)) goto bad;
    py_py_line = PyLong_FromLong(py_line);
    if (unlikely(!py_py_line)) goto bad;
    py_funcname = PyUnicode_FromString(funcname);
    if (unlikely(!py_funcname)) goto bad;
    dict = PyDict_New();
    if (unlikely(!dict)) goto bad;
    {
        PyObject *old_code_object = code_object;
        code_object = __Pyx_PyCode_Replace_For_AddTraceback(code_object, dict, py_py_line, py_funcname);
        Py_DECREF(old_code_object);
    }
    if (unlikely(!code_object)) goto bad;
    getframe = PySys_GetObject("_getframe");
    if (unlikely(!getframe)) goto bad;
    if (unlikely(PyDict_SetItemString(dict, "_getframe", getframe))) goto bad;
    frame = PyEval_EvalCode(code_object, dict, dict);
    if (unlikely(!frame) || frame == Py_None) goto bad;
    success = 1;
  bad:
    PyErr_Restore(exc_type, exc_value, exc_traceback);
    Py_XDECREF(code_object);
    Py_XDECREF(py_py_line);
    Py_XDECREF(py_funcname);
    Py_XDECREF(dict);
    Py_XDECREF(replace);
    if (success) {
        PyTraceBack_Here(
            (struct _frame*)frame);
    }
    Py_XDECREF(frame);
}
#else
static PyCodeObject* __Pyx_CreateCodeObjectForTraceback(
            const char *funcname, int c_line,
            int py_line, const char *filename) {
    PyCodeObject *py_code = NULL;
    PyObject *py_funcname = NULL;
    #if PY_MAJOR_VERSION < 3
    PyObject *py_srcfile = NULL;
    py_srcfile = PyString_FromString(filename);
    if (!py_srcfile) goto bad;
    #endif
    if (c_line) {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        if (!py_funcname) goto bad;
        #else
        py_funcname = PyUnicode_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        if (!py_funcname) goto bad;
        funcname = PyUnicode_AsUTF8(py_funcname);
        if (!funcname) goto bad;
        #endif
    }
    else {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromString(funcname);
        if (!py_funcname) goto bad;
        #endif
    }
    #if PY_MAJOR_VERSION < 3
    py_code = __Pyx_PyCode_New(
        0,
        0,
        0,
        0,
        0,
        0,
        __pyx_empty_bytes, /*PyObject *code,*/
        __pyx_empty_tuple, /*PyObject *consts,*/
        __pyx_empty_tuple, /*PyObject *names,*/
        __pyx_empty_tuple, /*PyObject *varnames,*/
        __pyx_empty_tuple, /*PyObject *freevars,*/
        __pyx_empty_tuple, /*PyObject *cellvars,*/
        py_srcfile,   /*PyObject *filename,*/
        py_funcname,  /*PyObject *name,*/
        py_line,
        __pyx_empty_bytes  /*PyObject *lnotab*/
    );
    Py_DECREF(py_srcfile);
    #else
    py_code = PyCode_NewEmpty(filename, funcname, py_line);
    #endif
    Py_XDECREF(py_funcname);
    return py_code;
bad:
    Py_XDECREF(py_funcname);
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(py_srcfile);
    #endif
    return NULL;
}
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyFrameObject *py_frame = 0;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    PyObject *ptype, *pvalue, *ptraceback;
    if (c_line) {
        c_line = __Pyx_CLineForTraceback(tstate, c_line);
    }
    py_code = __pyx_find_code_object(c_line ? -c_line : py_line);
    if (!py_code) {
        __Pyx_ErrFetchInState(tstate, &ptype, &pvalue, &ptraceback);
        py_code = __Pyx_CreateCodeObjectForTraceback(
            funcname, c_line, py_line, filename);
        if (!py_code) {
            /* If the code object creation fails, then we should clear the
               fetched exception references and propagate the new exception */
            Py_XDECREF(ptype);
            Py_XDECREF(pvalue);
            Py_XDECREF(ptraceback);
            goto bad;
        }
        __Pyx_ErrRestoreInState(tstate, ptype, pvalue, ptraceback);
        __pyx_insert_code_object(c_line ? -c_line : py_line, py_code);
    }
    py_frame = PyFrame_New(
        tstate,            /*PyThreadState *tstate,*/
        py_code,           /*PyCodeObject *code,*/
        __pyx_d,    /*PyObject *globals,*/
        0                  /*PyObject *locals*/
    );
    if (!py_frame) goto bad;
    __Pyx_PyFrame_SetLineNumber(py_frame, py_line);
    PyTraceBack_Here(py_frame);
bad:
    Py_XDECREF(py_code);
    Py_XDECREF(py_frame);
}
#endif

/* CIntFromPyVerify */
#define __PYX_VERIFY_RETURN_INT(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 0)
#define __PYX_VERIFY_RETURN_INT_EXC(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 1)
#define __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, exc)\
    {\
        func_type value = func_value;\
        if (sizeof(target_type) < sizeof(func_type)) {\
            if (unlikely(value != (func_type) (target_type) value)) {\
                func_type zero = 0;\
                if (exc && unlikely(value == (func_type)-1 && PyErr_Occurred()))\
                    return (target_type) -1;\
                if (is_unsigned && unlikely(value < zero))\
                    goto raise_neg_overflow;\
                else\
                    goto raise_overflow;\
            }\
        }\
        return (target_type) value;\
    }

/* CIntFromPy */
static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const long neg_one = (long) -1, const_zero = (long) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if ((sizeof(long) < sizeof(long))) {
            __PYX_VERIFY_RETURN_INT(long, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (long) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            if (unlikely(__Pyx_PyLong_IsNeg(x))) {
                goto raise_neg_overflow;
            } else if (__Pyx_PyLong_IsCompact(x)) {
                __PYX_VERIFY_RETURN_INT(long, __Pyx_compact_upylong, __Pyx_PyLong_CompactValueUnsigned(x))
            } else {
                const digit* digits = __Pyx_PyLong_Digits(x);
                assert(__Pyx_PyLong_DigitCount(x) > 1);
                switch (__Pyx_PyLong_DigitCount(x)) {
                    case 2:
                        if ((8 * sizeof(long) > 1 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(long) >= 2 * PyLong_SHIFT)) {
                                return (long) (((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                            }
                        }
                        break;
                    case 3:
                        if ((8 * sizeof(long) > 2 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(long) >= 3 * PyLong_SHIFT)) {
                                return (long) (((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                            }
                        }
                        break;
                    case 4:
                        if ((8 * sizeof(long) > 3 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(long) >= 4 * PyLong_SHIFT)) {
                                return (long) (((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                            }
                        }
                        break;
                }
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A7
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (long) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if ((sizeof(long) <= sizeof(unsigned long))) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if ((sizeof(long) <= sizeof(unsigned PY_LONG_LONG))) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            if (__Pyx_PyLong_IsCompact(x)) {
                __PYX_VERIFY_RETURN_INT(long, __Pyx_compact_pylong, __Pyx_PyLong_CompactValue(x))
            } else {
                const digit* digits = __Pyx_PyLong_Digits(x);
                assert(__Pyx_PyLong_DigitCount(x) > 1);
                switch (__Pyx_PyLong_SignedDigitCount(x)) {
                    case -2:
                        if ((8 * sizeof(long) - 1 > 1 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(long) - 1 > 2 * PyLong_SHIFT)) {
                                return (long) (((long)-1)*(((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                            }
                        }
                        break;
                    case 2:
                        if ((8 * sizeof(long) > 1 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(long) - 1 > 2 * PyLong_SHIFT)) {
                                return (long) ((((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                            }
                        }
                        break;
                    case -3:
                        if ((8 * sizeof(long) - 1 > 2 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(long) - 1 > 3 * PyLong_SHIFT)) {
                                return (long) (((long)-1)*(((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                            }
                        }
                        break;
                    case 3:
                        if ((8 * sizeof(long) > 2 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(long) - 1 > 3 * PyLong_SHIFT)) {
                                return (long) ((((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                            }
                        }
                        break;
                    case -4:
                        if ((8 * sizeof(long) - 1 > 3 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(long) - 1 > 4 * PyLong_SHIFT)) {
                                return (long) (((long)-1)*(((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                            }
                        }
                        break;
                    case 4:
                        if ((8 * sizeof(long) > 3 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(long) - 1 > 4 * PyLong_SHIFT)) {
                                return (long) ((((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                            }
                        }
                        break;
                }
            }
#endif
            if ((sizeof(long) <= sizeof(long))) {
                __PYX_VERIFY_RETURN_INT_EXC(long, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if ((sizeof(long) <= sizeof(PY_LONG_LONG))) {
                __PYX_VERIFY_RETURN_INT_EXC(long, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
            long val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
#if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
#endif
            if (likely(v)) {
                int ret = -1;
#if PY_VERSION_HEX < 0x030d0000 && !(CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API) || defined(_PyLong_AsByteArray)
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                ret = _PyLong_AsByteArray((PyLongObject *)v,
                                           bytes, sizeof(val),
                                           is_little, !is_unsigned);
#else
                PyObject *stepval = NULL, *mask = NULL, *shift = NULL;
                int bits, remaining_bits, is_negative = 0;
                long idigit;
                int chunk_size = (sizeof(long) < 8) ? 30 : 62;
                if (unlikely(!PyLong_CheckExact(v))) {
                    PyObject *tmp = v;
                    v = PyNumber_Long(v);
                    assert(PyLong_CheckExact(v));
                    Py_DECREF(tmp);
                    if (unlikely(!v)) return (long) -1;
                }
#if CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030B0000
                if (Py_SIZE(x) == 0)
                    return (long) 0;
                is_negative = Py_SIZE(x) < 0;
#else
                {
                    int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                    if (unlikely(result < 0))
                        return (long) -1;
                    is_negative = result == 1;
                }
#endif
                if (is_unsigned && unlikely(is_negative)) {
                    goto raise_neg_overflow;
                } else if (is_negative) {
                    stepval = PyNumber_Invert(v);
                    if (unlikely(!stepval))
                        return (long) -1;
                } else {
                    stepval = __Pyx_NewRef(v);
                }
                val = (long) 0;
                mask = PyLong_FromLong((1L << chunk_size) - 1); if (unlikely(!mask)) goto done;
                shift = PyLong_FromLong(chunk_size); if (unlikely(!shift)) goto done;
                for (bits = 0; bits < (int) sizeof(long) * 8 - chunk_size; bits += chunk_size) {
                    PyObject *tmp, *digit;
                    digit = PyNumber_And(stepval, mask);
                    if (unlikely(!digit)) goto done;
                    idigit = PyLong_AsLong(digit);
                    Py_DECREF(digit);
                    if (unlikely(idigit < 0)) goto done;
                    tmp = PyNumber_Rshift(stepval, shift);
                    if (unlikely(!tmp)) goto done;
                    Py_DECREF(stepval); stepval = tmp;
                    val |= ((long) idigit) << bits;
                    #if CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030B0000
                    if (Py_SIZE(stepval) == 0)
                        goto unpacking_done;
                    #endif
                }
                idigit = PyLong_AsLong(stepval);
                if (unlikely(idigit < 0)) goto done;
                remaining_bits = ((int) sizeof(long) * 8) - bits - (is_unsigned ? 0 : 1);
                if (unlikely(idigit >= (1L << remaining_bits)))
                    goto raise_overflow;
                val |= ((long) idigit) << bits;
            #if CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030B0000
            unpacking_done:
            #endif
                if (!is_unsigned) {
                    if (unlikely(val & (((long) 1) << (sizeof(long) * 8 - 1))))
                        goto raise_overflow;
                    if (is_negative)
                        val = ~val;
                }
                ret = 0;
            done:
                Py_XDECREF(shift);
                Py_XDECREF(mask);
                Py_XDECREF(stepval);
#endif
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
            return (long) -1;
        }
    } else {
        long val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (long) -1;
        val = __Pyx_PyInt_As_long(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to long");
    return (long) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to long");
    return (long) -1;
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int64_t(int64_t value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const int64_t neg_one = (int64_t) -1, const_zero = (int64_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(int64_t) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(int64_t) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int64_t) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(int64_t) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int64_t) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        return _PyLong_FromByteArray(bytes, sizeof(int64_t),
                                     little, !is_unsigned);
#else
        PyObject *from_bytes, *result = NULL;
        PyObject *py_bytes = NULL, *arg_tuple = NULL, *kwds = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(int64_t));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        arg_tuple = PyTuple_Pack(2, py_bytes, order_str);
        if (!arg_tuple) goto limited_bad;
        if (!is_unsigned) {
            kwds = PyDict_New();
            if (!kwds) goto limited_bad;
            if (PyDict_SetItemString(kwds, "signed", __Pyx_NewRef(Py_True))) goto limited_bad;
        }
        result = PyObject_Call(from_bytes, arg_tuple, kwds);
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(arg_tuple);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_uint64_t(uint64_t value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const uint64_t neg_one = (uint64_t) -1, const_zero = (uint64_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(uint64_t) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(uint64_t) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(uint64_t) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(uint64_t) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(uint64_t) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        return _PyLong_FromByteArray(bytes, sizeof(uint64_t),
                                     little, !is_unsigned);
#else
        PyObject *from_bytes, *result = NULL;
        PyObject *py_bytes = NULL, *arg_tuple = NULL, *kwds = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(uint64_t));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        arg_tuple = PyTuple_Pack(2, py_bytes, order_str);
        if (!arg_tuple) goto limited_bad;
        if (!is_unsigned) {
            kwds = PyDict_New();
            if (!kwds) goto limited_bad;
            if (PyDict_SetItemString(kwds, "signed", __Pyx_NewRef(Py_True))) goto limited_bad;
        }
        result = PyObject_Call(from_bytes, arg_tuple, kwds);
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(arg_tuple);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const long neg_one = (long) -1, const_zero = (long) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(long) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(long) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(long) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        return _PyLong_FromByteArray(bytes, sizeof(long),
                                     little, !is_unsigned);
#else
        PyObject *from_bytes, *result = NULL;
        PyObject *py_bytes = NULL, *arg_tuple = NULL, *kwds = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(long));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        arg_tuple = PyTuple_Pack(2, py_bytes, order_str);
        if (!arg_tuple) goto limited_bad;
        if (!is_unsigned) {
            kwds = PyDict_New();
            if (!kwds) goto limited_bad;
            if (PyDict_SetItemString(kwds, "signed", __Pyx_NewRef(Py_True))) goto limited_bad;
        }
        result = PyObject_Call(from_bytes, arg_tuple, kwds);
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(arg_tuple);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int(int value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const int neg_one = (int) -1, const_zero = (int) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(int) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(int) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(int) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        return _PyLong_FromByteArray(bytes, sizeof(int),
                                     little, !is_unsigned);
#else
        PyObject *from_bytes, *result = NULL;
        PyObject *py_bytes = NULL, *arg_tuple = NULL, *kwds = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(int));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        arg_tuple = PyTuple_Pack(2, py_bytes, order_str);
        if (!arg_tuple) goto limited_bad;
        if (!is_unsigned) {
            kwds = PyDict_New();
            if (!kwds) goto limited_bad;
            if (PyDict_SetItemString(kwds, "signed", __Pyx_NewRef(Py_True))) goto limited_bad;
        }
        result = PyObject_Call(from_bytes, arg_tuple, kwds);
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(arg_tuple);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* FormatTypeName */
#if CYTHON_COMPILING_IN_LIMITED_API
static __Pyx_TypeName
__Pyx_PyType_GetName(PyTypeObject* tp)
{
    PyObject *name = __Pyx_PyObject_GetAttrStr((PyObject *)tp,
                                               __pyx_n_s_name_2);
    if (unlikely(name == NULL) || unlikely(!PyUnicode_Check(name))) {
        PyErr_Clear();
        Py_XDECREF(name);
        name = __Pyx_NewRef(__pyx_n_s__45);
    }
    return name;
}
#endif

/* CIntFromPy */
static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const int neg_one = (int) -1, const_zero = (int) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if ((sizeof(int) < sizeof(long))) {
            __PYX_VERIFY_RETURN_INT(int, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (int) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            if (unlikely(__Pyx_PyLong_IsNeg(x))) {
                goto raise_neg_overflow;
            } else if (__Pyx_PyLong_IsCompact(x)) {
                __PYX_VERIFY_RETURN_INT(int, __Pyx_compact_upylong, __Pyx_PyLong_CompactValueUnsigned(x))
            } else {
                const digit* digits = __Pyx_PyLong_Digits(x);
                assert(__Pyx_PyLong_DigitCount(x) > 1);
                switch (__Pyx_PyLong_DigitCount(x)) {
                    case 2:
                        if ((8 * sizeof(int) > 1 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(int) >= 2 * PyLong_SHIFT)) {
                                return (int) (((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                            }
                        }
                        break;
                    case 3:
                        if ((8 * sizeof(int) > 2 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(int) >= 3 * PyLong_SHIFT)) {
                                return (int) (((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                            }
                        }
                        break;
                    case 4:
                        if ((8 * sizeof(int) > 3 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(int) >= 4 * PyLong_SHIFT)) {
                                return (int) (((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                            }
                        }
                        break;
                }
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A7
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (int) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if ((sizeof(int) <= sizeof(unsigned long))) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if ((sizeof(int) <= sizeof(unsigned PY_LONG_LONG))) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            if (__Pyx_PyLong_IsCompact(x)) {
                __PYX_VERIFY_RETURN_INT(int, __Pyx_compact_pylong, __Pyx_PyLong_CompactValue(x))
            } else {
                const digit* digits = __Pyx_PyLong_Digits(x);
                assert(__Pyx_PyLong_DigitCount(x) > 1);
                switch (__Pyx_PyLong_SignedDigitCount(x)) {
                    case -2:
                        if ((8 * sizeof(int) - 1 > 1 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(int) - 1 > 2 * PyLong_SHIFT)) {
                                return (int) (((int)-1)*(((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                            }
                        }
                        break;
                    case 2:
                        if ((8 * sizeof(int) > 1 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(int) - 1 > 2 * PyLong_SHIFT)) {
                                return (int) ((((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                            }
                        }
                        break;
                    case -3:
                        if ((8 * sizeof(int) - 1 > 2 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(int) - 1 > 3 * PyLong_SHIFT)) {
                                return (int) (((int)-1)*(((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                            }
                        }
                        break;
                    case 3:
                        if ((8 * sizeof(int) > 2 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(int) - 1 > 3 * PyLong_SHIFT)) {
                                return (int) ((((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                            }
                        }
                        break;
                    case -4:
                        if ((8 * sizeof(int) - 1 > 3 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(int) - 1 > 4 * PyLong_SHIFT)) {
                                return (int) (((int)-1)*(((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                            }
                        }
                        break;
                    case 4:
                        if ((8 * sizeof(int) > 3 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(int) - 1 > 4 * PyLong_SHIFT)) {
                                return (int) ((((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                            }
                        }
                        break;
                }
            }
#endif
            if ((sizeof(int) <= sizeof(long))) {
                __PYX_VERIFY_RETURN_INT_EXC(int, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if ((sizeof(int) <= sizeof(PY_LONG_LONG))) {
                __PYX_VERIFY_RETURN_INT_EXC(int, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
            int val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
#if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
#endif
            if (likely(v)) {
                int ret = -1;
#if PY_VERSION_HEX < 0x030d0000 && !(CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API) || defined(_PyLong_AsByteArray)
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                ret = _PyLong_AsByteArray((PyLongObject *)v,
                                           bytes, sizeof(val),
                                           is_little, !is_unsigned);
#else
                PyObject *stepval = NULL, *mask = NULL, *shift = NULL;
                int bits, remaining_bits, is_negative = 0;
                long idigit;
                int chunk_size = (sizeof(long) < 8) ? 30 : 62;
                if (unlikely(!PyLong_CheckExact(v))) {
                    PyObject *tmp = v;
                    v = PyNumber_Long(v);
                    assert(PyLong_CheckExact(v));
                    Py_DECREF(tmp);
                    if (unlikely(!v)) return (int) -1;
                }
#if CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030B0000
                if (Py_SIZE(x) == 0)
                    return (int) 0;
                is_negative = Py_SIZE(x) < 0;
#else
                {
                    int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                    if (unlikely(result < 0))
                        return (int) -1;
                    is_negative = result == 1;
                }
#endif
                if (is_unsigned && unlikely(is_negative)) {
                    goto raise_neg_overflow;
                } else if (is_negative) {
                    stepval = PyNumber_Invert(v);
                    if (unlikely(!stepval))
                        return (int) -1;
                } else {
                    stepval = __Pyx_NewRef(v);
                }
                val = (int) 0;
                mask = PyLong_FromLong((1L << chunk_size) - 1); if (unlikely(!mask)) goto done;
                shift = PyLong_FromLong(chunk_size); if (unlikely(!shift)) goto done;
                for (bits = 0; bits < (int) sizeof(int) * 8 - chunk_size; bits += chunk_size) {
                    PyObject *tmp, *digit;
                    digit = PyNumber_And(stepval, mask);
                    if (unlikely(!digit)) goto done;
                    idigit = PyLong_AsLong(digit);
                    Py_DECREF(digit);
                    if (unlikely(idigit < 0)) goto done;
                    tmp = PyNumber_Rshift(stepval, shift);
                    if (unlikely(!tmp)) goto done;
                    Py_DECREF(stepval); stepval = tmp;
                    val |= ((int) idigit) << bits;
                    #if CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030B0000
                    if (Py_SIZE(stepval) == 0)
                        goto unpacking_done;
                    #endif
                }
                idigit = PyLong_AsLong(stepval);
                if (unlikely(idigit < 0)) goto done;
                remaining_bits = ((int) sizeof(int) * 8) - bits - (is_unsigned ? 0 : 1);
                if (unlikely(idigit >= (1L << remaining_bits)))
                    goto raise_overflow;
                val |= ((int) idigit) << bits;
            #if CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030B0000
            unpacking_done:
            #endif
                if (!is_unsigned) {
                    if (unlikely(val & (((int) 1) << (sizeof(int) * 8 - 1))))
                        goto raise_overflow;
                    if (is_negative)
                        val = ~val;
                }
                ret = 0;
            done:
                Py_XDECREF(shift);
                Py_XDECREF(mask);
                Py_XDECREF(stepval);
#endif
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
            return (int) -1;
        }
    } else {
        int val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (int) -1;
        val = __Pyx_PyInt_As_int(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to int");
    return (int) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to int");
    return (int) -1;
}

/* CoroutineBase */
#include <frameobject.h>
#if PY_VERSION_HEX >= 0x030b00a6
  #ifndef Py_BUILD_CORE
    #define Py_BUILD_CORE 1
  #endif
  #include "internal/pycore_frame.h"
#endif
#define __Pyx_Coroutine_Undelegate(gen) Py_CLEAR((gen)->yieldfrom)
static int __Pyx_PyGen__FetchStopIterationValue(PyThreadState *__pyx_tstate, PyObject **pvalue) {
    PyObject *et, *ev, *tb;
    PyObject *value = NULL;
    CYTHON_UNUSED_VAR(__pyx_tstate);
    __Pyx_ErrFetch(&et, &ev, &tb);
    if (!et) {
        Py_XDECREF(tb);
        Py_XDECREF(ev);
        Py_INCREF(Py_None);
        *pvalue = Py_None;
        return 0;
    }
    if (likely(et == PyExc_StopIteration)) {
        if (!ev) {
            Py_INCREF(Py_None);
            value = Py_None;
        }
#if PY_VERSION_HEX >= 0x030300A0
        else if (likely(__Pyx_IS_TYPE(ev, (PyTypeObject*)PyExc_StopIteration))) {
            value = ((PyStopIterationObject *)ev)->value;
            Py_INCREF(value);
            Py_DECREF(ev);
        }
#endif
        else if (unlikely(PyTuple_Check(ev))) {
            if (PyTuple_GET_SIZE(ev) >= 1) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                value = PyTuple_GET_ITEM(ev, 0);
                Py_INCREF(value);
#else
                value = PySequence_ITEM(ev, 0);
#endif
            } else {
                Py_INCREF(Py_None);
                value = Py_None;
            }
            Py_DECREF(ev);
        }
        else if (!__Pyx_TypeCheck(ev, (PyTypeObject*)PyExc_StopIteration)) {
            value = ev;
        }
        if (likely(value)) {
            Py_XDECREF(tb);
            Py_DECREF(et);
            *pvalue = value;
            return 0;
        }
    } else if (!__Pyx_PyErr_GivenExceptionMatches(et, PyExc_StopIteration)) {
        __Pyx_ErrRestore(et, ev, tb);
        return -1;
    }
    PyErr_NormalizeException(&et, &ev, &tb);
    if (unlikely(!PyObject_TypeCheck(ev, (PyTypeObject*)PyExc_StopIteration))) {
        __Pyx_ErrRestore(et, ev, tb);
        return -1;
    }
    Py_XDECREF(tb);
    Py_DECREF(et);
#if PY_VERSION_HEX >= 0x030300A0
    value = ((PyStopIterationObject *)ev)->value;
    Py_INCREF(value);
    Py_DECREF(ev);
#else
    {
        PyObject* args = __Pyx_PyObject_GetAttrStr(ev, __pyx_n_s_args);
        Py_DECREF(ev);
        if (likely(args)) {
            value = PySequence_GetItem(args, 0);
            Py_DECREF(args);
        }
        if (unlikely(!value)) {
            __Pyx_ErrRestore(NULL, NULL, NULL);
            Py_INCREF(Py_None);
            value = Py_None;
        }
    }
#endif
    *pvalue = value;
    return 0;
}
static CYTHON_INLINE
void __Pyx_Coroutine_ExceptionClear(__Pyx_ExcInfoStruct *exc_state) {
#if PY_VERSION_HEX >= 0x030B00a4
    Py_CLEAR(exc_state->exc_value);
#else
    PyObject *t, *v, *tb;
    t = exc_state->exc_type;
    v = exc_state->exc_value;
    tb = exc_state->exc_traceback;
    exc_state->exc_type = NULL;
    exc_state->exc_value = NULL;
    exc_state->exc_traceback = NULL;
    Py_XDECREF(t);
    Py_XDECREF(v);
    Py_XDECREF(tb);
#endif
}
#define __Pyx_Coroutine_AlreadyRunningError(gen)  (__Pyx__Coroutine_AlreadyRunningError(gen), (PyObject*)NULL)
static void __Pyx__Coroutine_AlreadyRunningError(__pyx_CoroutineObject *gen) {
    const char *msg;
    CYTHON_MAYBE_UNUSED_VAR(gen);
    if ((0)) {
    #ifdef __Pyx_Coroutine_USED
    } else if (__Pyx_Coroutine_Check((PyObject*)gen)) {
        msg = "coroutine already executing";
    #endif
    #ifdef __Pyx_AsyncGen_USED
    } else if (__Pyx_AsyncGen_CheckExact((PyObject*)gen)) {
        msg = "async generator already executing";
    #endif
    } else {
        msg = "generator already executing";
    }
    PyErr_SetString(PyExc_ValueError, msg);
}
#define __Pyx_Coroutine_NotStartedError(gen)  (__Pyx__Coroutine_NotStartedError(gen), (PyObject*)NULL)
static void __Pyx__Coroutine_NotStartedError(PyObject *gen) {
    const char *msg;
    CYTHON_MAYBE_UNUSED_VAR(gen);
    if ((0)) {
    #ifdef __Pyx_Coroutine_USED
    } else if (__Pyx_Coroutine_Check(gen)) {
        msg = "can't send non-None value to a just-started coroutine";
    #endif
    #ifdef __Pyx_AsyncGen_USED
    } else if (__Pyx_AsyncGen_CheckExact(gen)) {
        msg = "can't send non-None value to a just-started async generator";
    #endif
    } else {
        msg = "can't send non-None value to a just-started generator";
    }
    PyErr_SetString(PyExc_TypeError, msg);
}
#define __Pyx_Coroutine_AlreadyTerminatedError(gen, value, closing)  (__Pyx__Coroutine_AlreadyTerminatedError(gen, value, closing), (PyObject*)NULL)
static void __Pyx__Coroutine_AlreadyTerminatedError(PyObject *gen, PyObject *value, int closing) {
    CYTHON_MAYBE_UNUSED_VAR(gen);
    CYTHON_MAYBE_UNUSED_VAR(closing);
    #ifdef __Pyx_Coroutine_USED
    if (!closing && __Pyx_Coroutine_Check(gen)) {
        PyErr_SetString(PyExc_RuntimeError, "cannot reuse already awaited coroutine");
    } else
    #endif
    if (value) {
        #ifdef __Pyx_AsyncGen_USED
        if (__Pyx_AsyncGen_CheckExact(gen))
            PyErr_SetNone(__Pyx_PyExc_StopAsyncIteration);
        else
        #endif
        PyErr_SetNone(PyExc_StopIteration);
    }
}
static
PyObject *__Pyx_Coroutine_SendEx(__pyx_CoroutineObject *self, PyObject *value, int closing) {
    __Pyx_PyThreadState_declare
    PyThreadState *tstate;
    __Pyx_ExcInfoStruct *exc_state;
    PyObject *retval;
    assert(!self->is_running);
    if (unlikely(self->resume_label == 0)) {
        if (unlikely(value && value != Py_None)) {
            return __Pyx_Coroutine_NotStartedError((PyObject*)self);
        }
    }
    if (unlikely(self->resume_label == -1)) {
        return __Pyx_Coroutine_AlreadyTerminatedError((PyObject*)self, value, closing);
    }
#if CYTHON_FAST_THREAD_STATE
    __Pyx_PyThreadState_assign
    tstate = __pyx_tstate;
#else
    tstate = __Pyx_PyThreadState_Current;
#endif
    exc_state = &self->gi_exc_state;
    if (exc_state->exc_value) {
        #if CYTHON_COMPILING_IN_PYPY
        #else
        PyObject *exc_tb;
        #if PY_VERSION_HEX >= 0x030B00a4 && !CYTHON_COMPILING_IN_CPYTHON
        exc_tb = PyException_GetTraceback(exc_state->exc_value);
        #elif PY_VERSION_HEX >= 0x030B00a4
        exc_tb = ((PyBaseExceptionObject*) exc_state->exc_value)->traceback;
        #else
        exc_tb = exc_state->exc_traceback;
        #endif
        if (exc_tb) {
            PyTracebackObject *tb = (PyTracebackObject *) exc_tb;
            PyFrameObject *f = tb->tb_frame;
            assert(f->f_back == NULL);
            #if PY_VERSION_HEX >= 0x030B00A1
            f->f_back = PyThreadState_GetFrame(tstate);
            #else
            Py_XINCREF(tstate->frame);
            f->f_back = tstate->frame;
            #endif
            #if PY_VERSION_HEX >= 0x030B00a4 && !CYTHON_COMPILING_IN_CPYTHON
            Py_DECREF(exc_tb);
            #endif
        }
        #endif
    }
#if CYTHON_USE_EXC_INFO_STACK
    exc_state->previous_item = tstate->exc_info;
    tstate->exc_info = exc_state;
#else
    if (exc_state->exc_type) {
        __Pyx_ExceptionSwap(&exc_state->exc_type, &exc_state->exc_value, &exc_state->exc_traceback);
    } else {
        __Pyx_Coroutine_ExceptionClear(exc_state);
        __Pyx_ExceptionSave(&exc_state->exc_type, &exc_state->exc_value, &exc_state->exc_traceback);
    }
#endif
    self->is_running = 1;
    retval = self->body(self, tstate, value);
    self->is_running = 0;
#if CYTHON_USE_EXC_INFO_STACK
    exc_state = &self->gi_exc_state;
    tstate->exc_info = exc_state->previous_item;
    exc_state->previous_item = NULL;
    __Pyx_Coroutine_ResetFrameBackpointer(exc_state);
#endif
    return retval;
}
static CYTHON_INLINE void __Pyx_Coroutine_ResetFrameBackpointer(__Pyx_ExcInfoStruct *exc_state) {
#if CYTHON_COMPILING_IN_PYPY
    CYTHON_UNUSED_VAR(exc_state);
#else
    PyObject *exc_tb;
    #if PY_VERSION_HEX >= 0x030B00a4
    if (!exc_state->exc_value) return;
    exc_tb = PyException_GetTraceback(exc_state->exc_value);
    #else
    exc_tb = exc_state->exc_traceback;
    #endif
    if (likely(exc_tb)) {
        PyTracebackObject *tb = (PyTracebackObject *) exc_tb;
        PyFrameObject *f = tb->tb_frame;
        Py_CLEAR(f->f_back);
        #if PY_VERSION_HEX >= 0x030B00a4
        Py_DECREF(exc_tb);
        #endif
    }
#endif
}
static CYTHON_INLINE
PyObject *__Pyx_Coroutine_MethodReturn(PyObject* gen, PyObject *retval) {
    CYTHON_MAYBE_UNUSED_VAR(gen);
    if (unlikely(!retval)) {
        __Pyx_PyThreadState_declare
        __Pyx_PyThreadState_assign
        if (!__Pyx_PyErr_Occurred()) {
            PyObject *exc = PyExc_StopIteration;
            #ifdef __Pyx_AsyncGen_USED
            if (__Pyx_AsyncGen_CheckExact(gen))
                exc = __Pyx_PyExc_StopAsyncIteration;
            #endif
            __Pyx_PyErr_SetNone(exc);
        }
    }
    return retval;
}
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03030000 && (defined(__linux__) || PY_VERSION_HEX >= 0x030600B3)
static CYTHON_INLINE
PyObject *__Pyx_PyGen_Send(PyGenObject *gen, PyObject *arg) {
#if PY_VERSION_HEX <= 0x030A00A1
    return _PyGen_Send(gen, arg);
#else
    PyObject *result;
    if (PyIter_Send((PyObject*)gen, arg ? arg : Py_None, &result) == PYGEN_RETURN) {
        if (PyAsyncGen_CheckExact(gen)) {
            assert(result == Py_None);
            PyErr_SetNone(PyExc_StopAsyncIteration);
        }
        else if (result == Py_None) {
            PyErr_SetNone(PyExc_StopIteration);
        }
        else {
#if PY_VERSION_HEX < 0x030d00A1
            _PyGen_SetStopIterationValue(result);
#else
            if (!PyTuple_Check(result) && !PyExceptionInstance_Check(result)) {
                PyErr_SetObject(PyExc_StopIteration, result);
            } else {
                PyObject *exc = __Pyx_PyObject_CallOneArg(PyExc_StopIteration, result);
                if (likely(exc != NULL)) {
                    PyErr_SetObject(PyExc_StopIteration, exc);
                    Py_DECREF(exc);
                }
            }
#endif
        }
        Py_DECREF(result);
        result = NULL;
    }
    return result;
#endif
}
#endif
static CYTHON_INLINE
PyObject *__Pyx_Coroutine_FinishDelegation(__pyx_CoroutineObject *gen) {
    PyObject *ret;
    PyObject *val = NULL;
    __Pyx_Coroutine_Undelegate(gen);
    __Pyx_PyGen__FetchStopIterationValue(__Pyx_PyThreadState_Current, &val);
    ret = __Pyx_Coroutine_SendEx(gen, val, 0);
    Py_XDECREF(val);
    return ret;
}
static PyObject *__Pyx_Coroutine_Send(PyObject *self, PyObject *value) {
    PyObject *retval;
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject*) self;
    PyObject *yf = gen->yieldfrom;
    if (unlikely(gen->is_running))
        return __Pyx_Coroutine_AlreadyRunningError(gen);
    if (yf) {
        PyObject *ret;
        gen->is_running = 1;
        #ifdef __Pyx_Generator_USED
        if (__Pyx_Generator_CheckExact(yf)) {
            ret = __Pyx_Coroutine_Send(yf, value);
        } else
        #endif
        #ifdef __Pyx_Coroutine_USED
        if (__Pyx_Coroutine_Check(yf)) {
            ret = __Pyx_Coroutine_Send(yf, value);
        } else
        #endif
        #ifdef __Pyx_AsyncGen_USED
        if (__pyx_PyAsyncGenASend_CheckExact(yf)) {
            ret = __Pyx_async_gen_asend_send(yf, value);
        } else
        #endif
        #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03030000 && (defined(__linux__) || PY_VERSION_HEX >= 0x030600B3)
        if (PyGen_CheckExact(yf)) {
            ret = __Pyx_PyGen_Send((PyGenObject*)yf, value == Py_None ? NULL : value);
        } else
        #endif
        #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03050000 && defined(PyCoro_CheckExact) && (defined(__linux__) || PY_VERSION_HEX >= 0x030600B3)
        if (PyCoro_CheckExact(yf)) {
            ret = __Pyx_PyGen_Send((PyGenObject*)yf, value == Py_None ? NULL : value);
        } else
        #endif
        {
            if (value == Py_None)
                ret = __Pyx_PyObject_GetIterNextFunc(yf)(yf);
            else
                ret = __Pyx_PyObject_CallMethod1(yf, __pyx_n_s_send, value);
        }
        gen->is_running = 0;
        if (likely(ret)) {
            return ret;
        }
        retval = __Pyx_Coroutine_FinishDelegation(gen);
    } else {
        retval = __Pyx_Coroutine_SendEx(gen, value, 0);
    }
    return __Pyx_Coroutine_MethodReturn(self, retval);
}
static int __Pyx_Coroutine_CloseIter(__pyx_CoroutineObject *gen, PyObject *yf) {
    PyObject *retval = NULL;
    int err = 0;
    #ifdef __Pyx_Generator_USED
    if (__Pyx_Generator_CheckExact(yf)) {
        retval = __Pyx_Coroutine_Close(yf);
        if (!retval)
            return -1;
    } else
    #endif
    #ifdef __Pyx_Coroutine_USED
    if (__Pyx_Coroutine_Check(yf)) {
        retval = __Pyx_Coroutine_Close(yf);
        if (!retval)
            return -1;
    } else
    if (__Pyx_CoroutineAwait_CheckExact(yf)) {
        retval = __Pyx_CoroutineAwait_Close((__pyx_CoroutineAwaitObject*)yf, NULL);
        if (!retval)
            return -1;
    } else
    #endif
    #ifdef __Pyx_AsyncGen_USED
    if (__pyx_PyAsyncGenASend_CheckExact(yf)) {
        retval = __Pyx_async_gen_asend_close(yf, NULL);
    } else
    if (__pyx_PyAsyncGenAThrow_CheckExact(yf)) {
        retval = __Pyx_async_gen_athrow_close(yf, NULL);
    } else
    #endif
    {
        PyObject *meth;
        gen->is_running = 1;
        meth = __Pyx_PyObject_GetAttrStrNoError(yf, __pyx_n_s_close);
        if (unlikely(!meth)) {
            if (unlikely(PyErr_Occurred())) {
                PyErr_WriteUnraisable(yf);
            }
        } else {
            retval = __Pyx_PyObject_CallNoArg(meth);
            Py_DECREF(meth);
            if (unlikely(!retval))
                err = -1;
        }
        gen->is_running = 0;
    }
    Py_XDECREF(retval);
    return err;
}
static PyObject *__Pyx_Generator_Next(PyObject *self) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject*) self;
    PyObject *yf = gen->yieldfrom;
    if (unlikely(gen->is_running))
        return __Pyx_Coroutine_AlreadyRunningError(gen);
    if (yf) {
        PyObject *ret;
        gen->is_running = 1;
        #ifdef __Pyx_Generator_USED
        if (__Pyx_Generator_CheckExact(yf)) {
            ret = __Pyx_Generator_Next(yf);
        } else
        #endif
        #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03030000 && (defined(__linux__) || PY_VERSION_HEX >= 0x030600B3)
        if (PyGen_CheckExact(yf)) {
            ret = __Pyx_PyGen_Send((PyGenObject*)yf, NULL);
        } else
        #endif
        #ifdef __Pyx_Coroutine_USED
        if (__Pyx_Coroutine_Check(yf)) {
            ret = __Pyx_Coroutine_Send(yf, Py_None);
        } else
        #endif
            ret = __Pyx_PyObject_GetIterNextFunc(yf)(yf);
        gen->is_running = 0;
        if (likely(ret)) {
            return ret;
        }
        return __Pyx_Coroutine_FinishDelegation(gen);
    }
    return __Pyx_Coroutine_SendEx(gen, Py_None, 0);
}
static PyObject *__Pyx_Coroutine_Close_Method(PyObject *self, PyObject *arg) {
    CYTHON_UNUSED_VAR(arg);
    return __Pyx_Coroutine_Close(self);
}
static PyObject *__Pyx_Coroutine_Close(PyObject *self) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    PyObject *retval, *raised_exception;
    PyObject *yf = gen->yieldfrom;
    int err = 0;
    if (unlikely(gen->is_running))
        return __Pyx_Coroutine_AlreadyRunningError(gen);
    if (yf) {
        Py_INCREF(yf);
        err = __Pyx_Coroutine_CloseIter(gen, yf);
        __Pyx_Coroutine_Undelegate(gen);
        Py_DECREF(yf);
    }
    if (err == 0)
        PyErr_SetNone(PyExc_GeneratorExit);
    retval = __Pyx_Coroutine_SendEx(gen, NULL, 1);
    if (unlikely(retval)) {
        const char *msg;
        Py_DECREF(retval);
        if ((0)) {
        #ifdef __Pyx_Coroutine_USED
        } else if (__Pyx_Coroutine_Check(self)) {
            msg = "coroutine ignored GeneratorExit";
        #endif
        #ifdef __Pyx_AsyncGen_USED
        } else if (__Pyx_AsyncGen_CheckExact(self)) {
#if PY_VERSION_HEX < 0x03060000
            msg = "async generator ignored GeneratorExit - might require Python 3.6+ finalisation (PEP 525)";
#else
            msg = "async generator ignored GeneratorExit";
#endif
        #endif
        } else {
            msg = "generator ignored GeneratorExit";
        }
        PyErr_SetString(PyExc_RuntimeError, msg);
        return NULL;
    }
    raised_exception = PyErr_Occurred();
    if (likely(!raised_exception || __Pyx_PyErr_GivenExceptionMatches2(raised_exception, PyExc_GeneratorExit, PyExc_StopIteration))) {
        if (raised_exception) PyErr_Clear();
        Py_INCREF(Py_None);
        return Py_None;
    }
    return NULL;
}
static PyObject *__Pyx__Coroutine_Throw(PyObject *self, PyObject *typ, PyObject *val, PyObject *tb,
                                        PyObject *args, int close_on_genexit) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    PyObject *yf = gen->yieldfrom;
    if (unlikely(gen->is_running))
        return __Pyx_Coroutine_AlreadyRunningError(gen);
    if (yf) {
        PyObject *ret;
        Py_INCREF(yf);
        if (__Pyx_PyErr_GivenExceptionMatches(typ, PyExc_GeneratorExit) && close_on_genexit) {
            int err = __Pyx_Coroutine_CloseIter(gen, yf);
            Py_DECREF(yf);
            __Pyx_Coroutine_Undelegate(gen);
            if (err < 0)
                return __Pyx_Coroutine_MethodReturn(self, __Pyx_Coroutine_SendEx(gen, NULL, 0));
            goto throw_here;
        }
        gen->is_running = 1;
        if (0
        #ifdef __Pyx_Generator_USED
            || __Pyx_Generator_CheckExact(yf)
        #endif
        #ifdef __Pyx_Coroutine_USED
            || __Pyx_Coroutine_Check(yf)
        #endif
            ) {
            ret = __Pyx__Coroutine_Throw(yf, typ, val, tb, args, close_on_genexit);
        #ifdef __Pyx_Coroutine_USED
        } else if (__Pyx_CoroutineAwait_CheckExact(yf)) {
            ret = __Pyx__Coroutine_Throw(((__pyx_CoroutineAwaitObject*)yf)->coroutine, typ, val, tb, args, close_on_genexit);
        #endif
        } else {
            PyObject *meth = __Pyx_PyObject_GetAttrStrNoError(yf, __pyx_n_s_throw);
            if (unlikely(!meth)) {
                Py_DECREF(yf);
                if (unlikely(PyErr_Occurred())) {
                    gen->is_running = 0;
                    return NULL;
                }
                __Pyx_Coroutine_Undelegate(gen);
                gen->is_running = 0;
                goto throw_here;
            }
            if (likely(args)) {
                ret = __Pyx_PyObject_Call(meth, args, NULL);
            } else {
                PyObject *cargs[4] = {NULL, typ, val, tb};
                ret = __Pyx_PyObject_FastCall(meth, cargs+1, 3 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET);
            }
            Py_DECREF(meth);
        }
        gen->is_running = 0;
        Py_DECREF(yf);
        if (!ret) {
            ret = __Pyx_Coroutine_FinishDelegation(gen);
        }
        return __Pyx_Coroutine_MethodReturn(self, ret);
    }
throw_here:
    __Pyx_Raise(typ, val, tb, NULL);
    return __Pyx_Coroutine_MethodReturn(self, __Pyx_Coroutine_SendEx(gen, NULL, 0));
}
static PyObject *__Pyx_Coroutine_Throw(PyObject *self, PyObject *args) {
    PyObject *typ;
    PyObject *val = NULL;
    PyObject *tb = NULL;
    if (unlikely(!PyArg_UnpackTuple(args, (char *)"throw", 1, 3, &typ, &val, &tb)))
        return NULL;
    return __Pyx__Coroutine_Throw(self, typ, val, tb, args, 1);
}
static CYTHON_INLINE int __Pyx_Coroutine_traverse_excstate(__Pyx_ExcInfoStruct *exc_state, visitproc visit, void *arg) {
#if PY_VERSION_HEX >= 0x030B00a4
    Py_VISIT(exc_state->exc_value);
#else
    Py_VISIT(exc_state->exc_type);
    Py_VISIT(exc_state->exc_value);
    Py_VISIT(exc_state->exc_traceback);
#endif
    return 0;
}
static int __Pyx_Coroutine_traverse(__pyx_CoroutineObject *gen, visitproc visit, void *arg) {
    Py_VISIT(gen->closure);
    Py_VISIT(gen->classobj);
    Py_VISIT(gen->yieldfrom);
    return __Pyx_Coroutine_traverse_excstate(&gen->gi_exc_state, visit, arg);
}
static int __Pyx_Coroutine_clear(PyObject *self) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    Py_CLEAR(gen->closure);
    Py_CLEAR(gen->classobj);
    Py_CLEAR(gen->yieldfrom);
    __Pyx_Coroutine_ExceptionClear(&gen->gi_exc_state);
#ifdef __Pyx_AsyncGen_USED
    if (__Pyx_AsyncGen_CheckExact(self)) {
        Py_CLEAR(((__pyx_PyAsyncGenObject*)gen)->ag_finalizer);
    }
#endif
    Py_CLEAR(gen->gi_code);
    Py_CLEAR(gen->gi_frame);
    Py_CLEAR(gen->gi_name);
    Py_CLEAR(gen->gi_qualname);
    Py_CLEAR(gen->gi_modulename);
    return 0;
}
static void __Pyx_Coroutine_dealloc(PyObject *self) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    PyObject_GC_UnTrack(gen);
    if (gen->gi_weakreflist != NULL)
        PyObject_ClearWeakRefs(self);
    if (gen->resume_label >= 0) {
        PyObject_GC_Track(self);
#if PY_VERSION_HEX >= 0x030400a1 && CYTHON_USE_TP_FINALIZE
        if (unlikely(PyObject_CallFinalizerFromDealloc(self)))
#else
        Py_TYPE(gen)->tp_del(self);
        if (unlikely(Py_REFCNT(self) > 0))
#endif
        {
            return;
        }
        PyObject_GC_UnTrack(self);
    }
#ifdef __Pyx_AsyncGen_USED
    if (__Pyx_AsyncGen_CheckExact(self)) {
        /* We have to handle this case for asynchronous generators
           right here, because this code has to be between UNTRACK
           and GC_Del. */
        Py_CLEAR(((__pyx_PyAsyncGenObject*)self)->ag_finalizer);
    }
#endif
    __Pyx_Coroutine_clear(self);
    __Pyx_PyHeapTypeObject_GC_Del(gen);
}
static void __Pyx_Coroutine_del(PyObject *self) {
    PyObject *error_type, *error_value, *error_traceback;
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    __Pyx_PyThreadState_declare
    if (gen->resume_label < 0) {
        return;
    }
#if !CYTHON_USE_TP_FINALIZE
    assert(self->ob_refcnt == 0);
    __Pyx_SET_REFCNT(self, 1);
#endif
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&error_type, &error_value, &error_traceback);
#ifdef __Pyx_AsyncGen_USED
    if (__Pyx_AsyncGen_CheckExact(self)) {
        __pyx_PyAsyncGenObject *agen = (__pyx_PyAsyncGenObject*)self;
        PyObject *finalizer = agen->ag_finalizer;
        if (finalizer && !agen->ag_closed) {
            PyObject *res = __Pyx_PyObject_CallOneArg(finalizer, self);
            if (unlikely(!res)) {
                PyErr_WriteUnraisable(self);
            } else {
                Py_DECREF(res);
            }
            __Pyx_ErrRestore(error_type, error_value, error_traceback);
            return;
        }
    }
#endif
    if (unlikely(gen->resume_label == 0 && !error_value)) {
#ifdef __Pyx_Coroutine_USED
#ifdef __Pyx_Generator_USED
    if (!__Pyx_Generator_CheckExact(self))
#endif
        {
        PyObject_GC_UnTrack(self);
#if PY_MAJOR_VERSION >= 3  || defined(PyErr_WarnFormat)
        if (unlikely(PyErr_WarnFormat(PyExc_RuntimeWarning, 1, "coroutine '%.50S' was never awaited", gen->gi_qualname) < 0))
            PyErr_WriteUnraisable(self);
#else
        {PyObject *msg;
        char *cmsg;
        #if CYTHON_COMPILING_IN_PYPY
        msg = NULL;
        cmsg = (char*) "coroutine was never awaited";
        #else
        char *cname;
        PyObject *qualname;
        qualname = gen->gi_qualname;
        cname = PyString_AS_STRING(qualname);
        msg = PyString_FromFormat("coroutine '%.50s' was never awaited", cname);
        if (unlikely(!msg)) {
            PyErr_Clear();
            cmsg = (char*) "coroutine was never awaited";
        } else {
            cmsg = PyString_AS_STRING(msg);
        }
        #endif
        if (unlikely(PyErr_WarnEx(PyExc_RuntimeWarning, cmsg, 1) < 0))
            PyErr_WriteUnraisable(self);
        Py_XDECREF(msg);}
#endif
        PyObject_GC_Track(self);
        }
#endif
    } else {
        PyObject *res = __Pyx_Coroutine_Close(self);
        if (unlikely(!res)) {
            if (PyErr_Occurred())
                PyErr_WriteUnraisable(self);
        } else {
            Py_DECREF(res);
        }
    }
    __Pyx_ErrRestore(error_type, error_value, error_traceback);
#if !CYTHON_USE_TP_FINALIZE
    assert(Py_REFCNT(self) > 0);
    if (likely(--self->ob_refcnt == 0)) {
        return;
    }
    {
        Py_ssize_t refcnt = Py_REFCNT(self);
        _Py_NewReference(self);
        __Pyx_SET_REFCNT(self, refcnt);
    }
#if CYTHON_COMPILING_IN_CPYTHON
    assert(PyType_IS_GC(Py_TYPE(self)) &&
           _Py_AS_GC(self)->gc.gc_refs != _PyGC_REFS_UNTRACKED);
    _Py_DEC_REFTOTAL;
#endif
#ifdef COUNT_ALLOCS
    --Py_TYPE(self)->tp_frees;
    --Py_TYPE(self)->tp_allocs;
#endif
#endif
}
static PyObject *
__Pyx_Coroutine_get_name(__pyx_CoroutineObject *self, void *context)
{
    PyObject *name = self->gi_name;
    CYTHON_UNUSED_VAR(context);
    if (unlikely(!name)) name = Py_None;
    Py_INCREF(name);
    return name;
}
static int
__Pyx_Coroutine_set_name(__pyx_CoroutineObject *self, PyObject *value, void *context)
{
    CYTHON_UNUSED_VAR(context);
#if PY_MAJOR_VERSION >= 3
    if (unlikely(value == NULL || !PyUnicode_Check(value)))
#else
    if (unlikely(value == NULL || !PyString_Check(value)))
#endif
    {
        PyErr_SetString(PyExc_TypeError,
                        "__name__ must be set to a string object");
        return -1;
    }
    Py_INCREF(value);
    __Pyx_Py_XDECREF_SET(self->gi_name, value);
    return 0;
}
static PyObject *
__Pyx_Coroutine_get_qualname(__pyx_CoroutineObject *self, void *context)
{
    PyObject *name = self->gi_qualname;
    CYTHON_UNUSED_VAR(context);
    if (unlikely(!name)) name = Py_None;
    Py_INCREF(name);
    return name;
}
static int
__Pyx_Coroutine_set_qualname(__pyx_CoroutineObject *self, PyObject *value, void *context)
{
    CYTHON_UNUSED_VAR(context);
#if PY_MAJOR_VERSION >= 3
    if (unlikely(value == NULL || !PyUnicode_Check(value)))
#else
    if (unlikely(value == NULL || !PyString_Check(value)))
#endif
    {
        PyErr_SetString(PyExc_TypeError,
                        "__qualname__ must be set to a string object");
        return -1;
    }
    Py_INCREF(value);
    __Pyx_Py_XDECREF_SET(self->gi_qualname, value);
    return 0;
}
static PyObject *
__Pyx_Coroutine_get_frame(__pyx_CoroutineObject *self, void *context)
{
    PyObject *frame = self->gi_frame;
    CYTHON_UNUSED_VAR(context);
    if (!frame) {
        if (unlikely(!self->gi_code)) {
            Py_RETURN_NONE;
        }
        frame = (PyObject *) PyFrame_New(
            PyThreadState_Get(),            /*PyThreadState *tstate,*/
            (PyCodeObject*) self->gi_code,  /*PyCodeObject *code,*/
            __pyx_d,                 /*PyObject *globals,*/
            0                               /*PyObject *locals*/
        );
        if (unlikely(!frame))
            return NULL;
        self->gi_frame = frame;
    }
    Py_INCREF(frame);
    return frame;
}
static __pyx_CoroutineObject *__Pyx__Coroutine_New(
            PyTypeObject* type, __pyx_coroutine_body_t body, PyObject *code, PyObject *closure,
            PyObject *name, PyObject *qualname, PyObject *module_name) {
    __pyx_CoroutineObject *gen = PyObject_GC_New(__pyx_CoroutineObject, type);
    if (unlikely(!gen))
        return NULL;
    return __Pyx__Coroutine_NewInit(gen, body, code, closure, name, qualname, module_name);
}
static __pyx_CoroutineObject *__Pyx__Coroutine_NewInit(
            __pyx_CoroutineObject *gen, __pyx_coroutine_body_t body, PyObject *code, PyObject *closure,
            PyObject *name, PyObject *qualname, PyObject *module_name) {
    gen->body = body;
    gen->closure = closure;
    Py_XINCREF(closure);
    gen->is_running = 0;
    gen->resume_label = 0;
    gen->classobj = NULL;
    gen->yieldfrom = NULL;
    #if PY_VERSION_HEX >= 0x030B00a4
    gen->gi_exc_state.exc_value = NULL;
    #else
    gen->gi_exc_state.exc_type = NULL;
    gen->gi_exc_state.exc_value = NULL;
    gen->gi_exc_state.exc_traceback = NULL;
    #endif
#if CYTHON_USE_EXC_INFO_STACK
    gen->gi_exc_state.previous_item = NULL;
#endif
    gen->gi_weakreflist = NULL;
    Py_XINCREF(qualname);
    gen->gi_qualname = qualname;
    Py_XINCREF(name);
    gen->gi_name = name;
    Py_XINCREF(module_name);
    gen->gi_modulename = module_name;
    Py_XINCREF(code);
    gen->gi_code = code;
    gen->gi_frame = NULL;
    PyObject_GC_Track(gen);
    return gen;
}

/* PatchModuleWithCoroutine */
static PyObject* __Pyx_Coroutine_patch_module(PyObject* module, const char* py_code) {
#if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
    int result;
    PyObject *globals, *result_obj;
    globals = PyDict_New();  if (unlikely(!globals)) goto ignore;
    result = PyDict_SetItemString(globals, "_cython_coroutine_type",
    #ifdef __Pyx_Coroutine_USED
        (PyObject*)__pyx_CoroutineType);
    #else
        Py_None);
    #endif
    if (unlikely(result < 0)) goto ignore;
    result = PyDict_SetItemString(globals, "_cython_generator_type",
    #ifdef __Pyx_Generator_USED
        (PyObject*)__pyx_GeneratorType);
    #else
        Py_None);
    #endif
    if (unlikely(result < 0)) goto ignore;
    if (unlikely(PyDict_SetItemString(globals, "_module", module) < 0)) goto ignore;
    if (unlikely(PyDict_SetItemString(globals, "__builtins__", __pyx_b) < 0)) goto ignore;
    result_obj = PyRun_String(py_code, Py_file_input, globals, globals);
    if (unlikely(!result_obj)) goto ignore;
    Py_DECREF(result_obj);
    Py_DECREF(globals);
    return module;
ignore:
    Py_XDECREF(globals);
    PyErr_WriteUnraisable(module);
    if (unlikely(PyErr_WarnEx(PyExc_RuntimeWarning, "Cython module failed to patch module with custom type", 1) < 0)) {
        Py_DECREF(module);
        module = NULL;
    }
#else
    py_code++;
#endif
    return module;
}

/* PatchGeneratorABC */
#ifndef CYTHON_REGISTER_ABCS
#define CYTHON_REGISTER_ABCS 1
#endif
#if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
static PyObject* __Pyx_patch_abc_module(PyObject *module);
static PyObject* __Pyx_patch_abc_module(PyObject *module) {
    module = __Pyx_Coroutine_patch_module(
        module, ""
"if _cython_generator_type is not None:\n"
"    try: Generator = _module.Generator\n"
"    except AttributeError: pass\n"
"    else: Generator.register(_cython_generator_type)\n"
"if _cython_coroutine_type is not None:\n"
"    try: Coroutine = _module.Coroutine\n"
"    except AttributeError: pass\n"
"    else: Coroutine.register(_cython_coroutine_type)\n"
    );
    return module;
}
#endif
static int __Pyx_patch_abc(void) {
#if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
    static int abc_patched = 0;
    if (CYTHON_REGISTER_ABCS && !abc_patched) {
        PyObject *module;
        module = PyImport_ImportModule((PY_MAJOR_VERSION >= 3) ? "collections.abc" : "collections");
        if (unlikely(!module)) {
            PyErr_WriteUnraisable(NULL);
            if (unlikely(PyErr_WarnEx(PyExc_RuntimeWarning,
                    ((PY_MAJOR_VERSION >= 3) ?
                        "Cython module failed to register with collections.abc module" :
                        "Cython module failed to register with collections module"), 1) < 0)) {
                return -1;
            }
        } else {
            module = __Pyx_patch_abc_module(module);
            abc_patched = 1;
            if (unlikely(!module))
                return -1;
            Py_DECREF(module);
        }
        module = PyImport_ImportModule("backports_abc");
        if (module) {
            module = __Pyx_patch_abc_module(module);
            Py_XDECREF(module);
        }
        if (!module) {
            PyErr_Clear();
        }
    }
#else
    if ((0)) __Pyx_Coroutine_patch_module(NULL, NULL);
#endif
    return 0;
}

/* Generator */
static PyMethodDef __pyx_Generator_methods[] = {
    {"send", (PyCFunction) __Pyx_Coroutine_Send, METH_O,
     (char*) PyDoc_STR("send(arg) -> send 'arg' into generator,\nreturn next yielded value or raise StopIteration.")},
    {"throw", (PyCFunction) __Pyx_Coroutine_Throw, METH_VARARGS,
     (char*) PyDoc_STR("throw(typ[,val[,tb]]) -> raise exception in generator,\nreturn next yielded value or raise StopIteration.")},
    {"close", (PyCFunction) __Pyx_Coroutine_Close_Method, METH_NOARGS,
     (char*) PyDoc_STR("close() -> raise GeneratorExit inside generator.")},
    {0, 0, 0, 0}
};
static PyMemberDef __pyx_Generator_memberlist[] = {
    {(char *) "gi_running", T_BOOL, offsetof(__pyx_CoroutineObject, is_running), READONLY, NULL},
    {(char*) "gi_yieldfrom", T_OBJECT, offsetof(__pyx_CoroutineObject, yieldfrom), READONLY,
     (char*) PyDoc_STR("object being iterated by 'yield from', or None")},
    {(char*) "gi_code", T_OBJECT, offsetof(__pyx_CoroutineObject, gi_code), READONLY, NULL},
    {(char *) "__module__", T_OBJECT, offsetof(__pyx_CoroutineObject, gi_modulename), 0, 0},
#if CYTHON_USE_TYPE_SPECS
    {(char *) "__weaklistoffset__", T_PYSSIZET, offsetof(__pyx_CoroutineObject, gi_weakreflist), READONLY, 0},
#endif
    {0, 0, 0, 0, 0}
};
static PyGetSetDef __pyx_Generator_getsets[] = {
    {(char *) "__name__", (getter)__Pyx_Coroutine_get_name, (setter)__Pyx_Coroutine_set_name,
     (char*) PyDoc_STR("name of the generator"), 0},
    {(char *) "__qualname__", (getter)__Pyx_Coroutine_get_qualname, (setter)__Pyx_Coroutine_set_qualname,
     (char*) PyDoc_STR("qualified name of the generator"), 0},
    {(char *) "gi_frame", (getter)__Pyx_Coroutine_get_frame, NULL,
     (char*) PyDoc_STR("Frame of the generator"), 0},
    {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_GeneratorType_slots[] = {
    {Py_tp_dealloc, (void *)__Pyx_Coroutine_dealloc},
    {Py_tp_traverse, (void *)__Pyx_Coroutine_traverse},
    {Py_tp_iter, (void *)PyObject_SelfIter},
    {Py_tp_iternext, (void *)__Pyx_Generator_Next},
    {Py_tp_methods, (void *)__pyx_Generator_methods},
    {Py_tp_members, (void *)__pyx_Generator_memberlist},
    {Py_tp_getset, (void *)__pyx_Generator_getsets},
    {Py_tp_getattro, (void *) __Pyx_PyObject_GenericGetAttrNoDict},
#if CYTHON_USE_TP_FINALIZE
    {Py_tp_finalize, (void *)__Pyx_Coroutine_del},
#endif
    {0, 0},
};
static PyType_Spec __pyx_GeneratorType_spec = {
    __PYX_TYPE_MODULE_PREFIX "generator",
    sizeof(__pyx_CoroutineObject),
    0,
    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC | Py_TPFLAGS_HAVE_FINALIZE,
    __pyx_GeneratorType_slots
};
#else
static PyTypeObject __pyx_GeneratorType_type = {
    PyVarObject_HEAD_INIT(0, 0)
    __PYX_TYPE_MODULE_PREFIX "generator",
    sizeof(__pyx_CoroutineObject),
    0,
    (destructor) __Pyx_Coroutine_dealloc,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC | Py_TPFLAGS_HAVE_FINALIZE,
    0,
    (traverseproc) __Pyx_Coroutine_traverse,
    0,
    0,
    offsetof(__pyx_CoroutineObject, gi_weakreflist),
    0,
    (iternextfunc) __Pyx_Generator_Next,
    __pyx_Generator_methods,
    __pyx_Generator_memberlist,
    __pyx_Generator_getsets,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
#if CYTHON_USE_TP_FINALIZE
    0,
#else
    __Pyx_Coroutine_del,
#endif
    0,
#if CYTHON_USE_TP_FINALIZE
    __Pyx_Coroutine_del,
#elif PY_VERSION_HEX >= 0x030400a1
    0,
#endif
#if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
    0,
#endif
#if __PYX_NEED_TP_PRINT_SLOT
    0,
#endif
#if PY_VERSION_HEX >= 0x030C0000
    0,
#endif
#if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
    0,
#endif
};
#endif
static int __pyx_Generator_init(PyObject *module) {
#if CYTHON_USE_TYPE_SPECS
    __pyx_GeneratorType = __Pyx_FetchCommonTypeFromSpec(module, &__pyx_GeneratorType_spec, NULL);
#else
    CYTHON_UNUSED_VAR(module);
    __pyx_GeneratorType_type.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
    __pyx_GeneratorType_type.tp_iter = PyObject_SelfIter;
    __pyx_GeneratorType = __Pyx_FetchCommonType(&__pyx_GeneratorType_type);
#endif
    if (unlikely(!__pyx_GeneratorType)) {
        return -1;
    }
    return 0;
}

/* CStringEquals */
static CYTHON_INLINE int __Pyx_StrEq(const char *s1, const char *s2) {
    while (*s1 != '\0' && *s1 == *s2) { s1++; s2++; }
    return *s1 == *s2;
}

/* CheckBinaryVersion */
static unsigned long __Pyx_get_runtime_version(void) {
#if __PYX_LIMITED_VERSION_HEX >= 0x030B00A4
    return Py_Version & ~0xFFUL;
#else
    const char* rt_version = Py_GetVersion();
    unsigned long version = 0;
    unsigned long factor = 0x01000000UL;
    unsigned int digit = 0;
    int i = 0;
    while (factor) {
        while ('0' <= rt_version[i] && rt_version[i] <= '9') {
            digit = digit * 10 + (unsigned int) (rt_version[i] - '0');
            ++i;
        }
        version += factor * digit;
        if (rt_version[i] != '.')
            break;
        digit = 0;
        factor >>= 8;
        ++i;
    }
    return version;
#endif
}
static int __Pyx_check_binary_version(unsigned long ct_version, unsigned long rt_version, int allow_newer) {
    const unsigned long MAJOR_MINOR = 0xFFFF0000UL;
    if ((rt_version & MAJOR_MINOR) == (ct_version & MAJOR_MINOR))
        return 0;
    if (likely(allow_newer && (rt_version & MAJOR_MINOR) > (ct_version & MAJOR_MINOR)))
        return 1;
    {
        char message[200];
        PyOS_snprintf(message, sizeof(message),
                      "compile time Python version %d.%d "
                      "of module '%.100s' "
                      "%s "
                      "runtime version %d.%d",
                       (int) (ct_version >> 24), (int) ((ct_version >> 16) & 0xFF),
                       __Pyx_MODULE_NAME,
                       (allow_newer) ? "was newer than" : "does not match",
                       (int) (rt_version >> 24), (int) ((rt_version >> 16) & 0xFF)
       );
        return PyErr_WarnEx(NULL, message, 1);
    }
}

/* InitStrings */
#if PY_MAJOR_VERSION >= 3
static int __Pyx_InitString(__Pyx_StringTabEntry t, PyObject **str) {
    if (t.is_unicode | t.is_str) {
        if (t.intern) {
            *str = PyUnicode_InternFromString(t.s);
        } else if (t.encoding) {
            *str = PyUnicode_Decode(t.s, t.n - 1, t.encoding, NULL);
        } else {
            *str = PyUnicode_FromStringAndSize(t.s, t.n - 1);
        }
    } else {
        *str = PyBytes_FromStringAndSize(t.s, t.n - 1);
    }
    if (!*str)
        return -1;
    if (PyObject_Hash(*str) == -1)
        return -1;
    return 0;
}
#endif
static int __Pyx_InitStrings(__Pyx_StringTabEntry *t) {
    while (t->p) {
        #if PY_MAJOR_VERSION >= 3
        __Pyx_InitString(*t, t->p);
        #else
        if (t->is_unicode) {
            *t->p = PyUnicode_DecodeUTF8(t->s, t->n - 1, NULL);
        } else if (t->intern) {
            *t->p = PyString_InternFromString(t->s);
        } else {
            *t->p = PyString_FromStringAndSize(t->s, t->n - 1);
        }
        if (!*t->p)
            return -1;
        if (PyObject_Hash(*t->p) == -1)
            return -1;
        #endif
        ++t;
    }
    return 0;
}

#include <string.h>
static CYTHON_INLINE Py_ssize_t __Pyx_ssize_strlen(const char *s) {
    size_t len = strlen(s);
    if (unlikely(len > (size_t) PY_SSIZE_T_MAX)) {
        PyErr_SetString(PyExc_OverflowError, "byte string is too long");
        return -1;
    }
    return (Py_ssize_t) len;
}
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char* c_str) {
    Py_ssize_t len = __Pyx_ssize_strlen(c_str);
    if (unlikely(len < 0)) return NULL;
    return __Pyx_PyUnicode_FromStringAndSize(c_str, len);
}
static CYTHON_INLINE PyObject* __Pyx_PyByteArray_FromString(const char* c_str) {
    Py_ssize_t len = __Pyx_ssize_strlen(c_str);
    if (unlikely(len < 0)) return NULL;
    return PyByteArray_FromStringAndSize(c_str, len);
}
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject* o) {
    Py_ssize_t ignore;
    return __Pyx_PyObject_AsStringAndSize(o, &ignore);
}
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
#if !CYTHON_PEP393_ENABLED
static const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    char* defenc_c;
    PyObject* defenc = _PyUnicode_AsDefaultEncodedString(o, NULL);
    if (!defenc) return NULL;
    defenc_c = PyBytes_AS_STRING(defenc);
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    {
        char* end = defenc_c + PyBytes_GET_SIZE(defenc);
        char* c;
        for (c = defenc_c; c < end; c++) {
            if ((unsigned char) (*c) >= 128) {
                PyUnicode_AsASCIIString(o);
                return NULL;
            }
        }
    }
#endif
    *length = PyBytes_GET_SIZE(defenc);
    return defenc_c;
}
#else
static CYTHON_INLINE const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    if (unlikely(__Pyx_PyUnicode_READY(o) == -1)) return NULL;
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    if (likely(PyUnicode_IS_ASCII(o))) {
        *length = PyUnicode_GET_LENGTH(o);
        return PyUnicode_AsUTF8(o);
    } else {
        PyUnicode_AsASCIIString(o);
        return NULL;
    }
#else
    return PyUnicode_AsUTF8AndSize(o, length);
#endif
}
#endif
#endif
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
    if (
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
            __Pyx_sys_getdefaultencoding_not_ascii &&
#endif
            PyUnicode_Check(o)) {
        return __Pyx_PyUnicode_AsStringAndSize(o, length);
    } else
#endif
#if (!CYTHON_COMPILING_IN_PYPY && !CYTHON_COMPILING_IN_LIMITED_API) || (defined(PyByteArray_AS_STRING) && defined(PyByteArray_GET_SIZE))
    if (PyByteArray_Check(o)) {
        *length = PyByteArray_GET_SIZE(o);
        return PyByteArray_AS_STRING(o);
    } else
#endif
    {
        char* result;
        int r = PyBytes_AsStringAndSize(o, &result, length);
        if (unlikely(r < 0)) {
            return NULL;
        } else {
            return result;
        }
    }
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject* x) {
   int is_true = x == Py_True;
   if (is_true | (x == Py_False) | (x == Py_None)) return is_true;
   else return PyObject_IsTrue(x);
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject* x) {
    int retval;
    if (unlikely(!x)) return -1;
    retval = __Pyx_PyObject_IsTrue(x);
    Py_DECREF(x);
    return retval;
}
static PyObject* __Pyx_PyNumber_IntOrLongWrongResultType(PyObject* result, const char* type_name) {
    __Pyx_TypeName result_type_name = __Pyx_PyType_GetName(Py_TYPE(result));
#if PY_MAJOR_VERSION >= 3
    if (PyLong_Check(result)) {
        if (PyErr_WarnFormat(PyExc_DeprecationWarning, 1,
                "__int__ returned non-int (type " __Pyx_FMT_TYPENAME ").  "
                "The ability to return an instance of a strict subclass of int is deprecated, "
                "and may be removed in a future version of Python.",
                result_type_name)) {
            __Pyx_DECREF_TypeName(result_type_name);
            Py_DECREF(result);
            return NULL;
        }
        __Pyx_DECREF_TypeName(result_type_name);
        return result;
    }
#endif
    PyErr_Format(PyExc_TypeError,
                 "__%.4s__ returned non-%.4s (type " __Pyx_FMT_TYPENAME ")",
                 type_name, type_name, result_type_name);
    __Pyx_DECREF_TypeName(result_type_name);
    Py_DECREF(result);
    return NULL;
}
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x) {
#if CYTHON_USE_TYPE_SLOTS
  PyNumberMethods *m;
#endif
  const char *name = NULL;
  PyObject *res = NULL;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_Check(x) || PyLong_Check(x)))
#else
  if (likely(PyLong_Check(x)))
#endif
    return __Pyx_NewRef(x);
#if CYTHON_USE_TYPE_SLOTS
  m = Py_TYPE(x)->tp_as_number;
  #if PY_MAJOR_VERSION < 3
  if (m && m->nb_int) {
    name = "int";
    res = m->nb_int(x);
  }
  else if (m && m->nb_long) {
    name = "long";
    res = m->nb_long(x);
  }
  #else
  if (likely(m && m->nb_int)) {
    name = "int";
    res = m->nb_int(x);
  }
  #endif
#else
  if (!PyBytes_CheckExact(x) && !PyUnicode_CheckExact(x)) {
    res = PyNumber_Int(x);
  }
#endif
  if (likely(res)) {
#if PY_MAJOR_VERSION < 3
    if (unlikely(!PyInt_Check(res) && !PyLong_Check(res))) {
#else
    if (unlikely(!PyLong_CheckExact(res))) {
#endif
        return __Pyx_PyNumber_IntOrLongWrongResultType(res, name);
    }
  }
  else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_TypeError,
                    "an integer is required");
  }
  return res;
}
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject* b) {
  Py_ssize_t ival;
  PyObject *x;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_CheckExact(b))) {
    if (sizeof(Py_ssize_t) >= sizeof(long))
        return PyInt_AS_LONG(b);
    else
        return PyInt_AsSsize_t(b);
  }
#endif
  if (likely(PyLong_CheckExact(b))) {
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(__Pyx_PyLong_IsCompact(b))) {
        return __Pyx_PyLong_CompactValue(b);
    } else {
      const digit* digits = __Pyx_PyLong_Digits(b);
      const Py_ssize_t size = __Pyx_PyLong_SignedDigitCount(b);
      switch (size) {
         case 2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
      }
    }
    #endif
    return PyLong_AsSsize_t(b);
  }
  x = PyNumber_Index(b);
  if (!x) return -1;
  ival = PyInt_AsSsize_t(x);
  Py_DECREF(x);
  return ival;
}
static CYTHON_INLINE Py_hash_t __Pyx_PyIndex_AsHash_t(PyObject* o) {
  if (sizeof(Py_hash_t) == sizeof(Py_ssize_t)) {
    return (Py_hash_t) __Pyx_PyIndex_AsSsize_t(o);
#if PY_MAJOR_VERSION < 3
  } else if (likely(PyInt_CheckExact(o))) {
    return PyInt_AS_LONG(o);
#endif
  } else {
    Py_ssize_t ival;
    PyObject *x;
    x = PyNumber_Index(o);
    if (!x) return -1;
    ival = PyInt_AsLong(x);
    Py_DECREF(x);
    return ival;
  }
}
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b) {
  return b ? __Pyx_NewRef(Py_True) : __Pyx_NewRef(Py_False);
}
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t ival) {
    return PyInt_FromSize_t(ival);
}


/* #### Code section: utility_code_pragmas_end ### */
#ifdef _MSC_VER
#pragma warning( pop )
#endif



/* #### Code section: end ### */
#endif /* Py_PYTHON_H */
